/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: console.c,v 1.6 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h> /* for isatty() and getopt() */

#include "console.h"

/* borrowed from ../object.h */
#define SILC_SCALAR_NAME(o) \
 (((o)->precision == SILC_INT) ? "int" : \
  ((o)->precision == SILC_LONG) ? "long" : \
  ((o)->precision == SILC_FLOAT) ? "float" : \
  ((o)->precision == SILC_DOUBLE) ? "double" : \
  ((o)->precision == SILC_COMPLEX) ? "complex" : \
  ((o)->precision == SILC_DOUBLE_COMPLEX) ? "double complex" : "unknown")

#include "client.h"

extern int silc_debug; /* client.c */

#include <sys/types.h> /* for int32_t and int64_t */
#ifdef _sx
typedef long int64_t;
#else
#ifdef _WIN32
typedef int int32_t;
typedef long long int64_t;
#endif /* _WIN32 */
#endif /* _sx */

#include "mm_read.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

int
fetch_object(char *ss_expr, silc_envelope_t *object)
{
  char buf[1024];

  sprintf(buf, "_=%s", ss_expr);
  if (SILC_EXEC(buf) < 0) {
    fprintf(stderr, "Error: SILC_EXEC failed\n");
    return -1;
  }
  object->v = NULL; /* to automatically allocate memory */
  if (SILC_GET(object, "_") < 0) {
    fprintf(stderr, "Error: SILC_GET failed\n");
    return -1;
  }
  return 0;
}

int
free_object(silc_envelope_t *object)
{
  free(object->v);
  if (object->type == SILC_MATRIX_TYPE &&
      strcmp(object->format, SILC_FORMAT_CRS) == 0) {
    free(object->row);
    free(object->index);
  }
  return 0;
}

int
compare_objects(cond_op_t op, silc_envelope_t *lhs, silc_envelope_t *rhs)
{
  if (lhs->type != SILC_SCALAR_TYPE ||
      rhs->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "Error: comparisons must be of scalar values\n");
    return -1;
  }
#define _cmp_1(t1, t2) \
  switch (op) {								\
  case EqualOp:        return *((t1 *)lhs->v) == *((t2 *)rhs->v);	\
  case NotEqualOp:     return *((t1 *)lhs->v) != *((t2 *)rhs->v);	\
  case LessOp:         return *((t1 *)lhs->v) <  *((t2 *)rhs->v);	\
  case LessEqualOp:    return *((t1 *)lhs->v) <= *((t2 *)rhs->v);	\
  case GreaterOp:      return *((t1 *)lhs->v) >  *((t2 *)rhs->v);	\
  case GreaterEqualOp: return *((t1 *)lhs->v) >= *((t2 *)rhs->v);	\
  }
#define _cmp_2(t1, t2) \
  switch (op) {								\
  case EqualOp:    return ((t1 *)lhs->v)[0] == ((t2 *)rhs->v)[0] &&	\
	                  ((t1 *)lhs->v)[1] == ((t2 *)rhs->v)[1];	\
  case NotEqualOp: return ((t1 *)lhs->v)[0] != ((t2 *)rhs->v)[0] ||	\
	                  ((t1 *)lhs->v)[1] != ((t2 *)rhs->v)[1];	\
  }
  switch (lhs->precision) {
  case SILC_INT:
    switch (rhs->precision) {
    case SILC_INT:    _cmp_1(int32_t, int32_t); break;
    case SILC_LONG:   _cmp_1(int32_t, int64_t); break;
    case SILC_FLOAT:  _cmp_1(int32_t, float);   break;
    case SILC_DOUBLE: _cmp_1(int32_t, double);  break;
    }
    break;
  case SILC_LONG:
    switch (rhs->precision) {
    case SILC_INT:    _cmp_1(int64_t, int32_t); break;
    case SILC_LONG:   _cmp_1(int64_t, int64_t); break;
    case SILC_FLOAT:  _cmp_1(int64_t, float);   break;
    case SILC_DOUBLE: _cmp_1(int64_t, double);  break;
    }
    break;
  case SILC_FLOAT:
    switch (rhs->precision) {
    case SILC_INT:    _cmp_1(float, int32_t); break;
    case SILC_LONG:   _cmp_1(float, int64_t); break;
    case SILC_FLOAT:  _cmp_1(float, float);   break;
    case SILC_DOUBLE: _cmp_1(float, double);  break;
    }
    break;
  case SILC_DOUBLE:
    switch (rhs->precision) {
    case SILC_INT:    _cmp_1(double, int32_t); break;
    case SILC_LONG:   _cmp_1(double, int64_t); break;
    case SILC_FLOAT:  _cmp_1(double, float);   break;
    case SILC_DOUBLE: _cmp_1(double, double);  break;
    }
    break;
  case SILC_COMPLEX:
    switch (rhs->precision) {
    case SILC_COMPLEX:        _cmp_2(float, float);  break;
    case SILC_DOUBLE_COMPLEX: _cmp_2(float, double); break;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    switch (rhs->precision) {
    case SILC_COMPLEX:        _cmp_2(double, float);  break;
    case SILC_DOUBLE_COMPLEX: _cmp_2(double, double); break;
    }
    break;
  }
  fprintf(stderr, "Error: invalid comparisons\n");
  return -1;
}

int
eval_expression(expr_t *expr)
{
  int result;
  silc_envelope_t lhs, rhs;

  switch (expr->type) {
  case BoolExprType:
    return BOOL(expr)->bool;
  case UnaryCondExprType:
    switch (UNARY_COND(expr)->op) {
    case NotOp:
      result = eval_expression(UNARY_COND(expr)->operand);
      if (result < 0)
	return -1;
      return !result;
    }
    break;
  case BinaryCondExprType:
    switch (BINARY_COND(expr)->op) {
    case AndOp:
      result = eval_expression(BINARY_COND(expr)->lhs);
      if (result < 0)
	return -1;
      if (!result)
	return 0;
      result = eval_expression(BINARY_COND(expr)->rhs);
      if (result < 0)
	return -1;
      if (!result)
	return 0;
      return 1;
    case OrOp:
      result = eval_expression(BINARY_COND(expr)->lhs);
      if (result < 0)
	return -1;
      if (result)
	return 1;
      result = eval_expression(BINARY_COND(expr)->rhs);
      if (result < 0)
	return -1;
      if (result)
	return 1;
      return 0;
    case EqualOp:
    case NotEqualOp:
    case LessOp:
    case LessEqualOp:
    case GreaterOp:
    case GreaterEqualOp:
      if (fetch_object(SS_EXPR(BINARY_COND(expr)->lhs)->expr, &lhs) < 0)
	return -1;
      if (fetch_object(SS_EXPR(BINARY_COND(expr)->rhs)->expr, &rhs) < 0) {
	free_object(&lhs);
	return -1;
      }
      result = compare_objects(BINARY_COND(expr)->op, &lhs, &rhs);
      free_object(&lhs);
      free_object(&rhs);
      return result;
    }
    break;
  }
  should_not_reach_here;
  return -1;
}

int pprint_limit = 100;

int
pprint_object(silc_envelope_t *object)
{
  int i, j, k, n;

  switch (object->type) {
  case SILC_SCALAR_TYPE:
    switch (object->precision) {
    case SILC_INT:
      printf("int: %d\n", *((int32_t *)object->v));
      break;
    case SILC_LONG:
      printf("long: %lld\n", *((int64_t *)object->v));
      break;
    case SILC_FLOAT:
      printf("float: %e\n", *((float *)object->v));
      break;
    case SILC_DOUBLE:
      printf("double: %e\n", *((double *)object->v));
      break;
    case SILC_COMPLEX:
      printf("complex: %e%+ei\n",
	     *((float *)object->v),
	     *((float *)object->v + 1));
      break;
    case SILC_DOUBLE_COMPLEX:
      printf("double complex: %e%+ei\n",
	     *((double *)object->v),
	     *((double *)object->v + 1));
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_ROW_VECTOR_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
    printf("%s vector, %ld elements of %s\n",
	   object->type == SILC_ROW_VECTOR_TYPE ? "row" : "column",
	   object->length,
	   SILC_SCALAR_NAME(object));
    if (object->length > pprint_limit)
      break;
    for (i = 0; i < object->length; i++) {
      printf("  ");
      switch (object->precision) {
      case SILC_INT:
	printf("[%d] = %d\n", i+1, *((int32_t *)object->v +i)); break;
      case SILC_LONG:
	printf("[%d] = %lld\n", i+1, *((int64_t *)object->v + i)); break;
      case SILC_FLOAT:
	printf("[%d] = %e\n", i+1, *((float *)object->v + i)); break;
      case SILC_DOUBLE:
	printf("[%d] = %e\n", i+1, *((double *)object->v + i)); break;
      case SILC_COMPLEX:
	printf("[%d] = %e%+ei\n", i+1,
	       *((float *)object->v + i * 2),
	       *((float *)object->v + i * 2 + 1)); break;
      case SILC_DOUBLE_COMPLEX:
	printf("[%d] = %e%+ei\n", i+1,
	       *((double *)object->v + i * 2),
	       *((double *)object->v + i * 2 + 1)); break;
      default:
	should_not_reach_here;
      }
    }
    break;
  case SILC_MATRIX_TYPE:
    printf("%ld x %ld matrix of %s in \"%s\" format\n",
	   object->m,
	   object->n,
	   SILC_SCALAR_NAME(object),
	   object->format);
    if (strcmp(object->format, SILC_FORMAT_DENSE) == 0) {
      if (object->m * object->n > pprint_limit)
	break;
      n = 0;
      for (j = 0; j < object->n; j++) {
	for (i = 0; i < object->m; i++) {
	  printf("  ");
	  switch (object->precision) {
	  case SILC_INT:
	    printf("[%d,%d] = %d\n", i+1, j+1,
		   *((int32_t *)object->v + n)); break;
	  case SILC_LONG:
	    printf("[%d,%d] = %lld\n", i+1, j+1,
		   *((int64_t *)object->v + n)); break;
	  case SILC_FLOAT:
	    printf("[%d,%d] = %e\n", i+1, j+1,
		   *((float *)object->v + n)); break;
	  case SILC_DOUBLE:
	    printf("[%d,%d] = %e\n", i+1, j+1,
		   *((double *)object->v + n)); break;
	  case SILC_COMPLEX:
	    printf("[%d,%d] = %e%+ei\n", i+1, j+1,
		   *((float *)object->v + n * 2),
		   *((float *)object->v + n * 2 + 1)); break;
	  case SILC_DOUBLE_COMPLEX:
	    printf("[%d,%d] = %e%+ei\n", i+1, j+1,
		   *((double *)object->v + n * 2),
		   *((double *)object->v + n * 2 + 1)); break;
	  default:
	    should_not_reach_here;
	  }
	  n++;
	}
      }
    } else if (strcmp(object->format, SILC_FORMAT_CRS) == 0) {
      if (object->nnz > pprint_limit)
	break;
      for (i = 0; i < object->m; i++) {
	for (j = object->row[i]; j < object->row[i+1]; j++) {
	  printf("  ");
	  switch (object->precision) {
	  case SILC_INT:
	    printf("[%d,%d] = %d\n", i+1, object->index[j]+1,
		   *((int32_t *)object->v + j)); break;
	  case SILC_LONG:
	    printf("[%d,%d] = %lld\n", i+1, object->index[j]+1,
		   *((int64_t *)object->v + j)); break;
	  case SILC_FLOAT:
	    printf("[%d,%d] = %e\n", i+1, object->index[j]+1,
		   *((float *)object->v + j)); break;
	  case SILC_DOUBLE:
	    printf("[%d,%d] = %e\n", i+1, object->index[j]+1,
		   *((double *)object->v + j)); break;
	  case SILC_COMPLEX:
	    printf("[%d,%d] = %e%+ei\n", i+1, object->index[j]+1,
		   *((float *)object->v + j * 2),
		   *((float *)object->v + j * 2 + 1));
	    break;
	  case SILC_DOUBLE_COMPLEX:
	    printf("[%d,%d] = %e%+ei\n", i+1, object->index[j]+1,
		   *((double *)object->v + j * 2),
		   *((double *)object->v + j * 2 + 1));
	    break;
	  default:
	    should_not_reach_here;
	  }
	}
      }
    } else {
      printf("Error: unknown matrix storage format: %s\n", object->format);
    }
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    printf("cubic_array, %ld x %ld x %ld elements of %s\n",
	   object->l,
	   object->m,
	   object->n,
	   SILC_SCALAR_NAME(object));
    if (object->l * object->m * object->n > pprint_limit)
      break;
    n = 0;
    for (i = 0; i < object->l; i++) {
      for (k = 0; k < object->n; k++) {
	for (j = 0; j < object->m; j++) {
	  printf("  ");
	  switch (object->precision) {
	  case SILC_INT:
	    printf("[%d,%d,%d] = %d\n", i+1, j+1, k+1,
		   *((int32_t *)object->v + n));
	    n++;
	    break;
	  case SILC_LONG:
	    printf("[%d,%d,%d] = %lld\n", i+1, j+1, k+1,
		   *((int64_t *)object->v + n));
	    n++;
	    break;
	  case SILC_FLOAT:
	    printf("[%d,%d,%d] = %e\n", i+1, j+1, k+1,
		   *((float *)object->v + n));
	    n++;
	    break;
	  case SILC_DOUBLE:
	    printf("[%d,%d,%d] = %e\n", i+1, j+1, k+1,
		   *((double *)object->v + n));
	    n++;
	    break;
	  case SILC_COMPLEX:
	    printf("[%d,%d,%d] = %e%+ei\n", i+1, j+1, k+1,
		   *((float *)object->v + n),
		   *((float *)object->v + n + 1));
	    n += 2;
	    break;
	  case SILC_DOUBLE_COMPLEX:
	    printf("[%d,%d,%d] = %e%+ei\n", i+1, j+1, k+1,
		   *((double *)object->v + n),
		   *((double *)object->v + n + 1));
	    n += 2;
	    break;
	  default:
	    should_not_reach_here;
	  }
	}
      }
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return 0;
}

int
load(char *path, int n_names, char *name[])
{
  FILE *fp;
  silc_envelope_t object;
  int len, result;

  len = strlen(path);
  if (len > 4 && strcmp(path + len - 4, ".mtx") == 0) {
    mm_header header;
    double *value;
    int *row, *index;

    if (n_names != 1) {
      fprintf(stderr, "Error: unexpected number of names\n");
      return -1;
    } 
    if ((fp = fopen(path, "rb")) == NULL) {
      fprintf(stderr, "Error: cannot open file: %s\n", path);
      return -1;
    }
    if (mm_read_header(fp, &header) < 0)
      return -1;
    if (header.format == mm_array) {
      if (mm_read_array(fp, &header, &value) < 0)
	return -1;
      fclose(fp);
      printf("%s: %d x %d matrix\n", path, header.m, header.n);
      object.v = value;
      object.precision = SILC_DOUBLE;
      if (header.m > 1 && header.n > 1) {
	object.type = SILC_MATRIX_TYPE;
	object.format = SILC_FORMAT_DENSE;
	object.m = header.m;
	object.n = header.n;
      } else if (header.m > 1) {
	object.type = SILC_COLUMN_VECTOR_TYPE;
	object.length = header.m;
      } else if (header.n > 1) {
	object.type = SILC_ROW_VECTOR_TYPE;
	object.length = header.n;
      } else {
	should_not_reach_here;
      }
      result = SILC_PUT(name[0], &object);
      free(value);
    } else if (header.format == mm_coordinate) {
      if (mm_read_crs(fp, &header, &value, &row, &index) < 0)
	return -1;
      fclose(fp);
      printf("%s: %d x %d matrix (%d elements)\n",
	     path, header.m, header.n, row[header.m]);
      object.v = value;
      object.type = SILC_MATRIX_TYPE;
      object.format = SILC_FORMAT_CRS;
      object.precision = SILC_DOUBLE;
      object.m = header.m;
      object.n = header.n;
      object.nnz = row[header.m];
      object.row = row;
      object.index = index;
      result = SILC_PUT(name[0], &object);
      free(value);
      free(row);
      free(index);
    }
    if (result < 0) {
      fprintf(stderr, "Error: SILC_PUT failed\n");
      return -1;
    }
  } else {
    fprintf(stderr, "Error: unsupported file format\n");
    return -1;
  }
  return 0;
}

int
save(char *path, int n_names, char *name[])
{
  FILE *fp;
  silc_envelope_t object;
  int i, j, len;

  len = strlen(path);
  if (len > 4 && strcmp(path + len - 4, ".mtx") == 0) {
    if (n_names != 1) {
      fprintf(stderr, "Error: unexpected number of names\n");
      return -1;
    } 
    object.v = NULL;
    if (SILC_GET(&object, name[0]) < 0) {
      fprintf(stderr, "Error: SILC_GET failed\n");
      return -1;
    }
    if ((fp = fopen(path, "wb")) == NULL) {
      fprintf(stderr, "Error: cannot open file: %s\n", path);
      return -1;
    }
    if (object.type == SILC_COLUMN_VECTOR_TYPE) {
      switch (object.precision) {
      case SILC_INT:
	fprintf(fp, "%%%%MatrixMarket matrix array integer general\n");
	fprintf(fp, "%ld 1\n", object.length);
	for (i = 0; i < object.length; i++)
	  fprintf(fp, "%d\n", ((int32_t *)object.v)[i]);
	break;
      case SILC_LONG:
	fprintf(fp, "%%%%MatrixMarket matrix array integer general\n");
	fprintf(fp, "%ld 1\n", object.length);
	for (i = 0; i < object.length; i++)
	  fprintf(fp, "%ld\n", ((int64_t *)object.v)[i]);
	break;
      case SILC_FLOAT:
	fprintf(fp, "%%%%MatrixMarket matrix array real general\n");
	fprintf(fp, "%ld 1\n", object.length);
	for (i = 0; i < object.length; i++)
	  fprintf(fp, "%e\n", ((float *)object.v)[i]);
	break;
      case SILC_DOUBLE:
	fprintf(fp, "%%%%MatrixMarket matrix array real general\n");
	fprintf(fp, "%ld 1\n", object.length);
	for (i = 0; i < object.length; i++)
	  fprintf(fp, "%e\n", ((double *)object.v)[i]);
	break;
      case SILC_COMPLEX:
	fprintf(fp, "%%%%MatrixMarket matrix array complex general\n");
	fprintf(fp, "%ld 1\n", object.length);
	for (i = 0; i < object.length; i++)
	  fprintf(fp, "%e %e\n",
		  ((float *)object.v)[i*2],
		  ((float *)object.v)[i*2+1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	fprintf(fp, "%%%%MatrixMarket matrix array complex general\n");
	fprintf(fp, "%ld 1\n", object.length);
	for (i = 0; i < object.length; i++)
	  fprintf(fp, "%e %e\n",
		  ((double *)object.v)[i*2],
		  ((double *)object.v)[i*2+1]);
	break;
      default:
	should_not_reach_here;
      }
      fclose(fp);
      free(object.v);
    } else if (object.type == SILC_MATRIX_TYPE &&
	       strcmp(object.format, SILC_FORMAT_DENSE) == 0) {
      switch (object.precision) {
      case SILC_INT:
	fprintf(fp, "%%%%MatrixMarket matrix array integer general\n");
	fprintf(fp, "%ld %ld\n", object.m, object.n);
	for (j = 0; j < object.n; j++)
	  for (i = 0; i < object.m; i++)
	    fprintf(fp, "%d\n", ((int32_t *)object.v)[i + object.m * j]);
	break;
      case SILC_LONG:
	fprintf(fp, "%%%%MatrixMarket matrix array integer general\n");
	fprintf(fp, "%ld %ld\n", object.m, object.n);
	for (j = 0; j < object.n; j++)
	  for (i = 0; i < object.m; i++)
	    fprintf(fp, "%ld\n", ((int64_t *)object.v)[i + object.m * j]);
	break;
      case SILC_FLOAT:
	fprintf(fp, "%%%%MatrixMarket matrix array real general\n");
	fprintf(fp, "%ld %ld\n", object.m, object.n);
	for (j = 0; j < object.n; j++)
	  for (i = 0; i < object.m; i++)
	    fprintf(fp, "%e\n", ((float *)object.v)[i + object.m * j]);
	break;
      case SILC_DOUBLE:
	fprintf(fp, "%%%%MatrixMarket matrix array real general\n");
	fprintf(fp, "%ld %ld\n", object.m, object.n);
	for (j = 0; j < object.n; j++)
	  for (i = 0; i < object.m; i++)
	    fprintf(fp, "%e\n", ((double *)object.v)[i + object.m * j]);
	break;
      case SILC_COMPLEX:
	fprintf(fp, "%%%%MatrixMarket matrix array complex general\n");
	fprintf(fp, "%ld %ld\n", object.m, object.n);
	for (j = 0; j < object.n; j++)
	  for (i = 0; i < object.m; i++)
	    fprintf(fp, "%e %e\n",
		    ((float *)object.v)[(i + object.m * j) * 2],
		    ((float *)object.v)[(i + object.m * j) * 2 + 1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	fprintf(fp, "%%%%MatrixMarket matrix array complex general\n");
	fprintf(fp, "%ld %ld\n", object.m, object.n);
	for (j = 0; j < object.n; j++)
	  for (i = 0; i < object.m; i++)
	    fprintf(fp, "%e %e\n",
		    ((double *)object.v)[(i + object.m * j) * 2],
		    ((double *)object.v)[(i + object.m * j) * 2 + 1]);
	break;
      default:
	should_not_reach_here;
      }
      fclose(fp);
      free(object.v);
    } else if (object.type == SILC_MATRIX_TYPE &&
	       strcmp(object.format, SILC_FORMAT_CRS) == 0) {
      switch (object.precision) {
      case SILC_INT:
	fprintf(fp, "%%%%MatrixMarket matrix coordinate integer general\n");
	fprintf(fp, "%ld %ld %ld\n", object.m, object.n, object.nnz);
	for (i = 0; i < object.m; i++)
	  for (j = object.row[i]; j < object.row[i+1]; j++)
	    fprintf(fp, "%d %d %d\n", i+1, object.index[j]+1,
		    ((int32_t *)object.v)[j]);
	break;
      case SILC_LONG:
	fprintf(fp, "%%%%MatrixMarket matrix coordinate integer general\n");
	fprintf(fp, "%ld %ld %ld\n", object.m, object.n, object.nnz);
	for (i = 0; i < object.m; i++)
	  for (j = object.row[i]; j < object.row[i+1]; j++)
	    fprintf(fp, "%d %d %ld\n", i+1, object.index[j]+1,
		    ((int64_t *)object.v)[j]);
	break;
      case SILC_FLOAT:
	fprintf(fp, "%%%%MatrixMarket matrix coordinate real general\n");
	fprintf(fp, "%ld %ld %ld\n", object.m, object.n, object.nnz);
	for (i = 0; i < object.m; i++)
	  for (j = object.row[i]; j < object.row[i+1]; j++)
	    fprintf(fp, "%d %d %e\n", i+1, object.index[j]+1,
		    ((float *)object.v)[j]);
	break;
      case SILC_DOUBLE:
	fprintf(fp, "%%%%MatrixMarket matrix coordinate real general\n");
	fprintf(fp, "%ld %ld %ld\n", object.m, object.n, object.nnz);
	for (i = 0; i < object.m; i++)
	  for (j = object.row[i]; j < object.row[i+1]; j++)
	    fprintf(fp, "%d %d %e\n", i+1, object.index[j]+1,
		    ((double *)object.v)[j]);
	break;
      case SILC_COMPLEX:
	fprintf(fp, "%%%%MatrixMarket matrix coordinate complex general\n");
	fprintf(fp, "%ld %ld %ld\n", object.m, object.n, object.nnz);
	for (i = 0; i < object.m; i++)
	  for (j = object.row[i]; j < object.row[i+1]; j++)
	    fprintf(fp, "%d %d %e %e\n", i+1, object.index[j]+1,
		    ((float *)object.v)[j*2],
		    ((float *)object.v)[j*2+1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	fprintf(fp, "%%%%MatrixMarket matrix coordinate complex general\n");
	fprintf(fp, "%ld %ld %ld\n", object.m, object.n, object.nnz);
	for (i = 0; i < object.m; i++)
	  for (j = object.row[i]; j < object.row[i+1]; j++)
	    fprintf(fp, "%d %d %e %e\n", i+1, object.index[j]+1,
		    ((double *)object.v)[j*2],
		    ((double *)object.v)[j*2+1]);
	break;
      default:
	should_not_reach_here;
      }
      fclose(fp);
      free(object.v);
      free(object.row);
      free(object.index);
    } else {
      fprintf(stderr, "Error: unsupported data type\n");
      fclose(fp);
      free(object.v);
      return -1;
    }
  } else {
    fprintf(stderr, "Error: unsupported file format\n");
    return -1;
  }
  return 0;
}

int
pprint(char *expr)
{
  silc_envelope_t object;

  if (fetch_object(expr, &object) < 0)
    return -1;
  if (pprint_object(&object) < 0)
    return -1;
  free_object(&object);
  return 0;
}

int within_while;

#define CONTINUE_WHILE  100
#define BREAK_WHILE     200

int
exec_statement(stmt_t *stmt)
{
  int i, j, len, result;
  char buf[1024];
  block_stmt_t *block;

  switch(stmt->type) {
  case BlockStmtType:
    i = 0;
    while (i < BLOCK(stmt)->n_stmts) {
      if (BLOCK(stmt)->stmt[i]->type != ServerSideStmtType) {
	result = exec_statement(BLOCK(stmt)->stmt[i]);
	if (result < 0)
	  return -1;
	if (result == CONTINUE_WHILE || result == BREAK_WHILE)
	  return result;
	i++;
	continue;
      }
      len = strlen(SS(BLOCK(stmt)->stmt[i])->stmt);
      if (len == 0) {
	i++;
	continue;
      }
      strcpy(buf, SS(BLOCK(stmt)->stmt[i])->stmt);
      i++;
      while (i < BLOCK(stmt)->n_stmts &&
	     BLOCK(stmt)->stmt[i]->type == ServerSideStmtType) {
	len = strlen(SS(BLOCK(stmt)->stmt[i])->stmt);
	if (len == 0) {
	  i++;
	  continue;
	}
	if (!(strlen(buf) + 1 + len < sizeof(buf)))
	  break;
	strcat(buf, ";");
	strcat(buf, SS(BLOCK(stmt)->stmt[i])->stmt);
	i++;
      }
      if (SILC_EXEC(buf) < 0) {
	fprintf(stderr, "Error: SILC_EXEC failed\n");
	return -1;
      }
    }
    break;
  case IfStmtType:
    result = eval_expression(IF(stmt)->cond_expr);
    if (result < 0)
      return -1;
    if (result)
      return exec_statement(IF(stmt)->then_block);
    else if (IF(stmt)->else_block)
      return exec_statement(IF(stmt)->else_block);
    break;
  case WhileStmtType:
    within_while++;
    while (1) {
      result = eval_expression(WHILE(stmt)->cond_expr);
      if (result < 0)
	return -1;
      if (!result)
	break;
      result = exec_statement(WHILE(stmt)->body);
      if (result < 0)
	return -1;
      if (result == CONTINUE_WHILE)
	continue;
      if (result == BREAK_WHILE)
	break;
    }
    within_while--;
    break;
  case ContinueStmtType:
    if (within_while < 1)
      return -1;
    return CONTINUE_WHILE;
  case BreakStmtType:
    if (within_while < 1)
      return -1;
    return BREAK_WHILE;
  case LoadStmtType:
    if (load(LOAD(stmt)->path, LOAD(stmt)->n_names, LOAD(stmt)->name) < 0)
      return -1;
    break;
  case SaveStmtType:
    if (save(SAVE(stmt)->path, SAVE(stmt)->n_names, SAVE(stmt)->name) < 0)
      return -1;
    break;
  case MessageStmtType:
    printf("*** %s\n", MESSAGE(stmt)->msg);
    break;
  case PPrintStmtType:
    pprint(PPRINT(stmt)->expr);
    break;
  }
  return 0;
}

int
run_string(char *buf)
{
  stmt_t *block;
  int result;

  block = parse_script(buf);
  if (!block) {
    fprintf(stderr, "Error: %s in line %d\n", parser_error, parser_lineno);
    return -1;
  }
#ifdef DEBUG
  if (silc_debug)
    dump_statement(block, 0);
#endif
  within_while = 0;
  result = exec_statement(block);
  free_statement(block);
  return result;
}

int
run_script(char *path)
{
  FILE *fp;
  char *buf, *p;
  size_t buf_size, n;
  int result;

  fp = fopen(path, "r");
  if (!fp) {
    fprintf(stderr, "Error: cannot open file: %s\n", path);
    return -1;
  }
  n = 0;
  buf_size = 1024;
  buf = (char *)malloc(buf_size);
  if (!buf) {
    fprintf(stderr, "Error: out of memory\n");
    return -1;
  }
  while (1) {
    p = fgets(buf + n, buf_size - n, fp);
    if (!p)
      break;
    n += strlen(p);
    if (buf_size - n < 1024) {
      buf_size += 1024;
      buf = (char *)realloc(buf, buf_size);
      if (!buf) {
	fprintf(stderr, "Error: out of memory\n");
	return -1;
      }
    }
  }
  fclose(fp);
  SILC_INIT();
  result = run_string(buf);
  SILC_FINALIZE();
  free(buf);
  return result;
}

int
interact(void)
{
  char *buf;
  size_t buf_size, n;
  int result;

  SILC_INIT();
#ifdef _WIN32
  printf("SILC console (Ctrl-Z to exit)\n");
#else
  printf("SILC console (Ctrl-D to exit)\n");
#endif
  n = 0;
  buf_size = 1024;
  buf = (char *)malloc(buf_size);
  if (!buf) {
    fprintf(stderr, "Error: out of memory\n");
    return -1;
  }
  while (1) {
    if (buf_size - n < 1024) {
      buf_size += 1024;
      buf = (char *)realloc(buf, buf_size);
      if (!buf) {
	fprintf(stderr, "Error: out of memory\n");
	return -1;
      }
    }
    if (isatty(0))
      printf("%s ", (n == 0) ? ">" : " ");
    if (fgets(buf + n, buf_size - n, stdin) == NULL)
      break;
    result = is_partial(buf);
    if (result < 0) {
      fprintf(stderr, "Error: %s\n", parser_error);
      n = 0;
    } else if (!result) {
      run_string(buf);
      n = 0;
    } else {
      n = strlen(buf);
    }
  }
  SILC_FINALIZE();
#ifndef _WIN32
  printf("\n");
#endif
  return 0;
}

void
usage(char *name)
{
  printf("Usage: %s [options] [script_file]\n", name);
  printf("Options:\n");
  printf("  -p NUM  pprint data size threshold (default: %d)\n", pprint_limit);
  printf("  -i      enter the interactive mode after running script\n");
  printf("  -v      verbose mode\n");
  printf("  -h      show this message\n");
}

int
main(int argc, char *argv[])
{
  int opt, interactive, verbose;

  interactive = verbose = 0;
  while ((opt = getopt(argc, argv, "p:ivh")) != EOF) {
    switch (opt) {
    case 'p':
      pprint_limit = atoi(optarg);
      break;
    case 'i':
      interactive = 1;
      break;
    case 'v':
      verbose = 1;
      break;
    case 'h':
    default:
      usage(argv[0]);
      return -1;
    }
  }
  if (optind+1 < argc) {
    usage(argv[0]);
    return -1;
  }
#ifdef DEBUG
  if (!verbose)
    silc_debug = 0;
#endif
  if (optind == argc-1) {
    if (run_script(argv[optind]) < 0)
      return -1;
  }
  if (optind == argc || interactive) {
    if (interact() < 0)
      return -1;
  }
  return 0;
}
