%{
/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: console_parser.y,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "console.h"

const char *parser_error;

void yyerror(const char *s)
{
  parser_error = s;
}

int yywrap(void)
{
  return 1;
} 

static stmt_t *last;

%}

%union 
{
  stmt_t *stmt;
  expr_t *expr;
  cond_op_t cond_op;
  char string[1024];
  struct {
    char **v;
    int n;
  } argument_list;
}

%token <string> TOK_IDENTIFIER
%token <string> TOK_STRING_LITERAL
%token <string> TOK_INT4_LITERAL
%token <string> TOK_INT8_LITERAL
%token <string> TOK_REAL4_LITERAL
%token <string> TOK_REAL8_LITERAL

%token TOK_EQUAL
%token TOK_PLUS
%token TOK_PLUS_EQUAL
%token TOK_MINUS
%token TOK_MINUS_EQUAL
%token TOK_MULTIPLY
%token TOK_MULTIPLY_EQUAL
%token TOK_DIVIDE
%token TOK_DIVIDE_EQUAL
%token TOK_REMAINDER
%token TOK_REMAINDER_EQUAL
%token TOK_LEQ
%token TOK_CONJUGATE_TRANSPOSE
%token TOK_COMPLEX_CONJUGATE
%token TOK_TRANSPOSE
%token TOK_MULTIPLY_ELEMENTWISE
%token TOK_MULTIPLY_ELEMENTWISE_EQUAL
%token TOK_DIVIDE_ELEMENTWISE
%token TOK_DIVIDE_ELEMENTWISE_EQUAL
%token TOK_DOUBLE_SEMICOLON
%token TOK_LESS_EQUAL
%token TOK_LESS
%token TOK_GREATER_EQUAL
%token TOK_GREATER
%token TOK_NOT_EQUAL
%token TOK_DOUBLE_EQUAL
%token TOK_IF
%token TOK_ELSE
%token TOK_WHILE
%token TOK_BREAK
%token TOK_CONTINUE
%token TOK_AND
%token TOK_OR
%token TOK_NOT
%token TOK_TRUE
%token TOK_FALSE
%token TOK_LOAD
%token TOK_SAVE
%token TOK_MESSAGE
%token TOK_PPRINT
%token TOK_PREFER
%token TOK_EOL

%type <stmt> statement statement_list if_statement else_block
%type <stmt> while_statement break_statement continue_statement
%type <stmt> load_statement save_statement message_statement pprint_statement
%type <stmt> procedure_call prefer_statement assignment
%type <expr> cond_expr and_cond_expr not_cond_expr simple_cond_expr
%type <cond_op> binary_cond_op
%type <string> assignment_op subscript
%type <string> expression range_expr range_start range_end
%type <string> multiplication_expr elementwise_multiplication_expr
%type <string> leq_expr pre_unary_expr post_unary_expr simple_expr
%type <string> cubic_array row_vector column_vector range_vector
%type <argument_list> argument_list

%%

statement_list: statement
{
  block_stmt_t *stmt;
  stmt = (block_stmt_t *)malloc(sizeof(block_stmt_t));
  stmt->type = BlockStmtType;
  stmt->n_stmts = 1;
  stmt->stmt = (stmt_t **)malloc(sizeof(stmt_t *));
  stmt->stmt[0] = $1;
  $$ = last = (stmt_t *)stmt;
}
| statement_list statement_separator statement
{
  block_stmt_t *stmt;
  int i;
  stmt = (block_stmt_t *)$1;
  i = stmt->n_stmts++;
  stmt->stmt =
    (stmt_t **)realloc(stmt->stmt, sizeof(stmt_t *) * stmt->n_stmts);
  stmt->stmt[i] = $3;
  $$ = last = (stmt_t *)stmt;
}
;

statement_separator: ';' | TOK_EOL
;

statement: /* empty */
{
  ss_stmt_t *stmt;
  stmt = (ss_stmt_t *)malloc(sizeof(ss_stmt_t));
  stmt->type = ServerSideStmtType;
  strcpy(stmt->stmt, "");
  $$ = (stmt_t *)stmt;
}
| if_statement
| while_statement
| continue_statement
| break_statement
| load_statement
| save_statement
| message_statement
| pprint_statement
| procedure_call
| prefer_statement
| assignment
;

if_statement: TOK_IF '(' cond_expr ')' '{' statement_list '}' else_block
{
  if_stmt_t *stmt;
  stmt = (if_stmt_t *)malloc(sizeof(if_stmt_t));
  stmt->type = IfStmtType;
  stmt->cond_expr = $3;
  stmt->then_block = $6;
  stmt->else_block = $8;
  $$ = (stmt_t *)stmt;
}
;

else_block: /* empty */
{
  $$ = NULL;
}
| TOK_ELSE if_statement
{
  $$ = $2;
}
| TOK_ELSE '{' statement_list '}'
{
  $$ = $3;
}
;

while_statement: TOK_WHILE '(' cond_expr ')' '{' statement_list '}'
{
  while_stmt_t *stmt;

  stmt = (while_stmt_t *)malloc(sizeof(while_stmt_t));
  stmt->type = WhileStmtType;
  stmt->cond_expr = $3;
  stmt->body = $6;
  $$ = (stmt_t *)stmt;
}
;

continue_statement: TOK_CONTINUE
{
  continue_stmt_t *stmt;
  stmt = (continue_stmt_t *)malloc(sizeof(continue_stmt_t));
  stmt->type = ContinueStmtType;
  $$ = (stmt_t *)stmt;
}
;

break_statement: TOK_BREAK
{
  break_stmt_t *stmt;
  stmt = (break_stmt_t *)malloc(sizeof(break_stmt_t));
  stmt->type = BreakStmtType;
  $$ = (stmt_t *)stmt;
}
;

cond_expr: and_cond_expr
| and_cond_expr TOK_OR and_cond_expr
{
  binary_cond_expr_t *expr;
  expr = (binary_cond_expr_t *)malloc(sizeof(binary_cond_expr_t));
  expr->type = BinaryCondExprType;
  expr->op = OrOp;
  expr->lhs = $1;
  expr->rhs = $3;
  $$ = (expr_t *)expr;
}
;

and_cond_expr: not_cond_expr
| not_cond_expr TOK_AND not_cond_expr
{
  binary_cond_expr_t *expr;
  expr = (binary_cond_expr_t *)malloc(sizeof(binary_cond_expr_t));
  expr->type = BinaryCondExprType;
  expr->op = AndOp;
  expr->lhs = $1;
  expr->rhs = $3;
  $$ = (expr_t *)expr;
}
;

not_cond_expr: simple_cond_expr
| TOK_NOT simple_cond_expr
{
  unary_cond_expr_t *expr;
  expr = (unary_cond_expr_t *)malloc(sizeof(unary_cond_expr_t));
  expr->type = UnaryCondExprType;
  expr->op = NotOp;
  expr->operand = $2;
  $$ = (expr_t *)expr;
}
;

simple_cond_expr: expression binary_cond_op expression
{
  binary_cond_expr_t *expr;
  expr = (binary_cond_expr_t *)malloc(sizeof(binary_cond_expr_t));
  expr->type = BinaryCondExprType;
  expr->op = $2;
  expr->lhs = (expr_t *)malloc(sizeof(ss_expr_t));
  expr->lhs->type = ServerSideExprType;
  strcpy(SS_EXPR(expr->lhs)->expr, $1);
  expr->rhs = (expr_t *)malloc(sizeof(ss_expr_t));
  expr->rhs->type = ServerSideExprType;
  strcpy(SS_EXPR(expr->rhs)->expr, $3);
  $$ = (expr_t *)expr;
}
| TOK_TRUE
{
  bool_expr_t *expr;
  expr = (bool_expr_t *)malloc(sizeof(bool_expr_t));
  expr->type = BoolExprType;
  expr->bool = 1;
  $$ = (expr_t *)expr;
}
| TOK_FALSE
{
  bool_expr_t *expr;
  expr = (bool_expr_t *)malloc(sizeof(bool_expr_t));
  expr->type = BoolExprType;
  expr->bool = 0;
  $$ = (expr_t *)expr;
}
;

binary_cond_op: TOK_LESS { $$ = LessOp; }
| TOK_LESS_EQUAL         { $$ = LessEqualOp; }
| TOK_GREATER            { $$ = GreaterOp; }
| TOK_GREATER_EQUAL      { $$ = GreaterEqualOp; }
| TOK_DOUBLE_EQUAL       { $$ = EqualOp; }
| TOK_NOT_EQUAL          { $$ = NotEqualOp; }
;

load_statement: TOK_LOAD TOK_STRING_LITERAL ',' argument_list
{
  load_stmt_t *stmt;
  stmt = (load_stmt_t *)malloc(sizeof(load_stmt_t));
  stmt->type = LoadStmtType;
  strcpy(stmt->path, $2);
  stmt->n_names = $4.n;
  stmt->name = $4.v;
  $$ = (stmt_t *)stmt;
}
;

save_statement: TOK_SAVE TOK_STRING_LITERAL ',' argument_list
{
  save_stmt_t *stmt;
  stmt = (save_stmt_t *)malloc(sizeof(save_stmt_t));
  stmt->type = SaveStmtType;
  strcpy(stmt->path, $2);
  stmt->n_names = $4.n;
  stmt->name = $4.v;
  $$ = (stmt_t *)stmt;
}
;

message_statement: TOK_MESSAGE TOK_STRING_LITERAL
{
  message_stmt_t *stmt;
  stmt = (message_stmt_t *)malloc(sizeof(message_stmt_t));
  stmt->type = MessageStmtType;
  strcpy(stmt->msg, $2);
  $$ = (stmt_t *)stmt;
}
;

pprint_statement: TOK_PPRINT expression
{
  pprint_stmt_t *stmt;
  stmt = (pprint_stmt_t *)malloc(sizeof(pprint_stmt_t));
  stmt->type = PPrintStmtType;
  strcpy(stmt->expr, $2);
  $$ = (stmt_t *)stmt;
}
;

procedure_call: TOK_IDENTIFIER '(' ')'
{
  ss_stmt_t *stmt;
  stmt = (ss_stmt_t *)malloc(sizeof(ss_stmt_t));
  stmt->type = ServerSideStmtType;
  strcpy(stmt->stmt, $1);
  strcat(stmt->stmt, "()");
  $$ = (stmt_t *)stmt;
}
| TOK_IDENTIFIER '(' argument_list ')'
{
  ss_stmt_t *stmt;
  int i;

  stmt = (ss_stmt_t *)malloc(sizeof(ss_stmt_t));
  stmt->type = ServerSideStmtType;
  strcpy(stmt->stmt, $1);
  strcat(stmt->stmt, "(");
  strcat(stmt->stmt, $3.v[0]);
  for (i = 1; i < $3.n; i++) {
    strcat(stmt->stmt, ",");
    strcat(stmt->stmt, $3.v[i]);
    free($3.v[i]);
  }
  free($3.v);
  strcat(stmt->stmt, ")");
  $$ = (stmt_t *)stmt;
}
;

argument_list: expression
{
  $$.v = (char **)malloc(sizeof(char *));
  $$.v[0] = strdup($1);
  $$.n = 1;
}
| argument_list ',' expression
{
  $$.v = (char **)realloc($1.v, ($1.n + 1) * sizeof(char *));
  $$.v[$1.n] = strdup($3);
  $$.n = $1.n + 1;
}
;

prefer_statement: TOK_PREFER TOK_IDENTIFIER
{
  ss_stmt_t *stmt;
  stmt = (ss_stmt_t *)malloc(sizeof(ss_stmt_t));
  stmt->type = ServerSideStmtType;
  strcpy(stmt->stmt, "prefer ");
  strcat(stmt->stmt, $2);
  $$ = (stmt_t *)stmt;
}
;

assignment: TOK_IDENTIFIER assignment_op expression
{
  ss_stmt_t *stmt;
  stmt = (ss_stmt_t *)malloc(sizeof(ss_stmt_t));
  stmt->type = ServerSideStmtType;
  strcpy(stmt->stmt, $1);
  strcat(stmt->stmt, $2);
  strcat(stmt->stmt, $3);
  $$ = (stmt_t *)stmt;
}
| TOK_IDENTIFIER subscript assignment_op expression
{
  ss_stmt_t *stmt;
  stmt = (ss_stmt_t *)malloc(sizeof(ss_stmt_t));
  stmt->type = ServerSideStmtType;
  strcpy(stmt->stmt, $1);
  strcat(stmt->stmt, $2);
  strcat(stmt->stmt, $3);
  strcat(stmt->stmt, $4);
  $$ = (stmt_t *)stmt;
}
;

assignment_op: TOK_EQUAL         { strcpy($$, "="); }
| TOK_PLUS_EQUAL                 { strcpy($$, "+="); }
| TOK_MINUS_EQUAL		 { strcpy($$, "-="); }
| TOK_MULTIPLY_EQUAL	         { strcpy($$, "*="); }
| TOK_MULTIPLY_ELEMENTWISE_EQUAL { strcpy($$, "*@="); }
| TOK_DIVIDE_EQUAL		 { strcpy($$, "/="); }
| TOK_DIVIDE_ELEMENTWISE_EQUAL   { strcpy($$, "/@="); }
| TOK_REMAINDER_EQUAL            { strcpy($$, "%="); }
;

subscript: '[' range_expr ']'
{
  strcpy($$, "[");
  strcat($$, $2);
  strcat($$, "]");
}  
| '[' range_expr ',' range_expr ']'
{
  strcpy($$, "[");
  strcat($$, $2);
  strcat($$, ",");
  strcat($$, $4);
  strcat($$, "]");
}  
| '[' range_expr ',' range_expr ',' range_expr ']'
{
  strcpy($$, "[");
  strcat($$, $2);
  strcat($$, ",");
  strcat($$, $4);
  strcat($$, ",");
  strcat($$, $6);
  strcat($$, "]");
}  
;

range_expr: expression
| range_start ':' range_end
{
  strcpy($$, $1);
  strcat($$, ":");
  strcat($$, $3);
} 
;

range_start: expression
| /* empty */
{
  strcpy($$, "");
}
;

range_end: expression
| /* empty */
{
  strcpy($$, "");
}
;

expression: multiplication_expr
| expression TOK_PLUS multiplication_expr
{
  strcpy($$, $1);
  strcat($$, "+");
  strcat($$, $3);
}
| expression TOK_MINUS multiplication_expr
{
  strcpy($$, $1);
  strcat($$, "-");
  strcat($$, $3);
}
;

multiplication_expr: elementwise_multiplication_expr
| multiplication_expr TOK_MULTIPLY elementwise_multiplication_expr
{
  strcpy($$, $1);
  strcat($$, "*");
  strcat($$, $3);
}
| multiplication_expr TOK_DIVIDE elementwise_multiplication_expr
{
  strcpy($$, $1);
  strcat($$, "/");
  strcat($$, $3);
}
| multiplication_expr TOK_REMAINDER elementwise_multiplication_expr
{
  strcpy($$, $1);
  strcat($$, "%");
  strcat($$, $3);
}
;

elementwise_multiplication_expr: leq_expr
| elementwise_multiplication_expr TOK_MULTIPLY_ELEMENTWISE pre_unary_expr
{
  strcpy($$, $1);
  strcat($$, "*@");
  strcat($$, $3);
}
| elementwise_multiplication_expr TOK_DIVIDE_ELEMENTWISE pre_unary_expr
{
  strcpy($$, $1);
  strcat($$, "/@");
  strcat($$, $3);
}
;

leq_expr: pre_unary_expr
|  pre_unary_expr TOK_LEQ leq_expr
{
  strcpy($$, $1);
  strcat($$, "\\");
  strcat($$, $3);
}
;

pre_unary_expr: post_unary_expr
| TOK_PLUS pre_unary_expr
{
  strcpy($$, "+");
  strcat($$, $2);
}
| TOK_MINUS pre_unary_expr
{
  strcpy($$, "-");
  strcat($$, $2);
}
;

post_unary_expr: simple_expr
| post_unary_expr TOK_CONJUGATE_TRANSPOSE
{
  strcpy($$, $1);
  strcat($$, "'");
}
| post_unary_expr TOK_COMPLEX_CONJUGATE
{
  strcpy($$, $1);
  strcat($$, "~");
}
| post_unary_expr TOK_TRANSPOSE
{
  strcpy($$, $1);
  strcat($$, "'~");
}
| post_unary_expr subscript
{
  strcpy($$, $1);
  strcat($$, $2);
}
;

simple_expr: TOK_IDENTIFIER '(' ')'
{
  strcpy($$, $1);
  strcat($$, "()");
} 
| TOK_IDENTIFIER '(' argument_list ')'
{
  int i;

  strcpy($$, $1);
  strcat($$, "(");
  strcat($$, $3.v[0]);
  for (i = 1; i < $3.n; i++) {
    strcat($$, ",");
    strcat($$, $3.v[i]);
    free($3.v[i]);
  }
  free($3.v);
  strcat($$, ")");
} 
| TOK_IDENTIFIER
{
  strcpy($$, $1);
}  
| TOK_INT4_LITERAL
{
  strcpy($$, $1);
}  
| TOK_INT8_LITERAL
{
  strcpy($$, $1);
}  
| TOK_REAL4_LITERAL
{
  strcpy($$, $1);
}  
| TOK_REAL8_LITERAL
{
  strcpy($$, $1);
}  
| '{' cubic_array '}'
{
  strcpy($$, "{");
  strcat($$, $2);
  strcat($$, "}");
}
| '(' expression ')'
{
  strcpy($$, "(");
  strcat($$, $2);
  strcat($$, ")");
}
;

cubic_array: row_vector
| cubic_array TOK_DOUBLE_SEMICOLON row_vector
{
  strcpy($$, $1);
  strcat($$, ";;");
  strcat($$, $3);
}
;

row_vector: column_vector
| row_vector ';' column_vector
{
  strcpy($$, $1);
  strcat($$, ";");
  strcat($$, $3);
}
;

column_vector: range_vector
| column_vector ',' range_vector
{
  strcpy($$, $1);
  strcat($$, ",");
  strcat($$, $3);
}
;

range_vector: expression
| expression ':' expression
{
  strcpy($$, $1);
  strcat($$, ":");
  strcat($$, $3);
}
;

%%

int
is_partial(char *s)
{
  int token, nested, lineno;
  int yylex(void);

  if (strcmp(s + strlen(s) - 2, "\\\n") == 0)
    return 1;
  set_text(s);
  lineno = parser_lineno;
  nested = 0;
  while ((token = yylex()) != 0) {
    switch (token){
    case '{': nested++; break;
    case '}': nested--; break;
    }
  }
  parser_lineno = lineno;
  if (nested < 0)
    yyerror("unmatched braces");
  if (parser_error)
    return -1;
  return nested > 0;
}

stmt_t *
parse_script(char *s)
{
  set_text(s);
  yyparse();
  if (parser_error)
    return NULL;
  return last;
}

static void
free_expression(expr_t *expr)
{
  switch(expr->type) {
  case UnaryCondExprType:
    free_expression(UNARY_COND(expr)->operand);
    break;
  case BinaryCondExprType:
    free_expression(BINARY_COND(expr)->lhs);
    free_expression(BINARY_COND(expr)->rhs);
    break;
  }
  free(expr);
}

void
free_statement(stmt_t *stmt)
{
  int i;

  switch(stmt->type) {
  case BlockStmtType:
    for (i = 0; i < BLOCK(stmt)->n_stmts; i++)
      free(BLOCK(stmt)->stmt[i]);
    free(BLOCK(stmt)->stmt);
    break;
  case IfStmtType:
    free_expression(IF(stmt)->cond_expr);
    free_statement(IF(stmt)->then_block);
    if (IF(stmt)->else_block)
      free_statement(IF(stmt)->else_block);
    break;
  case WhileStmtType:
    free_expression(WHILE(stmt)->cond_expr);
    free_statement(WHILE(stmt)->body);
    break;
  case LoadStmtType:
    for (i = 0; i < LOAD(stmt)->n_names; i++)
      free(LOAD(stmt)->name[i]);
    free(LOAD(stmt)->name);
    break;
  case SaveStmtType:
    for (i = 0; i < SAVE(stmt)->n_names; i++)
      free(SAVE(stmt)->name[i]);
    free(LOAD(stmt)->name);
    break;
  }
  free(stmt);
}

#ifdef DEBUG
void
dump_expression(expr_t *expr)
{
  switch(expr->type) {
  case ServerSideExprType:
    printf("%s", SS_EXPR(expr)->expr);
    break;
  case BoolExprType:
    if (BOOL(expr)->bool)
      printf("true");
    else
      printf("false");
    break;
  case UnaryCondExprType:
    switch (UNARY_COND(expr)->op) {
    case NotOp:
      printf("not ");
      dump_expression(UNARY_COND(expr)->operand);
      break;
    }
    break;
  case BinaryCondExprType:
    dump_expression(BINARY_COND(expr)->lhs);
    switch (UNARY_COND(expr)->op) {
    case AndOp:
      printf(" and ");
      break;
    case OrOp:
      printf(" or ");
      break;
    case EqualOp:
      printf(" == ");
      break;
    case NotEqualOp:
      printf(" != ");
      break;
    case LessOp:
      printf(" < ");
      break;
    case LessEqualOp:
      printf(" <= ");
      break;
    case GreaterOp:
      printf(" > ");
      break;
    case GreaterEqualOp:
      printf(" >= ");
      break;
    }
    dump_expression(BINARY_COND(expr)->rhs);
    break;
  }
}

void
dump_statement(stmt_t *stmt, int level)
{
  int i, j, len;
  ss_stmt_t *s;
  char buf[1024];

#define indent(lv) { \
    int k;			\
    printf("parser: ");		\
    for (k = 0; k < lv; k++)	\
      printf("    ");		\
  }
  switch(stmt->type) {
  case ServerSideStmtType:
    if (strlen(SS(stmt)->stmt) > 0) {
      indent(level);
      printf("%s\n", SS(stmt)->stmt);
    }
    break;
  case BlockStmtType:
    if (level > 0)
      printf("{\n");
    for (i = 0; i < BLOCK(stmt)->n_stmts; i++)
      dump_statement(BLOCK(stmt)->stmt[i], level);
    if (level > 0) {
      indent(level-1);
      printf("}");
    }
    break;
  case IfStmtType:
    indent(level);
    printf("if (");
    dump_expression(IF(stmt)->cond_expr);
    printf(") ");
    dump_statement(IF(stmt)->then_block, level+1);
    if (IF(stmt)->else_block) {
      printf(" else ");
      dump_statement(IF(stmt)->else_block, level+1);
    }
    printf("\n");
    break;
  case WhileStmtType:
    indent(level);
    printf("while (");
    dump_expression(WHILE(stmt)->cond_expr);
    printf(") ");
    dump_statement(WHILE(stmt)->body, level+1);
    printf("\n");
    break;
  case BreakStmtType:
    indent(level);
    printf("break\n");
    break;
  case LoadStmtType:
    indent(level);
    printf("load \"%s\"", LOAD(stmt)->path);
    for (i = 0; i < LOAD(stmt)->n_names; i++)
      printf(", %s", LOAD(stmt)->name[i]);
    printf("\n");
    break;
  case SaveStmtType:
    indent(level);
    printf("save \"%s\"", SAVE(stmt)->path);
    for (i = 0; i < SAVE(stmt)->n_names; i++)
      printf(", %s", SAVE(stmt)->name[i]);
    printf("\n");
    break;
  case MessageStmtType:
    indent(level);
    printf("message \"%s\"\n", MESSAGE(stmt)->msg);
    break;
  case PPrintStmtType:
    indent(level);
    printf("pprint %s\n", PPRINT(stmt)->expr);
    break;
  }
}
#endif
