/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: demo1.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <malloc.h>

#include "client.h"

void
print_matrix(const char *name, silc_envelope_t object)
{
  int i, j;
  double *v;
 
  if (object.type != SILC_MATRIX_TYPE ||
      object.precision != SILC_DOUBLE) {
    fprintf(stderr, "%s: not a matrix of double\n", name);
    return;
  }
  v = object.v;
  printf("MATRIX %s\n", name);
  for (i = 0; i < object.m; i++) {
    printf("(");
    for (j = 0; j < object.n; j++)
      printf("\t%lf", v[i + object.m * j]);
    printf("\t)\n");
  }
}

void
print_vector(const char *name, silc_envelope_t object)
{
  int i;
  double *v;
 
  if (object.type != SILC_COLUMN_VECTOR_TYPE ||
      object.precision != SILC_DOUBLE) {
    fprintf(stderr, "%s: not a column vector of double\n", name);
    return;
  }
  v = object.v;
  printf("COLUMN VECTOR %s\n", name);
  printf("(");
  for (i = 0; i < object.length; i++)
    printf("\t%lf", *v++);
  printf("\t)\n");
}

#define N 4

int main(int argc, char **argv)
{
  int i, j;
  silc_envelope_t object;
  double A[N * N];
  double b[N], x[N];
  double z[2];

  for (i = 0; i < N * N; i++) A[i] = (double)(i + 1);
  for (i = 0; i < N; i++) b[i] = (double)(i * 2 + 1);

  SILC_INIT();

  object.v = A;
  object.type = SILC_MATRIX_TYPE;
  object.precision = SILC_DOUBLE;
  object.format = SILC_FORMAT_DENSE;
  object.m = N;
  object.n = N;
  print_matrix("A", object);
  SILC_PUT("A", &object);

  object.v = b;
  object.type = SILC_COLUMN_VECTOR_TYPE;
  object.precision = SILC_DOUBLE;
  object.length = N;
  print_vector("b", object);
  SILC_PUT("b", &object);

  SILC_EXEC("prefer blasmodule");
  SILC_EXEC("x = A * b");
  object.v = x;
  SILC_GET(&object, "x");
  print_vector("x", object);

  SILC_EXEC("d = diagvec(A)");
  SILC_GET(&object, "d");
  print_vector("d", object);

  SILC_EXEC("split(A, L, D, U)");
  object.v = A;
  SILC_GET(&object, "L");
  print_matrix("L", object);
  SILC_GET(&object, "D");
  print_matrix("D", object);
  SILC_GET(&object, "U");
  print_matrix("U", object);

  SILC_EXEC("SUM = D");
  SILC_EXEC("SUM += LL");
  SILC_EXEC("SUMm += L");
  SILC_GET(&object, "SUM");
  print_matrix("SUM", object);
  SILC_EXEC("SUM += U");
  SILC_GET(&object, "SUM");
  print_matrix("SUM", object);

  SILC_EXEC("_SUB = A - A");
  SILC_GET(&object, "_SUB");
  print_matrix("_SUB", object);
  SILC_EXEC("_ADD = A + A");
  SILC_GET(&object, "_ADD");
  print_matrix("_ADD", object);
  SILC_EXEC("_SUB = b - b");
  SILC_GET(&object, "_SUB");
  print_vector("_SUB", object);
  SILC_EXEC("_ADD = b + b");
  SILC_GET(&object, "_ADD");
  print_vector("_ADD", object);

  SILC_EXEC("y = 1 + 2 * 3");
  SILC_GET(&object, "y");
  printf("SCALAR y\n");
  if (object.type == SILC_SCALAR_TYPE &&
      object.precision == SILC_INT) {
    printf("%d\n", *(int *)object.v);
  } else {
    printf("not a scalar of int\n");
  }

  SILC_EXEC("z = 3 - 2 * i");
  object.v = z;
  SILC_GET(&object, "z");
  printf("SCALAR z\n");
  if (object.type == SILC_SCALAR_TYPE &&
      object.precision == SILC_DOUBLE_COMPLEX) {
    printf("%f%+fi\n", z[0], z[1]);
  } else {
    printf("not a scalar of double complex\n");
  }

  object.v = b;
  SILC_EXEC("B = b[1]");
  SILC_GET(&object, "B");
  if (object.type == SILC_SCALAR_TYPE &&
      object.precision == SILC_DOUBLE) {
    printf("%f\n", b[0]);
  } else {
    printf("not a scalar of double\n");
  }
  SILC_EXEC("B = b[2:3]");
  SILC_GET(&object, "B");
  print_vector("B", object);
  SILC_EXEC("C = {1:5}[2:4]");

  SILC_EXEC("C = A[:2,3:]");
  SILC_GET(&object, "C");
  print_matrix("C", object);
  SILC_EXEC("t = {4,3,2,1}; C = A[t,t]");
  SILC_GET(&object, "C");
  print_matrix("C", object);

  SILC_EXEC("c1 = {1.0,2.0,3.0}");
  SILC_EXEC("c2 = {4.0,5.0,6.0}");
  SILC_EXEC("c3 = {7.0,8.0,9.0}");
  SILC_EXEC("m = {c1;c2;c3}");
  SILC_EXEC("A3 = {m;;m;;m}");

  SILC_EXEC("a = A3[1,1,1]");
  SILC_EXEC("a = A3[2,2,2]");
  SILC_EXEC("a = A3[1,:,:]");
  SILC_EXEC("a = A3[:,1,:]");
  SILC_EXEC("a = A3[:,:,1]");

  SILC_EXEC("C = A'");
  SILC_GET(&object, "C");
  print_matrix("C", object);

  SILC_EXEC("v = size(A)");
  SILC_EXEC("m = size(A, 1)");
  SILC_EXEC("n = size(A, 2)");

  SILC_FINALIZE();
  return 0;
}
