/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: linpack.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>

#include "client.h"

extern double wtime(void);

void
generate_matrix(double *A, double *b, int n)
{
  int seed, i, j, k;

#define next(seed) (3125 * seed % 65536)

  seed = 1325;
  k = 0;
  for (j = 0; j < n; j++) {
    for (i = 0; i < n; i++) {
      seed = next(seed);
      A[k++] = (seed - 32768.0) / 16384.0;
    }
    seed = next(seed); /* XXX */
  }
  for (i = 0; i < n; i++)
    b[i] = 0.0;
  k = 0;
  for (j = 0; j < n; j++)
    for (i = 0; i < n; i++)
      b[i] += A[k++];
}

double
compute_residual(double *A, double *b, double *x, int n)
{
  double norm, d;
  int i, j;

  norm = 0.0;
  for (i = 0; i < n; i++) {
    d = b[i];
    for (j = 0; j < n; j++)
      d -= A[i + j * n] * x[j];
    norm += d * d;
  }
  return sqrt(norm);
}

int
main(int argc, char *argv[])
{
  silc_envelope_t object;
  double *A, *b, *x, t, t_total, t_comp, t_comm, mflops, norm, d;
  int n;
  size_t n_bytes;

  n = 1000;
  if (argc == 2) {
    n = atol(argv[1]);
    if (n < 1) {
      fprintf(stderr, "Matrix size N must be greater than 1\n");
      return -1;
    }
  }
  A = (double *)malloc(sizeof(double) * n * n);
  b = (double *)malloc(sizeof(double) * n);
  x = (double *)malloc(sizeof(double) * n);
  generate_matrix(A, b, n);

  t = wtime();

  SILC_INIT();

  /* put matrix A */
  object.v = A;
  object.type = SILC_MATRIX_TYPE;
  object.precision = SILC_DOUBLE;
  object.format = SILC_FORMAT_DENSE;
  object.m = object.n = n;
  SILC_PUT("A", &object);

  /* put vector b */
  object.v = b;
  object.type = SILC_COLUMN_VECTOR_TYPE;
  object.precision = SILC_DOUBLE;
  object.length = n;
  SILC_PUT("b", &object);

  /* run LINPACK benchmark */
  SILC_EXEC("linpack(A, b, x, t)");

  /* get vector x */
  object.v = x;
  SILC_GET(&object, "x");

  t_total = wtime() - t;

  /* get scalar t */
  object.v = &t_comp;
  SILC_GET(&object, "t");

  SILC_FINALIZE();

  t_comm = t_total - t_comp;
  n_bytes = 128; /* INIT */
  n_bytes += (32 + sizeof(int) * 8 + 32) * 3; /* GET/PUT */
  n_bytes += 32; /* SILC */
  n_bytes += sizeof(size_t) * 2 + sizeof(double) * n * n; /* A */
  n_bytes += sizeof(size_t) + sizeof(double) * n; /* b */
  n_bytes += sizeof(size_t) + sizeof(double) * n; /* x */
  /*
   * column 1: matrix size N
   * column 2: elapsed time (in seconds)
   * column 3: performance (Mflops)
   * column 4: transferred data size (in bytes)
   * column 5: communication time (in seconds)
   * column 6: throughput (Mbps)
   */
  mflops = ((2.0 * n * n * n) / 3.0 + 2.0 * n * n) / (1.0e+6 * t_total);
  printf("%d %.3lf %.3lf %ld %.3lf %.3lf\n",
	 n, t_total, mflops, n_bytes, t_comm, n_bytes * 8 / (1.0e6 * t_comm));
#ifdef DEBUG
  printf("||b-Ax|| = %e\n", compute_residual(A, b, x, n));
#endif
  free(x);
  free(b);
  free(A);
  return 0;
}
