/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mm_band_sa.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

/* mm_band_sa.c - a "stand-alone" version of mm_band.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <unistd.h> /* for getopt() */

#include "client.h"

#include "mm_read.h"

/* blasdriver.o */
void
call_dgbmv(char trans, int m, int n, int kl, int ku, double alpha, double *a,
	   int lda, double *x, int incx, double beta, double *y, int incy);

/* lapackdriver.o */
void
call_dgbtrf(int m, int n, int kl, int ku, double *a, int lda, int *ipiv,
	    int *info);
void
call_dgbtrs(char trans, int n, int kl, int ku, int nrhs, double *a, int lda,
	    int *ipiv, double *b, int ldb, int *info);

extern double wtime(void); /* wtime.o */

int
main(int argc, char *argv[])
{
  FILE *fp;
  mm_header header;
  double *A, *b, *x;
  int i, kl, ku, lda;
  int opt, nrhs, n;
  int *ipiv, info;
  double *p;
  double t, elapsed;

#define USAGE "Usage: %s [-r nrhs] filename.mtx\n"

  nrhs = 1;
  while ((opt = getopt(argc, argv, "r:")) != EOF) {
    switch (opt) {
    case 'r':
      if ((nrhs = atoi(optarg)) <= 0) {
	printf("Error: nrhs <= 0\n");
	return -1;
      }
      break;
    default:
      printf(USAGE, argv[0]);
      return -1;
    }
  }
  if (optind == argc) {
    printf(USAGE, argv[0]);
    return -1;
  }

  /* read a Matrix Market matrix */
  if ((fp = fopen(argv[optind], "rb")) == NULL) {
    perror(argv[optind]);
    return -1;
  }
  if (mm_read_header(fp, &header) < 0)
    return -1;
  if (mm_read_band(fp, &header, &A, &kl, &ku, 1) < 0)
    return -1;
  fclose(fp);
  printf("%s: %d x %d matrix\n", argv[optind], header.m, header.n);
  lda = kl * 2 + ku + 1;
  printf("band width = %d (lower = %d, upper = %d, lda = %d, %ld bytes)\n",
	 kl + ku + 1, kl, ku, lda, sizeof(double) * lda * header.n);

  /* create vector b */
  b = (double *)malloc(sizeof(double) * header.n);
  x = (double *)malloc(sizeof(double) * header.n * nrhs);
  for (i = 0; i < header.n; i++)
    x[i] = 1.0;
  /* b = A * x */
  call_dgbmv('N', header.m, header.n, kl, ku, 1.0,
	     A + kl, /* additional 'kl' rows for LU factorization */
	     lda, x, 1, 0.0, b, 1);
  for (n = 0; n < nrhs; n++)
    memcpy(x + header.n * n, b, sizeof(double) * header.n);

  t = wtime();

  /* solve "Ax = b" */
  if (header.m == 0 && header.n == 0)
    n = 1;
  else
    n = (header.m < header.n) ? header.m : header.n;
  ipiv = (int *)malloc(sizeof(int) * n);
  call_dgbtrf(header.m, header.n, kl, ku, A, lda, ipiv, &info);
  if (info) {
    fprintf(stderr, "DGBTRF failed (info = %d)\n", info);
    return -1;
  }
  call_dgbtrs('N', header.n, kl, ku, nrhs, A, lda, ipiv, x, header.n, &info);
  if (info) {
    fprintf(stderr, "DGBTRS failed (info = %d)\n", info);
    return -1;
  }
  free(ipiv);

  elapsed = wtime() - t;
  printf("%8lf sec.\n", elapsed);

#ifdef DEBUG
  free(A);
  if ((fp = fopen(argv[optind], "rb")) == NULL)
    return -1;
  if (mm_read_header(fp, &header) < 0)
    return -1;
  if (mm_read_band(fp, &header, &A, &kl, &ku, 1) < 0)
    return -1;
  fclose(fp);
  p = (double *)malloc(sizeof(double) * header.n);
  for (n = 0; n < nrhs; n++) {
    double dot;
    memcpy(p, b, sizeof(double) * header.n);
    /* b = b - A * x */
    call_dgbmv('N', header.m, header.n, kl, ku, -1.0, A + kl, lda,
	       x + header.n * n, 1, 1.0, p, 1);
    dot = 0.0;
    for (i = 0; i < header.n; i++)
      dot += p[i] * p[i];
    printf("||b-Ax|| = %le (%d)\n", sqrt(dot), n);
  }
  free(p);
#endif

  free(A);
  free(b);
  free(x);
  return 0;
}
