/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mm_crs.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <unistd.h> /* for getopt() */

#include "client.h"

#include "mm_read.h"

extern double wtime(void); /* wtime.o */

int
main(int argc, char *argv[])
{
  FILE *fp;
  mm_header header;
  silc_envelope_t object;
  double *value, *b, *x;
  int i, j, *row, *index;
  int opt, rhs_type, rhs_cos, nrhs, n;
  char name[32], expr[256], *expr_A;
  double t, elapsed;

#define USAGE "Usage: %s [-b] [-C] [-m] [-r nrhs] filename.mtx\n"

  expr_A = "A";
  rhs_type = SILC_COLUMN_VECTOR_TYPE; rhs_cos = 0;
  nrhs = 1;
  while ((opt = getopt(argc, argv, "bCmr:")) != EOF) {
    switch (opt) {
    case 'b':
      expr_A = "band(A)";
      break;
    case 'C':
      rhs_cos = 1;
      break;
    case 'm':
      rhs_type = SILC_MATRIX_TYPE;
      break;
    case 'r':
      if ((nrhs = atoi(optarg)) <= 0) {
	printf("Error: nrhs <= 0\n");
	return -1;
      }
      break;
    default:
      printf(USAGE, argv[0]);
      return -1;
    }
  }
  if (optind == argc) {
    printf(USAGE, argv[0]);
    return -1;
  }

  /* read a Matrix Market matrix */
  if ((fp = fopen(argv[optind], "rb")) == NULL) {
    perror(argv[optind]);
    return -1;
  }
  if (mm_read_header(fp, &header) < 0)
    return -1;
  if (mm_read_crs(fp, &header, &value, &row, &index) < 0)
    return -1;
  fclose(fp);
  printf("%s: %d x %d matrix (%d elements)\n",
	 argv[optind], header.m, header.n, row[header.m]);

  /* create vector b */
  if (rhs_type == SILC_COLUMN_VECTOR_TYPE) {
    b = (double *)malloc(sizeof(double) * header.n);
    x = (double *)malloc(sizeof(double) * header.n);
  } else {
    b = (double *)malloc(sizeof(double) * header.n * nrhs);
    x = (double *)malloc(sizeof(double) * header.n * nrhs);
  }
  if (rhs_cos) {
    for (i = 0; i < header.n; i++)
      b[i] = cos(M_PI / (header.n - 1) * i); /* cos(theta), 0 <= theta <= pi */
  } else {
    for (i = 0; i < header.n; i++) {
      b[i] = 0.0;
      for (j = row[i]; j < row[i+1]; j++)
	b[i] += value[j];
    }
  }
  if (rhs_type == SILC_MATRIX_TYPE) {
    for (n = 1; n < nrhs; n++)
      memcpy(b + header.n * n, b, sizeof(double) * header.n);
  }

  t = wtime();

  /* initialize SILC */
  SILC_INIT();

  /* put matrix A */
  object.v = value;
  object.type = SILC_MATRIX_TYPE;
  object.format = SILC_FORMAT_CRS;
  object.precision = SILC_DOUBLE;
  object.m = header.m;
  object.n = header.n;
  object.nnz = row[header.m];
  object.row = row;
  object.index = index;
  SILC_PUT("A", &object);

  /* put right-hand side(s) */
  object.v = b;
  object.type = rhs_type;
  object.precision = SILC_DOUBLE;
  if (rhs_type == SILC_COLUMN_VECTOR_TYPE) {
    object.length = header.n;
    for (i = 0; i < nrhs; i++) {
      sprintf(name, "b%d", i);
      SILC_PUT(name, &object);
      /* solve "Ax=b" */
      sprintf(expr, "x%d = %s \\ b%d", i, expr_A, i);
      SILC_EXEC(expr);
    }
    /* get solution(s) */
    object.v = x;
    for (i = 0; i < nrhs; i++) {
      sprintf(name, "x%d", i);
      SILC_GET(&object, name);
    }
  } else { /* SILC_MATRIX_TYPE */
    object.format = SILC_FORMAT_DENSE;
    object.m = header.n;
    object.n = nrhs;
    SILC_PUT("b", &object);
    /* solve "Ax=b" */
    sprintf(expr, "x = %s \\ b", expr_A);
    SILC_EXEC(expr);
    /* get solution(s) */
    object.v = x;
    SILC_GET(&object, "x");
  }

  elapsed = wtime() - t;
  printf("%8lf sec.\n", elapsed);

#ifdef DEBUG
  /* check solution(s) */
  if (rhs_type == SILC_COLUMN_VECTOR_TYPE) {
    double norm;
    object.v = &norm;
    for (n = 0; n < nrhs; n++) {
      sprintf(expr, "d = b%d - A * x%d", n, n);
      SILC_EXEC(expr);
      SILC_EXEC("norm = sqrt(d' * d)");
      SILC_GET(&object, "norm");
      printf("||b-Ax|| = %le (%d)\n", norm, n);
    }
  } else { /* SILC_MATRIX_TYPE */
    double *p, *q, dot, d;
    for (n = 0; n < nrhs; n++) {
      p = b + header.n * n;
      q = x + header.n * n;
      dot = 0.0;
      for (i = 0; i < header.m; i++) {
	d = p[i];
	for (j = row[i]; j < row[i+1]; j++)
	  d -= value[j] * q[index[j]];
	dot += d * d;
      }
      printf("||b-Ax|| = %le (%d)\n", sqrt(dot), n);
    }
  }
#endif

  /* finalize SILC */
  SILC_FINALIZE();

  free(value);
  free(row);
  free(index);
  free(b);
  free(x);
  return 0;
}
