/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mm_crs_sa.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

/* mm_crs_sa.c - a "stand-alone" version of mm_crs.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <unistd.h> /* for getopt() */

#include "ssi_leqlib.h"

#include "mm_read.h"

extern double wtime(void); /* wtime.o */

int
main(int argc, char *argv[])
{
  FILE *fp;
  mm_header header;
  SSI_MATRIX mat_A;
  double ssi_params[SSI_PARAMS_LEN], *work;
  int ssi_options[SSI_OPTIONS_LEN], status;
  double *value, *b, *x, d, norm;
  int i, j, *row, *index;
  int opt, verbose, rhs_cos, nrhs, maxiter, n;
  double t, elapsed;

#define USAGE "Usage: %s [-v] [-C] [-r nrhs] [-X maxiter] filename.mtx\n"

  nrhs = 1;
  verbose = rhs_cos = 0;
  maxiter = -1;
  while ((opt = getopt(argc, argv, "vCr:X:")) != EOF) {
    switch (opt) {
    case 'v':
      verbose = 1;
      break;
    case 'C':
      rhs_cos = 1;
      break;
    case 'r':
      if ((nrhs = atoi(optarg)) <= 0) {
	printf("Error: nrhs <= 0\n");
	return -1;
      }
      break;
    case 'X':
      if ((maxiter = atoi(optarg)) <= 0) {
	printf("Error: maxiter <= 0\n");
	return -1;
      }
      break;
    default:
      printf(USAGE, argv[0]);
      return -1;
    }
  }
  if (optind == argc) {
    printf(USAGE, argv[0]);
    return -1;
  }

  /* read a Matrix Market matrix */
  if ((fp = fopen(argv[optind], "rb")) == NULL) {
    perror(argv[optind]);
    return -1;
  }
  if (mm_read_header(fp, &header) < 0)
    return -1;
  if (mm_read_crs(fp, &header, &value, &row, &index) < 0)
    return -1;
  fclose(fp);
  printf("%s: %d x %d matrix (%d elements)\n",
	 argv[optind], header.m, header.n, row[header.m]);
  if (maxiter < 0)
    maxiter = header.m;

  /* create vectors b and x */
  b = (double *)malloc(sizeof(double) * header.n * nrhs);
  x = (double *)malloc(sizeof(double) * header.n * nrhs);
  if (rhs_cos) {
    for (i = 0; i < header.n; i++)
      b[i] = cos(M_PI / (header.n - 1) * i); /* cos(theta), 0 <= theta <= pi */
  } else {
    for (i = 0; i < header.n; i++) {
      b[i] = 0.0;
      for (j = row[i]; j < row[i+1]; j++)
	b[i] += value[j];
    }
  }
  for (n = 1; n < nrhs; n++)
    memcpy(b + header.n * n, b, sizeof(double) * header.n);
  for (i = 0; i < header.n * nrhs; i++)
    x[i] = 0.0;

  mat_A.matrix_type = SSI_MATRIX_TYPE_CRS;
  mat_A.array_base  = SSI_MATRIX_ARRAY_BASE_0;
  mat_A.matrix_base = SSI_MATRIX_ARRAY_BASE_0;
  mat_A.n     = header.m;
  mat_A.nnz   = row[header.m];
  mat_A.value = value;
  mat_A.index = index;
  mat_A.row   = row;
  work = (double *)malloc(sizeof(double) * header.m * 4);
  ssi_params[SSI_PARAMS_RESID] = 1.0e-12;
  ssi_options[SSI_OPTIONS_OUTPUT] = verbose;
  t = wtime();
  for (n = 0; n < nrhs; n++) {
    ssi_options[SSI_OPTIONS_MAXITER] = maxiter;
    status = ssi_cg(b + header.n * n, x + header.n * n,
		    work, ssi_params, ssi_options, &mat_A, NULL);
    if (status == SSIRET_MAXITER) {
      fprintf(stderr, "Error: too many iterations (abort)\n");
      break;
    }
  }
  elapsed = wtime() - t;
  printf("%8lf sec.\n", elapsed);
  free(work);

#ifdef DEBUG
  /* check solution */
  norm = 0.0;
  for (i = 0; i < header.m; i++) {
    d = b[i];
    for (j = row[i]; j < row[i+1]; j++)
      d -= value[j] * x[index[j]];
    norm += d * d;
  }
  printf("||b-Ax|| = %e\n", sqrt(norm));
#endif

  free(value);
  free(b);
  free(x);
  return 0;
}
