/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mm_lis.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <math.h>
#include <unistd.h> /* for getopt() */

#include "lis.h"

#include "mm_read.h"

int
main(int argc, char *argv[])
{
  FILE *fp;
  mm_header header;
  LIS_MATRIX lis_A;
  LIS_VECTOR lis_b, lis_x;
  LIS_SCALAR params[LIS_PARAMS_LEN];
  LIS_SCALAR status[LIS_STATUS_LEN];
  int options[LIS_OPTIONS_LEN];
  double *value, v, elapsed, Mflop;
  int i, j, *row, *index, rhs_cos, verbose, opt, iter;

#define USAGE "Usage: %s [-C] [-v] filename.mtx [-- lis_options]\n"\
  "-C  defines RHS vector b based on cos(theta), 0 <= theta <= pi\n"\
  "-v  verbose mode\n"

  rhs_cos = verbose = 0;
  while ((opt = getopt(argc, argv, "Cv")) != EOF) {
    switch (opt) {
    case 'C':
      rhs_cos = 1;
      break;
    case 'v':
      verbose = 1;
      break;
    default:
      printf(USAGE, argv[0]);
      return -1;
    }
  }
  if (optind == argc) {
    printf(USAGE, argv[0]);
    return -1;
  }

  /* read a Matrix Market matrix */
  if ((fp = fopen(argv[optind], "rb")) == NULL) {
    perror(argv[optind]);
    return -1;
  }
  if (mm_read_header(fp, &header) < 0)
    return -1;
  if (mm_read_crs(fp, &header, &value, &row, &index) < 0)
    return -1;
  fclose(fp);
  printf("%s: %d x %d matrix (%d elements)\n",
	 argv[optind], header.m, header.n, row[header.m]);

  lis_initialize(argc - optind, argv + optind, NULL, options, params);
  lis_matrix_create(header.m, 0, 0, &lis_A);
#if 1
  lis_matrix_set_crs(row[header.m], row, index, value, lis_A);
#else
  for (i = 0; i < header.m; i++)
    for (j = row[i]; j < row[i+1]; j++)
      lis_matrix_set(i, index[j], value[j], lis_A);
#endif
  lis_matrix_assemble(&lis_A, LIS_MATRIX_CRS);
  lis_vector_create(0, header.n, 0, &lis_b);
  for (i = 0; i < header.n; i++) {
    if (rhs_cos) {
      v = cos(M_PI / (header.n - 1) * i); /* cos(theta), 0 <= theta <= pi */
    } else {
      v = 0.0;
      for (j = row[i]; j < row[i+1]; j++)
	v += value[j];
    }
    lis_vector_set(i, v, lis_b);
  }
  lis_vector_create(0, header.n, 0, &lis_x);
  lis_vector_set_all(0.0, lis_x);
  lis_solve(lis_A, lis_b, lis_x, params, options, status);
  if (verbose) {
    elapsed = status[LIS_STATUS_TIMES];
    switch ((int)status[LIS_STATUS_RET]) {
    case LIS_MAXITER:
      iter = options[LIS_OPTIONS_MAXITER];
      break;
    default:
      iter = (int)status[LIS_STATUS_ITER];
    }
    Mflop = 4 * header.n;
    Mflop += (double)iter * (2 * row[header.m] + 12 * header.n + 3);
    Mflop /= 1024 * 1024;
    printf("%lf sec\n", elapsed);
    printf("#iters = %d\n", iter);
    printf("||b-Ax|| = %.6e\n", status[LIS_STATUS_RESID]);
    printf("%lf Mflop, %lf Mflops\n", Mflop, Mflop / elapsed);
  }
  lis_matrix_destroy(lis_A);
  lis_vector_destroy(lis_b);
  lis_vector_destroy(lis_x);
  lis_finalize();

  return 0;
}
