/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mm_mp_crs.c,v 1.3 2007/10/31 06:22:07 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <unistd.h> /* for getopt() */

#include "client.h"

#include "mm_read.h"

int
main(int argc, char *argv[])
{
  FILE *fp;
  mm_header header;
  silc_envelope_t object;
  double *value, *b, *x, *h;
  int i, j, *row, *index;
  int opt, rhs_cos, history, maxiter, prec, n;
  char expr[256];
  double t, elapsed, wtime(void);

#define USAGE "Usage: %s [-C] [-H] [-I maxiter] [-P prec] filename.mtx\n"

  rhs_cos = history = 0; maxiter = 1000; prec = 64;
  while ((opt = getopt(argc, argv, "CHI:P:")) != EOF) {
    switch (opt) {
    case 'C':
      rhs_cos = 1;
      break;
    case 'H':
      history = 1;
      break;
    case 'I':
      if ((maxiter = atoi(optarg)) == 0) {
	printf("Error: maxiter == 0\n");
	return -1;
      }
      break;
    case 'P':
      if ((prec = atoi(optarg)) == 0) {
	printf("Error: prec == 0\n");
	return -1;
      }
      break;
    default:
      printf(USAGE, argv[0]);
      return -1;
    }
  }
  if (optind == argc) {
    printf(USAGE, argv[0]);
    return -1;
  }

  /* read a Matrix Market matrix */
  if ((fp = fopen(argv[optind], "rb")) == NULL) {
    perror(argv[optind]);
    return -1;
  }
  if (mm_read_header(fp, &header) < 0)
    return -1;
  if (mm_read_crs(fp, &header, &value, &row, &index) < 0)
    return -1;
  fclose(fp);
  printf("%s: %d x %d matrix (%d elements)\n",
	 argv[optind], header.m, header.n, row[header.m]);

  /* create vector b */
  b = (double *)malloc(sizeof(double) * header.n);
  x = (double *)malloc(sizeof(double) * header.n);
  if (rhs_cos) {
    for (i = 0; i < header.n; i++)
      b[i] = cos(M_PI / (header.n - 1) * i); /* cos(theta), 0 <= theta <= pi */
  } else {
    for (i = 0; i < header.n; i++) {
      b[i] = 0.0;
      for (j = row[i]; j < row[i+1]; j++)
	b[i] += value[j];
    }
  }

  t = wtime();

  /* initialize SILC */
  SILC_INIT();

  /* initialize the leq_mp module */
  SILC_EXEC("prefer leq_mp");
  sprintf(expr, "mp_options[MP_OPTIONS_PREC] = %d", prec);
  SILC_EXEC(expr);
  sprintf(expr, "mp_options[MP_OPTIONS_MAXITER] = %d", maxiter);
  SILC_EXEC(expr);

  /* put matrix A */
  object.v = value;
  object.type = SILC_MATRIX_TYPE;
  object.format = SILC_FORMAT_CRS;
  object.precision = SILC_DOUBLE;
  object.m = header.m;
  object.n = header.n;
  object.nnz = row[header.m];
  object.row = row;
  object.index = index;
  SILC_PUT("A", &object);

  /* put vector b */
  object.v = b;
  object.type = SILC_COLUMN_VECTOR_TYPE;
  object.precision = SILC_DOUBLE;
  object.length = header.n;
  SILC_PUT("b", &object);

  /* solve Ax = b */
  SILC_EXEC("x = A \\ b");

  /* get solution x */
  object.v = x;
  SILC_GET(&object, "x");

  elapsed = wtime() - t;
  printf("%8lf sec.\n", elapsed);

  /* get residual history if necessary */
  if (history) {
    SILC_EXEC("h = get_residual_history()");
    object.v = NULL;
    if (SILC_GET(&object, "h") < 0) {
      fprintf(stderr, "Error: failed to get residual history\n");
      return -1;
    }
    h = object.v;
    for (i = 0; i < object.length; i++)
      printf("%d\t%.6e\n", i + 1, h[i]);
    free(h);
  }

  /* finalize SILC */
  SILC_FINALIZE();

  free(value);
  free(row);
  free(index);
  free(b);
  free(x);
  return 0;
}
