# Copyright (C) 2002-2007 The SSI Project. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the project nor the names of its contributors 
#       may be used to endorse or promote products derived from this software 
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# $Id: silc.py,v 1.4 2007/10/31 06:19:47 kajiyama Exp $

import os
import socket
import struct

# type
SILC_SCALAR_TYPE        = 1
SILC_RANGE_TYPE         = 2
SILC_COLUMN_VECTOR_TYPE = 4
SILC_ROW_VECTOR_TYPE    = 8
SILC_MATRIX_TYPE        = 16
SILC_CUBIC_ARRAY_TYPE   = 32

# precision
SILC_FLOAT          = 1
SILC_DOUBLE         = 2
SILC_COMPLEX        = 4
SILC_DOUBLE_COMPLEX = 8
SILC_INT            = 16
SILC_LONG           = 32

# format
SILC_FORMAT_DENSE = "SILC:dense (column major)"
SILC_FORMAT_CRS   = "SILC:CRS"
SILC_FORMAT_BAND  = "SILC:Band"

supported_formats = [
    SILC_FORMAT_DENSE,
    SILC_FORMAT_CRS,
    SILC_FORMAT_BAND,
    ]

class Envelope:
    formats = []
    def __init__(self):
        self.type = 0
        self.precision = 0
        self.format = 0
        self.v = None
        self.length = 0
        self.l = 0
        self.m = 0
        self.n = 0
        self.u = 0
        self.nnz = 0
        self.index = []
        self.row = []
    def get_headers(self):
        if self.type == SILC_MATRIX_TYPE:
            if self.format not in supported_formats:
                raise ValueError, "unsupported format: " + repr(self.format)
            try:
                format = Envelope.formats.index(self.format)
            except ValueError:
                raise ValueError, "unknown format: " + repr(self.format)
        else:
            format = 0
        return self.type, self.precision, format, 0, 0, 0, 0, 0
    def set_headers(self, headers):
        self.type, self.precision, format, x, x, x, x, x = headers
        self.format = Envelope.formats[format]

class Client:
    def __init__(self):
        self.packer = {
            SILC_SCALAR_TYPE:        self.pack_scalar,
            SILC_COLUMN_VECTOR_TYPE: self.pack_vector,
            SILC_ROW_VECTOR_TYPE:    self.pack_vector,
            SILC_MATRIX_TYPE:        self.pack_matrix,
            SILC_CUBIC_ARRAY_TYPE:   self.pack_cubic_array}
        self.unpacker = {
            SILC_SCALAR_TYPE:        self.unpack_scalar,
            SILC_COLUMN_VECTOR_TYPE: self.unpack_vector,
            SILC_ROW_VECTOR_TYPE:    self.unpack_vector,
            SILC_MATRIX_TYPE:        self.unpack_matrix,
            SILC_CUBIC_ARRAY_TYPE:   self.unpack_cubic_array}
    def INIT(self, host=None, port=None, debug=False):
        path = os.path.expanduser(os.path.join("~", ".silc"))
        try:
            conf = open(path).readline()
        except IOError:
            pass
        else:
            conf_host, conf_port = conf.split()
            if host is None:
                host = conf_host
            if port is None:
                port = int(conf_port)
        self.debug = debug
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((host, port))
        # protocol version (1)
        self.send_request("V1\n")
        resp = self.recv_response()
        if resp[:3] != "200":
            raise RuntimeError, "failed to inform protocol version"
        # network (big endian)
        self.send_request("B\n")
        resp = self.recv_response()
        if resp[:3] != "200":
            raise RuntimeError, "failed to inform endianness"
        # origin (0)
        self.send_request("0\n")
        resp = self.recv_response()
        if resp[:3] != "200":
            raise RuntimeError, "failed to inform origin"
        # size of size_t
        self.sizeof_size = struct.calcsize("!L")
        self.send_request("S%d\n" % self.sizeof_size)
        resp = self.recv_response()
        if resp[:3] != "200":
            raise RuntimeError, "failed to inform size of size_t"
        # list of format names
        self.send_request("F\n")
        resp = self.recv_response()
        if resp[:3] != "200":
            raise RuntimeError, "failed to receive list of format names"
        (size,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        data = self.recv_all(size)
        Envelope.formats = data.split("\n")[:-1]
        self.matrix_packer = {
            SILC_FORMAT_DENSE: self.pack_dense,
            SILC_FORMAT_CRS:   self.pack_crs,
            SILC_FORMAT_BAND:  self.pack_band}
        self.matrix_unpacker = {
            SILC_FORMAT_DENSE: self.unpack_dense,
            SILC_FORMAT_CRS:   self.unpack_crs,
            SILC_FORMAT_BAND:  self.unpack_band}
        # number of processes
        self.send_request("T1\n")
        resp = self.recv_response()
        if resp[:3] != "200":
            raise RuntimeError, "failed to inform number of processes"
    def SHUTDOWN(self):
        self.send_request("Q\n")
    def FINALIZE(self):
        self.sock.close()
    def EXEC(self, cmd):
        n = len(cmd)
        if n == 0:
            return
        elif n > 4096:
            raise ValueError, "too long command"
        self.send_request(":%d:%s\n" % (n, cmd))
        self.recv_response()
    def send_request(self, request):
        if self.debug:
            print "Request:", request[:request.index("\n")]
        if len(request) < 32:
            request = request + "\0" * (32 - len(request))
        self.sock.send(request)
    def recv_response(self):
        response = self.recv_all(32)
        if self.debug:
            print "Response:", response[:response.index("\n")]
        return response
    def send_all(self, data, flags=0):
        size = len(data)
        while data:
            n = self.sock.send(data, flags)
            data = data[n:]
        return size
    def recv_all(self, size, flags=0):
        buffer = []
        m = size
        while m > 0:
            data = self.sock.recv(m, flags)
            if not data:
                raise RuntimeError, "unexpected end of stream"
            buffer.append(data)
            m = m - len(data)
        return "".join(buffer)
    def send_ack(self):
        self.send_all("********************************")
    def recv_ack(self):
        self.recv_all(32)
    precision_format = {
        SILC_FLOAT:          ("f", 1),
        SILC_DOUBLE:         ("d", 1),
        SILC_COMPLEX:        ("f", 2),
        SILC_DOUBLE_COMPLEX: ("d", 2),
        SILC_INT:            ("i", 1), # int32
        SILC_LONG:           ("q", 1), # int64
    }
    def get_precision_format(self, precision, count=None):
        try:
            format, n = self.precision_format[precision]
        except KeyError:
            raise ValueError, "unknown precision: %s" % str(object.precision)
        if count is not None:
            n = n * count
        return "!%d%s" % (n, format)
    def PUT(self, name, object):
        assert isinstance(object, Envelope)
        self.send_request(">%s\n" % name)
        self.recv_ack()
        self.send_all(struct.pack("!8I", *object.get_headers()))
        self.recv_ack()
        packer = self.packer.get(object.type)
        if packer is None:
            raise ValueError, "unknown object type: %s" % str(object.type)
        self.send_all(packer(object))
        self.recv_response()
    def pack_scalar(self, object):
        format = self.get_precision_format(object.precision)
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            value = complex(object.v)
            return struct.pack(format, value.real, value.imag)
        return struct.pack(format, object.v)
    def pack_vector(self, object):
        self.send_all(struct.pack("!L", object.length))
        self.recv_ack()
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            v = []
            for i in range(object.length):
                value = complex(object.v[i])
                v.append(value.real)
                v.append(value.imag)
        else:
            v = object.v
        format = self.get_precision_format(object.precision, object.length)
        return struct.pack(format, *v)
    def pack_matrix(self, object):
        self.send_all(struct.pack("!L", object.m))
        self.recv_ack()
        self.send_all(struct.pack("!L", object.n))
        self.recv_ack()
        packer = self.matrix_packer[object.format]
        return packer(object)
    def pack_dense(self, object):
        # convert from row major to column major
        v = []
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            for j in range(object.n):
                for i in range(object.m):
                    value = complex(object.v[i][j])
                    v.append(value.real)
                    v.append(value.imag)
        else:
            for j in range(object.n):
                for i in range(object.m):
                    v.append(object.v[i][j])
        format = self.get_precision_format(object.precision,
                                           object.m * object.n)
        return struct.pack(format, *v)
    def pack_crs(self, object):
        assert len(object.v) == object.nnz
        assert len(object.row) == object.m + 1
        assert len(object.index) == object.nnz
        self.send_all(struct.pack("!L", object.nnz))
        self.recv_ack()
        format = self.get_precision_format(SILC_INT, object.m + 1)
        self.send_all(struct.pack(format, *object.row))
        self.recv_ack()
        format = self.get_precision_format(SILC_INT, object.nnz)
        self.send_all(struct.pack(format, *object.index))
        self.recv_ack()
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            v = []
            for i in range(object.nnz):
                value = complex(object.v[i])
                v.append(value.real)
                v.append(value.imag)
        else:
            v = object.v
        format = self.get_precision_format(object.precision, object.nnz)
        return struct.pack(format, *v)
    def pack_band(self, object):
        self.send_all(struct.pack("!L", object.l))
        self.recv_ack()
        self.send_all(struct.pack("!L", object.u))
        self.recv_ack()
        # convert from row major to column major
        v = []
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            for j in range(object.n):
                for i in range(object.l + object.u + 1):
                    value = complex(object.v[i][j])
                    v.append(value.real)
                    v.append(value.imag)
        else:
            for j in range(object.n):
                for i in range(object.l + object.u + 1):
                    v.append(object.v[i][j])
        ld = object.l + object.u + 1 # leading dimension
        format = self.get_precision_format(object.precision, ld * object.n)
        return struct.pack(format, *v)
    def pack_cubic_array(self, object):
        self.send_all(struct.pack("!L", object.l))
        self.recv_ack()
        self.send_all(struct.pack("!L", object.m))
        self.recv_ack()
        self.send_all(struct.pack("!L", object.n))
        self.recv_ack()
        # convert from row major to column major
        v = []
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            for i in range(object.l):
                for k in range(object.n):
                    for j in range(object.m):
                        value = complex(object.v[i][j][k])
                        v.append(value.real)
                        v.append(value.imag)
        else:
            for i in range(object.l):
                for k in range(object.n):
                    for j in range(object.m):
                        v.append(object.v[i][j][k])
        format = self.get_precision_format(object.precision,
                                           object.l * object.m * object.n)
        return struct.pack(format, *v)
    def GET(self, object, name):
        assert isinstance(object, Envelope)
        self.send_request("<%s\n" % name)
        self.recv_response()
        self.send_ack()
        object.set_headers(struct.unpack("!8I", self.recv_all(32)))
        self.send_ack()
        unpacker = self.unpacker.get(object.type)
        if unpacker is None:
            raise ValueError, "unknown object type: %s" % str(object.type)
        unpacker(object)
    def unpack_scalar(self, object):
        format = self.get_precision_format(object.precision)
        buffer = struct.unpack(format, self.recv_all(struct.calcsize(format)))
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            object.v = complex(buffer[0], buffer[1])
        else:
            object.v = buffer[0]
    def unpack_vector(self, object):
        (object.length,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        format = self.get_precision_format(object.precision, object.length)
        buffer = struct.unpack(format, self.recv_all(struct.calcsize(format)))
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            object.v = []
            for i in range(object.length):
                object.v.append(complex(buffer[i * 2], buffer[i * 2 + 1]))
        else:
            object.v = list(buffer)
    def unpack_matrix(self, object):
        (object.m,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        (object.n,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        unpacker = self.matrix_unpacker[object.format]
        unpacker(object)
    def unpack_dense(self, object):
        format = self.get_precision_format(object.precision,
                                           object.m * object.n)
        buffer = struct.unpack(format, self.recv_all(struct.calcsize(format)))
        # convert from column major to row major
        object.v = []
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            for i in range(object.m):
                object.v.append([])
                for j in range(object.n):
                    k = (i + object.m * j) * 2
                    object.v[-1].append(complex(buffer[k], buffer[k + 1]))
        else:
            for i in range(object.m):
                object.v.append([])
                for j in range(object.n):
                    object.v[-1].append(buffer[i + object.m * j])
    def unpack_crs(self, object):
        (object.nnz,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        format = self.get_precision_format(SILC_INT, object.m + 1)
        object.row = list(struct.unpack(
            format, self.recv_all(struct.calcsize(format))))
        self.send_ack()
        format = self.get_precision_format(SILC_INT, object.nnz)
        object.index = list(struct.unpack(
            format, self.recv_all(struct.calcsize(format))))
        self.send_ack()
        format = self.get_precision_format(object.precision, object.nnz)
        buffer = struct.unpack(format, self.recv_all(struct.calcsize(format)))
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            object.v = []
            for i in range(object.nnz):
                object.v.append(complex(buffer[i * 2], buffer[i * 2 + 1]))
        else:
            object.v = list(buffer)
    def unpack_band(self, object):
        (object.l,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        (object.u,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        ld = object.l + object.u + 1 # leading dimension
        format = self.get_precision_format(object.precision, ld * object.n)
        buffer = struct.unpack(format, self.recv_all(struct.calcsize(format)))
        # convert from column major to row major
        object.v = []
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            for i in range(ld):
                object.v.append([])
                for j in range(object.n):
                    k = (i + ld * j) * 2
                    object.v[-1].append(complex(buffer[k], buffer[k + 1]))
        else:
            for i in range(ld):
                object.v.append([])
                for j in range(object.n):
                    object.v[-1].append(buffer[i + ld * j])
    def unpack_cubic_array(self, object):
        (object.l,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        (object.m,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        (object.n,) = struct.unpack("!L", self.recv_all(self.sizeof_size))
        self.send_ack()
        format = self.get_precision_format(object.precision,
                                           object.l * object.m * object.n)
        buffer = struct.unpack(format, self.recv_all(struct.calcsize(format)))
        if object.precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
            def get(i, j, k):
                k = (object.m * object.n * i + j + object.m * k) * 2
                return complex(buffer[k], buffer[k + 1])
        else:
            def get(i, j, k):
                return buffer[object.m * object.n * i + j + object.m * k]
        # convert from column major to row major
        object.v = []
        for i in range(object.l):
            object.v.append([])
            for j in range(object.m):
                object.v[-1].append([])
                for k in range(object.n):
                    object.v[-1][-1].append(get(i, j, k))

_default_client = Client()

def SILC_INIT(host=None, port=None):
    _default_client.INIT(host, port)

def SILC_PUT(name, object):
    _default_client.PUT(name, object)

def SILC_EXEC(cmd):
    _default_client.EXEC(cmd)

def SILC_GET(object, name):
    _default_client.GET(object, name)

def SILC_FINALIZE():
    _default_client.FINALIZE()

def SILC_SHUTDOWN():
    _default_client.SHUTDOWN()

def SILC_SCALAR_NAME(object, dict={SILC_INT: "int",
                                   SILC_LONG: "long",
                                   SILC_FLOAT: "float",
                                   SILC_DOUBLE: "double",
                                   SILC_COMPLEX: "complex",
                                   SILC_DOUBLE_COMPLEX: "double complex"}):
    return dict.get(object.precision, "unknown")
