#!/usr/bin/env python
#
# Copyright (C) 2002-2007 The SSI Project. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the project nor the names of its contributors 
#       may be used to endorse or promote products derived from this software 
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# $Id: silc_cg.py,v 1.4 2007/10/31 06:19:47 kajiyama Exp $

from silc import *

import time

verbose = False

def silc_cg(n, value, row, index, b, x, maxiter, epsilon):
    object = Envelope()

    SILC_INIT()

    object.v = value
    object.type = SILC_MATRIX_TYPE
    object.precision = SILC_DOUBLE
    object.format = SILC_FORMAT_CRS
    object.m = object.n = n
    object.nnz = row[n]
    object.row = row
    object.index = index
    SILC_PUT("A", object)

    object.v = b
    object.type = SILC_COLUMN_VECTOR_TYPE
    object.precision = SILC_DOUBLE
    object.length = n
    SILC_PUT("b", object)

    t = time.time()
    SILC_EXEC("rho_old = 1.0")
    SILC_EXEC("n = length(b)")
    SILC_EXEC("p = zeros(n, 1)")
    SILC_EXEC("x = zeros(n, 1)")
    SILC_EXEC("r = b")
    SILC_EXEC("bnrm2 = 1.0 / norm2(b)")
    for i in range(maxiter):
        SILC_EXEC("rho = r' * r")
        SILC_EXEC("beta = rho / rho_old")
        SILC_EXEC("p = r + beta * p")
        SILC_EXEC("q = A * p")
        SILC_EXEC("alpha = rho / (p' * q)")
        SILC_EXEC("r = r - alpha * q")
        SILC_EXEC("nrm2 = norm2(r) * bnrm2")
        SILC_EXEC("x = x + alpha * p")
        # check convergence
        SILC_GET(object, "nrm2") 
        nrm2 = object.v
        if verbose:
            print "iter: %d, res: %.6e" % (i+1, nrm2)
        if nrm2 <= epsilon:
            break
        SILC_EXEC("rho_old = rho")
    else:
        i = -1
    print "solve: %.6f sec." % (time.time() - t)

    SILC_GET(object, "x")

    SILC_FINALIZE()

    return i+1

def main():
    import sys, getopt, math
    import mm

    USAGE = "Usage: %s [-v] [-C] [-X maxiter] filename.mtx\n" % sys.argv[0]
    try:
        options, args = getopt.getopt(sys.argv[1:], "vCX:")
    except getopt.error, e:
        sys.stderr.write(USAGE)
        sys.exit(1)
    rhs_cos = False
    maxiter = -1
    for opt, val in options:
        if opt == "-v":
            global verbose
            verbose = True
        elif opt == "-C":
            rhs_cos = True
        elif opt == "-X":
            maxiter = int(val)
    if len(args) != 1:
        sys.stderr.write(USAGE)
        sys.exit(1)

    # read a Matrix Market matrix
    file = open(args[0])
    header = mm.read_header(file)
    n = header.m; assert n == header.n
    value, row, index = mm.read_crs(file, header)
    print "%s: %d x %d matrix (%d elements)" % (args[0], n, n, row[n])
    if maxiter < 0:
        maxiter = n

    # create vectors b and x
    if rhs_cos:
        b = [math.cos(math.pi / (n - 1) * i) for i in range(n)]
    else:
        b = [0.0] * n
        for i in range(n):
            for j in range(row[i], row[i+1]):
                b[i] += value[j]
    x = [0.0] * n

    # solve Ax=b with the CG method
    t = time.time()
    num_iters = silc_cg(n, value, row, index, b, x, maxiter, 1.0e-12)
    elapsed = time.time() - t
    if num_iters < 0:
        print "reached to the maximum iteration count"
        num_iters = maxiter
    Mflop = 4.0 * n
    Mflop += num_iters * (2 * row[n] + 12 * n + 3)
    Mflop /= 1024 * 1024
    print "total: %.6f sec." % elapsed
    print "#iters =", num_iters
    print "%.6f Mflop, %.6f Mflops" % (Mflop, Mflop / elapsed)

if __name__ == "__main__":
    main()
