# Copyright (C) 2002-2007 The SSI Project. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the project nor the names of its contributors 
#       may be used to endorse or promote products derived from this software 
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# $Id: test_augmented.py,v 1.4 2007/10/31 06:19:47 kajiyama Exp $

import silctest

from silc import *
precision_table = [(int, SILC_INT),
                   (int, SILC_LONG),
                   (float, SILC_FLOAT),
                   (float, SILC_DOUBLE),
                   (complex, SILC_COMPLEX),
                   (complex, SILC_DOUBLE_COMPLEX)]

ADD_OP  = 1
SUB_OP  = 2
MUL_OP  = 3
DIV_OP  = 4
REM_OP  = 5
EMUL_OP = 6
EDIV_OP = 7

assignment = {
    ADD_OP: "+=",
    SUB_OP: "-=",
    MUL_OP: "*=", EMUL_OP: "*@=",
    DIV_OP: "/=", EDIV_OP: "/@=",
    REM_OP: "%="}

class ElementwiseTestCase(silctest.TestCase):
    def binop_add(x, y): return x + y
    def binop_sub(x, y): return x - y
    def binop_mul(x, y): return x * y
    def binop_div(x, y): return x / y
    binop = {
        ADD_OP: binop_add,
        SUB_OP: binop_sub,
        EMUL_OP: binop_mul,
        EDIV_OP: binop_div}
    def test_scalar_scalar(self):
        for op in [ADD_OP, SUB_OP, EMUL_OP, EDIV_OP]:
            if op in [EMUL_OP, EDIV_OP]:
                continue # XXX
            for conv, precision in precision_table:
                if precision in [SILC_INT, SILC_LONG]:
                    a, b = 5, 2
                elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                    a, b = 5.0, 2.0
                else:
                    a, b = 5+1j, 2+1j
                self.put_scalar("a", a, precision)
                self.put_scalar("b", b, precision)
                self.execute("a %s b" % assignment[op])
                obj = self.get("a")
                answer = self.binop[op](a, b)
                assert obj.v == answer
    def test_vector_vector(self):
        for vector_type in [SILC_COLUMN_VECTOR_TYPE, SILC_ROW_VECTOR_TYPE]:
            put_vector = self.put_method[vector_type]
            for op in [ADD_OP, SUB_OP, EMUL_OP, EDIV_OP]:
                if op in [EMUL_OP, EDIV_OP]:
                    continue # XXX
                binop = self.binop[op]
                for conv, precision in precision_table:
                    if precision in [SILC_INT, SILC_LONG]:
                        a = [2, 4, 6]
                        b = [1, 2, 3]
                    elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                        a = [2.0, 4.0, 6.0]
                        b = [1.0, 2.0, 3.0]
                    else:
                        a = [2+2j, 4+4j, 6+6j]
                        b = [1+1j, 2+2j, 3+3j]
                    put_vector("a", a, precision)
                    put_vector("b", b, precision)
                    self.execute("a %s b" % assignment[op])
                    obj = self.get("a")
                    answer = [binop(a[i], b[i]) for i in range(3)]
                    assert obj.v == answer
    def test_matrix_matrix(self):
        for op in [ADD_OP, SUB_OP, EMUL_OP, EDIV_OP]:
            if op in [EMUL_OP, EDIV_OP]:
                continue # XXX
            binop = self.binop[op]
            for conv, precision in precision_table:
                if precision in [SILC_INT, SILC_LONG]:
                    a = [[2, 4], [6, 8]]
                    b = [[1, 2], [3, 4]]
                elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                    a = [[2.0, 4.0], [6.0, 8.0]]
                    b = [[1.0, 2.0], [3.0, 4.0]]
                else:
                    a = [[2+2j, 4+4j], [6+6j, 8+8j]]
                    b = [[1+1j, 2+2j], [3+3j, 4+4j]]
                self.put_matrix("a", 2, 2, a, precision)
                self.put_matrix("b", 2, 2, b, precision)
                self.execute("a %s b" % assignment[op])
                answer = [[binop(a[i][j], b[i][j]) for j in range(2)] for i in range(2)]
                obj = self.get("a")
                assert obj.v == answer

class MultiplicationTestCase(silctest.TestCase):
    def test_scalar_scalar(self):
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                a, b = 5, 2
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                a, b = 5.0, 2.0
            else:
                a, b = 5+1j, 2+1j
            self.put_scalar("a", a, precision)
            self.put_scalar("b", b, precision)
            self.execute("a *= b")
            obj = self.get("a")
            assert obj.v == a * b
    def test_scalar_vector(self):
        for vector_type in [SILC_COLUMN_VECTOR_TYPE, SILC_ROW_VECTOR_TYPE]:
            put_vector = self.put_method[vector_type]
            for conv, precision in precision_table:
                if precision in [SILC_INT, SILC_LONG]:
                    vector = [1, 2, 3]
                    scalar = 2
                elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                    vector = [1.0, 2.0, 3.0]
                    scalar = 2.0
                else:
                    vector = [1+1j, 2+2j, 3+3j]
                    scalar = 2+2j
                put_vector("a", vector, precision)
                self.put_scalar("b", scalar, precision)
                self.execute("c = a; c *= b") # vector * scalar
                obj1 = self.get("c")
                self.execute("c = b; c *= a") # scalar * vector
                obj2 = self.get("c")
                answer = [vector[i] * scalar for i in range(3)]
                assert obj1.v == obj2.v == answer
    def test_scalar_matrix(self):
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                matrix = [[1, 2], [3, 4]]
                scalar = 2
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                matrix = [[1.0, 2.0], [3.0, 4.0]]
                scalar = 2.0
            else:
                matrix = [[1+1j, 2+2j], [3+3j, 4+4j]]
                scalar = 2+2j
            self.put_matrix("a", 2, 2, matrix, precision)
            self.put_scalar("b", scalar, precision)
            self.execute("c = a; c *= b") # matrix * scalar
            obj1 = self.get("c")
            self.execute("c = b; c *= a") # scalar * matrix
            obj2 = self.get("c")
            answer = [[matrix[i][j] * scalar for j in range(2)] for i in range(2)]
            assert obj1.v == obj2.v == answer
    def test_vector_vector(self):
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                r = [1, 2, 3]
                c = [2, 4, 6]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                r = [1.0, 2.0, 3.0]
                c = [2.0, 4.0, 6.0]
            else:
                r = [1+1j, 2+2j, 3+3j]
                c = [2+2j, 4+4j, 6+6j]
            self.put_row_vector("r", r, precision)
            self.put_column_vector("c", c, precision)
            self.execute("a = r; a *= c") # row * column => scalar
            obj = self.get("a")
            answer = conv(0)
            for i in range(3):
                answer = answer + r[i] * c[i]
            assert obj.v == answer
            self.execute("a = c; a *= r") # column * row => matrix
            obj = self.get("a")
            answer = [[c[i] * r[j] for j in range(3)] for i in range(3)]
            assert obj.v == answer
    def test_matrix_vector(self):
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                v = [1, 2, 3]
                m = [[2] * 3, [4] * 3, [6] * 3]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                v = [1.0, 2.0, 3.0]
                m = [[2.0] * 3, [4.0] * 3, [6.0] * 3]
            else:
                v = [1+1j, 2+2j, 3+3j]
                m = [[2+2j] * 3, [4+4j] * 3, [6+6j] * 3]
            self.put_row_vector("r", v, precision)
            self.put_column_vector("c", v, precision)
            self.put_matrix("m", 3, 3, m, precision)
            self.execute("a = r; a *= m") # row * matrix => row
            obj = self.get("a")
            answer = [conv(0)] * 3
            for i in range(3):
                for j in range(3):
                    answer[i] = answer[i] + v[j] * m[j][i]
            assert obj.v == answer and obj.type == SILC_ROW_VECTOR_TYPE
            self.execute("a = m; a *= c") # matrix * column => column
            obj = self.get("a")
            answer = [conv(0)] * 3
            for i in range(3):
                for j in range(3):
                    answer[i] = answer[i] + m[i][j] * v[j]
            assert obj.v == answer and obj.type == SILC_COLUMN_VECTOR_TYPE
    def test_matrix_matrix(self):
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                m32 = [[1, 2], [3, 4], [5, 6]]
                m23 = [[1, 2, 3], [4, 5, 6]]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                m32 = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]
                m23 = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
            else:
                m32 = [[1+1j, 2+2j], [3+3j, 4+4j], [5+5j, 6+6j]]
                m23 = [[1+1j, 2+2j, 3+3j], [4+4j, 5+5j, 6+6j]]
            self.put_matrix("m32", 3, 2, m32, precision)
            self.put_matrix("m23", 2, 3, m23, precision)
            self.execute("a = m32; a *= m23") # => 3x3 matrix
            obj = self.get("a")
            answer = [[conv(0) for j in range(3)] for i in range(3)]
            for i in range(3):
                for j in range(3):
                    for k in range(2):
                        answer[i][j] = answer[i][j] + m32[i][k] * m23[k][j]
            assert obj.v == answer
            self.execute("a = m23; a *= m32") # => 2x2 matrix
            obj = self.get("a")
            answer = [[conv(0) for j in range(2)] for i in range(2)]
            for i in range(2):
                for j in range(2):
                    for k in range(3):
                        answer[i][j] = answer[i][j] + m23[i][k] * m32[k][j]
            assert obj.v == answer

class DivisionTestCase(silctest.TestCase):
    def test_scalar_scalar(self):
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                a, b = 3, 2
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                a, b = 3.0, 2.0
            else:
                a, b = 3+3j, 2+2j
            self.put_scalar("a", a, precision)
            self.put_scalar("b", b, precision)
            self.execute("a /= b")
            obj = self.get("a")
            assert obj.v == a / b
    def test_vector_matrix(self):
        self.execute("prefer blasmodule")
        for conv, precision in precision_table:
            if precision in [SILC_INT, SILC_LONG]:
                continue # no solver for integer systems
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                matrix = [[ 2.0, -1.0,  0.0,  0.0],
                          [-1.0,  2.0, -1.0,  0.0],
                          [ 0.0, -1.0,  2.0, -1.0],
                          [ 0.0,  0.0, -1.0,  2.0]]
                vector = [-1.0, 0.0, 0.0, -1.0]
                answer = [-1.0, -1.0, -1.0, -1.0]
            else:
                matrix = [[ 2+0j, -1+0j,  0+0j,  0+0j],
                          [-1+0j,  2+0j, -1+0j,  0+0j],
                          [ 0+0j, -1+0j,  2+0j, -1+0j],
                          [ 0+0j,  0+0j, -1+0j,  2+0j]]
                vector = [-1+0j, 0+0j, 0+0j, -1+0j]
                answer = [-1+0j, -1+0j, -1+0j, -1+0j]
            self.put_matrix("a", 4, 4, matrix, precision)
            self.put_row_vector("b", vector, precision)
            self.execute("b /= a")
            obj = self.get("b")
            for i in range(4):
                assert abs(obj.v[i] - answer[i]) < 1e-6

class RemainderTestCase(silctest.TestCase):
    def runTest(self):
        for precision in [SILC_INT, SILC_LONG]:
            self.put_scalar("a", 8, precision)
            self.put_scalar("b", 5, precision)
            self.execute("a %= b")
            obj = self.get("a")
            assert obj.v == 3

if __name__ == "__main__":
    silctest.main()
