# Copyright (C) 2002-2007 The SSI Project. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the project nor the names of its contributors 
#       may be used to endorse or promote products derived from this software 
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
# $Id: test_concat.py,v 1.4 2007/10/31 06:19:47 kajiyama Exp $

import silctest

from silc import *
precision_table = [(int, SILC_INT),
                   (int, SILC_LONG),
                   (float, SILC_FLOAT),
                   (float, SILC_DOUBLE),
                   (complex, SILC_COMPLEX),
                   (complex, SILC_DOUBLE_COMPLEX)]

class ScalarConcatTestCase(silctest.TestCase):
    def runTest(self):
        for conv, precision in precision_table:
            if precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
                v = [1+1j, 2+1j, 3+1j, 4+1j, 5+1j, 6+1j]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                v = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
            else:
                v = [1, 2, 3, 4, 5, 6]
            for n, value in enumerate(v):
                self.put_scalar("s%d" % (n+1), value, precision)
            obj = self.evaluate("{s1;s2;s3;s4;s5;s6}")
            assert obj.v == v and obj.type == SILC_ROW_VECTOR_TYPE
            obj = self.evaluate("{s1,s2,s3,s4,s5,s6}")
            assert obj.v == v and obj.type == SILC_COLUMN_VECTOR_TYPE
            obj = self.evaluate("{s1,s2,s3;s4,s5,s6}")
            assert obj.v == [list(x) for x in zip(v[0:3], v[3:6])]
            obj = self.evaluate("{s1,s2;s3,s4;s5,s6}")
            assert obj.v == [list(x) for x in zip(v[0:2], v[2:4], v[4:6])]
            obj = self.evaluate("{{s1;s2;s3},{s4;s5;s6}}")
            assert obj.v == [v[0:3], v[3:6]]
            obj = self.evaluate("{{s1;s2},{s3;s4},{s5;s6}}")
            assert obj.v == [v[0:2], v[2:4], v[4:6]]

class RangeConcatTestCase(silctest.TestCase):
    def runTest(self):
        obj = self.evaluate("{1:3,4:6,7:9}")
        assert obj.type == SILC_COLUMN_VECTOR_TYPE and \
               obj.v == [1, 2, 3, 4, 5, 6, 7, 8, 9]
        obj = self.evaluate("{1:3;4:6;7:9}")
        assert obj.type == SILC_MATRIX_TYPE and \
               obj.v == [[1, 4, 7], [2, 5, 8], [3, 6, 9]]

class VectorConcatTestCase(silctest.TestCase):
    def runTest(self):
        for conv, precision in precision_table:
            if precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
                v1, v2 = [1+1j, 2+1j, 3+1j], [4+1j, 5+1j, 6+1j]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                v1, v2 = [1.0, 2.0, 3.0], [4.0, 5.0, 6.0]
            else:
                v1, v2 = [1, 2, 3], [4, 5, 6]
            for vector_type in [SILC_COLUMN_VECTOR_TYPE, SILC_ROW_VECTOR_TYPE]:
                put_vector = self.put_method[vector_type]
                put_vector("v1", v1, precision)
                put_vector("v2", v2, precision)
                obj = self.evaluate("{v1,v2}")
                if vector_type == SILC_COLUMN_VECTOR_TYPE:
                    assert obj.v == v1 + v2
                else:
                    assert obj.v == [v1, v2]
                obj = self.evaluate("{v1;v2}")
                if vector_type == SILC_COLUMN_VECTOR_TYPE:
                    assert obj.v == [list(x) for x in zip(v1, v2)]
                else:
                    assert obj.v == v1 + v2
                obj = self.evaluate("{v1,v2;v1,v2}")
                if vector_type == SILC_COLUMN_VECTOR_TYPE:
                    assert obj.v == [list(x) for x in zip(v1 + v2, v1 + v2)]
                else:
                    assert obj.v == [v1 + v1, v2 + v2]
                obj = self.evaluate("{{v1;v2},{v1;v2}}")
                if vector_type == SILC_COLUMN_VECTOR_TYPE:
                    assert obj.v == [list(x) for x in zip(v1 + v1, v2 + v2)]
                else:
                    assert obj.v == [v1 + v2, v1 + v2]

class MatrixConcatTestCase(silctest.TestCase):
    def runTest(self):
        for conv, precision in precision_table:
            if precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
                m1 = [[1+1j, 2+1j], [3+1j, 4+1j]]
                m2 = [[5+1j, 6+1j], [7+1j, 8+1j]]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                m1 = [[1.0, 2.0], [3.0, 4.0]]
                m2 = [[5.0, 6.0], [7.0, 8.0]]
            else:
                m1 = [[1, 2], [3, 4]]
                m2 = [[5, 6], [7, 8]]
            self.put_matrix("m1", 2, 2, m1, precision)
            self.put_matrix("m2", 2, 2, m2, precision)
            obj = self.evaluate("{m1,m2}")
            assert obj.v == m1 + m2
            obj = self.evaluate("{m1;m2}")
            assert obj.v == [m1[0] + m2[0], m1[1] + m2[1]]
            obj = self.evaluate("{m1,m2;m1,m2}")
            assert obj.v == [m1[0] + m1[0], m1[1] + m1[1],
                             m2[0] + m2[0], m2[1] + m2[1]]
            obj = self.evaluate("{{m1;m2},{m1;m2}}")
            assert obj.v == [m1[0] + m2[0], m1[1] + m2[1],
                             m1[0] + m2[0], m1[1] + m2[1]]

class ConcatWithModifiersVectorTestCase(silctest.TestCase):
    def test_matrix_column_vector(self):
        testset1 = [("{ m; v}",     0),
                    ("{-m;-v}",     1),
                    ("{ m'~, v'~}", 2),
                    ("{-m'~,-v'~}", 3),
                    ("{ m~; v~}",   4),
                    ("{-m~;-v~}",   5),
                    ("{ m', v'}",   6),
                    ("{-m',-v'}",   7)]
        testset2 = [("{ m[%d,:]; v[%d]; s}",       0),
                    ("{-m[%d,:];-v[%d];-s}",       1),
                    ("{ m[%d,:]'~, v[%d]'~, s'~}", 2),
                    ("{-m[%d,:]'~,-v[%d]'~,-s'~}", 3),
                    ("{ m[%d,:]~; v[%d]~; s~}",    4),
                    ("{-m[%d,:]~;-v[%d]~;-s~}",    5),
                    ("{ m[%d,:]', v[%d]', s'}",    6),
                    ("{-m[%d,:]',-v[%d]',-s'}",    7)]
        for conv, precision in precision_table:
            if precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
                m = [[1+1j, 4+1j], [2+1j, 5+1j], [3+1j, 6+1j]]
                v = [7+1j, 8+1j, 9+1j]
                s = 10+1j
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                m = [[1.0, 4.0], [2.0, 5.0], [3.0, 6.0]]
                v = [7.0, 8.0, 9.0]
                s = 10.0
            else:
                m = [[1, 4], [2, 5], [3, 6]]
                v = [7, 8, 9]
                s = 10
            self.put_matrix("m", 3, 2, m, precision)
            self.put_column_vector("v", v, precision)
            self.put_scalar("s", s, precision)
            for expr, flags in testset1:
                answer = [[m[i][0], m[i][1], v[i]] for i in range(3)]
                answer = self.modify_matrix(flags, answer)
                obj = self.evaluate(expr)
                assert obj.v == answer
            for expr, flags in testset2:
                for i in range(3):
                    answer = [m[i][0], m[i][1], v[i], s]
                    answer = self.modify_vector(flags, answer)
                    obj = self.evaluate(expr % (i+1, i+1))
                    assert obj.v == answer
    def test_matrix_row_vector(self):
        testset1 = [("{ m, v}",     0),
                    ("{-m,-v}",     1),
                    ("{ m'~; v'~}", 2),
                    ("{-m'~;-v'~}", 3),
                    ("{ m~, v~}",   4),
                    ("{-m~,-v~}",   5),
                    ("{ m'; v'}",   6),
                    ("{-m';-v'}",   7)]
        testset2 = [("{ m[:,%d], v[%d], s}",       0),
                    ("{-m[:,%d],-v[%d],-s}",       1),
                    ("{ m[:,%d]'~; v[%d]'~; s'~}", 2),
                    ("{-m[:,%d]'~;-v[%d]'~;-s'~}", 3),
                    ("{ m[:,%d]~, v[%d]~, s~}",    4),
                    ("{-m[:,%d]~,-v[%d]~,-s~}",    5),
                    ("{ m[:,%d]'; v[%d]'; s'}",    6),
                    ("{-m[:,%d]';-v[%d]';-s'}",    7)]
        for conv, precision in precision_table:
            if precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
                m = [[1+1j, 2+1j, 3+1j], [4+1j, 5+1j, 6+1j]]
                v = [7+1j, 8+1j, 9+1j]
                s = 10+1j
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                m = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
                v = [7.0, 8.0, 9.0]
                s = 10.0
            else:
                m = [[1, 2, 3], [4, 5, 6]]
                v = [7, 8, 9]
                s = 10
            self.put_matrix("m", 2, 3, m, precision)
            self.put_row_vector("v", v, precision)
            self.put_scalar("s", s, precision)
            for expr, flags in testset1:
                answer = [m[0], m[1], v]
                answer = self.modify_matrix(flags, answer)
                obj = self.evaluate(expr)
                assert obj.v == answer
            for expr, flags in testset2:
                for i in range(3):
                    answer = [m[0][i], m[1][i], v[i], s]
                    answer = self.modify_vector(flags, answer)
                    obj = self.evaluate(expr % (i+1, i+1))
                    assert obj.v == answer
    def modify_vector(self, flags, vector):
        return [self.modify(flags, x) for x in vector]
    def modify_matrix(self, flags, matrix):
        if flags & 2:
            matrix = [[matrix[j][i] for j in range(3)] for i in range(3)]
        return [[self.modify(flags, x) for x in row] for row in matrix]
    def modify(self, flags, scalar):
        if type(scalar) is complex and flags & 4:
            scalar = scalar.conjugate()
        if flags & 1:
            scalar = -scalar
        return scalar

class ConcatWithModifiersMatrixTestCase(silctest.TestCase):
    def runTest(self):
        for conv, precision in precision_table:
            if precision in [SILC_COMPLEX, SILC_DOUBLE_COMPLEX]:
                m = [[1+1j,2+1j,3+1j], [4+1j,5+1j,6+1j], [7+1j,8+1j,9+1j]]
                v = [10+1j,11+1j,12+1j,13+1j,14+1j]
            elif precision in [SILC_FLOAT, SILC_DOUBLE]:
                m = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
                v = [10.0, 11.0, 12.0, 13.0, 14.0]
            else:
                m = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
                v = [10, 11, 12, 13, 14]
            self.put_matrix("m", 3, 3, m, precision)
            self.put_row_vector("r", v, precision)
            self.put_column_vector("c", v, precision)
            testset1 = [("{ m[%d,:], r[%d:%d]}",     0),
                        ("{-m[%d,:],-r[%d:%d]}",     1),
                        ("{ m[%d,:]'~; r[%d:%d]'~}", 2),
                        ("{-m[%d,:]'~;-r[%d:%d]'~}", 3),
                        ("{ m[%d,:]~, r[%d:%d]~}",   4),
                        ("{-m[%d,:]~,-r[%d:%d]~}",   5),
                        ("{ m[%d,:]'; r[%d:%d]'}",   6),
                        ("{-m[%d,:]';-r[%d:%d]'}",   7)]
            testset2 = [("{ m[:,%d]; c[%d:%d]}",     0),
                        ("{-m[:,%d];-c[%d:%d]}",     1),
                        ("{ m[:,%d]'~, c[%d:%d]'~}", 2),
                        ("{-m[:,%d]'~,-c[%d:%d]'~}", 3),
                        ("{ m[:,%d]~; c[%d:%d]~}",   4),
                        ("{-m[:,%d]~;-c[%d:%d]~}",   5),
                        ("{ m[:,%d]', c[%d:%d]'}",   6),
                        ("{-m[:,%d]',-c[%d:%d]'}",   7)]
            for expr, flags in testset1:
                for i in range(3):
                    answer = [m[i], v[i:i+3]]
                    answer = self.modify_matrix(flags, 2, 3, answer)
                    obj = self.evaluate(expr % (i+1, i+1, i+3))
                    assert obj.v == answer
            for expr, flags in testset2:
                for i in range(3):
                    answer = [[m[j][i], v[i+j]] for j in range(3)]
                    answer = self.modify_matrix(flags, 3, 2, answer)
                    obj = self.evaluate(expr % (i+1, i+1, i+3))
                    assert obj.v == answer
    def modify_matrix(self, flags, m, n, matrix):
        if flags & 2:
            matrix = [[matrix[i][j] for i in range(m)] for j in range(n)]
        return [[self.modify(flags, x) for x in row] for row in matrix]
    def modify(self, flags, scalar):
        if type(scalar) is complex and flags & 4:
            scalar = scalar.conjugate()
        if flags & 1:
            scalar = -scalar
        return scalar

if __name__ == "__main__":
    silctest.main()
