/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: silc.c,v 1.4 2007/10/31 06:22:07 kajiyama Exp $
 */

/* a simple SILC script runner */

/* todo: fully implement the `save' command */
/* todo: add support for sparse matrix (CCS) */
/* todo: add support for 3-dimensional array */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "client.h"

#define SILC_MAXNAMELEN 63 /* defined in ../const.h */

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

char **
split_args(char *buf, int *argc)
{
  int len;
  char **argv;

  *argc = 0;
  while (*buf == ' ')
    buf++;
  if (*buf == '\0')
    return NULL;
  argv = (char **)malloc(sizeof(char *));
  while (*buf) {
    len = 0;
    while (buf[len] && buf[len] != ' ')
      len++;
    if (*argc == 0)
      argv = (char **)malloc(sizeof(char *));
    else
      argv = (char **)realloc(argv, sizeof(char *) * (*argc + 1));
    argv[*argc] = (char *)malloc(len + 1);
    strncpy(argv[*argc], buf, len);
    argv[*argc][len] = '\0';
    *argc += 1;
    buf += len;
    while (*buf == ' ')
      buf++;
  }
  return argv;
}

void
free_args(int argc, char *argv[])
{
  int i;

  for (i = 0; i < argc; i++)
    free(argv[i]);
  free(argv);
}

int byte_swapping;

unsigned short
unpack_int16(char *p)
{
  char buf[8];

  if (byte_swapping) {
    buf[0] = p[1]; buf[1] = p[0];
    p = buf;
  }
  return *(unsigned short *)p;
}

unsigned int
unpack_int32(char *p)
{
  char buf[8];

  if (byte_swapping) {
    buf[0] = p[3]; buf[1] = p[2]; buf[2] = p[1]; buf[3] = p[0];
    p = buf;
  }
  return *(unsigned int *)p;
}

unsigned long
unpack_int64(char *p)
{
  char buf[8];

  if (byte_swapping) {
    buf[0] = p[7]; buf[1] = p[6]; buf[2] = p[5]; buf[3] = p[4];
    buf[4] = p[3]; buf[5] = p[2]; buf[6] = p[1]; buf[7] = p[0];
    p = buf;
  }
  return *(unsigned long *)p;
}

float
unpack_single(char *p)
{
  char buf[8];

  if (byte_swapping) {
    buf[0] = p[3]; buf[1] = p[2]; buf[2] = p[1]; buf[3] = p[0];
    p = buf;
  }
  return *(float *)p;
}

double
unpack_double(char *p)
{
  char buf[8];

  if (byte_swapping) {
    buf[0] = p[7]; buf[1] = p[6]; buf[2] = p[5]; buf[3] = p[4];
    buf[4] = p[3]; buf[5] = p[2]; buf[6] = p[1]; buf[7] = p[0];
    p = buf;
  }
  return *(double *)p;
}

#define miINT8    1
#define miUINT8   2
#define miINT16   3
#define miUINT16  4
#define miINT32   5
#define miUINT32  6
#define miSINGLE  7
#define miDOUBLE  9
#define miINT64   12
#define miUINT64  13
#define miMATRIX  14

#define mxINT32_CLASS   12
#define mxSINGLE_CLASS  7
#define mxDOUBLE_CLASS  6
#define mxSPARSE_CLASS  5

typedef struct {
  int num_elements;
  struct _element {
    char name[SILC_MAXNAMELEN+1];
    silc_envelope_t object;
  } *element;
} mat_file_t;

void
free_mat_file(mat_file_t *mat_file)
{
  int i;

  if (mat_file->num_elements > 0) {
    for (i = 0; i < mat_file->num_elements; i++)
      if (mat_file->element[i].object.v)
	free(mat_file->element[i].object.v);
    free(mat_file->element);
  }
  free(mat_file);
}

int
read_tag(FILE *fp, int *type, int *size)
{
  char tag[8];

  if (fread(tag, sizeof(char), 8, fp) < 8)
    return -1;
  *type = unpack_int32(tag);
  *size = unpack_int32(tag + 4);
  return 0;
}

void *
read_array(FILE *fp, int precision, int size)
{
  void *v;
  char buf[64];
  int i, j;
  int type, element_size;

  switch (precision) {
  case SILC_INT:
    v = (void *)malloc(sizeof(int) * size);
    {
      int *p = (int *)v;
      read_tag(fp, &type, &element_size);
      if (type != miINT32 || element_size / sizeof(int) != size)
	goto onError;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(int), 1, fp);
	p[i] = unpack_int32(buf);
      }
    }
    break;
  case SILC_FLOAT:
    v = (void *)malloc(sizeof(float) * size);
    {
      float *p = (float *)v;
      read_tag(fp, &type, &element_size);
      if (type != miSINGLE || element_size / sizeof(float) != size)
	goto onError;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(float), 1, fp);
	p[i] = unpack_single(buf);
      }
      if (size % 2)
	fread(buf, sizeof(float), 1, fp); /* padding */
    }
    break;
  case SILC_DOUBLE:
    v = (void *)malloc(sizeof(double) * size);
    {
      double *p = (double *)v;
      read_tag(fp, &type, &element_size);
      if (type != miDOUBLE || element_size / sizeof(double) != size)
	goto onError;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(double), 1, fp);
	p[i] = unpack_double(buf);
      }
    }
    break;
  case SILC_COMPLEX:
    v = (void *)malloc(sizeof(float) * size * 2);
    {
      float *p = (float *)v;
      read_tag(fp, &type, &element_size); /* real part */
      if (type != miSINGLE || element_size / sizeof(float) != size)
	goto onError;
      j = 0;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(float), 1, fp);
	p[j] = unpack_single(buf);
	j += 2;
      }
      if (size % 2)
	fread(buf, sizeof(float), 1, fp); /* padding */
      read_tag(fp, &type, &element_size); /* imaginary part */
      if (type != miSINGLE || element_size / sizeof(float) != size)
	goto onError;
      j = 1;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(float), 1, fp);
	p[j] = unpack_single(buf);
	j += 2;
      }
      if (size % 2)
	fread(buf, sizeof(float), 1, fp); /* padding */
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    v = (void *)malloc(sizeof(double) * size * 2);
    {
      double *p = (double *)v;
      read_tag(fp, &type, &element_size); /* real part */
      if (type != miDOUBLE || element_size / sizeof(double) != size)
	goto onError;
      j = 0;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(double), 1, fp);
	p[j] = unpack_double(buf);
	j += 2;
      }
      read_tag(fp, &type, &element_size); /* imaginary part */
      if (type != miDOUBLE || element_size / sizeof(double) != size)
	goto onError;
      j = 1;
      for (i = 0; i < size; i++) {
	fread(buf, sizeof(double), 1, fp);
	p[j] = unpack_double(buf);
	j += 2;
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  return v;

 onError:
  free(v);
  fprintf(stderr, "read_array: failed to read data\n");
  return NULL;
}

mat_file_t *
read_mat_file(char *filename)
{
  FILE *fp;
  char buf[128];
  int type, size, i, j;
  mat_file_t *mat_file;
  int is_complex, array_type, m, n, nnz;
  char array_name[SILC_MAXNAMELEN+1];
  int precision;

  fp = fopen(filename, "rb");
  if (!fp)
    return NULL;
  mat_file = (mat_file_t *)malloc(sizeof(mat_file_t));
  mat_file->num_elements = 0;
  if (fread(buf, sizeof(char), 128, fp) < 128) {
    fprintf(stderr, "read_mat_file: cannot read MAT file header\n");
    goto onError;
  }
  if (strncmp(buf, "MATL", 4)) {
    fprintf(stderr, "read_mat_file: not a MATLAB 5.0 MAT-file\n");
    goto onError;
  }
  if (strncmp(buf + 126, "MI", 2) == 0)
    byte_swapping = 0;
  else if (strncmp(buf + 126, "IM", 2) == 0)
    byte_swapping = 1;
  else {
    fprintf(stderr, "read_mat_file: invalid endian indicator\n");
    goto onError;
  }
  if (unpack_int16(buf + 124) != 0x0100) {
    fprintf(stderr, "read_mat_file: invalid version number\n");
    goto onError;
  }
  for (;;) {
    if (read_tag(fp, &type, &size) < 0)
      break;
    if (type != miMATRIX) {  /* XXX */
      fseek(fp, size, SEEK_CUR);
      continue;
    }
    read_tag(fp, &type, &size);
    if (type == miUINT32 && size == 8) {
      int flags;
      fread(buf, sizeof(char), 8, fp);
      flags = unpack_int32(buf);
      is_complex = flags & 0x800;
      array_type = flags & 0xff;
      if (array_type == mxSPARSE_CLASS)
	nnz = unpack_int32(buf + 4);
    } else {
      fprintf(stderr, "read_mat_file: unsupported Array Flags\n");
      goto onError;
    }
    read_tag(fp, &type, &size);
    if (type == miINT32 && size == 8) { /* XXX only 2-dimentional array */
      fread(buf, sizeof(char), 8, fp);
      m = (int)unpack_int32(buf);
      n = (int)unpack_int32(buf + 4);
    } else {
      fprintf(stderr, "read_mat_file: unsupported Dimensions Array\n");
      goto onError;
    }
    fread(buf, sizeof(char), 8, fp);
    type = unpack_int32(buf);
    if (type & 0xffff0000) { /* small data element */
      size = type >> 16;
      type = type & 0xffff;
      if (type != miINT8) {
	fprintf(stderr, "read_mat_file: unsupported Array Name\n");
	goto onError;
      } else if (size > SILC_MAXNAMELEN) {
	fprintf(stderr, "read_mat_file: too long name: %d\n", size);
	goto onError;
      }
      strncpy(array_name, buf + 4, size);
      array_name[size] = '\0';
    } else {
      size = unpack_int32(buf + 4);
      if (type != miINT8) {
	fprintf(stderr, "read_mat_file: unsupported Array Name\n");
	goto onError;
      } else if (size > SILC_MAXNAMELEN) {
	fprintf(stderr, "read_mat_file: too long name: %d\n", size);
	goto onError;
      }
      fread(array_name, sizeof(char), size, fp);
      array_name[size] = '\0';
      size %= 8;
      if (size)
	fread(buf, sizeof(char), 8 - size, fp); /* padding */
    }
    if (array_type == mxSPARSE_CLASS) {
      precision = -1; /* unknown yet */
    } else if (array_type == mxINT32_CLASS) {
      precision = SILC_INT;
    } else if (array_type == mxSINGLE_CLASS && !is_complex) {
      precision = SILC_FLOAT;
    } else if (array_type == mxDOUBLE_CLASS && !is_complex) {
      precision = SILC_DOUBLE;
    } else if (array_type == mxSINGLE_CLASS && is_complex) {
      precision = SILC_COMPLEX;
    } else if (array_type == mxDOUBLE_CLASS && is_complex) {
      precision = SILC_DOUBLE_COMPLEX;
    } else {
      fprintf(stderr, "read_mat_file: unsupported precision\n");
      goto onError;
    }
    i = mat_file->num_elements++;
    if (mat_file->num_elements == 1)
      mat_file->element = (struct _element *)malloc(sizeof(struct _element));
    else
      mat_file->element = (struct _element *)realloc
	(mat_file->element, mat_file->num_elements * sizeof(struct _element));
    mat_file->element[i].object.v = NULL;
    strcpy(mat_file->element[i].name, array_name);
    if (array_type == mxSPARSE_CLASS) {
      fprintf(stderr, "read_mat_file: mxSPARSE_CLASS not implemented yet\n");
      goto onError;
    } else if (m > 1 && n > 1) {
      mat_file->element[i].object.type = SILC_MATRIX_TYPE;
      mat_file->element[i].object.format = SILC_FORMAT_DENSE;
      mat_file->element[i].object.precision = precision;
      mat_file->element[i].object.m = m;
      mat_file->element[i].object.n = n;
      mat_file->element[i].object.v = read_array(fp, precision, m * n);
    } else if (m > 1 && n == 1) {
      mat_file->element[i].object.type = SILC_COLUMN_VECTOR_TYPE;
      mat_file->element[i].object.precision = precision;
      mat_file->element[i].object.length = m;
      mat_file->element[i].object.v = read_array(fp, precision, m);
    } else if (m == 1 && n > 1) {
      mat_file->element[i].object.type = SILC_ROW_VECTOR_TYPE;
      mat_file->element[i].object.precision = precision;
      mat_file->element[i].object.length = n;
      mat_file->element[i].object.v = read_array(fp, precision, n);
    } else if (m == 1 && n == 1) {
      mat_file->element[i].object.type = SILC_SCALAR_TYPE;
      mat_file->element[i].object.precision = precision;
      mat_file->element[i].object.v = read_array(fp, precision, 1);
    } else {
      fprintf(stderr, "read_mat_file: unexpected dimensions (%d, %d)\n", m, n);
      goto onError;
    }
  }
  return mat_file;

 onError:
  fclose(fp);
  free_mat_file(mat_file);
  return NULL;
}

int
load(char *buf)
{
  int argc, num_names, i, j;
  char **argv, *filename, **name;
  int opt_ascii, opt_mat;
  mat_file_t *mat_file;

  argv = split_args(buf, &argc);
  if (!argv) {
    fprintf(stderr, "load: cannot parse command line\n");
    return -1;
  }
  if (argc < 2) {
    fprintf(stderr, "load: insufficient arguments\n");
    goto onError;
  }
  filename = argv[1];
  name = argv + 2;
  num_names = 0;
  for (i = 2; i < argc; i++) {
    if (argv[i][0] == '-')
      break;
    num_names++;
  }
  opt_ascii = opt_mat = 0;
  for (; i < argc; i++) {
    if (strcmp(argv[i], "-ascii") == 0)
      opt_ascii = 1;
    else if (strcmp(argv[i], "-mat") == 0)
      opt_mat = 1;
    else {
      fprintf(stderr, "load: unknown option: %s\n", argv[i]);
      goto onError;
    }
  }
#ifdef DEBUG
  printf("load: \"%s\"", filename);
  if (opt_ascii)  printf(" -ascii");
  if (opt_mat)    printf(" -mat");
  printf("\n");
  for (i = 0; i < num_names; i++)
    printf("  %d: \"%s\"\n", i, name[i]);
#endif
  mat_file = read_mat_file(filename);
  if (!mat_file) {
    fprintf(stderr, "load: failed to read data from %s\n", filename);
    goto onError;
  }
  if (num_names > 0) {
    /* check if all of user-specified names are defined */
    for (i = 0; i < num_names; i++) {
      for (j = 0; j < mat_file->num_elements; j++)
	if (strcmp(name[i], mat_file->element[j].name) == 0)
	  break;
      if (j == mat_file->num_elements) {
	fprintf(stderr, "load: \"%s\" is not in \"%s\"\n", name[i], filename);
	free_mat_file(mat_file);
	goto onError;
      }
    }
    /* put only the data with the specified names */
    for (i = 0; i < num_names; i++) {
      for (j = 0; j < mat_file->num_elements; j++) {
	if (strcmp(name[i], mat_file->element[j].name) == 0) {
	  SILC_PUT(name[i], &mat_file->element[j].object);
	  break;
	}
      }
    }
  } else {
    /* put all data in the MAT-file */
    for (j = 0; j < mat_file->num_elements; j++)
      SILC_PUT(mat_file->element[j].name, &mat_file->element[j].object);
  }
  free_mat_file(mat_file);
  free_args(argc, argv);
  return 0;

 onError:
  free_args(argc, argv);
  return -1;
}

int
save(char *buf)
{
  int argc, i, num_names;
  char **argv, *filename;
  int opt_ascii, opt_double, opt_tabs, opt_append, opt_v4;

  argv = split_args(buf, &argc);
  if (!argv) {
    fprintf(stderr, "save: cannot parse command line\n");
    return -1;
  }
  if (argc < 2) {
    fprintf(stderr, "save: insufficient arguments\n");
    goto onError;
  }
  filename = argv[1];
  num_names = 0;
  for (i = 2; i < argc; i++) {
    if (argv[i][0] == '-')
      break;
    num_names++;
  }
  opt_ascii = opt_double = opt_tabs = opt_append = opt_v4 = 0;
  for (; i < argc; i++) {
    if (strcmp(argv[i], "-ascii") == 0)
      opt_ascii = 1;
    else if (strcmp(argv[i], "-double") == 0)
      opt_double = 1;
    else if (strcmp(argv[i], "-tabs") == 0)
      opt_tabs = 1;
    else if (strcmp(argv[i], "-append") == 0)
      opt_append = 1;
    else if (strcmp(argv[i], "-v4") == 0)
      opt_v4 = 1;
    else {
      fprintf(stderr, "save: unknown option: %s\n", argv[i]);
      goto onError;
    }
  }
#ifdef DEBUG
  printf("save: \"%s\"", filename);
  if (opt_ascii)  printf(" -ascii");
  if (opt_double) printf(" -double");
  if (opt_tabs)   printf(" -tabs");
  if (opt_append) printf(" -append");
  if (opt_v4)     printf(" -v4");
  printf("\n");
  for (i = 0; i < num_names; i++)
    printf("  %d: \"%s\"\n", i, argv[i+2]);
#endif

  free_args(argc, argv);
  return 0;

 onError:
  free_args(argc, argv);
  return -1;
}

int
exec(char *buf)
{
  return SILC_EXEC(buf);
}

int
main(int argc, char *argv[])
{
  FILE *fp;
  char buf[1024]; /* SILC_PROGRAM_MAX + 1 */
  char *p;
  int i, stat;
  
  if (argc != 2) {
    p = strrchr(argv[0], '/');
    fprintf(stderr, "Usage: %s <script>\n", p ? p + 1 : argv[0]);
    return -1;
  }
  fp = fopen(argv[1], "r");
  if (fp == NULL) {
    fprintf(stderr, "cannot open %s\n", argv[1]);
    return -1;
  }
  SILC_INIT();
  while (fgets(buf, sizeof(buf), fp)) {
    if (buf[0] == '%') /* skip comment */
      continue;
    for (p = buf; *p; p++) {
      if (*p == '\n') {
	*p = '\0';
	break;
      }
      if (*p == '\t')
	*p = ' ';
    }
    for (p = buf; *p == ' '; p++)
      ;
    if (*p == '\0') /* empty line and comment */
      continue;
    else if (strncmp(p, "save ", 5) == 0)
      stat = save(p);
    else if (strncmp(p, "load ", 5) == 0)
      stat = load(p);
    else
      stat = exec(p);
    if (stat < 0)
      break;
  }
  fclose(fp);
  printf("Hit ENTER to exit...\n");
  fgets(buf, sizeof(buf), stdin);
  SILC_FINALIZE();
  return 0;
}
