/* 
 *  Copyright (c) 2006 Tamito KAJIYAMA.  All rights reserved.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation, and that the name of Tamito KAJIYAMA not be
 *  used in advertising or publicity pertaining to distribution of the
 *  software without specific, written prior permission.
 *
 *  TAMITO KAJIYAMA DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 *  EVENT SHALL TAMITO KAJIYAMA BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 *  USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 *  PERFORMANCE OF THIS SOFTWARE.
 *
 *  $Id: mp_bicg.c,v 1.1 2007/08/25 22:47:06 kajiyama Exp $
 */

#include "mp_crs.h"

int
mp_bicg(mp_crs_t *A, mp_vector_t *b, mp_vector_t *x,
        int maxiter, mpf_t epsilon, int verbose)
{
  mp_vector_t *p, *q, *r, *pp, *qq, *rr;
  mpf_t alpha, beta, rho, rho_old, bnrm2, nrm2, zero, one, tmp;
  int n, iter;

  n = mp_crs_n(A);
  mpf_init_set_ui(zero, 0);
  mpf_init_set_ui(one, 1);
  mpf_init(tmp);

  /* rho_old = 1 */
  mpf_init_set(rho_old, one);

  /* p = pp = {0, ..., 0} */
  p = mp_vector_new(n);
  mp_vector_set(p, zero);
  pp = mp_vector_new(n);
  mp_vector_set(pp, zero);

  /* r = rr = b */
  r = mp_vector_new(n);
  mp_vector_copy(r, b);
  rr = mp_vector_new(n);
  mp_vector_copy(rr, b);

  /* bnrm2 = 1 / norm2(b) */
  mpf_init(bnrm2);
  mp_vector_norm2(tmp, b);
  mpf_div(bnrm2, one, tmp);

  q = mp_vector_new(n);
  qq = mp_vector_new(n);
  mpf_init(alpha);
  mpf_init(beta);
  mpf_init(rho);
  mpf_init(nrm2);
  for (iter = 1; iter <= maxiter; iter++) { 
    /* rho = r' * rr */
    mp_vector_dot(rho, r, rr);

    /* check breakdown */
    if (mpf_cmp(rho, zero) == 0)
      break;

    /* beta = rho / rho_old */
    mpf_div(beta, rho, rho_old);

    /* p = r + beta * p; pp = rr + beta * pp */
    mp_vector_xpay(r, beta, p);
    mp_vector_xpay(rr, beta, pp);

    /* q = A * p; qq = A' * pp */
    mp_crs_matvec(one, A, p, zero, q);
    mp_crs_tmatvec(one, A, pp, zero, qq);

    /* alpha = rho / (pp' * q) */
    mp_vector_dot(tmp, pp, q);
    mpf_div(alpha, rho, tmp);

    /* r = r - alpha * q; rr = rr - alpha * qq */
    mpf_neg(tmp, alpha);
    mp_vector_axpy(tmp, q, r);
    mp_vector_axpy(tmp, qq, rr);

    /* x = x + alpha * p */
    mp_vector_axpy(alpha, p, x);

    /* nrm2 = norm2(r) * bnrm2 */
    mp_vector_norm2(tmp, r);
    mpf_mul(nrm2, tmp, bnrm2);
#ifdef USE_HISTORY
    if (mp_residual_history)
      mpf_set(mp_vector_value(mp_residual_history, iter-1), nrm2);
#endif

    /* check convergence */
    if (verbose)
      gmp_printf("iter: %d, res: %.6Fe\n", iter, nrm2);
    if (mpf_cmp(nrm2, epsilon) <= 0) /* nrm2 <= epsilon */
      break;

    mpf_set(rho_old, rho);
  }
  mp_vector_free(p); mp_vector_free(pp);
  mp_vector_free(q); mp_vector_free(qq);
  mp_vector_free(r); mp_vector_free(rr);
  mpf_clear(alpha);
  mpf_clear(beta);
  mpf_clear(rho);
  mpf_clear(rho_old);
  mpf_clear(bnrm2);
  mpf_clear(nrm2);
  mpf_clear(zero);
  mpf_clear(one);
  mpf_clear(tmp);
  return iter;
}
