/* 
 *  Copyright (c) 2006 Tamito KAJIYAMA.  All rights reserved.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation, and that the name of Tamito KAJIYAMA not be
 *  used in advertising or publicity pertaining to distribution of the
 *  software without specific, written prior permission.
 *
 *  TAMITO KAJIYAMA DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 *  EVENT SHALL TAMITO KAJIYAMA BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 *  USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 *  PERFORMANCE OF THIS SOFTWARE.
 *
 *  $Id: mp_crs.c,v 1.1 2007/08/25 22:47:06 kajiyama Exp $
 */

#include <malloc.h>

#ifdef _OPENMP
#include <omp.h>
#else
#define omp_get_max_threads()  (1)
#define omp_get_thread_num()   (0)
#endif

#include "mp_crs.h"

mp_crs_t *
mp_crs_new(int m, int n, int nnz)
{
  mp_crs_t *p;
  int i;

  p = (mp_crs_t *)malloc(sizeof(mp_crs_t));
  if (!p)
    return NULL;
  p->value = (mpf_t *)malloc(sizeof(mpf_t) * nnz);
  if (!p->value) {
    free(p);
    return NULL;
  }
  p->row = (int *)malloc(sizeof(int) * (m + 1));
  if (!p->row) {
    free(p->value);
    free(p);
    return NULL;
  }
  p->index = (int *)malloc(sizeof(int) * nnz);
  if (!p->index) {
    free(p->value);
    free(p->row);
    free(p);
    return NULL;
  }
#pragma omp parallel for
  for (i = 0; i < nnz; i++)
    mpf_init(p->value[i]);
  p->m = m;
  p->n = n;
  p->nnz = nnz;
  return p;
}

int
mp_crs_free(mp_crs_t *p)
{
  int i;

#pragma omp parallel for
  for (i = 0; i < p->nnz; i++)
    mpf_clear(p->value[i]);
  free(p->value);
  free(p->row);
  free(p->index);
  free(p);
  return 0;
}

mp_vector_t *
mp_vector_new(int length)
{
  mp_vector_t *p;
  int i;

  p = (mp_vector_t *)malloc(sizeof(mp_vector_t));
  if (!p)
    return NULL;
  p->value = (mpf_t *)malloc(sizeof(mpf_t) * length);
  if (!p->value) {
    free(p);
    return NULL;
  }
#pragma omp parallel for
  for (i = 0; i < length; i++)
    mpf_init(p->value[i]);
  p->length = length;
  return p;
}

int
mp_vector_free(mp_vector_t *p)
{
  int i;

#pragma omp parallel for
  for (i = 0; i < p->length; i++)
    mpf_clear(p->value[i]);
  free(p->value);
  free(p);
  return 0;
}

int
mp_vector_copy(mp_vector_t *rop, mp_vector_t *op)
{
  int i;

  if (rop->length != op->length)
    return -1;
#pragma omp parallel for
  for (i = 0; i < op->length; i++)
    mpf_set(rop->value[i], op->value[i]);
  return 0;
}

int
mp_vector_set(mp_vector_t *op1, mpf_t op2)
{
  int i;

#pragma omp parallel for
  for (i = 0; i < op1->length; i++)
    mpf_set(op1->value[i], op2);
  return 0;
}

/* y = alpha * x + y */
int
mp_vector_axpy(mpf_t alpha, mp_vector_t *x, mp_vector_t *y)
{
  int i;

  if (x->length != y->length)
    return -1;
#pragma omp parallel
  {
    mpf_t tmp;

    mpf_init(tmp);
#pragma omp for
    for (i = 0; i < x->length; i++) {
      mpf_mul(tmp, alpha, x->value[i]);
      mpf_add(y->value[i], y->value[i], tmp);
    }
    mpf_clear(tmp);
  }
  return 0;
}

/* y = x + alpha * y */
int
mp_vector_xpay(mp_vector_t *x, mpf_t alpha, mp_vector_t *y)
{
  int i;

  if (x->length != y->length)
    return -1;
#pragma omp parallel for
  for (i = 0; i < x->length; i++) {
    mpf_mul(y->value[i], y->value[i], alpha);
    mpf_add(y->value[i], y->value[i], x->value[i]);
  }
  return 0;
}

int
mp_vector_dot(mpf_t rop, mp_vector_t *op1, mp_vector_t *op2)
{
  mpf_t sum[128];
  int i, p;

  if (sizeof(sum) / sizeof(mpf_t) < omp_get_max_threads())
    return -1;
  if (op1->length != op2->length)
    return -1;
#pragma omp parallel private(p)
  {
    mpf_t tmp;

    p = omp_get_thread_num();
    mpf_init(sum[p]);
    mpf_init(tmp);
#pragma omp for
    for (i = 0; i < op1->length; i++) {
      mpf_mul(tmp, op1->value[i], op2->value[i]);
      mpf_add(sum[p], sum[p], tmp);
    }
  }
  mpf_set(rop, sum[0]);
  p = omp_get_max_threads();
  for (i = 1; i < p; i++)
    mpf_add(rop, rop, sum[i]);
#pragma omp for
  for (i = 0; i < p; i++)
    mpf_clear(sum[i]);
  return 0;
}

int
mp_vector_norm2(mpf_t rop, mp_vector_t *op)
{
  mpf_t tmp;

  mpf_init(tmp);
  if (mp_vector_dot(tmp, op, op) < 0) {
    mpf_clear(tmp);
    return -1;
  }
  mpf_sqrt(rop, tmp);
  mpf_clear(tmp);
  return 0;
}

/* y = alpha * A * x + beta * y */
int
mp_crs_matvec(mpf_t alpha, mp_crs_t *A, mp_vector_t *x, mpf_t beta,
              mp_vector_t *y)
{
  int i;

  if (A->n != x->length || x->length != y->length)
    return -1;
#pragma omp parallel
  {
    mpf_t tmp;

    mpf_init(tmp);
#pragma omp for
    for (i = 0; i < A->m; i++) {
      int j;

      mpf_mul(y->value[i], y->value[i], beta);
      for (j = A->row[i]; j < A->row[i+1]; j++) {
        mpf_mul(tmp, A->value[j], alpha);
        mpf_mul(tmp, x->value[A->index[j]], tmp);
        mpf_add(y->value[i], y->value[i], tmp);
      }
    }
    mpf_clear(tmp);
  }
  return 0;
}

/* y = alpha * A' * x + beta * y */
int
mp_crs_tmatvec(mpf_t alpha, mp_crs_t *A, mp_vector_t *x, mpf_t beta,
               mp_vector_t *y)
{
  mp_vector_t *z, *buf[128];
  int i, j, p;

  if (sizeof(buf) / sizeof(mp_vector_t *) < omp_get_max_threads())
    return -1;
  if (A->m != x->length || x->length != y->length)
    return -1;
#pragma omp parallel private(j, p, z)
  {
    mpf_t tmp;
    int k;

    p = omp_get_thread_num();
    if (p == 0)
      z = y;
    else
      z = buf[p] = mp_vector_new(y->length);
    mpf_init(tmp);
#pragma omp for
    for (i = 0; i < A->m; i++)
      mpf_mul(y->value[i], y->value[i], beta);
#pragma omp for
    for (i = 0; i < A->m; i++) {
      for (j = A->row[i]; j < A->row[i+1]; j++) {
        k = A->index[j];
        mpf_mul(tmp, A->value[j], alpha);
        mpf_mul(tmp, x->value[i], tmp);
        mpf_add(z->value[k], z->value[k], tmp);
      }
    }
    mpf_clear(tmp);
  }
  p = omp_get_max_threads();
  for (i = 1; i < p; i++) {
    z = buf[i];
    for (j = 0; j < A->n; j++)
      mpf_add(y->value[j], y->value[j], z->value[j]);
    mp_vector_free(z);
  }
  return 0;
}

#ifdef USE_HISTORY

mp_vector_t *mp_residual_history;

int
mp_init_residual_history(int maxiter)
{
  mp_free_residual_history();
  mp_residual_history = mp_vector_new(maxiter);
  if (!mp_residual_history)
    return -1;
  return 0;
}

int
mp_free_residual_history()
{
  if (mp_residual_history)
    mp_vector_free(mp_residual_history);
  mp_residual_history = NULL;
  return 0;
} 

#endif /* USE_HISTORY */
