/* 
 *  Copyright (c) 2006 Tamito KAJIYAMA.  All rights reserved.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted,
 *  provided that the above copyright notice appear in all copies and that
 *  both that copyright notice and this permission notice appear in
 *  supporting documentation, and that the name of Tamito KAJIYAMA not be
 *  used in advertising or publicity pertaining to distribution of the
 *  software without specific, written prior permission.
 *
 *  TAMITO KAJIYAMA DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 *  EVENT SHALL TAMITO KAJIYAMA BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 *  USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 *  PERFORMANCE OF THIS SOFTWARE.
 *
 *  $Id: mp_solve_test.c,v 1.1 2007/08/25 22:47:06 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> /* for getopt() */

#ifdef _OPENMP
#include <omp.h>
#endif

#include "mp_crs.h"

int mp_bicg(mp_crs_t *, mp_vector_t *, mp_vector_t *, int, mpf_t, int);
int mp_cg(mp_crs_t *, mp_vector_t *, mp_vector_t *, int, mpf_t, int);
int mp_cgnr(mp_crs_t *, mp_vector_t *, mp_vector_t *, int, mpf_t, int);

#define TOEPLITZ 1

#ifdef _WIN32

#include <windows.h>

static LARGE_INTEGER freq;

double wtime(void)
{
  LARGE_INTEGER now;

  if (freq.QuadPart == 0)
    QueryPerformanceFrequency(&freq);
  QueryPerformanceCounter(&now);
  return (double)now.QuadPart / (double)freq.QuadPart;
}

#else /* _WIN32 */

#include <stdio.h>
#include <sys/time.h>

double wtime(void)
{
  struct timeval now;

  gettimeofday(&now, NULL);
  return now.tv_sec + (double)now.tv_usec * 1.0e-6;
}

#endif /* _WIN32 */

int
main(int argc, char *argv[])
{
  mp_crs_t *A;
  mp_vector_t *b, *x;
  mpf_t one, zero, epsilon;
  int i, j, n, nnz, opt, prec, method, iter, maxiter, verbose;
  double gamma;
  char *name;
  double t, elapsed;

#ifdef _OPENMP
  printf("omp_get_max_threads() => %d\n", omp_get_max_threads());
#endif

#define USAGE "Usage: %s [-g gamma] [-m method] [-p precision] [-x maxiter] [-v] dimension\nMethod: BiCG, CG, CGNR\n"

  gamma = 1.3;
  method = 1;
  prec = mpf_get_default_prec();
  maxiter = -1;
  verbose = 0;
  while ((opt = getopt(argc, argv, "g:m:p:x:v")) != EOF) {
    switch (opt) {
    case 'g':
      gamma = atof(optarg);
      break;
    case 'm':
      if (strcmp(optarg, "BiCG") == 0)
        method = 1;
      else if (strcmp(optarg, "CG") == 0)
        method = 2;
      else if (strcmp(optarg, "CGNR") == 0)
        method = 3;
      else {
        printf("unknown method: %s\n", optarg);
        return -1;
      }
      break;
    case 'p':
      prec = atoi(optarg);
      break;
    case 'x':
      maxiter = atoi(optarg);
      break;
    case 'v':
      verbose = 1;
      break;
    default:
      printf(USAGE, argv[0]);
      return -1;
    }
  }
  if (optind == argc) {
    printf(USAGE, argv[0]);
    return -1;
  }
  n = atoi(argv[optind]);
  if (n < 3) {
    printf("dimension must be >= 3\n");
    return -1;
  }    
  if (maxiter < 0)
    maxiter = n;

  mpf_set_default_prec(prec);
  printf("precision: at least %d bits\n", mpf_get_default_prec());

#ifdef TOEPLITZ
  A = mp_crs_new(n, n, 3 * n - 3);
#else
  A = mp_crs_new(n, n, 3 * n - 2);
#endif
  b = mp_vector_new(n);
  x = mp_vector_new(n);
  mpf_init_set_d(epsilon, 1.0e-12);
  nnz = 0;
  for (i = 0; i < n; i++) {
    mp_crs_row(A, i) = nnz;
#ifdef TOEPLITZ
    /*
     * A = | 2  1             |
     *     | 0  2  1          |
     *     | g  0  2  1       |
     *     |    g  0  2  .    |
     *     |       .  .  .  1 |
     *     |          g  0  2 |
     */
    if (i >= 2) {
      mpf_set_d(mp_crs_value(A, nnz), gamma);
      mp_crs_index(A, nnz) = i - 2;
      nnz++;
    }
    mpf_set_d(mp_crs_value(A, nnz), 2.0);
    mp_crs_index(A, nnz) = i;
    nnz++;
    if (i < n - 1) {
      mpf_set_d(mp_crs_value(A, nnz), 1.0);
      mp_crs_index(A, nnz) = i + 1;
      nnz++;
    }
#else
    /*
     * A = |  2 -1          |
     *     | -1  2 -1       |
     *     |    -1  2  .    |
     *     |        .  . -1 |
     *     |          -1  2 |
     */
    if (i > 0) {
      mpf_set_d(mp_crs_value(A, nnz), -1.0);
      mp_crs_index(A, nnz) = i - 1;
      nnz++;
    }
    mpf_set_d(mp_crs_value(A, nnz), 2.0);
    mp_crs_index(A, nnz) = i;
    nnz++;
    if (i < n - 1) {
      mpf_set_d(mp_crs_value(A, nnz), -1.0);
      mp_crs_index(A, nnz) = i + 1;
      nnz++;
    }
#endif /* TOEPLITZ */
  }
  mp_crs_row(A, n) = nnz;
  mpf_init_set_ui(one, 1);
  mpf_init_set_ui(zero, 0);
  mp_vector_set(x, one);
  mp_crs_matvec(one, A, x, zero, b);
  mp_vector_set(x, zero);
#if 0
  for (i = 0; i < n; i++)
    for (j = mp_crs_row(A, i); j < mp_crs_row(A, i+1); j++)
      gmp_printf("A[%d,%d] = %Fe\n", i, mp_crs_index(A,j), mp_crs_value(A,j));
  for (i = 0; i < n; i++)
    gmp_printf("b[%d] = %Fe\n", i, mp_vector_value(b,i));
#endif

  t = wtime();
  switch (method) {
  case 1:
    name = "BiCG";
    iter = mp_bicg(A, b, x, maxiter, epsilon, verbose);
    break;
  case 2:
    name = "CG";
    iter = mp_cg(A, b, x, maxiter, epsilon, verbose);
    break;
  case 3:
    name = "CGNR";
    iter = mp_cgnr(A, b, x, maxiter, epsilon, verbose);
    break;
  default:
    printf("unknown method: %d\n", method);
  }
  elapsed = wtime() - t;
  printf("%s: #iters = %d\n", name, iter);
  printf("%s: %.6lf sec\n", name, elapsed);
  if (iter > maxiter)
    printf("%s: reached to the maximum iteration count\n", name);

  mp_crs_free(A);
  mp_vector_free(b);
  mp_vector_free(x);
  mpf_clear(zero);
  mpf_clear(one);
  mpf_clear(epsilon);
  return 0;
}
