/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: blasmodule.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

/* todo: take x_mods and y_mods into consideration */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

#include "module.h"
#include "dense.h"
#include "sparse_crs.h"

#define MODULE_NAME "blasmodule"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#define _max(x, y) (((x) > (y)) ? (x) : (y))
#define _min(x, y) (((x) < (y)) ? (x) : (y))

void call_dgemv(char, int, int, double, double *, int, double *, int,
		double, double *, int);
void call_zgemv(char, int, int, double *, double *, int, double *, int,
		double *, double *, int);

void call_sgetrf(int, int, float *, int, int *, int *);
void call_dgetrf(int, int, double *, int, int *, int *);
void call_cgetrf(int, int, float *, int, int *, int *);
void call_zgetrf(int, int, double *, int, int *, int *);
void call_sgetrs(char, int, int, float *, int, int *, float *, int, int *);
void call_dgetrs(char, int, int, double *, int, int *, double *, int, int *);
void call_cgetrs(char, int, int, float *, int, int *, float *, int, int *);
void call_zgetrs(char, int, int, double *, int, int *, double *, int, int *);
void call_sgecon(char, int, float *, int, float, float *, float *,
		 int *, int *);
void call_dgecon(char, int, double *, int, double, double *, double *,
		 int *, int *);
void call_cgecon(char, int, float *, int, float, float *, float *,
		 float *, int *);
void call_zgecon(char, int, double *, int, double, double *, double *,
		 double *, int *);
float call_sdot(int, float *, int, float *, int);
double call_ddot(int, double *, int, double *, int);
void call_sgesvd(char, char, int, int, float *, int, float *, float *,
		 int, float *, int, float *, int, int *);
void call_dgesvd(char, char, int, int, double *, int, double *, double *,
		 int, double *, int, double *, int, int *);
void call_cgesvd(char, char, int, int, float *, int, float *, float *,
		 int, float *, int, float *, int, float *, int *);
void call_zgesvd(char, char, int, int, double *, int, double *, double *,
		 int, double *, int, double *, int, double *, int *);

int
blasmodule_double_mat_vec_product(int num_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  double *dv;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": double_mat_vec_product(0x%lx, 0x%lx)\n", x, y);
#endif
  dv = (double *)malloc(sizeof(double) * y->d.vector.length);
  call_dgemv('N', x->d.matrix.m, x->d.matrix.n, 1.0, SILC_DENSE(x)->v.d,
	     x->d.matrix.m, y->d.vector.v.d, 1, 0.0, dv, 1);
  free(y->d.vector.v.d);
  y->d.vector.v.d = dv;
  return 0;
}

int
blasmodule_doublecomplex_mat_vec_product(int num_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  double *zv;
  double alpha[2] = {1.0, 0.0}, beta[2] = {0.0, 0.0};

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME": doublecomplex_mat_vec_product(0x%lx, 0x%lx)\n", x, y);
#endif
  zv = (double *)malloc(sizeof(double) * 2 * y->d.vector.length);
  call_zgemv('N', x->d.matrix.m, x->d.matrix.n, alpha, SILC_DENSE(x)->v.z,
	     x->d.matrix.m, y->d.vector.v.z, 1, beta, zv, 1);
  free(y->d.vector.v.z);
  y->d.vector.v.z = zv;
  return 0;
}

static int
solve_1(silc_object_t *A, silc_object_t *b)
{
  silc_dense_t *A_in;
  silc_crs_t *in;
  silc_format_t *table;
  int i, j, k, m, n, *ipiv, info;
  const char *format_name;

#ifdef DEBUG
  printf(MODULE_NAME ": solve_1(0x%lx, 0x%lx)\n", A, b);
#endif
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return -1;
  }
  if (A->d.matrix.m != b->d.vector.length) {
    fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
    return -1;
  }
  format_name = silc_find_format_name(A->format);
  if (strcmp(format_name, SILC_FORMAT_DENSE) == 0) {
    A = silc_object_duplicate(A, 1);
    if (!A)
      return -1;
    A_in = SILC_DENSE(A);
  } else if (strcmp(format_name, SILC_FORMAT_CRS) == 0) {
    in = SILC_CRS(A);
    A = silc_object_new_matrix(A->precision, 
			       silc_find_format_id(SILC_FORMAT_DENSE),
			       A->d.matrix.m,
			       A->d.matrix.n);
    table = silc_find_format_proc_table(A->format);
    if (!table)
      return -1;
    if (!table->proc_new) {
      fprintf(stderr, MODULE_NAME ": proc_new() not defined");
      return -1;
    }
    if (table->proc_new(A) < 0) {
      fprintf(stderr, MODULE_NAME ": proc_new() failed");
      return -1;
    }
    A_in = SILC_DENSE(A);
    switch (A->precision) {
    case SILC_FLOAT:
      for (i = 0; i < A->d.matrix.m; i++) {
	for (j = in->row[i]; j < in->row[i + 1]; j++) {
	  A_in->v.s[i + in->index[j] * A->d.matrix.m] = in->v.s[j];
	}
      }
      break;
    case SILC_DOUBLE:
      for (i = 0; i < A->d.matrix.m; i++) {
	for (j = in->row[i]; j < in->row[i + 1]; j++) {
	  A_in->v.d[i + in->index[j] * A->d.matrix.m] = in->v.d[j];
	}
      }
      break;
    case SILC_COMPLEX:
      for (i = 0; i < A->d.matrix.m; i++) {
	for (j = in->row[i]; j < in->row[i + 1]; j++) {
	  k = i + in->index[j] * A->d.matrix.m;
	  A_in->v.c[k * 2]     = in->v.c[j * 2];
	  A_in->v.c[k * 2 + 1] = in->v.c[j * 2 + 1];
	}
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      for (i = 0; i < A->d.matrix.m; i++) {
	for (j = in->row[i]; j < in->row[i + 1]; j++) {
	  k = i + in->index[j] * A->d.matrix.m;
	  A_in->v.z[k * 2]     = in->v.z[j * 2];
	  A_in->v.z[k * 2 + 1] = in->v.z[j * 2 + 1];
	}
      }
      break;
    default:
      should_not_reach_here;
    }
  }
  m = A->d.matrix.m;
  n = A->d.matrix.n;
  ipiv = (int *)malloc(sizeof(int) * ((m < n) ? m : n));
  if (!ipiv) {
    fprintf(stderr, MODULE_NAME ": out of memory (ipiv)");
    silc_object_free(A);
    return -1;
  }
  switch (A->precision) {
  case SILC_FLOAT:
    call_sgetrf(m, n, A_in->v.s, m, ipiv, &info); break;
  case SILC_DOUBLE:
    call_dgetrf(m, n, A_in->v.d, m, ipiv, &info); break;
  case SILC_COMPLEX:
    call_cgetrf(m, n, A_in->v.c, m, ipiv, &info); break;
  case SILC_DOUBLE_COMPLEX:
    call_zgetrf(m, n, A_in->v.z, m, ipiv, &info); break;
  default:
    should_not_reach_here;
  }
  if (info == 0) {
    switch (A->precision) {
    case SILC_FLOAT:
      call_sgetrs('N', n, 1, A_in->v.s, m, ipiv,
		  b->d.vector.v.s, b->d.vector.length, &info);
      break;
    case SILC_DOUBLE:
      call_dgetrs('N', n, 1, A_in->v.d, m, ipiv,
		  b->d.vector.v.d, b->d.vector.length, &info);
      break;
    case SILC_COMPLEX:
      call_cgetrs('N', n, 1, A_in->v.c, m, ipiv,
		  b->d.vector.v.c, b->d.vector.length, &info);
      break;
    case SILC_DOUBLE_COMPLEX:
      call_zgetrs('N', n, 1, A_in->v.z, m, ipiv,
		  b->d.vector.v.z, b->d.vector.length, &info);
      break;
    default:
      should_not_reach_here;
    }
  }
  free(ipiv);
  silc_object_free(A);
  return (info == 0) ? 0 : -1;
}

/* out: 1 */
int
blasmodule_solve(int n_params, silc_mfunc_param_t *param)
{
  return solve_1(param[0].o, param[1].o);
}

/* out: 0 */
int
blasmodule_divide(int n_params, silc_mfunc_param_t *param)
{
  return solve_1(param[1].o, param[0].o);
}

/* out: 2 */
int
blasmodule_dot_product(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *new_object;
  silc_modifier_t *x_mods, *y_mods;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": dot_product(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_FLOAT:
    param[2].o = silc_object_new_float(call_sdot(x->d.vector.length,
						 x->d.vector.v.s, 1,
						 y->d.vector.v.s, 1));
    break;
  case SILC_DOUBLE:
    param[2].o = silc_object_new_double(call_ddot(x->d.vector.length,
						  x->d.vector.v.d, 1,
						  y->d.vector.v.d, 1));
    break;
  default:
    should_not_reach_here;
  }
  return (param[2].o) ? 0 : -1;
}

/* rcond - the reciprocal of the condition of a matrix */

int
blasmodule_func_rcond(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  int i, j, k, m, n, *ipiv, info;
  const char *format_name;
  union {
    float *s; double *d;
    float *c; double *z;
  } v;
  union {
    float s;
    double d;
  } norm, rcond;

  object = param[0].o;
#ifdef DEBUG
  printf("blasmodule_func_rcond(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  format_name = silc_find_format_name(object->format);
  if (strcmp(format_name, SILC_FORMAT_DENSE) &&
      strcmp(format_name, SILC_FORMAT_CRS)) {
    /* XXX need a workaround */
    fprintf(stderr, MODULE_NAME ": rcond: neigher in dense nor in CRS\n");
    return -1;
  }
  if (object->d.matrix.m != object->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": rcond: not a square matrix\n");
    return -1;
  }
  if (object->precision == SILC_INT ||
      object->precision == SILC_LONG) {
    fprintf(stderr, MODULE_NAME ": rcond: integer matrix is not supported\n");
    return -1;
  }

  m = object->d.matrix.m;
  n = object->d.matrix.n;
  if (strcmp(format_name, SILC_FORMAT_DENSE) == 0) {
    silc_dense_t *in;

    in = SILC_DENSE(object);
    switch (object->precision) {
    case SILC_FLOAT:
      {
	float sum;
	v.s = (float *)malloc(sizeof(float) * m * n);
	if (!v.s) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.s)");
	  return -1;
	}
	memcpy(v.s, in->v.s, sizeof(float) * m * n);
	norm.s = 0.0;
	for (j = 0; j < n; j++) {
	  sum = 0.0;
	  for (i = 0; i < m; i++)
	    sum += fabs(in->v.s[i + j * m]);
	  if (norm.s < sum)
	    norm.s = sum;
	}
      }
      break;
    case SILC_DOUBLE:
      {
	double sum;
	v.d = (double *)malloc(sizeof(double) * m * n);
	if (!v.d) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.d)");
	  return -1;
	}
	memcpy(v.d, in->v.d, sizeof(double) * m * n);
	norm.d = 0.0;
	for (j = 0; j < n; j++) {
	  sum = 0.0;
	  for (i = 0; i < m; i++)
	    sum += fabs(in->v.d[i + j * m]);
	  if (norm.d < sum)
	    norm.d = sum;
	}
      }
      break;
    case SILC_COMPLEX:
      {
	float re, im, sum;
	v.c = (float *)malloc(sizeof(float) * m * n * 2);
	if (!v.c) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.c)");
	  return -1;
	}
	memcpy(v.c, in->v.c, sizeof(float) * m * n * 2);
	norm.s = 0.0;
	for (j = 0; j < n; j++) {
	  sum = 0.0;
	  for (i = 0; i < m; i++) {
	    k = i + j * m * 2;
	    re = in->v.c[k * 2];
	    im = in->v.c[k * 2 + 1];
	    sum += sqrt(re * re + im * im);
	  }
	  if (norm.s < sum)
	    norm.s = sum;
	}
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double re, im, sum;
	v.z = (double *)malloc(sizeof(double) * m * n * 2);
	if (!v.z) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.z)");
	  return -1;
	}
	memcpy(v.z, in->v.z, sizeof(double) * m * n * 2);
	norm.d = 0.0;
	for (j = 0; j < n; j++) {
	  sum = 0.0;
	  for (i = 0; i < m; i++) {
	    k = i + j * m;
	    re = in->v.z[k * 2];
	    im = in->v.z[k * 2 + 1];
	    sum += sqrt(re * re + im * im);
	  }
	  if (norm.d < sum)
	    norm.d = sum;
	}
      }
      break;
    default:
      should_not_reach_here;
    }
  } else if (strcmp(format_name, SILC_FORMAT_CRS) == 0) {
    silc_crs_t *in;

    in = SILC_CRS(object);
    switch (object->precision) {
    case SILC_FLOAT:
      {
	float *sum;
	v.s = (float *)calloc(m * n, sizeof(float));
	if (!v.s) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.s)");
	  return -1;
	}
	sum = (float *)calloc(n, sizeof(float));
	for (i = 0; i < m; i++) {
	  for (j = in->row[i]; j < in->row[i + 1]; j++) {
	    v.s[i + m * in->index[j]] = in->v.s[j];
	    sum[in->index[j]] += fabs(in->v.s[j]);
	  }
	}
	norm.s = sum[0];
	for (i = 1; i < n; i++) {
	  if (norm.s < sum[i])
	    norm.s = sum[i];
	}
	free(sum);
      }
      break;
    case SILC_DOUBLE:
      {
	double *sum;
	v.d = (double *)calloc(m * n, sizeof(double));
	if (!v.d) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.d)");
	  return -1;
	}
	sum = (double *)calloc(n, sizeof(double));
	for (i = 0; i < m; i++) {
	  for (j = in->row[i]; j < in->row[i + 1]; j++) {
	    v.d[i + m * in->index[j]] = in->v.d[j];
	    sum[in->index[j]] += fabs(in->v.d[j]);
	  }
	}
	norm.d = sum[0];
	for (i = 1; i < n; i++) {
	  if (norm.d < sum[i])
	    norm.d = sum[i];
	}
	free(sum);
      }
      break;
    case SILC_COMPLEX:
      {
	float re, im, *sum;
	v.c = (float *)calloc(m * n * 2, sizeof(float));
	if (!v.c) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.c)");
	  return -1;
	}
	sum = (float *)calloc(n, sizeof(float));
	for (i = 0; i < m; i++) {
	  for (j = in->row[i]; j < in->row[i + 1]; j++) {
	    k = i + m * in->index[j];
	    re = v.c[k * 2]     = in->v.c[j * 2];
	    im = v.c[k * 2 + 1] = in->v.c[j * 2 + 1];
	    sum[in->index[j]] += sqrt(re * re + im * im);
	  }
	}
	norm.s = sum[0];
	for (i = 1; i < n; i++) {
	  if (norm.s < sum[i])
	    norm.s = sum[i];
	}
	free(sum);
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double re, im, *sum;
	v.z = (double *)calloc(m * n * 2, sizeof(double));
	if (!v.z) {
	  fprintf(stderr, MODULE_NAME ": rcond: out of memory (v.z)");
	  return -1;
	}
	sum = (double *)calloc(n, sizeof(double));
	for (i = 0; i < m; i++) {
	  for (j = in->row[i]; j < in->row[i + 1]; j++) {
	    k = i + m * in->index[j];
	    re = v.z[k * 2]     = in->v.z[j * 2];
	    im = v.z[k * 2 + 1] = in->v.z[j * 2 + 1];
	    sum[in->index[j]] += sqrt(re * re + im * im);
	  }
	}
	norm.d = sum[0];
	for (i = 1; i < n; i++) {
	  if (norm.d < sum[i])
	    norm.d = sum[i];
	}
	free(sum);
      }
      break;
    default:
      should_not_reach_here;
    }
  }
#ifdef DEBUG
  switch (object->precision) {
  case SILC_FLOAT:
  case SILC_COMPLEX:
    printf(MODULE_NAME ": rcond: anorm = %e\n", norm.s); break;
  case SILC_DOUBLE:
  case SILC_DOUBLE_COMPLEX:
    printf(MODULE_NAME ": rcond: anorm = %le\n", norm.d); break;
  default:
    should_not_reach_here;
  }
#endif
  ipiv = (int *)malloc(sizeof(int) * ((m < n) ? m : n));
  if (!ipiv) {
    fprintf(stderr, MODULE_NAME ": rcond: out of memory (ipiv)");
    free(v.s);
    return -1;
  }
  /* form an LU factorization of an M-by-N full matrix */
  switch (object->precision) {
  case SILC_FLOAT:
    call_sgetrf(m, n, v.s, m, ipiv, &info); break;
  case SILC_DOUBLE:
    call_dgetrf(m, n, v.d, m, ipiv, &info); break;
  case SILC_COMPLEX:
    call_cgetrf(m, n, v.c, m, ipiv, &info); break;
  case SILC_DOUBLE_COMPLEX:
    call_zgetrf(m, n, v.z, m, ipiv, &info); break;
  default:
    should_not_reach_here;
  }
  free(ipiv);
  if (info == 0) {
    /* estimate the reciprocal of the condition of the matrix in 1-norm */
    switch (object->precision) {
    case SILC_FLOAT:
      {
	float *work;
	int *iwork;
	work = (float *)malloc(sizeof(float) * 4 * n);
	iwork = (int *)malloc(sizeof(int) * n);
	call_sgecon('1', n, v.s, m, norm.s, &rcond.s, work, iwork, &info);
	free(work);
	free(iwork);
      }
      new_object = silc_object_new_float(rcond.s);
      break;
    case SILC_DOUBLE:
      {
	double *work;
	int *iwork;
	work = (double *)malloc(sizeof(double) * 4 * n);
	iwork = (int *)malloc(sizeof(int) * n);
	call_dgecon('1', n, v.d, m, norm.d, &rcond.d, work, iwork, &info);
	free(work);
	free(iwork);
      }
      new_object = silc_object_new_double(rcond.d);
      break;
    case SILC_COMPLEX:
      {
	float *work, *rwork;
	work = (float *)malloc(sizeof(float) * 4 * n);
	rwork = (float *)malloc(sizeof(float) * 2 * n);
	call_cgecon('1', n, v.c, m, norm.s, &rcond.s, work, rwork, &info);
	free(work);
	free(rwork);
      }
      new_object = silc_object_new_float(rcond.s);
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double *work, *rwork;
	work = (double *)malloc(sizeof(double) * 4 * n);
	rwork = (double *)malloc(sizeof(double) * 2 * n);
	call_zgecon('1', n, v.z, m, norm.d, &rcond.d, work, rwork, &info);
	free(work);
	free(rwork);
      }
      new_object = silc_object_new_double(rcond.d);
      break;
    default:
      should_not_reach_here;
    }
  }
  free(v.s);
  if (info != 0)
    return -1;
  param[1].o = new_object;
  return 0;
}

int
blasmodule_proc_svd(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *u, *s, *v;
  silc_dense_t *x_in, *u_in, *s_in, *v_in;
  int i, j, k, m, n, lwork, info;
  const char *format_name;

  x = param[0].o;
#ifdef DEBUG
  printf("blasmodule_proc_svd(0x%lx, U, S, V)\n", x);
#endif
  if (x->type != SILC_MATRIX_TYPE)
    return -1;
  format_name = silc_find_format_name(x->format);
  if (strcmp(format_name, SILC_FORMAT_DENSE)) {
    fprintf(stderr, MODULE_NAME ": svd: not a dense matrix\n");
    return -1;
  }
  m = x->d.matrix.m;
  n = x->d.matrix.n;
  x_in = SILC_DENSE(x);
  switch (x->precision) {
  case SILC_INT:
  case SILC_FLOAT:
    u = silc_object_new_matrix(SILC_FLOAT, x->format, m, m);
    s = silc_object_new_matrix(SILC_FLOAT, x->format, m, n);
    v = silc_object_new_matrix(SILC_FLOAT, x->format, n, n);
    break;
  case SILC_LONG:
  case SILC_DOUBLE:
    u = silc_object_new_matrix(SILC_DOUBLE, x->format, m, m);
    s = silc_object_new_matrix(SILC_DOUBLE, x->format, m, n);
    v = silc_object_new_matrix(SILC_DOUBLE, x->format, n, n);
    break;
  case SILC_COMPLEX:
    u = silc_object_new_matrix(SILC_COMPLEX, x->format, m, m);
    s = silc_object_new_matrix(SILC_FLOAT,   x->format, m, n);
    v = silc_object_new_matrix(SILC_COMPLEX, x->format, n, n);
    break;
  case SILC_DOUBLE_COMPLEX:
    u = silc_object_new_matrix(SILC_DOUBLE_COMPLEX, x->format, m, m);
    s = silc_object_new_matrix(SILC_DOUBLE,         x->format, m, n);
    v = silc_object_new_matrix(SILC_DOUBLE_COMPLEX, x->format, n, n);
    break;
  default:
    should_not_reach_here;
  }
  u_in = silc_dense_allocate(u->precision, m, m);
  s_in = silc_dense_allocate(s->precision, m, n);
  v_in = silc_dense_allocate(v->precision, n, n);
  u->d.matrix.in = (void *)u_in;
  s->d.matrix.in = (void *)s_in;
  v->d.matrix.in = (void *)v_in;
  switch (x->precision) {
  case SILC_INT:
  case SILC_FLOAT:
    {
      float *a, *sv, *vt, *work;

      a = (float *)malloc(sizeof(float) * m * n);
      if (!a) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (a)");
	return -1;
      }
      switch (x->precision) {
      case SILC_INT:
#pragma omp parallel for private(i)
	for (j = 0; j < n; j++)
	  for (i = 0; i < m; i++)
	    a[i + m * j] = x_in->v.i[i + m * j];
	break;
      case SILC_FLOAT:
#pragma omp parallel for private(i)
	for (j = 0; j < n; j++)
	  for (i = 0; i < m; i++)
	    a[i + m * j] = x_in->v.s[i + m * j];
	break;
      default:
	should_not_reach_here;
      }
      sv = (float *)malloc(sizeof(float) * _min(m, n));
      if (!sv) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (sv)");
	return -1;
      }
      vt = (float *)malloc(sizeof(float) * n * n);
      if (!vt) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (vt)");
	return -1;
      }
      lwork = _max(3 * _min(m, n) + _max(m, n), 5 * _min(m, n));
      work = (float *)malloc(sizeof(float) * lwork);
      if (!work) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (work)");
	return -1;
      }
      call_sgesvd('A', 'A', m, n, a, m, sv, u_in->v.s, m, vt, n,
		  work, lwork, &info);
#pragma omp parallel for
      for (j = 0; j < n; j++)
	memset(s_in->v.s + m * j, 0, sizeof(float) * m);
      k = _min(m, n);
#pragma omp parallel for
      for (j = 0; j < k; j++)
	s_in->v.s[j + m * j] = sv[j];
#pragma omp parallel for private(i)
      for (j = 0; j < n; j++) {
	for (i = 0; i < n; i++)
	  v_in->v.s[j + n * i] = vt[i + n * j];
      }
      free(a);
      free(sv);
      free(vt);
      free(work);
    }
    break;
  case SILC_LONG:
  case SILC_DOUBLE:
    {
      double *a, *sv, *vt, *work;

      a = (double *)malloc(sizeof(double) * m * n);
      if (!a) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (a)");
	return -1;
      }
      switch (x->precision) {
      case SILC_LONG:
#pragma omp parallel for private(i)
	for (j = 0; j < n; j++)
	  for (i = 0; i < m; i++)
	    a[i + m * j] = x_in->v.l[i + m * j];
	break;
      case SILC_DOUBLE:
#pragma omp parallel for private(i)
	for (j = 0; j < n; j++)
	  for (i = 0; i < m; i++)
	    a[i + m * j] = x_in->v.d[i + m * j];
	break;
      default:
	should_not_reach_here;
      }
      sv = (double *)malloc(sizeof(double) * _min(m, n));
      if (!sv) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (sv)");
	return -1;
      }
      vt = (double *)malloc(sizeof(double) * n * n);
      if (!vt) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (vt)");
	return -1;
      }
      lwork = _max(3 * _min(m, n) + _max(m, n), 5 * _min(m, n));
      work = (double *)malloc(sizeof(double) * lwork);
      if (!work) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (work)");
	return -1;
      }
      call_dgesvd('A', 'A', m, n, a, m, sv, u_in->v.d, m, vt, n,
		  work, lwork, &info);
#pragma omp parallel for
      for (j = 0; j < n; j++)
	memset(s_in->v.d + m * j, 0, sizeof(double) * m);
      k = _min(m, n);
#pragma omp parallel for
      for (j = 0; j < k; j++)
	s_in->v.d[j + m * j] = sv[j];
#pragma omp parallel for private(i)
      for (j = 0; j < n; j++) {
	for (i = 0; i < n; i++)
	  v_in->v.d[j + n * i] = vt[i + n * j];
      }
      free(a);
      free(sv);
      free(vt);
      free(work);
    }
    break;
  case SILC_COMPLEX:
    {
      float *a, *sv, *vt, *work, *rwork;

      a = (float *)malloc(sizeof(float) * 2 * m * n);
      if (!a) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (a)");
	return -1;
      }
#pragma omp parallel for
      for (j = 0; j < n * 2; j += 2)
	memcpy(a + m * j, x_in->v.c + m * j, sizeof(float) * 2 * m);
      sv = (float *)malloc(sizeof(float) * _min(m, n));
      if (!sv) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (sv)");
	return -1;
      }
      vt = (float *)malloc(sizeof(float) * 2 * n * n);
      if (!vt) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (vt)");
	return -1;
      }
      lwork = 2 * _min(m, n) + _max(m, n);
      work = (float *)malloc(sizeof(float) * 2 * lwork);
      if (!work) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (work)");
	return -1;
      }
      rwork = (float *)malloc(sizeof(float) * 5 * _min(m, n));
      if (!rwork) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (rwork)");
	return -1;
      }
      call_cgesvd('A', 'A', m, n, a, m, sv, u_in->v.c, m, vt, n,
		  work, lwork, rwork, &info);
#pragma omp parallel for
      for (j = 0; j < n; j++)
	memset(s_in->v.s + m * j, 0, sizeof(float) * m);
      k = _min(m, n);
#pragma omp parallel for
      for (j = 0; j < k; j++)
	s_in->v.s[j + m * j] = sv[j];
#pragma omp parallel for private(i)
      for (j = 0; j < n * 2; j += 2) {
	for (i = 0; i < n * 2; i += 2) {
	  v_in->v.c[j + n * i]     =  vt[i + n * j];
	  v_in->v.c[j + n * i + 1] = -vt[i + n * j + 1];
	}
      }
      free(a);
      free(sv);
      free(vt);
      free(work);
      free(rwork);
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    {
      double *a, *sv, *vt, *work, *rwork;

      a = (double *)malloc(sizeof(double) * 2 * m * n);
      if (!a) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (a)");
	return -1;
      }
#pragma omp parallel for
      for (j = 0; j < n * 2; j += 2)
	memcpy(a + m * j, x_in->v.z + m * j, sizeof(double) * 2 * m);
      sv = (double *)malloc(sizeof(double) * _min(m, n));
      if (!sv) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (sv)");
	return -1;
      }
      vt = (double *)malloc(sizeof(double) * 2 * n * n);
      if (!vt) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (vt)");
	return -1;
      }
      lwork = 2 * _min(m, n) + _max(m, n);
      work = (double *)malloc(sizeof(double) * 2 * lwork);
      if (!work) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (work)");
	return -1;
      }
      rwork = (double *)malloc(sizeof(double) * 5 * _min(m, n));
      if (!rwork) {
	fprintf(stderr, MODULE_NAME ": svd: out of memory (rwork)");
	return -1;
      }
      call_zgesvd('A', 'A', m, n, a, m, sv, u_in->v.z, m, vt, n,
		  work, lwork, rwork, &info);
#pragma omp parallel for
      for (j = 0; j < n; j++)
	memset(s_in->v.d + m * j, 0, sizeof(double) * m);
      k = _min(m, n);
#pragma omp parallel for
      for (j = 0; j < k; j++)
	s_in->v.d[j + m * j] = sv[j];
#pragma omp parallel for private(i)
      for (j = 0; j < n * 2; j += 2) {
	for (i = 0; i < n * 2; i += 2) {
	  v_in->v.z[j + n * i]     =  vt[i + n * j];
	  v_in->v.z[j + n * i + 1] = -vt[i + n * j + 1];
	}
      }
      free(a);
      free(sv);
      free(vt);
      free(work);
      free(rwork);
    }
    break;
  default:
    should_not_reach_here;
  }
  if (info != 0)
    return -1;
  param[1].o = u;
  param[2].o = s;
  param[3].o = v;
  return 0;
}

#define SILC_ANY_INTEGER   (SILC_INT|SILC_LONG)
#define SILC_ANY_REAL      (SILC_FLOAT|SILC_DOUBLE)
#define SILC_ANY_COMPLEX   (SILC_COMPLEX|SILC_DOUBLE_COMPLEX)
#define SILC_ANY_PRECISION (SILC_ANY_INTEGER|SILC_ANY_REAL|SILC_ANY_COMPLEX)

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {blasmodule_double_mat_vec_product, "blasmodule_double_mat_vec_product",
   SILC_MULTIPLY_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {blasmodule_doublecomplex_mat_vec_product,
   "blasmodule_doublecomplex_mat_vec_product",
   SILC_MULTIPLY_OP, SILC_DOUBLE_COMPLEX,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {blasmodule_solve, "blasmodule_solve",
   SILC_LEQ_OP, SILC_ANY_REAL|SILC_ANY_COMPLEX,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {blasmodule_solve, "blasmodule_solve",
   SILC_LEQ_OP, SILC_ANY_REAL|SILC_ANY_COMPLEX,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {blasmodule_divide, "blasmodule_divide",
   SILC_DIVIDE_OP, SILC_ANY_REAL|SILC_ANY_COMPLEX,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 0, 0},
  {blasmodule_divide, "blasmodule_divide",
   SILC_DIVIDE_OP, SILC_ANY_REAL|SILC_ANY_COMPLEX,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, 0, 0},

  {blasmodule_dot_product, "blasmodule_dot_product",
   SILC_MULTIPLY_OP, SILC_ANY_REAL,
   SILC_ROW_VECTOR_TYPE, 0, SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},

  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {blasmodule_func_rcond, "blasmodule_func_rcond", "scalar rcond(matrix)", 0},
  {NULL} /* sentinel */
};

static procedure_table_t _procedure_table[] = {
  /* name, sig_text, flags */
  {blasmodule_solve, "blasmodule_solve", "solve(matrix in, vector inout)", 0},
  {blasmodule_proc_svd, "blasmodule_proc_svd",
   "svd(matrix in, matrix out, matrix out, matrix out)", 0},
  {NULL} /* sentinel */
};

void
init_blasmodule()
{
  silc_register_module(MODULE_NAME,
		       NULL,   /* _unary_op_table */
		       _binary_op_table,
		       _function_table,
		       _procedure_table);
}
