/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: clist.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

/* clist.c - compressed list */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "clist.h"

DECLSPEC clist_t *
clist_new()
{
  clist_t *list;

  list = (clist_t *)malloc(sizeof(clist_t));
  list->type = CListType;
  list->next = NULL;
  clist_reset_defaults(list);
  return list;
}

DECLSPEC void
clist_free(clist_t *list)
{
  clist_t *item;

  while (list) {
    item = list;
    list = list->next;
    free(item);
  }
}

DECLSPEC clist_t *
clist_append_point(clist_t *list, int n)
{
  clist_t *element;

  if (n < 0)
    return NULL;
  while (list->next)
    list = list->next;
  element = list->next = (clist_t *)malloc(sizeof(clist_point_t));
  element->type = CListPointType;
  element->next = NULL;
  CLIST_POINT(element)->n = n;
  return element;
}

DECLSPEC clist_t *
clist_append_range(clist_t *list, int start, int end)
{
  clist_t *element;

  while (list->next)
    list = list->next;
  element = list->next = (clist_t *)malloc(sizeof(clist_range_t));
  element->type = CListRangeType;
  element->next = NULL;
  CLIST_RANGE(element)->start = start;
  CLIST_RANGE(element)->end = end;
  return element;
}

DECLSPEC void
clist_set_defaults(clist_t *list, int start, int end)
{
  list->start = start;
  list->end = end;
}

DECLSPEC void
clist_reset_defaults(clist_t *list)
{
  list->start = list->end = -1;
}

#define START(list, default) \
	((CLIST_RANGE(list)->start < 0) ? default : CLIST_RANGE(list)->start)
#define END(list, default) \
	((CLIST_RANGE(list)->end   < 0) ? default : CLIST_RANGE(list)->end)

DECLSPEC int
clist_range(clist_t *list, int *start, int *end)
{
  clist_t *p;

  p = list->next;
  if (p && p->type == CListRangeType && !p->next) {
    *start = START(p, list->start); *end = END(p, list->end);
    return 1;
  }
  return 0;
}

DECLSPEC int
clist_length(clist_t *list)
{
  int n;
  clist_t *p;

  n = 0;
  for (p = list->next; p; p = p->next) {
    switch (p->type) {
    case CListPointType:
      n = n + 1;
      break;
    case CListRangeType:
      n = n + END(p, list->end) - START(p, list->start) + 1;
      break;
    }
  }
  return n;
}

DECLSPEC int
clist_nth(clist_t *list, int n) /* 0 origin */
{
  int c;
  clist_t *p;

  if (n < 0)
    return -1;
  for (p = list->next; p; p = p->next) {
    switch (p->type) {
    case CListPointType:
      if (n == 0)
	return CLIST_POINT(p)->n;
      n = n - 1;
      break;
    case CListRangeType:
      c = END(p, list->end) - START(p, list->start) + 1;
      if (n < c)
	return START(p, list->start) + n;
      n = n - c;
      break;
    }
  }
  return -1;
}

DECLSPEC int
clist_min(clist_t *list)
{
  int c, v;
  clist_t *p;

  v = -1;
  for (p = list->next; p; p = p->next) {
    switch (p->type) {
    case CListPointType:
      if (v < 0 || v > CLIST_POINT(p)->n)
	v = CLIST_POINT(p)->n;
      break;
    case CListRangeType:
      if (v < 0 || v > START(p, list->start))
	v = START(p, list->start);
      break;
    }
  }
  return v;
}

DECLSPEC int
clist_max(clist_t *list)
{
  int c, v;
  clist_t *p;

  v = -1;
  for (p = list->next; p; p = p->next) {
    switch (p->type) {
    case CListPointType:
      if (v < CLIST_POINT(p)->n)
	v = CLIST_POINT(p)->n;
      break;
    case CListRangeType:
      if (v < END(p, list->end))
	v = END(p, list->end);
      break;
    }
  }
  return v;
}

DECLSPEC int *
clist_expand(clist_t *list)
{
  int *v, i, n = 0;
  clist_t *p;

  v = (int *)malloc(sizeof(int) * clist_length(list));
  for (p = list->next; p; p = p->next) {
    switch (p->type) {
    case CListPointType:
      v[n++] = CLIST_POINT(p)->n;
      break;
    case CListRangeType:
      for (i = START(p, list->start); i <= END(p, list->end); i++)
	v[n++] = i;
      break;
    }
  }
  return v;
}

DECLSPEC clist_t *
clist_compress(int *array, int len)
{
  clist_t *list;
  int i, n;

  list = clist_new();
  for (i = 0; i < len; i++) {
    n = i;
    while (i + 1 < len && array[i] + 1 == array[i + 1])
      i++;
    if (i == n) {
      clist_append_point(list, array[n]);
    } else if (i == n + 1) {
      clist_append_point(list, array[n]);
      clist_append_point(list, array[i]);
    } else {
      clist_append_range(list, array[n], array[i]);
    }
  }
  return list;
}

DECLSPEC clist_t *
clist_sublist(clist_t *list, clist_t *index_list)
{
  clist_t *sublist;
  int *v, m, n, i, j;

  m = clist_length(list);
  n = clist_length(index_list);
  if (m > n)
    m = n;
  v = (int *)malloc(m * sizeof(int));
  for (n = i = 0; i < m; i++)
    if ((j = clist_nth(list, clist_nth(index_list, i))) >= 0)
      v[n++] = j;
  sublist = clist_compress(v, n);
  free(v);
  return sublist;
}

DECLSPEC clist_t *
clist_duplicate(clist_t *list)
{
  clist_t *duplicated;

  duplicated = clist_new();
  clist_set_defaults(duplicated, list->start, list->end);
  while (list->next) {
    list = list->next;
    switch (list->type) {
    case CListPointType:
      clist_append_point(duplicated, CLIST_POINT(list)->n);
      break;
    case CListRangeType:
      clist_append_range(duplicated, CLIST_RANGE(list)->start,
			             CLIST_RANGE(list)->end);
      break;
    }
  }
  return duplicated;
}

DECLSPEC void
clist_pprint(clist_t *list)
{
  printf("{");
  while (list->next) {
    list = list->next;
    switch (list->type) {
    case CListPointType:
      printf("%d", CLIST_POINT(list)->n);
      break;
    case CListRangeType:
      if (CLIST_RANGE(list)->start < 0 && CLIST_RANGE(list)->end < 0)
	printf("-:-");
      else if (CLIST_RANGE(list)->start < 0)
	printf("-:%d", CLIST_RANGE(list)->end);
      else if (CLIST_RANGE(list)->end < 0)
	printf("%d:-", CLIST_RANGE(list)->start);
      else
	printf("%d:%d", CLIST_RANGE(list)->start, CLIST_RANGE(list)->end);
      break;
    }
    if (list->next)
      printf(",");
  }
  printf("}\n");
}
