/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cmd.h,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#ifndef SILC_PARSER_H
#define SILC_PARSER_H 1

#include "object.h"

#define SILC_PROGRAM_MAX 4096

/* expresilcon */

typedef enum {
  ExprType = 0,
  ScalarExprType,
  VariableExprType,
  CubicArrayExprType,
  RowVectorExprType,
  ColumnVectorExprType,
  SubscriptExprType,
  RangeExprType,
  UnaryOpExprType,
  BinaryOpExprType,
  FuncCallExprType,
} expr_type_t;

typedef struct {
  expr_type_t type;
} expr_t;

typedef struct {
  expr_type_t type;
  int precision;
  union {
    int32_t i;
    int64_t l;
    float s;
    double d;
  } v;
} scalar_expr_t; /* inherits expr_t */

#define SCALAR(p) ((scalar_expr_t *)(p))

typedef struct {
  expr_type_t type;
  char name[SILC_MAXNAMELEN+1];
} variable_expr_t; /* inherits expr_t */

#define VARIABLE(p) ((variable_expr_t *)(p))

typedef struct {
  expr_type_t type;
  expr_t **v;
  int length;
} cubic_array_expr_t; /* inherits expr_t */

#define CUBIC_ARRAY(p) ((cubic_array_expr_t *)(p))

typedef struct {
  expr_type_t type;
  expr_t **v;
  int length;
} row_vector_expr_t; /* inherits expr_t */

#define ROW_VECTOR(p) ((row_vector_expr_t *)(p))

typedef struct {
  expr_type_t type;
  expr_t **v;
  int length;
} column_vector_expr_t; /* inherits expr_t */

#define COLUMN_VECTOR(p) ((column_vector_expr_t *)(p))

typedef struct {
  expr_type_t type;
  expr_t *start, *end;
} range_expr_t; /* inherits expr_t */

#define RANGE(p) ((range_expr_t *)(p))

typedef struct {
  expr_t *v[3];
  int n;
} subscript_t;

typedef struct {
  expr_type_t type;
  expr_t *operand;
  subscript_t *op;
} subscript_expr_t; /* inherits expr_t */

#define SUBSCRIPT(p) ((subscript_expr_t *)(p))

/* unary operators */
#define SILC_NEGATE_OP			1
#define SILC_COMPLEX_CONJUGATE_OP	2
#define SILC_TRANSPOSE_OP		3
#define SILC_CONJUGATE_TRANSPOSE_OP	4

typedef struct {
  expr_type_t type;
  int op;
  expr_t *operand;
} unary_op_expr_t; /* inherits expr_t */

#define UNARY_OP(p) ((unary_op_expr_t *)(p))

/* binary operators */
#define SILC_ADD_OP			10
#define SILC_SUBTRACT_OP		11
#define SILC_MULTIPLY_OP		12
#define SILC_MULTIPLY_ELEMENTWISE_OP	13
#define SILC_DIVIDE_OP			14
#define SILC_DIVIDE_ELEMENTWISE_OP	15
#define SILC_REMAINDER_OP		16
#define SILC_LEQ_OP			17

typedef struct {
  expr_type_t type;
  int op;
  expr_t *lhs, *rhs;
} binary_op_expr_t; /* inherits expr_t */

#define BINARY_OP(p) ((binary_op_expr_t *)(p))

typedef struct {
  expr_type_t type;
  char name[SILC_MAXNAMELEN+1];
  expr_t **args;
  int n_args;
} func_call_expr_t; /* inherits expr_t */

#define FUNC_CALL(p) ((func_call_expr_t *)(p))

/* statement */

typedef enum {
  StmtType = 0,
  AssignmentStmtType,
  ProcCallStmtType,
  PreferStmtType,
  RenameStmtType,
} stmt_type_t;

typedef enum {
  AssignOp,
  Add_AssignOp,
  Subtract_AssignOp,
  Multiply_AssignOp,
  MultiplyElementwise_AssignOp,
  Divide_AssignOp,
  DivideElementwise_AssignOp,
  Remainder_AssignOp,
} assignment_op_t;
  
typedef struct {
  stmt_type_t type;
} stmt_t;

typedef struct {
  stmt_type_t type;
  assignment_op_t op;
  char name[SILC_MAXNAMELEN+1];
  subscript_t *subscript;
  expr_t *rhs;
} assignment_stmt_t; /* inherits stmt_t */

#define ASSIGNMENT(p) ((assignment_stmt_t *)(p))

typedef struct {
  stmt_type_t type;
  char name[SILC_MAXNAMELEN+1];
  expr_t **args;
  int n_args;
} proc_call_stmt_t; /* inherits stmt_t */

#define PROC_CALL(p) ((proc_call_stmt_t *)(p))

typedef struct {
  stmt_type_t type;
  char name[SILC_MAXNAMELEN+1];
} prefer_stmt_t; /* inherits stmt_t */

#define PREFER(p) ((prefer_stmt_t *)(p))

typedef struct {
  stmt_type_t type;
  char old_name[SILC_MAXNAMELEN+1];
  char new_name[SILC_MAXNAMELEN+1];
} rename_stmt_t; /* inherits stmt_t */

#define RENAME(p) ((rename_stmt_t *)(p))

/* program */

typedef struct _program_t {
  char *source;
  stmt_t **stmts;
  int n_stmts;
  struct _program_t *prev, *next;
} program_t;

program_t *parse_program(char *);
void free_program(program_t *);

#endif /* SILC_PARSER_H */
