%{
/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cmd_parser.y,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "cmd.h"
#include "wtime.h"

static int error;

void cmderror(const char *s)
{
  fprintf(stderr, "parser_program: %s\n", s);
  error = -1;
}

int cmdwrap(void)
{
  return 1;
} 

int cmdlex(void);

static program_t *prog;

static void
add_statement(stmt_t *stmt)
{
  if (prog->n_stmts == 0) {
    prog->stmts = (stmt_t **)malloc(sizeof(stmt_t *));
  } else {
    prog->stmts = (stmt_t **)
      realloc(prog->stmts, sizeof(stmt_t *) * (prog->n_stmts + 1));
  }
  prog->stmts[prog->n_stmts] = stmt;
  prog->n_stmts++;
}

static subscript_t *
new_subscript(int n)
{
  subscript_t *subscript;
  subscript = (subscript_t *)malloc(sizeof(subscript_t));
  subscript->n = n;
  return subscript;
}

static unary_op_expr_t *
new_unary_op_expr(int op, expr_t *operand)
{
  unary_op_expr_t *expr;

  expr = (unary_op_expr_t *)malloc(sizeof(unary_op_expr_t));
  expr->type = UnaryOpExprType;
  expr->op = op;
  expr->operand = operand;
  return expr;
}

static binary_op_expr_t *
new_binary_op_expr(int op, expr_t *lhs, expr_t *rhs)
{
  binary_op_expr_t *expr;

  expr = (binary_op_expr_t *)malloc(sizeof(binary_op_expr_t));
  expr->type = BinaryOpExprType;
  expr->op = op;
  expr->lhs = lhs;
  expr->rhs = rhs;
  return expr;
}

%}

%union 
{
  char string[SILC_MAXNAMELEN+1];
  int32_t num_int4;
  int64_t num_int8;
  float num_real4;
  double num_real8;
  stmt_t *stmt;
  expr_t *expr;
  struct {
    expr_t **v;
    int n;
  } argument_list;
  subscript_t *subscript;
  assignment_op_t assignment_op;
}

%token <string> TOK_IDENTIFIER
%token <num_int4> TOK_INT4_LITERAL
%token <num_int8> TOK_INT8_LITERAL
%token <num_real4> TOK_REAL4_LITERAL
%token <num_real8> TOK_REAL8_LITERAL

%token TOK_EQUAL
%token TOK_PLUS
%token TOK_PLUS_EQUAL
%token TOK_MINUS
%token TOK_MINUS_EQUAL
%token TOK_MULTIPLY
%token TOK_MULTIPLY_EQUAL
%token TOK_DIVIDE
%token TOK_DIVIDE_EQUAL
%token TOK_REMAINDER
%token TOK_REMAINDER_EQUAL
%token TOK_LEQ
%token TOK_CONJUGATE_TRANSPOSE
%token TOK_COMPLEX_CONJUGATE
%token TOK_TRANSPOSE
%token TOK_MULTIPLY_ELEMENTWISE
%token TOK_MULTIPLY_ELEMENTWISE_EQUAL
%token TOK_DIVIDE_ELEMENTWISE
%token TOK_DIVIDE_ELEMENTWISE_EQUAL
%token TOK_DOUBLE_SEMICOLON
%token TOK_PREFER
%token TOK_RENAME
%token TOK_EOL

%type <stmt> procedure_call prefer_statement rename_statement assignment
%type <argument_list> argument_list
%type <assignment_op> assignment_op
%type <expr> expression
%type <subscript> subscript
%type <expr> range_expr range_start range_end
%type <expr> multiplication_expr elementwise_multiplication_expr
%type <expr> leq_expr pre_unary_expr post_unary_expr simple_expr
%type <expr> cubic_array row_vector column_vector range_vector


%%

statement_list: statement | statement_list ';' statement
;

statement: /* empty */
| procedure_call    { add_statement($1); }
| prefer_statement  { add_statement($1); }
| rename_statement  { add_statement($1); }
| assignment        { add_statement($1); }
| error
;

procedure_call: TOK_IDENTIFIER '(' ')'
{
  proc_call_stmt_t *stmt;
  stmt = (proc_call_stmt_t *)malloc(sizeof(proc_call_stmt_t));
  stmt->type = ProcCallStmtType;
  strcpy(stmt->name, $1);
  stmt->args = NULL;
  stmt->n_args = 0;
  $$ = (stmt_t *)stmt;
}
| TOK_IDENTIFIER '(' argument_list ')'
{
  proc_call_stmt_t *stmt;
  stmt = (proc_call_stmt_t *)malloc(sizeof(proc_call_stmt_t));
  stmt->type = ProcCallStmtType;
  strcpy(stmt->name, $1);
  stmt->args = $3.v;
  stmt->n_args = $3.n;
  $$ = (stmt_t *)stmt;
}
;

argument_list: expression
{
  $$.v = (expr_t **)malloc(sizeof(expr_t *));
  $$.v[0] = $1;
  $$.n = 1;
}
| argument_list ',' expression
{
  $$.v = (expr_t **)realloc($1.v, ($1.n + 1) * sizeof(expr_t *));
  $$.v[$1.n] = $3;
  $$.n = $1.n + 1;
}
;

prefer_statement: TOK_PREFER TOK_IDENTIFIER
{
  prefer_stmt_t *stmt;
  stmt = (prefer_stmt_t *)malloc(sizeof(prefer_stmt_t));
  stmt->type = PreferStmtType;
  strcpy(stmt->name, $2);
  $$ = (stmt_t *)stmt;
}
;

rename_statement: TOK_RENAME TOK_IDENTIFIER TOK_IDENTIFIER
{
  rename_stmt_t *stmt;
  stmt = (rename_stmt_t *)malloc(sizeof(rename_stmt_t));
  stmt->type = RenameStmtType;
  strcpy(stmt->old_name, $2);
  strcpy(stmt->new_name, $3);
  $$ = (stmt_t *)stmt;
}
;

assignment: TOK_IDENTIFIER assignment_op expression
{
  assignment_stmt_t *stmt;
  stmt = (assignment_stmt_t *)malloc(sizeof(assignment_stmt_t));
  stmt->type = AssignmentStmtType;
  strcpy(stmt->name, $1);
  stmt->subscript = NULL;
  stmt->op = $2;
  stmt->rhs = $3;
  $$ = (stmt_t *)stmt;
}
| TOK_IDENTIFIER subscript assignment_op expression
{
  int i;

  assignment_stmt_t *stmt;
  stmt = (assignment_stmt_t *)malloc(sizeof(assignment_stmt_t));
  stmt->type = AssignmentStmtType;
  strcpy(stmt->name, $1);
  stmt->subscript = $2;
  stmt->op = $3;
  stmt->rhs = $4;
  $$ = (stmt_t *)stmt;
}
;

assignment_op: TOK_EQUAL         { $$ = AssignOp; }
| TOK_PLUS_EQUAL                 { $$ = Add_AssignOp; }
| TOK_MINUS_EQUAL		 { $$ = Subtract_AssignOp; }
| TOK_MULTIPLY_EQUAL	         { $$ = Multiply_AssignOp; }
| TOK_MULTIPLY_ELEMENTWISE_EQUAL { $$ = MultiplyElementwise_AssignOp; }
| TOK_DIVIDE_EQUAL		 { $$ = Divide_AssignOp; }
| TOK_DIVIDE_ELEMENTWISE_EQUAL   { $$ = DivideElementwise_AssignOp; }
| TOK_REMAINDER_EQUAL            { $$ = Remainder_AssignOp; }
;

subscript: '[' range_expr ']'
{
  $$ = new_subscript(1);
  $$->v[0] = $2;
}  
| '[' range_expr ',' range_expr ']'
{
  $$ = new_subscript(2);
  $$->v[0] = $2;
  $$->v[1] = $4;
}  
| '[' range_expr ',' range_expr ',' range_expr ']'
{
  $$ = new_subscript(3);
  $$->v[0] = $2;
  $$->v[1] = $4;
  $$->v[2] = $6;
}  
;

range_expr: expression
| range_start ':' range_end
{
  range_expr_t *expr;
  expr = (range_expr_t *)malloc(sizeof(range_expr_t));
  expr->type = RangeExprType;
  expr->start = $1;
  expr->end = $3;
  $$ = (expr_t *)expr;
} 
;

range_start: expression
| /* empty */
{
  scalar_expr_t *expr;
  expr = (scalar_expr_t *)malloc(sizeof(scalar_expr_t));
  expr->type = ScalarExprType;
  expr->precision = SILC_INT;
  expr->v.i = -1;
  $$ = (expr_t *)expr;
}
;

range_end: expression
| /* empty */
{
  scalar_expr_t *expr;
  expr = (scalar_expr_t *)malloc(sizeof(scalar_expr_t));
  expr->type = ScalarExprType;
  expr->precision = SILC_INT;
  expr->v.i = -1;
  $$ = (expr_t *)expr;
}
;

expression: multiplication_expr
| expression TOK_PLUS multiplication_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_ADD_OP, $1, $3);
}
| expression TOK_MINUS multiplication_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_SUBTRACT_OP, $1, $3);
}
;

multiplication_expr: elementwise_multiplication_expr
| multiplication_expr TOK_MULTIPLY elementwise_multiplication_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_MULTIPLY_OP, $1, $3);
}
| multiplication_expr TOK_DIVIDE elementwise_multiplication_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_DIVIDE_OP, $1, $3);
}
| multiplication_expr TOK_REMAINDER elementwise_multiplication_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_REMAINDER_OP, $1, $3);
}
;

elementwise_multiplication_expr: leq_expr
| elementwise_multiplication_expr TOK_MULTIPLY_ELEMENTWISE pre_unary_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_MULTIPLY_ELEMENTWISE_OP, $1, $3);
}
| elementwise_multiplication_expr TOK_DIVIDE_ELEMENTWISE pre_unary_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_DIVIDE_ELEMENTWISE_OP, $1, $3);
}
;

leq_expr: pre_unary_expr
|  pre_unary_expr TOK_LEQ leq_expr
{
  $$ = (expr_t *)new_binary_op_expr(SILC_LEQ_OP, $1, $3);
}
;

pre_unary_expr: post_unary_expr
| TOK_PLUS pre_unary_expr
{
  $$ = $2; /* simply ignore the plus sign */
}
| TOK_MINUS pre_unary_expr
{
  $$ = (expr_t *)new_unary_op_expr(SILC_NEGATE_OP, $2);
}
;

post_unary_expr: simple_expr
| post_unary_expr TOK_CONJUGATE_TRANSPOSE
{
  $$ = (expr_t *)new_unary_op_expr(SILC_CONJUGATE_TRANSPOSE_OP, $1);
}
| post_unary_expr TOK_COMPLEX_CONJUGATE
{
  $$ = (expr_t *)new_unary_op_expr(SILC_COMPLEX_CONJUGATE_OP, $1);
}
| post_unary_expr TOK_TRANSPOSE
{
  $$ = (expr_t *)new_unary_op_expr(SILC_TRANSPOSE_OP, $1);
}
| post_unary_expr subscript
{
  subscript_expr_t *expr;
  int i;

  expr = (subscript_expr_t *)malloc(sizeof(subscript_expr_t));
  expr->type = SubscriptExprType;
  expr->operand = $1;
  expr->op = $2;
  $$ = (expr_t *)expr;
}
;

simple_expr: TOK_IDENTIFIER '(' ')'
{
  func_call_expr_t *expr;
  expr = (func_call_expr_t *)malloc(sizeof(func_call_expr_t));
  expr->type = FuncCallExprType;
  strcpy(expr->name, $1);
  expr->args = NULL;
  expr->n_args = 0;
  $$ = (expr_t *)expr;
} 
| TOK_IDENTIFIER '(' argument_list ')'
{
  func_call_expr_t *expr;
  expr = (func_call_expr_t *)malloc(sizeof(func_call_expr_t));
  expr->type = FuncCallExprType;
  strcpy(expr->name, $1);
  expr->args = $3.v;
  expr->n_args = $3.n;
  $$ = (expr_t *)expr;
} 
| TOK_IDENTIFIER
{
  variable_expr_t *expr;
  expr = (variable_expr_t *)malloc(sizeof(variable_expr_t));
  expr->type = VariableExprType;
  strcpy(expr->name, $1);
  $$ = (expr_t *)expr;
}  
| TOK_INT4_LITERAL
{
  scalar_expr_t *expr;
  expr = (scalar_expr_t *)malloc(sizeof(scalar_expr_t));
  expr->type = ScalarExprType;
  expr->precision = SILC_INT;
  expr->v.i = $1;
  $$ = (expr_t *)expr;
}  
| TOK_INT8_LITERAL
{
  scalar_expr_t *expr;
  expr = (scalar_expr_t *)malloc(sizeof(scalar_expr_t));
  expr->type = ScalarExprType;
  expr->precision = SILC_LONG;
  expr->v.l = $1;
  $$ = (expr_t *)expr;
}  
| TOK_REAL4_LITERAL
{
  scalar_expr_t *expr;
  expr = (scalar_expr_t *)malloc(sizeof(scalar_expr_t));
  expr->type = ScalarExprType;
  expr->precision = SILC_FLOAT;
  expr->v.s = $1;
  $$ = (expr_t *)expr;
}  
| TOK_REAL8_LITERAL
{
  scalar_expr_t *expr;
  expr = (scalar_expr_t *)malloc(sizeof(scalar_expr_t));
  expr->type = ScalarExprType;
  expr->precision = SILC_DOUBLE;
  expr->v.d = $1;
  $$ = (expr_t *)expr;
}  
| '{' cubic_array '}'
{
  $$ = $2;
}
| '(' expression ')'
{
  $$ = $2;
}
;

cubic_array: row_vector
| cubic_array TOK_DOUBLE_SEMICOLON row_vector
{
  cubic_array_expr_t *expr;

  if ($1->type == CubicArrayExprType) {
    expr = CUBIC_ARRAY($1);
    expr->v = (expr_t **)
      realloc(expr->v, sizeof(expr_t *) * (expr->length + 1));
    expr->v[expr->length] = $3;
    expr->length++;
  } else {
    expr = (cubic_array_expr_t *)malloc(sizeof(cubic_array_expr_t));
    expr->type = CubicArrayExprType;
    expr->v = (expr_t **)malloc(sizeof(expr_t *) * 2);
    expr->v[0] = $1;
    expr->v[1] = $3;
    expr->length = 2;
  }
  $$ = (expr_t *)expr;
}
;

row_vector: column_vector
| row_vector ';' column_vector
{
  row_vector_expr_t *expr;

  if ($1->type == RowVectorExprType) {
    expr = ROW_VECTOR($1);
    expr->v = (expr_t **)
      realloc(expr->v, sizeof(expr_t *) * (expr->length + 1));
    expr->v[expr->length] = $3;
    expr->length++;
  } else {
    expr = (row_vector_expr_t *)malloc(sizeof(row_vector_expr_t));
    expr->type = RowVectorExprType;
    expr->v = (expr_t **)malloc(sizeof(expr_t *) * 2);
    expr->v[0] = $1;
    expr->v[1] = $3;
    expr->length = 2;
  }
  $$ = (expr_t *)expr;
}
;

column_vector: range_vector
| column_vector ',' range_vector
{
  column_vector_expr_t *expr;

  if ($1->type == ColumnVectorExprType) {
    expr = COLUMN_VECTOR($1);
    expr->v = (expr_t **)
      realloc(expr->v, sizeof(expr_t *) * (expr->length + 1));
    expr->v[expr->length] = $3;
    expr->length++;
  } else {
    expr = (column_vector_expr_t *)malloc(sizeof(column_vector_expr_t));
    expr->type = ColumnVectorExprType;
    expr->v = (expr_t **)malloc(sizeof(expr_t *) * 2);
    expr->v[0] = $1;
    expr->v[1] = $3;
    expr->length = 2;
  }
  $$ = (expr_t *)expr;
}
;

range_vector: expression
| expression ':' expression
{
  range_expr_t *expr;
  expr = (range_expr_t *)malloc(sizeof(range_expr_t));
  expr->type = RangeExprType;
  expr->start = $1;
  expr->end = $3;
  $$ = (expr_t *)expr;
}
;

%%

program_t *
parse_program(char *s)
{
  extern void set_cmd_text(char *);
#ifdef TIME_PARSE
  extern double _parse_time; /* server.c */
  double t, elapsed;

  t = silc_wtime();
#endif
  set_cmd_text(s);
  prog = (program_t *)malloc(sizeof(program_t));
  prog->source = strdup(s);
  prog->stmts = NULL;
  prog->n_stmts = 0;
  error = 0;
  yyparse();
#ifdef TIME_PARSE
  elapsed = silc_wtime() - t;
  _parse_time += elapsed;
#endif
  if (!error)
    return prog;
  free_program(prog);
  return NULL;
}

static void free_subscript(subscript_t *);

static void
free_expression(expr_t *expr)
{
  int i;

  switch(expr->type) {
  case CubicArrayExprType:
    for (i = 0; i < CUBIC_ARRAY(expr)->length; i++)
      free_expression(CUBIC_ARRAY(expr)->v[i]);
    break;
  case RowVectorExprType:
    for (i = 0; i < ROW_VECTOR(expr)->length; i++)
      free_expression(ROW_VECTOR(expr)->v[i]);
    break;
  case ColumnVectorExprType:
    for (i = 0; i < COLUMN_VECTOR(expr)->length; i++)
      free_expression(COLUMN_VECTOR(expr)->v[i]);
    break;
  case SubscriptExprType:
    free_expression(SUBSCRIPT(expr)->operand);
    free_subscript(SUBSCRIPT(expr)->op);
    break;
  case RangeExprType:
    free_expression(RANGE(expr)->start);
    free_expression(RANGE(expr)->end);
    break;
  case UnaryOpExprType:
    free_expression(UNARY_OP(expr)->operand);
    break;
  case BinaryOpExprType:
    free_expression(BINARY_OP(expr)->lhs);
    free_expression(BINARY_OP(expr)->rhs);
    break;
  case FuncCallExprType:
    if (FUNC_CALL(expr)->n_args > 0) {
      for (i = 0; i < FUNC_CALL(expr)->n_args; i++)
	free_expression(FUNC_CALL(expr)->args[i]);
      free(FUNC_CALL(expr)->args);
    }
    break;
  }
  free(expr);
}

static void
free_subscript(subscript_t *subscript)
{
  int i;

  for (i = 0; i < subscript->n; i++)
    free_expression(subscript->v[i]);
  free(subscript);
}

void
free_program(program_t *p)
{
  int i, j;

  for (i = 0; i < p->n_stmts; i++) {
    stmt_t *stmt = p->stmts[i];

    switch(stmt->type) {
    case AssignmentStmtType:
      if (ASSIGNMENT(stmt)->subscript)
	free_subscript(ASSIGNMENT(stmt)->subscript);
      free_expression(ASSIGNMENT(stmt)->rhs);
      break;
    case ProcCallStmtType:
      if (PROC_CALL(stmt)->n_args > 0) {
	for (j = 0; j < PROC_CALL(stmt)->n_args; j++)
	  free_expression(PROC_CALL(stmt)->args[j]);
	free(PROC_CALL(stmt)->args);
      }
      break;
    }
  }
  free(p);
}
