/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cmd_test.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>

#include "cmd.h"

#define indent() { int i;  for (i = 0; i < level; i++) printf("  "); }

void
print_subscript(subscript_t *subscript, int level)
{
  void print_expression(expr_t *, int);
  int i;

  for (i = 0; i < subscript->n; i++) {
    indent(); printf("v[%d]:\n", i);
    print_expression(subscript->v[i], level);
  }
}

void
print_expression(expr_t *expr, int level)
{
  int i;

  level++;
  switch(expr->type) {
  case ExprType:
    indent(); printf("none\n");
    break;
  case IntegerExprType:
    indent(); printf("integer: %d\n", INTEGER(expr)->v);
    break;
  case RealExprType:
    indent(); printf("real: %f\n", REAL(expr)->v);
    break;
  case CubicArrayExprType:
    indent(); printf("cubic array:\n");
    for (i = 0; i < CUBIC_ARRAY(expr)->length; i++) {
      indent(); printf("v[%d]:\n", i);
      print_expression(CUBIC_ARRAY(expr)->v[i], level);
    }
    break;
  case RowVectorExprType:
    indent(); printf("row vector:\n");
    for (i = 0; i < ROW_VECTOR(expr)->length; i++) {
      indent(); printf("v[%d]:\n", i);
      print_expression(ROW_VECTOR(expr)->v[i], level);
    }
    break;
  case ColumnVectorExprType:
    indent(); printf("column vector:\n");
    for (i = 0; i < COLUMN_VECTOR(expr)->length; i++) {
      indent(); printf("v[%d]:\n", i);
      print_expression(COLUMN_VECTOR(expr)->v[i], level);
    }
    break;
  case SubscriptExprType:
    indent(); printf("subscript expression:\n");
    indent(); printf("operand:\n");
    print_expression(SUBSCRIPT(expr)->operand, level);
    indent(); printf("subscript:\n");
    print_subscript(SUBSCRIPT(expr)->op, level);
    break;
  case RangeExprType:
    indent(); printf("range:\n");
    indent(); printf("start:\n");
    print_expression(RANGE(expr)->start, level);
    indent(); printf("end:\n");
    print_expression(RANGE(expr)->end, level);
    break;
  case VariableExprType:
    indent(); printf("variable: \"%s\"\n", VARIABLE(expr)->name);
    break;
  case UnaryOpExprType:
    indent(); printf("unary_op: ");
    switch(UNARY_OP(expr)->op) {
    case SILC_NEGATE_OP:
      printf("negate\n"); break;
    case SILC_INVERT_OP:
      printf("invert\n"); break;
    case SILC_CONJUGATE_TRANSPOSE_OP:
      printf("conjugate transpose\n"); break;
    case SILC_COMPLEX_CONJUGATE_OP:
      printf("complex conjugate\n"); break;
    case SILC_TRANSPOSE_OP:
      printf("transpose\n"); break;
    }
    indent(); printf("operand:\n");
    print_expression(UNARY_OP(expr)->operand, level);
    break;
  case BinaryOpExprType:
    indent(); printf("binary_op: ");
    switch(BINARY_OP(expr)->op) {
    case SILC_ADD_OP:
      printf("add\n"); break;
    case SILC_SUBTRACT_OP:
      printf("subtract\n"); break;
    case SILC_MULTIPLY_OP:
      printf("multiply\n"); break;
    case SILC_MULTIPLY_ELEMENTWISE_OP:
      printf("multiply elementwise\n"); break;
    case SILC_DIVIDE_OP:
      printf("divide\n"); break;
    case SILC_DIVIDE_ELEMENTWISE_OP:
      printf("divide elementwise\n"); break;
    case SILC_REMAINDER_OP:
      printf("remainder\n"); break;
    case SILC_INNER_PRODUCT_OP:
      printf("inner product\n"); break;
    case SILC_LEQ_OP:
      printf("linear equations\n"); break;
    }
    indent(); printf("left-hand side:\n");
    print_expression(BINARY_OP(expr)->lhs, level);
    indent(); printf("right-hand side:\n");
    print_expression(BINARY_OP(expr)->rhs, level);
    break;
  case FuncCallExprType:
    indent(); printf("funcion call:\n");
    indent(); printf("name: \"%s\"\n", FUNC_CALL(expr)->name);
    for (i = 0; i < FUNC_CALL(expr)->n_args; i++) {
      indent(); printf("args[%d]:\n", i);
      print_expression(FUNC_CALL(expr)->args[i], level);
    }
  }
}

void
print_statment(stmt_t *stmt)
{
  int i;

  switch(stmt->type) {
  case AssignmentStmtType:
    switch(ASSIGNMENT(stmt)->op) {
    case AssignOp:
      printf("assignment:\n"); break;
    case Add_AssignOp:
      printf("add assignment:\n"); break;
    case Subtract_AssignOp:
      printf("subtract assignment:\n"); break;
    case Multiply_AssignOp:
      printf("multiply assignment:\n"); break;
    case MultiplyElementwise_AssignOp:
      printf("multiply elementwise assignment:\n"); break;
    case Divide_AssignOp:
      printf("divide assignment:\n"); break;
    case DivideElementwise_AssignOp:
      printf("divide elementwise assignment:\n"); break;
    case Remainder_AssignOp:
      printf("remainder assignment:\n"); break;
    }
    printf("  name: \"%s\"\n", ASSIGNMENT(stmt)->name);
    if (ASSIGNMENT(stmt)->subscript) {
      printf("  subscript:\n");
      print_subscript(ASSIGNMENT(stmt)->subscript, 1);
    }
    printf("  right-hand side:\n");
    print_expression(ASSIGNMENT(stmt)->rhs, 1);
    break;
  case PreferStmtType:
    printf("  prefer:\n");
    printf("  name: \"%s\"\n", PREFER(stmt)->name);
    break;
  case RenameStmtType:
    printf("  rename:\n");
    printf("  old_name: \"%s\"\n", RENAME(stmt)->old_name);
    printf("  new_name: \"%s\"\n", RENAME(stmt)->new_name);
    break;
  case ProcCallStmtType:
    printf("  procedure call:\n");
    printf("  name: \"%s\"\n", PROC_CALL(stmt)->name);
    for (i = 0; i < PROC_CALL(stmt)->n_args; i++) {
      printf("  args[%d]:\n", i);
      print_expression(PROC_CALL(stmt)->args[i], 1);
    }
    break;
  }
}

int
main(int argc, char *argv[])
{
  program_t *p;
  int i;

  if (argc != 2) {
    fprintf(stderr, "Usage: %s <input_text>\n", argv[0]);
    return -1;
  }
  if ((p = parse_program(argv[1])) == NULL)
    return -1;
  if (p->n_stmts > 0)
    printf("n_stmts = %d\n", p->n_stmts);
  for (i = 0; i < p->n_stmts; i++) {
    printf("stmts[%d]:\n", i);
    print_statment(p->stmts[i]);
  }
  free_program(p);
  return 0;
} 
