/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: conditional.c,v 1.3 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "module.h"

#define MODULE_NAME "conditional"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#define OP_LT 1
#define OP_GT 2
#define OP_LE 3
#define OP_GE 4
#define OP_EQ 5
#define OP_NE 6

#define OP_NAME(o) (((o) == OP_LT) ? "lt" : \
		    ((o) == OP_GT) ? "gt" : \
		    ((o) == OP_LE) ? "le" : \
		    ((o) == OP_GE) ? "ge" : \
		    ((o) == OP_EQ) ? "eq" : \
		    ((o) == OP_NE) ? "ne" : "unknown")

silc_object_t *
_cmp_vector_vector(silc_object_t *x, silc_object_t *y, int op)
{
  silc_object_t *object;
  ssize_t i;

  if (x->d.vector.length != y->d.vector.length) {
    fprintf(stderr, "function %s: vector size mismatch\n", OP_NAME(op));
    return NULL;
  }
  if (x->precision != y->precision) {
    fprintf(stderr, "function %s: precision mismatch\n", OP_NAME(op));
    return NULL;
  }
  switch (x->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    object = silc_object_new_vector(x->precision, x->d.vector.length);
    break;
  case SILC_COMPLEX:
    object = silc_object_new_vector(SILC_FLOAT,   x->d.vector.length);
    break;
  case SILC_DOUBLE_COMPLEX:
    object = silc_object_new_vector(SILC_DOUBLE,  x->d.vector.length);
    break;
  default:
    should_not_reach_here;
    return NULL;
  }
  if (!object) {
    fprintf(stderr, "function %s: out of memory\n", OP_NAME(op));
    return NULL;
  }

#define _lt_vector_vector_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] < y->d.vector.v.pre[i];		\
  }
#define _gt_vector_vector_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] > y->d.vector.v.pre[i];		\
  }
#define _le_vector_vector_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] <= y->d.vector.v.pre[i];		\
  }
#define _ge_vector_vector_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] >= y->d.vector.v.pre[i];		\
  }
#define _eq_vector_vector_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] == y->d.vector.v.pre[i];		\
  }
#define _ne_vector_vector_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] != y->d.vector.v.pre[i];		\
  }
#define _lt_vector_vector_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] < y->d.vector.v.pre[ii];		\
  }
#define _gt_vector_vector_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] > y->d.vector.v.pre[ii];		\
  }
#define _le_vector_vector_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] <= y->d.vector.v.pre[ii];		\
  }
#define _ge_vector_vector_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] >= y->d.vector.v.pre[ii];		\
  }
#define _eq_vector_vector_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      (x->d.vector.v.pre[ii]   == y->d.vector.v.pre[ii] &&	\
       x->d.vector.v.pre[ii+1] == y->d.vector.v.pre[ii+1]);	\
  }
#define _ne_vector_vector_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      (x->d.vector.v.pre[ii]   != y->d.vector.v.pre[ii] &&	\
       x->d.vector.v.pre[ii+1] != y->d.vector.v.pre[ii+1]);	\
  }

  object->type = x->type;
  switch (object->precision) {
  case SILC_INT:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_vector_1(i);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_vector_1(i);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_vector_1(i);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_vector_1(i);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_vector_1(i);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_vector_1(i);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_vector_1(l);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_vector_1(l);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_vector_1(l);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_vector_1(l);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_vector_1(l);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_vector_1(l);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_vector_1(s);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_vector_1(s);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_vector_1(s);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_vector_1(s);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_vector_1(s);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_vector_1(s);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_vector_1(d);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_vector_1(d);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_vector_1(d);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_vector_1(d);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_vector_1(d);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_vector_1(d);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_vector_2(c);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_vector_2(c);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_vector_2(c);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_vector_2(c);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_vector_2(c);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_vector_2(c);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_vector_2(z);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_vector_2(z);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_vector_2(z);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_vector_2(z);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_vector_2(z);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_vector_2(z);
      break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return object;
}

silc_object_t *
_cmp_vector_scalar(silc_object_t *x, silc_object_t *y, int op)
{
  silc_object_t *object;
  ssize_t i;

  if (x->precision != y->precision) {
    fprintf(stderr, "function %s: precision mismatch\n", OP_NAME(op));
    return NULL;
  }
  switch (x->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    object = silc_object_new_vector(x->precision, x->d.vector.length);
    break;
  case SILC_COMPLEX:
    object = silc_object_new_vector(SILC_FLOAT,   x->d.vector.length);
    break;
  case SILC_DOUBLE_COMPLEX:
    object = silc_object_new_vector(SILC_DOUBLE,  x->d.vector.length);
    break;
  default:
    should_not_reach_here;
    return NULL;
  }
  if (!object) {
    fprintf(stderr, "function %s: out of memory\n", OP_NAME(op));
    return NULL;
  }

#define _lt_vector_scalar_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] < y->d.scalar.pre;			\
  }
#define _gt_vector_scalar_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] > y->d.scalar.pre;			\
  }
#define _le_vector_scalar_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] <= y->d.scalar.pre;			\
  }
#define _ge_vector_scalar_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] >= y->d.scalar.pre;			\
  }
#define _eq_vector_scalar_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] == y->d.scalar.pre;			\
  }
#define _ne_vector_scalar_1(pre) {				\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[i] != y->d.scalar.pre;			\
  }
#define _lt_vector_scalar_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] < y->d.scalar.pre[0];		\
  }
#define _gt_vector_scalar_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] > y->d.scalar.pre[0];		\
  }
#define _le_vector_scalar_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] <= y->d.scalar.pre[0];		\
  }
#define _ge_vector_scalar_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      x->d.vector.v.pre[ii] >= y->d.scalar.pre[0];		\
  }
#define _eq_vector_scalar_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      (x->d.vector.v.pre[ii]   == y->d.scalar.pre[0] &&		\
       x->d.vector.v.pre[ii+1] == y->d.scalar.pre[1]);		\
  }
#define _ne_vector_scalar_2(pre) {				\
    size_t ii; ii = i * 2;					\
    object->d.vector.v.pre[i] =					\
      (x->d.vector.v.pre[ii]   != y->d.scalar.pre[0] &&		\
       x->d.vector.v.pre[ii+1] != y->d.scalar.pre[1]);		\
  }

  object->type = x->type;
  switch (object->precision) {
  case SILC_INT:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_scalar_1(i);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_scalar_1(i);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_scalar_1(i);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_scalar_1(i);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_scalar_1(i);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_scalar_1(i);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_scalar_1(l);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_scalar_1(l);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_scalar_1(l);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_scalar_1(l);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_scalar_1(l);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_scalar_1(l);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_scalar_1(s);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_scalar_1(s);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_scalar_1(s);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_scalar_1(s);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_scalar_1(s);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_scalar_1(s);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_scalar_1(d);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_scalar_1(d);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_scalar_1(d);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_scalar_1(d);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_scalar_1(d);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_scalar_1(d);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_scalar_2(c);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_scalar_2(c);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_scalar_2(c);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_scalar_2(c);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_scalar_2(c);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_scalar_2(c);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    switch (op) {
    case OP_LT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_lt_vector_scalar_2(z);
      break;
    case OP_GT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_gt_vector_scalar_2(z);
      break;
    case OP_LE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_le_vector_scalar_2(z);
      break;
    case OP_GE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ge_vector_scalar_2(z);
      break;
    case OP_EQ:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_eq_vector_scalar_2(z);
      break;
    case OP_NE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	_ne_vector_scalar_2(z);
      break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return object;
}

int
func_lt(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_lt(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      param[2].o = _cmp_vector_vector(x, y, OP_LT);
      break;
    case SILC_SCALAR_TYPE:
      param[2].o = _cmp_vector_scalar(x, y, OP_LT);
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return (param[2].o) ? 0 : -1;
}

int
func_gt(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_gt(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      param[2].o = _cmp_vector_vector(x, y, OP_GT);
      break;
    case SILC_SCALAR_TYPE:
      param[2].o = _cmp_vector_scalar(x, y, OP_GT);
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return (param[2].o) ? 0 : -1;
}

int
func_le(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_le(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      param[2].o = _cmp_vector_vector(x, y, OP_LE);
      break;
    case SILC_SCALAR_TYPE:
      param[2].o = _cmp_vector_scalar(x, y, OP_LE);
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return (param[2].o) ? 0 : -1;
}

int
func_ge(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_ge(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      param[2].o = _cmp_vector_vector(x, y, OP_GE);
      break;
    case SILC_SCALAR_TYPE:
      param[2].o = _cmp_vector_scalar(x, y, OP_GE);
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return (param[2].o) ? 0 : -1;
}

int
func_eq(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_eq(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      param[2].o = _cmp_vector_vector(x, y, OP_EQ);
      break;
    case SILC_SCALAR_TYPE:
      param[2].o = _cmp_vector_scalar(x, y, OP_EQ);
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return (param[2].o) ? 0 : -1;
}

int
func_ne(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_ne(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      param[2].o = _cmp_vector_vector(x, y, OP_NE);
      break;
    case SILC_SCALAR_TYPE:
      param[2].o = _cmp_vector_scalar(x, y, OP_NE);
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return (param[2].o) ? 0 : -1;
}

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  /* vector-vector */
  {func_lt, "func_lt", "vector lt(vector, vector)", 0},
  {func_gt, "func_gt", "vector gt(vector, vector)", 0},
  {func_le, "func_le", "vector le(vector, vector)", 0},
  {func_ge, "func_ge", "vector ge(vector, vector)", 0},
  {func_eq, "func_eq", "vector eq(vector, vector)", 0},
  {func_ne, "func_ne", "vector ne(vector, vector)", 0},
  /* Vector-scalar */
  {func_lt, "func_lt", "vector lt(vector, scalar)", 0},
  {func_gt, "func_gt", "vector gt(vector, scalar)", 0},
  {func_le, "func_le", "vector le(vector, scalar)", 0},
  {func_ge, "func_ge", "vector ge(vector, scalar)", 0},
  {func_eq, "func_eq", "vector eq(vector, scalar)", 0},
  {func_ne, "func_ne", "vector ne(vector, scalar)", 0},
#if 0
  {func_not, "func_not", "scalar not(vector)", 0},
  {func_any, "func_any", "scalar any(vector)", 0},
  {func_all, "func_all", "scalar all(vector)", 0},
#endif
  {NULL} /* sentinel */
};

void
init_conditional()
{
  silc_register_module(MODULE_NAME,
		       NULL, /* _unary_op_table */
		       NULL, /* _binary_op_table */
		       _function_table,
		       NULL); /* _procedure_table */
}
