/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: coremodule.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

/* todo: take x_mods and y_mods into consideration */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "module.h"
#include "wtime.h"

#define MODULE_NAME "coremodule"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

static int
_complex_conjugate(silc_object_t *object)
{
  int i;

#define _complex_conjugate_one(v) ((v) = -(v))

#define _complex_conjugate_all(buf, n) \
	for (i = 0; i < n; i++) \
	  _complex_conjugate_one(buf[i*2+1])

  switch (object->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    break;
  case SILC_COMPLEX:
    switch (object->type) {
    case SILC_SCALAR_TYPE:
      _complex_conjugate_one(object->d.scalar.c[1]);
      break;
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      _complex_conjugate_all(object->d.vector.v.c,
			     object->d.vector.length);
      break;
    case SILC_CUBIC_ARRAY_TYPE:
      _complex_conjugate_all(object->d.cubic_array.v.c,
			     object->d.cubic_array.l * 
			     object->d.cubic_array.m * 
			     object->d.cubic_array.n);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    switch (object->type) {
    case SILC_SCALAR_TYPE:
      _complex_conjugate_one(object->d.scalar.z[1]);
      break;
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      _complex_conjugate_all(object->d.vector.v.z,
			     object->d.vector.length);
      break;
    case SILC_CUBIC_ARRAY_TYPE:
      _complex_conjugate_all(object->d.cubic_array.v.z,
			     object->d.cubic_array.l * 
			     object->d.cubic_array.m * 
			     object->d.cubic_array.n);
      break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static silc_object_t *
_transpose(silc_object_t *object)
{
  silc_object_t *modified;

  switch (object->type) {
  case SILC_SCALAR_TYPE:
    modified = silc_object_duplicate(object, 1);
    break;
  case SILC_COLUMN_VECTOR_TYPE:
    modified = silc_object_duplicate(object, 1);
    modified->type = SILC_ROW_VECTOR_TYPE;
    break;
  case SILC_ROW_VECTOR_TYPE:
    modified = silc_object_duplicate(object, 1);
    modified->type = SILC_COLUMN_VECTOR_TYPE;
    break;
  default:
    should_not_reach_here;
  }
  return modified; /* new object */
}

int
complex_conjugate(int n_params, silc_mfunc_param_t *param)
{
#ifdef DEBUG
  printf(MODULE_NAME ": complex_conjugate(0x%lx)\n", param[0].o);
#endif
  return _complex_conjugate(param[0].o);
}

int
transpose(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *modified;

#ifdef DEBUG
  printf(MODULE_NAME ": transpose(0x%lx)\n", param[0].o);
#endif
  modified = _transpose(param[0].o);
  if (!modified)
    return -1;
  param[1].o = modified;
  return 0;
}

int
conjugate_transpose(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *modified;

#ifdef DEBUG
  printf(MODULE_NAME ": conjugate_transpose(0x%lx)\n", param[0].o);
#endif
  modified = _transpose(param[0].o);
  if (!modified)
    return -1;
  if (_complex_conjugate(modified) < 0)
    return -1;
  param[1].o = modified;
  return 0;
}

/* out: 0 */
int
negate(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  size_t i, size;

  object = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": negate(0x%lx)\n", object);
#endif

#define _negate(v) ((v) = -(v))
#define _negate_all_1(buf) \
	for (i = 0; i < size; i++) _negate(buf[i])
#define _negate_all_2(buf) \
	for (i = 0; i < size; i++) _negate(buf[i*2])

  switch (object->type) {
  case SILC_SCALAR_TYPE:
    switch (object->precision) {
    case SILC_INT:
      _negate(object->d.scalar.i); break;
    case SILC_LONG:
      _negate(object->d.scalar.l); break;
    case SILC_FLOAT:
      _negate(object->d.scalar.s); break;
    case SILC_DOUBLE:
      _negate(object->d.scalar.d); break;
    case SILC_COMPLEX:
      _negate(object->d.scalar.c[0]); break;
    case SILC_DOUBLE_COMPLEX:
      _negate(object->d.scalar.z[0]); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    size = object->d.vector.length;
    switch (object->precision) {
    case SILC_INT:
      _negate_all_1(object->d.vector.v.i); break;
    case SILC_LONG:
      _negate_all_1(object->d.vector.v.l); break;
    case SILC_FLOAT:
      _negate_all_1(object->d.vector.v.s); break;
    case SILC_DOUBLE:
      _negate_all_1(object->d.vector.v.d); break;
    case SILC_COMPLEX:
      _negate_all_2(object->d.vector.v.c); break;
    case SILC_DOUBLE_COMPLEX:
      _negate_all_2(object->d.vector.v.z); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    size = object->d.cubic_array.l *
           object->d.cubic_array.m *
           object->d.cubic_array.n;
    switch (object->precision) {
    case SILC_INT:
      _negate_all_1(object->d.cubic_array.v.i); break;
    case SILC_LONG:
      _negate_all_1(object->d.cubic_array.v.l); break;
    case SILC_FLOAT:
      _negate_all_1(object->d.cubic_array.v.s); break;
    case SILC_DOUBLE:
      _negate_all_1(object->d.cubic_array.v.d); break;
    case SILC_COMPLEX:
      _negate_all_2(object->d.cubic_array.v.c); break;
    case SILC_DOUBLE_COMPLEX:
      _negate_all_2(object->d.cubic_array.v.z); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_add_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_add_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
    x->d.scalar.i = x->d.scalar.i + y->d.scalar.i; break;
  case SILC_LONG:
    x->d.scalar.l = x->d.scalar.l + y->d.scalar.l; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_sub_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_sub_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
    x->d.scalar.i = x->d.scalar.i - y->d.scalar.i; break;
  case SILC_LONG:
    x->d.scalar.l = x->d.scalar.l - y->d.scalar.l; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_mul_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_mul_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
    x->d.scalar.i = x->d.scalar.i * y->d.scalar.i; break;
  case SILC_LONG:
    x->d.scalar.l = x->d.scalar.l * y->d.scalar.l; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_div_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_div_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
    x->d.scalar.i = x->d.scalar.i / y->d.scalar.i; break;
  case SILC_LONG:
    x->d.scalar.l = x->d.scalar.l / y->d.scalar.l; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_mod_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_mod_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
    x->d.scalar.i = x->d.scalar.i % y->d.scalar.i; break;
  case SILC_LONG:
    x->d.scalar.l = x->d.scalar.l % y->d.scalar.l; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_leq_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_leq_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
    x->d.scalar.i = y->d.scalar.i / x->d.scalar.i; break;
  case SILC_LONG:
    x->d.scalar.l = y->d.scalar.l / x->d.scalar.l; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
real_add_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": real_add_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_FLOAT:
    x->d.scalar.s = x->d.scalar.s + y->d.scalar.s; break;
  case SILC_DOUBLE:
    x->d.scalar.d = x->d.scalar.d + y->d.scalar.d; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
real_sub_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": real_sub_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_FLOAT:
    x->d.scalar.s = x->d.scalar.s - y->d.scalar.s; break;
  case SILC_DOUBLE:
    x->d.scalar.d = x->d.scalar.d - y->d.scalar.d; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
real_mul_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": real_mul_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_FLOAT:
    x->d.scalar.s = x->d.scalar.s * y->d.scalar.s; break;
  case SILC_DOUBLE:
    x->d.scalar.d = x->d.scalar.d * y->d.scalar.d; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
real_div_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": real_div_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_FLOAT:
    x->d.scalar.s = x->d.scalar.s / y->d.scalar.s; break;
  case SILC_DOUBLE:
    x->d.scalar.d = x->d.scalar.d / y->d.scalar.d; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
real_leq_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": real_leq_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_FLOAT:
    x->d.scalar.s = y->d.scalar.s / x->d.scalar.s; break;
  case SILC_DOUBLE:
    x->d.scalar.d = y->d.scalar.d / x->d.scalar.d; break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_add_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": complex_add_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_COMPLEX:
    x->d.scalar.c[0] = x->d.scalar.c[0] + y->d.scalar.c[0];
    x->d.scalar.c[1] = x->d.scalar.c[1] + y->d.scalar.c[1];
    break;
  case SILC_DOUBLE_COMPLEX:
    x->d.scalar.z[0] = x->d.scalar.z[0] + y->d.scalar.z[0];
    x->d.scalar.z[1] = x->d.scalar.z[1] + y->d.scalar.z[1];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_sub_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": complex_sub_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_COMPLEX:
    x->d.scalar.c[0] = x->d.scalar.c[0] - y->d.scalar.c[0];
    x->d.scalar.c[1] = x->d.scalar.c[1] - y->d.scalar.c[1];
    break;
  case SILC_DOUBLE_COMPLEX:
    x->d.scalar.z[0] = x->d.scalar.z[0] - y->d.scalar.z[0];
    x->d.scalar.z[1] = x->d.scalar.z[1] - y->d.scalar.z[1];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_mul_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
  float c[2];
  double z[2];
#ifdef DEBUG
  printf(MODULE_NAME ": complex_mul_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_COMPLEX:
    c[0] = (x->d.scalar.c[0] * y->d.scalar.c[0] -
	    x->d.scalar.c[1] * y->d.scalar.c[1]);
    c[1] = (x->d.scalar.c[0] * y->d.scalar.c[1] +
	    x->d.scalar.c[1] * y->d.scalar.c[0]);
    x->d.scalar.c[0] = c[0];
    x->d.scalar.c[1] = c[1];
    break;
  case SILC_DOUBLE_COMPLEX:
    z[0] = (x->d.scalar.z[0] * y->d.scalar.z[0] -
	    x->d.scalar.z[1] * y->d.scalar.z[1]);
    z[1] = (x->d.scalar.z[0] * y->d.scalar.z[1] +
	    x->d.scalar.z[1] * y->d.scalar.z[0]);
    x->d.scalar.z[0] = z[0];
    x->d.scalar.z[1] = z[1];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_div_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
  float c[2], cdiv;
  double z[2], zdiv;
#ifdef DEBUG
  printf(MODULE_NAME ": complex_div_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_COMPLEX:
    cdiv = (y->d.scalar.c[0] * y->d.scalar.c[0] +
	    y->d.scalar.c[1] * y->d.scalar.c[1]);
    c[0] = (x->d.scalar.c[0] * y->d.scalar.c[0] +
	    x->d.scalar.c[1] * y->d.scalar.c[1]) / cdiv;
    c[1] = (x->d.scalar.c[1] * y->d.scalar.c[0] -
	    x->d.scalar.c[0] * y->d.scalar.c[1]) / cdiv;
    x->d.scalar.c[0] = c[0];
    x->d.scalar.c[1] = c[1];
    break;
  case SILC_DOUBLE_COMPLEX:
    zdiv = (y->d.scalar.z[0] * y->d.scalar.z[0] +
	    y->d.scalar.z[1] * y->d.scalar.z[1]);
    z[0] = (x->d.scalar.z[0] * y->d.scalar.z[0] +
	    x->d.scalar.z[1] * y->d.scalar.z[1]) / zdiv;
    z[1] = (x->d.scalar.z[1] * y->d.scalar.z[0] -
	    x->d.scalar.z[0] * y->d.scalar.z[1]) / zdiv;
    x->d.scalar.z[0] = z[0];
    x->d.scalar.z[1] = z[1];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_leq_scalar_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x = param[0].o;
  silc_object_t *y = param[1].o;
  float c[2], cdiv;
  double z[2], zdiv;
#ifdef DEBUG
  printf(MODULE_NAME ": complex_leq_scalar_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_COMPLEX:
    cdiv = (x->d.scalar.c[0] * x->d.scalar.c[0] +
	    x->d.scalar.c[1] * x->d.scalar.c[1]);
    c[0] = (y->d.scalar.c[0] * x->d.scalar.c[0] +
	    y->d.scalar.c[1] * x->d.scalar.c[1]) / cdiv;
    c[1] = (y->d.scalar.c[1] * x->d.scalar.c[0] -
	    y->d.scalar.c[0] * x->d.scalar.c[1]) / cdiv;
    x->d.scalar.c[0] = c[0];
    x->d.scalar.c[1] = c[1];
    break;
  case SILC_DOUBLE_COMPLEX:
    zdiv = (x->d.scalar.z[0] * x->d.scalar.z[0] +
	    x->d.scalar.z[1] * x->d.scalar.z[1]);
    z[0] = (y->d.scalar.z[0] * x->d.scalar.z[0] +
	    y->d.scalar.z[1] * x->d.scalar.z[1]) / zdiv;
    z[1] = (y->d.scalar.z[1] * x->d.scalar.z[0] -
	    y->d.scalar.z[0] * x->d.scalar.z[1]) / zdiv;
    x->d.scalar.z[0] = z[0];
    x->d.scalar.z[1] = z[1];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
multiply_vector_by_scalar(silc_mfunc_param_t *vector,
			  silc_mfunc_param_t *scalar)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k, n;

  x = vector->o; x_mods = vector->m;
  y = scalar->o; y_mods = scalar->m;
#ifdef DEBUG
  printf(MODULE_NAME ": multiply_vector_by_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.i[k] *= y->d.scalar.i;
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.l[k] *= y->d.scalar.l;
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.s[k] *= y->d.scalar.s;
    break;
  case SILC_DOUBLE:
    {
      double *dx = x->d.vector.v.d;
      double c = y->d.scalar.d;
#pragma omp parallel for
      for (k = 0; k < (ssize_t)x->d.vector.length; k++)
	dx[k] *= c;
    }
    break;
  case SILC_COMPLEX:
    {
      float real, imag;
#pragma omp parallel for private(k,real,imag)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++) {
	real = (x->d.vector.v.c[k*2]   * y->d.scalar.c[0] -
		x->d.vector.v.c[k*2+1] * y->d.scalar.c[1]);
	imag = (x->d.vector.v.c[k*2]   * y->d.scalar.c[1] +
		x->d.vector.v.c[k*2+1] * y->d.scalar.c[0]);
	x->d.vector.v.c[k*2]   = real;
	x->d.vector.v.c[k*2+1] = imag;
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    {
      double real, imag;
#pragma omp parallel for private(real,imag)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++) {
	real = (x->d.vector.v.z[k*2]   * y->d.scalar.z[0] -
		x->d.vector.v.z[k*2+1] * y->d.scalar.z[1]);
	imag = (x->d.vector.v.z[k*2]   * y->d.scalar.z[1] +
		x->d.vector.v.z[k*2+1] * y->d.scalar.z[0]);
	x->d.vector.v.z[k*2]   = real;
	x->d.vector.v.z[k*2+1] = imag;
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
mul_vector_scalar(int n_params, silc_mfunc_param_t *param)
{
  return multiply_vector_by_scalar(&param[0], &param[1]);
}

/* out: 1 */
int
mul_scalar_vector(int n_params, silc_mfunc_param_t *param)
{
  return multiply_vector_by_scalar(&param[1], &param[0]);
}

int
integer_add_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_add_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.vector.length != y->d.vector.length)
    return -1;
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.i[k] += y->d.vector.v.i[k];
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.l[k] += y->d.vector.v.l[k];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
integer_sub_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": integer_sub_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.vector.length != y->d.vector.length)
    return -1;
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.i[k] -= y->d.vector.v.i[k];
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.l[k] -= y->d.vector.v.l[k];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

void vadd(double *restrict x, double *y, ssize_t n)
{
  ssize_t k;
#pragma omp parallel for
  for (k = 0; k < n; k++)
    x[k] += y[k];
}
int
real_add_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": real_add_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.vector.length != y->d.vector.length)
    return -1;
  switch (x->precision) {
  case SILC_FLOAT:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.s[k] += y->d.vector.v.s[k];
    break;
  case SILC_DOUBLE:
#if 0
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.d[k] += y->d.vector.v.d[k];
#endif
    vadd(x->d.vector.v.d, y->d.vector.v.d, x->d.vector.length);
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
real_sub_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": real_sub_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.vector.length != y->d.vector.length)
    return -1;
  switch (x->precision) {
  case SILC_FLOAT:
#pragma omp parallel for
    for (k = 0; k < (ssize_t)x->d.vector.length; k++)
      x->d.vector.v.s[k] -= y->d.vector.v.s[k];
    break;
  case SILC_DOUBLE:
    {
      double *dx = x->d.vector.v.d;
      double *dy = y->d.vector.v.d;
#pragma omp parallel for
      for (k = 0; k < (ssize_t)x->d.vector.length; k++)
	dx[k] -= dy[k];
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_add_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k, c;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": complex_add_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.vector.length != y->d.vector.length)
    return -1;
  switch (x->precision) {
  case SILC_COMPLEX:
    c = x->d.vector.length * 2;
#pragma omp parallel for
    for (k = 0; k < c; k++)
      x->d.vector.v.c[k] += y->d.vector.v.c[k];
    break;
  case SILC_DOUBLE_COMPLEX:
    c = x->d.vector.length * 2;
#pragma omp parallel for
    for (k = 0; k < c; k++)
      x->d.vector.v.z[k] += y->d.vector.v.z[k];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

int
complex_sub_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k, c;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": complex_sub_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.vector.length != y->d.vector.length)
    return -1;
  switch (x->precision) {
  case SILC_COMPLEX:
    c = x->d.vector.length * 2;
#pragma omp parallel for
    for (k = 0; k < c; k++)
      x->d.vector.v.c[k] -= y->d.vector.v.c[k];
    break;
  case SILC_DOUBLE_COMPLEX:
    c = x->d.vector.length * 2;
#pragma omp parallel for
    for (k = 0; k < c; k++)
      x->d.vector.v.z[k] -= y->d.vector.v.z[k];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
mul_vector_vector_elementwise(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k, n;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": mul_vector_vector_elementwise(0x%lx, 0x%lx)\n", x, y);
#endif
  n = (ssize_t)x->d.vector.length;
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (k = 0; k < n; k++)
      x->d.vector.v.i[k] *= y->d.vector.v.i[k];
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (k = 0; k < n; k++)
      x->d.vector.v.l[k] *= y->d.vector.v.l[k];
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (k = 0; k < n; k++)
      x->d.vector.v.s[k] *= y->d.vector.v.s[k];
    break;
  case SILC_DOUBLE:
    {
      double *px = x->d.vector.v.d;
      double *py = y->d.vector.v.d;
#pragma omp parallel for
      for (k = 0; k < n; k++)
	px[k] *= py[k];
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (k = 0; k < n; k++) {
      float real, imag;
      real = (x->d.vector.v.c[k*2]   * y->d.vector.v.c[0] -
              x->d.vector.v.c[k*2+1] * y->d.vector.v.c[1]);
      imag = (x->d.vector.v.c[k*2]   * y->d.vector.v.c[1] +
              x->d.vector.v.c[k*2+1] * y->d.vector.v.c[0]);
      x->d.vector.v.c[k*2]   = real;
      x->d.vector.v.c[k*2+1] = imag;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (k = 0; k < n; k++) {
      double real, imag;
      real = (x->d.vector.v.z[k*2]   * y->d.vector.v.z[0] -
              x->d.vector.v.z[k*2+1] * y->d.vector.v.z[1]);
      imag = (x->d.vector.v.z[k*2]   * y->d.vector.v.z[1] +
              x->d.vector.v.z[k*2+1] * y->d.vector.v.z[0]);
      x->d.vector.v.z[k*2]   = real;
      x->d.vector.v.z[k*2+1] = imag;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
div_vector_vector_elementwise(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t k, n;
  int zero;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": div_vector_vector_elementwise(0x%lx, 0x%lx)\n", x, y);
#endif
  n = (ssize_t)x->d.vector.length;
  zero = 0;
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for reduction(||:zero)
    for (k = 0; k < n; k++) {
      zero = zero || (y->d.vector.v.i[k] == 0);
      if (!zero)
        x->d.vector.v.i[k] /= y->d.vector.v.i[k];
    }
    break;
  case SILC_LONG:
#pragma omp parallel for reduction(||:zero)
    for (k = 0; k < n; k++) {
      zero = zero || (y->d.vector.v.l[k] == 0);
      if (!zero)
        x->d.vector.v.l[k] /= y->d.vector.v.l[k];
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for reduction(||:zero)
    for (k = 0; k < n; k++) {
      zero = zero || (y->d.vector.v.s[k] == 0.0);
      if (!zero)
        x->d.vector.v.s[k] /= y->d.vector.v.s[k];
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for reduction(||:zero)
    for (k = 0; k < n; k++) {
      zero = zero || (y->d.vector.v.d[k] == 0.0);
      if (!zero)
        x->d.vector.v.d[k] /= y->d.vector.v.d[k];
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for reduction(||:zero)
    for (k = 0; k < n; k++) {
      float div, real, imag;
      div = (y->d.vector.v.c[k*2]   * y->d.vector.v.c[k*2] +
             y->d.vector.v.c[k*2+1] * y->d.vector.v.c[k*2+1]);
      zero = zero || (div == 0.0);
      if (!zero) {
        real = (x->d.vector.v.c[k*2]   * y->d.vector.v.c[k*2] +
                x->d.vector.v.c[k*2+1] * y->d.vector.v.c[k*2+1]) / div;
        imag = (x->d.vector.v.c[k*2+1] * y->d.vector.v.c[k*2] -
                x->d.vector.v.c[k*2]   * y->d.vector.v.c[k*2+1]) / div;
        x->d.vector.v.c[k*2]   = real;
        x->d.vector.v.c[k*2+1] = imag;
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for reduction(||:zero)
    for (k = 0; k < n; k++) {
      double div, real, imag;
      div = (y->d.vector.v.z[k*2]   * y->d.vector.v.z[k*2] +
             y->d.vector.v.z[k*2+1] * y->d.vector.v.z[k*2+1]);
      zero = zero || (div == 0.0);
      if (!zero) {
        real = (x->d.vector.v.z[k*2]   * y->d.vector.v.z[k*2] +
                x->d.vector.v.z[k*2+1] * y->d.vector.v.z[k*2+1]) / div;
        imag = (x->d.vector.v.z[k*2+1] * y->d.vector.v.z[k*2] -
                x->d.vector.v.z[k*2]   * y->d.vector.v.z[k*2+1]) / div;
        x->d.vector.v.z[k*2]   = real;
        x->d.vector.v.z[k*2+1] = imag;
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  if (zero) {
    fprintf(stderr, "div_vector_vector_elementwise: division by zero\n");
    return -1;
  }
  return 0;
}

/* dot(), sqrt(), norm2() */

static silc_object_t *
_dot(silc_object_t *x, silc_object_t *y)
{
  silc_object_t *v;
  ssize_t k, n;

  switch (x->precision) {
  case SILC_INT:
    {
      int32_t i = 0;
#pragma omp parallel for reduction(+:i)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++)
	i += x->d.vector.v.i[k] * y->d.vector.v.i[k];
      v = silc_object_new_int(i);
    }
    break;
  case SILC_LONG:
    {
      int64_t l = 0;
#pragma omp parallel for reduction(+:l)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++)
	l += x->d.vector.v.l[k] * y->d.vector.v.l[k];
      v = silc_object_new_long(l);
    }
    break;
  case SILC_FLOAT:
    {
      float s = 0.0;
#pragma omp parallel for reduction(+:s)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++)
	s += x->d.vector.v.s[k] * y->d.vector.v.s[k];
      v = silc_object_new_float(s);
    }
    break;
  case SILC_DOUBLE:
    {
      double d, *dx, *dy;
      d = 0.0;
      dx = x->d.vector.v.d;
      dy = y->d.vector.v.d;
#pragma omp parallel for reduction(+:d)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++)
	d += dx[k] * dy[k];
      v = silc_object_new_double(d);
    }
    break;
  case SILC_COMPLEX:
    {
      float real = 0.0;
      float imag = 0.0;
#pragma omp parallel for reduction(+:real,imag)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++) {
	n = k * 2;
	real += (x->d.vector.v.c[n]   * y->d.vector.v.c[n]   -
		 x->d.vector.v.c[n+1] * y->d.vector.v.c[n+1]);
	imag += (x->d.vector.v.c[n]   * y->d.vector.v.c[n+1] +
		 x->d.vector.v.c[n+1] * y->d.vector.v.c[n]);
      }
      v = silc_object_new_complex(real, imag);
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    {
      double real = 0.0;
      double imag = 0.0;
#pragma omp parallel for reduction(+:real,imag)
      for (k = 0; k < (ssize_t)x->d.vector.length; k++) {
	n = k * 2;
	real += (x->d.vector.v.z[n]   * y->d.vector.v.z[n]   -
		 x->d.vector.v.z[n+1] * y->d.vector.v.z[n+1]);
	imag += (x->d.vector.v.z[n]   * y->d.vector.v.z[n+1] +
		 x->d.vector.v.z[n+1] * y->d.vector.v.z[n]);
      }
      v = silc_object_new_double_complex(real, imag);
    }
    break;
  default:
    should_not_reach_here;
    return NULL;
  }
  return v;
}

/* out: 2 */
int
dot_product(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf(MODULE_NAME ": dot_product(0x%lx, 0x%lx)\n", x, y);
#endif
  param[2].o = _dot(x, y);
  return (param[2].o) ? 0 : -1;
}

/* complex_sqrt.c */
void c_sqrt(float *, float *);
void z_sqrt(double *, double *);

static silc_object_t *
_sqrt(silc_object_t *object)
{
  silc_object_t *v;

  if (object->type != SILC_SCALAR_TYPE)
    return NULL;
  switch (object->precision) {
  case SILC_INT:
    if (object->d.scalar.i < 0)
      v = silc_object_new_complex(0, sqrt(-object->d.scalar.i));
    else
      v = silc_object_new_float(sqrt(object->d.scalar.i));
    break;
  case SILC_LONG:
    if (object->d.scalar.l < 0)
      v = silc_object_new_double_complex(0, sqrt(-object->d.scalar.l));
    else
      v = silc_object_new_double(sqrt(object->d.scalar.l));
    break;
  case SILC_FLOAT:
    if (object->d.scalar.s < 0)
      v = silc_object_new_complex(0, sqrt(-object->d.scalar.s));
    else
      v = silc_object_new_float(sqrt(object->d.scalar.s));
    break;
  case SILC_DOUBLE:
    if (object->d.scalar.d < 0)
      v = silc_object_new_double_complex(0, sqrt(-object->d.scalar.d));
    else
      v = silc_object_new_double(sqrt(object->d.scalar.d));
    break;
  case SILC_COMPLEX:
    {
      float c[2];
      c_sqrt(object->d.scalar.c, c);
      v = silc_object_new_complex(c[0], c[1]);
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    {
      double z[2];
      z_sqrt(object->d.scalar.z, z);
      v = silc_object_new_double_complex(z[0], z[1]);
    }
    break;
  default:
    should_not_reach_here;
    return NULL;
  }
  return v;
}

int
func_sqrt(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *v;

  object = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_sqrt(0x%lx)\n", object);
#endif
  param[1].o = _sqrt(object);
  return (param[1].o) ? 0 : -1;
}

int
func_sqrtv(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *v;
  int negative, precision;
  ssize_t k, n;

  object = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_sqrtv(0x%lx)\n", object);
#endif
  if (object->type != SILC_COLUMN_VECTOR_TYPE &&
      object->type != SILC_ROW_VECTOR_TYPE)
    return -1;
  n = object->d.vector.length;

#define _sqrtv_1(pr1, pr2) { \
    v->d.vector.v.pr2[k] = sqrt(object->d.vector.v.pr1[k]);		\
  }
#define _sqrtv_2(pr1, pr2) { \
    if (object->d.vector.v.pr1[k] < 0) {				\
      v->d.vector.v.pr2[k*2]   = 0;					\
      v->d.vector.v.pr2[k*2+1] = sqrt(-object->d.vector.v.pr1[k]);	\
    } else {								\
      v->d.vector.v.pr2[k*2]   = sqrt(object->d.vector.v.pr1[k]);	\
      v->d.vector.v.pr2[k*2+1] = 0;					\
    }									\
  }

  switch (object->precision) {
  case SILC_INT:
    negative = 0;
#pragma omp parallel
    {
#pragma omp for reduction(||:negative)
      for (k = 0; k < n; k++)
	negative = negative || (object->d.vector.v.i[k] < 0);
#pragma omp single
      {
	precision = (!negative) ? SILC_FLOAT : SILC_COMPLEX;
	if (object->type == SILC_COLUMN_VECTOR_TYPE)
	  v = silc_object_new_column_vector(precision, n);
	else
	  v = silc_object_new_row_vector(precision, n);
      }
      if (!negative) {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_1(i, s);
      } else {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_2(i, c);
      }
    }
    break;
  case SILC_LONG:
    negative = 0;
#pragma omp parallel
    {
#pragma omp for reduction(||:negative)
      for (k = 0; k < n; k++)
	negative = negative || (object->d.vector.v.l[k] < 0);
#pragma omp single
      {
	precision = (!negative) ? SILC_DOUBLE : SILC_DOUBLE_COMPLEX;
	if (object->type == SILC_COLUMN_VECTOR_TYPE)
	  v = silc_object_new_column_vector(precision, n);
	else
	  v = silc_object_new_row_vector(precision, n);
      }
      if (!negative) {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_1(l, d);
      } else {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_2(l, z);
      }
    }
    break;
  case SILC_FLOAT:
    negative = 0;
#pragma omp parallel
    {
#pragma omp for reduction(||:negative)
      for (k = 0; k < n; k++)
	negative = negative || (object->d.vector.v.s[k] < 0);
#pragma omp single
      {
	precision = (!negative) ? SILC_FLOAT : SILC_COMPLEX;
	if (object->type == SILC_COLUMN_VECTOR_TYPE)
	  v = silc_object_new_column_vector(precision, n);
	else
	  v = silc_object_new_row_vector(precision, n);
      }
      if (!negative) {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_1(s, s);
      } else {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_2(s, c);
      }
    }
    break;
  case SILC_DOUBLE:
    negative = 0;
#pragma omp parallel
    {
#pragma omp for reduction(||:negative)
      for (k = 0; k < n; k++)
	negative = negative || (object->d.vector.v.d[k] < 0);
#pragma omp single
      {
	precision = (!negative) ? SILC_DOUBLE : SILC_DOUBLE_COMPLEX;
	if (object->type == SILC_COLUMN_VECTOR_TYPE)
	  v = silc_object_new_column_vector(precision, n);
	else
	  v = silc_object_new_row_vector(precision, n);
      }
      if (!negative) {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_1(d, d);
      } else {
#pragma omp for
	for (k = 0; k < n; k++) _sqrtv_2(d, z);
      }
    }
    break;
  case SILC_COMPLEX:
    v = silc_object_duplicate(object, 0);
#pragma omp parallel for
    for (k = 0; k < n; k++)
      c_sqrt(object->d.vector.v.c + k * 2, v->d.vector.v.c + k * 2);
    break;
  case SILC_DOUBLE_COMPLEX:
    v = silc_object_duplicate(object, 0);
#pragma omp parallel for
    for (k = 0; k < n; k++)
      z_sqrt(object->d.vector.v.z + k * 2, v->d.vector.v.z + k * 2);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  param[1].o = v;
  return (param[1].o) ? 0 : -1;
}

int
func_norm2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *v;

  x = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_norm2(0x%lx)\n", x);
#endif
  switch (x->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    v = _dot(x, x);
    break;
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    y = _transpose(x); _complex_conjugate(y);
    v = _dot(y, x);
    silc_object_free(y);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  if (!v)
    return -1;
  param[1].o = _sqrt(v); /* XXX: optimize here */
  silc_object_free(v);
  return (param[1].o) ? 0 : -1;
}

int
func_length(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;

  object = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_length(0x%lx)\n", object);
#endif
  param[1].o = silc_object_new_int(object->d.vector.length);
  return (param[1].o) ? 0 : -1;
}

int
func_time(int n_params, silc_mfunc_param_t *param)
{
#ifdef DEBUG
  printf(MODULE_NAME ": func_time()\n");
#endif
  param[0].o = silc_object_new_double(silc_wtime());
  return (param[0].o) ? 0 : -1;
}

int
func_max(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  ssize_t i;
  union _max {
    int32_t i;
    int64_t l;
    float s, c[2];
    double d, z[2];
  } *p;
  int num_threads;

  object = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": func_max(0x%lx)\n", object);
#endif
  if (object->type == SILC_COLUMN_VECTOR_TYPE ||
      object->type == SILC_ROW_VECTOR_TYPE) {
#ifdef _OPENMP
    num_threads = omp_get_max_threads();
#else
    num_threads = 1;
#endif
    p = (union _max *)malloc(sizeof(union _max) * num_threads);
    if (!p) {
      fprintf(stderr, "max: out of memory (p)\n");
      return -1;
    }
#pragma omp parallel
    {
      int n;

#ifdef _OPENMP
      n = omp_get_thread_num();
#else
      n = 0;
#endif
#define _max_1(pre) {						\
	if (p[n].pre < object->d.vector.v.pre[i])		\
	  p[n].pre = object->d.vector.v.pre[i];			\
      }
#define _max_2(pre, type) {					\
        size_t ii;						\
	type re, im, mag2;					\
								\
	ii = i * 2;						\
	re = object->d.vector.v.pre[ii];			\
	im = object->d.vector.v.pre[ii+1];			\
	mag2 = re * re + im * im;				\
	if (mag < mag2) {					\
	  p[n].pre[0] = re;					\
	  p[n].pre[1] = im;					\
	  mag = mag2;						\
	}							\
      }
      switch (object->precision) {
      case SILC_INT:
	p[n].i = object->d.vector.v.i[0];
#pragma omp for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  _max_1(i);
	break;
      case SILC_LONG:
	p[n].l = object->d.vector.v.l[0];
#pragma omp for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  _max_1(l);
	break;
      case SILC_FLOAT:
	p[n].s = object->d.vector.v.s[0];
#pragma omp for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  _max_1(s);
	break;
      case SILC_DOUBLE:
	p[n].d = object->d.vector.v.d[0];
#pragma omp for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  _max_1(d);
	break;
      case SILC_COMPLEX:
	{
	  float mag;

	  p[n].c[0] = object->d.vector.v.c[0];
	  p[n].c[1] = object->d.vector.v.c[1];
	  mag = p[n].c[0] * p[n].c[0] + p[n].c[1] * p[n].c[1];
#pragma omp for
	  for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	    _max_2(c, float);
	}
	break;
      case SILC_DOUBLE_COMPLEX:
	{
	  double mag;

	  p[n].z[0] = object->d.vector.v.z[0];
	  p[n].z[1] = object->d.vector.v.z[1];
	  mag = p[n].z[0] * p[n].z[0] + p[n].z[1] * p[n].z[1];
#pragma omp for
	  for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	    _max_2(z, double);
	}
	break;
      default:
	should_not_reach_here;
      }
    }
    new_object = silc_object_new_scalar(object->precision);
    if (new_object) {
#define _max_reduce_1(pre) {						\
	if (p[0].pre < p[i].pre)					\
	  p[0].pre = p[i].pre;						\
      }
#define _max_reduce_2(pre, type) {					\
	type mag2;							\
									\
	mag2 = p[i].pre[0] * p[i].pre[0] + p[i].pre[1] * p[i].pre[1];	\
	if (mag < mag2) {						\
	  p[0].pre[0] = p[i].pre[0];					\
	  p[0].pre[1] = p[i].pre[1];					\
	  mag = mag2;							\
	}								\
      }
      switch (object->precision) {
      case SILC_INT:
	for (i = 1; i < num_threads; i++)
	  _max_reduce_1(i);
	new_object->d.scalar.i = p[0].i;
	break;
      case SILC_LONG:
	for (i = 1; i < num_threads; i++)
	  _max_reduce_1(l);
	new_object->d.scalar.l = p[0].l;
	break;
      case SILC_FLOAT:
	for (i = 1; i < num_threads; i++)
	  _max_reduce_1(s);
	new_object->d.scalar.s = p[0].s;
	break;
      case SILC_DOUBLE:
	for (i = 1; i < num_threads; i++)
	  _max_reduce_1(d);
	new_object->d.scalar.d = p[0].d;
	break;
      case SILC_COMPLEX:
	{
	  float mag;

	  mag = p[0].c[0] * p[0].c[0] + p[0].c[1] * p[0].c[1];
	  for (i = 1; i < num_threads; i++)
	    _max_reduce_2(c, float);
	}
	new_object->d.scalar.c[0] = p[0].c[0];
	new_object->d.scalar.c[1] = p[0].c[1];
	break;
      case SILC_DOUBLE_COMPLEX:
	{
	  double mag;

	  mag = p[0].z[0] * p[0].z[0] + p[0].z[1] * p[0].z[1];
	  for (i = 1; i < num_threads; i++)
	    _max_reduce_2(z, double);
	}
	new_object->d.scalar.z[0] = p[0].z[0];
	new_object->d.scalar.z[1] = p[0].z[1];
	break;
      default:
	should_not_reach_here;
      }
    }
    free(p);
    param[1].o = new_object;
  } else if (object->type == SILC_SCALAR_TYPE) {
    param[1].o = silc_object_duplicate(object, 1);
  } else {
    return -1;
  }
  return (param[0].o) ? 0 : -1;
}

#define SILC_ANY_INTEGER   (SILC_INT|SILC_LONG)
#define SILC_ANY_REAL      (SILC_FLOAT|SILC_DOUBLE)
#define SILC_ANY_COMPLEX   (SILC_COMPLEX|SILC_DOUBLE_COMPLEX)
#define SILC_ANY_PRECISION (SILC_ANY_INTEGER|SILC_ANY_REAL|SILC_ANY_COMPLEX)

#define SILC_ANY_VECTOR_TYPE \
       (SILC_COLUMN_VECTOR_TYPE|SILC_ROW_VECTOR_TYPE)

static unary_op_table_t _unary_op_table[] = {
  /* name, op, precision, type, format, out, flags */
  {negate, "negate", SILC_NEGATE_OP,
   SILC_ANY_PRECISION,
   SILC_SCALAR_TYPE|SILC_ANY_VECTOR_TYPE|SILC_CUBIC_ARRAY_TYPE, 0, 0, 0},
  {complex_conjugate, "complex_conjugate", SILC_COMPLEX_CONJUGATE_OP,
   SILC_ANY_PRECISION,   
   SILC_SCALAR_TYPE|SILC_ANY_VECTOR_TYPE|SILC_CUBIC_ARRAY_TYPE, 0, 0, 0},
  {transpose, "transpose", SILC_TRANSPOSE_OP,
   SILC_ANY_PRECISION,
   SILC_SCALAR_TYPE|SILC_ANY_VECTOR_TYPE, 0, 1, 0},
  {conjugate_transpose, "conjugate_transpose", SILC_CONJUGATE_TRANSPOSE_OP,
   SILC_ANY_PRECISION,
   SILC_SCALAR_TYPE|SILC_ANY_VECTOR_TYPE, 0, 1, 0},
  {NULL} /* sentinel */
};

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {integer_add_scalar_scalar, "integer_add_scalar_scalar", SILC_ADD_OP,
   SILC_ANY_INTEGER, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {integer_sub_scalar_scalar, "integer_sub_scalar_scalar", SILC_SUBTRACT_OP,
   SILC_ANY_INTEGER, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {integer_mul_scalar_scalar, "integer_mul_scalar_scalar", SILC_MULTIPLY_OP,
   SILC_ANY_INTEGER, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {integer_div_scalar_scalar, "integer_div_scalar_scalar", SILC_DIVIDE_OP,
   SILC_ANY_INTEGER, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {integer_mod_scalar_scalar, "integer_mod_scalar_scalar", SILC_REMAINDER_OP,
   SILC_ANY_INTEGER, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {integer_leq_scalar_scalar, "integer_leq_scalar_scalar", SILC_LEQ_OP,
   SILC_ANY_INTEGER, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {real_add_scalar_scalar, "real_add_scalar_scalar", SILC_ADD_OP,
   SILC_ANY_REAL, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {real_sub_scalar_scalar, "real_sub_scalar_scalar", SILC_SUBTRACT_OP,
   SILC_ANY_REAL, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {real_mul_scalar_scalar, "real_mul_scalar_scalar", SILC_MULTIPLY_OP,
   SILC_ANY_REAL, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {real_div_scalar_scalar, "real_div_scalar_scalar", SILC_DIVIDE_OP,
   SILC_ANY_REAL, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {real_leq_scalar_scalar, "real_leq_scalar_scalar", SILC_LEQ_OP,
   SILC_ANY_REAL, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {complex_add_scalar_scalar, "complex_add_scalar_scalar", SILC_ADD_OP,
   SILC_ANY_COMPLEX, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {complex_sub_scalar_scalar, "complex_sub_scalar_scalar", SILC_SUBTRACT_OP,
   SILC_ANY_COMPLEX, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {complex_mul_scalar_scalar, "complex_mul_scalar_scalar", SILC_MULTIPLY_OP,
   SILC_ANY_COMPLEX, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {complex_div_scalar_scalar, "complex_div_scalar_scalar", SILC_DIVIDE_OP,
   SILC_ANY_COMPLEX, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {complex_leq_scalar_scalar, "complex_leq_scalar_scalar", SILC_LEQ_OP,
   SILC_ANY_COMPLEX, SILC_SCALAR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},

  {mul_vector_scalar, "mul_vector_scalar", SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_ANY_VECTOR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {mul_scalar_vector, "mul_scalar_vector", SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_SCALAR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 1, 0},

  {integer_add_vector_vector, "integer_add_vector_vector", SILC_ADD_OP,
   SILC_ANY_INTEGER, SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},
  {integer_sub_vector_vector, "integer_sub_vector_vector", SILC_SUBTRACT_OP,
   SILC_ANY_INTEGER, SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},
  {real_add_vector_vector, "real_add_vector_vector", SILC_ADD_OP,
   SILC_ANY_REAL, SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},
  {real_sub_vector_vector, "real_sub_vector_vector", SILC_SUBTRACT_OP,
   SILC_ANY_REAL, SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},
  {complex_add_vector_vector, "complex_add_vector_vector", SILC_ADD_OP,
   SILC_ANY_COMPLEX, SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},
  {complex_sub_vector_vector, "complex_sub_vector_vector", SILC_SUBTRACT_OP,
   SILC_ANY_COMPLEX, SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},

  {mul_vector_vector_elementwise, "mul_vector_vector_elementwise",
   SILC_MULTIPLY_ELEMENTWISE_OP, SILC_ANY_PRECISION,
   SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},
  {div_vector_vector_elementwise, "div_vector_vector_elementwise",
   SILC_DIVIDE_ELEMENTWISE_OP, SILC_ANY_PRECISION,
   SILC_ANY_VECTOR_TYPE, 0, SILC_ANY_VECTOR_TYPE, 0, 0, 0},

  {dot_product, "dot_product", SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_ROW_VECTOR_TYPE, 0, SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},

  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {dot_product, "dot_product", "scalar dot(row vector, column vector)", 0},
  {func_sqrt, "func_sqrt", "scalar sqrt(scalar)", 0},
  {func_sqrtv, "func_sqrtv", "vector sqrt(vector)", 0},
  {func_norm2, "func_norm2", "scalar norm2(vector)", 0},
  {func_length, "func_length", "scalar length(vector)", 0},
  {func_time, "func_time", "scalar time()", 0},
  {func_max, "func_max", "scalar max(scalar)", 0},
  {func_max, "func_max", "scalar max(vector)", 0},
  {NULL} /* sentinel */
};

static procedure_table_t _procedure_table[] = {
  /* name, sig_text, flags */

  {NULL} /* sentinel */
};

void
init_coremodule()
{
  silc_register_module(MODULE_NAME,
		       _unary_op_table,
		       _binary_op_table,
		       _function_table,
		       NULL);  /* _procedure_table */
}
