/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: dense_format.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "object.h"
#include "dense.h"
#include "module.h"
#include "net.h"

#define MODULE_NAME "dense"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

/* proc_new: create a zero matrix */
static int
proc_new(silc_object_t *object)
{
  silc_dense_t *in;
  size_t m;
  ssize_t i, n;

  in = silc_dense_allocate(object->precision, object->d.matrix.m, object->d.matrix.n);
  if (!in)
    return -1;
  m = object->d.matrix.m;
  n = object->d.matrix.n;
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      memset(in->v.i + m * i, 0, m * sizeof(int32_t));
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      memset(in->v.l + m * i, 0, m * sizeof(int64_t));
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      memset(in->v.s + m * i, 0, m * sizeof(float));
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      memset(in->v.d + m * i, 0, m * sizeof(double));
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      memset(in->v.c + 2 * m * i, 0, m * sizeof(float) * 2);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      memset(in->v.z + 2 * m * i, 0, m * sizeof(double) * 2);
    break;
  default:
    should_not_reach_here;
  }
  object->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_recv(int sock, silc_object_t *object)
{
  silc_dense_t *in;
  size_t size;

  in = silc_dense_allocate(object->precision, object->d.matrix.m, object->d.matrix.n);
  if (!in)
    return -1;
  size = object->d.matrix.m * object->d.matrix.n;
  switch (object->precision) {
  case SILC_INT:
    recv_32_array(sock, in->v.i, size, 0); break;
  case SILC_LONG:
    recv_64_array(sock, in->v.l, size, 0); break;
  case SILC_FLOAT:
    recv_32_array(sock, in->v.s, size, 0); break;
  case SILC_DOUBLE:
    recv_64_array(sock, in->v.d, size, 0); break;
  case SILC_COMPLEX:
    recv_32_array(sock, in->v.c, 2 * size, 0); break;
  case SILC_DOUBLE_COMPLEX:
    recv_64_array(sock, in->v.z, 2 * size, 0); break;
  default:
    should_not_reach_here;
  }
  object->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_send(int sock, silc_object_t *object)
{
  size_t size;

  size = object->d.matrix.m * object->d.matrix.n;
  switch (object->precision) {
  case SILC_INT:
    send_32_array(sock, SILC_DENSE(object)->v.i, size, 0); break;
  case SILC_LONG:
    send_64_array(sock, SILC_DENSE(object)->v.l, size, 0); break;
  case SILC_FLOAT:
    send_32_array(sock, SILC_DENSE(object)->v.s, size, 0); break;
  case SILC_DOUBLE:
    send_64_array(sock, SILC_DENSE(object)->v.d, size, 0); break;
  case SILC_COMPLEX:
    send_32_array(sock, SILC_DENSE(object)->v.c, 2 * size, 0); break;
  case SILC_DOUBLE_COMPLEX:
    send_64_array(sock, SILC_DENSE(object)->v.z, 2 * size, 0); break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_free(silc_object_t *object)
{
  silc_dense_t *in;

  in = SILC_DENSE(object);
  if (!in)
    return -1;
  free(in->v.i);
  free(in);
  return 0;
}

static int
proc_dup(silc_object_t *object, silc_object_t *duplicated, int copy)
{
  silc_dense_t *in;

  in = silc_dense_allocate(object->precision, object->d.matrix.m, object->d.matrix.n);
  if (!in)
    return -1;
  if (copy)
    memcpy(in->v.i, SILC_DENSE(object)->v.i,
	   SILC_SCALAR_SIZE(object) * object->d.matrix.m * object->d.matrix.n);
  duplicated->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_pconv(silc_object_t *object, int new_precision)
{
  silc_dense_t *in;
  size_t size;
  int i;
  
#define _pconv11(src,dest,type) { \
    type *buf;						\
    buf = (type *)malloc(sizeof(type) * size);		\
    for (i = 0; i < size; i++)				\
      buf[i] = (type)in->v.src[i];			\
    free(in->v.src);					\
    in->v.dest = buf;					\
  }
#define _pconv12(src,dest,type) { \
    type *buf;						\
    buf = (type *)malloc(sizeof(type) * 2 * size);	\
    for (i = 0; i < size; i++) {			\
      buf[i*2]	 = (type)in->v.src[i];			\
      buf[i*2+1] = (type)0;				\
    }							\
    free(in->v.src);					\
    in->v.dest = buf;					\
  }
#define _pconv22(src,dest,type) { \
    type *buf;						\
    buf = (type *)malloc(sizeof(type) * 2 * size);	\
    for (i = 0; i < size; i++) {			\
      buf[i*2]	 = (type)in->v.src[i*2];		\
      buf[i*2+1] = (type)in->v.src[i*2+1];		\
    }							\
    free(in->v.src);					\
    in->v.dest = buf;					\
  }
 
  in = SILC_DENSE(object);
  size = object->d.matrix.m * object->d.matrix.n;
  switch (object->precision) {
  case SILC_INT:
    switch (new_precision) {
    case SILC_LONG:
      _pconv11(i, l, int64_t); break;
    case SILC_FLOAT:
      _pconv11(i, s, float); break;
    case SILC_DOUBLE:
      _pconv11(i, d, double); break;
    case SILC_COMPLEX:
      _pconv12(i, c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(i, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (new_precision) {
    case SILC_DOUBLE:
      _pconv11(l, d, double); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(l, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (new_precision) {
    case SILC_DOUBLE:
      _pconv11(s, d, double); break;
    case SILC_COMPLEX:
      _pconv12(s, c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(s, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _pconv12(d, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _pconv22(c, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static silc_object_t *
proc_subscript(silc_object_t *object, clist_t *subscript[])
{
  silc_object_t *modified, *temp;
  silc_dense_t *in, *new_in;
  int c, i, j, k, length0, length1, *v0, *v1, *index;

  v0 = clist_expand(subscript[0]);
  v1 = clist_expand(subscript[1]);
  length0 = clist_length(subscript[0]);
  length1 = clist_length(subscript[1]);
  in = SILC_DENSE(object);
  if (length0 == 1 && length1 == 1) {
    i = v0[0] + v1[0] * object->d.matrix.m;
    switch (object->precision) {
    case SILC_INT:
      modified = silc_object_new_int(in->v.i[i]); break;
    case SILC_LONG:
      modified = silc_object_new_long(in->v.l[i]); break;
    case SILC_FLOAT:
      modified = silc_object_new_float(in->v.s[i]); break;
    case SILC_DOUBLE:
      modified = silc_object_new_double(in->v.d[i]); break;
    case SILC_COMPLEX:
      i = i * 2;
      modified = silc_object_new_complex(in->v.c[i], in->v.c[i+1]);
      break;
    case SILC_DOUBLE_COMPLEX:
      i = i * 2;
      modified = silc_object_new_double_complex(in->v.z[i], in->v.z[i+1]);
      break;
    default:
      should_not_reach_here;
    }
  } else if (length0 == 1) {
    modified = silc_object_new_row_vector(object->precision, length1);
    switch (object->precision) {
    case SILC_INT:
      for (i = 0; i < length1; i++) {
	j = v0[0] + v1[i] * object->d.matrix.m;
	modified->d.vector.v.i[i] = in->v.i[j];
      }
      break;
    case SILC_LONG:
      for (i = 0; i < length1; i++) {
	j = v0[0] + v1[i] * object->d.matrix.m;
	modified->d.vector.v.l[i] = in->v.l[j];
      }
      break;
    case SILC_FLOAT:
      for (i = 0; i < length1; i++) {
	j = v0[0] + v1[i] * object->d.matrix.m;
	modified->d.vector.v.s[i] = in->v.s[j];
      }
      break;
    case SILC_DOUBLE:
      for (i = 0; i < length1; i++) {
	j = v0[0] + v1[i] * object->d.matrix.m;
	modified->d.vector.v.d[i] = in->v.d[j];
      }
      break;
    case SILC_COMPLEX:
      c = 0;
      for (i = 0; i < length1; i++) {
	j = (v0[0] + v1[i] * object->d.matrix.m) * 2;
	modified->d.vector.v.c[c++] = in->v.c[j];
	modified->d.vector.v.c[c++] = in->v.c[j+1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      c = 0;
      for (i = 0; i < length1; i++) {
	j = (v0[0] + v1[i] * object->d.matrix.m) * 2;
	modified->d.vector.v.z[c++] = in->v.z[j];
	modified->d.vector.v.z[c++] = in->v.z[j+1];
      }
      break;
    default:
      should_not_reach_here;
    }
  } else if (length1 == 1) {
    modified = silc_object_new_column_vector(object->precision, length0);
    switch (object->precision) {
    case SILC_INT:
      for (i = 0; i < length0; i++) {
	j = v0[i] + v1[0] * object->d.matrix.m;
	modified->d.vector.v.i[i] = in->v.i[j];
      }
      break;
    case SILC_LONG:
      for (i = 0; i < length0; i++) {
	j = v0[i] + v1[0] * object->d.matrix.m;
	modified->d.vector.v.l[i] = in->v.l[j];
      }
      break;
    case SILC_FLOAT:
      for (i = 0; i < length0; i++) {
	j = v0[i] + v1[0] * object->d.matrix.m;
	modified->d.vector.v.s[i] = in->v.s[j];
      }
      break;
    case SILC_DOUBLE:
      for (i = 0; i < length0; i++) {
	j = v0[i] + v1[0] * object->d.matrix.m;
	modified->d.vector.v.d[i] = in->v.d[j];
      }
      break;
    case SILC_COMPLEX:
      c = 0;
      for (i = 0; i < length0; i++) {
	j = (v0[i] + v1[0] * object->d.matrix.m) * 2;
	modified->d.vector.v.c[c++] = in->v.c[j];
	modified->d.vector.v.c[c++] = in->v.c[j+1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      c = 0;
      for (i = 0; i < length0; i++) {
	j = (v0[i] + v1[0] * object->d.matrix.m) * 2;
	modified->d.vector.v.z[c++] = in->v.z[j];
	modified->d.vector.v.z[c++] = in->v.z[j+1];
      }
      break;
    default:
      should_not_reach_here;
    }
  } else {
    modified = silc_object_new_matrix(object->precision, object->format,
				      length0, length1); /* m, n */
    modified->d.matrix.in = (void *)silc_dense_allocate(modified->precision,
							modified->d.matrix.m,
							modified->d.matrix.n);
    new_in = SILC_DENSE(modified);
    c = 0;
    switch (object->precision) {
    case SILC_INT:
      for (j = 0; j < length1; j++) {
	for (i = 0; i < length0; i++) {
	  k = v0[i] + v1[j] * object->d.matrix.m;
	  new_in->v.i[c++] = in->v.i[k];
	}
      }
      break;
    case SILC_LONG:
      for (j = 0; j < length1; j++) {
	for (i = 0; i < length0; i++) {
	  k = v0[i] + v1[j] * object->d.matrix.m;
	  new_in->v.l[c++] = in->v.l[k];
	}
      }
      break;
    case SILC_FLOAT:
      for (j = 0; j < length1; j++) {
	for (i = 0; i < length0; i++) {
	  k = v0[i] + v1[j] * object->d.matrix.m;
	  new_in->v.s[c++] = in->v.s[k];
	}
      }
      break;
    case SILC_DOUBLE:
      for (j = 0; j < length1; j++) {
	for (i = 0; i < length0; i++) {
	  k = v0[i] + v1[j] * object->d.matrix.m;
	  new_in->v.d[c++] = in->v.d[k];
	}
      }
      break;
    case SILC_COMPLEX:
      for (j = 0; j < length1; j++) {
	for (i = 0; i < length0; i++) {
	  k = (v0[i] + v1[j] * object->d.matrix.m) * 2;
	  new_in->v.c[c++] = in->v.c[k];
	  new_in->v.c[c++] = in->v.c[k+1];
	}
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      for (j = 0; j < length1; j++) {
	for (i = 0; i < length0; i++) {
	  k = (v0[i] + v1[j] * object->d.matrix.m) * 2;
	  new_in->v.z[c++] = in->v.z[k];
	  new_in->v.z[c++] = in->v.z[k+1];
	}
      }
      break;
    default:
      should_not_reach_here;
    }
  }
  free(v0);
  free(v1);
  return modified;
}

/*
 *  vector[] = matrix
 *  vector[] = matrix[]
 *  matrix[] = scalar
 *  matrix[] = vector
 *  matrix[] = vector[]
 *  matrix[] = matrix
 *  matrix[] = matrix[]
 */
static int
proc_assign(silc_object_t *dest, clist_t *dest_subscript[3], int dest_n,
	    silc_object_t *src, silc_modifier_t *modifiers)
{
  silc_dense_t *in;
  int dest_length0, *dest_v0, src_length0, *src_v0;
  int dest_length1, *dest_v1, src_length1, *src_v1;
  int dest_length2, *dest_v2, src_length2, *src_v2;
  int src_n, copy_pattern, identity_subscript[1] = {0};
  void *dest_array, *src_array;
  size_t dest_lda, src_lda;
  ssize_t i;

#define _PAT_IDENT           0
#define _PAT_NEG             1
#define _PAT_TRANS           2
#define _PAT_TRANS_NEG       3
#define _PAT_CONJ            4
#define _PAT_CONJ_NEG        5
#define _PAT_CONJ_TRANS      6
#define _PAT_CONJ_TRANS_NEG  7

  printf("*** dense: dense_assign(0x%lx, 0x%lx)\n", dest, src);
  if (!modifiers) {
    src_n = 0;
    copy_pattern = 0;
  } else {
    src_n = modifiers->n;
    switch (src->precision) {
    case SILC_INT:
    case SILC_LONG:
    case SILC_FLOAT:
    case SILC_DOUBLE:
      copy_pattern = ((modifiers->transpose ? 2 : 0) +
		      (modifiers->negate    ? 1 : 0));
      break;
    case SILC_COMPLEX:
    case SILC_DOUBLE_COMPLEX:
      copy_pattern = ((modifiers->conjugate ? 4 : 0) +
		      (modifiers->transpose ? 2 : 0) +
		      (modifiers->negate    ? 1 : 0));
      break;
    default:
      should_not_reach_here;
    }
  }
  switch (src->type) {
  case SILC_SCALAR_TYPE:
    src_array = &src->d.scalar.i;
    src_lda = 1;
    break;
  case SILC_COLUMN_VECTOR_TYPE:
    src_array = src->d.vector.v.i;
    src_lda = src->d.vector.length;
    break;
  case SILC_ROW_VECTOR_TYPE:
    src_array = src->d.vector.v.i;
    src_lda = 1;
    break;
  case SILC_MATRIX_TYPE:
    src_array = SILC_DENSE(src)->v.i;
    src_lda = src->d.matrix.m;
    break;
  default:
    should_not_reach_here;
  }
  if (src_n == 0) { /* no subscript */
    switch (src->type) {
    case SILC_SCALAR_TYPE:
      src_length0 = 1;
      src_length1 = 1;
      break;
    case SILC_COLUMN_VECTOR_TYPE:
      src_length0 = src->d.vector.length;
      src_length1 = 1;
      break;
    case SILC_ROW_VECTOR_TYPE:
      src_length0 = 1;
      src_length1 = src->d.vector.length;
      break;
    case SILC_MATRIX_TYPE:
      src_length0 = src->d.matrix.m;
      src_length1 = src->d.matrix.n;
      break;
    default:
      should_not_reach_here;
    }
  } else { /* with subscript */
    switch (src->type) {
    case SILC_COLUMN_VECTOR_TYPE:
      src_v0 = clist_expand(modifiers->subscript[0]);
      src_v1 = identity_subscript;
      src_length0 = clist_length(modifiers->subscript[0]);
      src_length1 = 1;
      break;
    case SILC_ROW_VECTOR_TYPE:
      src_v0 = identity_subscript;
      src_v1 = clist_expand(modifiers->subscript[0]);
      src_length0 = 1;
      src_length1 = clist_length(modifiers->subscript[0]);
      break;
    case SILC_MATRIX_TYPE:
      src_v0 = clist_expand(modifiers->subscript[0]);
      src_v1 = clist_expand(modifiers->subscript[1]);
      src_length0 = clist_length(modifiers->subscript[0]);
      src_length1 = clist_length(modifiers->subscript[1]);
      break;
    default:
      should_not_reach_here;
    }
  }
  switch (dest->type) {
  case SILC_COLUMN_VECTOR_TYPE:
    dest_array = dest->d.vector.v.i;
    dest_lda = dest->d.vector.length;
    dest_v0 = clist_expand(dest_subscript[0]);
    dest_v1 = identity_subscript;
    dest_length0 = clist_length(dest_subscript[0]);
    dest_length1 = 1;
    break;
  case SILC_ROW_VECTOR_TYPE:
    dest_array = dest->d.vector.v.i;
    dest_lda = 1;
    dest_v0 = identity_subscript;
    dest_v1 = clist_expand(dest_subscript[0]);
    dest_length0 = 1;
    dest_length1 = clist_length(dest_subscript[0]);
    break;
  case SILC_MATRIX_TYPE:
    dest_array = SILC_DENSE(dest)->v.i;
    dest_lda = dest->d.matrix.m;
    dest_v0 = clist_expand(dest_subscript[0]);
    dest_v1 = clist_expand(dest_subscript[1]);
    dest_length0 = clist_length(dest_subscript[0]);
    dest_length1 = clist_length(dest_subscript[1]);
    break;
  default:
    should_not_reach_here;
  }
  if ((modifiers && modifiers->transpose) ?
      (dest_length0 != src_length1 || dest_length1 != src_length0) :
      (dest_length0 != src_length0 || dest_length1 != src_length1)) {
    fprintf(stderr, MODULE_NAME ": subscript size mismatch\n");
    return -1;
  }
  if (src_n == 0) { /* no subscript */
#define _matrix_simple_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + src_lda * i;		\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = q[j];				\
	}
#define _matrix_simple_neg_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + src_lda * i;		\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = -q[j];				\
	}
#define _matrix_simple_trans_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + i;				\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = q[src_lda * j];			\
	}
#define _matrix_simple_trans_neg_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + i;				\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = -q[src_lda * j];			\
	}
#define _matrix_simple_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * i * 2;		\
	  for (j = jj = 0; j < dest_length0; j++) {		\
	    kk = dest_v0[j] * 2;				\
	    p[kk]   = q[jj++];					\
	    p[kk+1] = q[jj++];					\
	  }							\
	}
#define _matrix_simple_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * i * 2;		\
	  for (j = jj = 0; j < dest_length0; j++) {		\
	    kk = dest_v0[j] * 2;				\
	    p[kk]   = -q[jj++];					\
	    p[kk+1] = -q[jj++];					\
	  }							\
	}
#define _matrix_simple_trans_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + i * 2;			\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * j * 2;				\
	    p[kk]   = q[jj];					\
	    p[kk+1] = q[jj+1];					\
	  }							\
	}
#define _matrix_simple_trans_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + i * 2;			\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * j * 2;				\
	    p[kk]   = -q[jj];					\
	    p[kk+1] = -q[jj+1];					\
	  }							\
	}
#define _matrix_simple_conj_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * i * 2;		\
	  for (j = jj = 0; j < dest_length0; j++) {		\
	    kk = dest_v0[j] * 2;				\
	    p[kk]   =  q[jj++];					\
	    p[kk+1] = -q[jj++];					\
	  }							\
	}
#define _matrix_simple_conj_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * i * 2;		\
	  for (j = jj = 0; j < dest_length0; j++) {		\
	    kk = dest_v0[j] * 2;				\
	    p[kk]   = -q[jj++];					\
	    p[kk+1] =  q[jj++];					\
	  }							\
	}
#define _matrix_simple_conj_trans_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + i * 2;			\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * j * 2;				\
	    p[kk]   =  q[jj];					\
	    p[kk+1] = -q[jj+1];					\
	  }							\
	}
#define _matrix_simple_conj_trans_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + i * 2;			\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * j * 2;				\
	    p[kk]   = -q[jj];					\
	    p[kk+1] =  q[jj+1];					\
	  }							\
	}
    switch (src->precision) {
    case SILC_INT:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_simple_1(i, int32_t);
	break;
      case _PAT_NEG:
	_matrix_simple_neg_1(i, int32_t);
	break;
      case _PAT_TRANS:
	_matrix_simple_trans_1(i, int32_t);
	break;
      case _PAT_TRANS_NEG:
	_matrix_simple_trans_neg_1(i, int32_t);
	break;
      }
      break;
    case SILC_LONG:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_simple_1(l, int64_t);
	break;
      case _PAT_NEG:
	_matrix_simple_neg_1(l, int64_t);
	break;
      case _PAT_TRANS:
	_matrix_simple_trans_1(l, int64_t);
	break;
      case _PAT_TRANS_NEG:
	_matrix_simple_trans_neg_1(l, int64_t);
	break;
      }
      break;
    case SILC_FLOAT:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_simple_1(s, float);
	break;
      case _PAT_NEG:
	_matrix_simple_neg_1(s, float);
	break;
      case _PAT_TRANS:
	_matrix_simple_trans_1(s, float);
	break;
      case _PAT_TRANS_NEG:
	_matrix_simple_trans_neg_1(s, float);
	break;
      }
      break;
    case SILC_DOUBLE:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_simple_1(d, double);
	break;
      case _PAT_NEG:
	_matrix_simple_neg_1(d, double);
	break;
      case _PAT_TRANS:
	_matrix_simple_trans_1(d, double);
	break;
      case _PAT_TRANS_NEG:
	_matrix_simple_trans_neg_1(d, double);
	break;
      }
      break;
    case SILC_COMPLEX:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_simple_2(c, float);
	break;
      case _PAT_NEG:
	_matrix_simple_neg_2(c, float);
	break;
      case _PAT_TRANS:
	_matrix_simple_trans_2(c, float);
	break;
      case _PAT_TRANS_NEG:
	_matrix_simple_trans_neg_2(c, float);
	break;
      case _PAT_CONJ:
	_matrix_simple_conj_2(c, float);
	break;
      case _PAT_CONJ_NEG:
	_matrix_simple_conj_neg_2(c, float);
	break;
      case _PAT_CONJ_TRANS:
	_matrix_simple_conj_trans_2(c, float);
	break;
      case _PAT_CONJ_TRANS_NEG:
	_matrix_simple_conj_trans_neg_2(c, float);
	break;
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_simple_2(z, double);
	break;
      case _PAT_NEG:
	_matrix_simple_neg_2(z, double);
	break;
      case _PAT_TRANS:
	_matrix_simple_trans_2(z, double);
	break;
      case _PAT_TRANS_NEG:
	_matrix_simple_trans_neg_2(z, double);
	break;
      case _PAT_CONJ:
	_matrix_simple_conj_2(z, double);
	break;
      case _PAT_CONJ_NEG:
	_matrix_simple_conj_neg_2(z, double);
	break;
      case _PAT_CONJ_TRANS:
	_matrix_simple_conj_trans_2(z, double);
	break;
      case _PAT_CONJ_TRANS_NEG:
	_matrix_simple_conj_trans_neg_2(z, double);
	break;
      }
      break;
    default:
      should_not_reach_here;
    }
  } else { /* with subscript */
#define _matrix_complex_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + src_lda * src_v1[i];	\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = q[src_v0[j]];			\
	}
#define _matrix_complex_neg_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + src_lda * src_v1[i];	\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = -q[src_v0[j]];			\
	}
#define _matrix_complex_trans_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + src_v0[i];			\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = q[src_lda * src_v1[j]];		\
	}
#define _matrix_complex_trans_neg_1(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j;						\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i];	\
	  q = ((type *)src_array) + src_v0[i];			\
	  for (j = 0; j < dest_length0; j++)			\
	    p[dest_v0[j]] = -q[src_lda * src_v1[j]];		\
	}
#define _matrix_complex_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * src_v1[i] * 2;	\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_v0[j] * 2;					\
	    p[kk]   = q[jj];					\
	    p[kk+1] = q[jj+1];					\
	  }							\
	}
#define _matrix_complex_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * src_v1[i] * 2;	\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_v0[j] * 2;					\
	    p[kk]   = -q[jj];					\
	    p[kk+1] = -q[jj+1];					\
	  }							\
	}
#define _matrix_complex_trans_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_v0[i] * 2;		\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * src_v1[j] * 2;			\
	    p[kk]   = q[jj];					\
	    p[kk+1] = q[jj+1];					\
	  }							\
	}
#define _matrix_complex_trans_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_v0[i] * 2;		\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * src_v1[j] * 2;			\
	    p[kk]   = -q[jj];					\
	    p[kk+1] = -q[jj+1];					\
	  }							\
	}
#define _matrix_complex_conj_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * src_v1[i] * 2;	\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_v0[j] * 2;					\
	    p[kk]   =  q[jj];					\
	    p[kk+1] = -q[jj+1];					\
	  }							\
	}
#define _matrix_complex_conj_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_lda * src_v1[i] * 2;	\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_v0[j] * 2;					\
	    p[kk]   = -q[jj];					\
	    p[kk+1] =  q[jj+1];					\
	  }							\
	}
#define _matrix_complex_conj_trans_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_v0[i] * 2;		\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * src_v1[j] * 2;			\
	    p[kk]   =  q[jj];					\
	    p[kk+1] = -q[jj+1];					\
	  }							\
	}
#define _matrix_complex_conj_trans_neg_2(pre, type) \
	for (i = 0; i < dest_length1; i++) {			\
	  size_t j, jj, kk;					\
	  type *p, *q;						\
	  p = ((type *)dest_array) + dest_lda * dest_v1[i] * 2;	\
	  q = ((type *)src_array) + src_v0[i] * 2;		\
	  for (j = 0; j < dest_length0; j++) {			\
	    kk = dest_v0[j] * 2;				\
	    jj = src_lda * src_v1[j] * 2;			\
	    p[kk]   = -q[jj];					\
	    p[kk+1] =  q[jj+1];					\
	  }							\
	}
    switch (src->precision) {
    case SILC_INT:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_complex_1(i, int32_t);
	break;
      case _PAT_NEG:
	_matrix_complex_neg_1(i, int32_t);
	break;
      case _PAT_TRANS:
	_matrix_complex_trans_1(i, int32_t);
	break;
      case _PAT_TRANS_NEG:
	_matrix_complex_trans_neg_1(i, int32_t);
	break;
      }
      break;
    case SILC_LONG:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_complex_1(l, int64_t);
	break;
      case _PAT_NEG:
	_matrix_complex_neg_1(l, int64_t);
	break;
      case _PAT_TRANS:
	_matrix_complex_trans_1(l, int64_t);
	break;
      case _PAT_TRANS_NEG:
	_matrix_complex_trans_neg_1(l, int64_t);
	break;
      }
      break;
    case SILC_FLOAT:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_complex_1(s, float);
	break;
      case _PAT_NEG:
	_matrix_complex_neg_1(s, float);
	break;
      case _PAT_TRANS:
	_matrix_complex_trans_1(s, float);
	break;
      case _PAT_TRANS_NEG:
	_matrix_complex_trans_neg_1(s, float);
	break;
      }
      break;
    case SILC_DOUBLE:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_complex_1(d, double);
	break;
      case _PAT_NEG:
	_matrix_complex_neg_1(d, double);
	break;
      case _PAT_TRANS:
	_matrix_complex_trans_1(d, double);
	break;
      case _PAT_TRANS_NEG:
	_matrix_complex_trans_neg_1(d, double);
	break;
      }
      break;
    case SILC_COMPLEX:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_complex_2(c, float);
	break;
      case _PAT_NEG:
	_matrix_complex_neg_2(c, float);
	break;
      case _PAT_TRANS:
	_matrix_complex_trans_2(c, float);
	break;
      case _PAT_TRANS_NEG:
	_matrix_complex_trans_neg_2(c, float);
	break;
      case _PAT_CONJ:
	_matrix_complex_conj_2(c, float);
	break;
      case _PAT_CONJ_NEG:
	_matrix_complex_conj_neg_2(c, float);
	break;
      case _PAT_CONJ_TRANS:
	_matrix_complex_conj_trans_2(c, float);
	break;
      case _PAT_CONJ_TRANS_NEG:
	_matrix_complex_conj_trans_neg_2(c, float);
	break;

      }
      break;
    case SILC_DOUBLE_COMPLEX:
      switch (copy_pattern) {
      case _PAT_IDENT:
	_matrix_complex_2(z, double);
	break;
      case _PAT_NEG:
	_matrix_complex_neg_2(z, double);
	break;
      case _PAT_TRANS:
	_matrix_complex_trans_2(z, double);
	break;
      case _PAT_TRANS_NEG:
	_matrix_complex_trans_neg_2(z, double);
	break;
      case _PAT_CONJ:
	_matrix_complex_conj_2(z, double);
	break;
      case _PAT_CONJ_NEG:
	_matrix_complex_conj_neg_2(z, double);
	break;
      case _PAT_CONJ_TRANS:
	_matrix_complex_conj_trans_2(z, double);
	break;
      case _PAT_CONJ_TRANS_NEG:
	_matrix_complex_conj_trans_neg_2(z, double);
	break;
      }
      break;
    default:
      should_not_reach_here;
    }
    if (src_v0 != identity_subscript) free(src_v0);
    if (src_v1 != identity_subscript) free(src_v1);
  }
  if (dest_v0 != identity_subscript) free(dest_v0);
  if (dest_v1 != identity_subscript) free(dest_v1);
  printf("*** dense: dense_assign() done.\n");
  return 0;
}

static int
proc_pprint(FILE *file, silc_object_t *object, int level)
{
  silc_dense_t *in;
  int c, i, j, k;

  in = SILC_DENSE(object);
  c = 0;
  for (j = 0; j < object->d.matrix.n; j++) {
    for (i = 0; i < object->d.matrix.m; i++) {
      for (k = 0; k < level; k++)
	fprintf(file, "  ");
      switch (object->precision) {
      case SILC_INT:
	fprintf(file, "v.i[%d,%d] = %d\n", i, j, in->v.i[c]); break;
      case SILC_LONG:
	fprintf(file, "v.l[%d,%d] = %lld\n", i, j, in->v.l[c]); break;
      case SILC_FLOAT:
	fprintf(file, "v.s[%d,%d] = %e\n", i, j, in->v.s[c]); break;
      case SILC_DOUBLE:
	fprintf(file, "v.d[%d,%d] = %e\n", i, j, in->v.d[c]); break;
      case SILC_COMPLEX:
	fprintf(file, "v.c[%d,%d] = %e%+ei\n", i, j,
		in->v.c[c*2],
		in->v.c[c*2+1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	fprintf(file, "v.z[%d,%d] = %e%+ei\n", i, j,
		in->v.z[c*2],
		in->v.z[c*2+1]);
	break;
      default:
	should_not_reach_here;
      }
      c++;
    }
  }
  return 0;
}

silc_format_t silc_dense_proc_table = {
  proc_new,
  proc_recv,
  proc_send,
  proc_free,
  proc_dup,
  proc_pconv,
  NULL,
  NULL,
  proc_assign,
  proc_subscript,
  proc_pprint
};

void init_dense_op(); /* dense_op.c */

void
init_dense()
{
  int format_id;

  format_id = silc_register_format(SILC_FORMAT_DENSE, &silc_dense_proc_table);

  init_dense_op(MODULE_NAME);
}
