/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: dense_op.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <time.h>

#include "object.h"
#include "dense.h"
#include "module.h"

#include "mt19937ar.h" /* for init_genrand() and genrand_real3() */

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

static int
_complex_conjugate(silc_object_t *object)
{
  silc_dense_t *in;
  ssize_t i;

#define _complex_conjugate_2(pre) { \
    size_t j, k;				\
    k = 1 + i * object->d.matrix.m * 2;		\
    for (j = 0; j < object->d.matrix.m; j++) {	\
      in->v.pre[k] = -in->v.pre[k]; k += 2;	\
    }						\
  }

  in = SILC_DENSE(object);
  switch (object->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _complex_conjugate_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _complex_conjugate_2(z);
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static silc_object_t *
_transpose(silc_object_t *object)
{
  silc_object_t *modified;
  silc_dense_t *in, *new_in;
  int c, i, j, k;

#define _transpose_1(pre) \
	c = 0;						\
	for (i = 0; i < object->d.matrix.m; i++) {	\
	  for (j = 0; j < object->d.matrix.n; j++) {	\
	    k = i + j * object->d.matrix.m;		\
	    new_in->v.pre[c++] = in->v.pre[k];		\
	  }						\
	}

#define _transpose_2(pre) \
	c = 0;						\
	for (i = 0; i < object->d.matrix.m; i++) {	\
	  for (j = 0; j < object->d.matrix.n; j++) {	\
	    k = (i + j * object->d.matrix.m) * 2;	\
	    new_in->v.pre[c++] = in->v.pre[k];		\
	    new_in->v.pre[c++] = in->v.pre[k+1];	\
	  }						\
	}

  in = SILC_DENSE(object);
  modified = silc_object_new_matrix(object->precision,
				    object->format,
				    object->d.matrix.n,  /* m */
				    object->d.matrix.m); /* n */
  modified->d.matrix.in = (void *)silc_dense_allocate(object->precision,
						      object->d.matrix.m,
						      object->d.matrix.n);
  new_in = SILC_DENSE(modified);
  switch (object->precision) {
  case SILC_INT:
    _transpose_1(i); break;
  case SILC_LONG:
    _transpose_1(l); break;
  case SILC_FLOAT:
    _transpose_1(s); break;
  case SILC_DOUBLE:
    _transpose_1(d); break;
  case SILC_COMPLEX:
    _transpose_2(c); break;
  case SILC_DOUBLE_COMPLEX:
    _transpose_2(z); break;
  default:
    should_not_reach_here;
  }
  return modified;
}

/* out: 0 */
int
silc_dense_complex_conjugate(int n_params, silc_mfunc_param_t *param)
{
#ifdef DEBUG
  printf("silc_dense_complex_conjugate(0x%lx)\n", param[0].o);
#endif
  return _complex_conjugate(param[0].o);
}

/* out: 1 */
int
silc_dense_transpose(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *modified;

#ifdef DEBUG
  printf("silc_dense_transpose(0x%lx)\n", param[0].o);
#endif
  modified = _transpose(param[0].o);
  if (!modified)
    return -1;
  param[1].o = modified;
  return 0;
}

/* out: 1 */
int
silc_dense_conjugate_transpose(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *modified;

#ifdef DEBUG
  printf("silc_dense_conjugate_transpose(0x%lx)\n", param[0].o);
#endif
  modified = _transpose(param[0].o);
  if (!modified)
    return -1;
  if (_complex_conjugate(modified) < 0)
    return -1;
  param[1].o = modified;
  return 0;
}

/* out: 0 */
int
silc_dense_negate(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  silc_dense_t *in;
  ssize_t i;

#define _negate_1(pre) { \
    size_t j, k;				\
    k = i * object->d.matrix.m;			\
    for (j = 0; j < object->d.matrix.m; j++) {	\
      in->v.pre[k] = -in->v.pre[k]; k++;	\
    }						\
  }

#define _negate_2(pre) { \
    size_t j, k;				\
    k = i * object->d.matrix.m * 2;		\
    for (j = 0; j < object->d.matrix.m; j++) {	\
      in->v.pre[k] = -in->v.pre[k]; k += 2;	\
    }						\
  }

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_negate(0x%lx)\n", object);
#endif
  in = SILC_DENSE(object);
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _negate_1(i);
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _negate_1(l);
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _negate_1(s);
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _negate_1(d);
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _negate_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _negate_2(z);
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
silc_dense_add_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_dense_t *x_in, *y_in;
  ssize_t size, k;

  x = param[0].o; x_mods = param[0].m; x_in = SILC_DENSE(x);
  y = param[1].o; y_mods = param[1].m; y_in = SILC_DENSE(y);
#ifdef DEBUG
  printf("silc_dense_add_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.matrix.m != y->d.matrix.m ||
      x->d.matrix.n != y->d.matrix.n)
    return -1;
  switch (x->precision) {
  case SILC_INT:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.i[k] += y_in->v.i[k];
    break;
  case SILC_LONG:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.l[k] += y_in->v.l[k];
    break;
  case SILC_FLOAT:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.s[k] += y_in->v.s[k];
    break;
  case SILC_DOUBLE:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.d[k] += y_in->v.d[k];
    break;
  case SILC_COMPLEX:
    size = x->d.matrix.m * y->d.matrix.n * 2;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.c[k] += y_in->v.c[k];
    break;
  case SILC_DOUBLE_COMPLEX:
    size = x->d.matrix.m * y->d.matrix.n * 2;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.z[k] += y_in->v.z[k];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
silc_dense_sub_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_dense_t *x_in, *y_in;
  ssize_t size, k;

  x = param[0].o; x_mods = param[0].m; x_in = SILC_DENSE(x);
  y = param[1].o; y_mods = param[1].m; y_in = SILC_DENSE(y);
#ifdef DEBUG
  printf("silc_dense_sub_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->d.matrix.m != y->d.matrix.m ||
      x->d.matrix.n != y->d.matrix.n)
    return -1;
  switch (x->precision) {
  case SILC_INT:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.i[k] -= y_in->v.i[k];
    break;
  case SILC_LONG:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.l[k] -= y_in->v.l[k];
    break;
  case SILC_FLOAT:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.s[k] -= y_in->v.s[k];
    break;
  case SILC_DOUBLE:
    size = x->d.matrix.m * y->d.matrix.n;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.d[k] -= y_in->v.d[k];
    break;
  case SILC_COMPLEX:
    size = x->d.matrix.m * y->d.matrix.n * 2;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.c[k] -= y_in->v.c[k];
    break;
  case SILC_DOUBLE_COMPLEX:
    size = x->d.matrix.m * y->d.matrix.n * 2;
#pragma omp parallel for
    for (k = 0; k < size; k++)
      x_in->v.z[k] -= y_in->v.z[k];
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 2 */
int
silc_dense_mul_matrix_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *z;
  silc_modifier_t *x_mods, *y_mods;
  silc_dense_t *in;
  ssize_t i;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_dense_mul_matrix_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_DENSE(x);
  z = silc_object_new_column_vector(x->precision, x->d.matrix.m);
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      int32_t e;
      size_t j, k;
      e = 0;
      k = i;
      for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
	e += in->v.i[k] * y->d.vector.v.i[j];
	k += x->d.matrix.m;
      }
      z->d.vector.v.i[i] = e;
    }
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      int64_t e;
      size_t j, k;
      e = 0;
      k = i;
      for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
	e += in->v.l[k] * y->d.vector.v.l[j];
	k += x->d.matrix.m;
      }
      z->d.vector.v.l[i] = e;
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      float e;
      size_t j, k;
      e = 0.0;
      k = i;
      for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
	e += in->v.s[k] * y->d.vector.v.s[j];
	k += x->d.matrix.m;
      }
      z->d.vector.v.s[i] = e;
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      double e;
      size_t j, k;
      e = 0.0;
      k = i;
      for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
	e += in->v.d[k] * y->d.vector.v.d[j];
	k += x->d.matrix.m;
      }
      z->d.vector.v.d[i] = e;
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m * 2; i += 2) {
      float x_re, y_re, e_re;
      float x_im, y_im, e_im;
      size_t j, k;
      e_re = e_im = 0.0;
      k = i;
      for (j = 0; j < (ssize_t)x->d.matrix.n * 2; j += 2) {
	x_re = in->v.c[k];   y_re = y->d.vector.v.c[j];
	x_im = in->v.c[k+1]; y_im = y->d.vector.v.c[j+1];
	e_re += (x_re * y_re - x_im * y_im);
	e_im += (x_re * y_im + x_im * y_re);
	k += x->d.matrix.m * 2;
      }
      z->d.vector.v.c[i]   = e_re;
      z->d.vector.v.c[i+1] = e_im;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m * 2; i += 2) {
      double x_re, y_re, e_re;
      double x_im, y_im, e_im;
      size_t j, k;
      e_re = e_im = 0.0;
      k = i;
      for (j = 0; j < (ssize_t)x->d.matrix.n * 2; j += 2) {
	x_re = in->v.z[k];   y_re = y->d.vector.v.z[j];
	x_im = in->v.z[k+1]; y_im = y->d.vector.v.z[j+1];
	e_re += (x_re * y_re - x_im * y_im);
	e_im += (x_re * y_im + x_im * y_re);
	k += x->d.matrix.m * 2;
      }
      z->d.vector.v.z[i]   = e_re;
      z->d.vector.v.z[i+1] = e_im;
    }
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = z;
  return 0;
}

/* out: 2 */
int
silc_dense_mul_vector_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *z;
  silc_modifier_t *x_mods, *y_mods;
  silc_dense_t *in;
  ssize_t i;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_dense_mul_vector_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_DENSE(y);
  z = silc_object_new_row_vector(y->precision, y->d.matrix.n);
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)y->d.matrix.n; i++) {
      int32_t e;
      size_t j, k;
      e = 0;
      for (j = 0, k = y->d.matrix.m * i; j < y->d.matrix.m; j++, k++)
	e += x->d.vector.v.i[j] * in->v.i[k];
      z->d.vector.v.i[i] = e;
    }
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)y->d.matrix.n; i++) {
      int64_t e;
      size_t j, k;
      e = 0;
      for (j = 0, k = y->d.matrix.m * i; j < y->d.matrix.m; j++, k++)
	e += x->d.vector.v.l[j] * in->v.l[k];
      z->d.vector.v.l[i] = e;
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)y->d.matrix.n; i++) {
      float e;
      size_t j, k;
      e = 0.0;
      for (j = 0, k = y->d.matrix.m * i; j < y->d.matrix.m; j++, k++)
	e += x->d.vector.v.s[j] * in->v.s[k];
      z->d.vector.v.s[i] = e;
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)y->d.matrix.n; i++) {
      double e;
      size_t j, k;
      e = 0.0;
      for (j = 0, k = y->d.matrix.m * i; j < y->d.matrix.m; j++, k++)
	e += x->d.vector.v.d[j] * in->v.d[k];
      z->d.vector.v.d[i] = e;
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)y->d.matrix.n * 2; i += 2) {
      float real, imag;
      size_t j, k;
      real = imag = 0.0;
      k = y->d.matrix.m * i;
      for (j = 0; j < y->d.matrix.m * 2; j += 2, k += 2) {
	real += (x->d.vector.v.c[j]   * in->v.c[k]   -
		 x->d.vector.v.c[j+1] * in->v.c[k+1]);
	imag += (x->d.vector.v.c[j]   * in->v.c[k+1] +
		 x->d.vector.v.c[j+1] * in->v.c[k]);
      }
      z->d.vector.v.c[i]   = real;
      z->d.vector.v.c[i+1] = imag;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)y->d.matrix.n * 2; i += 2) {
      double real, imag;
      size_t j, k;
      real = imag = 0.0;
      k = y->d.matrix.m * i;
      for (j = 0; j < y->d.matrix.m * 2; j += 2, k += 2) {
	real += (x->d.vector.v.z[j]   * in->v.z[k]   -
		 x->d.vector.v.z[j+1] * in->v.z[k+1]);
	imag += (x->d.vector.v.z[j]   * in->v.z[k+1] +
		 x->d.vector.v.z[j+1] * in->v.z[k]);
      }
      z->d.vector.v.z[i]   = real;
      z->d.vector.v.z[i+1] = imag;
    }
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = z;
  return 0;
}

/* out: 2 */
int
silc_dense_mul_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *z;
  silc_modifier_t *x_mods, *y_mods;
  silc_dense_t *x_in, *y_in;
  ssize_t k;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_dense_mul_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  z = silc_object_new_matrix(x->precision, x->format,
			     x->d.matrix.m,
			     y->d.matrix.n);
  if (!z)
    return -1;
  z->d.matrix.in = (void *)malloc(sizeof(silc_dense_t));
  if (!z->d.matrix.in)
    return -1;
  x_in = SILC_DENSE(x);
  y_in = SILC_DENSE(y);
  switch (x->precision) {
  case SILC_INT:
    {
      int32_t *v;
      v = (int32_t *)malloc(sizeof(int32_t) * x->d.matrix.m * y->d.matrix.n);
#pragma omp parallel for
      for (k = 0; k < (ssize_t)y->d.matrix.n; k++) {
	size_t i, j;
	for (i = 0; i < x->d.matrix.m; i++) {
	  int32_t e;
	  e = 0;
	  for (j = 0; j < x->d.matrix.n; j++)
	    e += (x_in->v.i[i + x->d.matrix.m * j] *
		  y_in->v.i[j + y->d.matrix.m * k]);
	  v[i + x->d.matrix.m * k] = e;
	}
      }
      SILC_DENSE(z)->v.i = v;
    }
    break;
  case SILC_LONG:
    {
      int64_t *v;
      v = (int64_t *)malloc(sizeof(int64_t) * x->d.matrix.m * y->d.matrix.n);
#pragma omp parallel for
      for (k = 0; k < (ssize_t)y->d.matrix.n; k++) {
	size_t i, j;
	for (i = 0; i < x->d.matrix.m; i++) {
	  int64_t e;
	  e = 0;
	  for (j = 0; j < x->d.matrix.n; j++)
	    e += (x_in->v.l[i + x->d.matrix.m * j] *
		  y_in->v.l[j + y->d.matrix.m * k]);
	  v[i + x->d.matrix.m * k] = e;
	}
      }
      SILC_DENSE(z)->v.l = v;
    }
    break;
  case SILC_FLOAT:
    {
      float *v;
      v = (float *)malloc(sizeof(float) * x->d.matrix.m * y->d.matrix.n);
#pragma omp parallel for
      for (k = 0; k < (ssize_t)y->d.matrix.n; k++) {
	size_t i, j;
	for (i = 0; i < x->d.matrix.m; i++) {
	  float e;
	  e = 0;
	  for (j = 0; j < x->d.matrix.n; j++)
	    e += (x_in->v.s[i + x->d.matrix.m * j] *
		  y_in->v.s[j + y->d.matrix.m * k]);
	  v[i + x->d.matrix.m * k] = e;
	}
      }
      SILC_DENSE(z)->v.s = v;
    }
    break;
  case SILC_DOUBLE:
    {
      double *v;
      v = (double *)malloc(sizeof(double) * x->d.matrix.m * y->d.matrix.n);
#pragma omp parallel for
      for (k = 0; k < (ssize_t)y->d.matrix.n; k++) {
	size_t i, j;
	for (i = 0; i < x->d.matrix.m; i++) {
	  double e;
	  e = 0;
	  for (j = 0; j < x->d.matrix.n; j++)
	    e += (x_in->v.d[i + x->d.matrix.m * j] *
		  y_in->v.d[j + y->d.matrix.m * k]);
	  v[i + x->d.matrix.m * k] = e;
	}
      }
      SILC_DENSE(z)->v.d = v;
    }
    break;
  case SILC_COMPLEX:
    {
      float *v;
      v = (float *)malloc(sizeof(float) * 2 * x->d.matrix.m * y->d.matrix.n);
#pragma omp parallel for
      for (k = 0; k < (ssize_t)y->d.matrix.n * 2; k += 2) {
	size_t i, j, p;
	for (i = 0; i < x->d.matrix.m * 2; i += 2) {
	  float x_re, y_re, e_re;
	  float x_im, y_im, e_im;
	  e_re = e_im = 0;
	  for (j = 0; j < x->d.matrix.n * 2; j += 2) {
	    p = i + x->d.matrix.m * j;
	    x_re = x_in->v.c[p]; x_im = x_in->v.c[p+1];
	    p = j + y->d.matrix.m * k;
	    y_re = y_in->v.c[p]; y_im = y_in->v.c[p+1];
	    e_re += (x_re * y_re - x_im * y_im);
	    e_im += (x_re * y_im + x_im * y_re);
	  }
	  p = i + x->d.matrix.m * k;
	  v[p]   = e_re;
	  v[p+1] = e_im;
	}
      }
      SILC_DENSE(z)->v.c = v;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    {
      double *v;
      v = (double *)malloc(sizeof(double) * 2 * x->d.matrix.m * y->d.matrix.n);
#pragma omp parallel for
      for (k = 0; k < (ssize_t)y->d.matrix.n * 2; k += 2) {
	size_t i, j, p;
	for (i = 0; i < x->d.matrix.m * 2; i += 2) {
	  double x_re, y_re, e_re;
	  double x_im, y_im, e_im;
	  e_re = e_im = 0;
	  for (j = 0; j < x->d.matrix.n * 2; j += 2) {
	    p = i + x->d.matrix.m * j;
	    x_re = x_in->v.z[p]; x_im = x_in->v.z[p+1];
	    p = j + y->d.matrix.m * k;
	    y_re = y_in->v.z[p]; y_im = y_in->v.z[p+1];
	    e_re += (x_re * y_re - x_im * y_im);
	    e_im += (x_re * y_im + x_im * y_re);
	  }
	  p = i + x->d.matrix.m * k;
	  v[p]   = e_re;
	  v[p+1] = e_im;
	}
      }
      SILC_DENSE(z)->v.z = v;
    }
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = z;
  return 0;
}

/* out: 2 */
int
silc_dense_mul_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *z;
  silc_modifier_t *x_mods, *y_mods;
  ssize_t i;

  x = param[0].o; x_mods = param[0].m;  /* column vector */
  y = param[1].o; y_mods = param[1].m;  /* row vector */
#ifdef DEBUG
  printf("silc_dense_mul_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  z = silc_object_new_matrix(x->precision,
			     silc_find_format_id(SILC_FORMAT_DENSE),
			     x->d.vector.length,  /* m */
			     y->d.vector.length); /* n */
  if (!z)
    return -1;
  z->d.matrix.in = (void *)malloc(sizeof(silc_dense_t));
  if (!z->d.matrix.in)
    return -1;
  switch (x->precision) {
  case SILC_INT:
    {
      int32_t *v;
      v = (int32_t *)malloc(sizeof(int32_t) * z->d.matrix.m * z->d.matrix.n);
#pragma omp parallel for
      for (i = 0; i < (ssize_t)y->d.vector.length; i++) {
	size_t j, k;
	k = z->d.matrix.m * i;
	for (j = 0; j < x->d.vector.length; j++)
	  v[k++] = x->d.vector.v.i[j] * y->d.vector.v.i[i];
      }
      SILC_DENSE(z)->v.i = v;
    }
    break;
  case SILC_LONG:
    {
      int64_t *v;
      v = (int64_t *)malloc(sizeof(int64_t) * z->d.matrix.m * z->d.matrix.n);
#pragma omp parallel for
      for (i = 0; i < (ssize_t)y->d.vector.length; i++) {
	size_t j, k;
	k = z->d.matrix.m * i;
	for (j = 0; j < x->d.vector.length; j++)
	  v[k++] = x->d.vector.v.l[j] * y->d.vector.v.l[i];
      }
      SILC_DENSE(z)->v.l = v;
    }
    break;
  case SILC_FLOAT:
    {
      float *v;
      v = (float *)malloc(sizeof(float) * z->d.matrix.m * z->d.matrix.n);
#pragma omp parallel for
      for (i = 0; i < (ssize_t)y->d.vector.length; i++) {
	size_t j, k;
	k = z->d.matrix.m * i;
	for (j = 0; j < x->d.vector.length; j++)
	  v[k++] = x->d.vector.v.s[j] * y->d.vector.v.s[i];
      }
      SILC_DENSE(z)->v.s = v;
    }
    break;
  case SILC_DOUBLE:
    {
      double *v;
      v = (double *)malloc(sizeof(double) * z->d.matrix.m * z->d.matrix.n);
#pragma omp parallel for
      for (i = 0; i < (ssize_t)y->d.vector.length; i++) {
	size_t j, k;
	k = z->d.matrix.m * i;
	for (j = 0; j < x->d.vector.length; j++)
	  v[k++] = x->d.vector.v.d[j] * y->d.vector.v.d[i];
      }
      SILC_DENSE(z)->v.d = v;
    }
    break;
  case SILC_COMPLEX:
    {
      float *v;
      v = (float *)malloc(sizeof(float) * 2 * z->d.matrix.m * z->d.matrix.n);
#pragma omp parallel for
      for (i = 0; i < (ssize_t)y->d.vector.length * 2; i += 2) {
	float x_re, y_re;
	float x_im, y_im;
	size_t j, k;
	y_re = y->d.vector.v.c[i];
	y_im = y->d.vector.v.c[i+1];
	k = z->d.matrix.m * i;
	for (j = 0; j < x->d.vector.length * 2; j += 2) {
	  x_re = x->d.vector.v.c[j];
	  x_im = x->d.vector.v.c[j+1];
	  v[k++] = (x_re * y_re - x_im * y_im);
	  v[k++] = (x_re * y_im + x_im * y_re);
	}
      }
      SILC_DENSE(z)->v.c = v;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    {
      double *v;
      v = (double *)malloc(sizeof(double) * 2 * z->d.matrix.m * z->d.matrix.n);
#pragma omp parallel for
      for (i = 0; i < (ssize_t)y->d.vector.length * 2; i += 2) {
	double x_re, y_re;
	double x_im, y_im;
	size_t j, k;
	y_re = y->d.vector.v.z[i];
	y_im = y->d.vector.v.z[i+1];
	k = z->d.matrix.m * i;
	for (j = 0; j < x->d.vector.length * 2; j += 2) {
	  x_re = x->d.vector.v.z[j];
	  x_im = x->d.vector.v.z[j+1];
	  v[k++] = (x_re * y_re - x_im * y_im);
	  v[k++] = (x_re * y_im + x_im * y_re);
	}
      }
      SILC_DENSE(z)->v.z = v;
    }
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = z;
  return 0;
}

static int
matrix_scalar_product(silc_mfunc_param_t *matrix,
		      silc_mfunc_param_t *scalar)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_dense_t *in;
  ssize_t i, j, k;

  x = matrix->o; x_mods = matrix->m;
  y = scalar->o; y_mods = scalar->m;
#ifdef DEBUG
  printf("dense_op: matrix_scalar_product(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_DENSE(x);
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for private(i, j, k)
    for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
      k = x->d.matrix.m * j;
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	in->v.i[k++] *= y->d.scalar.i;
    }
    break;
  case SILC_LONG:
#pragma omp parallel for private(i, j, k)
    for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
      k = x->d.matrix.m * j;
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	in->v.l[k++] *= y->d.scalar.l;
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for private(i, j, k)
    for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
      k = x->d.matrix.m * j;
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	in->v.s[k++] *= y->d.scalar.s;
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for private(i, j, k)
    for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
      k = x->d.matrix.m * j;
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	in->v.d[k++] *= y->d.scalar.d;
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for private(i, j, k)
    for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
      float real, imag;
      k = x->d.matrix.m * j * 2;
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
	real = (in->v.c[k]   * y->d.scalar.c[0] -
		in->v.c[k+1] * y->d.scalar.c[1]);
	imag = (in->v.c[k]   * y->d.scalar.c[1] +
		in->v.c[k+1] * y->d.scalar.c[0]);
	in->v.c[k++] = real;
	in->v.c[k++] = imag;
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for private(i, j, k)
    for (j = 0; j < (ssize_t)x->d.matrix.n; j++) {
      float real, imag;
      k = x->d.matrix.m * j * 2;
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
	real = (in->v.z[k]   * y->d.scalar.z[0] -
		in->v.z[k+1] * y->d.scalar.z[1]);
	imag = (in->v.z[k]   * y->d.scalar.z[1] +
		in->v.z[k+1] * y->d.scalar.z[0]);
	in->v.z[k++] = real;
	in->v.z[k++] = imag;
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
silc_dense_mul_matrix_scalar(int n_params, silc_mfunc_param_t *param)
{
  return matrix_scalar_product(&param[0], &param[1]);
}

/* out: 1 */
int
silc_dense_mul_scalar_matrix(int n_params, silc_mfunc_param_t *param)
{
  return matrix_scalar_product(&param[1], &param[0]);
}

/* functions */

static silc_object_t *
_zeros(double v, size_t m, size_t n)
{
  silc_object_t *object;
  silc_dense_t *in;
  int format;
  ssize_t i;

  if (m == 1 && n == 1) {
    object = silc_object_new_double(v);
  } else if (m > 1 && n == 1) {
    object = silc_object_new_column_vector(SILC_DOUBLE, m);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.vector.length; i++)
      object->d.vector.v.d[i] = v;
  } else if (m == 1 && n > 1) {
    object = silc_object_new_row_vector(SILC_DOUBLE, n);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.vector.length; i++)
      object->d.vector.v.d[i] = v;
  } else if (m > 1 && n > 1) {
    format = silc_find_format_id(SILC_FORMAT_DENSE);
    object = silc_object_new_matrix(SILC_DOUBLE, format, m, n);
    object->d.matrix.in = (void *)silc_dense_allocate(object->precision,
						      object->d.matrix.m,
						      object->d.matrix.n);
    in = SILC_DENSE(object);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++) {
      double *p; size_t j;
      p = in->v.d + m * i;
      for (j = 0; j < object->d.matrix.m; j++)
	*p++ = v;
    }
  } else {
    return NULL;  
  }
  return object;
}

int
silc_dense_func_zeros_1(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  size_t n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_func_zeros_1(0x%lx)\n", object);
#endif
  if (object->type != SILC_SCALAR_TYPE)
    return -1;
  switch (object->precision) {
  case SILC_INT:
    n = object->d.scalar.i; break;
  case SILC_LONG:
    n = object->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "zeros: argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "zeros: invalid dimension\n");
    return -1;
  }
  param[1].o = _zeros(0, n ,n);
  return (param[1].o) ? 0 : -1;
}

int
silc_dense_func_zeros_2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *size1, *size2;
  size_t m, n;

  size1 = param[0].o;
  size2 = param[1].o;
#ifdef DEBUG
  printf("silc_dense_func_zeros_2(0x%lx, 0x%lx)\n", size1, size2);
#endif
  if (size1->type != SILC_SCALAR_TYPE ||
      size2->type != SILC_SCALAR_TYPE)
    return -1;
  switch (size1->precision) {
  case SILC_INT:
    m = size1->d.scalar.i; break;
  case SILC_LONG:
    m = size1->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "zeros: 1st argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (m < 1) {
    fprintf(stderr, "zeros: 1st dimension is invalid\n");
    return -1;
  }
  switch (size2->precision) {
  case SILC_INT:
    n = size2->d.scalar.i; break;
  case SILC_LONG:
    n = size2->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "zeros: 2nd argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "zeros: 2nd dimension is invalid\n");
    return -1;
  }
  param[2].o = _zeros(0, m ,n);
  return (param[2].o) ? 0 : -1;
}

int
silc_dense_func_ones_1(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  size_t n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_func_ones_1(0x%lx)\n", object);
#endif
  if (object->type != SILC_SCALAR_TYPE)
    return -1;
  switch (object->precision) {
  case SILC_INT:
    n = object->d.scalar.i; break;
  case SILC_LONG:
    n = object->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "ones: argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "ones: invalid dimension\n");
    return -1;
  }
  param[1].o = _zeros(1, n ,n);
  return (param[1].o) ? 0 : -1;
}

int
silc_dense_func_ones_2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *size1, *size2;
  size_t m, n;

  size1 = param[0].o;
  size2 = param[1].o;
#ifdef DEBUG
  printf("silc_dense_func_ones_2(0x%lx, 0x%lx)\n", size1, size2);
#endif
  if (size1->type != SILC_SCALAR_TYPE ||
      size2->type != SILC_SCALAR_TYPE)
    return -1;
  switch (size1->precision) {
  case SILC_INT:
    m = size1->d.scalar.i; break;
  case SILC_LONG:
    m = size1->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "ones: 1st argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (m < 1) {
    fprintf(stderr, "ones: 1st dimension is invalid\n");
    return -1;
  }
  switch (size2->precision) {
  case SILC_INT:
    n = size2->d.scalar.i; break;
  case SILC_LONG:
    n = size2->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "ones: 2nd argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "ones: 2nd dimension is invalid\n");
    return -1;
  }
  param[2].o = _zeros(1, m ,n);
  return (param[2].o) ? 0 : -1;
}

static silc_object_t *
_random(size_t m, size_t n)
{
  silc_object_t *object;
  silc_dense_t *in;
  int format;
  size_t i, j;

  if (m == 1 && n == 1) {
    object = silc_object_new_double(genrand_real3());
  } else if (m > 1 && n == 1) {
    object = silc_object_new_column_vector(SILC_DOUBLE, m);
    for (i = 0; i < object->d.vector.length; i++)
      object->d.vector.v.d[i] = genrand_real3();
  } else if (m == 1 && n > 1) {
    object = silc_object_new_row_vector(SILC_DOUBLE, n);
    for (i = 0; i < object->d.vector.length; i++)
      object->d.vector.v.d[i] = genrand_real3();
  } else if (m > 1 && n > 1) {
    format = silc_find_format_id(SILC_FORMAT_DENSE);
    object = silc_object_new_matrix(SILC_DOUBLE, format, m, n);
    object->d.matrix.in = (void *)silc_dense_allocate(object->precision,
						      object->d.matrix.m,
						      object->d.matrix.n);
    in = SILC_DENSE(object);
    for (j = 0; j < object->d.matrix.n; j++)
      for (i = 0; i < object->d.matrix.m; i++)
	in->v.d[i + j * object->d.matrix.m] = genrand_real3();
  } else {
    return NULL;  
  }
  return object;
}

int
silc_dense_func_rand_1(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  size_t n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_func_rand_1(0x%lx)\n", object);
#endif
  if (object->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "rand: argument must be scalar\n");
    return -1;
  }
  switch (object->precision) {
  case SILC_INT:
    n = object->d.scalar.i; break;
  case SILC_LONG:
    n = object->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "rand: argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  if (n < 1) {
    fprintf(stderr, "rand: invalid dimension\n");
    return -1;
  }
  param[1].o = _random(n, n);
  return (param[1].o) ? 0 : -1;
}

int
silc_dense_func_rand_2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *size1, *size2;
  size_t m, n;

  size1 = param[0].o;
  size2 = param[1].o;
#ifdef DEBUG
  printf("silc_dense_func_rand_2(0x%lx, 0x%lx)\n", size1, size2);
#endif
  if (size1->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "rand: 1st argument must be scalar\n");
    return -1;
  }
  if (size2->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "rand: 2nd argument must be scalar\n");
    return -1;
  }
  switch (size1->precision) {
  case SILC_INT:
    m = size1->d.scalar.i; break;
  case SILC_LONG:
    m = size1->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "rand: 1st argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  if (m < 1) {
    fprintf(stderr, "rand: 1st dimension is invalid\n");
    return -1;
  }
  switch (size2->precision) {
  case SILC_INT:
    n = size2->d.scalar.i; break;
  case SILC_LONG:
    n = size2->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "rand: 2nd argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  if (n < 1) {
    fprintf(stderr, "rand: 2nd dimension is invalid\n");
    return -1;
  }
  param[2].o = _random(m, n);
  return (param[2].o) ? 0 : -1;
}

int
silc_dense_func_size(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *size;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_func_size(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  size = silc_object_new_column_vector(SILC_LONG, 2);
  if (!size)
    return -1;
  size->d.vector.v.l[0] = object->d.matrix.m;
  size->d.vector.v.l[1] = object->d.matrix.n;
  param[1].o = size;
  return 0;
}

int
silc_dense_func_size_dim(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *mat, *dim, *size;
  int dimension;

  mat = param[0].o;
  dim = param[1].o;
#ifdef DEBUG
  printf("silc_dense_func_size_dim(0x%lx, 0x%lx)\n", mat, dim);
#endif
  if (mat->type != SILC_MATRIX_TYPE || dim->type != SILC_SCALAR_TYPE)
    return -1;
  switch (dim->precision) {
  case SILC_INT:
    dimension = dim->d.scalar.i; break;
  case SILC_LONG:
    dimension = dim->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "size: dimension must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  switch (dimension) {
  case 1:
    size = silc_object_new_long(mat->d.matrix.m); break;
  case 2:
    size = silc_object_new_long(mat->d.matrix.n); break;
  default:
    fprintf(stderr, "size: invalid dimension: %d\n", dimension);
    return -1;
  }
  param[2].o = size;
  return 0;
}

int
silc_dense_func_diagvec(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_dense_t *in;
  int length, i, j;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_func_diagvec(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  if (object->d.matrix.m != object->d.matrix.n) {
    fprintf(stderr, "dense_op: diagvec: not a square matrix\n");
    return -1;
  }
  in = SILC_DENSE(object);
  length = object->d.matrix.n;
  new_object = silc_object_new_column_vector(object->precision, length);
  if (!new_object)
    return -1;
  switch (object->precision) {
  case SILC_INT:
    for (i = j = 0; i < length; i++, j += length + 1)
      new_object->d.vector.v.i[i] = in->v.i[j];
    break;
  case SILC_LONG:
    for (i = j = 0; i < length; i++, j += length + 1)
      new_object->d.vector.v.l[i] = in->v.l[j];
    break;
  case SILC_FLOAT:
    for (i = j = 0; i < length; i++, j += length + 1)
      new_object->d.vector.v.s[i] = in->v.s[j];
    break;
  case SILC_DOUBLE:
    for (i = j = 0; i < length; i++, j += length + 1)
      new_object->d.vector.v.d[i] = in->v.d[j];
    break;
  case SILC_COMPLEX:
    for (i = j = 0; i < length; i++, j += length + 1) {
      new_object->d.vector.v.c[i*2]   = in->v.c[j*2];
      new_object->d.vector.v.c[i*2+1] = in->v.c[j*2+1];
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    for (i = j = 0; i < length; i++, j += length + 1) {
      new_object->d.vector.v.z[i*2]   = in->v.z[j*2];
      new_object->d.vector.v.z[i*2+1] = in->v.z[j*2+1];
    }
    break;
  default:
    should_not_reach_here;
  }
  param[1].o = new_object;
  return 0;
}

int
silc_dense_func_full(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *new_object, *size1, *size2;
  silc_dense_t *in;
  ssize_t m, n, i, j;
  int64_t factor;

  size1 = param[0].o;
  size2 = param[1].o;
#ifdef DEBUG
  printf("silc_dense_func_full(0x%lx, 0x%lx)\n", size1, size2);
#endif
  if (size1->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "full: 1st argument must be scalar\n");
    return -1;
  }
  if (size2->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "full: 2nd argument must be scalar\n");
    return -1;
  }
  switch (size1->precision) {
  case SILC_INT:
    m = size1->d.scalar.i; break;
  case SILC_LONG:
    m = size1->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "full: 1st argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  if (m < 1) {
    fprintf(stderr, "full: 1st dimension is invalid\n");
    return -1;
  }
  switch (size2->precision) {
  case SILC_INT:
    n = size2->d.scalar.i; break;
  case SILC_LONG:
    n = size2->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "full: 2nd argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  if (n < 1) {
    fprintf(stderr, "full: 2nd dimension is invalid\n");
    return -1;
  }
  new_object = silc_object_new_matrix(SILC_LONG,
				      silc_find_format_id(SILC_FORMAT_DENSE),
				      m, n);
  new_object->d.matrix.in = (void *)
    silc_dense_allocate(new_object->precision,
			new_object->d.matrix.m,
			new_object->d.matrix.n);
  i = n;
  factor = 1;
  while (i > 0) {
    i /= 10;
    factor *= 10;
  }
  in = SILC_DENSE(new_object);
  for (j = 0; j < new_object->d.matrix.n; j++)
    for (i = 0; i < new_object->d.matrix.m; i++)
      in->v.l[i + j * new_object->d.matrix.m] = (i + 1) * factor + (j + 1);
  param[2].o = new_object;
  return 0;
}

int
silc_dense_func_min(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_dense_t *in;
  ssize_t i;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_func_min(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  if (object->format != silc_find_format_id(SILC_FORMAT_DENSE)){
    /* XXX need a workaround */
    fprintf(stderr, "dense_op: min: not in the dense format\n");
    return -1;
  }
  new_object = silc_object_new_row_vector(object->precision,
					  object->d.matrix.n);
  if (!new_object) {
    fprintf(stderr, "dense_op: min: out of memory\n");
    return -1;
  }

#define _min_1(pre, type) {				\
    size_t j, k;					\
    type v;						\
							\
    k = i * object->d.matrix.m;				\
    v = in->v.pre[k++];					\
    for (j = 1; j < object->d.matrix.m; j++) {		\
      if (v > in->v.pre[k])				\
	v = in->v.pre[k];				\
      k++;						\
    }							\
    new_object->d.vector.v.pre[i] = v;			\
  }
#define _min_2(pre, type) {				\
    size_t j, k;					\
    type re, im, mag, re2, im2, mag2;			\
							\
    k = i * object->d.matrix.m * 2;			\
    re = in->v.pre[k++];				\
    im = in->v.pre[k++];				\
    mag = re * re + im * im;				\
    for (j = 1; j < object->d.matrix.m; j++) {		\
      re2 = in->v.pre[k];				\
      im2 = in->v.pre[k+1];				\
      mag2 = re2 * re2 + im2 * im2;			\
      if (mag > mag2) {					\
	re = re2;					\
	im = im2;					\
	mag = mag2;					\
      }							\
    }							\
    new_object->d.vector.v.pre[i*2]   = re;		\
    new_object->d.vector.v.pre[i*2+1] = im;		\
  }
  in = SILC_DENSE(object);
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _min_1(i, int32_t);
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _min_1(l, int64_t);
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _min_1(s, float);
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _min_1(d, double);
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _min_2(c, float);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _min_2(z, double);
    break;
  default:
    should_not_reach_here;
  }
  param[1].o = new_object;
  return 0;
}

/* procedures */

int
silc_dense_proc_srand(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  unsigned long s;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_proc_srand(0x%lx)\n", object);
#endif
  if (object->type != SILC_SCALAR_TYPE) {
    fprintf(stderr, "srand: argument must be scalar\n");
    return -1;
  }
  switch (object->precision) {
  case SILC_INT:
    s = object->d.scalar.i; break;
  case SILC_LONG:
    s = object->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "rand: argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  init_genrand(s); /* mt19937 */
  return 0;
}

int
silc_dense_proc_split(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  silc_dense_t *in, *new_in;
  size_t i, j, k, size;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_dense_proc_split(0x%lx, L, D, U)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  if (object->d.matrix.m != object->d.matrix.n) {
    fprintf(stderr, "dense_op: split: not a square matrix\n");
    return -1;
  }
  in = SILC_DENSE(object);
  /* lower rectangle */
  param[1].o = silc_object_duplicate(object, 1);
  new_in = SILC_DENSE(param[1].o);
  k = 0;
  for (i = 0; i < object->d.matrix.n; i++) {
    size = SILC_SCALAR_SIZE(object) * (i + 1);
    switch (object->precision) {
    case SILC_INT:
      memset(new_in->v.i + k, 0, size); break;
    case SILC_LONG:
      memset(new_in->v.l + k, 0, size); break;
    case SILC_FLOAT:
      memset(new_in->v.s + k, 0, size); break;
    case SILC_DOUBLE:
      memset(new_in->v.d + k, 0, size); break;
    case SILC_COMPLEX:
      memset(new_in->v.c + k * 2, 0, size); break;
    case SILC_DOUBLE_COMPLEX:
      memset(new_in->v.z + k * 2, 0, size); break;
    default:
      should_not_reach_here;
    }
    k += object->d.matrix.m;
  }
  /* diagonal */
  param[2].o = silc_object_duplicate(object, 0);
  new_in = SILC_DENSE(param[2].o);
  size = SILC_SCALAR_SIZE(object) * object->d.matrix.m * object->d.matrix.n;
  memset(new_in->v.i, 0, size);
  switch (object->precision) {
  case SILC_INT:
    for (i = k = 0; i < object->d.matrix.n; i++, k += object->d.matrix.m + 1)
      new_in->v.i[k] = in->v.i[k];
    break;
  case SILC_LONG:
    for (i = k = 0; i < object->d.matrix.n; i++, k += object->d.matrix.m + 1)
      new_in->v.l[k] = in->v.l[k];
    break;
  case SILC_FLOAT:
    for (i = k = 0; i < object->d.matrix.n; i++, k += object->d.matrix.m + 1)
      new_in->v.s[k] = in->v.s[k];
    break;
  case SILC_DOUBLE:
    for (i = k = 0; i < object->d.matrix.n; i++, k += object->d.matrix.m + 1)
      new_in->v.d[k] = in->v.d[k];
    break;
  case SILC_COMPLEX:
    for (i = k = 0; i < object->d.matrix.n; i++, k += object->d.matrix.m + 1) {
      new_in->v.c[k*2]   = in->v.c[k*2];
      new_in->v.c[k*2+1] = in->v.c[k*2+1];
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    for (i = k = 0; i < object->d.matrix.n; i++, k += object->d.matrix.m + 1) {
      new_in->v.z[k*2]   = in->v.z[k*2];
      new_in->v.z[k*2+1] = in->v.z[k*2+1];
    }
    break;
  default:
    should_not_reach_here;
  }
  /* upper rectangle */
  param[3].o = silc_object_duplicate(object, 1);
  new_in = SILC_DENSE(param[3].o);
  k = 0;
  for (i = 0; i < object->d.matrix.n; i++) {
    size = SILC_SCALAR_SIZE(object) * (object->d.matrix.m - i);
    switch (object->precision) {
    case SILC_INT:
      memset(new_in->v.i + (k + i), 0, size); break;
    case SILC_LONG:
      memset(new_in->v.l + (k + i), 0, size); break;
    case SILC_FLOAT:
      memset(new_in->v.s + (k + i), 0, size); break;
    case SILC_DOUBLE:
      memset(new_in->v.d + (k + i), 0, size); break;
    case SILC_COMPLEX:
      memset(new_in->v.c + (k + i) * 2, 0, size); break;
    case SILC_DOUBLE_COMPLEX:
      memset(new_in->v.z + (k + i) * 2, 0, size); break;
    default:
      should_not_reach_here;
    }
    k += object->d.matrix.m;
  }
  return 0;
}

#define SILC_ANY_INTEGER   (SILC_INT|SILC_LONG)
#define SILC_ANY_REAL      (SILC_FLOAT|SILC_DOUBLE)
#define SILC_ANY_COMPLEX   (SILC_COMPLEX|SILC_DOUBLE_COMPLEX)
#define SILC_ANY_PRECISION (SILC_ANY_INTEGER|SILC_ANY_REAL|SILC_ANY_COMPLEX)

static unary_op_table_t _unary_op_table[] = {
  /* name, op, precision, type, format, out, flags */
  {silc_dense_negate, "silc_dense_negate",
   SILC_NEGATE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   0, 0},
  {silc_dense_complex_conjugate, "silc_dense_complex_conjugate",
   SILC_COMPLEX_CONJUGATE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   0, 0},
  {silc_dense_transpose, "silc_dense_transpose",
   SILC_TRANSPOSE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   1, 0},
  {silc_dense_conjugate_transpose, "silc_dense_conjugate_transpose",
   SILC_CONJUGATE_TRANSPOSE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   1, 0},
  {NULL} /* sentinel */
};

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {silc_dense_add_matrix_matrix, "silc_dense_add_matrix_matrix",
   SILC_ADD_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   0, 0},
  {silc_dense_sub_matrix_matrix, "silc_dense_sub_matrix_matrix",
   SILC_SUBTRACT_OP,  SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   0, 0},
  {silc_dense_mul_matrix_vector, "silc_dense_mul_matrix_vector",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_COLUMN_VECTOR_TYPE, 0,
   2, 0},
  {silc_dense_mul_vector_matrix, "silc_dense_mul_vector_matrix",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_ROW_VECTOR_TYPE, 0, SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   2, 0},
  {silc_dense_mul_matrix_matrix, "silc_dense_mul_matrix_matrix",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   2, 0},
  {silc_dense_mul_vector_vector, "silc_dense_mul_vector_vector",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_COLUMN_VECTOR_TYPE, 0, SILC_ROW_VECTOR_TYPE, 0,
   2, 0},
  {silc_dense_mul_matrix_scalar, "silc_dense_mul_matrix_scalar",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_SCALAR_TYPE, 0,
   0, 0},
  {silc_dense_mul_scalar_matrix, "silc_dense_mul_scalar_matrix",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_SCALAR_TYPE, 0, SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   1, 0},
  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {silc_dense_func_diagvec, "silc_dense_func_diagvec",
   "vector diagvec(matrix)", 0},
  {silc_dense_func_zeros_1, "silc_dense_func_zeros_1",
   "matrix zeros(scalar)", 0},
  {silc_dense_func_zeros_2, "silc_dense_func_zeros_2",
   "matrix zeros(scalar, scalar)", 0},
  {silc_dense_func_ones_1, "silc_dense_func_ones_1",
   "matrix ones(scalar)", 0},
  {silc_dense_func_ones_2, "silc_dense_func_ones_2",
   "matrix ones(scalar, scalar)", 0},
  {silc_dense_func_rand_1, "silc_dense_func_rand_1",
   "matrix rand(scalar)", 0},
  {silc_dense_func_rand_2, "silc_dense_func_rand_2",
   "matrix rand(scalar, scalar)", 0},
  {silc_dense_func_size, "silc_dense_func_size",
   "column vector size(matrix)", 0},
  {silc_dense_func_size_dim, "silc_dense_func_size_dim",
   "scalar size(matrix, scalar)", 0},
  {silc_dense_func_full, "silc_dense_func_full",
   "matrix full(scalar, scalar)", 0},
  {silc_dense_func_min, "silc_dense_func_min",
   "vector min(matrix)", 0},
  {NULL} /* sentinel */
};

static procedure_table_t _procedure_table[] = {
  /* name, sig_text, flags */
  {silc_dense_proc_srand, "silc_dense_proc_srand",
   "srand(scalar)", 0},
  {silc_dense_proc_split, "silc_dense_proc_split",
   "split(matrix in, matrix out, matrix out, matrix out)", 0},
  {NULL} /* sentinel */
};

void
init_dense_op(char *module_name)
{
  init_genrand(time(NULL)); /* mt19937 */

  silc_register_module(module_name,
		       _unary_op_table,
		       _binary_op_table,
		       _function_table,
		       _procedure_table);
}
