/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: fftss.c,v 1.5 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "object.h"
#include "module.h"
#include "dense.h"

#include "fftss.h"

#define MODULE_NAME "fftss"

static fftss_plan p1f = NULL, p1b = NULL;
static fftss_plan p2f = NULL, p2b = NULL;
static fftss_plan p3f = NULL, p3b = NULL;

static long n1f = 0, n1b = 0;
static long nx2f = 0, nx2b = 0;
static long ny2f = 0, ny2b = 0;
static long nx3f = 0, nx3b = 0;
static long ny3f = 0, ny3b = 0;
static long nz3f = 0, nz3b = 0;

int
fft1d(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  size_t status;
  fftss_plan p;

  x = param[0].o;
#ifdef DEBUG
  printf("fft1d(0x%lx)\n", x);
#endif
  if (x->type != SILC_COLUMN_VECTOR_TYPE &&
      x->type != SILC_ROW_VECTOR_TYPE) {
    fprintf(stderr, MODULE_NAME ": not a vector\n");
    return -1;
  }
  if (x->precision != SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, MODULE_NAME ": only double complex is supported\n");
    return -1;
  }
  y = silc_object_duplicate(x, 0);
  if (n1f == (long)x->d.vector.length)
    p = p1f;
  else {
    if (p1f) fftss_destroy_plan(p1f);
    p = fftss_plan_dft_1d(x->d.vector.length, x->d.vector.v.z, y->d.vector.v.z,
			  FFTSS_FORWARD,
#ifdef DEBUG
			  FFTSS_VERBOSE|
#endif
			  FFTSS_MEASURE|FFTSS_PRESERVE_INPUT);
    if (p == NULL) {
      silc_object_free(y);
      return -1;
    }
    p1f = p;
    n1f = x->d.vector.length;
  }
  fftss_execute_dft(p, x->d.vector.v.z, y->d.vector.v.z);
  param[1].o = y;
  return 0;
}

int
ifft1d(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  size_t status;
  fftss_plan p;

  x = param[0].o;
#ifdef DEBUG
  printf("fft1d(0x%lx)\n", x);
#endif
  if (x->type != SILC_COLUMN_VECTOR_TYPE &&
      x->type != SILC_ROW_VECTOR_TYPE) {
    fprintf(stderr, MODULE_NAME ": not a vector\n");
    return -1;
  }
  if (x->precision != SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, MODULE_NAME ": only double complex is supported\n");
    return -1;
  }
  y = silc_object_duplicate(x, 0);
  if (n1b == (long)x->d.vector.length)
    p = p1b;
  else {
    if (p1b) fftss_destroy_plan(p1b);
    p = fftss_plan_dft_1d(x->d.vector.length, x->d.vector.v.z, y->d.vector.v.z,
			  FFTSS_BACKWARD,
#ifdef DEBUG
			  FFTSS_VERBOSE|
#endif
			  FFTSS_MEASURE|FFTSS_PRESERVE_INPUT);
    if (p == NULL) {
      silc_object_free(y);
      return -1;
    }
    p1b = p;
    n1b = x->d.vector.length;
  }
  fftss_execute_dft(p, x->d.vector.v.z, y->d.vector.v.z);
  param[1].o = y;
  return 0;
}

#if !defined(FFTSS_VERSION) || FFTSS_VERSION > 1

int
fft2d(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  size_t status;
  fftss_plan p;
#ifdef DEBUG
  double t;
#endif

  x = param[0].o;
#ifdef DEBUG
  printf("fft2d(0x%lx)\n", x);
#endif
  if (x->type != SILC_MATRIX_TYPE ||
      x->format != silc_find_format_id(SILC_FORMAT_DENSE)) {
    fprintf(stderr, MODULE_NAME ": not a dense matrix\n");
    return -1;
  }
  if (x->precision != SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, MODULE_NAME ": only double complex is supported\n");
    return -1;
  }
  y = silc_object_duplicate(x, 0);
  if (nx2f == (long)x->d.matrix.m &&
      ny2f == (long)x->d.matrix.n)
    p = p2f;
  else {
    if (p2f) fftss_destroy_plan(p2f);
    p = fftss_plan_dft_2d(x->d.matrix.m, x->d.matrix.n, x->d.matrix.m,
			  SILC_DENSE(x)->v.z, SILC_DENSE(y)->v.z,
			  FFTSS_FORWARD,
#ifdef DEBUG
			  FFTSS_VERBOSE|
#endif
			  FFTSS_MEASURE|FFTSS_PRESERVE_INPUT);
    if (p == NULL) {
      silc_object_free(y);
      return -1;
    }
    p2f = p;
    nx2f = x->d.matrix.m;
    ny2f = x->d.matrix.n;
  }
#ifdef DEBUG
  t = fftss_get_wtime();
#endif
  fftss_execute_dft(p, SILC_DENSE(x)->v.z, SILC_DENSE(y)->v.z);
#ifdef DEBUG
  t = fftss_get_wtime() - t;
  printf("%lf sec, %lf GFLOPS.\n", t,
	 5.0 * (double)nx2f * (double)ny2f 
	 * log((double)nx2f * (double)ny2f) / log((double)2.0) / t
	 * 0.000000001);
#endif
  param[1].o = y;
  return 0;
}

int
ifft2d(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  size_t status;
  fftss_plan p;

  x = param[0].o;
#ifdef DEBUG
  printf("fft(0x%lx)\n", x);
#endif
  if (x->type != SILC_MATRIX_TYPE ||
      x->format != silc_find_format_id(SILC_FORMAT_DENSE)) {
    fprintf(stderr, MODULE_NAME ": not a dense matrix\n");
    return -1;
  }
  if (x->precision != SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, MODULE_NAME ": only double complex is supported\n");
    return -1;
  }
  y = silc_object_duplicate(x, 0);
  if (nx2b == (long)x->d.matrix.m &&
      ny2b == (long)x->d.matrix.n)
    p = p2b;
  else {
    if (p2b) fftss_destroy_plan(p2b);
    p = fftss_plan_dft_2d(x->d.matrix.m, x->d.matrix.n, x->d.matrix.m,
			  SILC_DENSE(x)->v.z, SILC_DENSE(y)->v.z,
			  FFTSS_BACKWARD,
#ifdef DEBUG
			  FFTSS_VERBOSE|
#endif
			  FFTSS_MEASURE|FFTSS_PRESERVE_INPUT);
    if (p == NULL) {
      silc_object_free(y);
      return -1;
    }
    p2b = p;
    nx2b = x->d.matrix.m;
    ny2b = x->d.matrix.n;
  }
  fftss_execute_dft(p, SILC_DENSE(x)->v.z, SILC_DENSE(y)->v.z);
  param[1].o = y;
  return 0;
}

#endif /* !defined(FFTSS_VERSION) || FFTSS_VERSION > 1 */

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {fft1d, "fft1d", "vector fft(vector)", 0},
  {ifft1d, "ifft1d", "vector ifft(vector)", 0},
#if !defined(FFTSS_VERSION) || FFTSS_VERSION > 1
  {fft2d, "fft2d", "matrix fft(matrix)", 0},
  {ifft2d, "ifft2d", "matrix ifft(matrix)", 0},
#endif
  {NULL} /* sentinel */
};

void
init_fftss()
{
  silc_register_module(MODULE_NAME,
		       NULL,   /* _unary_op_table */
		       NULL,   /* _binary_op_table */
		       _function_table,
		       NULL);  /* _procedure_table */
}
