/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: interp.c,v 1.5 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h> /* for exp() and M_PI */
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "interp.h"
#include "dense.h"
#include "sparse_crs.h"
#include "wtime.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

typedef struct {
  char name[SILC_MAXNAMELEN+1];
  int is_temp;
  silc_modifier_t *modifiers;
} value_t;

static value_t *value_new(char *, int);
static value_t *evaluate_expression(expr_t *);
static void value_free(value_t *);
static value_t **evaluate_expression_list(expr_t **, int);
static void value_list_free(value_t **, int);
static int evaluate_subscript(subscript_t *, clist_t *[]);

static silc_object_t *value_get_object(value_t *);
static silc_object_t *value_get_modified_object(value_t *);
static silc_modifier_t *value_get_modifiers(value_t *);

silc_object_t *do_binary_op(int, value_t *, value_t *);

#ifdef TIME_FIND_MFUNC
extern double _find_mfunc_time; /* server.c */
#endif
#ifdef TIME_CALL_MFUNC
extern double _call_mfunc_time; /* server.c */
#endif
#ifdef TIME_SUBSCRIPT_VECTOR
extern double _vector_subsasgn_time; /* server.c */
#endif

static int
call_mfunc_impl(silc_mfunc_t *mfunc, int n_params, silc_mfunc_param_t *param)
{
  int status;
#ifdef TIME_CALL_MFUNC
  double t, elapsed;

  t = silc_wtime();
#endif
  status = (mfunc->impl)(n_params, param);
#ifdef TIME_CALL_MFUNC
  elapsed = silc_wtime() - t;
  _call_mfunc_time += elapsed;
#if defined(DEBUG) || TIME_CALL_MFUNC > 1
  printf("mfunc: %s in %e sec.\n", mfunc->mfunc_name, elapsed);
#endif
#endif
  return status;
}

void
error(char *format, ...)
{
  va_list ap;
  char message[1024];

  va_start(ap, format);
  vsnprintf(message, sizeof(message), format, ap);
  va_end(ap);
  fprintf(stderr, "Error: %s\n", message);
}

#define ordinal_suffix(n) ((i / 10 == 1) ? "th" : \
			   (i % 10 == 0) ? "st" : \
			   (i % 10 == 1) ? "nd" : \
			   (i % 10 == 2) ? "rd" : "th")

#define print_value(v) \
        printf(((v)->modifiers == NULL) ? "%s" : "mod(%s)", (v)->name)
#define print_assignment(op) { int i; \
	printf("do" op "assign("); \
	printf((n == 0) ? "%s" : "%s:subscript[%d]", name, n); \
	printf(", "); print_value(rhs); printf(") not implemented\n"); \
        for (i = 0; i < n; i++) { \
          printf("  %s:subscript[%d]: ",name,i); clist_pprint(subscript[i]); \
        }}

static void
do_assign_vector(silc_object_t *dest, clist_t *dest_subscript[3], int dest_n,
		 silc_object_t *src, silc_modifier_t *modifiers)
{
  int *dest_v, *src_v;
  ssize_t dest_length, i, j, k;
  void *src_array;
  int src_n, copy_pattern;
#ifdef TIME_SUBSCRIPT_VECTOR
  double t, elapsed;

  t = silc_wtime();
#endif

#define _PAT_IDENT           0
#define _PAT_NEG             1
#define _PAT_CONJ            4
#define _PAT_CONJ_NEG        5

  if (dest_n != 1)
    should_not_reach_here;
  dest_v = clist_expand(dest_subscript[0]);
  dest_length = clist_length(dest_subscript[0]);
  if (!modifiers) {
    src_n = 0;
    copy_pattern = 0;
  } else {
    src_n = modifiers->n;
    switch (src->precision) {
    case SILC_INT:
    case SILC_LONG:
    case SILC_FLOAT:
    case SILC_DOUBLE:
      copy_pattern = ((modifiers->negate    ? 1 : 0));
      break;
    case SILC_COMPLEX:
    case SILC_DOUBLE_COMPLEX:
      copy_pattern = ((modifiers->conjugate ? 4 : 0) +
		      (modifiers->negate    ? 1 : 0));
      break;
    default:
      should_not_reach_here;
    }
  }
  switch (src->type) {
  case SILC_SCALAR_TYPE:
    if (src_n > 0)
      should_not_reach_here;
    src_array = &src->d.scalar.i;
    break;
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    src_array = src->d.vector.v.i;
    if (src_n > 0) {
      if (src_n != 1)
	should_not_reach_here;
      clist_set_defaults(modifiers->subscript[0],
			 0, src->d.vector.length - 1);
      src_v = clist_expand(modifiers->subscript[0]);
    }
  }
  if (src_n == 0) { /* no subscript */
    switch (dest->precision) {
    case SILC_INT:
      {
	int32_t *p = dest->d.vector.v.i, *q = (int32_t *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[i];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[i];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_LONG:
      {
	int64_t *p = dest->d.vector.v.l, *q = (int64_t *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[i];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[i];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_FLOAT:
      {
	float *p = dest->d.vector.v.s, *q = (float *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[i];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[i];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_DOUBLE:
      {
	double *p = dest->d.vector.v.d, *q = (double *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[i];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[i];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_COMPLEX:
      {
	float *p = dest->d.vector.v.c, *q = (float *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   = q[j];
	    p[k+1] = q[j+1];
	  }
	  break;
	case _PAT_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   = -q[j];
	    p[k+1] = -q[j+1];
	  }
	  break;
	case _PAT_CONJ:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   =  q[j];
	    p[k+1] = -q[j+1];
	  }
	  break;
	case _PAT_CONJ_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   = -q[j];
	    p[k+1] =  q[j+1];
	  }
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double *p = dest->d.vector.v.z, *q = (double *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   = q[j];
	    p[k+1] = q[j+1];
	  }
	  break;
	case _PAT_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   = -q[j];
	    p[k+1] = -q[j+1];
	  }
	  break;
	case _PAT_CONJ:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   =  q[j];
	    p[k+1] = -q[j+1];
	  }
	  break;
	case _PAT_CONJ_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = i * 2;
	    k = dest_v[i] * 2;
	    p[k]   = -q[j];
	    p[k+1] =  q[j+1];
	  }
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    default:
      should_not_reach_here;
    }
  } else { /* RHS with subscript */
    switch (dest->precision) {
    case SILC_INT:
      {
	int32_t *p = dest->d.vector.v.i, *q = (int32_t *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[src_v[i]];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[src_v[i]];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_LONG:
      {
	int64_t *p = dest->d.vector.v.l, *q = (int64_t *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[src_v[i]];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[src_v[i]];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_FLOAT:
      {
	float *p = dest->d.vector.v.s, *q = (float *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[src_v[i]];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[src_v[i]];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_DOUBLE:
      {
	double *p = dest->d.vector.v.d, *q = (double *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = q[src_v[i]];
	  break;
	case _PAT_NEG:
#pragma omp parallel for
	  for (i = 0; i < dest_length; i++)
	    p[dest_v[i]] = -q[src_v[i]];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_COMPLEX:
      {
	float *p = dest->d.vector.v.c, *q = (float *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   = q[k];
	    p[j+1] = q[k+1];
	  }
	  break;
	case _PAT_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   = -q[k];
	    p[j+1] = -q[k+1];
	  }
	  break;
	case _PAT_CONJ:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   =  q[k];
	    p[j+1] = -q[k+1];
	  }
	  break;
	case _PAT_CONJ_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   = -q[k];
	    p[j+1] =  q[k+1];
	  }
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double *p = dest->d.vector.v.z, *q = (double *)src_array;
	switch (copy_pattern) {
	case _PAT_IDENT:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   = q[k];
	    p[j+1] = q[k+1];
	  }
	  break;
	case _PAT_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   = -q[k];
	    p[j+1] = -q[k+1];
	  }
	  break;
	case _PAT_CONJ:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   =  q[k];
	    p[j+1] = -q[k+1];
	  }
	  break;
	case _PAT_CONJ_NEG:
#pragma omp parallel for private(j, k)
	  for (i = 0; i < dest_length; i++) {
	    j = dest_v[i] * 2;
	    k = src_v[i] * 2;
	    p[j]   = -q[k];
	    p[j+1] =  q[k+1];
	  }
	  break;
	default:
	  should_not_reach_here;
	}
      }
      break;
    default:
      should_not_reach_here;
    }
    free(src_v);
  }
  free(dest_v);
#ifdef TIME_SUBSCRIPT_VECTOR
  elapsed = silc_wtime() - t;
  _vector_subsasgn_time += elapsed;
#ifdef DEBUG
  printf("do_assign_vector: %e sec.\n", elapsed);
#endif
#endif
}

static int
do_assign(char *name, clist_t *subscript[], int n, value_t *rhs)
{
  silc_object_t *dest, *src, *object;
  silc_format_t *table;
  int with_matrix;

  if (n == 0) { /* the simplest assignment */
    object = value_get_modified_object(rhs);
    if (!object) {
      error("undefined variable: %s", rhs->name);
      return -1;
    }
    if (var_get(name))
      var_free(name);
    var_add(name, object);
#ifdef DEBUG
    printf("assigned to %s:\n", name);
    silc_object_pprint(object, 1);
#endif
  } else {
    object = var_get(rhs->name);
    if (!object) {
      should_not_reach_here;
      return -1;
    }
    dest = var_get(name);
    if (!dest) {
      should_not_reach_here;
      return -1;
    }
    if (dest->precision != object->precision) {
      error("*** precision mismatch"); /* XXX need a workaround */
      return -1;
    }
    /*
     * - implemented
     *   vector[] = scalar    (core)
     *   vector[] = vector    (core)
     *   vector[] = vector[]  (core)
     *   vector[] = matrix    (storage format module)
     *   vector[] = matrix[]  (storage format module)
     *   matrix[] = scalar    (storage format module)
     *   matrix[] = vector    (storage format module)
     *   matrix[] = vector[]  (storage format module)
     *   matrix[] = matrix    (storage format module)
     *   matrix[] = matrix[]  (storage format module)
     * - pending
     *   carray[] = scalar    (core)
     *   carray[] = vector    (core)
     *   carray[] = vector[]  (core)
     *   carray[] = matrix    (storage format module)
     *   carray[] = matrix[]  (storage format module)
     *   carray[] = carray    (core)
     *   carray[] = carray[]  (core)
     */
    with_matrix = 0;
    switch (dest->type) {
    case SILC_RANGE_TYPE:
      dest = silc_object_new_column_vector_from_range(var_get_free(name));
      var_add(name, dest);
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      break;
    case SILC_MATRIX_TYPE:
      with_matrix = 1;
      break;
    default:
      error("invalid assignment");
      return -1;
    }
    src = object;
    switch (src->type) {
    case SILC_SCALAR_TYPE:
      break;
    case SILC_RANGE_TYPE:
      src = silc_object_new_column_vector_from_range(object);
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      break;
    case SILC_MATRIX_TYPE:
      if (dest->type == SILC_MATRIX_TYPE &&
	  dest->format != src->format) {
	error("*** storage format mismatch"); /* XXX need a workaround */
	return -1;
      }
      with_matrix = 1;
      break;
    default:
      error("invalid assignment");
      return -1;
    }
    if (with_matrix) {
      table = silc_find_format_proc_table
	((dest->type == SILC_MATRIX_TYPE) ? dest->format : src->format);
      if (!table)
	return -1;
      if (!table->proc_assign) {
	error("proc_assign() not defined"); /* XXX need a workaround */
	return -1;
      }
      if (table->proc_assign(dest, subscript, n, src, rhs->modifiers) < 0) {
	error("proc_assign() failed");
	return -1; /* XXX assume that the dest object is unchanged */
      }
    } else {
      switch (dest->type) {
      case SILC_COLUMN_VECTOR_TYPE:
      case SILC_ROW_VECTOR_TYPE:
	do_assign_vector(dest, subscript, n, src, rhs->modifiers);
	break;
      default:
	should_not_reach_here;
      }
    }
    if (src != object)
      silc_object_free(src);
  }
  return 0;
}

static int
do_prefer(char *name)
{
  if (silc_prefer_module(name) < 0) {
    error("module not found: %s", name);
    return -1;
  }
  return 0;
}

static int
do_rename(char *old_name, char *new_name)
{
  silc_object_t *object;

  object = var_get_free(old_name);
  if (!object) {
    error("undefined variable: %s", old_name);
    return -1;
  }
  var_free(new_name);
  var_add(new_name, object);
  return 0;
}

#define proc_param_error(i, iotype, msg) \
        error("%d%s %s argument: %s", i+1, ordinal_suffix(i), iotype, msg)
	
static int
do_proc_call(char *name, int n_params, value_t **value)
{
  signature_t *sig;
  silc_object_t *object;
  silc_mfunc_t *mfunc;
  silc_mfunc_param_t *param;
  int status, i, j;
#if defined(TIME_FIND_MFUNC) ||  defined(TIME_CALL_MFUNC)
  double t, elapsed;
#endif

  sig = silc_find_procedure(name);
  if (!sig) {
    error("procedure not found: %s", name);
    return -1;
  }
  if (sig->n_params != n_params) {
    error("illegal number of arguments: %s", name);
    return -1;
  }
#ifdef TIME_FIND_MFUNC
  t = silc_wtime();
#endif
  mfunc = silc_find_procedure_impl(name);
  if (!mfunc) {
    error("FATAL: no module function for %s found", name);
    return -1;
  }
#ifdef TIME_FIND_MFUNC
  elapsed = silc_wtime() - t;
  _find_mfunc_time += elapsed;
#endif
  param = (silc_mfunc_param_t *)malloc(sizeof(silc_mfunc_param_t) * n_params);
  for (i = 0; i < n_params; i++) {
    param[i].o = var_get(value[i]->name);
    param[i].m = value[i]->modifiers;
    switch (sig->param[i].iotype) {
    case SILC_PARAM_IN:
      if (!param[i].o) {
	proc_param_error(i, "in", "undefined variable");
	goto invalid_param;
      }
      break;
    case SILC_PARAM_OUT:
      if (param[i].m) {
	proc_param_error(i, "out", "cannot specify modifier(s)");
	goto invalid_param;
      }
      break;
    case SILC_PARAM_INOUT:
      if (!param[i].o) {
	proc_param_error(i, "inout", "undefined variable");
	goto invalid_param;
      }
      if (param[i].m) {
	proc_param_error(i, "inout", "cannot specify modifier(s)");
	goto invalid_param;
      }
      break;
    default:
      should_not_reach_here;
    }
    /* XXX: check types of parameters */
  }
 invalid_param:
  if (i < n_params) {
    free(param);
    free(mfunc);
    return -1;
  }
#if 1 /* XXX strip all modifiers */
  for (i = 0; i < n_params; i++) {
    if (value[i]->modifiers) {
      param[i].o = modifier_apply_all(param[i].o, value[i]->modifiers);
      param[i].m = NULL;
    } else if (param[i].o->type == SILC_RANGE_TYPE) {
#ifdef TIME_CALL_MFUNC
      t = silc_wtime();
#endif
      param[i].o = silc_object_new_column_vector_from_range(param[i].o);
#ifdef TIME_CALL_MFUNC
      elapsed = silc_wtime() - t;
      _call_mfunc_time += elapsed;
#endif
    }
  }
#endif
#ifdef VANILLA_MPI
  if (silc_ipc_mpi_mfunc(mfunc->flags))
    silc_ipc_init(mfunc->module_name, mfunc->mfunc_name);
#endif
  status = call_mfunc_impl(mfunc, n_params, param);
  if (status == 0) {
    for (i = 0; i < n_params; i++) {
      switch (sig->param[i].iotype) {
      case SILC_PARAM_IN:
	break;
      case SILC_PARAM_OUT:
	if (param[i].o) {
	  if (var_get(value[i]->name))
	    var_free(value[i]->name);
	  var_add(value[i]->name, param[i].o);
	} else {
	  proc_param_error(i, "out", "NULL (illegal module function)");
	}
	break;
      case SILC_PARAM_INOUT:
	break;
      default:
	should_not_reach_here;
      }
    }
  }
#if 1 /* XXX */
  for (i = 0; i < n_params; i++) {
    if (value[i]->modifiers || param[i].o->type == SILC_RANGE_TYPE)
      silc_object_free(param[i].o);
  }
#endif
  free(param);
  free(mfunc);
  if (status < 0) {
    error("FATAL: procedure %s failed", name);
    return -1;
  }
  return 0;
}

#define dimension_error(i, msg)			\
	error("%s dimension: %s",		\
	      (i == 0) ? "1st" :		\
	      (i == 1) ? "2nd" :		\
	      (i == 2) ? "3rd" : "<unknown>",	\
	      msg);

int
execute_statment(stmt_t *stmt)
{
  char *name, temp[SILC_MAXNAMELEN+1];
  silc_object_t *dest, *src;
  clist_t *subscript[3];
  int i, n, status;
  value_t *lhs, *rhs, **list;

  switch(stmt->type) {
  case AssignmentStmtType:
    name = ASSIGNMENT(stmt)->name;
    if (ASSIGNMENT(stmt)->subscript) {
      int dimension[3];

      /* validate the LHS subscript */
      dest = var_get(name);
      if (!dest) {
	error("undefined variable: %s", name);
	return -1;
      }
      n = ASSIGNMENT(stmt)->subscript->n;
      switch (dest->type) {
      case SILC_SCALAR_TYPE:
	error("cannot apply subscript to scalar"); /* XXX */
	return -1;
      case SILC_RANGE_TYPE:
      case SILC_COLUMN_VECTOR_TYPE:
      case SILC_ROW_VECTOR_TYPE:
	if (n != 1) {
	  error("cannot apply %d-dimensional subscript to vector", n);
	  return -1;
	}
	dimension[0] = dest->d.vector.length;
	break;
      case SILC_MATRIX_TYPE:
	if (n != 2) {
	  error("cannot apply %d-dimensional subscript to matrix", n);
	  return -1;
	}
	dimension[0] = dest->d.matrix.m;
	dimension[1] = dest->d.matrix.n;
	break;
      case SILC_CUBIC_ARRAY_TYPE:
	if (n != 3) {
	  error("cannot apply %d-dimensional subscript to cubic array", n);
	  return -1;
	}
	dimension[0] = dest->d.cubic_array.l;
	dimension[1] = dest->d.cubic_array.m;
	dimension[2] = dest->d.cubic_array.n;
	break;
      default:
	should_not_reach_here;
      }
      if (evaluate_subscript(ASSIGNMENT(stmt)->subscript, subscript) < 0)
	return -1;
      for (i = 0; i < n; i++) {
	clist_set_defaults(subscript[i], 0, dimension[i] - 1);
	if (clist_min(subscript[i]) < 0 ||
	    clist_max(subscript[i]) > dimension[i] - 1) {
	  dimension_error(i, "subscript index out of range");
	  for (i = 0; i < n; i++)
	    clist_free(subscript[i]);
	  return -1;
	}
      }
    } else {
      n = 0;
    }
    rhs = evaluate_expression(ASSIGNMENT(stmt)->rhs);
    if (rhs == NULL)
      return -1;
    if (ASSIGNMENT(stmt)->op == AssignOp) {
      if (strcmp(rhs->name, name) == 0 && n > 0) {
	/* if both LHS (with subscript) and RHS refer to the same object */
	if (rhs->is_temp)
	  should_not_reach_here;
	src = value_get_modified_object(rhs); /* make a duplicated object */
	value_free(rhs);
	rhs = value_new(var_add_temp(src, temp), 1);
      }
    } else { /* augmented assignment */
      lhs = value_new(name, 0);
      lhs->modifiers = modifier_new();
      lhs->modifiers->n = n;
      for (i = 0; i < n; i++)
	lhs->modifiers->subscript[i] = clist_duplicate(subscript[i]);
#if 1 /* XXX strip all modifiers */
      if (n > 0) {
	dest = value_get_modified_object(lhs);
	value_free(lhs);
	lhs = value_new(var_add_temp(dest, temp), 1);
      }
      if (rhs->modifiers) {
	src = value_get_modified_object(rhs);
	value_free(rhs);
	rhs = value_new(var_add_temp(src, temp), 1);
      } 
#endif
      switch (ASSIGNMENT(stmt)->op) {
      case Add_AssignOp:
	src = do_binary_op(SILC_ADD_OP, lhs, rhs);
	break;
      case Subtract_AssignOp:
	src = do_binary_op(SILC_SUBTRACT_OP, lhs, rhs);
	break;
      case Multiply_AssignOp:
	src = do_binary_op(SILC_MULTIPLY_OP, lhs, rhs);
	break;
      case MultiplyElementwise_AssignOp:
	src = do_binary_op(SILC_MULTIPLY_ELEMENTWISE_OP, lhs, rhs);
	break;
      case Divide_AssignOp:
	src = do_binary_op(SILC_DIVIDE_OP, lhs, rhs);
	break;
      case DivideElementwise_AssignOp:
	src = do_binary_op(SILC_DIVIDE_ELEMENTWISE_OP, lhs, rhs);
	break;
      case Remainder_AssignOp:
	src = do_binary_op(SILC_REMAINDER_OP, lhs, rhs);
	break;
      default:
	should_not_reach_here;
      }
      value_free(lhs);
      value_free(rhs);
      if (!src)
	return -1;
      rhs = value_new(var_add_temp(src, temp), 1);
    }
    if (n > 0) {
      int dest_length[3], src_length[3];
      src = var_get(rhs->name);
      if (!src)
	should_not_reach_here;
      if (rhs->modifiers && rhs->modifiers->n > 0) { /* RHS with subscript */
	switch (src->type) {
	case SILC_COLUMN_VECTOR_TYPE:
	  clist_set_defaults(rhs->modifiers->subscript[0],
			     0, src->d.vector.length - 1);
	  src_length[0] = clist_length(rhs->modifiers->subscript[0]);
	  src_length[1] = 1;
	  src_length[2] = 0;
	  break;
	case SILC_ROW_VECTOR_TYPE:
	  clist_set_defaults(rhs->modifiers->subscript[0],
			     0, src->d.vector.length - 1);
	  src_length[0] = 1;
	  src_length[1] = clist_length(rhs->modifiers->subscript[0]);
	  src_length[2] = 0;
	  break;
	case SILC_MATRIX_TYPE:
	  clist_set_defaults(rhs->modifiers->subscript[0],
			     0, src->d.matrix.m - 1);
	  clist_set_defaults(rhs->modifiers->subscript[1],
			     0, src->d.matrix.n - 1);
	  src_length[0] = clist_length(rhs->modifiers->subscript[0]);
	  src_length[1] = clist_length(rhs->modifiers->subscript[1]);
	  src_length[2] = 0;
	  break;
	case SILC_CUBIC_ARRAY_TYPE:
	  clist_set_defaults(rhs->modifiers->subscript[0],
			     0, src->d.cubic_array.l - 1);
	  clist_set_defaults(rhs->modifiers->subscript[1],
			     0, src->d.cubic_array.m - 1);
	  clist_set_defaults(rhs->modifiers->subscript[2],
			     0, src->d.cubic_array.n - 1);
	  src_length[0] = clist_length(rhs->modifiers->subscript[0]);
	  src_length[1] = clist_length(rhs->modifiers->subscript[1]);
	  src_length[2] = clist_length(rhs->modifiers->subscript[2]);
	  break;
	default:
	  should_not_reach_here;
	}
      } else { /* RHS without subscript */
	switch (src->type) {
	case SILC_SCALAR_TYPE:
	  src_length[0] = 1;
	  src_length[1] = 1;
	  src_length[2] = 0;
	  break;
	case SILC_COLUMN_VECTOR_TYPE:
	  src_length[0] = src->d.vector.length;
	  src_length[1] = 1;
	  src_length[2] = 0;
	  break;
	case SILC_ROW_VECTOR_TYPE:
	  src_length[0] = 1;
	  src_length[1] = src->d.vector.length;
	  src_length[2] = 0;
	  break;
	case SILC_MATRIX_TYPE:
	  src_length[0] = src->d.matrix.m;
	  src_length[1] = src->d.matrix.n;
	  src_length[2] = 0;
	  break;
	case SILC_CUBIC_ARRAY_TYPE:
	  src_length[0] = src->d.cubic_array.l;
	  src_length[1] = src->d.cubic_array.m;
	  src_length[2] = src->d.cubic_array.n;
	default:
	  should_not_reach_here;
	}
      }
      if (rhs->modifiers && rhs->modifiers->transpose) {
	switch (src->type) {
	case SILC_SCALAR_TYPE:
	case SILC_COLUMN_VECTOR_TYPE:
	case SILC_ROW_VECTOR_TYPE:
	case SILC_MATRIX_TYPE:
	  i = src_length[0]; src_length[0] = src_length[1]; src_length[1] = i;
	  break;
	default:
	  should_not_reach_here;
	}
      }
      switch (dest->type) {
      case SILC_RANGE_TYPE:
      case SILC_COLUMN_VECTOR_TYPE:
	dest_length[0] = clist_length(subscript[0]);
	dest_length[1] = 1;
	dest_length[2] = 0;
	break;
      case SILC_ROW_VECTOR_TYPE:
	dest_length[0] = 1;
	dest_length[1] = clist_length(subscript[0]);
	dest_length[2] = 0;
	break;
      case SILC_MATRIX_TYPE:
	dest_length[0] = clist_length(subscript[0]);
	dest_length[1] = clist_length(subscript[1]);
	dest_length[2] = 0;
	break;
      case SILC_CUBIC_ARRAY_TYPE:
	dest_length[0] = clist_length(subscript[0]);
	dest_length[1] = clist_length(subscript[1]);
	dest_length[2] = clist_length(subscript[2]);
	break;
      default:
	should_not_reach_here;
      }
      for (i = 0; i < 3; i++) {
	if (dest_length[i] != src_length[i]) {
	  dimension_error(i, "subscript size mismatch");
	  for (i = 0; i < n; i++)
	    clist_free(subscript[i]);
	  return -1;
	}
      }
    }
    status = do_assign(name, subscript, n, rhs);
    value_free(rhs);
    for (i = 0; i < n; i++)
      clist_free(subscript[i]);
    break;
  case PreferStmtType:
    status = do_prefer(PREFER(stmt)->name);
    break;
  case RenameStmtType:
    status = do_rename(RENAME(stmt)->old_name, RENAME(stmt)->new_name);
    break;
  case ProcCallStmtType:
    list = evaluate_expression_list(PROC_CALL(stmt)->args,
				    PROC_CALL(stmt)->n_args);
    status = do_proc_call(PROC_CALL(stmt)->name,
			  PROC_CALL(stmt)->n_args,
			  list);
    value_list_free(list, PROC_CALL(stmt)->n_args);
    break;
  default:
    should_not_reach_here;
  }
  return status;
}

/* TODO: add support for sparse matrix */
static silc_object_t *
new_cubic_array(expr_t *v[], int length)
{
  value_t *value;
  silc_object_t *object, **element;
  int i, j, k, l, m, n, matched;
  size_t msize;

  element = (silc_object_t **)malloc(sizeof(silc_object_t *) * length);
  for (i = 0; i < length; i++) {
    value = evaluate_expression(v[i]);
    if (!value)
      break;
    object = value_get_modified_object(value);
    value_free(value);
    if (!object)
      break;
    element[i] = object;
  }
  if (i < length) {
    while (i > 0)
      silc_object_free(element[--i]);
    return NULL;
  }
  switch (element[0]->type) {
  case SILC_SCALAR_TYPE:
    l = m = n = 1;
    break;
  case SILC_RANGE_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
    l = m = 1;
    n = element[0]->d.vector.length;
    break;
  case SILC_ROW_VECTOR_TYPE:
    l = n = 1;
    m = element[0]->d.vector.length;
    break;
  case SILC_MATRIX_TYPE:
    l = 1;
    m = element[0]->d.matrix.m;
    n = element[0]->d.matrix.n;
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    l = element[0]->d.cubic_array.l;
    m = element[0]->d.cubic_array.m;
    n = element[0]->d.cubic_array.n;
    break;
  default:
    should_not_reach_here;
  }
  for (i = 1; i < length; i++) {
    if (element[i]->precision != element[0]->precision) {
      error("precision mismatch in cubic array elements");
      goto onError;
    }
    switch (element[i]->type) {
    case SILC_SCALAR_TYPE:
      matched = (m == 1 && n == 1);
      l++;
      break;
    case SILC_RANGE_TYPE:
    case SILC_COLUMN_VECTOR_TYPE:
      matched = (m == 1 && n == element[0]->d.vector.length);
      l++;
      break;
    case SILC_ROW_VECTOR_TYPE:
      matched = (m == element[i]->d.vector.length && n == 1);
      l++;
      break;
    case SILC_MATRIX_TYPE:
      matched = (m == element[i]->d.matrix.m &&
		 n == element[i]->d.matrix.n);
      l++;
      break;
    case SILC_CUBIC_ARRAY_TYPE:
      matched = (m == element[i]->d.cubic_array.m &&
		 n == element[i]->d.cubic_array.n);
      l += element[i]->d.cubic_array.l;
      break;
    default:
      should_not_reach_here;
    }
    if (!matched) {
      error("size mismatch in cubic array construction");
      goto onError;
    }
  }

  object = (silc_object_t *)malloc(sizeof(silc_object_t));
  object->type = SILC_CUBIC_ARRAY_TYPE;
  object->format = 0;
  object->precision = element[0]->precision;
  object->d.cubic_array.l = l;
  object->d.cubic_array.m = m;
  object->d.cubic_array.n = n;
  msize = m * n * SILC_SCALAR_SIZE(object);
  object->d.cubic_array.v.i = (int *)malloc(l * msize);
  k = 0;
  for (i = 0; i < length; i++) {
    switch (element[i]->type) {
    case SILC_SCALAR_TYPE:
      switch (object->precision) {
      case SILC_INT:
	object->d.cubic_array.v.i[k] = element[i]->d.scalar.i; break;
      case SILC_LONG:
	object->d.cubic_array.v.l[k] = element[i]->d.scalar.l; break;
      case SILC_FLOAT:
	object->d.cubic_array.v.s[k] = element[i]->d.scalar.s; break;
      case SILC_DOUBLE:
	object->d.cubic_array.v.d[k] = element[i]->d.scalar.d; break;
      case SILC_COMPLEX:
	object->d.cubic_array.v.c[k*2]   = element[i]->d.scalar.c[0];
	object->d.cubic_array.v.c[k*2+1] = element[i]->d.scalar.c[1];
	break;
      case SILC_DOUBLE_COMPLEX:
	object->d.cubic_array.v.z[k*2]   = element[i]->d.scalar.z[0];
	object->d.cubic_array.v.z[k*2+1] = element[i]->d.scalar.z[1];
	break;
      default:
	should_not_reach_here;
      }
      k++;
      break;
    case SILC_RANGE_TYPE:
      for (j = 0; j < element[i]->d.vector.length; j++, k++)
	object->d.cubic_array.v.i[k] = element[i]->d.vector.v.range.start + j;
      break;
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      switch (object->precision) {
      case SILC_INT:
	memcpy(object->d.cubic_array.v.i + k, element[i]->d.vector.v.i, msize);
	break;
      case SILC_LONG:
	memcpy(object->d.cubic_array.v.l + k, element[i]->d.vector.v.l, msize);
	break;
      case SILC_FLOAT:
	memcpy(object->d.cubic_array.v.s + k, element[i]->d.vector.v.s, msize);
	break;
      case SILC_DOUBLE:
	memcpy(object->d.cubic_array.v.d + k, element[i]->d.vector.v.d, msize);
	break;
      case SILC_COMPLEX:
	memcpy(object->d.cubic_array.v.c + k, element[i]->d.vector.v.c, msize);
	break;
      case SILC_DOUBLE_COMPLEX:
	memcpy(object->d.cubic_array.v.z + k, element[i]->d.vector.v.z, msize);
	break;
      default:
	should_not_reach_here;
      }
      k += m * n;
      break;
#if 0 /* XXX */
    case SILC_MATRIX_TYPE:
      switch (object->precision) {
      case SILC_INT:
	memcpy(object->d.cubic_array.v.i + k, element[i]->d.matrix.v.i, msize);
	break;
      case SILC_LONG:
	memcpy(object->d.cubic_array.v.l + k, element[i]->d.matrix.v.l, msize);
	break;
      case SILC_FLOAT:
	memcpy(object->d.cubic_array.v.s + k, element[i]->d.matrix.v.s, msize);
	break;
      case SILC_DOUBLE:
	memcpy(object->d.cubic_array.v.d + k, element[i]->d.matrix.v.d, msize);
	break;
      case SILC_COMPLEX:
	memcpy(object->d.cubic_array.v.c + k * 2, element[i]->d.matrix.v.c,
	       msize);
	break;
      case SILC_DOUBLE_COMPLEX:
	memcpy(object->d.cubic_array.v.z + k * 2, element[i]->d.matrix.v.z,
	       msize);
	break;
      default:
	should_not_reach_here;
      }
      k += m * n;
      break;
#endif
    case SILC_CUBIC_ARRAY_TYPE:
      switch (object->precision) {
      case SILC_INT:
	memcpy(object->d.cubic_array.v.i + k,
	       element[i]->d.cubic_array.v.i,
	       element[i]->d.cubic_array.l * msize); break;
      case SILC_LONG:
	memcpy(object->d.cubic_array.v.l + k,
	       element[i]->d.cubic_array.v.l,
	       element[i]->d.cubic_array.l * msize); break;
      case SILC_FLOAT:
	memcpy(object->d.cubic_array.v.s + k,
	       element[i]->d.cubic_array.v.s,
	       element[i]->d.cubic_array.l * msize); break;
      case SILC_DOUBLE:
	memcpy(object->d.cubic_array.v.d + k,
	       element[i]->d.cubic_array.v.d,
	       element[i]->d.cubic_array.l * msize); break;
      case SILC_COMPLEX:
	memcpy(object->d.cubic_array.v.c + k * 2,
	       element[i]->d.cubic_array.v.c,
	       element[i]->d.cubic_array.l * msize); break;
      case SILC_DOUBLE_COMPLEX:
	memcpy(object->d.cubic_array.v.z + k * 2,
	       element[i]->d.cubic_array.v.z,
	       element[i]->d.cubic_array.l * msize); break;
      default:
	should_not_reach_here;
      }
      k += element[i]->d.cubic_array.l * m * n;
      break;
    }
  }
#ifdef DEBUG
  printf("new_cubic_array() => 0x%lx\n", object);
#endif
  for (i = 0; i < length; i++)
    silc_object_free(element[i]);
  free(element);
  return object;

 onError:
  for (i = 0; i < length; i++)
    silc_object_free(element[i]);
  free(element);
  return NULL;
}

static silc_object_t *
new_array(expr_t *v[], int length, int vector_type)
{
  value_t **element;
  silc_object_t *new_object, *object, *src;
  silc_modifier_t *modifiers;
  silc_format_t *table;
  int i, j, k, m, n, n_columns, n_rows, is_sparse, precision, format, status;
  clist_t *subscript[2];

  element = (value_t **)malloc(sizeof(value_t *) * length);
  n_rows = n_columns = 0;
  format = silc_find_format_id(SILC_FORMAT_DENSE);
  is_sparse = 1;
  for (i = status = 0; i < length && status == 0; i++) {
    element[i] = evaluate_expression(v[i]);
    if (!element[i]) {
      error("%d%s element: failed to evaluate", i+1, ordinal_suffix(i));
      status = -1;
      break;
    }
    object = var_get(element[i]->name);
    if (!object) {
      error("%d%s element: undefined variable: %s", i+1, ordinal_suffix(i),
	    element[i]->name);
      status = -1;
      break;
    }
    modifiers = element[i]->modifiers;
    switch (object->type) {
    case SILC_SCALAR_TYPE:
      m = n = 1;
      is_sparse = 0;
      break;
    case SILC_RANGE_TYPE:
    case SILC_COLUMN_VECTOR_TYPE:
      m = object->d.vector.length;
      n = 1;
      if (modifiers && modifiers->n > 0) {
        clist_set_defaults(modifiers->subscript[0], 0, m - 1);
        m = clist_length(modifiers->subscript[0]);
      }
      is_sparse = 0;
      break;
    case SILC_ROW_VECTOR_TYPE:
      m = 1;
      n = object->d.vector.length;
      if (modifiers && modifiers->n > 0) {
        clist_set_defaults(modifiers->subscript[0], 0, n - 1);
        n = clist_length(modifiers->subscript[0]);
      }
      is_sparse = 0;
      break;
    case SILC_MATRIX_TYPE:
      m = object->d.matrix.m;
      n = object->d.matrix.n;
      if (modifiers && modifiers->n > 0) {
        clist_set_defaults(modifiers->subscript[0], 0, m - 1);
        clist_set_defaults(modifiers->subscript[1], 0, n - 1);
        m = clist_length(modifiers->subscript[0]);
        n = clist_length(modifiers->subscript[1]);
      }
      if (object->format == format)
	is_sparse = 0;
      break;
    default:
      should_not_reach_here;
    }
    if (modifiers && modifiers->transpose) {
      int t = m; m = n; n = t;
    }
    switch (vector_type) {
    case SILC_COLUMN_VECTOR_TYPE: /* vertical concatenation */
      if (n_columns == 0)
	n_columns = n;
      if (n_columns != n) {
	error("size mismatch among elements to be concatenated");
	status = -1;
      }
      n_rows += m;
      break;
    case SILC_ROW_VECTOR_TYPE: /* horizontal concatenation */
      if (n_rows == 0)
	n_rows = m;
      if (n_rows != m) {
	error("size mismatch among elements to be concatenated");
	status = -1;
      }
      n_columns += n;
      break;
    default:
      should_not_reach_here;
    }
  }
  if (status < 0) {
    while (i > 0)
      value_free(element[--i]);
    free(element);
    return NULL;
  }
  
  /* check if all elements are of the same precision */
  precision = var_get(element[0]->name)->precision;
  for (i = 1; i < length; i++)
    if (var_get(element[i]->name)->precision != precision)
      break;
  if (i < length) {
    error("precision mismatch among elements to be concatenated");
    goto onError;
  }

  if (is_sparse)
    format = silc_find_format_id(SILC_FORMAT_CRS);
  table = silc_find_format_proc_table(format);
  if (!table) {
    error("silc_find_format_proc_table() failed (fatal)");
    goto onError;
  }
  if (!table->proc_assign) {
    error("proc_assign() not defined (fatal)");
    goto onError;
  }

  /* concatenate elements */
  if (n_rows == 1 && n_columns == 1) {
    /*
     * create new scalar object
     */
    object = value_get_modified_object(element[0]);
  } else if (n_rows == 1 || n_columns == 1) {
    /*
     * create new vector object
     */
    if (n_rows < n_columns)
      new_object = silc_object_new_row_vector(precision, n_columns);
    else
      new_object = silc_object_new_column_vector(precision, n_rows);
    k = 0;
    for (i = 0; i < length; i++) {
      object = src = var_get(element[i]->name);
      modifiers = element[i]->modifiers;
      switch (object->type) {
      case SILC_SCALAR_TYPE:
	subscript[0] = clist_new();
	clist_append_point(subscript[0], k);
	do_assign_vector(new_object, subscript, 1, object, modifiers);
	clist_free(subscript[0]);
	k++;
	break;
      case SILC_RANGE_TYPE:
	object = silc_object_new_column_vector_from_range(object);
      case SILC_COLUMN_VECTOR_TYPE:
      case SILC_ROW_VECTOR_TYPE:
	if (modifiers && modifiers->n > 0)
	  m = clist_length(modifiers->subscript[0]);
	else
	  m = object->d.vector.length;
	subscript[0] = clist_new();
	clist_append_range(subscript[0], k, k + m - 1);
	do_assign_vector(new_object, subscript, 1, object, modifiers);
	clist_free(subscript[0]);
	k += m;
	break;
      case SILC_MATRIX_TYPE:
	if (!modifiers || modifiers->n == 0)
	  should_not_reach_here;
	if (n_rows < n_columns &&  modifiers->transpose ||
	    n_rows > n_columns && !modifiers->transpose)
	  m = clist_length(modifiers->subscript[0]);
	else
	  m = clist_length(modifiers->subscript[1]);
	subscript[0] = clist_new();
	clist_append_range(subscript[0], k, k + m - 1);
	status = table->proc_assign(new_object, subscript, 1, object,
				    modifiers);
	clist_free(subscript[0]);
	if (status < 0) {
	  error("proc_assign() failed");
	  silc_object_free(new_object);
	  goto onError;
	}
	k += m;
	break;
      default:
	should_not_reach_here;
      }
      if (object != src)
        silc_object_free(object);
    }
  } else {
    /*
     * create new matrix object
     */
    if (!table->proc_new) {
      error("proc_new() not defined (fatal)");
      goto onError;
    }
    new_object = silc_object_new_matrix(precision, format, n_rows, n_columns);
    if (table->proc_new(new_object) < 0) {
      error("proc_new() failed (fatal)");
      silc_object_free(new_object);
      goto onError;
    }
    k = 0;
    for (i = 0; i < length; i++) {
      object = src = var_get(element[i]->name);
      subscript[0] = clist_new();
      subscript[1] = clist_new();
      modifiers = element[i]->modifiers;
      switch (vector_type) {
      case SILC_COLUMN_VECTOR_TYPE:
	switch (object->type) {
	case SILC_COLUMN_VECTOR_TYPE:
	case SILC_ROW_VECTOR_TYPE:
	  if (object->type == SILC_COLUMN_VECTOR_TYPE)
	    if (!modifiers || !modifiers->transpose)
	      should_not_reach_here;
          /* new_object[k,:] = object */
          clist_append_point(subscript[0], k);
          clist_append_range(subscript[1], 0, n_columns - 1);
	  k++;
	  break;
	case SILC_MATRIX_TYPE:
	  if (!modifiers) {
	    m = object->d.matrix.m;
	  } else if (!modifiers->transpose) {
	    if (modifiers->n > 0)
	      m = clist_length(modifiers->subscript[0]);
	    else
	      m = object->d.matrix.m;
	  } else { /* transpose */
	    if (modifiers->n > 0)
	      m = clist_length(modifiers->subscript[1]);
	    else
	      m = object->d.matrix.n;
	  }
          /* new_object[k:k+m-1,:] = object */
          clist_append_range(subscript[0], k, k + m - 1);
          clist_append_range(subscript[1], 0, n_columns - 1);
	  k += m;
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_ROW_VECTOR_TYPE:
	switch (object->type) {
	case SILC_RANGE_TYPE:
	  object = silc_object_new_column_vector_from_range(object);
	case SILC_COLUMN_VECTOR_TYPE:
	case SILC_ROW_VECTOR_TYPE:
	  if (object->type == SILC_ROW_VECTOR_TYPE)
	    if (!modifiers || !modifiers->transpose)
	      should_not_reach_here;
          /* new_object[:,k] = object */
          clist_append_range(subscript[0], 0, n_rows - 1);
          clist_append_point(subscript[1], k);
	  k++;
	  break;
	case SILC_MATRIX_TYPE:
	  if (!modifiers) {
	    n = object->d.matrix.n;
	  } else if (!modifiers->transpose) {
	    if (modifiers->n > 0)
	      n = clist_length(modifiers->subscript[1]);
	    else
	      n = object->d.matrix.n;
	  } else { /* transpose */
	    if (modifiers->n > 0)
	      n = clist_length(modifiers->subscript[0]);
	    else
	      n = object->d.matrix.m;
	  }
          /* new_object[:,k:k+n-1] = object */
          clist_append_range(subscript[0], 0, n_rows - 1);
          clist_append_range(subscript[1], k, k + n - 1);
	  k += n;
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      default:
	should_not_reach_here;
      }
      status = table->proc_assign(new_object, subscript, 2, object,
				  modifiers);
      clist_free(subscript[0]);
      clist_free(subscript[1]);
      if (object != src)
        silc_object_free(object);
      if (status < 0) {
        error("proc_assign() failed");
        silc_object_free(new_object);
        goto onError;
      }
    }
  }
#ifdef DEBUG
  printf("new_array() => 0x%lx (%s)\n", new_object, SILC_TYPE_NAME(new_object));
#endif
  for (i = 0; i < length; i++)
    value_free(element[i]);
  free(element);
  return new_object;

 onError:
  for (i = 0; i < length; i++)
    value_free(element[i]);
  free(element);
  return NULL;
}

silc_object_t *
new_range(expr_t *lhs, expr_t *rhs)
{
  value_t *value;
  silc_object_t *object;
  int start, end;

  /* evaluate left-hand side */
  value = evaluate_expression(lhs);
  if (!value)
    return NULL;
  object = value_get_modified_object(value);
  value_free(value);
  if (!object)
    return NULL;
  if (object->type == SILC_SCALAR_TYPE &&
      object->precision == SILC_INT) {
    start = object->d.scalar.i;
  } else if (object->type == SILC_SCALAR_TYPE &&
	     object->precision == SILC_LONG) {
    start = object->d.scalar.l;
  } else {
    silc_object_free(object);
    error("the start of range must be a scalar integer");
    return NULL;
  }
  silc_object_free(object);
  /* evaluate right-hand side */
  value = evaluate_expression(rhs);
  if (!value)
    return NULL;
  object = value_get_modified_object(value);
  value_free(value);
  if (!object)
    return NULL;
  if (object->type == SILC_SCALAR_TYPE &&
      object->precision == SILC_INT) {
    end = object->d.scalar.i;
  } else if (object->type == SILC_SCALAR_TYPE &&
	     object->precision == SILC_LONG) {
    end = object->d.scalar.l;
  } else {
    silc_object_free(object);
    error("the end of range must be a scalar integer");
    return NULL;
  }
  silc_object_free(object);
  /* create range object */
  object = (silc_object_t *)malloc(sizeof(silc_object_t));
  object->type = SILC_RANGE_TYPE;
  object->format = 0;
  object->precision = SILC_INT;
  object->d.vector.length = end - start + 1;
  object->d.vector.v.range.start = start;
  object->d.vector.v.range.end = end;
#ifdef DEBUG
  printf("new_range() => 0x%lx\n", object);
#endif
  return object;
}

/*
 *  unary operators
 */
silc_object_t *
do_unary_op(int op, value_t *operand)
{
  silc_object_t *object, *new_object;
  silc_modifier_t *object_mods;
  silc_mfunc_t *mfunc;
  silc_mfunc_param_t *param;
  int object_is_copy, status;

  object = var_get(operand->name);
  if (!object) {
    if (operand->is_temp)
      should_not_reach_here;
    else
      error("undefined variable: %s", operand->name);
    return NULL;
  }
  object_is_copy = 0;
  object_mods = operand->modifiers; /* XXX: take modifiers into account */

  /* find module function */
  mfunc = silc_find_unary_op_impl
    (op, object->precision, object->type, object->format);
  if (!mfunc) {
    error("FATAL: no appropriate module function for %s (%s)",
	  ((op == SILC_NEGATE_OP) ? "negate" :
	   (op == SILC_COMPLEX_CONJUGATE_OP) ? "complex conjugate" :
	   (op == SILC_CONJUGATE_TRANSPOSE_OP) ? "conjugate transpose" :
	   (op == SILC_TRANSPOSE_OP) ? "transpose" : "<unknown>"),
	  SILC_TYPE_NAME(object));
    return NULL;
  }

  /* apply unary operator */
  switch (mfunc->out) {
  case 0:
    if (!operand->is_temp) {
      object = silc_object_duplicate(object, 1);
      if (!object) {
	free(mfunc);
	return NULL;
      }
      object_is_copy = 1;
    }
    break;
  case 1:
    break;
  default:
    should_not_reach_here;
  }
#ifdef VANILLA_MPI
  if (silc_ipc_mpi_mfunc(mfunc->flags))
    silc_ipc_init(mfunc->module_name, mfunc->mfunc_name);
#endif
  param = (silc_mfunc_param_t *)malloc(sizeof(silc_mfunc_param_t) * 2);
  param[0].o = object;
  param[0].m = NULL;
  status = call_mfunc_impl(mfunc, 1, param);
  if (status < 0) {
    free(param);
    free(mfunc);
    if (object_is_copy)
      silc_object_free(object);
    return NULL;
  }
  switch (mfunc->out) {
  case 0:
    new_object = object;
    if (operand->is_temp)
      var_get_free(operand->name);
    break;
  case 1:
    new_object = param[1].o;
    break;
  default:
    should_not_reach_here;
  }
  free(param);
  free(mfunc);
  return new_object;
}

/*
 *  precision conversion
 */
int
convert_scalar_precision(silc_object_t *object, int new_precision)
{
  switch (object->precision) {
  case SILC_INT:
    switch (new_precision) {
    case SILC_LONG:
      object->d.scalar.l = (int64_t)object->d.scalar.i; break;
    case SILC_FLOAT:
      object->d.scalar.s = (float)object->d.scalar.i; break;
    case SILC_DOUBLE:
      object->d.scalar.d = (double)object->d.scalar.i; break;
    case SILC_COMPLEX:
      object->d.scalar.c[0] = (float)object->d.scalar.i;
      object->d.scalar.c[1] = 0;
      break;
    case SILC_DOUBLE_COMPLEX:
      object->d.scalar.z[0] = (double)object->d.scalar.i;
      object->d.scalar.z[1] = 0;
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (new_precision) {
    case SILC_DOUBLE:
      object->d.scalar.d = (double)object->d.scalar.l; break;
    case SILC_DOUBLE_COMPLEX:
      object->d.scalar.z[0] = (double)object->d.scalar.l;
      object->d.scalar.z[1] = 0;
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (new_precision) {
    case SILC_DOUBLE:
      object->d.scalar.d = (double)object->d.scalar.s; break;
    case SILC_COMPLEX:
      object->d.scalar.c[0] = object->d.scalar.s;
      object->d.scalar.c[1] = 0;
      break;
    case SILC_DOUBLE_COMPLEX:
      object->d.scalar.z[0] = (double)object->d.scalar.s;
      object->d.scalar.z[1] = 0;
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      object->d.scalar.z[0] = object->d.scalar.d;
      object->d.scalar.z[1] = 0;
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      object->d.scalar.z[1] = (double)object->d.scalar.c[1];
      object->d.scalar.z[0] = (double)object->d.scalar.c[0];
      /* NOTE: keep this order since z[] and c[] are part of union */
      break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}
int
convert_range_precision(silc_object_t *object, int new_precision)
{
  int i;
  int64_t *lv;
  float *sv, *cv;
  double *dv, *zv;
  
#define _crp11(v2,buf,t)                                             \
        buf = (t *)malloc(object->d.vector.length * sizeof(t));      \
        if (!buf) return -1;					     \
	for (i = 0; i < object->d.vector.length; i++)                \
	  buf[i] = (t)(object->d.vector.v.range.start + i);          \
	v2 = buf;
#define _crp12(v2,buf,t)                                             \
        buf = (t *)malloc(object->d.vector.length * sizeof(t) * 2);  \
        if (!buf) return -1;					     \
	for (i = 0; i < object->d.vector.length; i++) {              \
	  buf[i*2]   = (t)(object->d.vector.v.range.start + i);      \
          buf[i*2+1] = (t)0;                                         \
	}                                                            \
	v2 = buf;
  switch (new_precision) {
  case SILC_LONG:
    _crp11(object->d.vector.v.l, lv, int64_t); break;
  case SILC_FLOAT:
    _crp11(object->d.vector.v.s, sv, float); break;
  case SILC_DOUBLE:
    _crp11(object->d.vector.v.d, dv, double); break;
  case SILC_COMPLEX:
    _crp12(object->d.vector.v.c, cv, float); break;
  case SILC_DOUBLE_COMPLEX:
    _crp12(object->d.vector.v.z, zv, double); break;
  default:
    should_not_reach_here;
  }
  object->type = SILC_COLUMN_VECTOR_TYPE;
  return 0;
}
int
convert_vector_precision(silc_object_t *object, int new_precision)
{
  int i;
  int64_t *lv;
  float *sv, *cv;
  double *dv, *zv;
  
#define _cvp11(v1,v2,buf,t)                                          \
        buf = (t *)malloc(object->d.vector.length * sizeof(t));      \
        if (!buf) return -1;					     \
	for (i = 0; i < object->d.vector.length; i++)                \
	  buf[i] = (t)v1[i];                                         \
	free(v1);                                                    \
	v2 = buf;
#define _cvp12(v1,v2,buf,t)                                          \
        buf = (t *)malloc(object->d.vector.length * sizeof(t) * 2);  \
        if (!buf) return -1;					     \
	for (i = 0; i < object->d.vector.length; i++) {              \
	  buf[i*2]   = (t)v1[i];                                     \
          buf[i*2+1] = (t)0;                                         \
	}                                                            \
	free(v1);                                                    \
	v2 = buf;
#define _cvp22(v1,v2,buf,t)                                          \
        buf = (t *)malloc(object->d.vector.length * sizeof(t) * 2);  \
        if (!buf) return -1;					     \
	for (i = 0; i < object->d.vector.length; i++) {              \
	  buf[i*2]   = (t)v1[i*2];                                   \
          buf[i*2+1] = (t)v1[i*2+1];                                 \
	}                                                            \
	free(v1);                                                    \
	v2 = buf;
  switch (object->precision) {
  case SILC_INT:
    switch (new_precision) {
    case SILC_LONG:
      _cvp11(object->d.vector.v.i, object->d.vector.v.l, lv, int64_t); break;
    case SILC_FLOAT:
      _cvp11(object->d.vector.v.i, object->d.vector.v.s, sv, float); break;
    case SILC_DOUBLE:
      _cvp11(object->d.vector.v.i, object->d.vector.v.d, dv, double); break;
    case SILC_COMPLEX:
      _cvp12(object->d.vector.v.i, object->d.vector.v.c, cv, float); break;
    case SILC_DOUBLE_COMPLEX:
      _cvp12(object->d.vector.v.i, object->d.vector.v.z, zv, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (new_precision) {
    case SILC_DOUBLE:
      _cvp11(object->d.vector.v.l, object->d.vector.v.d, dv, double); break;
    case SILC_DOUBLE_COMPLEX:
      _cvp12(object->d.vector.v.l, object->d.vector.v.z, zv, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (new_precision) {
    case SILC_DOUBLE:
      _cvp11(object->d.vector.v.s, object->d.vector.v.d, dv, double); break;
    case SILC_COMPLEX:
      _cvp12(object->d.vector.v.s, object->d.vector.v.c, cv, float); break;
    case SILC_DOUBLE_COMPLEX:
      _cvp12(object->d.vector.v.s, object->d.vector.v.z, zv, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _cvp12(object->d.vector.v.d, object->d.vector.v.z, zv, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _cvp22(object->d.vector.v.c, object->d.vector.v.z, zv, double); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}
int
convert_matrix_precision(silc_object_t *object, int new_precision)
{
  silc_format_t *table;

  table = silc_find_format_proc_table(object->format);
  if (!table)
    return -1;
  if (!table->proc_pconv) {
    error("proc_pconv() not defined (fatal)");
    return -1;
  }
  if (table->proc_pconv(object, new_precision) < 0) {
    error("proc_pconv() failed (fatal)");
    return -1;
  }
  return 0;
}
int
convert_cubic_array_precision(silc_object_t *object, int new_precision)
{
  int i, n_elements;
  int64_t *lv;
  float *sv, *cv;
  double *dv, *zv;
  
#define _cqp11(v1,v2,buf,t)                             \
        buf = (t *)malloc(n_elements * sizeof(t));      \
        if (!buf) return -1;				\
	for (i = 0; i < n_elements; i++)                \
	  buf[i] = (t)v1[i];                            \
	free(v1);                                       \
	v2 = buf;
#define _cqp12(v1,v2,buf,t)                             \
        buf = (t *)malloc(n_elements * sizeof(t) * 2);  \
        if (!buf) return -1;				\
	for (i = 0; i < n_elements; i++) {              \
	  buf[i*2]   = (t)v1[i];                        \
          buf[i*2+1] = (t)0;                            \
	}                                               \
	free(v1);                                       \
	v2 = buf;
#define _cqp22(v1,v2,buf,t)                             \
        buf = (t *)malloc(n_elements * sizeof(t) * 2);  \
        if (!buf) return -1;				\
	for (i = 0; i < n_elements; i++) {              \
	  buf[i*2]   = (t)v1[i*2];                      \
          buf[i*2+1] = (t)v1[i*2+1];                    \
	}                                               \
	free(v1);                                       \
	v2 = buf;
  n_elements = object->d.cubic_array.l *
	       object->d.cubic_array.m *
	       object->d.cubic_array.n;
  switch (object->precision) {
  case SILC_INT:
    switch (new_precision) {
    case SILC_LONG:
      _cqp11(object->d.cubic_array.v.i, object->d.cubic_array.v.l, lv, int64_t);
      break;
    case SILC_FLOAT:
      _cqp11(object->d.cubic_array.v.i, object->d.cubic_array.v.s, sv, float);
      break;
    case SILC_DOUBLE:
      _cqp11(object->d.cubic_array.v.i, object->d.cubic_array.v.d, dv, double);
      break;
    case SILC_COMPLEX:
      _cqp12(object->d.cubic_array.v.i, object->d.cubic_array.v.c, cv, float);
      break;
    case SILC_DOUBLE_COMPLEX:
      _cqp12(object->d.cubic_array.v.i, object->d.cubic_array.v.z, zv, double);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (new_precision) {
    case SILC_DOUBLE:
      _cqp11(object->d.cubic_array.v.l, object->d.cubic_array.v.d, dv, double);
      break;
    case SILC_DOUBLE_COMPLEX:
      _cqp12(object->d.cubic_array.v.l, object->d.cubic_array.v.z, zv, double);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (new_precision) {
    case SILC_DOUBLE:
      _cqp11(object->d.cubic_array.v.s, object->d.cubic_array.v.d, dv, double);
      break;
    case SILC_COMPLEX:
      _cqp12(object->d.cubic_array.v.s, object->d.cubic_array.v.c, cv, float);
      break;
    case SILC_DOUBLE_COMPLEX:
      _cqp12(object->d.cubic_array.v.s, object->d.cubic_array.v.z, zv, double);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _cqp12(object->d.cubic_array.v.d, object->d.cubic_array.v.z, zv, double);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _cqp22(object->d.cubic_array.v.c, object->d.cubic_array.v.z, zv, double);
      break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}
int
convert_precision(silc_object_t *object, int new_precision)
{
  int status;
#ifdef DEBUG
  char *p = SILC_SCALAR_NAME(object);
#endif

  switch (object->type) {
  case SILC_SCALAR_TYPE:
    if (convert_scalar_precision(object, new_precision) < 0)
      return -1;
    break;
  case SILC_RANGE_TYPE:
    if (convert_range_precision(object, new_precision) < 0)
      return -1;
    break;
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    if (convert_vector_precision(object, new_precision) < 0)
      return -1;
    break;
  case SILC_MATRIX_TYPE:
    if (convert_matrix_precision(object, new_precision) < 0)
      return -1;
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    if (convert_cubic_array_precision(object, new_precision) < 0)
      return -1;
    break;
  default:
    should_not_reach_here;
  }
  object->precision = new_precision;
#ifdef DEBUG
  printf("converted %s of %s (0x%lx) to %s\n",
	 SILC_TYPE_NAME(object), p, object, SILC_SCALAR_NAME(object));
  silc_object_pprint(object, 1);
#endif
  return 0;
}

/*
 *  binary operators
 */
static silc_mfunc_t *
binary_op_vector_vector(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
    if (x->type != y->type) {
      error("vector type mismatch");
      return NULL;
    }
    if (x->d.vector.length != y->d.vector.length) {
      error("vector size mismatch");
      return NULL;
    }
    break;
  case SILC_MULTIPLY_OP:
    /* XXX: take the transposition modifier into consideration */
    if (x->type == y->type) {
      error("vectors are of the same type");
      return NULL;
    }
    if (x->d.vector.length != y->d.vector.length) {
      error("vector size mismatch");
      return NULL;
    }
    break;
  case SILC_DIVIDE_OP:
  case SILC_REMAINDER_OP:
  case SILC_LEQ_OP:
    error("illegal operator on vectors");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, 0);
}
static silc_mfunc_t *
binary_op_vector_matrix(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
  case SILC_DIVIDE_OP:
    /* XXX: take the transposition modifier into consideration */
    if (x->type != SILC_ROW_VECTOR_TYPE) {
      error("not a row vector");
      return NULL;
    }
    if (x->d.vector.length != y->d.matrix.m) {
      error("size mismatch on vector and matrix");
      return NULL;
    }
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_REMAINDER_OP:
  case SILC_LEQ_OP:
    error("illegal operator on vector and matrix");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  SILC_ROW_VECTOR_TYPE, 0,
				  y->type, y->format);
}
static silc_mfunc_t *
binary_op_vector_scalar(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_REMAINDER_OP:
  case SILC_LEQ_OP:
    error("illegal operator on vector and scalar");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, 0);
}
static silc_mfunc_t *
binary_op_vector_carray(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  error("illegal operator on vector and cubic array");
  return NULL;
}
static silc_mfunc_t *
binary_op_matrix_vector(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    switch (y->type) {
    case SILC_ROW_VECTOR_TYPE:
      error("cannot multiply matrix by row vector");
      return NULL;
    case SILC_COLUMN_VECTOR_TYPE:
      if (x->d.matrix.n != y->d.vector.length) {
	error("cannot multiply matrix (#rows: %d) by vector (length: %d)",
	      x->d.matrix.n, y->d.vector.length);
	return NULL;
      }
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LEQ_OP:
    switch (y->type) {
    case SILC_ROW_VECTOR_TYPE:
      error("cannot solve LEQ: right-hand side is row vector"); /* XXX: OK? */
      return NULL;
    case SILC_COLUMN_VECTOR_TYPE:
      if (x->d.matrix.n != y->d.vector.length) {
	error("cannot solve LEQ: matrix-vector size mismatch");
	return NULL;
      }
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_REMAINDER_OP:
    error("illegal operator on matrix and vector");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, x->format,
				  SILC_COLUMN_VECTOR_TYPE, 0);
}
static silc_mfunc_t *
binary_op_matrix_matrix(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
    if (x->d.matrix.m != y->d.matrix.m ||
	x->d.matrix.n != y->d.matrix.n) {
      error("size mismatch on matrices (%dx%d and %dx%d)",
	    x->d.matrix.m, x->d.matrix.n,
	    y->d.matrix.m, y->d.matrix.n);
      return NULL;
    }
    break;
  case SILC_MULTIPLY_OP:
    if (x->d.matrix.n != y->d.matrix.m) {
      error("cannot multiply matrix (%d columns) by matrix (%d rows)",
	    x->d.matrix.n, y->d.matrix.m);
      return NULL;
    }
    break;
  case SILC_LEQ_OP:
  case SILC_DIVIDE_OP:
    if (x->d.matrix.n != y->d.matrix.m) {
      error("cannot solve LEQ: matrix-matrix size mismatch");
      return NULL;
    }
    break;
  case SILC_REMAINDER_OP:
    error("illegal operator on matrices");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, x->format,
				  y->type, y->format);
}
static silc_mfunc_t *
binary_op_matrix_scalar(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_REMAINDER_OP:
  case SILC_LEQ_OP:
    error("illegal operator on matrix and scalar");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, x->format,
				  y->type, 0);
}
static silc_mfunc_t *
binary_op_matrix_carray(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  error("illegal operator on matrix and cubic array");
  return NULL;
}
static silc_mfunc_t *
binary_op_scalar_vector(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_DIVIDE_OP:
  case SILC_REMAINDER_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_LEQ_OP:
    error("illegal operator on scalar and vector");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, 0);
}
static silc_mfunc_t *
binary_op_scalar_matrix(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_REMAINDER_OP:
  case SILC_LEQ_OP:
    error("illegal operator on scalar and matrix");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, y->format);
}
static silc_mfunc_t *
binary_op_scalar_scalar(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  int zero;

  switch(op) {
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_MULTIPLY_OP:
    break;
  case SILC_DIVIDE_OP:
    switch (y->precision) {
    case SILC_INT:
      zero = (y->d.scalar.i == 0); break;
    case SILC_LONG:
      zero = (y->d.scalar.l == 0); break;
    case SILC_FLOAT:
      zero = (y->d.scalar.s == 0.0); break;
    case SILC_DOUBLE:
      zero = (y->d.scalar.d == 0.0); break;
    case SILC_COMPLEX:
      zero = (y->d.scalar.c[0] == 0.0 && y->d.scalar.c[1] == 0.0); break;
    case SILC_DOUBLE_COMPLEX:
      zero = (y->d.scalar.z[0] == 0.0 && y->d.scalar.z[1] == 0.0); break;
    default:
      should_not_reach_here;
    }
    if (zero) {
      error("division by zero");
      return NULL;
    }
    break;
  case SILC_REMAINDER_OP:
    switch (y->precision) {
    case SILC_INT:
      zero = (y->d.scalar.i == 0); break;
    case SILC_LONG:
      zero = (y->d.scalar.l == 0); break;
    case SILC_FLOAT:
    case SILC_DOUBLE:
    case SILC_COMPLEX:
    case SILC_DOUBLE_COMPLEX:
      error("operands must be integers: \"%\" operator");
      return NULL;
    default:
      should_not_reach_here;
    }
    if (zero) {
      error("division by zero");
      return NULL;
    }
    break;
  case SILC_LEQ_OP:
    switch (x->precision) {
    case SILC_INT:
      zero = (x->d.scalar.i == 0); break;
    case SILC_LONG:
      zero = (x->d.scalar.l == 0); break;
    case SILC_FLOAT:
      zero = (x->d.scalar.s == 0.0); break;
    case SILC_DOUBLE:
      zero = (x->d.scalar.d == 0.0); break;
    case SILC_COMPLEX:
      zero = (x->d.scalar.c[0] == 0.0 && x->d.scalar.c[1] == 0.0); break;
    case SILC_DOUBLE_COMPLEX:
      zero = (x->d.scalar.z[0] == 0.0 && x->d.scalar.z[1] == 0.0); break;
    default:
      should_not_reach_here;
    }
    if (zero) {
      error("division by zero");
      return NULL;
    }
    break;
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
    error("illegal operator on scalars");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  SILC_SCALAR_TYPE, 0,
				  SILC_SCALAR_TYPE, 0);
}
static silc_mfunc_t *
binary_op_scalar_carray(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_DIVIDE_OP:
  case SILC_REMAINDER_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_LEQ_OP:
    error("illegal operator on scalar and cubic array");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, 0);
}
static silc_mfunc_t *
binary_op_carray_vector(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  error("illegal operator on cubic array and vector");
  return NULL;
}
static silc_mfunc_t *
binary_op_carray_matrix(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  error("illegal operator on cubic array and matrix");
  return NULL;
}
static silc_mfunc_t *
binary_op_carray_scalar(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_MULTIPLY_OP:
    break;
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
  case SILC_DIVIDE_OP:
  case SILC_REMAINDER_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_LEQ_OP:
    error("illegal operator on cubic array and scalar");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, 0);
}
static silc_mfunc_t *
binary_op_carray_carray(int op,
			silc_object_t *x, silc_modifier_t *x_mods,
			silc_object_t *y, silc_modifier_t *y_mods)
{
  switch(op) {
  case SILC_ADD_OP:
  case SILC_SUBTRACT_OP:
    if (x->d.cubic_array.l != y->d.cubic_array.l ||
	x->d.cubic_array.m != y->d.cubic_array.m ||
	x->d.cubic_array.n != y->d.cubic_array.n) {
      error("size mismatch on cubic arrays (%dx%dx%d and %dx%dx%d)",
	    x->d.cubic_array.l, x->d.cubic_array.m, x->d.cubic_array.n,
	    y->d.cubic_array.l, y->d.cubic_array.m, y->d.cubic_array.n);
    }
    break;
  case SILC_MULTIPLY_OP:
  case SILC_MULTIPLY_ELEMENTWISE_OP:
  case SILC_DIVIDE_OP:
  case SILC_DIVIDE_ELEMENTWISE_OP:
  case SILC_REMAINDER_OP:
  case SILC_LEQ_OP:
    error("illegal operator on cubic arrays");
    return NULL;
  default:
    should_not_reach_here;
  }
  return silc_find_binary_op_impl(op, x->precision,
				  x->type, 0,
				  y->type, 0);
}

silc_object_t *
do_binary_op(int op, value_t *lhs, value_t *rhs)
{
  silc_object_t *x, *y, *z, *p;
  silc_modifier_t *x_mods, *y_mods;
  silc_mfunc_t *mfunc;
  silc_mfunc_param_t *param;
  silc_format_t *x_table, *y_table;
  int x_is_copy, y_is_copy, dense_format, crs_format, status;
#ifdef TIME_FIND_MFUNC
  double t, elapsed;
#endif

  x = var_get(lhs->name);
  if (!x) {
    if (lhs->is_temp)
      should_not_reach_here;
    else
      error("undefined variable: %s", lhs->name);
    return NULL;
  }
  y = var_get(rhs->name);
  if (!y) {
    if (rhs->is_temp)
      should_not_reach_here;
    else
      error("undefined variable: %s", rhs->name);
    return NULL;
  }
  x_is_copy = 0;
  y_is_copy = 0;
  x_mods = lhs->modifiers;
  y_mods = rhs->modifiers;

  /* make sure that the two operands have the same precision */
#define convert_precision_x(precision) \
	if (!lhs->is_temp && !x_is_copy) {			\
	  if ((x = silc_object_duplicate(x, 1)) == NULL) {	\
	    if (y_is_copy)					\
	      silc_object_free(y);				\
	    return NULL;					\
	  }							\
	  x_is_copy = 1;					\
	}							\
	if (convert_precision(x, precision) < 0) {		\
	  if (x_is_copy) silc_object_free(x);			\
	  if (y_is_copy) silc_object_free(y);			\
	  return NULL;						\
	}
#define convert_precision_y(precision) \
	if (!rhs->is_temp && !y_is_copy) {			\
	  if ((y = silc_object_duplicate(y, 1)) == NULL) {	\
	    if (x_is_copy)					\
	      silc_object_free(x);				\
	    return NULL;					\
	  }							\
	  y_is_copy = 1;					\
	}							\
	if (convert_precision(y, precision) < 0) {		\
	  if (x_is_copy) silc_object_free(x);			\
	  if (y_is_copy) silc_object_free(y);			\
	  return NULL;						\
	}
  if (x->precision != y->precision) {
    switch (x->precision) {
    case SILC_INT:
      switch (y->precision) {
      case SILC_FLOAT:
      case SILC_COMPLEX:
      case SILC_LONG:
      case SILC_DOUBLE:
      case SILC_DOUBLE_COMPLEX:
	convert_precision_x(y->precision); break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_LONG:
      switch (y->precision) {
      case SILC_INT:
	convert_precision_y(x->precision); break;
      case SILC_DOUBLE:
      case SILC_DOUBLE_COMPLEX:
	convert_precision_x(y->precision); break;
      case SILC_FLOAT:          
	convert_precision_x(SILC_DOUBLE);
	convert_precision_y(SILC_DOUBLE);
	break;
      case SILC_COMPLEX:        
	convert_precision_x(SILC_DOUBLE_COMPLEX);
	convert_precision_y(SILC_DOUBLE_COMPLEX);
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_FLOAT:
      switch (y->precision) {
      case SILC_INT:
	convert_precision_y(x->precision); break;
      case SILC_COMPLEX:
      case SILC_DOUBLE:
      case SILC_DOUBLE_COMPLEX:
	convert_precision_x(y->precision); break;
      case SILC_LONG:
	convert_precision_x(SILC_DOUBLE);
	convert_precision_y(SILC_DOUBLE);
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_DOUBLE:
      switch (y->precision) {
      case SILC_INT:
      case SILC_FLOAT:
      case SILC_LONG:
	convert_precision_y(x->precision); break;
      case SILC_DOUBLE_COMPLEX:
	convert_precision_x(y->precision); break;
      case SILC_COMPLEX:        
	convert_precision_x(SILC_DOUBLE_COMPLEX);
	convert_precision_y(SILC_DOUBLE_COMPLEX);
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_COMPLEX:
      switch (y->precision) {
      case SILC_INT:
      case SILC_FLOAT:
	convert_precision_y(x->precision); break;
      case SILC_DOUBLE_COMPLEX:
	convert_precision_x(y->precision); break;
      case SILC_LONG:
      case SILC_DOUBLE:
	convert_precision_x(SILC_DOUBLE_COMPLEX);
	convert_precision_y(SILC_DOUBLE_COMPLEX);
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      switch (y->precision) {
      case SILC_INT:
      case SILC_FLOAT:
      case SILC_COMPLEX:
      case SILC_LONG:
      case SILC_DOUBLE:
	convert_precision_y(x->precision); break;
      default:
	should_not_reach_here;
      }
      break;
    default:
      should_not_reach_here;
    }
  }

  if (x->type == SILC_RANGE_TYPE) {
    if (x_is_copy)
      should_not_reach_here;
    x = silc_object_new_column_vector_from_range(x);
    x_is_copy = 1;
  }
  if (y->type == SILC_RANGE_TYPE) {
    if (y_is_copy)
      should_not_reach_here;
    y = silc_object_new_column_vector_from_range(y);
    y_is_copy = 1;
  }

#if 1 /* XXX strip all modifiers */
  if (x_mods) {
    p = modifier_apply_all(x, x_mods);
    if (x_is_copy)
      silc_object_free(x);
    x = p;
    x_is_copy = 1;
    x_mods = NULL;
  }
  if (y_mods) {
    p = modifier_apply_all(y, y_mods);
    if (y_is_copy)
      silc_object_free(y);
    y = p;
    y_is_copy = 1;
    y_mods = NULL;
  }
#endif

  /* find module function */
  status = 1;
#ifdef TIME_FIND_MFUNC
  t = silc_wtime();
#endif
  switch (x->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      mfunc = binary_op_vector_vector(op, x, x_mods, y, y_mods); break;
    case SILC_MATRIX_TYPE:
      mfunc = binary_op_vector_matrix(op, x, x_mods, y, y_mods); break;
    case SILC_SCALAR_TYPE:
      mfunc = binary_op_vector_scalar(op, x, x_mods, y, y_mods); break;
    case SILC_CUBIC_ARRAY_TYPE:
      mfunc = binary_op_vector_carray(op, x, x_mods, y, y_mods); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_MATRIX_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      mfunc = binary_op_matrix_vector(op, x, x_mods, y, y_mods); break;
    case SILC_MATRIX_TYPE:
      mfunc = binary_op_matrix_matrix(op, x, x_mods, y, y_mods);
      if (!mfunc && x->format != y->format) {
	x_table = silc_find_format_proc_table(x->format);
	if (!x_table) {
	  should_not_reach_here;
	  return NULL;
	}
	y_table = silc_find_format_proc_table(y->format);
	if (!y_table) {
	  should_not_reach_here;
	  return NULL;
	}
	dense_format = silc_find_format_id(SILC_FORMAT_DENSE);
	crs_format   = silc_find_format_id(SILC_FORMAT_CRS);
	if (x->format == dense_format ||
	    y->format == dense_format ||
	    x->format != crs_format && !x_table->proc_to_crs ||
	    y->format != crs_format && !y_table->proc_to_crs) {
	  if (x->format != dense_format) {
	    if (x_table->proc_to_dense == NULL) {
	      error("x: proc_to_dense() not found");
	      if (x_is_copy) silc_object_free(x);
	      if (y_is_copy) silc_object_free(y);
	      return NULL;
	    }
	    p = silc_object_new_matrix(x->precision, dense_format,
				       x->d.matrix.m,
				       x->d.matrix.n);
	    if (x_table->proc_to_dense(x, p) < 0) {
	      error("x: proc_to_dense() failed");
	      if (x_is_copy) silc_object_free(x);
	      if (y_is_copy) silc_object_free(y);
	      silc_object_free(p);
	      return NULL;
	    }
	    if (x_is_copy)
	      silc_object_free(x);
	    x = p;
	    x_is_copy = 1;
	  }
	  if (y->format != dense_format) {
	    if (y_table->proc_to_dense == NULL) {
	      error("y: proc_to_dense() not found");
	      if (x_is_copy) silc_object_free(x);
	      if (y_is_copy) silc_object_free(y);
	      return NULL;
	    }
	    p = silc_object_new_matrix(y->precision, dense_format,
				       y->d.matrix.m,
				       y->d.matrix.n);
	    if (y_table->proc_to_dense(y, p) < 0) {
	      error("y: proc_to_dense() failed");
	      if (x_is_copy) silc_object_free(x);
	      if (y_is_copy) silc_object_free(y);
	      silc_object_free(p);
	      return NULL;
	    }
	    if (y_is_copy)
	      silc_object_free(y);
	    y = p;
	    y_is_copy = 1;
	  }
	} else {
	  if (x->format != crs_format) {
	    /* convert x to CRS */
	    p = silc_object_new_matrix(x->precision, crs_format,
				       x->d.matrix.m,
				       x->d.matrix.n);
	    if (x_table->proc_to_crs(x, p) < 0) {
	      error("x: proc_to_crs() failed");
	      if (x_is_copy) silc_object_free(x);
	      if (y_is_copy) silc_object_free(y);
	      silc_object_free(p);
	      return NULL;
	    }
	    if (x_is_copy)
	      silc_object_free(x);
	    x = p;
	    x_is_copy = 1;
	  }
	  if (y->format != crs_format) {
	    /* convert y to CRS */
	    p = silc_object_new_matrix(y->precision, crs_format,
				       y->d.matrix.m,
				       y->d.matrix.n);
	    if (y_table->proc_to_crs(y, p) < 0) {
	      error("y: proc_to_crs() failed");
	      if (x_is_copy) silc_object_free(x);
	      if (y_is_copy) silc_object_free(y);
	      silc_object_free(p);
	      return NULL;
	    }
	    if (y_is_copy)
	      silc_object_free(y);
	    y = p;
	    y_is_copy = 1;
	  }
	}
	mfunc = binary_op_matrix_matrix(op, x, x_mods, y, y_mods);
      }
      break;
    case SILC_SCALAR_TYPE:
      mfunc = binary_op_matrix_scalar(op, x, x_mods, y, y_mods); break;
    case SILC_CUBIC_ARRAY_TYPE:
      mfunc = binary_op_matrix_carray(op, x, x_mods, y, y_mods); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_SCALAR_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      mfunc = binary_op_scalar_vector(op, x, x_mods, y, y_mods); break;
    case SILC_MATRIX_TYPE:
      mfunc = binary_op_scalar_matrix(op, x, x_mods, y, y_mods); break;
    case SILC_SCALAR_TYPE:
      mfunc = binary_op_scalar_scalar(op, x, x_mods, y, y_mods); break;
    case SILC_CUBIC_ARRAY_TYPE:
      mfunc = binary_op_scalar_carray(op, x, x_mods, y, y_mods); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    switch (y->type) {
    case SILC_COLUMN_VECTOR_TYPE:
    case SILC_ROW_VECTOR_TYPE:
      mfunc = binary_op_carray_vector(op, x, x_mods, y, y_mods); break;
    case SILC_MATRIX_TYPE:
      mfunc = binary_op_carray_matrix(op, x, x_mods, y, y_mods); break;
    case SILC_SCALAR_TYPE:
      mfunc = binary_op_carray_scalar(op, x, x_mods, y, y_mods); break;
    case SILC_CUBIC_ARRAY_TYPE:
      mfunc = binary_op_carray_carray(op, x, x_mods, y, y_mods); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  if (!mfunc) {
    error("FATAL: no appropriate module function for %s %s %s",
	  SILC_TYPE_NAME(x),
	  ((op == SILC_ADD_OP) ? "+" :
	   (op == SILC_SUBTRACT_OP) ? "-" :
	   (op == SILC_MULTIPLY_OP) ? "*" :
	   (op == SILC_MULTIPLY_ELEMENTWISE_OP) ? "*@" :
	   (op == SILC_DIVIDE_OP) ? "/" :
	   (op == SILC_DIVIDE_ELEMENTWISE_OP) ? "/@" :
	   (op == SILC_REMAINDER_OP) ? "%" :
	   (op == SILC_LEQ_OP) ? "\\" : "<unknown>"),
	  SILC_TYPE_NAME(y));
    if (x_is_copy) silc_object_free(x);
    if (y_is_copy) silc_object_free(y);
    return NULL;
  }
#ifdef TIME_FIND_MFUNC
  elapsed = silc_wtime() - t;
  _find_mfunc_time += elapsed;
#endif

  /* apply binary operator */
  switch (mfunc->out) {
  case 0:
    if (!lhs->is_temp && !x_is_copy) {
      if ((x = silc_object_duplicate(x, 1)) == NULL) {
	if (y_is_copy)
	  silc_object_free(y);
	free(mfunc);
	return NULL;
      }
      x_is_copy = 1;
    }
    break;
  case 1:
    if (!rhs->is_temp && !y_is_copy) {
      if ((y = silc_object_duplicate(y, 1)) == NULL) {
	if (x_is_copy)
	  silc_object_free(x);
	free(mfunc);
	return NULL;
      }
      y_is_copy = 1;
    }
    break;
  case 2:
    break;
  default:
    should_not_reach_here;
  }
#ifdef VANILLA_MPI
  if (silc_ipc_mpi_mfunc(mfunc->flags))
    silc_ipc_init(mfunc->module_name, mfunc->mfunc_name);
#endif
  param = (silc_mfunc_param_t *)malloc(sizeof(silc_mfunc_param_t) * 3);
  param[0].o = x; param[0].m = x_mods;
  param[1].o = y; param[1].m = y_mods;
  status = call_mfunc_impl(mfunc, 2, param);
  if (status < 0) {
    free(param);
    free(mfunc);
    if (x_is_copy) silc_object_free(x);
    if (y_is_copy) silc_object_free(y);
    return NULL;
  }
  switch (mfunc->out) {
  case 0:
    z = x;
    if (lhs->is_temp && !x_is_copy)
      var_get_free(lhs->name);
    if (y_is_copy) silc_object_free(y);
    break;
  case 1:
    z = y;
    if (rhs->is_temp && !y_is_copy)
      var_get_free(rhs->name);
    if (x_is_copy) silc_object_free(x);
    break;
  case 2:
    z = param[2].o;
    if (x_is_copy) silc_object_free(x);
    if (y_is_copy) silc_object_free(y);
    break;
  default:
    should_not_reach_here;
  }
  free(param);
  free(mfunc);
  return z;
}

static silc_object_t *
do_func_call(char *name, int n_params, value_t *value[])
{
  signature_t *sig;
  silc_object_t *object;
  silc_mfunc_t *mfunc;
  silc_mfunc_param_t *param;
  int *param_type;
  int status, i, j;
#if defined(TIME_FIND_MFUNC) ||  defined(TIME_CALL_MFUNC)
  double t, elapsed;
#endif

#define arg_error(n, msg) \
	error("%s: %d%s argument: %s", name, n,		\
	      (n != 11 && n % 10 == 1) ? "st" :		\
	      (n != 12 && n % 10 == 2) ? "nd" :		\
	      (n != 13 && n % 10 == 3) ? "rd" : "th",	\
	      msg)

  param_type = (int *)malloc(sizeof(int) * n_params);
  for (i = 0; i < n_params; i++) {
    object = var_get(value[i]->name);
    if (!object) {
      arg_error(i + 1, "undefined variable");
      free(param_type);
      return NULL;
    }
    if (object->type == SILC_RANGE_TYPE)
      param_type[i] = SILC_COLUMN_VECTOR_TYPE;
    else
      param_type[i] = object->type;
  }
#ifdef TIME_FIND_MFUNC
  t = silc_wtime();
#endif
  mfunc = silc_find_function_impl(name, n_params, param_type);
  free(param_type);
  if (!mfunc) {
    error("FATAL: no module function for %s found", name);
    return NULL;
  }
#ifdef TIME_FIND_MFUNC
  elapsed = silc_wtime() - t;
  _find_mfunc_time += elapsed;
#endif
#ifdef VANILLA_MPI
  if (silc_ipc_mpi_mfunc(mfunc->flags))
    silc_ipc_init(mfunc->module_name, mfunc->mfunc_name);
#endif
  param = (silc_mfunc_param_t *)
    malloc(sizeof(silc_mfunc_param_t) * (n_params + 1));
  for (i = 0; i < n_params; i++) {
    param[i].o = var_get(value[i]->name);
    param[i].m = value[i]->modifiers;
#if 1 /* XXX strip all modifiers */
    if (value[i]->modifiers) {
      param[i].o = modifier_apply_all(param[i].o, value[i]->modifiers);
      param[i].m = NULL;
    } else if (param[i].o->type == SILC_RANGE_TYPE) {
#ifdef TIME_CALL_MFUNC
      t = silc_wtime();
#endif
      param[i].o = silc_object_new_column_vector_from_range(param[i].o);
#ifdef TIME_CALL_MFUNC
      elapsed = silc_wtime() - t;
      _call_mfunc_time += elapsed;
#endif
    }
#endif
  }
  status = call_mfunc_impl(mfunc, n_params, param);
  object = param[n_params].o; /* returned value */
#if 1 /* XXX */
  for (i = 0; i < n_params; i++) {
    if (value[i]->modifiers || param[i].o->type == SILC_RANGE_TYPE)
      silc_object_free(param[i].o);
  }
#endif
  free(param);
  free(mfunc);
  if (status < 0) {
    error("FATAL: function %s failed", name);
    return NULL;
  }
  return object;
}

static value_t *
value_new(char *name, int is_temp)
{
  value_t *value;

#ifdef DEBUG
  if (is_temp) {
    silc_object_t *object = var_get(name);
    if (object) {
      printf("bound to %s:\n", name);
      silc_object_pprint(object, 1);
    }
  }
#endif
  value = (value_t *)malloc(sizeof(value_t));
  strcpy(value->name, name);
  value->is_temp = is_temp;
  value->modifiers = NULL;
  return value;
}

int
value_append_subscript(value_t *operand, int n, clist_t *subscript[3])
{
  silc_object_t *object;
  clist_t *list;
  int i, j, len, dimension[3];

  object = var_get(operand->name);
  if (!object) {
    error("undefined variable: %s", operand->name);
    return -1;
  }
  switch (object->type) {
  case SILC_SCALAR_TYPE:
    error("cannot apply subscript to scalar"); /* XXX */
    return -1;
  case SILC_RANGE_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    if (n != 1) {
      error("cannot apply %d-dimensional subscript to vector", n);
      return -1;
    }
    dimension[0] = object->d.vector.length;
    break;
  case SILC_MATRIX_TYPE:
    if (n != 2) {
      error("cannot apply %d-dimensional subscript to matrix", n);
      return -1;
    }
    dimension[0] = object->d.matrix.m;
    dimension[1] = object->d.matrix.n;
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    if (n != 3) {
      error("cannot apply %d-dimensional subscript to cubic array", n);
      return -1;
    }
    dimension[0] = object->d.cubic_array.l;
    dimension[1] = object->d.cubic_array.m;
    dimension[2] = object->d.cubic_array.n;
    break;
  default:
    should_not_reach_here;
  }
  if (operand->modifiers == NULL) {
    operand->modifiers = modifier_new();
    operand->modifiers->n = n;
    for (i = 0; i < n; i++) {
      clist_set_defaults(subscript[i], 0, dimension[i] - 1);
      if (clist_min(subscript[i]) < 0 ||
	  clist_max(subscript[i]) > dimension[i] - 1) {
	dimension_error(i, "subscript index out of range");
	return -1;
      }
      operand->modifiers->subscript[i] = subscript[i];
    }
  } else if (operand->modifiers->n == 0) {
    for (i = 0; i < n; i++) {
      if (n == 2 && operand->modifiers->transpose)
	j = (i == 0) ? 1 : 0;
      else
	j = i;
      clist_set_defaults(subscript[j], 0, dimension[i] - 1);
      if (clist_min(subscript[j]) < 0 ||
	  clist_max(subscript[j]) > dimension[i] - 1) {
	dimension_error(i, "subscript index out of range");
	return -1;
      }
      operand->modifiers->subscript[i] = subscript[j];
    }
    operand->modifiers->n = n;
  } else if (operand->modifiers->n != n) {
    error("subscript size mismatch");
    return -1;
  } else {
    for (i = 0; i < n; i++) {
      if (n == 2 && operand->modifiers->transpose)
	j = (i == 0) ? 1 : 0;
      else
	j = i;
      clist_set_defaults(operand->modifiers->subscript[i],
			 0, dimension[i] - 1);
      len = clist_length(operand->modifiers->subscript[i]);
      clist_set_defaults(subscript[j],
			 0, len - 1);
      if (clist_min(subscript[j]) < 0 ||
	  clist_max(subscript[j]) > len - 1) {
	dimension_error(i, "subscript index out of range");
	return -1;
      }
      list = clist_sublist(operand->modifiers->subscript[i], subscript[j]);
      clist_free(operand->modifiers->subscript[i]);
      operand->modifiers->subscript[i] = list;
    }
  }
#ifdef DEBUG
  printf("value_append_subscript: name=\"%s\" (0x%lx)\n",
	 operand->name, object);
  printf("  subscript[%d]:\n", operand->modifiers->n);
  for (i = 0; i < operand->modifiers->n; i++) {
    printf("    %d:", i); clist_pprint(operand->modifiers->subscript[i]);
  }
  printf("  conjugate: %d\n", operand->modifiers->conjugate);
  printf("  transpose: %d\n", operand->modifiers->transpose);
  printf("  negate: %d\n", operand->modifiers->negate);
#endif
  return 0;
}

static double
get_inf()
{
  unsigned short n = 0x1234;
  unsigned char *p = (unsigned char *)&n;
  unsigned char inf_be[8] = {0x7f, 0xf0, 0, 0, 0, 0, 0, 0};
  unsigned char inf_le[8] = {0, 0, 0, 0, 0, 0, 0xf0, 0x7f};
  return *(double *)((*p == 0x12) ? inf_be : inf_le);
}

static value_t *
evaluate_expression(expr_t *expr)
{
  value_t *value, *operand, *lhs, *rhs, **list;
  silc_object_t *object;
  clist_t *subscript[3];
  char temp[SILC_MAXNAMELEN+1];
  int i, n;

  switch(expr->type) {
  case VariableExprType:
    if (strcmp(VARIABLE(expr)->name, "i") == 0)
      object = silc_object_new_complex(0.0, 1.0);
    else if (strcmp(VARIABLE(expr)->name, "e") == 0)
      object = silc_object_new_double(exp(1.0));
    else if (strcmp(VARIABLE(expr)->name, "pi") == 0)
      object = silc_object_new_double(M_PI);
    else if (strcmp(VARIABLE(expr)->name, "inf") == 0)
      object = silc_object_new_double(get_inf());
    else
      return value_new(VARIABLE(expr)->name, 0);
    break;
  case ScalarExprType:
    switch (SCALAR(expr)->precision) {
    case SILC_INT:
      object = silc_object_new_int(SCALAR(expr)->v.i); break;
    case SILC_LONG:
      object = silc_object_new_long(SCALAR(expr)->v.l); break;
    case SILC_FLOAT:
      object = silc_object_new_float(SCALAR(expr)->v.s); break;
    case SILC_DOUBLE:
      object = silc_object_new_double(SCALAR(expr)->v.d); break;
    }
    break;
  case RangeExprType:
    object = new_range(RANGE(expr)->start, RANGE(expr)->end);
    break;
  case RowVectorExprType: /* horizontal concatenation */
    object = new_array(ROW_VECTOR(expr)->v,
		       ROW_VECTOR(expr)->length,
		       SILC_ROW_VECTOR_TYPE);
    break;
  case ColumnVectorExprType: /* vertical concatenation */
    object = new_array(COLUMN_VECTOR(expr)->v,
		       COLUMN_VECTOR(expr)->length,
		       SILC_COLUMN_VECTOR_TYPE);
    break;
  case CubicArrayExprType:
    object = new_cubic_array(CUBIC_ARRAY(expr)->v,
			     CUBIC_ARRAY(expr)->length);
    break;
  case SubscriptExprType:
    operand = evaluate_expression(SUBSCRIPT(expr)->operand);
    if (!operand)
      return NULL;
    if (evaluate_subscript(SUBSCRIPT(expr)->op, subscript) < 0) {
      value_free(operand);
      return NULL;
    }
    n = SUBSCRIPT(expr)->op->n;
    if (value_append_subscript(operand, n, subscript) < 0) {
      value_free(operand);
      for (i = 0; i < n; i++)
	clist_free(subscript[i]);
      return NULL;
    }
    return operand;
  case UnaryOpExprType:
    operand = evaluate_expression(UNARY_OP(expr)->operand);
    if (!operand)
      return NULL;
    if (operand->modifiers == NULL)
      operand->modifiers = modifier_new();
    switch(UNARY_OP(expr)->op) {
    case SILC_CONJUGATE_TRANSPOSE_OP:
      operand->modifiers->conjugate = !operand->modifiers->conjugate;
      operand->modifiers->transpose = !operand->modifiers->transpose;
      return operand;
    case SILC_COMPLEX_CONJUGATE_OP:
      operand->modifiers->conjugate = !operand->modifiers->conjugate;
      return operand;
    case SILC_TRANSPOSE_OP:
      operand->modifiers->transpose = !operand->modifiers->transpose;
      return operand;
    case SILC_NEGATE_OP:
      operand->modifiers->negate = !operand->modifiers->negate;
      return operand;
    default:
      should_not_reach_here;
    }
    break;
  case BinaryOpExprType:
    lhs = evaluate_expression(BINARY_OP(expr)->lhs);
    if (!lhs)
      return NULL;
    rhs = evaluate_expression(BINARY_OP(expr)->rhs);
    if (!rhs) {
      value_free(lhs);
      return NULL;
    }
    object = do_binary_op(BINARY_OP(expr)->op, lhs, rhs);
    value_free(lhs);
    value_free(rhs);
    break;
  case FuncCallExprType:
    list = evaluate_expression_list(FUNC_CALL(expr)->args,
				    FUNC_CALL(expr)->n_args);
    if (!list)
      return NULL;
    object = do_func_call(FUNC_CALL(expr)->name,
			  FUNC_CALL(expr)->n_args,
			  list);
    value_list_free(list, FUNC_CALL(expr)->n_args);
    break;
  default:
    should_not_reach_here;
  }
  if (!object)
    return NULL;
  return value_new(var_add_temp(object, temp), 1);
}

static clist_t *
evaluate_subscript_1(expr_t *expr)
{
  value_t *value;
  silc_object_t *object;
  clist_t *list;
  int *p, i;

  value = evaluate_expression(expr);
  if (!value)
    return NULL;
  object = value_get_modified_object(value);
  value_free(value);
  if (!object)
    return NULL;
  if (object->precision != SILC_INT) {
    error("invalid precision for subscript: %s", SILC_SCALAR_NAME(object));
    silc_object_free(object);
    return NULL;
  }
  /* convert the origin of indeces from 1 (in expressions) to 0 (in C) */
  switch (object->type) {
  case SILC_SCALAR_TYPE:
    list = clist_new();
    clist_append_point(list, object->d.scalar.i - 1);
    break;
  case SILC_RANGE_TYPE:
    list = clist_new();
    clist_append_range(list, object->d.vector.v.range.start - 1,
			     object->d.vector.v.range.end   - 1);
    break;
  case SILC_COLUMN_VECTOR_TYPE:
    p = (int *)malloc(sizeof(int) * object->d.vector.length);
    for (i = 0; i < object->d.vector.length; i++)
      p[i] = object->d.vector.v.i[i] - 1;
    list = clist_compress(p, object->d.vector.length);
    free(p);
    break;
  default:
    error("invalid object type for subscript: %s", SILC_TYPE_NAME(object));
    list = NULL;
  }
  silc_object_free(object);
  return list;  /* a new instance of clist_t */
}

/* returns new instances of clist_t through v */
static int
evaluate_subscript(subscript_t *subscript, clist_t *v[3])
{
  int i;

  for (i = 0; i < subscript->n; i++) {
    if ((v[i] = evaluate_subscript_1(subscript->v[i])) == NULL) {
      while (i > 0)
	clist_free(v[--i]);
      return -1;
    }
  }
  return 0;
}

static void
value_free(value_t *value)
{
  if (value->is_temp)
    var_free(value->name);
  if (value->modifiers)
    modifier_free(value->modifiers);
  free(value);
}

static value_t **
evaluate_expression_list(expr_t **expr, int n)
{
  value_t **list;
  int i;

  list = (value_t **)malloc(sizeof(value_t *) * n);
  for (i = 0; i < n; i++) {
    if ((list[i] = evaluate_expression(expr[i])) == NULL) {
      value_list_free(list, i);
      return NULL;
    }
  }
  return list;
}

static void
value_list_free(value_t **list, int n)
{
  int i;

  for (i = 0; i < n; i++)
    value_free(list[i]);
  free(list);
}

static silc_object_t *
value_get_object(value_t *value)
{
  silc_object_t *object;

  if (value->is_temp)
    return var_get_free(value->name);
  object = var_get(value->name);
  if (!object)
    return NULL;
  return silc_object_duplicate(object, 1);
}

static silc_object_t *
value_get_modified_object(value_t *value)
{
  silc_object_t *object;

  if (value->modifiers) {
    object = var_get(value->name);
    if (!object)
      return NULL;
    return modifier_apply_all(object, value->modifiers);
  }
  return value_get_object(value);
}

static silc_modifier_t *
value_get_modifiers(value_t *value)
{
  silc_modifier_t *modifiers;

  modifiers = value->modifiers;
  value->modifiers = NULL;
  return modifiers;
}
