/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: leq_cg.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <malloc.h>
#include <string.h>
#include <math.h>

#include "module.h"
#include "dense.h"
#include "sparse_crs.h"
#include "wtime.h"

#include "ssi_leqlib.h"

#define MODULE_NAME "leq_cg"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#ifdef DEBUG
static double
check_dense(silc_object_t *A, double *b, double *x)
{
  silc_dense_t *in;
  double dot, d;
  int i, j;

  in = SILC_DENSE(A);
  dot = 0.0;
  for (i = 0; i < A->d.matrix.m; i++) {
    d = b[i];
    for (j = 0; j < A->d.matrix.n; j++)
      d -= in->v.d[j * A->d.matrix.m + i] * x[j];
    dot += d * d;
  }
  return sqrt(dot);
}

static double
check_crs(silc_object_t *A, double *b, double *x)
{
  silc_crs_t *in;
  double dot, d;
  int i, j;

  in = SILC_CRS(A);
  dot = 0.0;
  for (i = 0; i < A->d.matrix.m; i++) {
    d = b[i];
    for (j = in->row[i]; j < in->row[i+1]; j++)
      d -= in->v.d[j] * x[in->index[j]];
    dot += d * d;
  }
  return sqrt(dot);
}
#endif /* DEBUG */

static silc_object_t *
solve_dense_1(silc_object_t *A, silc_object_t *b)
{
  silc_object_t *x;
  silc_dense_t *A_in;
  double *work, *p, *q;
  double ssi_params[SSI_PARAMS_LEN];
  int ssi_options[SSI_OPTIONS_LEN];
  int status, n;
  SSI_MATRIX mat_A;

#ifdef DEBUG
  printf(MODULE_NAME ": solve_dense_1(0x%lx, 0x%lx)\n", A, b);
#endif
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return NULL;
  }
  if (b->type == SILC_MATRIX_TYPE) {
    if (A->d.matrix.n != b->d.matrix.m) {
      fprintf(stderr, MODULE_NAME ": matrix-matrix size mismatch\n");
      return NULL;
    }
  } else {
    if (A->d.matrix.n != b->d.vector.length) {
      fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
      return NULL;
    }
  }
  x = silc_object_duplicate(b, 0);
  if (x->type == SILC_MATRIX_TYPE)
    memset(SILC_DENSE(x)->v.d, 0,
	   sizeof(double) * x->d.matrix.m * x->d.matrix.n);
  else
    memset(x->d.vector.v.d, 0, sizeof(double) * x->d.vector.length);
  A_in = SILC_DENSE(A);
  mat_A.matrix_type = SSI_MATRIX_TYPE_DENSE;
  mat_A.array_base  = SSI_MATRIX_ARRAY_BASE_0;
  mat_A.matrix_base = SSI_MATRIX_ARRAY_BASE_0;
  mat_A.n     = A->d.matrix.m;
  mat_A.value = A_in->v.d;
  work = (double *)malloc(sizeof(double) * A->d.matrix.m * 4);
  ssi_params[SSI_PARAMS_RESID] = 1.0e-10;
  ssi_options[SSI_OPTIONS_MAXITER] = A->d.matrix.m;
#ifdef DEBUG
  ssi_options[SSI_OPTIONS_OUTPUT] = 1;
#else
  ssi_options[SSI_OPTIONS_OUTPUT] = 0;
#endif
  if (b->type == SILC_MATRIX_TYPE) {
    p = SILC_DENSE(b)->v.d;
    q = SILC_DENSE(x)->v.d;
    for (n = 0; n < b->d.matrix.n; n++) {
      ssi_options[SSI_OPTIONS_MAXITER] = A->d.matrix.m;
      status = ssi_cg(p, q, work, ssi_params, ssi_options, &mat_A, NULL);
      if (status == SSIRET_MAXITER)
	break;
      p += b->d.matrix.m;
      q += x->d.matrix.m;
    }
  } else {
    ssi_options[SSI_OPTIONS_MAXITER] = A->d.matrix.m;
    status = ssi_cg(b->d.vector.v.d, x->d.vector.v.d, work,
		    ssi_params, ssi_options, &mat_A, NULL);
  }
  free(work);
  if (status == SSIRET_MAXITER) {
    silc_object_free(x);
    fprintf(stderr, MODULE_NAME ": too many iterations\n");
    return NULL;
  }
#ifdef DEBUG
  /* check solution */
  if (b->type == SILC_MATRIX_TYPE) {
    p = SILC_DENSE(b)->v.d;
    q = SILC_DENSE(x)->v.d;
    for (n = 0; n < b->d.matrix.n; n++) {
      printf(MODULE_NAME ": ||b-Ax|| = %e (%d)\n", check_dense(A, p, q), n);
      p += b->d.matrix.m;
      q += x->d.matrix.m;
    }
  } else {
    printf(MODULE_NAME ": ||b-Ax|| = %e\n",
	   check_dense(A, b->d.vector.v.d, x->d.vector.v.d));
  }
#endif
  return x;
}

int
leq_cg_solve_dense(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[0].o;
  b = param[1].o;
  param[2].o = solve_dense_1(A, b);
  return (param[2].o) ? 0 : -1;
}

int
leq_cg_divide_dense(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[1].o;
  b = param[0].o;
  if (b->type == SILC_MATRIX_TYPE)
    b = modifier_apply_transpose(b);
  param[2].o = solve_dense_1(A, b);
  if (b->type == SILC_MATRIX_TYPE)
    silc_object_free(b);
  return (param[2].o) ? 0 : -1;
}

#ifdef TIME_CRS_CG
extern double _crs_cg_time;  /* server.c */
#endif

static silc_object_t *
solve_crs_1(silc_object_t *A, silc_object_t *b)
{
  silc_object_t *x;
  silc_crs_t *A_in;
  double *work, *p, *q;
  double ssi_params[SSI_PARAMS_LEN];
  int ssi_options[SSI_OPTIONS_LEN];
  int status, n;
  SSI_MATRIX mat_A;
  double t, elapsed;

#ifdef DEBUG
  printf(MODULE_NAME ": solve_crs_1(0x%lx, 0x%lx)\n", A, b);
#endif
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return NULL;
  }
  if (b->type == SILC_MATRIX_TYPE) {
    if (A->d.matrix.n != b->d.matrix.m) {
      fprintf(stderr, MODULE_NAME ": matrix-matrix size mismatch\n");
      return NULL;
    }
  } else {
    if (A->d.matrix.n != b->d.vector.length) {
      fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
      return NULL;
    }
  }
  x = silc_object_duplicate(b, 0);
  if (x->type == SILC_MATRIX_TYPE)
    memset(SILC_DENSE(x)->v.d, 0,
	   sizeof(double) * x->d.matrix.m * x->d.matrix.n);
  else
    memset(x->d.vector.v.d, 0, sizeof(double) * x->d.vector.length);
  A_in = SILC_CRS(A);
  mat_A.matrix_type = SSI_MATRIX_TYPE_CRS;
  mat_A.array_base  = SSI_MATRIX_ARRAY_BASE_0;
  mat_A.matrix_base = SSI_MATRIX_ARRAY_BASE_0;
  mat_A.n     = A->d.matrix.m;
  mat_A.nnz   = A_in->nnz;
  mat_A.value = A_in->v.d;
  mat_A.index = A_in->index;
  mat_A.row   = A_in->row;
  work = (double *)malloc(sizeof(double) * A->d.matrix.m * 4);
  ssi_params[SSI_PARAMS_RESID] = 1.0e-10;
#ifdef DEBUG
  ssi_options[SSI_OPTIONS_OUTPUT] = 1;
#else
  ssi_options[SSI_OPTIONS_OUTPUT] = 0;
#endif
#ifdef TIME_CRS_CG
  t = silc_wtime();
#endif
  if (b->type == SILC_MATRIX_TYPE) {
    p = SILC_DENSE(b)->v.d;
    q = SILC_DENSE(x)->v.d;
    for (n = 0; n < b->d.matrix.n; n++) {
      ssi_options[SSI_OPTIONS_MAXITER] = A->d.matrix.m;
      status = ssi_cg(p, q, work, ssi_params, ssi_options, &mat_A, NULL);
      if (status == SSIRET_MAXITER)
	break;
      p += b->d.matrix.m;
      q += x->d.matrix.m;
    }
  } else {
    ssi_options[SSI_OPTIONS_MAXITER] = A->d.matrix.m;
    status = ssi_cg(b->d.vector.v.d, x->d.vector.v.d, work,
		    ssi_params, ssi_options, &mat_A, NULL);
  }
#ifdef TIME_CRS_CG
  elapsed = silc_wtime() - t;
#ifdef DEBUG
  printf(MODULE_NAME ": %lf sec. (iter = %d)\n",
	 elapsed, ssi_options[SSI_OPTIONS_MAXITER]);
#endif
  _crs_cg_time += elapsed;
#endif
  free(work);
#if 0
  if (status == SSIRET_MAXITER) {
    silc_object_free(x);
    fprintf(stderr, MODULE_NAME ": too many iterations\n");
    return NULL;
  }
#endif
#ifdef DEBUG
  /* check solution */
  if (b->type == SILC_MATRIX_TYPE) {
    p = SILC_DENSE(b)->v.d;
    q = SILC_DENSE(x)->v.d;
    for (n = 0; n < b->d.matrix.n; n++) {
      printf(MODULE_NAME ": ||b-Ax|| = %e (%d)\n", check_crs(A, p, q), n);
      p += b->d.matrix.m;
      q += x->d.matrix.m;
    }
  } else {
    printf(MODULE_NAME ": ||b-Ax|| = %e\n",
	   check_crs(A, b->d.vector.v.d, x->d.vector.v.d));
  }
#endif
  return x;
}

int
leq_cg_solve_crs(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[0].o;
  b = param[1].o;
  param[2].o = solve_crs_1(A, b);
  return (param[2].o) ? 0 : -1;
}

int
leq_cg_divide_crs(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[1].o;
  b = param[0].o;
  if (b->type == SILC_MATRIX_TYPE)
    b = modifier_apply_transpose(b);
  param[2].o = solve_crs_1(A, b);
  if (b->type == SILC_MATRIX_TYPE)
    silc_object_free(b);
  return (param[2].o) ? 0 : -1;
}

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {leq_cg_solve_dense, "leq_cg_solve_dense",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
  {leq_cg_solve_dense, "leq_cg_solve_dense",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_cg_divide_dense, "leq_cg_divide_dense",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_cg_divide_dense, "leq_cg_divide_dense",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_cg_solve_crs, "leq_cg_solve_crs",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
  {leq_cg_solve_crs, "leq_cg_solve_crs",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_cg_divide_crs, "leq_cg_divide_crs",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, 2, 0},
  {leq_cg_divide_crs, "leq_cg_divide_crs",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, 2, 0},
  {NULL} /* sentinel */
};

void
init_leq_cg()
{
  silc_register_module(MODULE_NAME,
		       NULL,   /* _unary_op_table */
		       _binary_op_table,
		       NULL,   /* _function_table */
		       NULL);  /* _procedure_table */
}
