/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: leq_lis.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>

#include "module.h"
#include "dense.h"
#include "sparse_crs.h"
#include "variable.h"

#include "lis.h"

#define MODULE_NAME "leq_lis"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

static void
load_options(int options[], LIS_SCALAR params[])
{
  silc_object_t *object;
  int i;

  object = var_get("lis_options");
  if (!object) {
    object = silc_object_new_column_vector(SILC_INT, LIS_OPTIONS_LEN);
    for (i = 0; i < LIS_OPTIONS_LEN; i++)
      object->d.vector.v.i[i] = options[i];
    var_add("lis_options", object);
  }
  if (object->type == SILC_COLUMN_VECTOR_TYPE &&
      object->precision == SILC_INT &&
      object->d.vector.length == LIS_OPTIONS_LEN) {
    for (i = 0; i < LIS_OPTIONS_LEN; i++)
      options[i] = object->d.vector.v.i[i];
  } else {
    fprintf(stderr, MODULE_NAME ": broken lis_options (ignored)\n");
  }
  object = var_get("lis_params");
  if (!object) {
    object = silc_object_new_column_vector(SILC_DOUBLE, LIS_PARAMS_LEN);
    for (i = 0; i < LIS_PARAMS_LEN; i++)
      object->d.vector.v.d[i] = params[i];
    var_add("lis_params", object);
  }
  if (object->type == SILC_COLUMN_VECTOR_TYPE &&
      object->precision == SILC_DOUBLE &&
      object->d.vector.length == LIS_PARAMS_LEN) {
    for (i = 0; i < LIS_PARAMS_LEN; i++)
      params[i] = object->d.vector.v.d[i];
  } else {
    fprintf(stderr, MODULE_NAME ": broken lis_params (ignored)\n");
  }
}

static void
store_status(LIS_SCALAR status[])
{
  silc_object_t *object;
  int i;

  object = var_get("lis_status");
  if (!object) {
    object = silc_object_new_column_vector(SILC_DOUBLE, LIS_STATUS_LEN);
    var_add("lis_status", object);
  }
  if (object->type == SILC_COLUMN_VECTOR_TYPE &&
      object->precision == SILC_DOUBLE &&
      object->d.vector.length == LIS_STATUS_LEN) {
    for (i = 0; i < LIS_STATUS_LEN; i++)
      object->d.vector.v.d[i] = status[i];
  } else {
    fprintf(stderr, MODULE_NAME ": broken lis_status (ignored)\n");
  }
}

/* binary operators */

static silc_object_t *
leq_lis_solve_1(int format, silc_object_t *A, silc_object_t *b)
{
  silc_object_t *x;
  double *p, *q;
  LIS_MATRIX lis_A;
  LIS_VECTOR lis_b, lis_x;
  LIS_SCALAR params[LIS_PARAMS_LEN];
  LIS_SCALAR status[LIS_STATUS_LEN];
  int ret, options[LIS_OPTIONS_LEN];
  size_t i, n;

  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return NULL;
  }
  if (b->type == SILC_MATRIX_TYPE) {
    if (A->d.matrix.n != b->d.matrix.m) {
      fprintf(stderr, MODULE_NAME ": matrix-matrix size mismatch\n");
      return NULL;
    }
  } else {
    if (A->d.matrix.n != b->d.vector.length) {
      fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
      return NULL;
    }
  }
  x = silc_object_duplicate(b, 0);
  lis_matrix_create(A->d.matrix.m, 0, 0, &lis_A);
  switch (format) {
  case LIS_MATRIX_DNS:
    {
      silc_dense_t *in;
      in = SILC_DENSE(A);
      lis_matrix_set_dns(in->v.d, lis_A);
    }
    break;
  case LIS_MATRIX_CRS:
    {
      silc_crs_t *in;
      in = SILC_CRS(A);
#if 0
      printf("n: %d , nnz: %d\n", (int)A->d.matrix.m, (int)in->nnz);
#endif
#if 1
      lis_matrix_set_crs(in->nnz, in->row, in->index, in->v.d, lis_A);
#else
      for (i = 0; i < A->d.matrix.m; i++) {
	int j;
	for (j = in->row[i]; j < in->row[i+1]; j++)
	  lis_matrix_set(i, in->index[j], in->v.d[j], lis_A);
      }
#endif
    }
    break;
  default:
    should_not_reach_here;
  }
  lis_matrix_assemble(&lis_A, format);
  lis_vector_create(A->d.matrix.n, 0, 0, &lis_b);
  lis_vector_create(A->d.matrix.n, 0, 0, &lis_x);
  load_options(options, params);
  if (b->type == SILC_MATRIX_TYPE) {
    p = SILC_DENSE(b)->v.d;
    q = SILC_DENSE(x)->v.d;
    for (n = 0; n < b->d.matrix.n; n++) {
      memcpy(lis_b->value, p, sizeof(double) * b->d.matrix.n);
      lis_vector_set_all(0.0, lis_x);
      lis_solve(lis_A, lis_b, lis_x, params, options, status);
      ret = (int)status[LIS_STATUS_RET];
      memcpy(q, lis_x->value, sizeof(double) * x->d.matrix.n);
      if (ret != LIS_SUCCESS)
	break;
      p += b->d.matrix.m;
      q += x->d.matrix.m;
    }
  } else {
    memcpy(lis_b->value, b->d.vector.v.d, sizeof(double) * b->d.vector.length);
    lis_vector_set_all(0.0, lis_x);
    lis_solve(lis_A, lis_b, lis_x, params, options, status);
    ret = (int)status[LIS_STATUS_RET];
    memcpy(x->d.vector.v.d, lis_x->value, sizeof(double) * x->d.vector.length);
  }
  store_status(status);
  switch (format) {
  case LIS_MATRIX_DNS:
    lis_A->value = NULL;
    break;
  case LIS_MATRIX_CRS:
    lis_A->value = NULL;
    lis_A->ptr = lis_A->index = NULL;
    break;
  default:
    should_not_reach_here;
  }
  lis_matrix_destroy(lis_A);
  lis_vector_destroy(lis_b);
  lis_vector_destroy(lis_x);
#if 1
  if (ret == LIS_SUCCESS) {
#else
  if (ret == LIS_SUCCESS || ret == LIS_MAXITER) {
#endif
#ifdef TIME_LIS_SOLVE
    double elapsed, Mflop;
    size_t nnz, iter;

    switch (format) {
    case LIS_MATRIX_DNS:
      nnz = A->d.matrix.m * A->d.matrix.n;
      break;
    case LIS_MATRIX_CRS:
      nnz = SILC_CRS(A)->nnz;
      break;
    default:
      should_not_reach_here;
    }
    elapsed = status[LIS_STATUS_TIMES];
    iter = (size_t)status[LIS_STATUS_ITER];
    Mflop = 4 * A->d.matrix.n;
    Mflop += (double)iter * (2 * nnz + 12 * A->d.matrix.n + 3);
    Mflop /= 1024 * 1024;
    printf(MODULE_NAME ": %lf sec (precon: %lf sec, iter: %lf sec)\n",
	   elapsed, status[LIS_STATUS_PTIMES], status[LIS_STATUS_ITIMES]);
    printf(MODULE_NAME ": #iters = %ld\n", iter);
    printf(MODULE_NAME ": ||b-Ax|| = %.6e\n", status[LIS_STATUS_RESID]);
    printf(MODULE_NAME ": %lf Mflop, %lf Mflops\n", Mflop, Mflop / elapsed);
#endif
  } else {
    switch (ret) {
    case LIS_ILL_OPTION:
      fprintf(stderr, MODULE_NAME ": illegal LIS_OPTIONS\n"); break;
    case LIS_BREAKDOWN:
      fprintf(stderr, MODULE_NAME ": breakdown\n"); break;
    case LIS_OUT_OF_MEMORY:
      fprintf(stderr, MODULE_NAME ": out of memory\n"); break;
    case LIS_MAXITER:
      fprintf(stderr, MODULE_NAME ": too many iterations (limit = %d)\n",
	      options[LIS_OPTIONS_MAXITER]);
      break;
    default:
      should_not_reach_here;
    }
    silc_object_free(x);
    return NULL;
  }
  return x;
}

int
leq_lis_solve_dense(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[0].o;
  b = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_solve_dense(0x%lx, 0x%lx)\n", A, b);
#endif
  param[2].o = leq_lis_solve_1(LIS_MATRIX_DNS, A, b);
  return (param[2].o) ? 0 : -1;
}

int
leq_lis_divide_dense(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[1].o;
  b = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_divide_dense(0x%lx, 0x%lx)\n", b, A);
#endif
  if (b->type == SILC_MATRIX_TYPE)
    b = modifier_apply_transpose(b);
  param[2].o = leq_lis_solve_1(LIS_MATRIX_DNS, A, b);
  if (b->type == SILC_MATRIX_TYPE)
    silc_object_free(b);
  return (param[2].o) ? 0 : -1;
}

int
leq_lis_solve_crs(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[0].o;
  b = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_solve_crs(0x%lx, 0x%lx)\n", A, b);
#endif
  param[2].o = leq_lis_solve_1(LIS_MATRIX_CRS, A, b);
  return (param[2].o) ? 0 : -1;
}

int
leq_lis_divide_crs(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  A = param[1].o;
  b = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_divide_crs(0x%lx, 0x%lx)\n", b, A);
#endif
  if (b->type == SILC_MATRIX_TYPE)
    b = modifier_apply_transpose(b);
  param[2].o = leq_lis_solve_1(LIS_MATRIX_CRS, A, b);
  if (b->type == SILC_MATRIX_TYPE)
    silc_object_free(b);
  return (param[2].o) ? 0 : -1;
}

#ifdef WITH_COCG

static int
lis_cocg(LIS_MATRIX A_r, LIS_MATRIX A_i,
	 LIS_VECTOR b_r, LIS_VECTOR b_i,
	 LIS_VECTOR x_r, LIS_VECTOR x_i,
	 LIS_VECTOR work[],
	 LIS_SCALAR params[], int options[],
	 LIS_SCALAR status[])
{
  LIS_VECTOR p_r, q_r, r_r, z_r, tmp_r;
  LIS_VECTOR p_i, q_i, r_i, z_i, tmp_i;
  LIS_SCALAR alpha_r, beta_r, rho_r, rho_old_r, dot_pq_r;
  LIS_SCALAR alpha_i, beta_i, rho_i, rho_old_i, dot_pq_i;
  LIS_SCALAR bnrm2, nrm2, tol, dot_1, dot_2, d;
  int i, iter, n, maxiter, output;
  double ptimes, times;

  n = A_r->n;
  maxiter = options[LIS_OPTIONS_MAXITER];
  output  = options[LIS_OPTIONS_OUTPUT];
  tol     = params[LIS_PARAMS_RESID];
  ptimes  = 0.0;

  p_r = work[0];
  p_i = work[1];
  q_r = work[2];
  q_i = work[3];
  r_r = work[4];
  r_i = work[5];
  z_r = work[6];
  z_i = work[7];
  tmp_r = work[8];
  tmp_i = work[9];

  rho_old_r = (LIS_SCALAR)1.0;
  rho_old_i = (LIS_SCALAR)0.0;
  beta_r = (LIS_SCALAR)0.0;
  beta_i = (LIS_SCALAR)0.0;

  lis_vector_set_all(0.0, p_r);
  lis_vector_set_all(0.0, p_i);

  /* ||b|| */
  lis_vector_dot(b_r, b_r, &dot_1);
  lis_vector_dot(b_i, b_i, &dot_2);
  bnrm2 = sqrt(dot_1 + dot_2);
  if (bnrm2 == 0.0)
    bnrm2 = 1.0;
  else
    bnrm2 = 1.0 / bnrm2;
  
  /* initial residual */
  lis_vector_copy(b_r, r_r);
  lis_vector_copy(b_i, r_i);
 
  for (iter = 1; iter <= maxiter; iter++) {
    /* z = M^-1 * r */
    times = lis_wtime();
    lis_vector_copy(r_r, z_r);
    lis_vector_copy(r_i, z_i);
    ptimes += lis_wtime() - times;

    /* rho = <r, z> */
    lis_vector_dot(r_r, z_r, &dot_1);
    lis_vector_dot(r_i, z_i, &dot_2);
    rho_r = dot_1 + dot_2;
    lis_vector_dot(r_r, z_i, &dot_1);
    lis_vector_dot(r_i, z_r, &dot_2);
    rho_i = dot_1 - dot_2;
    
    /* beta = rho / rho_old */
    d = rho_old_r * rho_old_r + rho_old_i * rho_old_i;
    beta_r = (rho_r * rho_old_r + rho_i * rho_old_i) / d;
    beta_i = (rho_i * rho_old_r - rho_r * rho_old_i) / d;

    /* p = z + beta * p */
    lis_vector_axpyz(-beta_i, p_i, z_r, tmp_r);
    lis_vector_axpyz( beta_i, p_r, z_i, tmp_i);
    lis_vector_xpay(tmp_r, beta_r, p_r);
    lis_vector_xpay(tmp_i, beta_r, p_i);

    /* q = A * p */
    lis_matvec(A_r, p_r, q_r);
    lis_matvec(A_i, p_i, tmp_r);
    lis_matvec(A_r, p_i, q_i);
    lis_matvec(A_i, p_r, tmp_i);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      q_r->value[i] -= tmp_r->value[i];
#pragma omp parallel for
    for (i = 0; i < n; i++)
      q_i->value[i] += tmp_i->value[i];

    /* dot_pq = <p, q> */
    lis_vector_dot(p_r, q_r, &dot_1);
    lis_vector_dot(p_i, q_i, &dot_2);
    dot_pq_r = dot_1 + dot_2;
    lis_vector_dot(p_r, q_i, &dot_1);
    lis_vector_dot(p_i, q_r, &dot_2);
    dot_pq_i = dot_1 - dot_2;

    /* breakdown check */
    d = dot_pq_r * dot_pq_r + dot_pq_i * dot_pq_i;
    if (d == 0.0) {
      status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
      status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
      status[LIS_STATUS_RESID] = nrm2;
      return LIS_BREAKDOWN;
    }

    /* alpha = rho / dot_pq */
    alpha_r = (rho_r * dot_pq_r + rho_i * dot_pq_i) / d;
    alpha_i = (rho_i * dot_pq_r - rho_r * dot_pq_i) / d;

    /* x = x + alpha * p */
    lis_vector_axpy( alpha_r, p_r, x_r);
    lis_vector_axpy(-alpha_i, p_i, x_r);
    lis_vector_axpy( alpha_r, p_i, x_i);
    lis_vector_axpy( alpha_i, p_r, x_i);

    /* r = r - alpha * q */
    lis_vector_axpy(-alpha_r, q_r, r_r);
    lis_vector_axpy( alpha_i, q_i, r_r);
    lis_vector_axpy(-alpha_r, q_i, r_i);
    lis_vector_axpy(-alpha_i, q_r, r_i);

    /* convergence check */
    lis_vector_dot(r_r, r_r, &dot_1);
    lis_vector_dot(r_i, r_i, &dot_2);
    nrm2 = sqrt(dot_1 + dot_2) * bnrm2;
    if (output & LIS_PRINT_OUT)
      printf("iter: %5d  residual = %e\n", iter, nrm2);
    if (tol >= nrm2) {
      status[LIS_STATUS_RET]    = (LIS_SCALAR)LIS_SUCCESS;
      status[LIS_STATUS_ITER]   = (LIS_SCALAR)iter;
      status[LIS_STATUS_RESID]  = nrm2;
      status[LIS_STATUS_PTIMES] = ptimes;
      return LIS_SUCCESS;
    }
    rho_old_r = rho_r;
    rho_old_i = rho_i;
  }
  
  status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
  status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
  status[LIS_STATUS_RESID] = nrm2;
  return LIS_MAXITER;
}

int
leq_lis_cocg(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;
  LIS_MATRIX lis_A_r, lis_A_i;
  LIS_VECTOR lis_b_r, lis_b_i, lis_x_r, lis_x_i, work[10];
  LIS_SCALAR params[LIS_PARAMS_LEN];
  LIS_SCALAR status[LIS_STATUS_LEN];
  int ret, options[LIS_OPTIONS_LEN];
  int i, j;

  A = param[0].o;
  b = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_cocg(0x%lx, 0x%lx)\n", A, b);
#endif
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return NULL;
  }
  if (A->d.matrix.n != b->d.vector.length) {
    fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
    return NULL;
  }
  x = silc_object_duplicate(b, 0);
  lis_matrix_create(A->d.matrix.m, 0, 0, &lis_A_r);
  lis_matrix_create(A->d.matrix.m, 0, 0, &lis_A_i);
  {
    silc_crs_t *in;
    in = SILC_CRS(A);
    for (i = 0; i < A->d.matrix.m; i++) {
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	lis_matrix_set(i, in->index[j], in->v.z[j*2],   lis_A_r);
	lis_matrix_set(i, in->index[j], in->v.z[j*2+1], lis_A_i);
      }
    }
  }
  lis_matrix_assemble(&lis_A_r, LIS_MATRIX_CRS);
  lis_matrix_assemble(&lis_A_i, LIS_MATRIX_CRS);
  lis_vector_create(A->d.matrix.n, 0, 0, &lis_b_r);
  lis_vector_create(A->d.matrix.n, 0, 0, &lis_b_i);
  lis_vector_create(A->d.matrix.n, 0, 0, &lis_x_r);
  lis_vector_create(A->d.matrix.n, 0, 0, &lis_x_i);
  for (i = 0; i < 10; i++)
    lis_vector_create(A->d.matrix.n, 0, 0, &work[i]);
#pragma omp parallel for
  for (i = 0; i < A->d.matrix.n; i++) {
    lis_b_r->value[i] = b->d.vector.v.z[i*2];
    lis_b_i->value[i] = b->d.vector.v.z[i*2+1];
  }
  lis_vector_set_all(0.0, lis_x_r);
  lis_vector_set_all(0.0, lis_x_i);
  load_options(options, params);
  lis_cocg(lis_A_r, lis_A_i,
	   lis_b_r, lis_b_i,
	   lis_x_r, lis_x_i,
	   work,
	   params, options, status);
  store_status(status);
#pragma omp parallel for
  for (i = 0; i < A->d.matrix.n; i++) {
    x->d.vector.v.z[i*2]   = lis_x_r->value[i];
    x->d.vector.v.z[i*2+1] = lis_x_i->value[i];
  }
  lis_matrix_destroy(lis_A_r);
  lis_matrix_destroy(lis_A_i);
  lis_vector_destroy(lis_b_r);
  lis_vector_destroy(lis_b_i);
  lis_vector_destroy(lis_x_r);
  lis_vector_destroy(lis_x_i);
  for (i = 0; i < 10; i++)
    lis_vector_destroy(work[i]);
  ret = (int)status[LIS_STATUS_RET];
  if (ret == LIS_SUCCESS /* || ret == LIS_MAXITER /**/) {
#ifdef TIME_LIS_SOLVE
    double elapsed, Mflop;
    size_t nnz, iter;

    nnz = SILC_CRS(A)->nnz;
    elapsed = status[LIS_STATUS_TIMES];
    iter = (size_t)status[LIS_STATUS_ITER];
    Mflop = 4 * A->d.matrix.n;
    Mflop += (double)iter * (2 * nnz + 12 * A->d.matrix.n + 3); /* XXX */
    Mflop /= 1024 * 1024;
    printf(MODULE_NAME ": %lf sec\n", elapsed);
    printf(MODULE_NAME ": #iters = %ld\n", iter);
    printf(MODULE_NAME ": ||b-Ax|| = %.6e\n", status[LIS_STATUS_RESID]);
    printf(MODULE_NAME ": %lf Mflop, %lf Mflops\n", Mflop, Mflop / elapsed);
#endif
  } else {
    switch (ret) {
    case LIS_ILL_OPTION:
      fprintf(stderr, MODULE_NAME ": illegal LIS_OPTIONS\n"); break;
    case LIS_BREAKDOWN:
      fprintf(stderr, MODULE_NAME ": breakdown\n"); break;
    case LIS_OUT_OF_MEMORY:
      fprintf(stderr, MODULE_NAME ": out of memory\n"); break;
    case LIS_MAXITER:
      fprintf(stderr, MODULE_NAME ": too many iterations (limit = %d)\n",
	      options[LIS_OPTIONS_MAXITER]);
      break;
    default:
      should_not_reach_here;
    }
    silc_object_free(x);
    return -1;
  }
  param[2].o = x;
  return 0;
}

#endif /* WITH_COCG */

int
leq_lis_dot(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  LIS_VECTOR lis_x, lis_y;
  LIS_SCALAR v;
  int i, n;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_dot(0x%lx, 0x%lx)\n", x, y);
#endif
  n = x->d.vector.length;
  lis_vector_create(n, 0, 0, &lis_x);
  lis_vector_create(n, 0, 0, &lis_y);
#ifdef ENABLE_SNATCH
  {
    LIS_SCALAR *t1 = lis_x->value; lis_x->value = x->d.vector.v.d;
    LIS_SCALAR *t2 = lis_y->value; lis_y->value = y->d.vector.v.d;
    lis_vector_dot(lis_x, lis_y, &v);
    lis_x->value = t1;
    lis_y->value = t2;
  }
#else
#pragma omp parallel for
  for (i = 0; i < n; i++) {
    lis_x->value[i] = x->d.vector.v.d[i];
    lis_y->value[i] = y->d.vector.v.d[i];
  }
  lis_vector_dot(lis_x, lis_y, &v);
#endif
  lis_vector_destroy(lis_x);
  lis_vector_destroy(lis_y);
  param[2].o = silc_object_new_double(v);
  return (param[2].o) ? 0 : -1;
}

static int
call_axpy(silc_object_t *alpha, silc_object_t *x)
{
  LIS_VECTOR lis_x, lis_y;
  int i, n;

  n = x->d.vector.length;
  lis_vector_create(n, 0, 0, &lis_x);
  lis_vector_create(n, 0, 0, &lis_y);
#ifdef ENABLE_SNATCH
  {
    LIS_SCALAR *temp = lis_x->value; lis_x->value = x->d.vector.v.d;
    lis_vector_set_all(0.0, lis_y);
    lis_vector_axpy(alpha->d.scalar.d, lis_x, lis_y);
    lis_x->value = temp;
  }
#else
#pragma omp parallel for
  for (i = 0; i < n; i++)
    lis_x->value[i] = x->d.vector.v.d[i];
  lis_vector_set_all(0.0, lis_y);
  lis_vector_axpy(alpha->d.scalar.d, lis_x, lis_y);
#endif
#pragma omp parallel for
  for (i = 0; i < n; i++)
    x->d.vector.v.d[i] = lis_y->value[i];
  lis_vector_destroy(lis_x);
  lis_vector_destroy(lis_y);
  return 0;
}

/* out: 1 */
int
leq_lis_mul_scalar_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_mul_scalar_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  return call_axpy(x, y);
}

/* out: 0 */
int
leq_lis_mul_vector_scalar(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_mul_vector_scalar(0x%lx, 0x%lx)\n", x, y);
#endif
  return call_axpy(y, x);
}

static int
call_xpay(silc_object_t *x, double alpha, silc_object_t *y)
{
  LIS_VECTOR lis_x, lis_y;
  int i, n;

  n = x->d.vector.length;
  lis_vector_create(n, 0, 0, &lis_x);
  lis_vector_create(n, 0, 0, &lis_y);
#ifdef ENABLE_SNATCH
  {
    LIS_SCALAR *t1 = lis_x->value; lis_x->value = x->d.vector.v.d;
    LIS_SCALAR *t2 = lis_y->value; lis_y->value = y->d.vector.v.d;
    lis_vector_xpay(lis_x, alpha, lis_y);
    lis_x->value = t1;
    lis_y->value = t2;
  }
#else
#pragma omp parallel for
  for (i = 0; i < n; i++) {
    lis_x->value[i] = x->d.vector.v.d[i];
    lis_y->value[i] = y->d.vector.v.d[i];
  }
  lis_vector_xpay(lis_x, alpha, lis_y);
#pragma omp parallel for
  for (i = 0; i < n; i++)
    y->d.vector.v.d[i] = lis_y->value[i];
#endif
  lis_vector_destroy(lis_x);
  lis_vector_destroy(lis_y);
  return 0;
}

/* out: 1 */
int
leq_lis_add_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *new_object;
  LIS_VECTOR lis_x, lis_y;
  int i, n;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_add_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  return call_xpay(x, 1.0, y);
}

/* out: 1 */
int
leq_lis_sub_vector_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *new_object;
  LIS_VECTOR lis_x, lis_y;
  int i, n;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_sub_vector_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  return call_xpay(x, -1.0, y);
}

static int
call_matvec(int format, silc_object_t *A, silc_object_t *x)
{
  LIS_MATRIX lis_A;
  LIS_VECTOR lis_x, lis_y;
  int i, n;

  if ((n = A->d.matrix.m) != A->d.matrix.n) {
    printf(MODULE_NAME ": call_matvec: not a square matrix\n");
    return -1;
  }
  lis_matrix_create(n, 0, 0, &lis_A);
  switch (format) {
  case LIS_MATRIX_DNS:
    {
      silc_dense_t *in;
      in = SILC_DENSE(A);
      lis_matrix_set_dns(in->v.d, lis_A);
    }
    break;
  case LIS_MATRIX_CRS:
    {
      silc_crs_t *in;
      in = SILC_CRS(A);
      lis_matrix_set_crs(in->nnz, in->row, in->index, in->v.d, lis_A);
    }
    break;
  default:
    should_not_reach_here;
  }
  lis_matrix_assemble(&lis_A, format);
  lis_vector_create(n, 0, 0, &lis_x);
  lis_vector_create(n, 0, 0, &lis_y);
#ifdef ENABLE_SNATCH
  {
    LIS_SCALAR *temp = lis_x->value; lis_x->value = x->d.vector.v.d;
    lis_matvec(lis_A, lis_x, lis_y);
    lis_x->value = temp;
  }
#else
#pragma omp parallel for
  for (i = 0; i < n; i++)
    lis_x->value[i] = x->d.vector.v.d[i];
  lis_matvec(lis_A, lis_x, lis_y);
#endif
#pragma omp parallel for
  for (i = 0; i < n; i++)
    x->d.vector.v.d[i] = lis_y->value[i];
  switch (format) {
  case LIS_MATRIX_DNS:
    lis_A->value = NULL;
    break;
  case LIS_MATRIX_CRS:
    lis_A->value = NULL;
    lis_A->ptr = lis_A->index = NULL;
    break;
  default:
    should_not_reach_here;
  }
  lis_matrix_destroy(lis_A);
  lis_vector_destroy(lis_x);
  lis_vector_destroy(lis_y);
  return 0;
}

/* out: 1 */
int
leq_lis_matvec_dense(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *x;

  A = param[0].o;
  x = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_matvec_dense(0x%lx, 0x%lx)\n", A, x);
#endif
  return call_matvec(LIS_MATRIX_DNS, A, x);
}

/* out: 1 */
int
leq_lis_matvec_crs(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *x;

  A = param[0].o;
  x = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_matvec_crs(0x%lx, 0x%lx)\n", A, x);
#endif
  return call_matvec(LIS_MATRIX_CRS, A, x);
}

/* functions */

int
leq_lis_func_norm2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x;
  LIS_VECTOR lis_x, lis_y;
  LIS_SCALAR v, dot_1, dot_2;
  int i, n;

  x = param[0].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_func_norm2(0x%lx)\n", x);
#endif
  n = x->d.vector.length;
  switch (x->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    lis_vector_create(n, 0, 0, &lis_x);
    switch (x->precision) {
    case SILC_INT:
#pragma omp parallel for
      for (i = 0; i < n; i++)
	lis_x->value[i] = x->d.vector.v.i[i];
      break;
    case SILC_LONG:
#pragma omp parallel for
      for (i = 0; i < n; i++)
	lis_x->value[i] = x->d.vector.v.l[i];
      break;
    case SILC_FLOAT:
#pragma omp parallel for
      for (i = 0; i < n; i++)
	lis_x->value[i] = x->d.vector.v.s[i];
      break;
    case SILC_DOUBLE:
#pragma omp parallel for
      for (i = 0; i < n; i++)
	lis_x->value[i] = x->d.vector.v.d[i];
      break;
    }
    lis_vector_nrm2(lis_x, &v);
    lis_vector_destroy(lis_x);
    break;
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    lis_vector_create(n, 0, 0, &lis_x);
    lis_vector_create(n, 0, 0, &lis_y);
    switch (x->precision) {
    case SILC_COMPLEX:
#pragma omp parallel for
      for (i = 0; i < n; i++) {
	lis_x->value[i] = (double)x->d.vector.v.c[i*2];
	lis_y->value[i] = (double)x->d.vector.v.c[i*2+1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
      for (i = 0; i < n; i++) {
	lis_x->value[i] = x->d.vector.v.z[i*2];
	lis_y->value[i] = x->d.vector.v.z[i*2+1];
      }
      break;
    }
    lis_vector_dot(lis_x, lis_x, &dot_1);
    lis_vector_dot(lis_y, lis_y, &dot_2);
    v = sqrt(dot_1 + dot_2);
    lis_vector_destroy(lis_x);
    lis_vector_destroy(lis_y);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  param[1].o = silc_object_new_double(v);
  return (param[1].o) ? 0 : -1;
}

int
leq_lis_get_residual_history(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  LIS_VECTOR history;
  LIS_SCALAR params[LIS_PARAMS_LEN];
  int iter, options[LIS_OPTIONS_LEN];

#ifdef DEBUG
  printf(MODULE_NAME ": leq_lis_get_residual_history()\n");
#endif
  load_options(options, params);
  if (!(options[LIS_OPTIONS_OUTPUT] == LIS_PRINT_MEM ||
	options[LIS_OPTIONS_OUTPUT] == LIS_PRINT_ALL)) {
    fprintf(stderr, MODULE_NAME ": residual history not available\n");
    return -1;
  }
  object = var_get("lis_status");
  if (!object) {
    fprintf(stderr, MODULE_NAME ": lis_status not found\n");
    return -1;
  }
  if (object->type != SILC_COLUMN_VECTOR_TYPE ||
      object->precision != SILC_DOUBLE ||
      object->d.vector.length != LIS_STATUS_LEN) {
    fprintf(stderr, MODULE_NAME ": broken lis_status\n");
    return -1;
  }
  switch ((int)object->d.vector.v.d[LIS_STATUS_RET]) {
  case LIS_SUCCESS:
    iter = (int)object->d.vector.v.d[LIS_STATUS_ITER];
    break;
  case LIS_MAXITER:
  case LIS_BREAKDOWN:
    iter = (int)object->d.vector.v.d[LIS_STATUS_ITER] - 1;
    break;
  default:
    iter = 0;
  }
  if (iter < 1) {
    fprintf(stderr, MODULE_NAME ": no residual history available\n");
    return -1;
  }
  lis_vector_create(iter, 0, 0, &history);
  lis_get_residual_history(history);
  object = silc_object_new_column_vector(SILC_DOUBLE, iter);
  memcpy(object->d.vector.v.d, history->value, sizeof(double) * iter);
  lis_vector_destroy(history);
  param[0].o = object;
  return 0;
}

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {leq_lis_solve_dense, "leq_lis_solve_dense",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
  {leq_lis_solve_dense, "leq_lis_solve_dense",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_lis_divide_dense, "leq_lis_divide_dense",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_lis_divide_dense, "leq_lis_divide_dense",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_lis_solve_crs, "leq_lis_solve_crs",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
  {leq_lis_solve_crs, "leq_lis_solve_crs",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 2, 0},
  {leq_lis_divide_crs, "leq_lis_divide_crs",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, 2, 0},
  {leq_lis_divide_crs, "leq_lis_divide_crs",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, 2, 0},
#ifdef WITH_COCG
  {leq_lis_cocg, "leq_lis_cocg",
   SILC_LEQ_OP, SILC_DOUBLE_COMPLEX,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
#endif
  {leq_lis_mul_scalar_vector, "leq_lis_mul_scalar_vector",
   SILC_MULTIPLY_OP, SILC_DOUBLE,
   SILC_SCALAR_TYPE, 0, SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {leq_lis_mul_vector_scalar, "leq_lis_mul_vector_scalar",
   SILC_MULTIPLY_OP, SILC_DOUBLE,
   SILC_COLUMN_VECTOR_TYPE, 0, SILC_SCALAR_TYPE, 0, 0, 0},
  {leq_lis_add_vector_vector, "leq_lis_add_vector_vector",
   SILC_ADD_OP, SILC_DOUBLE,
   SILC_COLUMN_VECTOR_TYPE, 0,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {leq_lis_sub_vector_vector, "leq_lis_sub_vector_vector",
   SILC_SUBTRACT_OP, SILC_DOUBLE,
   SILC_COLUMN_VECTOR_TYPE, 0,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {leq_lis_matvec_dense, "leq_lis_matvec_dense",
   SILC_MULTIPLY_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {leq_lis_matvec_crs, "leq_lis_matvec_crs",
   SILC_MULTIPLY_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {leq_lis_func_norm2, "leq_lis_func_norm2",
   "scalar norm2(vector)", 0},
  {leq_lis_get_residual_history, "leq_lis_get_residual_history",
   "vector lis_get_residual_history()", 0},
  {leq_lis_get_residual_history, "leq_lis_get_residual_history",
   "vector get_residual_history()", 0},
  {NULL} /* sentinel */
};

void
init_leq_lis()
{
  int options[LIS_OPTIONS_LEN];
  LIS_SCALAR params[LIS_PARAMS_LEN];
  char *env, *argv[] = { MODULE_NAME };

  lis_initialize(1, argv, NULL, options, params);
  env = getenv("LIS_OPTIONS");
  if (env)
    lis_set_options(env, options, params);
  load_options(options, params);

#define define_const(name, value) var_add(name, silc_object_new_int(value))

  define_const("LIS_PARAMS_RESID",	LIS_PARAMS_RESID+1);
  define_const("LIS_PARAMS_W",		LIS_PARAMS_W+1);
  define_const("LIS_PARAMS_RELAX",	LIS_PARAMS_RELAX+1);
  define_const("LIS_PARAMS_DROP",	LIS_PARAMS_DROP+1);
  define_const("LIS_PARAMS_ALPHA",	LIS_PARAMS_ALPHA+1);
  define_const("LIS_PARAMS_TAU",	LIS_PARAMS_TAU+1);
  define_const("LIS_PARAMS_SIGMA",	LIS_PARAMS_SIGMA+1);
  define_const("LIS_PARAMS_GAMMA",	LIS_PARAMS_GAMMA+1);
  define_const("LIS_PARAMS_SSOR_W",	LIS_PARAMS_SSOR_W+1);
  define_const("LIS_PARAMS_PRESID",	LIS_PARAMS_PRESID+1);
  define_const("LIS_PARAMS_PW",		LIS_PARAMS_PW+1);
  define_const("LIS_OPTIONS_SOLVER",	LIS_OPTIONS_SOLVER+1);
  define_const("LIS_OPTIONS_PRECON",	LIS_OPTIONS_PRECON+1);
  define_const("LIS_OPTIONS_MAXITER",	LIS_OPTIONS_MAXITER+1);
  define_const("LIS_OPTIONS_OUTPUT",	LIS_OPTIONS_OUTPUT+1);
  define_const("LIS_OPTIONS_RESTART",	LIS_OPTIONS_RESTART+1);
  define_const("LIS_OPTIONS_ELL",	LIS_OPTIONS_ELL+1);
  define_const("LIS_OPTIONS_SCALE",	LIS_OPTIONS_SCALE+1);
  define_const("LIS_OPTIONS_FILL",	LIS_OPTIONS_FILL+1);
  define_const("LIS_OPTIONS_M",		LIS_OPTIONS_M+1);
  define_const("LIS_OPTIONS_PSOLVER",	LIS_OPTIONS_PSOLVER+1);
  define_const("LIS_OPTIONS_PMAXITER",	LIS_OPTIONS_PMAXITER+1);
  define_const("LIS_OPTIONS_PRESTART",	LIS_OPTIONS_PRESTART+1);
  define_const("LIS_OPTIONS_PELL",	LIS_OPTIONS_PELL+1);
  define_const("LIS_OPTIONS_PADDING",	LIS_OPTIONS_PADDING+1);
  define_const("LIS_OPTIONS_ISLEVEL",	LIS_OPTIONS_ISLEVEL+1);
  define_const("LIS_STATUS_RET",	LIS_STATUS_RET+1);
  define_const("LIS_STATUS_ITER",	LIS_STATUS_ITER+1);
  define_const("LIS_STATUS_RESID",	LIS_STATUS_RESID+1);
  define_const("LIS_STATUS_TIMES",	LIS_STATUS_TIMES+1);
  define_const("LIS_STATUS_PTIMES",	LIS_STATUS_PTIMES+1);
  define_const("LIS_STATUS_ITIMES",	LIS_STATUS_ITIMES+1);
  define_const("LIS_SOLVER_CG",		LIS_SOLVER_CG);
  define_const("LIS_SOLVER_BICG",	LIS_SOLVER_BICG);
  define_const("LIS_SOLVER_CGS",	LIS_SOLVER_CGS);
  define_const("LIS_SOLVER_BICGSTAB",	LIS_SOLVER_BICGSTAB);
  define_const("LIS_SOLVER_BICGSTABL",	LIS_SOLVER_BICGSTABL);
  define_const("LIS_SOLVER_GPBICG",	LIS_SOLVER_GPBICG);
  define_const("LIS_SOLVER_ORTHOMIN",	LIS_SOLVER_ORTHOMIN);
  define_const("LIS_SOLVER_GMRES",	LIS_SOLVER_GMRES);
  define_const("LIS_SOLVER_QMR",	LIS_SOLVER_QMR);
  define_const("LIS_SOLVER_JACOBI",	LIS_SOLVER_JACOBI);
  define_const("LIS_SOLVER_GS",		LIS_SOLVER_GS);
  define_const("LIS_SOLVER_SOR",	LIS_SOLVER_SOR);
  define_const("LIS_PRECON_TYPE_NONE",	LIS_PRECON_TYPE_NONE);
  define_const("LIS_PRECON_TYPE_JACOBI",LIS_PRECON_TYPE_JACOBI);
  define_const("LIS_PRECON_TYPE_ILU",	LIS_PRECON_TYPE_ILU);
  define_const("LIS_PRECON_TYPE_SSOR",	LIS_PRECON_TYPE_SSOR);
  define_const("LIS_PRECON_TYPE_HYBRID",LIS_PRECON_TYPE_HYBRID);
  define_const("LIS_PRECON_TYPE_IS",	LIS_PRECON_TYPE_IS);
  define_const("LIS_PRECON_TYPE_SAI",	LIS_PRECON_TYPE_SAI);
  define_const("LIS_PRECON_TYPE_SAAMG",	LIS_PRECON_TYPE_SAAMG);
  define_const("LIS_PRECON_TYPE_RIC2S",	LIS_PRECON_TYPE_RIC2S);
  define_const("LIS_SCALE_NONE",	LIS_SCALE_NONE);
  define_const("LIS_SCALE_JACOBI",	LIS_SCALE_JACOBI);
  define_const("LIS_SCALE_SYMM_DIAG",	LIS_SCALE_SYMM_DIAG);
  define_const("LIS_PRINT_NONE",	LIS_PRINT_NONE);
  define_const("LIS_PRINT_MEM",		LIS_PRINT_MEM);
  define_const("LIS_PRINT_OUT",		LIS_PRINT_OUT);
  define_const("LIS_PRINT_ALL",		LIS_PRINT_ALL);

  silc_register_module(MODULE_NAME,
		       NULL,   /* _unary_op_table */
		       _binary_op_table,
		       _function_table,
		       NULL);  /* _procedure_table */
}
