/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: leq_lu.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "module.h"
#include "object.h"
#include "dense.h"
#include "sparse_crs.h"

#define MODULE_NAME "leq_lu"

static int
solve_1(silc_object_t *A, silc_object_t *b)
{
  silc_object_t *x;
  silc_dense_t *A_in;
  silc_crs_t *in;
  double *dv;
  size_t m, n;
  int i, j, k, ii, jj, kk;
  const char *format_name;

#ifdef DEBUG
  printf(MODULE_NAME ": solve_1(0x%lx, 0x%lx)\n", A, b);
#endif
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return -1;
  }
  if (A->d.matrix.m != b->d.vector.length) {
    fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
    return -1;
  }
  A = silc_object_duplicate(A, 1);
  if (!A)
    return -1;
  format_name = silc_find_format_name(A->format);
  if (strcmp(format_name, SILC_FORMAT_DENSE) == 0) {
    A_in = SILC_DENSE(A);
  } else if (strcmp(format_name, SILC_FORMAT_CRS) == 0) {
    in = SILC_CRS(A);
    dv = (double *)calloc(A->d.matrix.m * A->d.matrix.n, sizeof(double));
    for (i = 0; i < A->d.matrix.m; i++)
      for (j = in->row[i]; j < in->row[i + 1]; j++)
	dv[i + in->index[j] * A->d.matrix.m] = in->v.d[j];
    free(in->v.d);
    free(in->index);
    free(in->row);
    A->type = SILC_MATRIX_TYPE;
    A->format = silc_find_format_id(SILC_FORMAT_DENSE);
    A_in = SILC_DENSE(A);
    A_in->v.d = dv;
  }
  m = A->d.matrix.m;
  n = A->d.matrix.n;
#if 0
  /* dumb implementation */
  for (i = 0; i < m; i++) {
    for (j = i + 1; j < n; j++) {
      if ((int)A_in->v.d[i + m * i] == 0) {
	silc_object_free(A);
	return -1;
      }
      A_in->v.d[j + m * i] /=
	A_in->v.d[i + m * i];
      for (k = i + 1; k < n; k++) {
	A_in->v.d[j + m * k] -= A_in->v.d[i + m * k] * A_in->v.d[j + m * i];
      }
    }
  }
  for (i = 0; i < m; i++)
    for (j = i + 1; j < n; j++)
      b->d.vector.v.d[j] -=
	b->d.vector.v.d[i] * A_in->v.d[j + m * i];
  for (i = m - 1; i >= 0; i--) {
    for (j = i + 1; j < n; j++)
      b->d.vector.v.d[i] -=
	A_in->v.d[i + m * j] * b->d.vector.v.d[j];
    b->d.vector.v.d[i] /= A_in->v.d[i + m * i];
  }    
#else
  /* slightly optimized */
  for (i = ii = 0; i < m; i++, ii += m) {
    for (j = i + 1; j < n; j++) {
      if (A_in->v.d[i + ii] == 0.0) {
	silc_object_free(A);
	return -1;
      }
      A_in->v.d[j + ii] /= A_in->v.d[i + ii];
      for (k = i + 1, kk = m * k; k < n; k++, kk += m) {
	A_in->v.d[j + kk] -= A_in->v.d[i + kk] * A_in->v.d[j + ii];
      }
    }
  }
  for (i = ii = 0; i < m; i++, ii += m)
    for (j = i + 1; j < n; j++)
      b->d.vector.v.d[j] -= b->d.vector.v.d[i] * A_in->v.d[j + ii];
  for (i = m - 1; i >= 0; i--) {
    for (j = i + 1, jj = m * j; j < n; j++, jj += m)
      b->d.vector.v.d[i] -= A_in->v.d[i + jj] * b->d.vector.v.d[j];
    b->d.vector.v.d[i] /= A_in->v.d[i + m * i];
  }    
#endif
  silc_object_free(A);
  return 0;
}

/* out: 1 */
int
leq_lu_solve(int n_params, silc_mfunc_param_t *param)
{
  return solve_1(param[0].o, param[1].o);
}

/* out: 0 */
int
leq_lu_divide(int n_params, silc_mfunc_param_t *param)
{
  return solve_1(param[1].o, param[0].o);
}

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {leq_lu_solve, "leq_lu_solve",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {leq_lu_solve, "leq_lu_solve",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 1, 0},
  {leq_lu_divide, "leq_lu_divide",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, 0, 0},
  {leq_lu_divide, "leq_lu_divide",
   SILC_DIVIDE_OP, SILC_DOUBLE,
   SILC_ROW_VECTOR_TYPE, 0,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, 0, 0},
  {NULL} /* sentinel */
};

static procedure_table_t _procedure_table[] = {
  /* name, sig_text, flags */
  {leq_lu_solve, "leq_lu_solve", "leq_lu(matrix in, vector inout)", 0},
  {NULL} /* sentinel */
};

void
init_leq_lu()
{
  silc_register_module(MODULE_NAME,
		       NULL,  /* _unary_op_table */
		       _binary_op_table,
		       NULL,  /* _function_table */
		       _procedure_table);
}
