/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: leq_mp.c,v 1.3 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>

#include "module.h"
#include "object.h"
#include "sparse_crs.h"
#include "variable.h"
#include "wtime.h"

#define MODULE_NAME "leq_mp"

#include "mp_crs.h"

int mp_bicg(mp_crs_t *, mp_vector_t *, mp_vector_t *, int, mpf_t, int);
int mp_cg(mp_crs_t *, mp_vector_t *, mp_vector_t *, int, mpf_t, int);
int mp_cgnr(mp_crs_t *, mp_vector_t *, mp_vector_t *, int, mpf_t, int);

#define MP_OPTIONS_SOLVER   1
#define MP_OPTIONS_PREC     2
#define MP_OPTIONS_MAXITER  3
#define MP_OPTIONS_LEN      4

#define MP_PARAMS_RESID     1
#define MP_PARAMS_LEN       2

#define MP_STATUS_TIME      1
#define MP_STATUS_ITER      2
#define MP_STATUS_RESID     3
#define MP_STATUS_LEN       4

#define MP_SOLVER_BICG      1
#define MP_SOLVER_CG        2
#define MP_SOLVER_CGNR      3

static void
load_options(int options[], double params[])
{
  silc_object_t *object;
  int i;

  object = var_get("mp_options");
  if (!object) {
    object = silc_object_new_column_vector(SILC_INT, MP_OPTIONS_LEN);
    object->d.vector.v.i[0] = 0; /* reserved */
    for (i = 1; i < MP_OPTIONS_LEN; i++)
      object->d.vector.v.i[i] = options[i];
    var_add("mp_options", object);
  }
  if (object->type == SILC_COLUMN_VECTOR_TYPE &&
      object->precision == SILC_INT &&
      object->d.vector.length == MP_OPTIONS_LEN) {
    for (i = 1; i < MP_OPTIONS_LEN; i++)
      options[i] = object->d.vector.v.i[i];
  } else {
    fprintf(stderr, MODULE_NAME ": broken mp_options (ignored)\n");
  }
  object = var_get("mp_params");
  if (!object) {
    object = silc_object_new_column_vector(SILC_DOUBLE, MP_PARAMS_LEN);
    object->d.vector.v.d[0] = 0; /* reserved */
    for (i = 1; i < MP_PARAMS_LEN; i++)
      object->d.vector.v.d[i] = params[i];
    var_add("mp_params", object);
  }
  if (object->type == SILC_COLUMN_VECTOR_TYPE &&
      object->precision == SILC_DOUBLE &&
      object->d.vector.length == MP_PARAMS_LEN) {
    for (i = 1; i < MP_PARAMS_LEN; i++)
      params[i] = object->d.vector.v.d[i];
  } else {
    fprintf(stderr, MODULE_NAME ": broken mp_params (ignored)\n");
  }
}

static void
store_status(double status[])
{
  silc_object_t *object;
  int i;

  object = var_get("mp_status");
  if (!object) {
    object = silc_object_new_column_vector(SILC_DOUBLE, MP_STATUS_LEN);
    object->d.vector.v.d[0] = 0; /* reserved */
    for (i = 1; i < MP_STATUS_LEN; i++)
      object->d.vector.v.d[i] = 0;
    var_add("mp_status", object);
  }
  if (object->type == SILC_COLUMN_VECTOR_TYPE &&
      object->precision == SILC_DOUBLE &&
      object->d.vector.length == MP_STATUS_LEN) {
    for (i = 1; i < MP_STATUS_LEN; i++)
      object->d.vector.v.d[i] = status[i];
  } else {
    fprintf(stderr, MODULE_NAME ": broken leq_mp_status (ignored)\n");
  }
}

int
leq_mp_solve(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;
  silc_crs_t *in;
  mp_crs_t *mp_A;
  mp_vector_t *mp_b, *mp_x;
  mpf_t mp_epsilon, mp_tmp, mp_tmp2, mp_bnrm2;
  int options[MP_OPTIONS_LEN];
  double params[MP_PARAMS_LEN], status[MP_STATUS_LEN];
  double residual, t1, t2;
  int i, n, prec, solver, maxiter, iter, verbose;

  A = param[0].o;
  b = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": leq_mp_solve(0x%lx, 0x%lx)\n", A, b);
#endif
  n = A->d.matrix.n;
  if (n != A->d.matrix.m) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return -1;
  }
  if (n != b->d.vector.length) {
    fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
    return -1;
  }
  if (A->precision != SILC_DOUBLE || b->precision != SILC_DOUBLE) {
    fprintf(stderr, MODULE_NAME ": unsupported precision\n");
    return -1;
  }
  if (strcmp(silc_find_format_name(A->format), SILC_FORMAT_CRS)) {
    fprintf(stderr, MODULE_NAME ": unsupported matrix storage format\n");
    return -1;
  }
  load_options(options, params);
  solver = options[MP_OPTIONS_SOLVER];
  switch (solver) {
  case MP_SOLVER_BICG:
  case MP_SOLVER_CG:
  case MP_SOLVER_CGNR:
#ifdef DEBUG
    printf(MODULE_NAME ": solver = %s\n",
	   (solver == MP_SOLVER_BICG) ? "BiCG" :
	   (solver == MP_SOLVER_CG)   ? "CG" :
	   (solver == MP_SOLVER_CGNR) ? "CGNR" : "unknown");
#endif
    break;
  default:
    fprintf(stderr, MODULE_NAME ": unknown solver: %d\n", solver);
    return -1;
  }
  prec = options[MP_OPTIONS_PREC];
  if (prec != mpf_get_default_prec()) {
    mp_free_residual_history();
    mpf_set_default_prec(prec);
  }
  maxiter = options[MP_OPTIONS_MAXITER];
  if (mp_init_residual_history(maxiter) < 0) {
    fprintf(stderr, MODULE_NAME ": mp_init_residual_history() failed\n");
    return -1;
  }
  mpf_init_set_d(mp_epsilon, params[MP_PARAMS_RESID]);
#ifdef DEBUG
  verbose = 1;
#else
  verbose = 0;
#endif  

  in = SILC_CRS(A);
  mp_A = mp_crs_new(n, n, in->nnz);
  mp_b = mp_vector_new(n);
  mp_x = mp_vector_new(n);
#pragma omp parallel for
  for (i = 0; i < n; i++) {
    int j;
    mp_crs_row(mp_A, i) = in->row[i];
    for (j = in->row[i]; j < in->row[i+1]; j++) {
      mp_crs_index(mp_A, j) = in->index[j];
      mpf_set_d(mp_crs_value(mp_A, j), in->v.d[j]);
    }
    mpf_set_d(mp_vector_value(mp_b, i), b->d.vector.v.d[i]);
    mpf_set_d(mp_vector_value(mp_x, i), 0);
  }
  mp_crs_row(mp_A, n) = in->nnz;

  t1 = silc_wtime();
  switch (solver) {
  case MP_SOLVER_BICG:
    iter = mp_bicg(mp_A, mp_b, mp_x, maxiter, mp_epsilon, verbose);
    break;
  case MP_SOLVER_CG:
    iter = mp_cg(mp_A, mp_b, mp_x, maxiter, mp_epsilon, verbose);
    break;
  case MP_SOLVER_CGNR:
    iter = mp_cgnr(mp_A, mp_b, mp_x, maxiter, mp_epsilon, verbose);
    break;
  default:
    fprintf(stderr, MODULE_NAME ": unknown solver: %d\n", solver);
  }
  t2 = silc_wtime();

  status[MP_STATUS_TIME] = t2 - t1;
  status[MP_STATUS_ITER] = iter;
  mpf_init_set_d(mp_tmp, -1);
  mpf_init_set_d(mp_tmp2, 1);
  mpf_init(mp_bnrm2);
  mp_vector_norm2(mp_bnrm2, mp_b);
  mp_crs_matvec(mp_tmp, mp_A, mp_x, mp_tmp2, mp_b);
  mp_vector_norm2(mp_tmp2, mp_b);
  residual = mpf_get_d(mp_tmp2);
  mpf_div(mp_tmp, mp_tmp2, mp_bnrm2);
  status[MP_STATUS_RESID] = mpf_get_d(mp_tmp);
  mp_crs_free(mp_A);
  mp_vector_free(mp_b);
  mpf_clear(mp_epsilon);
  mpf_clear(mp_tmp);
  mpf_clear(mp_tmp2);
  mpf_clear(mp_bnrm2);
#ifdef DEBUG
  printf(MODULE_NAME ": %e sec\n", status[MP_STATUS_TIME]);
  printf(MODULE_NAME ": #iters = %ld\n", (int)status[MP_STATUS_ITER]);
  printf(MODULE_NAME ": residual = %e (abs.), %e (rel.)\n",
	 residual, status[MP_STATUS_RESID]);
#endif
  store_status(status);
  if (iter > maxiter) {
    fprintf(stderr, MODULE_NAME ": reached to the maximum iteration count\n");
    mp_vector_free(mp_x);
    return -1;
  }
  x = silc_object_duplicate(b, 0);
#pragma omp parallel for
  for (i = 0; i < n; i++)
    x->d.vector.v.d[i] = mpf_get_d(mp_vector_value(mp_x, i));
  mp_vector_free(mp_x);
  param[2].o = x;
  return 0;
}

int
leq_mp_get_residual_history(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  int i, iter, options[MP_OPTIONS_LEN];
  double params[MP_PARAMS_LEN];

#ifdef DEBUG
  printf(MODULE_NAME ": leq_mp_get_residual_history()\n");
#endif
  object = var_get("mp_status");
  if (!object) {
    fprintf(stderr, MODULE_NAME ": mp_status not found\n");
    return -1;
  }
  if (object->type != SILC_COLUMN_VECTOR_TYPE ||
      object->precision != SILC_DOUBLE ||
      object->d.vector.length != MP_STATUS_LEN) {
    fprintf(stderr, MODULE_NAME ": broken mp_status\n");
    return -1;
  }
  iter = (int)object->d.vector.v.d[MP_STATUS_ITER];
  if (iter < 1) {
    fprintf(stderr, MODULE_NAME ": no residual history available\n");
    return -1;
  }
  load_options(options, params);
  if (iter > options[MP_OPTIONS_MAXITER]) {
    iter = options[MP_OPTIONS_MAXITER];
  }
  object = silc_object_new_column_vector(SILC_DOUBLE, iter);
  if (!object) {
    fprintf(stderr, MODULE_NAME ": out of memory\n");
    return -1;
  }
  for (i = 0; i < iter; i++)
    object->d.vector.v.d[i] =
      mpf_get_d(mp_vector_value(mp_residual_history, i));
  param[0].o = object;
  return 0;
}

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {leq_mp_solve, "leq_mp_solve",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {leq_mp_get_residual_history, "leq_mp_get_residual_history",
   "vector get_residual_history()", 0},
  {NULL} /* sentinel */
};

void
init_leq_mp()
{
  int options[MP_OPTIONS_LEN];
  double params[MP_PARAMS_LEN];

  options[MP_OPTIONS_SOLVER] = MP_SOLVER_BICG;
  options[MP_OPTIONS_PREC] = 64;
  options[MP_OPTIONS_MAXITER] = 1000;
  params[MP_PARAMS_RESID] = 1.0e-12;
  load_options(options, params);

#define define_const(name, value) var_add(name, silc_object_new_int(value))

  define_const("MP_OPTIONS_SOLVER",  MP_OPTIONS_SOLVER+1);
  define_const("MP_OPTIONS_PREC",    MP_OPTIONS_PREC+1);
  define_const("MP_OPTIONS_MAXITER", MP_OPTIONS_MAXITER+1);
  define_const("MP_PARAMSS_RESID",   MP_PARAMS_RESID+1);
  define_const("MP_STATUS_TIME",     MP_STATUS_TIME+1);
  define_const("MP_STATUS_ITER",     MP_STATUS_ITER+1);
  define_const("MP_STATUS_RESID",    MP_STATUS_RESID+1);
  define_const("MP_SOLVER_BICG",     MP_SOLVER_BICG);
  define_const("MP_SOLVER_CG",       MP_SOLVER_CG);
  define_const("MP_SOLVER_CGNR",     MP_SOLVER_CGNR);

  silc_register_module(MODULE_NAME,
		       NULL,   /* _unary_op_table */
		       _binary_op_table,
		       _function_table,
		       NULL);  /* _procedure_table */
}
