/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: leq_scalapack.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

#include "PBblacs.h" /* BLACS (part of ScaLAPACK) */

#include "module.h"
#include "dense.h"

#define MODULE_NAME "leq_scalapack"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#define BLACS_BUF(x) ((char *)x)

#define _ceil(x, y) (((x) + (y) - 1) / (y))

typedef struct {
  /* local array */
  int nrows, ncols, ld; /* leading dimension */
  double *value;
  /* data distribution */
  int np_row, np_col; /* process grid */
  int bf_row, bf_col; /* blocking factors */
  int src_row, src_col;
} local_array_t;

static void
distribute_matrix(int context, int row, int column,
		  int m, int n, double *value,
		  int np_row, int np_col,
		  int bf_row, int bf_col,
		  local_array_t *loc)
{
  int nblocks_row, extra_row, nrows, max_nrows;
  int nblocks_col, extra_col, ncols, max_ncols;
  int i, j, k, ii, jj, kk, vi, vj;
  double *loc_value;

  max_nrows = _ceil(_ceil(m, bf_row), np_row) * bf_row;
  max_ncols = _ceil(_ceil(n, bf_col), np_col) * bf_col;
#ifdef DEBUG
  if (row == 0 && column == 0)
    printf("distribute: m=%d, n=%d, max_nrows=%d, max_ncols=%d\n",
	   m, n, max_nrows, max_ncols);
#endif

  nblocks_row = m / bf_row; extra_row = nblocks_row % np_row;
  nblocks_col = n / bf_col; extra_col = nblocks_col % np_col;
  
  for (i = 0; i < np_row; i++) { /* for each row in the process grid */

    /* number of rows in the local array for process (i, j) */
    nrows = nblocks_row / np_row * bf_row;
    if (i < extra_row)
      nrows += bf_row;
    else if (i == extra_row)
      nrows += m % bf_row;

    for (j = 0; j < np_col; j++) { /* for each process in the process row */

      /* number of columns in the local array for process (i, j) */
      ncols = nblocks_col / np_col * bf_col;
      if (j < extra_col)
	ncols += bf_col;
      else if (j == extra_col)
	ncols += n % bf_col;

      /* create the local array for process (i, j) */
      if (row == 0 && column == 0 ||
	  row == i && column == j) {
	loc_value = (double *)malloc(sizeof(double) * max_nrows * max_ncols);
      }
      if (row == 0 && column == 0) {
	vj = bf_col * j;
	for (jj = 0; jj < ncols; jj++) {
	  kk = max_nrows * jj;
	  vi = bf_row * i;
	  for (ii = 0; ii < nrows; ii++) {
	    loc_value[kk++] = value[vi + m * vj];
	    vi ++;
	    if (vi % bf_row == 0)
	      vi += bf_row * (np_row - 1);
	  }
	  vj ++;
	  if (vj % bf_col == 0)
	    vj += bf_col * (np_col - 1);
	}
      }

      /* transfer the local array to process (i, j) */
      if (i == 0 && j == 0) {
	/* no need for data communications */
      } else if (row == 0 && column == 0) {
	Cdgesd2d(context, nrows, ncols, BLACS_BUF(loc_value), max_nrows, i, j);
      } else if (row == i && column == j) {
	Cdgerv2d(context, nrows, ncols, BLACS_BUF(loc_value), max_nrows, 0, 0);
      }
      if (row == i && column == j) {
	loc->nrows = nrows;
	loc->ncols = ncols;
	loc->ld = max_nrows;
	loc->value = loc_value;
	loc->np_row = np_row;
	loc->np_col = np_col;
	loc->bf_row = bf_row;
	loc->bf_col = bf_col;
	loc->src_row = 0;
	loc->src_col = 0;
#ifdef DEBUG
	printf("(%d, %d) nrows=%d, ncols=%d\n", i, j, nrows, ncols);
	for (jj = 0; jj < ncols; jj++) {
	  kk = max_nrows * jj;
	  for (ii = 0; ii < nrows; ii++) {
	    printf("(%d, %d) loc_value[%d, %d] = %.1f\n", i, j, ii, jj,
		   loc_value[kk++]);
	  }
	}
#endif
      }
#ifdef DEBUG
      fflush(stdout);
      Cblacs_barrier(context, "All");
#endif
    }
  }
}

static void
collect_matrix(int context, int row, int column,
	       int m, int n, double *value,
	       int np_row, int np_col,
	       int bf_row, int bf_col,
	       local_array_t *loc)
{
  int nblocks_row, extra_row, nrows, max_nrows;
  int nblocks_col, extra_col, ncols, max_ncols;
  int i, j, k, ii, jj, kk, vi, vj;
  double *loc_value, *buffer;

  max_nrows = _ceil(_ceil(m, bf_row), np_row) * bf_row;
  max_ncols = _ceil(_ceil(n, bf_col), np_col) * bf_col;
  if (row == 0 && column == 0) {
#ifdef DEBUG
    printf("collect: max_nrows=%d, max_ncols=%d\n", max_nrows, max_ncols);
#endif
    buffer = (double *)malloc(sizeof(double) * max_nrows * max_ncols);
  } else {
    buffer = loc->value;
  }

  nblocks_row = m / bf_row; extra_row = nblocks_row % np_row;
  nblocks_col = n / bf_col; extra_col = nblocks_col % np_col;
  
  for (i = 0; i < np_row; i++) { /* for each row in the process grid */

    /* number of rows in the local array for process (i, j) */
    nrows = nblocks_row / np_row * bf_row;
    if (i < extra_row)
      nrows += bf_row;
    else if (i == extra_row)
      nrows += m % bf_row;

    for (j = 0; j < np_col; j++) { /* for each process in the process row */

      /* number of columns in the local array for process (i, j) */
      ncols = nblocks_col / np_col * bf_col;
      if (j < extra_col)
	ncols += bf_col;
      else if (j == extra_col)
	ncols += n % bf_col;

#ifdef DEBUG
      if (row == i && column == j)
	printf("(%d, %d) nrows = %d, ncols = %d\n",
	       row, column, nrows, ncols);
#endif

      /* transfer the local array to process (0, 0) */
      if (i == 0 && j == 0) {
	/* no need for data communications */
      } else if (row == 0 && column == 0) {
	Cdgerv2d(context, nrows, ncols, BLACS_BUF(buffer), max_nrows, i, j);
      } else if (row == i && column == j) {
	Cdgesd2d(context, nrows, ncols, BLACS_BUF(buffer), max_nrows, 0, 0);
      }

      /* unpack the local array */
      if (row == 0 && column == 0) {
	if (i == 0 && j == 0)
	  loc_value = loc->value;
	else
	  loc_value = buffer;
#ifdef DEBUG
	for (jj = 0; jj < ncols; jj++) {
	  kk = max_nrows * jj;
	  for (ii = 0; ii < nrows; ii++) {
	    printf("(%d, %d) loc_value[%d, %d] = %.1f\n", i, j, ii, jj,
		   loc_value[kk++]);
	  }
	}
#endif
	vj = bf_col * j;
	for (jj = 0; jj < ncols; jj++) {
	  kk = max_nrows * jj;
	  vi = bf_row * i;
	  for (ii = 0; ii < nrows; ii++) {
	    value[vi + m * vj] = loc_value[kk++];
	    vi ++;
	    if (vi % bf_row == 0)
	      vi += bf_row * (np_row - 1);
	  }
	  vj ++;
	  if (vj % bf_col == 0)
	    vj += bf_col * (np_col - 1);
	}
      }
#ifdef DEBUG
      fflush(stdout);
      Cblacs_barrier(context, "All");
#endif
    }
  }
  if (row == 0 && column == 0)
    free(buffer);
}

static int
call_pdgesv(int context, int n, int nrhs,
	    int np_row, int np_col, int bf_row, int bf_col,
	    local_array_t *loc_A, local_array_t *loc_b)
{
  int desc_A[9], desc_b[9], info, i, j, max_nrows, *ipiv;
  /* TOOLS/descinit.f */
  void descinit_(int *, int *, int *, int *, int *, int *, int *, int *,
		 int *, int *);
  /* SRC/{pdgesv,pdgetrf,pdgetrs}.f */
  void pdgesv_(int *, int *, double *, int *, int *, int *, int *, double *,
	       int *, int *, int *, int *);
  void pdgetrf_(int *, int *, double *, int *, int *, int *, int *, int *);
  void pdgetrs_(char *, int *, int *, double *, int *, int *, int *, int *,
		double *, int *, int *, int *, int *);

  i = j = 1;
  max_nrows = _ceil(_ceil(n, bf_row), np_row) * bf_row;
  ipiv = (int *)malloc(sizeof(int) * (max_nrows + bf_row));
  descinit_(desc_A, &n, &n,
	    &bf_row, &bf_col, &loc_A->src_row, &loc_A->src_col, &context,
	    &loc_A->ld, &info);
  descinit_(desc_b, &n, &nrhs,
	    &bf_row, &bf_col, &loc_b->src_row, &loc_b->src_col, &context,
	    &loc_b->ld, &info);
#if 1
  pdgesv_(&n, &nrhs,
	  loc_A->value, &i, &j, desc_A, ipiv,
	  loc_b->value, &i, &j, desc_b, &info);
#else
  pdgetrf_(&n, &n, loc_A->value, &i, &j, desc_A, ipiv, &info);
  if (info == 0) {
    char trans = 'N';
    pdgetrs_(&trans, &n, &nrhs,
	     loc_A->value, &i, &j, desc_A, ipiv,
	     loc_b->value, &i, &j, desc_b, &info);
  }
#endif
  free(ipiv);
  return info;
}

static silc_object_t *
solve_1(silc_object_t *A, silc_object_t *b)
{
  silc_object_t *x;
  double *A_value, *b_value, *x_value;
  int NPROW = 2, NPCOL = 3; /* process grid */
  int BFROW = 2, BFCOL = 2; /* blocking factors (i.e. block size) */
  int context, n_rows, n_cols, row, column, n, nrhs, info[2], status;
  local_array_t loc_A, loc_b;

  Cblacs_get(-1, 0, &context); /* get the default system context */
  Cblacs_gridinit(&context, "Row-major", NPROW, NPCOL);
  Cblacs_gridinfo(context, &n_rows, &n_cols, &row, &column);
  if (row < 0) /* not in the process grid */
    return NULL;
  if (row == 0 && column == 0) {
#ifdef DEBUG
    printf("BLACS context = %d\n", context);
#endif
    A_value = SILC_DENSE(A)->v.d;
    n = A->d.matrix.n;
    if (A->d.matrix.m != n)
      should_not_reach_here;
    x = silc_object_duplicate(b, 0);
    switch (b->type) {
    case SILC_ROW_VECTOR_TYPE:
    case SILC_COLUMN_VECTOR_TYPE:
      if (b->d.vector.length != n)
	should_not_reach_here;
      b_value = b->d.vector.v.d;
      x_value = x->d.vector.v.d;
      nrhs = 1;
      break;
    case SILC_MATRIX_TYPE:
      if (b->d.matrix.m != n)
	should_not_reach_here;
      b_value = SILC_DENSE(b)->v.d;
      x_value = SILC_DENSE(x)->v.d;
      nrhs = b->d.matrix.n;
      break;
    default:
      should_not_reach_here;
    }
    info[0] = n;
    info[1] = nrhs;
    Cigebs2d(context, "All", " ", 2, 1, BLACS_BUF(info), 2);
  } else {
    Cigebr2d(context, "All", " ", 2, 1, BLACS_BUF(info), 2, 0, 0);
    n    = info[0];
    nrhs = info[1];
  }
#ifdef DEBUG
  printf("(%d, %d) n=%d, nrhs=%d\n", row, column, n, nrhs);
  fflush(stdout);
  Cblacs_barrier(context, "All");
#endif  
  distribute_matrix(context, row, column, n, n,
		    A_value, NPROW, NPCOL, BFROW, BFCOL, &loc_A);
  distribute_matrix(context, row, column, n, nrhs,
		    b_value, NPROW, NPCOL, BFROW, BFCOL, &loc_b);
  status = call_pdgesv(context, n, nrhs,
		       NPROW, NPCOL, BFROW, BFCOL, &loc_A, &loc_b);
#ifdef DEBUG
  if (row == 0 && column == 0)
    printf("solve_1: status = %d\n", status);
#endif
  collect_matrix(context, row, column, n, nrhs,
		 x_value, NPROW, NPCOL, BFROW, BFCOL, &loc_b);
  free(loc_A.value);
  free(loc_b.value);
  Cblacs_gridexit(context);

  return x;
}

int
leq_scalapack_solve(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;

  if (silc_ipc_master()) {
    A = param[0].o;
    b = param[1].o;
    x = solve_1(A, b);
    if (!x)
      return -1;
    param[2].o = x;
  } else {
    solve_1(NULL, NULL);
  }
  return 0;
}

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {leq_scalapack_solve, "leq_scalapack_solve",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   SILC_COLUMN_VECTOR_TYPE, 0,
   2, SILC_MPI_MFUNC},
  {NULL} /* sentinel */
};

void
init_leq_scalapack()
{
  int proc_num, n_procs;

  Cblacs_pinfo(&proc_num, &n_procs); /* initialize BLACS */

  silc_register_module(MODULE_NAME,
		       NULL,   /* _unary_op_table */
		       _binary_op_table,
		       NULL,   /* _function_table */
		       NULL);  /* _procedure_table */
}
