/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: leq_smsamg.c,v 1.3 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>

#include "module.h"
#include "object.h"
#include "sparse_crs.h"
#include "wtime.h"

#define MODULE_NAME "leq_smsamg"

extern int SMSAMG3D(double *, double *, int *, double *, double *,
		    double *, int *, int *, int *, int *, int, int,
		    int, int, double, int, int, int, double, double,
		    int, int);

#ifndef MSTP
#define MSTP 100 /* maximum number of iterations */
#endif

#ifndef EPS
#define EPS 1.0e-12 /* convergence criteria */
#endif

#ifndef LEPSSW
#define LEPSSW 1 /* check convergence using relative residual 2-norm */
#endif

#ifndef LOP1
#define LOP1 1 /* no scaling */
#endif

#ifndef RTO
#define RTO 0.2
#endif

#ifndef OMG
#define OMG 1.0
#endif

#ifndef METHOD
#define METHOD 0 /* Residual Cutting (RC) method */
#endif

static int
call_smsamg3d(silc_object_t *matrix_A, /* SILC_CRS */
	      silc_object_t *vector_B,
	      silc_object_t *vector_X)
{
  silc_crs_t *in;
  double *AD, *ALU, *B, *X, ABRS;
  int ND, NS, NSTP, *LND, *LNT, *LPV, *LPM, RTC;
  int i, j, k;
  double t, elapsed;

  in = SILC_CRS(matrix_A);
  ND = matrix_A->d.matrix.n; /* dimension */
  NS = in->nnz - ND; /* number of non-zero non-diagonal elements */

  /* coefficient matrix */
  AD = (double *)malloc(sizeof(double) * ND);
  ALU = (double *)malloc(sizeof(double) * NS);
  LNT = (int *)malloc(sizeof(int) * NS);
  LND = (int *)malloc(sizeof(int) * ND);
  k = 0;
  for (i = 0; i < ND; i++) {
    for (j = in->row[i]; j < in->row[i+1]; j++) {
      if (in->index[j] == i) {
	AD[i] = in->v.d[j]; /* diagonal element */
      } else {
	ALU[k] = in->v.d[j]; /* non-zero non-diagonal element */
	LNT[k] = in->index[j] + 1; /* column index (1 origin) */
	++k;
      }
    }
    /* number of non-zero non-diagonal elements per row */
    LND[i] = in->row[i+1] - in->row[i] - 1;
  }

  /* right-hand side */
  B = (double *)malloc(sizeof(double) * ND);
  for (i = 0; i < ND; i++)
    B[i] = vector_B->d.vector.v.d[i];

  /* solution */
  X = vector_X->d.vector.v.d;
  for (i = 0; i < ND; i++)
    X[i] = 0;

  /* classify unknowns based on physical quantities and mesh nodes */
  LPV = (int *)malloc(sizeof(int) * ND);
  LPM = (int *)malloc(sizeof(int) * ND);
  for (i = 0; i < ND; i++) {
    LPV[i] = 1; /* assume only one physical quantity */
    LPM[i] = i+1; /* assume a 1-1 relationship between unknowns and nodes */
  }

#ifdef DEBUG
  t = silc_wtime();
#endif
  RTC = SMSAMG3D
    (X,        /* double *X */
     &ABRS,    /* double *Abrs; relative residual 2-norm */
     &NSTP,    /* int *Nstp; number of iterations */
     AD,       /* double *AD */
     ALU,      /* double *ALU */
     B,        /* double *B */
     LNT,      /* int *LNT */
     LND,      /* int *LND */
     LPV,      /* int *LPV */
     LPM,      /* int *LPM */
     ND,       /* int ND */
     NS,       /* int NS */
     1,        /* int NV; number of physical quantities */
     MSTP,     /* int Mstp */
     EPS,      /* double EPS */
     LEPSSW,   /* int Lepssw */
     LOP1,     /* int Lop1 */
     1,        /* int Lu0sw; 1 = a zero initial guess */
     RTO,      /* double RTO */
     OMG,      /* double Omg */
     0,        /* int Mfmt; 0 = full, 1 = upper triangle */
     METHOD);  /* int Method; 0 = RC, 1 = CG */
#ifdef DEBUG
  elapsed = silc_wtime() - t;
  printf("SMSAMG3D: RTC = %d\n", RTC);
  printf("SMSAMG3D: NSTP = %d\n", NSTP);
  printf("SMSAMG3D: ABRS = %e\n", ABRS);
  printf("SMSAMG3D: %.6f sec\n", elapsed);
#endif
  free(AD);
  free(ALU);
  free(LNT);
  free(LND);
  free(B);
  free(LPV);
  free(LPM);

  return RTC;
}

/* out: 2 */
int
leq_smsamg_solve(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;
  silc_crs_t *in;
  int i, j, ret;

  A = param[0].o;
  b = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": solve_1(0x%lx, 0x%lx)\n", A, b);
#endif
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return -1;
  }
  if (A->d.matrix.m != b->d.vector.length) {
    fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
    return -1;
  }
  in = SILC_CRS(A);
  for (i = 0; i < A->d.matrix.n; i++) {
    for (j = in->row[i]; j < in->row[i+1]; j++) {
      if (in->index[j] == i)
	goto found;
    }
    fprintf(stderr, MODULE_NAME ": zero diagonal element(s)\n");
    return -1;
  found:
    ;
  }
  x = silc_object_duplicate(b, 0);

  ret = call_smsamg3d(A, b, x);
  switch (ret) {
  case 0:
    break;
  case 100: /* license file */
  case 200: case 201: case 203: /* log file */
  case 801: case 802: case 803: /* dump file */
    fprintf(stderr, "leq_smsamg_solve: error code %d (ignored)\n", ret);
    break;
  default:
    fprintf(stderr, "leq_smsamg_solve: error code %d\n", ret);
    silc_object_free(x);
    return -1;
  }
  param[2].o = x;
  return 0;
}

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {leq_smsamg_solve, "leq_smsamg_solve",
   SILC_LEQ_OP, SILC_DOUBLE,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   SILC_COLUMN_VECTOR_TYPE, 0, 2, 0},
  {NULL} /* sentinel */
};

void
init_leq_smsamg()
{
  silc_register_module(MODULE_NAME,
		       NULL,  /* _unary_op_table */
		       _binary_op_table,
		       NULL,  /* _function_table */
		       NULL); /* _procedure_table */
}
