/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: linpackmodule.c,v 1.4 2007/10/31 05:50:17 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

#include "module.h"
#include "dense.h"
#include "wtime.h"

#define MODULE_NAME "linpackmodule"

void dgefa_(double *, int *, int *, int *, int *);
void dgesl_(double *, int *, int *, int *, double *, int *);

int
linpack(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *A, *b, *x;
  silc_dense_t *in;
  double *value, t1, t2, total, mflops;
  int n, *ip, info, job;

  A = param[0].o;
  b = param[1].o;
#ifdef DEBUG
  printf(MODULE_NAME ": linpack(0x%lx, 0x%lx, out)\n", A, b);
#endif
  if (A->type != SILC_MATRIX_TYPE || b->type != SILC_COLUMN_VECTOR_TYPE ||
      A->format != silc_find_format_id(SILC_FORMAT_DENSE))
    return -1;
  if (A->precision != SILC_DOUBLE ||
      b->precision != SILC_DOUBLE) {
    fprintf(stderr, MODULE_NAME ": unsupported precision\n");
    return -1;
  }
  if (A->d.matrix.m != A->d.matrix.n) {
    fprintf(stderr, MODULE_NAME ": not a square matrix\n");
    return -1;
  }
  if (A->d.matrix.m != b->d.vector.length) {
    fprintf(stderr, MODULE_NAME ": matrix-vector size mismatch\n");
    return -1;
  }
  n = A->d.matrix.n;
  value = (double *)malloc(sizeof(double) * n * n);
  in = SILC_DENSE(A);
  memcpy(value, in->v.d, sizeof(double) * n * n);
  ip = (int *)malloc(sizeof(int) * n);
  t1 = silc_wtime();
  dgefa_(value, &n, &n, ip, &info);
  t1 = silc_wtime() - t1;
  if (info > 0)
    fprintf(stderr, MODULE_NAME ": linpack: info > 0\n");
  x = silc_object_duplicate(b, 1);
  job = 0;
  t2 = silc_wtime();
  dgesl_(value, &n, &n, ip, x->d.vector.v.d, &job);
  t2 = silc_wtime() - t2;
  total = t1 + t2;
  mflops = ((2.0 * n * n * n) / 3.0 + 2.0 * n * n) / (1.0e+6 * total);
  printf("computation time: %.3lfs (%.3lf Mflops)\n", total, mflops);
#ifdef DEBUG
  {
    double norm, d;
    int i, j;

    norm = 0.0;
    for (i = 0; i < n; i++) {
      d = b->d.vector.v.d[i];
      for (j = 0; j < n; j++)
	d -= in->v.d[i + j * n] * x->d.vector.v.d[j];
      norm += d * d;
    }
    printf("||b-Ax|| = %e\n", sqrt(norm));
  }
#endif
  free(ip);
  free(value);
  param[2].o = x;
  param[3].o = silc_object_new_double(total);
  return 0;
}

#define Q(x) " \"" x "\" "

static procedure_table_t _procedure_table[] = {
  /* name, sig_text, flags */
  {linpack, "linpack", "linpack(matrix" Q(SILC_FORMAT_DENSE) "in, vector in, vector out, scalar out)", 0},
  {NULL} /* sentinel */
};

void
init_linpackmodule()
{
  silc_register_module(MODULE_NAME,
		       NULL,  /* _unary_op_table */
		       NULL,  /* _binary_op_table */
		       NULL,  /* _function_table */
		       _procedure_table);
}
