/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: modifier.c,v 1.5 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <malloc.h>

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "modifier.h"
#include "interp.h" /* for error() */
#include "wtime.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#ifdef TIME_SUBSCRIPT_VECTOR
extern double _vector_subsref_time; /* server.c */
#endif

silc_modifier_t *
modifier_new(void)
{
  silc_modifier_t *mod;

  mod = (silc_modifier_t *)malloc(sizeof(silc_modifier_t));
  mod->n = 0;
  mod->conjugate = 0;
  mod->transpose = 0;
  mod->negate = 0;
  return mod;
}

void
modifier_free(silc_modifier_t *mod)
{
  int i;

  for (i = 0; i < mod->n; i++)
    clist_free(mod->subscript[i]);
  free(mod);
}

/* subscript */

static silc_object_t *
apply_subscript_to_vector(silc_object_t *object, clist_t *subscript[])
{
  silc_object_t *modified, *temp;
  int i, length, *v, start, end;
#ifdef TIME_SUBSCRIPT_VECTOR
  double t, elapsed;

  t = silc_wtime();
#endif
  length = clist_length(subscript[0]);
  if (length == 1) {
    v = clist_expand(subscript[0]);
    switch (object->precision) {
    case SILC_INT:
      modified = silc_object_new_int(object->d.vector.v.i[v[0]]);
      break;
    case SILC_LONG:
      modified = silc_object_new_long(object->d.vector.v.l[v[0]]);
      break;
    case SILC_FLOAT:
      modified = silc_object_new_float(object->d.vector.v.s[v[0]]);
      break;
    case SILC_DOUBLE:
      modified = silc_object_new_double(object->d.vector.v.d[v[0]]);
      break;
    case SILC_COMPLEX:
      i = v[0] * 2;
      modified = silc_object_new_complex
	(object->d.vector.v.c[i], object->d.vector.v.c[i+1]);
      break;
    case SILC_DOUBLE_COMPLEX:
      i = v[0] * 2;
      modified = silc_object_new_double_complex
	(object->d.vector.v.z[i], object->d.vector.v.z[i+1]);
      break;
    default:
      should_not_reach_here;
    }
    free(v);
  } else if (object->type == SILC_RANGE_TYPE) {
    modified = silc_object_new_column_vector(object->precision, length);
    temp = silc_object_new_column_vector_from_range(object);
    v = clist_expand(subscript[0]);
#pragma omp parallel for
    for (i = 0; i < length; i++)
      modified->d.vector.v.i[i] = temp->d.vector.v.i[v[i]];
    silc_object_free(temp);
    free(v);
  } else if (clist_range(subscript[0], &start, &end)) {
    if (object->type == SILC_ROW_VECTOR_TYPE)
      modified = silc_object_new_row_vector(object->precision, length);
    else
      modified = silc_object_new_column_vector(object->precision, length);
    switch (object->precision) {
    case SILC_INT:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.i[i] = object->d.vector.v.i[start+i];
      break;
    case SILC_LONG:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.l[i] = object->d.vector.v.l[start+i];
      break;
    case SILC_FLOAT:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.s[i] = object->d.vector.v.s[start+i];
      break;
    case SILC_DOUBLE:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.d[i] = object->d.vector.v.d[start+i];
      break;
    case SILC_COMPLEX:
#pragma omp parallel for
      for (i = 0; i < length; i++) {
	int ii, jj;
        ii = i * 2; jj = ii + start * 2;
	modified->d.vector.v.c[ii]   = object->d.vector.v.c[jj];
	modified->d.vector.v.c[ii+1] = object->d.vector.v.c[jj+1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
      for (i = 0; i < length; i++) {
	int ii, jj;
        ii = i * 2; jj = ii + start * 2;
	modified->d.vector.v.z[ii]   = object->d.vector.v.z[jj];
	modified->d.vector.v.z[ii+1] = object->d.vector.v.z[jj+1];
      }
      break;
    default:
      should_not_reach_here;
    }
  } else {
    if (object->type == SILC_ROW_VECTOR_TYPE)
      modified = silc_object_new_row_vector(object->precision, length);
    else
      modified = silc_object_new_column_vector(object->precision, length);
    switch (object->precision) {
    case SILC_INT:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.i[i] = object->d.vector.v.i[v[i]];
      break;
    case SILC_LONG:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.l[i] = object->d.vector.v.l[v[i]];
      break;
    case SILC_FLOAT:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.s[i] = object->d.vector.v.s[v[i]];
      break;
    case SILC_DOUBLE:
#pragma omp parallel for
      for (i = 0; i < length; i++)
	modified->d.vector.v.d[i] = object->d.vector.v.d[v[i]];
      break;
    case SILC_COMPLEX:
#pragma omp parallel for
      for (i = 0; i < length; i++) {
	int ii, jj;
        ii = i * 2; jj = v[i] * 2;
	modified->d.vector.v.c[ii]   = object->d.vector.v.c[jj];
	modified->d.vector.v.c[ii+1] = object->d.vector.v.c[jj+1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
      for (i = 0; i < length; i++) {
	int ii, jj;
        ii = i * 2; jj = v[i] * 2;
	modified->d.vector.v.z[ii]   = object->d.vector.v.z[jj];
	modified->d.vector.v.z[ii+1] = object->d.vector.v.z[jj+1];
      }
      break;
    default:
      should_not_reach_here;
    }
    free(v);
  }
#ifdef TIME_SUBSCRIPT_VECTOR
  elapsed = silc_wtime() - t;
  _vector_subsref_time += elapsed;
#ifdef DEBUG
  printf("apply_subscript_to_vector: %e sec.\n", elapsed);
#endif
#endif
  return modified;
}

static silc_object_t *
apply_subscript_to_matrix(silc_object_t *object, clist_t *subscript[])
{
  silc_object_t *modified;
  silc_format_t *table;

  table = silc_find_format_proc_table(object->format);
  if (!table)
    return NULL;
  if (!table->proc_subscript) {
    error("proc_subscript() not defined (fatal)");
    return NULL;
  }
  modified = table->proc_subscript(object, subscript);
  if (!modified) {
    error("proc_subscript() failed (fatal)");
    return NULL;
  }
  return modified;
}

static silc_object_t *
apply_subscript_to_cubic_array(silc_object_t *object, clist_t *subscript[])
{
  silc_object_t *modified;
  int c, i0, i1, i2, j, s0, s1, s2, *v0, *v1, *v2;
  size_t plane_size;

  plane_size = object->d.cubic_array.m * object->d.cubic_array.n;
  v0 = clist_expand(subscript[0]); s0 = clist_length(subscript[0]);
  v1 = clist_expand(subscript[1]); s1 = clist_length(subscript[1]);
  v2 = clist_expand(subscript[2]); s2 = clist_length(subscript[2]);
  if (s0 == 1 && s1 == 1 && s2 == 1) {
    j = (v0[0] * plane_size + v1[0] + v2[0] * object->d.cubic_array.m);
    switch (object->precision) {
    case SILC_INT:
      modified = silc_object_new_int(object->d.cubic_array.v.i[j]); break;
    case SILC_LONG:
      modified = silc_object_new_long(object->d.cubic_array.v.l[j]); break;
    case SILC_FLOAT:
      modified = silc_object_new_float(object->d.cubic_array.v.s[j]); break;
    case SILC_DOUBLE:
      modified = silc_object_new_double(object->d.cubic_array.v.d[j]); break;
    case SILC_COMPLEX:
      j = j * 2;
      modified = silc_object_new_complex
	(object->d.cubic_array.v.c[j], object->d.cubic_array.v.c[j+1]);
      break;
    case SILC_DOUBLE_COMPLEX:
      j = j * 2;
      modified = silc_object_new_double_complex
	(object->d.cubic_array.v.z[j], object->d.cubic_array.v.z[j+1]);
      break;
    default:
      should_not_reach_here;
    }
  } else {
    modified = silc_object_new_cubic_array(object->precision, s0, s1, s2);
    c = 0;
    switch (object->precision) {
    case SILC_INT:
      for (i0 = 0; i0 < s0; i0++) {
	for (i2 = 0; i2 < s2; i2++) {
	  for (i1 = 0; i1 < s1; i1++) {
	    j = (v0[i0] * plane_size +
		 v1[i1] + v2[i2] * object->d.cubic_array.m);
	    modified->d.cubic_array.v.i[c++] = object->d.cubic_array.v.i[j];
	  }
	}
      }
      break;
    case SILC_LONG:
      for (i0 = 0; i0 < s0; i0++) {
	for (i2 = 0; i2 < s2; i2++) {
	  for (i1 = 0; i1 < s1; i1++) {
	    j = (v0[i0] * plane_size +
		 v1[i1] + v2[i2] * object->d.cubic_array.m);
	    modified->d.cubic_array.v.l[c++] = object->d.cubic_array.v.l[j];
	  }
	}
      }
      break;
    case SILC_FLOAT:
      for (i0 = 0; i0 < s0; i0++) {
	for (i2 = 0; i2 < s2; i2++) {
	  for (i1 = 0; i1 < s1; i1++) {
	    j = (v0[i0] * plane_size +
		 v1[i1] + v2[i2] * object->d.cubic_array.m);
	    modified->d.cubic_array.v.s[c++] = object->d.cubic_array.v.s[j];
	  }
	}
      }
      break;
    case SILC_DOUBLE:
      for (i0 = 0; i0 < s0; i0++) {
	for (i2 = 0; i2 < s2; i2++) {
	  for (i1 = 0; i1 < s1; i1++) {
	    j = (v0[i0] * plane_size +
		 v1[i1] + v2[i2] * object->d.cubic_array.m);
	    modified->d.cubic_array.v.d[c++] = object->d.cubic_array.v.d[j];
	  }
	}
      }
      break;
    case SILC_COMPLEX:
      for (i0 = 0; i0 < s0; i0++) {
	for (i2 = 0; i2 < s2; i2++) {
	  for (i1 = 0; i1 < s1; i1++) {
	    j = (v0[i0] * plane_size +
		 v1[i1] + v2[i2] * object->d.cubic_array.m) * 2;
	    modified->d.cubic_array.v.c[c++] = object->d.cubic_array.v.c[j];
	    modified->d.cubic_array.v.c[c++] = object->d.cubic_array.v.c[j+1];
	  }
	}
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      for (i0 = 0; i0 < s0; i0++) {
	for (i2 = 0; i2 < s2; i2++) {
	  for (i1 = 0; i1 < s1; i1++) {
	    j = (v0[i0] * plane_size +
		 v1[i1] + v2[i2] * object->d.cubic_array.m) * 2;
	    modified->d.cubic_array.v.z[c++] = object->d.cubic_array.v.z[j];
	    modified->d.cubic_array.v.z[c++] = object->d.cubic_array.v.z[j+1];
	  }
	}
      }
      break;
    default:
      should_not_reach_here;
    }
  }
  free(v0);
  free(v1);
  free(v2);
  return modified;
}

DECLSPEC silc_object_t *
modifier_apply_subscript(silc_object_t *object, int n, clist_t *subscript[])
{
#ifdef DEBUG
  printf("modifier_apply_subscript(0x%lx)\n", object);
#endif
  switch (object->type) {
  case SILC_SCALAR_TYPE:
    error("cannot apply subscript to scalar");
    return NULL;
  case SILC_RANGE_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
    if (n != 1) {
      error("invalid subscript size");
      return NULL;
    }
    clist_set_defaults(subscript[0], 0, object->d.vector.length - 1);
    if (!(clist_max(subscript[0]) < object->d.vector.length)) {
      error("subscript out of range");
      return NULL;
    }
    return apply_subscript_to_vector(object, subscript);
  case SILC_MATRIX_TYPE:
    if (n != 2) {
      error("invalid subscript size");
      return NULL;
    }
    clist_set_defaults(subscript[0], 0, object->d.matrix.m - 1);
    clist_set_defaults(subscript[1], 0, object->d.matrix.n - 1);
    if (!(clist_max(subscript[0]) < object->d.matrix.m) ||
	!(clist_max(subscript[1]) < object->d.matrix.n)) {
      error("subscript out of range");
      return NULL;
    }
    return apply_subscript_to_matrix(object, subscript);
  case SILC_CUBIC_ARRAY_TYPE:
    if (n != 3) {
      error("invalid subscript size");
      return NULL;
    }
    clist_set_defaults(subscript[0], 0, object->d.cubic_array.l - 1);
    clist_set_defaults(subscript[1], 0, object->d.cubic_array.m - 1);
    clist_set_defaults(subscript[2], 0, object->d.cubic_array.n - 1);
    if (!(clist_max(subscript[0]) < object->d.cubic_array.l) ||
	!(clist_max(subscript[1]) < object->d.cubic_array.m) ||
	!(clist_max(subscript[2]) < object->d.cubic_array.n)) {
      error("subscript out of range");
      return NULL;
    }
    return apply_subscript_to_cubic_array(object, subscript);
  default:
    should_not_reach_here;
  }
  return NULL;
}

/* unary ops: complex conjugate, transpose and conjugate transpose */

#ifdef TIME_FIND_MFUNC
extern double _find_mfunc_time; /* server.c */
#endif
#ifdef TIME_CALL_MFUNC
extern double _call_mfunc_time; /* server.c */
#endif

static silc_object_t *
apply_unary_op(int op, silc_object_t *object)
{
  silc_object_t *operand, *modified;
  silc_mfunc_t *mfunc;
  silc_mfunc_param_t *param;
  int status;
#if defined(TIME_CALL_MFUNC) || defined(TIME_FIND_MFUNC)
  double t, elapsed;
#endif

  if (object->type == SILC_RANGE_TYPE)
    operand = silc_object_new_column_vector_from_range(object);
  else
    operand = object;
#ifdef TIME_FIND_MFUNC
  t = silc_wtime();
#endif
  mfunc = silc_find_unary_op_impl
    (op, operand->precision, operand->type, operand->format);
  if (!mfunc) {
    error("FATAL: no appropriate module function for %s (%s)",
	  ((op == SILC_NEGATE_OP) ? "negate" :
	   (op == SILC_COMPLEX_CONJUGATE_OP) ? "complex conjugate" :
	   (op == SILC_CONJUGATE_TRANSPOSE_OP) ? "conjugate transpose" :
	   (op == SILC_TRANSPOSE_OP) ? "transpose" : "<unknown>"),
	  SILC_TYPE_NAME(operand));
    return NULL;
  }
#ifdef TIME_FIND_MFUNC
  elapsed = silc_wtime() - t;
  _find_mfunc_time += elapsed;
#endif
  param = (silc_mfunc_param_t *)malloc(sizeof(silc_mfunc_param_t) * 2);
  switch (mfunc->out) {
  case 0:
    if (object->type == SILC_RANGE_TYPE)
      param[0].o = operand; /* already duplicated */
    else
      param[0].o = silc_object_duplicate(operand, 1);
    param[0].m = NULL;
    break;
  case 1:
    param[0].o = operand;
    param[0].m = NULL;
    break;
  default:
    should_not_reach_here;
  }
#ifdef TIME_CALL_MFUNC
  t = silc_wtime();
#endif
  status = (mfunc->impl)(1, param);
#ifdef TIME_CALL_MFUNC
  elapsed = silc_wtime() - t;
  _call_mfunc_time += elapsed;
#if defined(DEBUG) || TIME_CALL_MFUNC > 1
  printf("mfunc: %s in %e sec.\n", mfunc->mfunc_name, elapsed);
#endif
#endif
  if (object->type == SILC_RANGE_TYPE && mfunc->out == 1)
    silc_object_free(operand);
  modified = param[mfunc->out].o;
  free(param);
  free(mfunc);
  if (status < 0) {
    if (mfunc->out == 0)
      silc_object_free(modified); /* i.e. param[0].o */
    return NULL;
  }
  return modified; /* new object */
}

DECLSPEC silc_object_t *
modifier_apply_complex_conjugate(silc_object_t *object)
{
#ifdef DEBUG
  printf("modifier_apply_complex_conjugate(0x%lx)\n", object);
#endif
  return apply_unary_op(SILC_COMPLEX_CONJUGATE_OP, object);
}

DECLSPEC silc_object_t *
modifier_apply_transpose(silc_object_t *object)
{
#ifdef DEBUG
  printf("modifier_apply_transpose(0x%lx)\n", object);
#endif
  return apply_unary_op(SILC_TRANSPOSE_OP, object);
}

DECLSPEC silc_object_t *
modifier_apply_conjugate_transpose(silc_object_t *object)
{
#ifdef DEBUG
  printf("modifier_apply_conjugate_transpose(0x%lx)\n", object);
#endif
  return apply_unary_op(SILC_CONJUGATE_TRANSPOSE_OP, object);
}

DECLSPEC silc_object_t *
modifier_apply_negate(silc_object_t *object)
{
#ifdef DEBUG
  printf("modifier_apply_negate(0x%lx)\n", object);
#endif
  return apply_unary_op(SILC_NEGATE_OP, object);
}

/* utility functions */

DECLSPEC silc_object_t *
modifier_apply_all(silc_object_t *object, silc_modifier_t *modifier)
{
  silc_object_t *modified;

#define _apply(func) \
  if (modified == object) {		\
    modified = func(object);		\
  } else {				\
    silc_object_t *p = func(modified);	\
    silc_object_free(modified);		\
    modified = p;			\
  }					\

  if (modifier->n == 0)
    modified = object;
  else
    modified = modifier_apply_subscript(object,
					modifier->n, modifier->subscript);
  if (modifier->conjugate && modifier->transpose) {
    _apply(modifier_apply_conjugate_transpose);
  } else if (modifier->conjugate) {
    _apply(modifier_apply_complex_conjugate);
  } else if (modifier->transpose) {
    _apply(modifier_apply_transpose);
  }
  if (modifier->negate) {
    _apply(modifier_apply_negate);
  }
  if (modified == object)
    modified = silc_object_duplicate(object, 1);
  return modified;
}
