/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: module.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <limits.h> /* for PATH_MAX */
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#ifdef VANILLA_MPI
#include <mpi.h>
#endif

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "sig.h"
#include "module.h"
#include "interp.h"
#include "dl.h"
#include "log.h"

static module_t module_stack;
static char module_path[PATH_MAX];

void
init_module_stack()
{
  module_stack.next = NULL;
}

void
free_module_stack()
{
  module_t *p, *next;

  p = module_stack.next;
  while (p) {
    next = p->next;
    free(p->module_name);
    free(p->module_path);
    free(p->unary_op_table);
    free(p->binary_op_table);
    free(p->function_table);
    free(p->procedure_table);
    free(p);
    p = next;
  }
  module_stack.next = NULL;
}

int
load_modules(const char *p)
{
  DIR *dp;
  struct dirent *de;
  struct stat st;
  char fname[PATH_MAX];
  char *init_module_name, *suffix;
  void *handle;
  void (*init_module)();

#ifdef DEBUG
#ifdef VANILLA_MPI
  if (silc_ipc_master())
#endif
    printf("load_modules(\"%s\")\n", p);
#endif
  strcpy(module_path, p);
  dp = opendir(module_path);
  if (!dp) {
    silc_log("load_modules: cannot open module directory: %s", module_path);
    return -1;
  }
  while ((de = readdir(dp)) != NULL) {
    if (de->d_name[0] == '.')
      continue;
    sprintf(fname, "%s/%s", module_path, de->d_name);
    if (stat(fname, &st) < 0) {
      silc_log("load_modules: %s: %s", strerror(errno), fname);
      continue;
    }
    if (st.st_mode & S_IFDIR)
      continue;
    handle = silc_dlopen(fname); 
    if (!handle) {
      silc_log("load_modules: %s", silc_dlerror());
      continue;
    }
    suffix = strchr(de->d_name, '.');
    if (!suffix) {
      silc_log("load_modules: suffix not found: %s", de->d_name);
      continue;
    }
    init_module_name = (char *)malloc(5 + suffix - de->d_name + 1);
    strcpy(init_module_name, "init_");
    strncpy(init_module_name + 5, de->d_name, suffix - de->d_name);
    init_module_name[5 + suffix - de->d_name] = '\0';
    init_module = (void (*)())silc_dlsym(handle, init_module_name);
    free(init_module_name);
    if (!init_module) {
      silc_log("load_modules: init_module not found: %s", fname);
      continue;
    }
    (*init_module)();
  }
  closedir(dp);
  return 0;
}

static char *
fully_qualified_module_path(char *module_name)
{
  static char buf[PATH_MAX];

  strcpy(buf, module_path);
  strcat(buf, "/");
  strcat(buf, module_name);
  strcat(buf, ".so");  /* XXX: OS dependent */
  return strdup(buf);
}

/* callback */
DECLSPEC void
silc_register_module(char *module_name,
		     unary_op_table_t *unary_op_table,
		     binary_op_table_t *binary_op_table,
		     function_table_t *function_table,
		     procedure_table_t *procedure_table)
{
  module_t *p;
  int i, j, n;
  size_t size;

#ifdef DEBUG
#ifdef VANILLA_MPI
  if (silc_ipc_master())
#endif
    printf("silc_register_module(\"%s\")\n", module_name);
#endif
  for (p = &module_stack; p->next; p = p->next);
  p->next = (module_t *)malloc(sizeof(module_t));
  p->next->next = NULL;
  p->next->module_name = strdup(module_name);
  p->next->module_path = fully_qualified_module_path(module_name);
  p->next->loaded = 0;
  p->next->blocked = 0;

  if (!unary_op_table) {
    p->next->unary_op_table = NULL;
  } else {
    unary_op_table_t *table;

    for (n = 0; unary_op_table[n].impl; n++);
    size = sizeof(unary_op_table_t) * (n + 1);
    table = (unary_op_table_t *)malloc(size);
    memcpy(table, unary_op_table, size);
    table[n].impl = NULL;

    /* replace format names to format IDs */
    for (i = 0; i < n; i++)
      table[i].format.id = (table[i].type == SILC_MATRIX_TYPE) ?
	silc_find_format_id(table[i].format.name) : -1;

    p->next->unary_op_table = table;
  }

  if (!binary_op_table) {
    p->next->binary_op_table = NULL;
  } else {
    binary_op_table_t *table;

    for (n = 0; binary_op_table[n].impl; n++);
    size = sizeof(binary_op_table_t) * (n + 1);
    table = (binary_op_table_t *)malloc(size);
    memcpy(table, binary_op_table, size);
    table[n].impl = NULL;

    /* replace format names to format IDs */
    for (i = 0; i < n; i++) {
      table[i].x_format.id = (table[i].x_type == SILC_MATRIX_TYPE) ?
	silc_find_format_id(table[i].x_format.name) : -1;
      table[i].y_format.id = (table[i].y_type == SILC_MATRIX_TYPE) ?
	silc_find_format_id(table[i].y_format.name) : -1;
    }

    p->next->binary_op_table = table;
  }

  if (!function_table) {
    p->next->function_table = NULL;
  } else {
    function_table_t *table, *entry;
    signature_t *sig;

    for (n = 0; function_table[n].impl; n++);
    table = malloc(sizeof(function_table_t) * (n + 1));
    i = 0;
    for (n = 0; function_table[n].impl; n++) {
      sig = &table[i].sig;
      if (parse_signature(function_table[n].sig_text, sig) < 0 ||
	  sig->return_type == 0) {
	fprintf(stderr, "Error: illegal signature in %s: %s\n",
		module_name, function_table[n].sig_text);
	continue;
      }

      /* replace format names to format IDs */
      for (j = 0; j < sig->n_params; j++)
	sig->param[j].format = (sig->param[j].type == SILC_MATRIX_TYPE) ?
	  silc_find_format_id(sig->param[j].format_name) : -1;
      sig->return_format = (sig->return_type == SILC_MATRIX_TYPE) ?
	silc_find_format_id(sig->return_format_name) : -1;

      table[i].impl  = function_table[n].impl;
      table[i].flags = function_table[n].flags;
      strcpy(table[i].name, function_table[n].name);
      strcpy(table[i].sig_text, function_table[n].sig_text);
      i++;
    }
    table[i].impl = NULL;
    p->next->function_table = table;
  }

  if (!procedure_table) {
    p->next->procedure_table = NULL;
  } else {
    procedure_table_t *table;
    signature_t *sig;

    for (n = 0; procedure_table[n].impl; n++);
    table = malloc(sizeof(procedure_table_t) * (n + 1));
    i = 0;
    for (n = 0; procedure_table[n].impl; n++) {
      sig = &table[i].sig;
      if (parse_signature(procedure_table[n].sig_text, sig) < 0 ||
	  sig->return_type != 0) {
	fprintf(stderr, "Error: illegal signature in %s: %s\n",
		module_name, procedure_table[n].sig_text);
	continue;
      }

      /* replace format names to format IDs */
      for (j = 0; j < sig->n_params; j++)
	sig->param[j].format = (sig->param[j].type == SILC_MATRIX_TYPE) ?
	  silc_find_format_id(sig->param[j].format_name) : -1;

      table[i].impl  = procedure_table[n].impl;
      table[i].flags = procedure_table[n].flags;
      strcpy(table[i].name, procedure_table[n].name);
      strcpy(table[i].sig_text, procedure_table[n].sig_text);
      i++;
    }
    table[i].impl = NULL;
    p->next->procedure_table = table;
  }
}

DECLSPEC int
silc_prefer_module(char *module_name)
{
  module_t *table;
  module_t volatile *p; /* disable -fno-alias */
  char buf[32];

  p = &module_stack;
  while (p->next) {
    if (strcmp(p->next->module_name, module_name) == 0) {
      table = p->next;
      p->next = p->next->next;
      table->next = module_stack.next;
      module_stack.next = table;
#ifdef DEBUG
      printf("silc_prefer_module(\"%s\")\n", module_name);
      printf("module search order:\n");
      for (p = module_stack.next; p; p = p->next)
	printf("  %s\n", p->module_name);
#endif
      return 0;
    }
    p = p->next;
  }
  return -1;
}

module_t *
get_module(const char *module_name)
{
  module_t *p;

  p = module_stack.next;
  while (p) {
    if (strcmp(p->module_name, module_name) == 0) 
      return p;
    p = p->next;
  }
  return NULL;
}

silc_mfunc_impl_t
get_module_function(char *module_name, char *mfunc_name)
{
  module_t *p;
  silc_mfunc_impl_t impl;

  p = get_module(module_name);
  if (!p)
    return NULL;
#ifdef STATIC_MODULES
  impl = NULL;
  if (p->unary_op_table) {
    unary_op_table_t *table;
    for (table = p->unary_op_table; table->impl; table++) {
      if (strcmp(table->name, mfunc_name) == 0) {
	impl = table->impl;
	goto found;
      }
    }
  }
  if (p->binary_op_table) {
    binary_op_table_t *table;
    for (table = p->binary_op_table; table->impl; table++) {
      if (strcmp(table->name, mfunc_name) == 0) {
	impl = table->impl;
	goto found;
      }
    }
  }
  if (p->function_table) {
    function_table_t *table;
    for (table = p->function_table; table->impl; table++) {
      if (strcmp(table->name, mfunc_name) == 0) {
	impl = table->impl;
	goto found;
      }
    }
  }
  if (p->procedure_table) {
    procedure_table_t *table;
    for (table = p->procedure_table; table->impl; table++) {
      if (strcmp(table->name, mfunc_name) == 0) {
	impl = table->impl;
	goto found;
      }
    }
  }
 found:
#else /* STATIC_MODULES */
  if (!p->loaded) {
    p->handle = silc_dlopen(p->module_path);
    if (!p->handle) {
      fprintf(stderr, "cannot load module: %s\n", p->module_path);
      return NULL;
    }
    p->loaded = 1;
  }
  impl = (silc_mfunc_impl_t)silc_dlsym(p->handle, mfunc_name);
  if (!impl) {
    fprintf(stderr, "cannot find symbol: %s\n", mfunc_name);
    return NULL;
  }
#endif /* STATIC_MODULES */
  return impl;
}

#if 0 /* unused */
int
unload_module(module_t *module)
{
  char fname[PATH_MAX];
  void (*finalize_module)();

  module->handle = silc_dlopen(module->module_path);
  
  finalize_module = (void (*)())silc_dlsym(module->handle, "finalize_module");
  if (finalize_module)
    (*finalize_module)();
 
  module->loaded = 1;
  module->blocked = 0;
}
#endif

DECLSPEC silc_mfunc_t *
silc_find_unary_op_impl(int op, int precision, int type, int format)
{
  module_t *p;
  unary_op_table_t *table;
  silc_mfunc_impl_t impl;
  silc_mfunc_t *mfunc;
  int i;

#ifdef DEBUG
  printf("silc_find_unary_op_impl()\n");
#endif
  p = module_stack.next;
  while (p) {
    table = p->unary_op_table;
    if (table) {
      for (i = 0; table[i].impl; i++) {
	if (table[i].op == op &&
	    (table[i].precision & precision) &&
	    (table[i].type & type) &&
	    (table[i].format.id < 0 || table[i].format.id == format)) {
#ifdef STATIC_MODULES
	  impl = table[i].impl;
#else
	  if (!p->loaded) {
	    p->handle = silc_dlopen(p->module_path);
	    if (!p->handle) {
	      fprintf(stderr, "cannot load module: %s\n", p->module_path);
	      return NULL;
	    }
	    p->loaded = 1;
	  }
	  impl = (silc_mfunc_impl_t)silc_dlsym(p->handle, table[i].name);
	  if (!impl) {
	    fprintf(stderr, "cannot find symbol: %s\n", table[i].name);
	    return NULL;
	  }
#endif
	  mfunc = (silc_mfunc_t *)malloc(sizeof(silc_mfunc_t));
	  mfunc->impl        = impl;
	  mfunc->out         = table[i].out;
	  mfunc->flags       = table[i].flags;
	  mfunc->mfunc_name  = table[i].name;
	  mfunc->module_name = p->module_name;
	  return mfunc;
	}
      }
    }
    p = p->next;
  }
  return NULL;
}

DECLSPEC silc_mfunc_t *
silc_find_binary_op_impl(int op, int precision,
			 int x_type, int x_format,
			 int y_type, int y_format)
{
  module_t *p;
  binary_op_table_t *table;
  silc_mfunc_impl_t impl;
  silc_mfunc_t *mfunc;
  int i;

#ifdef DEBUG
  printf("silc_find_binary_op_impl()\n");
#endif
  p = module_stack.next;
  while (p) {
    table = p->binary_op_table;
    if (table) {
      for (i = 0; table[i].impl; i++) {
	if (table[i].op == op &&
	    (table[i].precision & precision) &&
	    (table[i].x_type & x_type) &&
	    (table[i].y_type & y_type) &&
	    (table[i].x_format.id < 0 || table[i].x_format.id == x_format) &&
	    (table[i].y_format.id < 0 || table[i].y_format.id == y_format)) {
#ifdef STATIC_MODULES
	  impl = table[i].impl;
#else
	  if (!p->loaded) {
	    p->handle = silc_dlopen(p->module_path);
	    if (!p->handle) {
	      fprintf(stderr, "cannot load module: %s\n", p->module_path);
	      return NULL;
	    }
	    p->loaded = 1;
	  }
	  impl = (silc_mfunc_impl_t)silc_dlsym(p->handle, table[i].name);
	  if (!impl) {
	    fprintf(stderr, "cannot find symbol: %s\n", table[i].name);
	    return NULL;
	  }
#endif
	  mfunc = (silc_mfunc_t *)malloc(sizeof(silc_mfunc_t));
	  mfunc->impl        = impl;
	  mfunc->out         = table[i].out;
	  mfunc->flags       = table[i].flags;
	  mfunc->mfunc_name  = table[i].name;
	  mfunc->module_name = p->module_name;
	  return mfunc;
	}
      }
    }
    p = p->next;
  }
  return NULL;
}

DECLSPEC silc_mfunc_t *
silc_find_function_impl(char *name, int n_params, int *param_type)
{
  module_t *p;
  function_table_t *table;
  silc_mfunc_impl_t impl;
  silc_mfunc_t *mfunc;
  int i, j, sig_error, nth;

#define arg_error(n, msg) \
	error("%s: %d%s argument: %s", name, n,		\
	      (n != 11 && n % 10 == 1) ? "st" :		\
	      (n != 12 && n % 10 == 2) ? "nd" :		\
	      (n != 13 && n % 10 == 3) ? "rd" : "th",	\
	      msg)

#ifdef DEBUG
  printf("silc_find_function_impl(\"%s\")\n", name);
#endif
  p = module_stack.next;
  while (p) {
    table = p->function_table;
    if (table) {
      for (i = 0; table[i].impl; i++) {
	if (strcmp(table[i].sig.name, name))
	  continue;
	if (table[i].sig.n_params != n_params) {
	  sig_error = 1; /* illegal number of arguments */
	  continue;
	}
	for (j = 0; j < n_params; j++) {
	  if ((table[i].sig.param[j].type & param_type[j]) == 0)
	    break;
	}
	if (j < n_params) {
	  sig_error = 2; nth = j + 1; /* unexpected data type */
	  continue;
	}
#ifdef STATIC_MODULES
	impl = table[i].impl;
#else
	if (!p->loaded) {
	  p->handle = silc_dlopen(p->module_path);
	  if (!p->handle) {
	    fprintf(stderr, "cannot load module: %s\n", p->module_path);
	    return NULL;
	  }
	  p->loaded = 1;
	}
	impl = (silc_mfunc_impl_t)silc_dlsym(p->handle, table[i].name);
	if (!impl) {
	  fprintf(stderr, "cannot find symbol: %s\n", table[i].name);
	  return NULL;
	}
#endif
	mfunc = (silc_mfunc_t *)malloc(sizeof(silc_mfunc_t));
	mfunc->impl        = impl;
	mfunc->out         = 0; /* not used */
	mfunc->flags       = table[i].flags;
	mfunc->mfunc_name  = table[i].name;
	mfunc->module_name = p->module_name;
	return mfunc;
      }
    }
    p = p->next;
  }
  switch (sig_error) {
  case 1:
    error("%s: illegal number of arguments", name);
    break;
  case 2:
    arg_error(nth, "unexpected data type");
    break;
  default:
    error("%s: function not found", name);
  }
  return NULL;
}

DECLSPEC signature_t *
silc_find_procedure(char *name)
{
  module_t *p;
  procedure_table_t *table;
  int i;

  p = module_stack.next;
  while (p) {
    table = p->procedure_table;
    if (table) {
      for (i = 0; table[i].impl; i++) {
	if (strcmp(table[i].sig.name, name) == 0) {
	  return &table[i].sig;
	}
      }
    }
    p = p->next;
  }
  return NULL;
}

DECLSPEC silc_mfunc_t *
silc_find_procedure_impl(char *name)
{
  module_t *p;
  procedure_table_t *table;
  silc_mfunc_impl_t impl;
  silc_mfunc_t *mfunc;
  int i;

#ifdef DEBUG
  printf("silc_find_procedure_impl(\"%s\")\n", name);
#endif
  p = module_stack.next;
  while (p) {
    table = p->procedure_table;
    if (table) {
      for (i = 0; table[i].impl; i++) {
	if (strcmp(table[i].sig.name, name) == 0) { /* XXX: parameters */
#ifdef STATIC_MODULES
	  impl = table[i].impl;
#else
	  if (!p->loaded) {
	    p->handle = silc_dlopen(p->module_path);
	    if (!p->handle) {
	      fprintf(stderr, "cannot load module: %s\n", p->module_path);
	      return NULL;
	    }
	    p->loaded = 1;
	  }
	  impl = (silc_mfunc_impl_t)silc_dlsym(p->handle, table[i].name);
	  if (!impl) {
	    fprintf(stderr, "cannot find symbol: %s\n", table[i].name);
	    return NULL;
	  }
#endif
	  mfunc = (silc_mfunc_t *)malloc(sizeof(silc_mfunc_t));
	  mfunc->impl        = impl;
	  mfunc->out         = 0; /* not used */
	  mfunc->flags       = table[i].flags;
	  mfunc->mfunc_name  = table[i].name;
	  mfunc->module_name = p->module_name;
	  return mfunc;
	}
      }
    }
    p = p->next;
  }
  return NULL;
}

/* callback */

static struct _format {
  char name[SILC_MAXNAMELEN+1];
  silc_format_t table;
} *format;

static int num_formats = 0;

DECLSPEC int
silc_register_format(char *name, silc_format_t *table)
{
  int id;
  size_t size;
  silc_format_t *p;

#ifdef DEBUG
#ifdef VANILLA_MPI
  if (silc_ipc_master())
#endif
    printf("silc_register_format(\"%s\")\n", name);
#endif
  id = num_formats;
  num_formats++;
  size = num_formats * sizeof(struct _format);
  if (num_formats == 1)
    format = (struct _format *)malloc(size);
  else
    format = (struct _format *)realloc(format, size);
  if (!format)
    return -1;
  strcpy(format[id].name, name);
  format[id].table = *table;
  return id;
}

DECLSPEC const char *
silc_find_format_name(int id)
{
  if (id < num_formats)
    return format[id].name;
  return NULL;
}

DECLSPEC silc_format_t *
silc_find_format_proc_table(int id)
{
  if (id < num_formats)
    return &format[id].table;
  return NULL;
}

DECLSPEC int
silc_find_format_id(const char *name)
{
  int i;

  for (i = 0; i < num_formats; i++)
    if (strcmp(format[i].name, name) == 0)
      return i;
  return -1;
}

DECLSPEC int
silc_get_number_of_formats()
{
  return num_formats;
}
