/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: module.h,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#ifndef SILC_MODULE_H
#define SILC_MODULE_H 1

#include "object.h"
#include "modifier.h"
#include "cmd.h"
#include "sig.h"

/* type definitions */

typedef struct {
  silc_object_t *o;
  silc_modifier_t *m;
} silc_mfunc_param_t;

typedef int (*silc_mfunc_impl_t)(int, silc_mfunc_param_t *);

typedef struct {
  silc_mfunc_impl_t impl;
  char name[SILC_MAXNAMELEN+1];
  int op, precision, type;
  union { char *name; int id; } format;
  int out, flags;
} unary_op_table_t;

typedef struct {
  silc_mfunc_impl_t impl;
  char name[SILC_MAXNAMELEN+1];
  int op, precision;
  int x_type; union { char *name; int id; } x_format;
  int y_type; union { char *name; int id; } y_format;
  int out, flags;
} binary_op_table_t;

typedef struct {
  silc_mfunc_impl_t impl;
  char name[SILC_MAXNAMELEN+1];
  char sig_text[SILC_SIGNATURE_MAX+1];
  int flags;
  /* internal */
  signature_t sig;
} function_table_t;

typedef struct {
  silc_mfunc_impl_t impl;
  char name[SILC_MAXNAMELEN+1];
  char sig_text[SILC_SIGNATURE_MAX+1];
  int flags;
  /* internal */
  signature_t sig;
} procedure_table_t;

typedef struct _module {
  struct _module *next;
  char *module_name;
  char *module_path;
  int loaded;
  int blocked;
  void *handle;
  unary_op_table_t *unary_op_table;
  binary_op_table_t *binary_op_table;
  function_table_t *function_table;
  procedure_table_t *procedure_table;
} module_t;

typedef struct {
  silc_mfunc_impl_t impl;
  int out, flags;
  char *module_name, *mfunc_name;
} silc_mfunc_t;

#ifdef VANILLA_MPI
int silc_ipc_master();
void silc_ipc_init(char *, char *);
void silc_ipc_finalize();

/* flags */
#define SILC_MPI_MFUNC 1
#define silc_ipc_mpi_mfunc(x) ((x) & SILC_MPI_MFUNC)
#endif

/* function declarations */

#if defined(_WIN32) && !defined(STATIC_MODULES)
#ifdef _EXPORTING
#define DECLSPEC __declspec(dllexport)
#else
#define DECLSPEC __declspec(dllimport)
#endif
#else
#define DECLSPEC
#endif

void init_module_stack();
void free_module_stack();
DECLSPEC void silc_register_module(char *, unary_op_table_t *, binary_op_table_t *,
					 function_table_t *, procedure_table_t *);
int load_modules(const char *);
DECLSPEC int silc_prefer_module(char *);
module_t *get_module(const char *);
silc_mfunc_impl_t get_module_function(char *, char *);
int unload_module(module_t *);

DECLSPEC silc_mfunc_t *silc_find_unary_op_impl(int, int, int, int);
DECLSPEC silc_mfunc_t *silc_find_binary_op_impl(int, int, int, int, int, int);
DECLSPEC silc_mfunc_t *silc_find_function_impl(char *, int, int *);
DECLSPEC signature_t *silc_find_procedure(char *);
DECLSPEC silc_mfunc_t *silc_find_procedure_impl(char *);

/* for storage format modules */

typedef int (*silc_format_new)(silc_object_t *);
typedef int (*silc_format_recv)(int, silc_object_t *);
typedef int (*silc_format_send)(int, silc_object_t *);
typedef int (*silc_format_free)(silc_object_t *);
typedef int (*silc_format_dup)(silc_object_t *, silc_object_t *, int);
typedef int (*silc_format_pconv)(silc_object_t *, int);
typedef int (*silc_format_to_dense)(silc_object_t *, silc_object_t *);
typedef int (*silc_format_to_crs)(silc_object_t *, silc_object_t *);
typedef int (*silc_format_assign)(silc_object_t *, clist_t **, int,
				  silc_object_t *, silc_modifier_t *);
typedef silc_object_t * (*silc_format_subscript)(silc_object_t *, clist_t **);
typedef int (*silc_format_pprint)(FILE *, silc_object_t *, int);

typedef struct {
  silc_format_new        proc_new;
  silc_format_recv       proc_recv;
  silc_format_send       proc_send;
  silc_format_free       proc_free;
  silc_format_dup        proc_dup;
  silc_format_pconv      proc_pconv;      /* precision conversion */
  silc_format_to_dense   proc_to_dense;   /* NULL if dense */
  silc_format_to_crs     proc_to_crs;     /* NULL if dense or CRS */
  silc_format_assign     proc_assign;
  silc_format_subscript  proc_subscript;
  silc_format_pprint     proc_pprint;     /* pretty printer */
} silc_format_t;

/* auxiliary functions */

DECLSPEC int silc_register_format(char *, silc_format_t *);
DECLSPEC const char *silc_find_format_name(int);
DECLSPEC silc_format_t *silc_find_format_proc_table(int);
DECLSPEC int silc_find_format_id(const char *);
DECLSPEC int silc_get_number_of_formats();

DECLSPEC int silc_get_origin(); /* XXX server.c */

#endif /* SILC_MODULE_H */
