/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: net.c,v 1.5 2007/10/31 05:50:18 kajiyama Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#ifdef _WIN32
#include <winsock2.h>
typedef unsigned int uint32_t;
typedef unsigned long long uint64_t;
#else
#ifdef _sx
typedef unsigned long uint64_t;
#else
#include <sys/socket.h>
#include <netinet/in.h>
#endif /* _sx */
#endif /* _WIN32 */

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "net.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

static void (*conv_endian_32)(unsigned char *);
static void (*conv_endian_64)(unsigned char *);
static void (*conv_endian_32_array)(unsigned char *, int);
static void (*conv_endian_64_array)(unsigned char *, int);
static int sizeof_size_t;

static void
reverse_endian_32(unsigned char *p)
{
  unsigned char b0, b1, b2, b3;

  b0 = p[0];
  b1 = p[1];
  b2 = p[2];
  b3 = p[3];
  *p++ = b3;
  *p++ = b2;
  *p++ = b1;
  *p = b0;
}

static void
reverse_endian_64(unsigned char *p)
{
  unsigned char b0, b1, b2, b3, b4, b5, b6, b7;

  b0 = p[0];
  b1 = p[1];
  b2 = p[2];
  b3 = p[3];
  b4 = p[4];
  b5 = p[5];
  b6 = p[6];
  b7 = p[7];
  *p++ = b7;
  *p++ = b6;
  *p++ = b5;
  *p++ = b4;
  *p++ = b3;
  *p++ = b2;
  *p++ = b1;
  *p = b0;
}

static void
reverse_endian_32_array(unsigned char *p, int n)
{
  int i;
  unsigned char b0, b1, b2, b3;

  for (i = 0; i < n; i++) {
    b0 = p[0];
    b1 = p[1];
    b2 = p[2];
    b3 = p[3];
    *p++ = b3;
    *p++ = b2;
    *p++ = b1;
    *p++ = b0;
  }
}

static void
reverse_endian_64_array(unsigned char *p, int n)
{
  int i;
  unsigned char b0, b1, b2, b3, b4, b5, b6, b7;

  for (i = 0; i < n; i++) {
    b0 = p[0];
    b1 = p[1];
    b2 = p[2];
    b3 = p[3];
    b4 = p[4];
    b5 = p[5];
    b6 = p[6];
    b7 = p[7];
    *p++ = b7;
    *p++ = b6;
    *p++ = b5;
    *p++ = b4;
    *p++ = b3;
    *p++ = b2;
    *p++ = b1;
    *p++ = b0;
  }
}

static void straight_endian_32(unsigned char *p){}
static void straight_endian_64(unsigned char *p){}
static void straight_endian_32_array(unsigned char *p, int n){}
static void straight_endian_64_array(unsigned char *p, int n){}

void
set_endian(int endian)
{
  unsigned short n = 0x1234;
  unsigned char *p = (unsigned char *)&n;

#ifdef DEBUG
  printf("set_endian()\n");
  printf("server: %s endian\n", (*p == 0x12) ? "big" : "little");
  printf("client: %s\n",
         (endian == SILC_BIG_ENDIAN) ? "big endian" :
         (endian == SILC_LITTLE_ENDIAN) ? "little endian" : "unknown");
#endif
  if (*p == 0x12 && endian == SILC_BIG_ENDIAN ||
      *p == 0x34 && endian == SILC_LITTLE_ENDIAN) {
    conv_endian_32 = straight_endian_32;
    conv_endian_64 = straight_endian_64;
    conv_endian_32_array = straight_endian_32_array;
    conv_endian_64_array = straight_endian_64_array;
  } else {
    conv_endian_32 = reverse_endian_32;
    conv_endian_64 = reverse_endian_64;
    conv_endian_32_array = reverse_endian_32_array;
    conv_endian_64_array = reverse_endian_64_array;
  }
}

void
set_size(int n)
{
  sizeof_size_t = n;
}

DECLSPEC ssize_t
send_all(int sock, void *buf, size_t len, int flags)
{
  unsigned char *p;
  size_t m;
  ssize_t n;

  p = buf;
  m = len;
  while (m > 0) {
    n = send(sock, p, m, flags);
    if (n < 0)
      return -1;
    p += n;
    m -= n;
  }
  return len;
}

DECLSPEC ssize_t
recv_all(int sock, void *buf, size_t len, int flags)
{
  unsigned char *p;
  size_t m;
  ssize_t n;

  p = buf;
  m = len;
  while (m > 0) {
    n = recv(sock, p, m, flags);
    if (n < 0)
      return -1;
    p += n;
    m -= n;
  }
  return len;
}

DECLSPEC void
send_ack(int sock)
{
  char buf[32] = "ACK";
  send_all(sock, buf, 32, 0);
}

DECLSPEC void
recv_ack(int sock)
{
  char buf[32];
  recv_all(sock, buf, 32, 0);
}

DECLSPEC int
recv_32_array(int sock, void *buf, size_t len, int flags)
{
  unsigned char *p;
  int m, n;

  p = buf;
  while (len > 0) {
    m = (len > 1024) ? 1024 : len;
    n = recv_all(sock, p, m * 4, flags); /* at most 4096 bytes */
    if (n < 0)
      return -1;
#ifdef DEBUG
    if (n < m * 4)
      printf("recv_32_array: received %d out of %d bytes\n", n, m * 4);
#endif
    conv_endian_32_array(p, m);
    len -= m;
    p += n;
  }
  return 0;
}

#if 0 /* experimental */
#include <zlib.h>
#define CHUNK 65536
int
bulk_recv_32_array(int sock, void *buf, size_t len, int flags)
{
  z_stream strm;
  unsigned char in[CHUNK];
  unsigned char *p;
  int status, n;
  size_t size;

#ifdef DEBUG
  printf("recv_32_array: len=%ld\n", len);
#endif
  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  if (inflateInit(&strm) != Z_OK)
    return -1;
  strm.next_in = in;
  strm.avail_in = 0;
  strm.next_out = p = buf;
  strm.avail_out = len * 4;
  for (;;) {
    if (strm.avail_in == 0) {
      if (recv_size(sock, &size, 0) < 0)
	goto inflateError;
      send_ack(sock);
      if (recv_all(sock, in, size, flags) < 0)
	goto inflateError;
      send_ack(sock);
      strm.next_in = in;
      strm.avail_in = size;
    }
    status = inflate(&strm, Z_NO_FLUSH);
    n = (p - strm.next_out) / 4;
    if (n > 0) {
      conv_endian_32_array(p, n);
      p += n * 4;
    }
    if (status == Z_STREAM_END)
      break;
    if (status != Z_OK)
      goto inflateError;
  }
  deflateEnd(&strm);
  recv_ack(sock);
  return 0;

 inflateError:
  deflateEnd(&strm);
  return -1;
}
#endif /* experimental */

DECLSPEC int
recv_64_array(int sock, void *buf, size_t len, int flags)
{
  unsigned char *p;
  int m, n;

  p = buf;
  while (len > 0) {
    m = (len > 512) ? 512 : len;
    n = recv_all(sock, p, m * 8, flags); /* at most 4096 bytes */
    if (n < 0)
      return -1;
#ifdef DEBUG
    if (n < m * 8)
      printf("recv_64_array: received %d out of %d bytes\n", n, m * 8);
#endif
    conv_endian_64_array(p, m);
    len -= m;
    p += n;
  }
  return 0;
}

DECLSPEC int
recv_size(int sock, size_t *p, int flags)
{
  unsigned char buf[8];
  int n;

  n = recv_all(sock, buf, sizeof_size_t, flags);
  if (n < 0)
    return -1;
  switch (sizeof_size_t) {
  case 4:
    conv_endian_32(buf);
    *p = (size_t)(*(uint32_t *)buf);
    break;
  case 8:
    conv_endian_64(buf);
    *p = (size_t)(*(uint64_t *)buf);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  return 0;
}

DECLSPEC int
send_32_array(int sock, void *buf, size_t len, int flags)
{
  unsigned char tmp[4096], *p;
  int m, n;

  p = buf;
  while (len > 0) {
    m = (len > 1024) ? 1024 : len;
    memcpy(tmp, p, m * 4);
    p += m * 4;
    conv_endian_32_array(tmp, m);
    n = send_all(sock, tmp, m * 4, flags); /* at most 4096 bytes */
    if (n < 0)
      return -1;
#ifdef DEBUG
    if (n < m * 4)
      printf("send_32_array: sent %d out of %d bytes\n", n, m * 4);
#endif
    len -= m;
  }
  return 0;
}

DECLSPEC int
send_64_array(int sock, void *buf, size_t len, int flags)
{
  unsigned char tmp[4096], *p;
  int m, n;

  p = buf;
  while (len > 0) {
    m = (len > 512) ? 512 : len;
    memcpy(tmp, p, m * 8);
    p += m * 8;
    conv_endian_64_array(tmp, m);
    n = send_all(sock, tmp, m * 8, flags); /* at most 4096 bytes */
    if (n < 0)
      return -1;
#ifdef DEBUG
    if (n < m * 8)
      printf("send_64_array: sent %d out of %d bytes\n", n, m * 8);
#endif
    len -= m;
  }
  return 0;
}

DECLSPEC int
send_size(int sock, size_t *p, int flags)
{
  unsigned char buf[8];
  int n;

  switch (sizeof_size_t) {
  case 4:
    *(uint32_t *)buf = (uint32_t)*p;
    conv_endian_32(buf);
    break;
  case 8:
    *(uint64_t *)buf = (uint64_t)*p;
    conv_endian_64(buf);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  n = send_all(sock, buf, sizeof_size_t, flags);
  if (n < 0)
    return -1;
  return 0;
}
