/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: object.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "object.h"
#include "interp.h"
#include "module.h"
#include "wtime.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#ifdef TIME_FREE
extern double _free_time; /* server.c */
#endif
#ifdef TIME_DUPLICATE
extern double _duplicate_time; /* server.c */
#endif
#ifdef TIME_VECTOR_CACHE
extern double _vector_cache_time; /* server.c */
#endif

#ifndef VECTOR_CACHE_LIMIT
#define VECTOR_CACHE_LIMIT 100
#endif

#ifdef SINGLE_THREAD
#define init_mutex_vector_cache()
#define lock_mutex_vector_cache()
#define unlock_mutex_vector_cache()
#define destroy_mutex_vector_cache()
#else
#ifdef _WIN32
#include <windows.h>
static HANDLE mutex_vector_cache;
#define init_mutex_vector_cache() { mutex_vector_cache = CreateMutex(NULL, FALSE, NULL); }
#define lock_mutex_vector_cache() WaitForSingleObject(mutex_vector_cache, INFINITE)
#define unlock_mutex_vector_cache() ReleaseMutex(mutex_vector_cache)
#define destroy_mutex_vector_cache() CloseHandle(mutex_vector_cache)
#else
#include <pthread.h>
static pthread_mutex_t mutex_vector_cache;
#define init_mutex_vector_cache() pthread_mutex_init(&mutex_vector_cache, NULL)
#define lock_mutex_vector_cache() pthread_mutex_lock(&mutex_vector_cache)
#define unlock_mutex_vector_cache() pthread_mutex_unlock(&mutex_vector_cache)
#define destroy_mutex_vector_cache() pthread_mutex_destroy(&mutex_vector_cache)
#endif
#endif

typedef struct _vector_cache {
  struct _vector_cache *prev, *next;
  silc_object_t *object;
} vector_cache_t;

static vector_cache_t vector_cache;
static int vector_cache_size;

void
vector_cache_init(void)
{
  vector_cache.next = vector_cache.prev = &vector_cache;
  vector_cache_size = 0;
  init_mutex_vector_cache();
}

void
vector_cache_finalize(void)
{
  destroy_mutex_vector_cache();
}

DECLSPEC int
silc_object_free(silc_object_t *object)
{
  int i;
  silc_format_t *table;
  vector_cache_t *p;
#if defined(TIME_VECTOR_CACHE) || defined(TIME_FREE)
  double t, elapsed;
#endif

#ifdef DEBUG
  printf("silc_object_free(0x%lx)\n", object);
#endif
  switch (object->type) {
  case SILC_COLUMN_VECTOR_TYPE:
  case SILC_ROW_VECTOR_TYPE:
#if VECTOR_CACHE_LIMIT > 0
#ifdef TIME_VECTOR_CACHE
    t = silc_wtime();
#endif
    lock_mutex_vector_cache();
    p = (vector_cache_t *)malloc(sizeof(vector_cache_t));
    p->object = object;
    p->next = &vector_cache;
    p->prev = vector_cache.prev;
    vector_cache.prev->next = p;
    vector_cache.prev = p;
    if (vector_cache_size < VECTOR_CACHE_LIMIT) {
      vector_cache_size++;
      object = NULL;
    } else {
      p = vector_cache.next;
      vector_cache.next = p->next;
      vector_cache.next->prev = &vector_cache;
      object = p->object;
      free(p);
    }
    unlock_mutex_vector_cache();
#ifdef TIME_VECTOR_CACHE
    elapsed = silc_wtime() - t;
    _vector_cache_time += elapsed;
#endif
    if (!object)
      return 0;
#endif
#ifdef TIME_FREE
    t = silc_wtime();
#endif
    free(object->d.vector.v.i);
#ifdef TIME_FREE
    elapsed = silc_wtime() - t;
    _free_time += elapsed;
#ifdef DEBUG
    printf("free vector %e\n", elapsed);
#endif
#endif
    break;
  case SILC_MATRIX_TYPE:
#ifdef TIME_FREE
    t = silc_wtime();
#endif
    table = silc_find_format_proc_table(object->format);
    if (!table)
      return -1;
    if (table->proc_free && table->proc_free(object) < 0) {
      error("proc_free() failed (fatal)");
      return -1;
    }
#ifdef TIME_FREE
    elapsed = silc_wtime() - t;
    _free_time += elapsed;
#ifdef DEBUG
    printf("free matrix %e\n", elapsed);
#endif
#endif
    break;
  case SILC_CUBIC_ARRAY_TYPE:
#ifdef TIME_FREE
    t = silc_wtime();
#endif
    free(object->d.cubic_array.v.i);
#ifdef TIME_FREE
    elapsed = silc_wtime() - t;
    _free_time += elapsed;
#ifdef DEBUG
    printf("free carray %e\n", elapsed);
#endif
#endif
    break;
  }
#ifdef TIME_FREE
  t = silc_wtime();
#endif
  free(object);
#ifdef TIME_FREE
  elapsed = silc_wtime() - t;
  _free_time += elapsed;
#ifdef DEBUG
  printf("free header %e\n", elapsed);
#endif
#endif
  return 0;
}

DECLSPEC silc_object_t *
silc_object_duplicate(silc_object_t *object, int copy)
{
  silc_object_t *duplicated;
  size_t size;
  ssize_t i;
  silc_format_t *table;
#ifdef TIME_DUPLICATE
  double t, elapsed;

  t = silc_wtime();
#endif
  duplicated = (silc_object_t *)malloc(sizeof(silc_object_t));
  duplicated->type = object->type;
  duplicated->format = object->format;
  duplicated->precision = object->precision;
  switch (object->type) {
  case SILC_SCALAR_TYPE:
    if (copy) {
      switch (object->precision) {
      case SILC_INT:
	duplicated->d.scalar.i = object->d.scalar.i; break;
      case SILC_LONG:
	duplicated->d.scalar.l = object->d.scalar.l; break;
      case SILC_FLOAT:
	duplicated->d.scalar.s = object->d.scalar.s; break;
      case SILC_DOUBLE:
	duplicated->d.scalar.d = object->d.scalar.d; break;
      case SILC_COMPLEX:
	duplicated->d.scalar.c[0] = object->d.scalar.c[0]; /* real */
	duplicated->d.scalar.c[1] = object->d.scalar.c[1]; /* image */
	break;
      case SILC_DOUBLE_COMPLEX:
	duplicated->d.scalar.z[0] = object->d.scalar.z[0]; /* real */
	duplicated->d.scalar.z[1] = object->d.scalar.z[1]; /* image */
	break;
      default:
	should_not_reach_here;
      }
    }
    break;
  case SILC_RANGE_TYPE:
    duplicated->d.vector.length = object->d.vector.length;
    duplicated->d.vector.v.range.start = object->d.vector.v.range.start;
    duplicated->d.vector.v.range.end = object->d.vector.v.range.end;
    break;
  case SILC_ROW_VECTOR_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
    size = SILC_SCALAR_SIZE(object) * object->d.vector.length;
    duplicated->d.vector.length = object->d.vector.length;
    duplicated->d.vector.v.i = (int *)malloc(size);
    if (copy) {
      switch (object->precision) {
      case SILC_INT:
	{
	  int32_t *dest = duplicated->d.vector.v.i;
	  int32_t *src = object->d.vector.v.i;
	  ssize_t length = object->d.vector.length;
#pragma omp parallel for
	  for (i = 0; i < length; i++)
	    dest[i] = src[i];
	}
	break;
      case SILC_LONG:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  duplicated->d.vector.v.l[i] = object->d.vector.v.l[i];
	break;
      case SILC_FLOAT:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  duplicated->d.vector.v.s[i] = object->d.vector.v.s[i];
	break;
      case SILC_DOUBLE:
	{
	  double *dest = duplicated->d.vector.v.d;
	  double *src = object->d.vector.v.d;
	  ssize_t length = object->d.vector.length;
#pragma omp parallel for
	  for (i = 0; i < length; i++)
	    dest[i] = src[i];
	}
	break;
      case SILC_COMPLEX:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++) {
	  int ii; ii = i * 2;
	  duplicated->d.vector.v.c[ii]   = object->d.vector.v.c[ii];
	  duplicated->d.vector.v.c[ii+1] = object->d.vector.v.c[ii+1];
	}
	break;
      case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++) {
	  int ii; ii = i * 2;
	  duplicated->d.vector.v.z[ii]   = object->d.vector.v.z[ii];
	  duplicated->d.vector.v.z[ii+1] = object->d.vector.v.z[ii+1];
	}
	break;
      default:
	should_not_reach_here;
      }
#ifdef ENABLE_NUMA
    } else {
      switch (object->precision) {
      case SILC_INT:
	{
	  int32_t *dest = duplicated->d.vector.v.i;
	  ssize_t length = object->d.vector.length;
#pragma omp parallel for
	  for (i = 0; i < length; i++)
	    dest[i] = 0;
	}
	break;
      case SILC_LONG:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  duplicated->d.vector.v.l[i] = 0;
	break;
      case SILC_FLOAT:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++)
	  duplicated->d.vector.v.s[i] = 0.0;
	break;
      case SILC_DOUBLE:
	{
	  double *dest = duplicated->d.vector.v.d;
	  ssize_t length = object->d.vector.length;
#pragma omp parallel for
	  for (i = 0; i < length; i++)
	    dest[i] = 0.0;
	}
	break;
      case SILC_COMPLEX:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++) {
	  int ii; ii = i * 2;
	  duplicated->d.vector.v.c[ii] = duplicated->d.vector.v.c[ii+1] = 0.0;
	}
	break;
      case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
	for (i = 0; i < (ssize_t)object->d.vector.length; i++) {
	  int ii; ii = i * 2;
	  duplicated->d.vector.v.z[ii] = duplicated->d.vector.v.z[ii+1] = 0.0;
	}
	break;
      default:
	should_not_reach_here;
      }
#endif /* ENABLE_NUMA */
    }
    break;
  case SILC_MATRIX_TYPE:
    duplicated->d.matrix.m = object->d.matrix.m;
    duplicated->d.matrix.n = object->d.matrix.n;
    table = silc_find_format_proc_table(object->format);
    if (!table)
      return NULL;
    if (!table->proc_dup) {
      error("proc_dup() not defined (fatal)");
      return NULL;
    }
    if (table->proc_dup(object, duplicated, copy) < 0) {
      error("proc_dup() failed (fatal)");
      return NULL;
    }
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    size = SILC_SCALAR_SIZE(object)
      * object->d.cubic_array.l
      * object->d.cubic_array.m
      * object->d.cubic_array.n;
    duplicated->d.cubic_array.l = object->d.cubic_array.l;
    duplicated->d.cubic_array.m = object->d.cubic_array.m;
    duplicated->d.cubic_array.n = object->d.cubic_array.n;
    duplicated->d.cubic_array.v.i = (int *)malloc(size);
    if (copy)
      memcpy(duplicated->d.cubic_array.v.i, object->d.cubic_array.v.i, size);
    break;
  default:
    should_not_reach_here;
  }
#ifdef TIME_DUPLICATE
  elapsed = silc_wtime() - t;
  _duplicate_time += elapsed;
#endif
#ifdef DEBUG
#ifdef TIME_DUPLICATE
  printf("silc_object_duplicate: %s, %s, %e sec.\n",
	 SILC_TYPE_NAME(object), SILC_SCALAR_NAME(object), elapsed);
#else
  printf("silc_object_duplicate(0x%lx, %d) %s, %s => 0x%lx\n", object, copy,
	 SILC_TYPE_NAME(object), SILC_SCALAR_NAME(object), duplicated);
#endif
#endif
  return duplicated;
}

#ifndef PPRINT_LIMIT
#define PPRINT_LIMIT 100
#endif

DECLSPEC void
silc_object_pprint(silc_object_t *object, int level)
{
  int i, j, k, n;
  silc_format_t *table;

#define indent() { int i;  for (i = 0; i < level; i++) printf("  "); }

  switch (object->type) {
  case SILC_SCALAR_TYPE:
    switch (object->precision) {
    case SILC_INT:
      indent(); printf("int: %d (0x%lx)\n", object->d.scalar.i, object);
      break;
    case SILC_LONG:
      indent(); printf("long: %lld (0x%lx)\n", object->d.scalar.l, object);
      break;
    case SILC_FLOAT:
      indent(); printf("float: %e (0x%lx)\n", object->d.scalar.s, object);
      break;
    case SILC_DOUBLE:
      indent(); printf("double: %e (0x%lx)\n", object->d.scalar.d, object);
      break;
    case SILC_COMPLEX:
      indent(); printf("complex: %e%+ei (0x%lx)\n",
		       object->d.scalar.c[0],
		       object->d.scalar.c[1],
		       object);
      break;
    case SILC_DOUBLE_COMPLEX:
      indent(); printf("double complex: %e%+ei (0x%lx)\n",
		       object->d.scalar.z[0],
		       object->d.scalar.z[1],
		       object);
      break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_RANGE_TYPE:
    indent(); printf("range: %d to %d (0x%lx)\n",
		     object->d.vector.v.range.start,
		     object->d.vector.v.range.end,
		     object);
    break;
  case SILC_ROW_VECTOR_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
    indent(); printf("%s vector, %ld elements of %s (0x%lx)\n",
		     object->type == SILC_ROW_VECTOR_TYPE ? "row" : "column",
		     object->d.vector.length,
		     SILC_SCALAR_NAME(object),
		     object);
    if (object->d.vector.length > PPRINT_LIMIT)
      break;
    for (i = 0; i < object->d.vector.length; i++) {
      indent();
      switch (object->precision) {
      case SILC_INT:
	printf("v.i[%d] = %d\n", i, object->d.vector.v.i[i]); break;
      case SILC_LONG:
	printf("v.l[%d] = %lld\n", i, object->d.vector.v.l[i]); break;
      case SILC_FLOAT:
	printf("v.s[%d] = %e\n", i, object->d.vector.v.s[i]); break;
      case SILC_DOUBLE:
	printf("v.d[%d] = %e\n", i, object->d.vector.v.d[i]); break;
      case SILC_COMPLEX:
	printf("v.c[%d] = %e%+ei\n", i,
	       object->d.vector.v.c[i*2],
	       object->d.vector.v.c[i*2+1]); break;
      case SILC_DOUBLE_COMPLEX:
	printf("v.z[%d] = %e%+ei\n", i,
	       object->d.vector.v.z[i*2],
	       object->d.vector.v.z[i*2+1]); break;
      default:
	should_not_reach_here;
      }
    }
    break;
  case SILC_MATRIX_TYPE:
    indent(); printf("%ld x %ld matrix of %s in \"%s\" format (0x%lx)\n",
		     object->d.matrix.m,
		     object->d.matrix.n,
		     SILC_SCALAR_NAME(object),
		     silc_find_format_name(object->format),
		     object);
    if (object->d.matrix.m * object->d.matrix.n > PPRINT_LIMIT)
      break;
    table = silc_find_format_proc_table(object->format);
    if (table &&
	table->proc_pprint &&
	table->proc_pprint(stdout, object, level) < 0)
      fprintf(stderr, "proc_pprint() failed (ignore)\n");
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    indent(); printf("cubic_array, %ld x %ld x %ld elements of %s (0x%lx)\n",
		     object->d.cubic_array.l,
		     object->d.cubic_array.m,
		     object->d.cubic_array.n,
		     SILC_SCALAR_NAME(object),
		     object);
    if (object->d.cubic_array.l *
	object->d.cubic_array.m *
	object->d.cubic_array.n > PPRINT_LIMIT)
      break;
    n = 0;
    for (i = 0; i < object->d.cubic_array.l; i++) {
      for (k = 0; k < object->d.cubic_array.n; k++) {
	for (j = 0; j < object->d.cubic_array.m; j++) {
	  indent();
	  switch (object->precision) {
	  case SILC_INT:
	    printf("v.i[%d,%d,%d] = %d\n", i, j, k,
		   object->d.cubic_array.v.i[n++]); break;
	  case SILC_LONG:
	    printf("v.l[%d,%d,%d] = %lld\n", i, j, k,
		   object->d.cubic_array.v.l[n++]); break;
	  case SILC_FLOAT:
	    printf("v.s[%d,%d,%d] = %e\n", i, j, k,
		   object->d.cubic_array.v.s[n++]); break;
	  case SILC_DOUBLE:
	    printf("v.d[%d,%d,%d] = %e\n", i, j, k,
		   object->d.cubic_array.v.d[n++]); break;
	  case SILC_COMPLEX:
	    printf("v.c[%d,%d,%d] = %e%+ei\n", i, j, k,
		   object->d.cubic_array.v.c[n],
		   object->d.cubic_array.v.c[n+1]);
	    n += 2;
	    break;
	  case SILC_DOUBLE_COMPLEX:
	    printf("v.z[%d,%d,%d] = %e%+ei\n", i, j, k,
		   object->d.cubic_array.v.z[n],
		   object->d.cubic_array.v.z[n+1]);
	    n += 2;
	    break;
	  default:
	    should_not_reach_here;
	  }
	}
      }
    }
    break;
  default:
    should_not_reach_here;
  }
}

DECLSPEC silc_object_t *
silc_object_new_scalar(int precision)
{
  silc_object_t *object;

  object = (silc_object_t *)malloc(sizeof(silc_object_t));
  object->type = SILC_SCALAR_TYPE;
  object->format = 0;
  object->precision = precision;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_int(int32_t value)
{
  silc_object_t *object;

  object = silc_object_new_scalar(SILC_INT);
  if (object)
    object->d.scalar.i = value;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_long(int64_t value)
{
  silc_object_t *object;

  object = silc_object_new_scalar(SILC_LONG);
  if (object)
    object->d.scalar.l = value;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_float(float value)
{
  silc_object_t *object;

  object = silc_object_new_scalar(SILC_FLOAT);
  if (object)
    object->d.scalar.s = value;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_double(double value)
{
  silc_object_t *object;

  object = silc_object_new_scalar(SILC_DOUBLE);
  if (object)
    object->d.scalar.d = value;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_complex(float real, float image)
{
  silc_object_t *object;

  object = silc_object_new_scalar(SILC_COMPLEX);
  if (object) {
    object->d.scalar.c[0] = real;
    object->d.scalar.c[1] = image;
  }
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_double_complex(double real, double image)
{
  silc_object_t *object;

  object = silc_object_new_scalar(SILC_DOUBLE_COMPLEX);
  if (object) {
    object->d.scalar.z[0] = real;
    object->d.scalar.z[1] = image;
  }
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_vector(int precision, size_t length)
{
  vector_cache_t *p;
  silc_object_t *object;
  ssize_t i;
#ifdef TIME_VECTOR_CACHE
  double t, elapsed;

  t = silc_wtime();
#endif
#if VECTOR_CACHE_LIMIT > 0
  lock_mutex_vector_cache();
#if 0
  {
    size_t size = 0;

    p = vector_cache.next;
    while (p != &vector_cache) {
      object = p->object;
      size += SILC_SCALAR_SIZE(object) * object->d.vector.length;
      p = p->next;
    }
    printf("currently %d vector%s cached, %.3f MB used\n",
	   vector_cache_size, (vector_cache_size > 1) ? "s" : "",
	   size / 1024.0 / 1024.0);
  }
#endif
  p = vector_cache.next;
  while (p != &vector_cache) {
    object = p->object;
    if (object->d.vector.length == length &&
	object->precision == precision) {
      p->next->prev = p->prev;
      p->prev->next = p->next;
      free(p);
      vector_cache_size--;
      goto hit;
    }
    p = p->next;
  }
  object = NULL;
 hit:
  unlock_mutex_vector_cache();
#ifdef TIME_VECTOR_CACHE
  elapsed = silc_wtime() - t;
  _vector_cache_time += elapsed;
#endif
  if (object) {
#ifdef DEBUG
#ifdef TIME_VECTOR_CACHE
    printf("vector_cache: %p reused, %e sec.\n", object, elapsed);
#else
    printf("vector_cache: %p reused\n", object);
#endif
#endif
    return object;
  }
#ifdef DEBUG
#ifdef TIME_VECTOR_CACHE
  printf("vector_cache: miss, %e sec.\n", elapsed);
#else
  printf("vector_cache: miss\n");
#endif
#endif
#endif /* VECTOR_CACHE_LIMIT > 0 */
  object = (silc_object_t *)malloc(sizeof(silc_object_t));
  object->format = 0;
  object->precision = precision;
  object->d.vector.length = length;
  switch (object->precision) {
  case SILC_INT:
    object->d.vector.v.i = (int32_t *)malloc(sizeof(int32_t) * length);
#ifdef ENABLE_NUMA
    {
      int32_t *dest = object->d.vector.v.i;
#pragma omp parallel for
      for (i = 0; i < (ssize_t)length; i++)
	dest[i] = 0;
    }
#endif
    break;
  case SILC_LONG:
    object->d.vector.v.l = (int64_t *)malloc(sizeof(int64_t) * length);
#ifdef ENABLE_NUMA
#pragma omp parallel for
    for (i = 0; i < (ssize_t)length; i++)
      object->d.vector.v.l[i] = 0;
#endif
    break;
  case SILC_FLOAT:
    object->d.vector.v.s = (float *)malloc(sizeof(float) * length);
#ifdef ENABLE_NUMA
#pragma omp parallel for
    for (i = 0; i < (ssize_t)length; i++)
      object->d.vector.v.s[i] = 0.0;
#endif
    break;
  case SILC_DOUBLE:
    object->d.vector.v.d = (double *)malloc(sizeof(double) * length);
#ifdef ENABLE_NUMA
    {
      double *dest = object->d.vector.v.d;
#pragma omp parallel for
      for (i = 0; i < (ssize_t)length; i++)
	dest[i] = 0.0;
    }
#endif
    break;
  case SILC_COMPLEX:
    object->d.vector.v.c = (float *)malloc(sizeof(float) * 2 * length);
#ifdef ENABLE_NUMA
#pragma omp parallel for
    for (i = 0; i < (ssize_t)length; i++) {
      size_t ii; ii = i * 2;
      object->d.vector.v.c[ii] = object->d.vector.v.c[ii+1] = 0.0;
    }
#endif
    break;
  case SILC_DOUBLE_COMPLEX:
    object->d.vector.v.z = (double *)malloc(sizeof(double) * 2 * length);
#ifdef ENABLE_NUMA
#pragma omp parallel for
    for (i = 0; i < (ssize_t)length; i++) {
      size_t ii; ii = i * 2;
      object->d.vector.v.z[ii] = object->d.vector.v.z[ii+1] = 0.0;
    }
#endif
    break;
  default:
    should_not_reach_here;
  }
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_row_vector(int precision, size_t length)
{
  silc_object_t *object;

  object = silc_object_new_vector(precision, length);
  if (object)
    object->type = SILC_ROW_VECTOR_TYPE;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_column_vector(int precision, size_t length)
{
  silc_object_t *object;

  object = silc_object_new_vector(precision, length);
  if (object)
    object->type = SILC_COLUMN_VECTOR_TYPE;
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_column_vector_from_range(silc_object_t *range)
{
  silc_object_t *vector;
  ssize_t i;

  vector = silc_object_new_column_vector(range->precision,
					 range->d.vector.length);
#pragma omp parallel for
  for (i = 0; i < (ssize_t)range->d.vector.length; i++)
    vector->d.vector.v.i[i] = range->d.vector.v.range.start + i;
  return vector;
}

DECLSPEC silc_object_t *
silc_object_new_matrix(int precision, int format, size_t m, size_t n)
{
  silc_object_t *object;

  object = (silc_object_t *)malloc(sizeof(silc_object_t));
  object->type = SILC_MATRIX_TYPE;
  object->format = format;
  object->precision = precision;
  object->d.matrix.m = m; /* row */
  object->d.matrix.n = n; /* column */
  object->d.matrix.in = NULL; /* to be set by a storage module procedure */
  return object;
}

DECLSPEC silc_object_t *
silc_object_new_cubic_array(int precision, size_t l, size_t m, size_t n)
{
  silc_object_t *object;

  object = (silc_object_t *)malloc(sizeof(silc_object_t));
  object->type = SILC_CUBIC_ARRAY_TYPE;
  object->format = 0;
  object->precision = precision;
  object->d.cubic_array.l = l; /* number of planes */
  object->d.cubic_array.m = m; /* number of rows */
  object->d.cubic_array.n = n; /* number of columns */
  object->d.cubic_array.v.i =
    (int *)malloc(SILC_SCALAR_SIZE(object) * l * m * n);
  return object;
}
