/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: object.h,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#ifndef SILC_OBJECT_H
#define SILC_OBJECT_H 1

#include <stdio.h>
#include <stdlib.h> /* for size_t */
#include <sys/types.h> /* for int32_t and int64_t */
#ifdef _sx
typedef long int64_t;
#else
#ifdef _WIN32
typedef int int32_t;
typedef long long int64_t;
#endif /* _WIN32 */
#endif /* _sx */

#include "const.h"
#include "clist.h"

typedef struct {
  int type;
  int precision;
  int format;
  int reserved[5];
  union {
    union {
      float s;
      float c[2];
      double d;
      double z[2];
      int32_t i;
      int64_t l;
    } scalar;
    struct {
      size_t length;
      union {
	float *s;
	float *c;
	double *d;
	double *z;
	int32_t *i;
	int64_t *l;
	struct {
	  int start, end;
	} range;
      } v;
    } vector;
    struct {
      size_t m; /* number of rows */
      size_t n; /* number of columns */
      void *in;
    } matrix;
    struct {
      size_t l, m, n;
      union {
	float *s;
	float *c;
	double *d;
	double *z;
	int32_t *i;
	int64_t *l;
      } v;
    } cubic_array;
  } d;
} silc_object_t;

#define SILC_TYPE_NAME(o) \
 (((o)->type == SILC_SCALAR_TYPE) ? "scalar" : \
  ((o)->type == SILC_RANGE_TYPE) ? "range" : \
  ((o)->type == SILC_COLUMN_VECTOR_TYPE) ? "column vector" : \
  ((o)->type == SILC_ROW_VECTOR_TYPE) ? "row vector" : \
  ((o)->type == SILC_MATRIX_TYPE) ? "matrix" : \
  ((o)->type == SILC_CUBIC_ARRAY_TYPE) ? "cubic array" : "unknown")

#define SILC_SCALAR_NAME(o) \
 (((o)->precision == SILC_INT) ? "int" : \
  ((o)->precision == SILC_LONG) ? "long" : \
  ((o)->precision == SILC_FLOAT) ? "float" : \
  ((o)->precision == SILC_DOUBLE) ? "double" : \
  ((o)->precision == SILC_COMPLEX) ? "complex" : \
  ((o)->precision == SILC_DOUBLE_COMPLEX) ? "double complex" : "unknown")

#define SILC_SCALAR_SIZE(o) \
 (((o)->precision == SILC_INT) ? sizeof(int32_t) : \
  ((o)->precision == SILC_LONG) ? sizeof(int64_t) : \
  ((o)->precision == SILC_FLOAT) ? sizeof(float) : \
  ((o)->precision == SILC_DOUBLE) ? sizeof(double) : \
  ((o)->precision == SILC_COMPLEX) ? (sizeof(float) * 2) : \
  ((o)->precision == SILC_DOUBLE_COMPLEX) ? (sizeof(double) * 2) : 0)

#if defined(_WIN32) && !defined(STATIC_MODULES)
#ifdef _EXPORTING
#define DECLSPEC __declspec(dllexport)
#else
#define DECLSPEC __declspec(dllimport)
#endif
#else
#define DECLSPEC
#endif

DECLSPEC silc_object_t *silc_object_duplicate(silc_object_t *, int);
DECLSPEC int silc_object_free(silc_object_t *);
DECLSPEC void silc_object_pprint(silc_object_t *, int);

DECLSPEC silc_object_t *silc_object_new_scalar(int);
DECLSPEC silc_object_t *silc_object_new_int(int32_t);
DECLSPEC silc_object_t *silc_object_new_long(int64_t);
DECLSPEC silc_object_t *silc_object_new_float(float);
DECLSPEC silc_object_t *silc_object_new_double(double);
DECLSPEC silc_object_t *silc_object_new_complex(float, float);
DECLSPEC silc_object_t *silc_object_new_double_complex(double, double);
DECLSPEC silc_object_t *silc_object_new_vector(int, size_t);
DECLSPEC silc_object_t *silc_object_new_row_vector(int, size_t);
DECLSPEC silc_object_t *silc_object_new_column_vector(int, size_t);
DECLSPEC silc_object_t *silc_object_new_column_vector_from_range(silc_object_t *);
DECLSPEC silc_object_t *silc_object_new_matrix(int, int, size_t, size_t);
DECLSPEC silc_object_t *silc_object_new_cubic_array(int, size_t, size_t, size_t);

void vector_cache_init(void);
void vector_cache_finalize(void);

#endif /* SILC_OBJECT_H */
