/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: queue.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#ifdef TIME_WAIT
double _wait_time;
#endif

#ifndef SINGLE_THREAD

#include <string.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "interp.h"
#include "queue.h"
#include "wtime.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#ifdef _WIN32
static HANDLE l_program_queue;
static HANDLE c_program_queue;
static HANDLE c2_program_queue;
#else
static pthread_mutex_t l_program_queue;
static pthread_cond_t c_program_queue;
static pthread_cond_t c2_program_queue;
#endif

static program_t program_queue_top, program_queue_last;
static int n_program;

int program_queue_init(void)
{
#ifdef _WIN32
  l_program_queue = CreateMutex(NULL, FALSE, NULL);
  c_program_queue = CreateEvent(NULL, FALSE, FALSE, NULL);
  c2_program_queue = CreateEvent(NULL, FALSE, FALSE, NULL);
#else
  pthread_mutex_init(&l_program_queue, NULL);
  pthread_cond_init(&c_program_queue, NULL);
  pthread_cond_init(&c2_program_queue, NULL);
#endif
  
  program_queue_top.next = &program_queue_last;
  program_queue_last.prev = &program_queue_top;
  program_queue_top.prev = program_queue_last.next = NULL;
 
  n_program = 0;

  return 0;
}

int program_queue_size(void)
{
  return n_program;
}

int program_queue_add(program_t *p)
{
#ifdef _WIN32
  WaitForSingleObject(l_program_queue, INFINITE);
#else
  pthread_mutex_lock(&l_program_queue);
#endif
  p->prev = program_queue_last.prev;
  p->next = &program_queue_last;
  p->prev->next = p;
  program_queue_last.prev = p;
  n_program ++;
#ifdef _WIN32
  ReleaseMutex(l_program_queue);
  SetEvent(c_program_queue);
#else
  pthread_mutex_unlock(&l_program_queue);
  pthread_cond_signal(&c_program_queue);
#endif
  return 0;
}

program_t *program_queue_next(void)
{
  program_t *p;

#ifdef _WIN32
  WaitForSingleObject(l_program_queue, INFINITE);
#else
  pthread_mutex_lock(&l_program_queue);
#endif
  if (n_program == 0) {
#ifdef _WIN32
    ResetEvent(c_program_queue);
    SignalObjectAndWait(l_program_queue, c_program_queue, INFINITE, FALSE);
#else
    pthread_cond_wait(&c_program_queue, &l_program_queue);
#endif
  }
  p = program_queue_top.next;
#ifdef _WIN32
  ReleaseMutex(l_program_queue);
#else
  pthread_mutex_unlock(&l_program_queue);
#endif
  return p;
} 

program_t *program_queue_pop(void)
{
  program_t *p;

#ifdef _WIN32
  WaitForSingleObject(l_program_queue, INFINITE);
#else
  pthread_mutex_lock(&l_program_queue);
#endif
  if (n_program == 0) {
#ifdef _WIN32
    ResetEvent(c_program_queue);
    SignalObjectAndWait(l_program_queue, c_program_queue, INFINITE, FALSE);
#else
    pthread_cond_wait(&c_program_queue, &l_program_queue);
#endif
  }
  p = program_queue_top.next;
  p->next->prev = &program_queue_top;
  program_queue_top.next = p->next;
  p->next = p->prev = NULL;

  n_program --;

#ifdef _WIN32
  ReleaseMutex(l_program_queue);
  SetEvent(c2_program_queue);
#else
  pthread_mutex_unlock(&l_program_queue);
  pthread_cond_signal(&c2_program_queue);
#endif
#if 0
  fprintf(stderr, "pop:%d\n", n_program);
#endif
  return p;
}

#include "interp.h"

int program_queue_wait(const char *name)
{
  program_t *p;
  int noneedtowait = 0;
  int i;
#ifdef TIME_WAIT
  double t;
#endif

#ifdef _WIN32
  WaitForSingleObject(l_program_queue, INFINITE);
#else
  pthread_mutex_lock(&l_program_queue);
#endif
  p = program_queue_last.prev;
  while (p != &program_queue_top) {
    for (i = 0; i < p->n_stmts; i++) {
      if (p->stmts[i]->type == RenameStmtType) {
	if (strcmp(RENAME(p->stmts[i])->old_name, name) == 0 ||
	    strcmp(RENAME(p->stmts[i])->new_name, name) == 0)
	  break;
      } else if (p->stmts[i]->type == AssignmentStmtType) {
	if (strcmp(ASSIGNMENT(p->stmts[i])->name, name) == 0)
	  break;
      } else if (p->stmts[i]->type == ProcCallStmtType) {
	break; /* XXX: implemented yet */
      } else if (p->stmts[i]->type == PreferStmtType) {
	/* no need to wait */
      } else {
	should_not_reach_here;
      }
    }
    if (i != p->n_stmts) break;
    noneedtowait ++;
    p = p->prev;
  }

#ifdef TIME_WAIT
  t = silc_wtime();
#endif

  while (n_program > noneedtowait) { 
    //    fprintf(stderr, "wait:%d\n", n_program);
#ifdef _WIN32
    ResetEvent(c_program_queue);
    SignalObjectAndWait(l_program_queue, c2_program_queue, INFINITE, FALSE);
#else
    pthread_cond_wait(&c2_program_queue, &l_program_queue);
#endif
  }

#ifdef TIME_WAIT
  _wait_time += silc_wtime() - t;
#endif

#ifdef _WIN32
  ReleaseMutex(l_program_queue);
#else
  pthread_mutex_unlock(&l_program_queue);
#endif

  return 0;
}

int
program_queue_finalize(void)
{
#ifdef _WIN32
  CloseHandle(l_program_queue);
  CloseHandle(c_program_queue);
  CloseHandle(c2_program_queue);
#else
  pthread_mutex_destroy(&l_program_queue);
  pthread_cond_destroy(&c_program_queue);
  pthread_cond_destroy(&c2_program_queue);
#endif
#ifdef DEBUG
  printf("program queue finalized\n");
#endif
  return 0;
}

#endif /* SINGLE_THREAD */
