/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: server.c,v 1.7 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <malloc.h>
#include <limits.h> /* for PATH_MAX */
#include <sys/types.h>
#ifdef _WIN32
#include <winsock2.h>
#define SHUT_WR SD_SEND
typedef int socklen_t;
#else
#define closesocket close
#ifdef _sx
#define SHUT_WR 1
typedef int socklen_t;
#endif
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif /* _WIN32 */

#ifndef SINGLE_THREAD
#ifdef _WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif
#endif
#ifdef _OPENMP
#include <omp.h>
#ifdef _sgi
#include <cpumemsets.h>
#endif
#endif
#ifdef VANILLA_MPI
#include <mpi.h>
#endif
#include <math.h>

#ifndef MODULE_DIR
#define MODULE_DIR "./modules/formats:./modules"
#endif

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "interp.h"
#include "net.h"
#include "log.h"
#include "queue.h"
#include "wtime.h"

#include "dense.h" /* for recv_object() */
#ifdef COMMTEST
#include "sparse_crs.h"
#include "sparse_band.h"
#endif

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

silc_object_t *
recv_object(int sock)
{
  silc_object_t header, *object, *tmp;
  size_t size;
  size_t length, l, m, n;
  int i, j, k, format;
  silc_format_t *table;
  silc_dense_t *in;

  send_ack(sock);
  recv_32_array(sock, &header, 8, 0);
  send_ack(sock);
  switch (header.type) {
  case SILC_SCALAR_TYPE:
    object = silc_object_new_scalar(header.precision);
    switch (object->precision) {
    case SILC_INT:
      recv_32_array(sock, &object->d.scalar.i, 1, 0); break;
    case SILC_LONG:
      recv_64_array(sock, &object->d.scalar.l, 1, 0); break;
    case SILC_FLOAT:
      recv_32_array(sock, &object->d.scalar.s, 1, 0); break;
    case SILC_DOUBLE:
      recv_64_array(sock, &object->d.scalar.d, 1, 0); break;
    case SILC_COMPLEX:
      recv_32_array(sock, &object->d.scalar.c, 2, 0); break;
    case SILC_DOUBLE_COMPLEX:
      recv_64_array(sock, &object->d.scalar.z, 2, 0); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_ROW_VECTOR_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
    recv_size(sock, &length, 0);
    send_ack(sock);
    object = silc_object_new_vector(header.precision, length);
    object->type = header.type;
#ifdef ENDIANTEST_BIG
    set_endian(SILC_LITTLE_ENDIAN);
#endif
#ifdef ENDIANTEST_LITTLE
    set_endian(SILC_BIG_ENDIAN);
#endif
    switch (object->precision) {
    case SILC_INT:
      recv_32_array(sock, object->d.vector.v.i, length, 0); break;
    case SILC_LONG:
      recv_64_array(sock, object->d.vector.v.l, length, 0); break;
    case SILC_FLOAT:
      recv_32_array(sock, object->d.vector.v.s, length, 0); break;
    case SILC_DOUBLE:
      recv_64_array(sock, object->d.vector.v.d, length, 0); break;
    case SILC_COMPLEX:
      recv_32_array(sock, object->d.vector.v.c, length * 2, 0); break;
    case SILC_DOUBLE_COMPLEX:
      recv_64_array(sock, object->d.vector.v.z, length * 2, 0); break;
    default:
      should_not_reach_here;
    }
#ifdef ENDIANTEST_BIG
    set_endian(SILC_BIG_ENDIAN);
#endif
#ifdef ENDIANTEST_LITTLE
    set_endian(SILC_LITTLE_ENDIAN);
#endif
    break;
  case SILC_MATRIX_TYPE:
    recv_size(sock, &m, 0); send_ack(sock);
    recv_size(sock, &n, 0); send_ack(sock);
    table = silc_find_format_proc_table(header.format);
    if (!table) {
      fprintf(stderr, "recv_object: proc table not fount\n");
      return NULL;
    }
    if (!table->proc_recv) {
      fprintf(stderr, "recv_object: proc_recv() not defined\n");
      return NULL;
    }
    object = silc_object_new_matrix(header.precision, header.format, m, n);
    if (table->proc_recv(sock, object) < 0) {
      fprintf(stderr, "recv_object: proc_recv() failed\n");
      silc_object_free(object);
      return NULL;
    }
    if (m == 1 || n == 1) {
      format = silc_find_format_id(SILC_FORMAT_DENSE);
      if (object->format == format) {
	tmp = object;
      } else {
	if (table->proc_to_dense == NULL) {
	  error("recv_object: proc_to_dense() not found");
	  silc_object_free(object);
	  return NULL;
	}
	tmp = silc_object_new_matrix(object->precision, format,
				     object->d.matrix.m,
				     object->d.matrix.n);
	if (table->proc_to_dense(object, tmp) < 0) {
	  error("recv_object: proc_to_dense() failed");
	  silc_object_free(object);
	  silc_object_free(tmp);
	  return NULL;
	}
	silc_object_free(object);
      }
      in = SILC_DENSE(tmp);
      if (m == 1 && n == 1) {
	switch (object->precision) {
	case SILC_INT:
	  object = silc_object_new_int(in->v.i[0]);
	  break;
	case SILC_LONG:
	  object = silc_object_new_long(in->v.l[0]);
	  break;
	case SILC_FLOAT:
	  object = silc_object_new_float(in->v.s[0]);
	  break;
	case SILC_DOUBLE:
	  object = silc_object_new_double(in->v.d[0]);
	  break;
	case SILC_COMPLEX:
	  object = silc_object_new_complex(in->v.c[0], in->v.c[1]);
	  break;
	case SILC_DOUBLE_COMPLEX:
	  object = silc_object_new_double_complex(in->v.z[0], in->v.z[1]);
	  break;
	default:
	  should_not_reach_here;
	}
      } else {
	if (m == 1)
	  object = silc_object_new_row_vector(tmp->precision, n);
	else
	  object = silc_object_new_column_vector(tmp->precision, m);
	switch (object->precision) {
	case SILC_INT:
	  for (i = 0; i < object->d.vector.length; i++)
	    object->d.vector.v.i[i] = in->v.i[i];
	  break;
	case SILC_LONG:
	  for (i = 0; i < object->d.vector.length; i++)
	    object->d.vector.v.l[i] = in->v.l[i];
	  break;
	case SILC_FLOAT:
	  for (i = 0; i < object->d.vector.length; i++)
	    object->d.vector.v.s[i] = in->v.s[i];
	  break;
	case SILC_DOUBLE:
	  for (i = 0; i < object->d.vector.length; i++)
	    object->d.vector.v.d[i] = in->v.d[i];
	  break;
	case SILC_COMPLEX:
	  for (i = 0; i < object->d.vector.length * 2; i++)
	    object->d.vector.v.c[i] = in->v.c[i];
	  break;
	case SILC_DOUBLE_COMPLEX:
	  for (i = 0; i < object->d.vector.length * 2; i++)
	    object->d.vector.v.z[i] = in->v.z[i];
	  break;
	default:
	  should_not_reach_here;
	}
      }
      silc_object_free(tmp);
    }
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    recv_size(sock, &l, 0); send_ack(sock);
    recv_size(sock, &m, 0); send_ack(sock);
    recv_size(sock, &n, 0); send_ack(sock);
    object = silc_object_new_cubic_array(header.precision, l, m, n);
    size = object->d.cubic_array.l *
           object->d.cubic_array.m *
           object->d.cubic_array.n;
    switch (object->precision) {
    case SILC_INT:
      recv_32_array(sock, object->d.cubic_array.v.i, size, 0); break;
    case SILC_LONG:
      recv_64_array(sock, object->d.cubic_array.v.l, size, 0); break;
    case SILC_FLOAT:
      recv_32_array(sock, object->d.cubic_array.v.s, size, 0); break;
    case SILC_DOUBLE:
      recv_64_array(sock, object->d.cubic_array.v.d, size, 0); break;
    case SILC_COMPLEX:
      recv_32_array(sock, object->d.cubic_array.v.c, size * 2, 0); break;
    case SILC_DOUBLE_COMPLEX:
      recv_64_array(sock, object->d.cubic_array.v.z, size * 2, 0); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return object;
}

int
send_object(int sock, silc_object_t *object)
{
  size_t size;
  int i, j, *p;
  silc_format_t *table;

  recv_ack(sock);
  send_32_array(sock, object, 8, 0);
  recv_ack(sock);
  switch (object->type) {
  case SILC_SCALAR_TYPE:
    switch (object->precision) {
    case SILC_INT:
      send_32_array(sock, &object->d.scalar.i, 1, 0); break;
    case SILC_LONG:
      send_64_array(sock, &object->d.scalar.l, 1, 0); break;
    case SILC_FLOAT:
      send_32_array(sock, &object->d.scalar.s, 1, 0); break;
    case SILC_DOUBLE:
      send_64_array(sock, &object->d.scalar.d, 1, 0); break;
    case SILC_COMPLEX:
      send_32_array(sock, &object->d.scalar.c, 2, 0); break;
    case SILC_DOUBLE_COMPLEX:
      send_64_array(sock, &object->d.scalar.z, 2, 0); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_ROW_VECTOR_TYPE:
  case SILC_COLUMN_VECTOR_TYPE:
    send_size(sock, &object->d.vector.length, 0);
    recv_ack(sock);    
    size = object->d.vector.length;
#ifdef ENDIANTEST_BIG
    set_endian(SILC_LITTLE_ENDIAN);
#endif
#ifdef ENDIANTEST_LITTLE
    set_endian(SILC_BIG_ENDIAN);
#endif
    switch (object->precision) {
    case SILC_INT:
      send_32_array(sock, object->d.vector.v.i, size, 0); break;
    case SILC_LONG:
      send_64_array(sock, object->d.vector.v.l, size, 0); break;
    case SILC_FLOAT:
      send_32_array(sock, object->d.vector.v.s, size, 0); break;
    case SILC_DOUBLE:
      send_64_array(sock, object->d.vector.v.d, size, 0); break;
    case SILC_COMPLEX:
      send_32_array(sock, object->d.vector.v.c, size * 2, 0); break;
    case SILC_DOUBLE_COMPLEX:
      send_64_array(sock, object->d.vector.v.z, size * 2, 0); break;
    default:
      should_not_reach_here;
    }
#ifdef ENDIANTEST_BIG
    set_endian(SILC_BIG_ENDIAN);
#endif
#ifdef ENDIANTEST_LITTLE
    set_endian(SILC_LITTLE_ENDIAN);
#endif
    break;
  case SILC_MATRIX_TYPE:
    send_size(sock, &object->d.matrix.m, 0); recv_ack(sock);    
    send_size(sock, &object->d.matrix.n, 0); recv_ack(sock);    
    table = silc_find_format_proc_table(object->format);
    if (!table)
      return -1;
    if (!table->proc_send) {
      fprintf(stderr, "proc_send() not defined (fatal)\n");
      return -1;
    }
    if (table->proc_send(sock, object) < 0) {
      fprintf(stderr, "proc_send() failed (fatal)\n");
      return -1;
    }
    break;
  case SILC_CUBIC_ARRAY_TYPE:
    send_size(sock, &object->d.cubic_array.l, 0); recv_ack(sock);    
    send_size(sock, &object->d.cubic_array.m, 0); recv_ack(sock);    
    send_size(sock, &object->d.cubic_array.n, 0); recv_ack(sock);    
    size = object->d.cubic_array.l *
           object->d.cubic_array.m *
           object->d.cubic_array.n;
    switch (object->precision) {
    case SILC_INT:
      send_32_array(sock, object->d.cubic_array.v.i, size, 0); break;
    case SILC_LONG:
      send_64_array(sock, object->d.cubic_array.v.l, size, 0); break;
    case SILC_FLOAT:
      send_32_array(sock, object->d.cubic_array.v.s, size, 0); break;
    case SILC_DOUBLE:
      send_64_array(sock, object->d.cubic_array.v.d, size, 0); break;
    case SILC_COMPLEX:
      send_32_array(sock, object->d.cubic_array.v.c, size * 2, 0); break;
    case SILC_DOUBLE_COMPLEX:
      send_64_array(sock, object->d.cubic_array.v.z, size * 2, 0); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
#ifdef COMMTEST
  recv_ack(sock);
#endif
  return 0;
}

#define COLUMN_MAJOR 1
#define ROW_MAJOR    2

int origin, array_format;

void
set_array_format(int format)
{
#ifdef DEBUG_FRONTEND
  printf("set_array_format(%s major)\n",
	 (format == ROW_MAJOR) ? "row" : "column");
#endif
  array_format = format;
}

void
set_origin(int n)
{
#ifdef DEBUG_FRONTEND
  printf("set_origin(%d)\n", n);
#endif
  origin = n;
}

DECLSPEC int
silc_get_origin()
{
  return origin;
}

static program_t *
create_rename_request(char *old_name, char *new_name)
{
  char buf[7+SILC_MAXNAMELEN+1+SILC_MAXNAMELEN+1];

  sprintf(buf, "rename %s %s", old_name, new_name);
  return parse_program(buf);
}

#ifdef TIME_PARSE
double _parse_time; /* cmd_parser.y */
#endif
#ifdef TIME_WAIT
extern double _wait_time; /* queue.c */
#endif
#ifdef TIME_EXEC_STMT
double _exec_time;
#endif
#ifdef TIME_VAR_LOOKUP
double _var_lookup_time; /* variable.c */
#endif
#ifdef TIME_FIND_MFUNC
double _find_mfunc_time; /* interp.c, modifier.c */
#endif
#ifdef TIME_CALL_MFUNC
double _call_mfunc_time; /* interp.c, modifier.c */
#endif
#ifdef TIME_FREE
double _free_time; /* object.c */
#endif
#ifdef TIME_DUPLICATE
double _duplicate_time; /* object.c */
#endif
#ifdef TIME_VECTOR_CACHE
double _vector_cache_time; /* object.c */
#endif
#ifdef TIME_SUBSCRIPT_VECTOR
double _vector_subsref_time; /* modifier.c */
double _vector_subsasgn_time; /* interp.c */
#endif
#ifdef TIME_BANDED_MV
double _banded_mv_time; /* sparse_band_op.c */
#endif
#ifdef TIME_BANDED_LU
double _banded_lu_time; /* sparse_band_op.c */
#endif
#ifdef TIME_CRS_CG
double _crs_cg_time; /* leq_cg.c */
#endif

int
cmd_loop(int sock)
{
  char buf[32+SILC_PROGRAM_MAX+1], cmd[SILC_PROGRAM_MAX+1];
  char name[SILC_MAXNAMELEN+1], temp[SILC_MAXNAMELEN+1];
  char *p;
  int i, j, n, error;
  size_t size;
  program_t *program;
  silc_object_t *object, header;
#ifdef TIME_EXEC_STMT
  double t, elapsed;
#endif
#ifdef COMMTEST
  double t, elapsed, Mbits;
#endif

  while (1) {
    n = recv(sock, buf, 32, 0);
    if (n <= 0) {
#ifdef DEBUG_FRONTEND
      printf("connection closed\n");
      fflush(stdout);
#endif
      return 1;
    } else if (n != 32) {
      fprintf(stderr, "Error: received %d byte%s.\n", n, (n < 2) ? "" : "s");
      return 1;
    }
#ifdef COMMTEST
    t = silc_wtime();
#endif
    if (buf[0] == 'Q') { /* shutdown */
      return 0;
    } else if (buf[0] == 'B') {
      set_endian(SILC_BIG_ENDIAN);
      send_all(sock, "200 OK\n", 32, 0);
    } else if (buf[0] == 'L') {
      set_endian(SILC_LITTLE_ENDIAN);
      send_all(sock, "200 OK\n", 32, 0);
    } else if (buf[0] == 'R') {
      set_array_format(ROW_MAJOR);
      send_all(sock, "200 OK\n", 32, 0);
    } else if (buf[0] == 'C') {
      set_array_format(COLUMN_MAJOR);
      send_all(sock, "200 OK\n", 32, 0);
    } else if (buf[0] == '0') {
      set_origin(0);
      send_all(sock, "200 OK\n", 32, 0);
    } else if (buf[0] == '1') {
      set_origin(1);
      send_all(sock, "200 OK\n", 32, 0);
    } else if (buf[0] == 'V') {
      p = strchr(buf + 1, '\n');
      if (p)
	*p = '\0';
      n = atoi(buf + 1);
      if (n == 1) {
	send_all(sock, "200 OK\n", 32, 0);
      } else {
	send_all(sock, "400 Bad Request\n", 32, 0);
      }
    } else if (buf[0] == 'S') {
      p = strchr(buf + 1, '\n');
      if (p)
	*p = '\0';
      n = atoi(buf + 1);
      if (n == 4 || n == 8) { /* cf. recv_size() */
	set_size(n);
	send_all(sock, "200 OK\n", 32, 0);
      } else {
	send_all(sock, "400 Bad Request\n", 32, 0);
      }
    } else if (buf[0] == 'T') {
      p = strchr(buf + 1, '\n');
      if (p)
	*p = '\0';
      n = atoi(buf + 1);
      if (n == 1) {
	send_all(sock, "200 OK\n", 32, 0);
      } else {
	send_all(sock, "400 Bad Request\n", 32, 0);
      }
    } else if (buf[0] == 'F') {
      send_all(sock, "200 OK\n", 32, 0);
      n = silc_get_number_of_formats();
      size = 0;
      for (i = 0; i < n; i++)
	size += strlen(silc_find_format_name(i)) + 1;
      send_size(sock, &size, 0);
      p = (char *)malloc(size);
      for (i = 0; i < n; i++) {
        strcpy(p, silc_find_format_name(i));
        strcat(p, "\n");
        send_all(sock, p, strlen(p), 0);
      }
      free(p);
    } else if (buf[0] == '>') { /* PUT */
      sscanf(buf, ">%s\n", name);
#ifdef DEBUG_FRONTEND
      printf("PUT \"%s\":\n", name);
#endif
      object = recv_object(sock);
      if (!object) {
	send_all(sock, "400 Bad Request\n", 32, 0);
      } else {
	send_all(sock, "200 OK\n", 32, 0);
#ifdef SINGLE_THREAD
	if (var_get(name))
	  var_free(name);
	var_add(name, object);
#else
	p = var_add_temp(object, temp);
	program = create_rename_request(p, name);
	program_queue_add(program);
#endif
#ifdef DEBUG_FRONTEND
	silc_object_pprint(object, 1);
#endif
      }
    } else if (buf[0] == '<') { /* GET */
      sscanf(buf, "<%s\n", name);
#ifdef DEBUG_FRONTEND
      printf("GET \"%s\":\n", name);
#endif
#ifndef SINGLE_THREAD
      program_queue_wait(name);
#endif
      object = var_get(name);
      if (!object) {
	send_all(sock, "404 Not Found\n", 32, 0);
      } else {
	send_all(sock, "200 OK\n", 32, 0);
	send_object(sock, object);
#ifdef DEBUG_FRONTEND
	silc_object_pprint(object, 1);
#endif
      }
    } else if (buf[0] == ':') { /* EXEC */
      error = 1;
      p = strchr(buf+1, ':');
      if (p) {
	strncpy(cmd, buf+1, p-buf-1);
	cmd[p-buf-1] = '\0';
	n = atoi(cmd);
	if (n >= 0 && n <= SILC_PROGRAM_MAX) {
	  i = p - buf + 1;
	  if (i + n + 1 > 32)
	    recv_all(sock, buf+32, i+n+1-32, 0);
	  buf[i+n+1] = '\0';
	  p = strchr(buf+i, '\n');
	  if (p) {
	    strncpy(cmd, buf+i, p-buf-i);
	    cmd[p-buf-i] = '\0';
	    program = parse_program(cmd);
	    if (program)
	      error = 0;
	  }
	}
      }
      if (!error) {
#ifdef DEBUG_FRONTEND
	printf("SILC \"%s\"\n", cmd);
#endif
	send_all(sock, "200 OK\n", 32, 0);
#ifdef SINGLE_THREAD
#ifdef TIME_EXEC_STMT
	t = silc_wtime();
#endif
	for (i = 0; i < program->n_stmts; i++)
	  if (execute_statment(program->stmts[i]) < 0)
	    break;
#ifdef TIME_EXEC_STMT
	elapsed = silc_wtime() - t;
	_exec_time += elapsed;
#endif
#if TIME_EXEC_STMT > 1
	printf("exec: \"%s\" in %e sec.\n", program->source, elapsed);
#else
#ifdef DEBUG
	printf("exec: \"%s\" completed.\n", program->source);
#endif
#endif
	free_program(program);
#else
	program_queue_add(program);
#endif
      } else {
	send_all(sock, "400 Bad Request\n", 32, 0);
      }
    } else if (buf[0] == 'Y') { /* latency test */
      int i;
      char buf;
      send_all(sock, "200 OK\n", 32, 0);
      for (i = 0; i < 100; i++) {
	recv(sock, &buf, 1, 0);
	send(sock, &buf, 1, 0);
      }
      printf("latency test: done.\n");
    } else {
      fprintf(stderr, "Error: bad request: \"%s\"\n", buf);
      send_all(sock, "400 Bad Request\n", 32, 0);
    }
#ifdef COMMTEST
    elapsed = silc_wtime() - t;
    if (buf[0] == '>' || buf[0] == '<') {
      switch (object->type) {
      case SILC_SCALAR_TYPE:
	size = sizeof(double);
	printf("scalar");
	break;
      case SILC_ROW_VECTOR_TYPE:
      case SILC_COLUMN_VECTOR_TYPE:
	size = sizeof(double) * object->d.vector.length;
	printf("vector");
	break;
      case SILC_MATRIX_TYPE:
	p = (char *)silc_find_format_name(object->format);
	if (strcmp(p, SILC_FORMAT_DENSE) == 0) {
	  printf("dense");
	  size = sizeof(double) * object->d.matrix.m * object->d.matrix.n;
	} else if (strcmp(p, SILC_FORMAT_BAND) == 0) {
	  silc_band_t *in = SILC_BAND(object);
	  size = sizeof(double) * (in->l + in->u + 1) * object->d.matrix.n;
	  printf("banded");
	} else if (strcmp(p, SILC_FORMAT_CRS) == 0) {
	  silc_crs_t *in = SILC_CRS(object);
	  size = (sizeof(double) * in->nnz +
		  sizeof(int) * (object->d.matrix.m + 1 + in->nnz));
	  printf("CRS");
	} else {
	  should_not_reach_here;
	}
	break;
      default:
	should_not_reach_here;
      }
      Mbits = 8 * (double)size / 1024 / 1024;
      printf(" %s", (buf[0] == '>') ? "PUT" : "GET");
      printf(" %e Mbits, %e sec, %e Mbps\n", Mbits, elapsed, Mbits / elapsed);
      fflush(stdout);
    }
#endif
  }
}

#ifdef STATIC_MODULES
extern void init_coremodule();
extern void init_dense();
extern void init_sparse_crs();
extern void init_sparse_jds();
extern void init_sparse_band();
extern void init_leq_lis();
extern void init_leq_cg();
extern void init_leq_gs();
extern void init_leq_lu();
extern void init_blasmodule();
extern void init_linpackmodule();
extern void init_conditional();
extern void init_kumamodule();

struct _builtin_module {
  void (*init)();
  char name[SILC_MAXNAMELEN+1];
} builtin_modules[] = {
  {init_coremodule, "coremodule"},
  {init_dense, "dense"},
  {init_sparse_crs, "sparse_crs"},
  {init_sparse_jds, "sparse_jds"},
  {init_sparse_band, "sparse_band"},
#ifdef _WIN32
  {init_leq_lis, "leq_lis"},
#endif
  {init_leq_cg, "leq_cg"},
  {init_leq_gs, "leq_gs"},
  {init_leq_lu, "leq_lu"},
  {init_blasmodule, "blasmodule"},
  {init_linpackmodule, "linpackmodule"},
  {init_conditional, "conditional"},
#ifdef KUMA
  {init_kumamodule, "kumamodule"},
#endif
  {NULL}, /* sentinel */
};

int
register_modules()
{
  struct _builtin_module *p;

  for (p = builtin_modules; p->init; p++)
    (*p->init)();
  return 0;
}

#else /* STATIC_MODULES */

int
register_modules()
{
  char *list, *p, path[PATH_MAX];
  size_t size;
  int i, j, num_dirs;

  list = getenv("SILC_MODULE_DIR");
  if (!list)
    list = MODULE_DIR;
  size = strlen(list);
  if (size == 0)
    return 0;
  num_dirs = 1;
  for (p = list; p < list + size; p++) {
    if (*p == ':')
      num_dirs++;
  }
  p = list;
  for (i = 0; i < num_dirs; i++) {
    for (j = 0; p + j < list + size && p[j] != ':'; j++);
    if (j < PATH_MAX) {
      strncpy(path, p, j);
      path[j] = '\0';
      if (load_modules(path) < 0) {
	silc_log("failed to load modules from %s\n", path);
	return -1;
      }
    }
    p += j + 1;
  }
  return 0;
}
#endif /* STATIC_MODULES */

void
sighandler(int signum)
{
  switch (signum) {
  case SIGINT:
#ifdef _WIN32
    WSACleanup();
#endif
#ifdef VANILLA_MPI
    MPI_Finalize();
#endif
    silc_log_finish();
    _exit(1);
    break;
  }
}

#ifndef SINGLE_THREAD
#ifdef _WIN32
unsigned long __stdcall
#else
void *
#endif
backend(void *arg)
{
  program_t *p;
  int i;
#ifdef TIME_EXEC_STMT
  double t, elapsed;
#endif

#ifdef DEBUG
  printf("Starting backend thread (pid: %d).\n", getpid());
#endif
#ifdef _OPENMP
#ifdef _sgi
#pragma omp parallel 
  cpubind(omp_get_thread_num());
#endif
#endif
  while (1) {
    p = program_queue_next();
#ifdef TIME_EXEC_STMT
    t = silc_wtime();
#endif
    for (i = 0; i < p->n_stmts; i++)
      if (execute_statment(p->stmts[i]) < 0)
	break;
#ifdef TIME_EXEC_STMT
    elapsed = silc_wtime() - t;
    _exec_time += elapsed;
#endif
#if TIME_EXEC_STMT > 1
    printf("exec: \"%s\" in %e sec.\n", p->source, elapsed);
#else
#ifdef DEBUG
    printf("exec: \"%s\" completed.\n", p->source);
#endif
#endif
    program_queue_pop();
    free_program(p);
  }
}
#endif /* SINGLE_THREAD */

void *frontend(void *arg)
{
  char host[256];
  FILE *fp;
  char buf[256];
  int s0, s1;
  int i, cont;
  socklen_t len;
  struct sockaddr_in sa, sb;
  char *port;
  int sndbuf;
  socklen_t sizeof_sndbuf;
#ifdef _WIN32
  WSADATA wsaData;

  WSAStartup(MAKEWORD(2, 0), &wsaData);
#endif
  s0 = socket(AF_INET, SOCK_STREAM, 0);
  sa.sin_family = AF_INET;
  sa.sin_port = ((port = getenv("SILC_PORT")) == NULL) ? 0 : htons(atoi(port));
  sa.sin_addr.s_addr = INADDR_ANY;
  memset(sa.sin_zero, 0, sizeof(sa.sin_zero));
  bind(s0, (struct sockaddr *)&sa, sizeof(sa));
  len = sizeof(sb);
  getsockname(s0, (struct sockaddr *)&sb, &len);
  listen(s0, 1);
  gethostname(host, sizeof(host));
#ifdef _WIN32
  sprintf(buf, "%s/.silc", getenv("USERPROFILE"));
#else
  sprintf(buf, "%s/.silc", getenv("HOME"));
#endif
  fp = fopen(buf, "w");
  fprintf(fp, "%s %d\n", host, htons(sb.sin_port));
  fclose(fp);

#if 0
  sndbuf = 32;
  if (setsockopt(s0, SOL_SOCKET, SO_SNDBUF, &sndbuf, sizeof(int)) < 0)
    printf("cannot set SO_SNDBUF to 32\n");
  sizeof_sndbuf = sizeof(int);
  if (getsockopt(s0, SOL_SOCKET, SO_SNDBUF, &sndbuf, &sizeof_sndbuf) < 0)
    printf("cannot get SO_SNDBUF\n");
  else if (sizeof_sndbuf != sizeof(int))
    printf("sndbuf is not `int'\n");
  else
    printf("sndbuf=%d\n", sndbuf);
  {
    struct protoent *proto;
    int status, prototcp;
    status = 1;
    if (proto = getprotobyname("tcp")) 
      prototcp = proto->p_proto;
    else
      prototcp = IPPROTO_TCP;
    if (setsockopt(s0, prototcp, TCP_NODELAY, &status, sizeof(status)) < 0)
      perror("setsockopt");
  }
#endif

#ifdef DEBUG_FRONTEND
  printf("Serving SILC on port %d...\n", htons(sb.sin_port));
#endif
  cont = 1;
  while (cont) {
#ifdef TIME_PARSE
    _parse_time = 0;
#endif
#ifdef TIME_WAIT
    _wait_time = 0;
#endif
#ifdef TIME_EXEC_STMT
    _exec_time = 0;
#endif
#ifdef TIME_VAR_LOOKUP
    _var_lookup_time = 0;
#endif
#ifdef TIME_FIND_MFUNC
    _find_mfunc_time = 0;
#endif
#ifdef TIME_CALL_MFUNC
    _call_mfunc_time = 0;
#endif
#ifdef TIME_FREE
    _free_time = 0;
#endif
#ifdef TIME_DUPLICATE
    _duplicate_time = 0;
#endif
#ifdef TIME_VECTOR_CACHE
    _vector_cache_time = 0;
#endif
#ifdef TIME_SUBSCRIPT_VECTOR
    _vector_subsref_time = _vector_subsasgn_time = 0;
#endif
#ifdef TIME_BANDED_MV
    _banded_mv_time = 0;
#endif
#ifdef TIME_BANDED_LU
    _banded_lu_time = 0;
#endif
#ifdef TIME_CRS_CG
    _crs_cg_time = 0;
#endif

    /* initialize client-specific variables */
    set_size(4);
    set_origin(0);
    set_array_format(COLUMN_MAJOR);

    /* accept connection */
    s1 = accept(s0, (struct sockaddr*)&sb, &len);
#if 0
    {
      struct timeval t;
      socklen_t s = sizeof(struct timeval);
      if (getsockopt(s1, SOL_SOCKET, SO_SNDTIMEO, &t, &s) < 0)
	printf("cannot get SO_SNDTIMEO\n");
      else
	printf("SO_SNDTIMEO=%d.%06d\n", t.tv_sec, t.tv_usec);
    }
#endif
    cont = cmd_loop(s1);
    shutdown(s1, SHUT_WR);
    while (1) {
      i = recv(s1, buf, sizeof(buf), 0);
      if (i <= 0) {
#ifdef DEBUG
        if (i < 0)
          fprintf(stderr, "Error: recv(2) failed while closing client socket (ignored)\n");
#endif
        break;
      }
    }
    closesocket(s1);
#ifdef TIME_PARSE
    printf("_parse_time = %e\n", _parse_time);
#endif
#ifdef TIME_WAIT
    printf("_wait_time  = %e\n", _wait_time);
#endif
#ifdef TIME_EXEC_STMT
    printf("_exec_time  = %e\n", _exec_time);
#endif
#ifdef TIME_VAR_LOOKUP
    printf("_var_lookup_time = %e\n", _var_lookup_time);
#endif
#ifdef TIME_FIND_MFUNC
    printf("_find_mfunc_time = %e\n", _find_mfunc_time);
#endif
#ifdef TIME_CALL_MFUNC
    printf("_call_mfunc_time = %e\n", _call_mfunc_time);
#endif
#ifdef TIME_FREE
    printf("_free_time  = %e\n", _free_time);
#endif
#ifdef TIME_DUPLICATE
    printf("_duplicate_time  = %e\n", _duplicate_time);
#endif
#ifdef TIME_VECTOR_CACHE
    printf("_vector_cache_time = %e\n", _vector_cache_time);
#endif
#ifdef TIME_SUBSCRIPT_VECTOR
    printf("_vector_subsref_time   = %e\n", _vector_subsref_time);
    printf("_vector_subsasgn_time  = %e\n", _vector_subsasgn_time);
#endif
#ifdef TIME_BANDED_MV
    if (_banded_mv_time > 0)
      printf("_banded_mv_time  = %.6lf\n", _banded_mv_time);
#endif
#ifdef TIME_BANDED_LU
    if (_banded_lu_time > 0)
      printf("_banded_lu_time  = %.6lf\n", _banded_lu_time);
#endif
#ifdef TIME_CRS_CG
    if (_crs_cg_time > 0)
      printf("_crs_cg_time  = %.6lf\n", _crs_cg_time);
#endif
#if 0
    fflush(stdout);
#endif

#ifdef DEBUG
    /* check if all temporary variables got freed */
    i = var_count_temp();
    if (i > 0)
      printf("*** %d temporary variable%s leaked\n", i, (i < 2) ? "" : "s");
#endif

#if 0
    var_free_all(); /* XXX */
#endif
  }
  closesocket(s0);
#ifdef _WIN32
  WSACleanup();
#endif

  silc_log_finish();

  return NULL;
}

#ifdef VANILLA_MPI
int
silc_ipc_master()
{
  int rank, status;

  status = MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  if (status != MPI_SUCCESS) {
    error("silc_ipc_master: MPI_Comm_rank() failed (abort)\n");
    MPI_Finalize();
    exit(1);
  }
  return rank == 0;
}

void
silc_ipc_init(char *module, char *mfunc)
{
  int status;

  status = MPI_Bcast(module, SILC_MAXNAMELEN+1, MPI_CHARACTER, 0,
		     MPI_COMM_WORLD);
  if (status != MPI_SUCCESS) {
    MPI_Finalize();
    error("silc_ipc_init: MPI_Bcast(module) failed (abort)\n");
    exit(1);
  }
  status = MPI_Bcast(mfunc, SILC_MAXNAMELEN+1, MPI_CHARACTER, 0,
		     MPI_COMM_WORLD);
  if (status != MPI_SUCCESS) {
    MPI_Finalize();
    error("silc_ipc_init: MPI_Bcast(mfunc) failed (abort)\n");
    exit(1);
  }
}

void
silc_ipc_finalize()
{
  int status;

  status = MPI_Barrier(MPI_COMM_WORLD);
  if (status != MPI_SUCCESS) {
    MPI_Finalize();
    error("silc_ipc_finalize: MPI_Barrier() failed (abort)\n");
    exit(1);
  }
}

static void
ipc_worker()
{
  char module[SILC_MAXNAMELEN+1], mfunc[SILC_MAXNAMELEN+1];
  silc_mfunc_impl_t impl;
  int status;

  for (;;) {
    silc_ipc_init(module, mfunc);
    impl = get_module_function(module, mfunc);
    if (!impl) {
      error("\"%s\" not found in module \"%s\"\n", mfunc, module);
      continue;
    }
    status = impl(0, NULL);
    if (status < 0) {
      error("module function %s in %s failed\n", mfunc, module);
      continue;
    }
  }
}
#endif /* VANILLA_MPI */

int main(int argc, char **argv)
{
#ifndef SINGLE_THREAD
#ifdef _WIN32
  unsigned long bid;
  HANDLE handle;
#else
  pthread_t bid;
#endif
#endif

  signal(SIGINT, sighandler);
#ifdef DEBUG
#ifdef SINGLE_THREAD
  printf("single thread\n");
#endif
#ifdef STATIC_MODULES
  printf("static modules\n");
#endif
#ifdef COMMTEST
  printf("COMMTEST enabled\n");
#endif
#endif
#ifdef _OPENMP
#ifdef DEBUG
  printf("%d processors\n", omp_get_num_procs());
#pragma omp parallel 
  printf("pid: %d\n", getpid());
#endif
#ifdef _sgi
#pragma omp parallel 
  cpubind(omp_get_thread_num());
#endif
#ifdef ENABLE_NUMA
#ifdef DEBUG
  printf("NUMA support enabled\n");
#endif
#endif
#endif
#ifdef VANILLA_MPI
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) {
    fprintf(stderr, "Error: MPI_Init() failed (abort)\n");
    return -1;
  }
#endif
  var_init();
  vector_cache_init();
  init_module_stack();
  if (register_modules() < 0)
    error("some of the modules may not have been loaded\n");
#ifdef VANILLA_MPI
  silc_ipc_finalize(); /* XXX */
  if (silc_ipc_master()) {
#ifdef DEBUG
    printf("ipc: master\n");
#endif
  } else {
#ifdef DEBUG
    int rank;
    if (MPI_Comm_rank(MPI_COMM_WORLD, &rank) == MPI_SUCCESS)
      printf("ipc: worker #%d\n", rank);
#endif
    ipc_worker();
  }
#endif
  if (silc_log_init() < 0)
    fprintf(stderr, "Error: cannot open log file (ignored)\n");
#ifdef SINGLE_THREAD
  frontend(NULL);
#else /* SINGLE_THREAD */
  program_queue_init();
#if 1
#ifdef _WIN32
  handle = CreateThread(NULL, 0, backend, NULL, 0, &bid);
#else
  pthread_create(&bid, NULL, backend, NULL);
#endif
  frontend(NULL);
#ifdef _WIN32
  TerminateThread(handle, 0);
  CloseHandle(handle);
#else
  pthread_detach(bid);
#endif
#else
#pragma omp parallel sections
  {
#pragma omp section
    {
      backend(NULL);
    }
#pragma omp section
    {
      frontend(NULL);
    }
  }
#endif
  program_queue_finalize();
  vector_cache_finalize();
  var_finalize();
#endif /* SINGLE_THREAD */
  return 0;
}
