/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sparse_band_format.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "object.h"
#include "sparse_band.h"
#include "module.h"
#include "net.h"

#include "dense.h"       /* for proc_to_dense() */
#include "sparse_crs.h"  /* for proc_to_crs() */

#define MODULE_NAME "sparse_band"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#define _max(x, y) (((x) > (y)) ? (x) : (y))
#define _min(x, y) (((x) < (y)) ? (x) : (y))

silc_band_t *
allocate(int precision, size_t l, size_t u, size_t n)
{
  silc_band_t *in;
  size_t size;

  in = (silc_band_t *)malloc(sizeof(silc_band_t));
  in->l = l;
  in->u = u;
  size = (l + u + 1) * n;
  switch (precision) {
  case SILC_INT:
    in->v.i = (int32_t *)malloc(sizeof(int32_t) * size); break;
  case SILC_LONG:
    in->v.l = (int64_t *)malloc(sizeof(int64_t) * size); break;
  case SILC_FLOAT:
    in->v.s = (float *)malloc(sizeof(float) * size); break;
  case SILC_DOUBLE:
    in->v.d = (double *)malloc(sizeof(double) * size); break;
  case SILC_COMPLEX:
    in->v.c = (float *)malloc(sizeof(float) * 2 * size); break;
  case SILC_DOUBLE_COMPLEX:
    in->v.z = (double *)malloc(sizeof(double) * 2 * size); break;
  default:
    should_not_reach_here;
  }
  return in;
}

static int
proc_new(silc_object_t *object)
{
  silc_band_t *in;

  in = allocate(object->precision, 0, 0, object->d.matrix.n);
  if (!in)
    return -1;
  switch (object->precision) {
  case SILC_INT:
    memset(in->v.i, 0, object->d.matrix.n * sizeof(int32_t)); break;
  case SILC_LONG:
    memset(in->v.l, 0, object->d.matrix.n * sizeof(int64_t)); break;
  case SILC_FLOAT:
    memset(in->v.s, 0, object->d.matrix.n * sizeof(float)); break;
  case SILC_DOUBLE:
    memset(in->v.d, 0, object->d.matrix.n * sizeof(double)); break;
  case SILC_COMPLEX:
    memset(in->v.c, 0, object->d.matrix.n * sizeof(float) * 2); break;
  case SILC_DOUBLE_COMPLEX:
    memset(in->v.z, 0, object->d.matrix.n * sizeof(double) * 2); break;
  default:
    should_not_reach_here;
  }
  object->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_recv(int sock, silc_object_t *object)
{
  silc_band_t *in;
  size_t l, u, size;

  recv_size(sock, &l, 0); send_ack(sock);
  recv_size(sock, &u, 0); send_ack(sock);
  in = allocate(object->precision, l, u, object->d.matrix.n);
  if (!in)
    return -1;
  size = (l + u + 1) * object->d.matrix.n;
  switch (object->precision) {
  case SILC_INT:
    recv_32_array(sock, in->v.i, size, 0); break;
  case SILC_LONG:
    recv_64_array(sock, in->v.l, size, 0); break;
  case SILC_FLOAT:
    recv_32_array(sock, in->v.s, size, 0); break;
  case SILC_DOUBLE:
    recv_64_array(sock, in->v.d, size, 0); break;
  case SILC_COMPLEX:
    recv_32_array(sock, in->v.c, size * 2, 0); break;
  case SILC_DOUBLE_COMPLEX:
    recv_64_array(sock, in->v.z, size * 2, 0); break;
  default:
    should_not_reach_here;
  }
  object->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_send(int sock, silc_object_t *object)
{
  silc_band_t *in;
  size_t size;

  in = SILC_BAND(object);
  send_size(sock, &in->l, 0); recv_ack(sock);
  send_size(sock, &in->u, 0); recv_ack(sock);
  size = (in->l + in->u + 1) * object->d.matrix.n;
  switch (object->precision) {
  case SILC_INT:
    send_32_array(sock, in->v.i, size, 0); break;
  case SILC_LONG:
    send_64_array(sock, in->v.l, size, 0); break;
  case SILC_FLOAT:
    send_32_array(sock, in->v.s, size, 0); break;
  case SILC_DOUBLE:
    send_64_array(sock, in->v.d, size, 0); break;
  case SILC_COMPLEX:
    send_32_array(sock, in->v.c, size * 2, 0); break;
  case SILC_DOUBLE_COMPLEX:
    send_64_array(sock, in->v.z, size * 2, 0); break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_free(silc_object_t *object)
{
  silc_band_t *in;

  in = SILC_BAND(object);
  if (!in)
    return -1;
  free(in->v.i);
  free(in);
  return 0;
}

static int
proc_dup(silc_object_t *object, silc_object_t *duplicated, int copy)
{
  silc_band_t *in, *new_in;
  size_t i, size;

  in = SILC_BAND(object);
  new_in = allocate(object->precision, in->l, in->u, object->d.matrix.n);
  if (!new_in)
    return -1;
  if (copy) {
    size = (in->l + in->u + 1) * object->d.matrix.n;
    switch (object->precision) {
    case SILC_INT:
      for (i = 0; i < size; i++)
	new_in->v.i[i] = in->v.i[i];
      break;
    case SILC_LONG:
      for (i = 0; i < size; i++)
	new_in->v.l[i] = in->v.l[i];
      break;
    case SILC_FLOAT:
      for (i = 0; i < size; i++)
	new_in->v.s[i] = in->v.s[i];
      break;
    case SILC_DOUBLE:
      for (i = 0; i < size; i++)
	new_in->v.d[i] = in->v.d[i];
      break;
    case SILC_COMPLEX:
      for (i = 0; i < size * 2; i++)
	new_in->v.c[i] = in->v.c[i];
      break;
    case SILC_DOUBLE_COMPLEX:
      for (i = 0; i < size * 2; i++)
	new_in->v.z[i] = in->v.z[i];
      break;
    default:
      should_not_reach_here;
    }
  }
  duplicated->d.matrix.in = (void *)new_in;
  return 0;
}

static int
proc_pconv(silc_object_t *object, int new_precision)
{
  silc_band_t *in, temp;
  size_t i, size;
  
#define _pconv11(src,dest,type) { \
    for (i = 0; i < size; i++)				\
      temp.v.dest[i] = (type)in->v.src[i];		\
    free(in->v.src);					\
    in->v.dest = temp.v.dest;				\
  }
#define _pconv12(src,dest,type) { \
    for (i = 0; i < size; i++) {			\
      temp.v.dest[i*2]   = (type)in->v.src[i];		\
      temp.v.dest[i*2+1] = (type)0;			\
    }							\
    free(in->v.src);					\
    in->v.dest = temp.v.dest;				\
  }
#define _pconv22(src,dest,type) { \
    for (i = 0; i < size; i++) {			\
      temp.v.dest[i*2]   = (type)in->v.src[i*2];	\
      temp.v.dest[i*2+1] = (type)in->v.src[i*2+1];	\
    }							\
    free(in->v.src);					\
    in->v.dest = temp.v.dest;				\
  }

  in = SILC_BAND(object);
  size = (in->l + in->u + 1) * object->d.matrix.n;
  switch (new_precision) {
  case SILC_LONG:
    temp.v.l = (int64_t *)malloc(sizeof(int64_t) * size); break;
  case SILC_FLOAT:
    temp.v.s = (float *)malloc(sizeof(float) * size); break;
  case SILC_DOUBLE:
    temp.v.d = (double *)malloc(sizeof(double) * size); break;
  case SILC_COMPLEX:
    temp.v.c = (float *)malloc(sizeof(float) * 2 * size); break;
  case SILC_DOUBLE_COMPLEX:
    temp.v.z = (double *)malloc(sizeof(double) * 2 * size); break;
  default:
    should_not_reach_here;
  }
  switch (object->precision) {
  case SILC_INT:
    switch (new_precision) {
    case SILC_LONG:
      _pconv11(i, l, int64_t); break;
    case SILC_FLOAT:
      _pconv11(i, s, float); break;
    case SILC_DOUBLE:
      _pconv11(i, d, double); break;
    case SILC_COMPLEX:
      _pconv12(i, c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(i, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (new_precision) {
    case SILC_DOUBLE:
      _pconv11(l, d, double); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(l, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (new_precision) {
    case SILC_DOUBLE:
      _pconv11(s, d, double); break;
    case SILC_COMPLEX:
      _pconv12(s, c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(s, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _pconv12(d, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _pconv22(c, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_to_dense(silc_object_t *object, silc_object_t *new)
{
  silc_band_t *in;
  silc_dense_t *new_in;
  silc_format_t *table;
  size_t ld;
  ssize_t i;

#ifdef DEBUG
  printf(MODULE_NAME ": proc_to_dense(0x%lx, 0x%lx)\n", object, new);
#endif
  table = silc_find_format_proc_table(new->format); /* dense */
  if (!table)
    return -1;
  if (!table->proc_new) {
    fprintf(stderr, MODULE_NAME ": proc_new() not found");
    return -1;
  }
  if (table->proc_new(new) < 0) {
    fprintf(stderr, MODULE_NAME ": proc_new() failed");
    return -1;
  }
  in = SILC_BAND(object);
  ld = in->l + in->u + 1;
  new_in = SILC_DENSE(new);

#define _conv_to_dense_1(pre) { \
    size_t start, end, new_k, k, j;			\
    start = _max(0, i - (ssize_t)in->u);		\
    end = _min(i + in->l + 1, object->d.matrix.m);	\
    new_k = start + object->d.matrix.m * i;		\
    k = _max(0, (ssize_t)in->u - i) + ld * i;		\
    for (j = start; j < end; j++)			\
      new_in->v.pre[new_k++] = in->v.pre[k++];		\
  }
#define _conv_to_dense_2(pre) { \
    size_t start, end, new_k, k, j;			\
    start = _max(0, i - (ssize_t)in->u);		\
    end = _min(i + in->l + 1, object->d.matrix.m);	\
    new_k = (start + object->d.matrix.m * i) * 2;	\
    k = (_max(0, (ssize_t)in->u - i) + ld * i) * 2;	\
    for (j = start; j < end; j++) {			\
      new_in->v.pre[new_k++] = in->v.pre[k++];		\
      new_in->v.pre[new_k++] = in->v.pre[k++];		\
    }							\
  }
  switch (new->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _conv_to_dense_1(i);
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _conv_to_dense_1(l);
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _conv_to_dense_1(s);
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _conv_to_dense_1(d);
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _conv_to_dense_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.n; i++)
      _conv_to_dense_2(z);
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_to_crs(silc_object_t *object, silc_object_t *new)
{
  silc_band_t *in;
  silc_crs_t *new_in;
  silc_format_t *table;
  size_t nnz, ld;
  ssize_t i;

#ifdef DEBUG
  printf(MODULE_NAME ": proc_to_crs(0x%lx, 0x%lx)\n", object, new);
#endif
  table = silc_find_format_proc_table(new->format); /* CRS */
  if (!table)
    return -1;
  if (!table->proc_new) {
    fprintf(stderr, MODULE_NAME ": proc_new() not found");
    return -1;
  }
  if (table->proc_new(new) < 0) {
    fprintf(stderr, MODULE_NAME ": proc_new() failed");
    return -1;
  }
  in = SILC_BAND(object);
  ld = in->l + in->u + 1; /* leading dimension */
  new_in = SILC_CRS(new);
  nnz = 0; /* number of non-zero elements */

#define _count_nonzeros(test) { \
    size_t column, c, j, k;						\
    for (i = 0; i < in->l; i++) {					\
      new_in->row[i] = nnz;						\
      c = _min(in->u + 1 + i, object->d.matrix.n);			\
      k = in->u + i;							\
      for (j = 0; j < c; j++) {						\
	if (test)							\
	  nnz++;							\
	k += ld - 1;							\
      }									\
    }									\
    for (i = in->l; i < object->d.matrix.m; i++) {			\
      new_in->row[i] = nnz;						\
      column = i - in->l;						\
      c = _min(in->u + 1 + in->l, object->d.matrix.n - column);		\
      k = ld * (column + 1) - 1;					\
      for (j = 0; j < c; j++) {						\
	if (test)							\
	  nnz++;							\
	k += ld - 1;							\
      }									\
    }									\
  }
  switch (object->precision) {
  case SILC_INT:
    _count_nonzeros(in->v.i[k]); break;
  case SILC_LONG:
    _count_nonzeros(in->v.l[k]); break;
  case SILC_FLOAT:
    _count_nonzeros(*(int32_t *)&in->v.s[k]); break;
  case SILC_DOUBLE:
    _count_nonzeros(*(int64_t *)&in->v.d[k]); break;
  case SILC_COMPLEX:
    _count_nonzeros(*(int32_t *)&in->v.c[k*2] ||
		    *(int32_t *)&in->v.c[k*2+1]);
    break;
  case SILC_DOUBLE_COMPLEX:
    _count_nonzeros(*(int64_t *)&in->v.z[k*2] ||
		    *(int64_t *)&in->v.z[k*2+1]);
    break;
  default:
    should_not_reach_here;
  }

  new_in->row[i] = new_in->nnz = nnz;
  new_in->index = (int *)realloc(new_in->index, sizeof(int) * nnz);

#define _conv_to_crs_1(pre, type, test) { \
    size_t column, c, j, k, p;					\
    if (i < in->l) {						\
      column = 0; k = in->u + i;				\
      c = _min(in->u + 1 + i, object->d.matrix.n);		\
    } else {							\
      column = i - in->l; k = ld * (column + 1) - 1;		\
      c = _min(in->u + 1 + in->l, object->d.matrix.n - column); \
    }								\
    p = new_in->row[i];						\
    for (j = 0; j < c; j++) {					\
      if (test) { 						\
        new_in->v.pre[p] = in->v.pre[k];			\
        new_in->index[p] = column;				\
	p++;							\
      }								\
      k += ld - 1; column++;					\
    }								\
  }
#define _conv_to_crs_2(pre, type, test) { \
    size_t column, c, j, k, p;					\
    if (i < in->l) {						\
      column = 0; k = in->u + i;				\
      c = _min(in->u + 1 + i, object->d.matrix.n);		\
    } else {							\
      column = i - in->l; k = ld * (column + 1) - 1;		\
      c = _min(in->u + 1 + in->l, object->d.matrix.n - column); \
    }								\
    p = new_in->row[i];						\
    for (j = 0; j < c; j++) {					\
      if (test) { 						\
        new_in->v.pre[p*2]   = in->v.pre[k*2];			\
        new_in->v.pre[p*2+1] = in->v.pre[k*2+1];		\
        new_in->index[p] = column;				\
	p++;							\
      }								\
      k += ld - 1; column++;					\
    }								\
  }
  switch (object->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)realloc(new_in->v.i, sizeof(int32_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _conv_to_crs_1(i, int32_t, in->v.i[k]);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)realloc(new_in->v.l, sizeof(int64_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _conv_to_crs_1(l, int64_t, in->v.l[k]);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)realloc(new_in->v.s, sizeof(float) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _conv_to_crs_1(s, int32_t, *(int32_t *)&in->v.s[k]);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)realloc(new_in->v.d, sizeof(double) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _conv_to_crs_1(d, int64_t, *(int64_t *)&in->v.d[k]);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)realloc(new_in->v.c, sizeof(float) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _conv_to_crs_2(c, int32_t, (*(int32_t *)&in->v.c[k*2] ||
				  *(int32_t *)&in->v.c[k*2+1]));
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)realloc(new_in->v.z, sizeof(double) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _conv_to_crs_2(z, int64_t, (*(int64_t *)&in->v.z[k*2] ||
				  *(int64_t *)&in->v.z[k*2+1]));
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_pprint(FILE *file, silc_object_t *object, int level)
{
  silc_band_t *in;
  size_t i, j, k, l, ld, column, count;

#define _min(x, y) (((x) < (y)) ? (x) : (y))

  in = SILC_BAND(object);
  ld = in->l + in->u + 1; /* leading dimension */
  for (i = 0; i < object->d.matrix.m; i++) {
    if (i < in->l) {
      column = 0;
      k = in->u + i;
      count = _min(object->d.matrix.n, in->u + 1 + i);
    } else {
      column = i - in->l;
      k = ld * (column + 1) - 1;
      count = _min(object->d.matrix.n - (i - in->l), in->u + 1 + in->l);
    }
    for (j = 0; j < count; j++) {
      for (l = 0; l < level; l++)
	fprintf(file, "  ");
      switch (object->precision) {
      case SILC_INT:
	fprintf(file, "v.i[%ld,%ld] = %d\n", i, column, in->v.i[k]); break;
      case SILC_LONG:
	fprintf(file, "v.l[%ld,%ld] = %lld\n", i, column, in->v.l[k]); break;
      case SILC_FLOAT:
	fprintf(file, "v.s[%ld,%ld] = %e\n", i, column, in->v.s[k]); break;
      case SILC_DOUBLE:
	fprintf(file, "v.d[%ld,%ld] = %e\n", i, column, in->v.d[k]); break;
      case SILC_COMPLEX:
	fprintf(file, "v.c[%ld,%ld] = %e%+ei\n", i, column,
		in->v.c[k*2],
		in->v.c[k*2+1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	fprintf(file, "v.z[%ld,%ld] = %e%+ei\n", i, column,
		in->v.z[k*2],
		in->v.z[k*2+1]);
	break;
      default:
	should_not_reach_here;
      }
      column++;
      k += ld - 1;
    }
  }
  return 0;
}

silc_format_t proc_table = {
  proc_new,
  proc_recv,
  proc_send,
  proc_free,
  proc_dup,
  proc_pconv,
  proc_to_dense,
  proc_to_crs,
  NULL, /* proc_assign */
  NULL, /* proc_subscript */
  proc_pprint
};

int format_id; /* XXX: how to use ? */

void
init_sparse_band()
{
  void init_sparse_band_op();

  format_id = silc_register_format(SILC_FORMAT_BAND, &proc_table);

  init_sparse_band_op(MODULE_NAME);
}
