/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sparse_crs_format.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "object.h"
#include "sparse_crs.h"
#include "module.h"
#include "net.h"

#include "dense.h" /* for proc_to_dense() */

#define MODULE_NAME "sparse_crs"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

#ifdef ENABLE_NUMA
static void
init(array_t v, int precision, size_t m, int *row)
{
  ssize_t i;

  switch (precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memset(v.i + row[i], 0, sizeof(int32_t) * (row[i+1] - row[i]));
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memset(v.l + row[i], 0, sizeof(int64_t) * (row[i+1] - row[i]));
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memset(v.s + row[i], 0, sizeof(float) * (row[i+1] - row[i]));
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memset(v.d + row[i], 0, sizeof(double) * (row[i+1] - row[i]));
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memset(v.c + row[i], 0, sizeof(float) * 2 * (row[i+1] - row[i]));
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memset(v.z + row[i], 0, sizeof(double) * 2 * (row[i+1] - row[i]));
    break;
  default:
    should_not_reach_here;
  }
}
#endif /* ENABLE_NUMA */

/* proc_new: create a zero matrix */
static int
proc_new(silc_object_t *object)
{
  silc_crs_t *in;
  ssize_t i;

  in = silc_crs_allocate(object->precision, object->d.matrix.m, 0);
  if (!in)
    return -1;
#pragma omp parallel for
  for (i = 0; i <= (ssize_t)object->d.matrix.m; i++)
    in->row[i] = 0;
  object->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_recv(int sock, silc_object_t *object)
{
  size_t nnz;
  silc_crs_t *in;
  ssize_t i;
  int origin;

  recv_size(sock, &nnz, 0);
  send_ack(sock);
  in = silc_crs_allocate(object->precision, object->d.matrix.m, nnz);
  if (!in)
    return -1;
  origin = silc_get_origin();
#ifdef ENABLE_NUMA
#pragma omp parallel for
  for (i = 0; i < (ssize_t)object->d.matrix.m + 1; i++)
    in->row[i] = 0;
#endif
  recv_32_array(sock, in->row, object->d.matrix.m + 1, 0);
  send_ack(sock);
  if (origin) {
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m + 1; i++)
      in->row[i] -= origin;
  }
#ifdef ENABLE_NUMA
#pragma omp parallel for
  for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
    int j;
    for (j = in->row[i]; j < in->row[i+1]; j++)
      in->index[j] = 0;
  }
#endif
  recv_32_array(sock, in->index, in->nnz, 0);
  send_ack(sock);
  if (origin) {
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	in->index[j] -= origin;
    }
  }
#ifdef ENABLE_NUMA
  init(in->v, object->precision, object->d.matrix.m, in->row);
#endif
  switch (object->precision) {
  case SILC_INT:
    recv_32_array(sock, in->v.i, in->nnz, 0); break;
  case SILC_LONG:
    recv_64_array(sock, in->v.l, in->nnz, 0); break;
  case SILC_FLOAT:
    recv_32_array(sock, in->v.s, in->nnz, 0); break;
  case SILC_DOUBLE:
    recv_64_array(sock, in->v.d, in->nnz, 0); break;
  case SILC_COMPLEX:
    recv_32_array(sock, in->v.c, 2 * in->nnz, 0); break;
  case SILC_DOUBLE_COMPLEX:
    recv_64_array(sock, in->v.z, 2 * in->nnz, 0); break;
  default:
    should_not_reach_here;
  }
  object->d.matrix.in = (void *)in;
  silc_crs_sort(object);
  return 0;
}

static int
proc_send(int sock, silc_object_t *object)
{
  silc_crs_t *in;
  int *p, origin;
  ssize_t i;

  in = SILC_CRS(object);
  send_size(sock, &in->nnz, 0);
  recv_ack(sock);
  origin = silc_get_origin();
  if (origin == 0) {
    send_32_array(sock, in->row, object->d.matrix.m + 1, 0);
    recv_ack(sock);
    send_32_array(sock, in->index, in->nnz, 0);
    recv_ack(sock);
  } else {
    p = (int *)malloc(sizeof(int) * (object->d.matrix.m + 1));
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m + 1; i++)
      p[i] = in->row[i] + origin;
    send_32_array(sock, p, object->d.matrix.m + 1, 0);
    recv_ack(sock);
    p = (int *)realloc(p, sizeof(int) * in->nnz);
#pragma omp parallel for private(i)
    for (i = 0; i < (ssize_t)in->nnz; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	p[j] = in->index[j] + origin;
    }
    send_32_array(sock, p, in->nnz, 0);
    recv_ack(sock);
    free(p);
  }
  switch (object->precision) {
  case SILC_INT:
    send_32_array(sock, in->v.i, in->nnz, 0); break;
  case SILC_LONG:
    send_64_array(sock, in->v.l, in->nnz, 0); break;
  case SILC_FLOAT:
    send_32_array(sock, in->v.s, in->nnz, 0); break;
  case SILC_DOUBLE:
    send_64_array(sock, in->v.d, in->nnz, 0); break;
  case SILC_COMPLEX:
    send_32_array(sock, in->v.c, 2 * in->nnz, 0); break;
  case SILC_DOUBLE_COMPLEX:
    send_64_array(sock, in->v.z, 2 * in->nnz, 0); break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_free(silc_object_t *object)
{
  silc_crs_t *in;

  in = SILC_CRS(object);
  if (!in)
    return -1;
  free(in->v.i);
  free(in->row);
  free(in->index);
  free(in);
  return 0;
}

static int
proc_dup(silc_object_t *object, silc_object_t *duplicated, int copy)
{
  silc_crs_t *in, *new_in;
  ssize_t i;

#define _copy1(pre) \
  for (j = in->row[i]; j < in->row[i+1]; j++)		\
    new_in->v.pre[j] = in->v.pre[j];
#define _copy2(pre) \
  for (j = in->row[i]; j < in->row[i+1]; j++) {		\
    new_in->v.pre[j*2]   = in->v.pre[j*2];		\
    new_in->v.pre[j*2+1] = in->v.pre[j*2+1];		\
  }

  in = SILC_CRS(object);
  new_in = silc_crs_allocate(object->precision, object->d.matrix.m, in->nnz);
  if (!new_in)
    return -1;
  if (copy) {
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m + 1; i++)
      new_in->row[i] = in->row[i];
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i + 1]; j++)
	new_in->index[j] = in->index[j];
    }
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) { /* localize rows */
      int j;
      switch (object->precision) {
      case SILC_INT:
	_copy1(i); break;
      case SILC_LONG:
	_copy1(l); break;
      case SILC_FLOAT:
	_copy1(s); break;
      case SILC_DOUBLE:
	_copy1(d); break;
      case SILC_COMPLEX:
	_copy2(c); break;
      case SILC_DOUBLE_COMPLEX:
	_copy2(z); break;
      default:
	should_not_reach_here;
      }
    }
#ifdef ENABLE_NUMA
  } else {
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m + 1; i++)
      in->row[i] = 0;
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	in->index[j] = 0;
    }
    init(in->v, object->precision, object->d.matrix.m, in->row);
#endif /* ENABLE_NUMA */
  }
  duplicated->d.matrix.in = (void *)new_in;
  return 0;
}

static int
proc_pconv(silc_object_t *object, int new_precision)
{
  silc_crs_t *in;
  array_t v;
  int i;
  
#define _pconv11(src,dest,type) { \
    for (i = 0; i < in->nnz; i++)			\
      v.dest[i] = (type)in->v.src[i];			\
    free(in->v.src);					\
    in->v.dest = v.dest;				\
  }
#define _pconv12(src,dest,type) { \
    for (i = 0; i < in->nnz; i++) {			\
      v.dest[i*2]   = (type)in->v.src[i];		\
      v.dest[i*2+1] = (type)0;				\
    }							\
    free(in->v.src);					\
    in->v.dest = v.dest;				\
  }
#define _pconv22(src,dest,type) { \
    for (i = 0; i < in->nnz; i++) {			\
      v.dest[i*2]   = (type)in->v.src[i*2];		\
      v.dest[i*2+1] = (type)in->v.src[i*2+1];		\
    }							\
    free(in->v.src);					\
    in->v.dest = v.dest;				\
  }

  in = SILC_CRS(object);
  switch (new_precision) {
  case SILC_LONG:
    v.l = (int64_t *)malloc(sizeof(int64_t) * in->nnz); break;
  case SILC_FLOAT:
    v.s = (float *)malloc(sizeof(float) * in->nnz); break;
  case SILC_DOUBLE:
    v.d = (double *)malloc(sizeof(double) * in->nnz); break;
  case SILC_COMPLEX:
    v.c = (float *)malloc(sizeof(float) * 2 * in->nnz); break;
  case SILC_DOUBLE_COMPLEX:
    v.z = (double *)malloc(sizeof(double) * 2 * in->nnz); break;
  default:
    should_not_reach_here;
  }
#ifdef ENABLE_NUMA  
  init(v, new_precision, object->d.matrix.m, in->row);
#endif
  switch (object->precision) {
  case SILC_INT:
    switch (new_precision) {
    case SILC_LONG:
      _pconv11(i, l, int64_t); break;
    case SILC_FLOAT:
      _pconv11(i, s, float); break;
    case SILC_DOUBLE:
      _pconv11(i, d, double); break;
    case SILC_COMPLEX:
      _pconv12(i, c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(i, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_LONG:
    switch (new_precision) {
    case SILC_DOUBLE:
      _pconv11(l, d, double); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(l, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_FLOAT:
    switch (new_precision) {
    case SILC_DOUBLE:
      _pconv11(s, d, double); break;
    case SILC_COMPLEX:
      _pconv12(s, c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _pconv12(s, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_DOUBLE:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _pconv12(d, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  case SILC_COMPLEX:
    switch (new_precision) {
    case SILC_DOUBLE_COMPLEX:
      _pconv22(c, z, double); break;
    default:
      should_not_reach_here;
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_to_dense(silc_object_t *object, silc_object_t *new)
{
  silc_crs_t *in;
  silc_dense_t *new_in;
  silc_format_t *table;
  ssize_t i;

#ifdef DEBUG
  printf(MODULE_NAME ": silc_crs_to_dense(0x%lx, 0x%lx)\n", object, new);
#endif
  table = silc_find_format_proc_table(new->format);
  if (!table)
    return -1;
  if (!table->proc_new) {
    fprintf(stderr, MODULE_NAME ": proc_new() not found");
    return -1;
  }
  if (table->proc_new(new) < 0) {
    fprintf(stderr, MODULE_NAME ": proc_new() failed");
    return -1;
  }
  in = SILC_CRS(object);
  new_in = SILC_DENSE(new);
  switch (new->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.i[i + in->index[j] * object->d.matrix.m] = in->v.i[j];
    }
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.l[i + in->index[j] * object->d.matrix.m] = in->v.l[j];
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.s[i + in->index[j] * object->d.matrix.m] = in->v.s[j];
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.d[i + in->index[j] * object->d.matrix.m] = in->v.d[j];
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j, jj, kk;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	kk = (i + in->index[j] * object->d.matrix.m) * 2;
	jj = j * 2;
	new_in->v.c[kk]   = in->v.c[jj];
	new_in->v.c[kk+1] = in->v.c[jj+1];
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j, jj, kk;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	kk = (i + in->index[j] * object->d.matrix.m) * 2;
	jj = j * 2;
	new_in->v.z[kk]   = in->v.z[jj];
	new_in->v.z[kk+1] = in->v.z[jj+1];
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static int
proc_pprint(FILE *file, silc_object_t *object, int level)
{
  silc_crs_t *in;
  int column, i, j, k;
  
  in = SILC_CRS(object);
  for (i = 0; i < object->d.matrix.m; i++) {
    for (j = in->row[i]; j < in->row[i+1]; j++) {
      for (k = 0; k < level; k++)
	fprintf(file, "  ");
      column = in->index[j];
      switch (object->precision) {
      case SILC_INT:
	fprintf(file, "v.i[%d,%d] = %d\n", i, column, in->v.i[j]); break;
      case SILC_LONG:
	fprintf(file, "v.l[%d,%d] = %lld\n", i, column, in->v.l[j]); break;
      case SILC_FLOAT:
	fprintf(file, "v.s[%d,%d] = %e\n", i, column, in->v.s[j]); break;
      case SILC_DOUBLE:
	fprintf(file, "v.d[%d,%d] = %e\n", i, column, in->v.d[j]); break;
      case SILC_COMPLEX:
	fprintf(file, "v.c[%d,%d] = %e%+ei\n", i, column,
		in->v.c[j*2],
		in->v.c[j*2+1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	fprintf(file, "v.z[%d,%d] = %e%+ei\n", i, column,
		in->v.z[j*2],
		in->v.z[j*2+1]);
	break;
      default:
	should_not_reach_here;
      }
    }
  }
  return 0;
}

#define _crs_to_row_vector_1(pre, type) \
  row = v0[0];								\
  for (i = 0; i < modified->d.vector.length; i++) {			\
    column = v1[i];							\
    for (j = in->row[row]; j < in->row[row+1]; j++)			\
      if (in->index[j] == column)					\
	break;								\
    if (j < in->row[row+1])						\
      modified->d.vector.v.pre[i] = in->v.pre[j];			\
    else								\
      modified->d.vector.v.pre[i] = (type)0;				\
  }

#define _crs_to_row_vector_2(pre, type) \
  row = v0[0];								\
  for (i = 0; i < modified->d.vector.length; i++) {			\
    column = v1[i];							\
    for (j = in->row[row]; j < in->row[row+1]; j++)			\
      if (in->index[j] == column)					\
	break;								\
    if (j < in->row[row+1]) {						\
      modified->d.vector.v.pre[i*2]   = in->v.pre[j*2];			\
      modified->d.vector.v.pre[i*2+1] = in->v.pre[j*2+1];		\
    } else {								\
      modified->d.vector.v.pre[i*2]   = (type)0;			\
      modified->d.vector.v.pre[i*2+1] = (type)0;			\
    }									\
  }

#define _crs_to_column_vector_1(pre, type) \
  column = v1[0];							\
  for (i = 0; i < modified->d.vector.length; i++) {			\
    row = v0[i];							\
    for (j = in->row[row]; j < in->row[row+1]; j++)			\
      if (in->index[j] == column)					\
	break;								\
    if (j < in->row[row+1])						\
      modified->d.vector.v.pre[i] = in->v.pre[j];			\
    else								\
      modified->d.vector.v.pre[i] = (type)0;				\
  }

#define _crs_to_column_vector_2(pre, type) \
  column = v1[0];							\
  for (i = 0; i < modified->d.vector.length; i++) {			\
    row = v0[i];							\
    for (j = in->row[row]; j < in->row[row+1]; j++)			\
      if (in->index[j] == column)					\
	break;								\
    if (j < in->row[row+1]) {						\
      modified->d.vector.v.pre[i*2]   = in->v.pre[j*2];			\
      modified->d.vector.v.pre[i*2+1] = in->v.pre[j*2+1];		\
    } else {								\
      modified->d.vector.v.pre[i*2]   = (type)0;			\
      modified->d.vector.v.pre[i*2+1] = (type)0;			\
    }									\
  }

#define _simple_subscript_crs_1(pre, type) \
  c = row = 0;								\
  for (i = s1; i <= e1; i++) {						\
    new_in->row[row] = c;						\
    for (j = in->row[i]; j < in->row[i+1]; j++) {			\
      if (s2 <= in->index[j] && in->index[j] <= e2) {			\
	new_in->v.pre[c] = in->v.pre[j];				\
	new_in->index[c] = in->index[j] - s2;				\
	c++;								\
      }									\
    }									\
    row++;								\
  }									\
  new_in->row[row] = new_in->nnz = c;					\
  new_in->v.pre = realloc(new_in->v.pre, new_in->nnz * sizeof(type));	\
  new_in->index = realloc(new_in->index, new_in->nnz * sizeof(int));

#define _simple_subscript_crs_2(pre, type) \
  c = row = 0;								\
  for (i = s1; i <= e1; i++) {						\
    new_in->row[row] = c;						\
    for (j = in->row[i]; j < in->row[i+1]; j++) {			\
      if (s2 <= in->index[j] && in->index[j] <= e2) {			\
	new_in->v.pre[c*2]   = in->v.pre[j*2];				\
	new_in->v.pre[c*2+1] = in->v.pre[j*2+1];			\
	new_in->index[c]     = in->index[j] - s2;			\
	c++;								\
      }									\
    }									\
    row++;								\
  }									\
  new_in->row[row] = new_in->nnz = c;					\
  new_in->v.pre = realloc(new_in->v.pre, new_in->nnz * sizeof(type) * 2); \
  new_in->index = realloc(new_in->index, new_in->nnz * sizeof(int));

#define _complex_subscript_crs_1(pre, type) \
  c = 0;								\
  for (i = 0; i < modified->d.matrix.m; i++) {				\
    new_in->row[i] = c;							\
    row = v0[i];							\
    for (j = 0; j < modified->d.matrix.n; j++) {			\
      column = v1[j];							\
      for (k = in->row[row]; k < in->row[row+1]; k++) {			\
	if (in->index[k] == column) {					\
	  if (c == new_in->nnz) {					\
	    new_in->nnz += in->nnz;					\
	    new_in->v.pre = realloc(new_in->v.pre,			\
				    new_in->nnz * sizeof(type));	\
	    new_in->index = realloc(new_in->index,			\
				    new_in->nnz * sizeof(int));		\
	  }								\
	  new_in->v.pre[c] = in->v.pre[k];				\
	  new_in->index[c] = j;						\
	  c = c + 1;							\
	}								\
      }									\
    }									\
  }									\
  new_in->row[modified->d.matrix.m] = new_in->nnz = c;			\
  new_in->v.pre = realloc(new_in->v.pre, new_in->nnz * sizeof(type));	\
  new_in->index = realloc(new_in->index, new_in->nnz * sizeof(int));
    
#define _complex_subscript_crs_2(pre, type) \
  c = 0;								\
  for (i = 0; i < modified->d.matrix.m; i++) {				\
    new_in->row[i] = c;							\
    row = v0[i];							\
    for (j = 0; j < modified->d.matrix.n; j++) {			\
      column = v1[j];							\
      for (k = in->row[row]; k < in->row[row+1]; k++) {			\
	if (in->index[k] == column) {					\
	  if (c == new_in->nnz) {					\
	    new_in->nnz += in->nnz;					\
	    new_in->v.pre = realloc(new_in->v.pre,			\
				    new_in->nnz * sizeof(type) * 2);	\
	    new_in->index = realloc(new_in->index,			\
				    new_in->nnz * sizeof(int));		\
	  }								\
	  new_in->v.pre[c*2]   = in->v.pre[k*2];			\
	  new_in->v.pre[c*2+1] = in->v.pre[k*2+1];			\
	  new_in->index[c] = j;						\
	  c = c + 1;							\
	}								\
      }									\
    }									\
  }									\
  new_in->row[modified->d.matrix.m] = new_in->nnz = c;			\
  new_in->v.pre = realloc(new_in->v.pre, new_in->nnz * sizeof(type) * 2); \
  new_in->index = realloc(new_in->index, new_in->nnz * sizeof(int));

static silc_object_t *
proc_subscript(silc_object_t *object, clist_t *subscript[])
{
  silc_object_t *modified;
  silc_crs_t *in, *new_in;
  int c, i, j, k, length0, length1, *v0, *v1, *index;
  int row, column;
  int s1, e1, s2, e2;

  in = SILC_CRS(object);
  length0 = clist_length(subscript[0]);
  length1 = clist_length(subscript[1]);
  if (length0 == 1 && length1 == 1) {
    v0 = clist_expand(subscript[0]);
    v1 = clist_expand(subscript[1]);
    i = v0[0];
    for (j = in->row[i]; j < in->row[i+1]; j++)
      if (in->index[j] == v1[0])
	break;
    if (j == in->row[i+1]) {
      switch (object->precision) {
      case SILC_INT:
	modified = silc_object_new_int(0); break;
      case SILC_LONG:
	modified = silc_object_new_long(0); break;
      case SILC_FLOAT:
	modified = silc_object_new_float(0.0); break;
      case SILC_DOUBLE:
	modified = silc_object_new_double(0.0); break;
      case SILC_COMPLEX:
	modified = silc_object_new_complex(0.0, 0.0); break;
      case SILC_DOUBLE_COMPLEX:
	modified = silc_object_new_double_complex(0.0, 0.0); break;
      default:
	should_not_reach_here;
      }
    } else {
      switch (object->precision) {
      case SILC_INT:
	modified = silc_object_new_int(in->v.i[j]); break;
      case SILC_LONG:
	modified = silc_object_new_long(in->v.l[j]); break;
      case SILC_FLOAT:
	modified = silc_object_new_float(in->v.s[j]); break;
      case SILC_DOUBLE:
	modified = silc_object_new_double(in->v.d[j]); break;
      case SILC_COMPLEX:
	j = j * 2;
	modified = silc_object_new_complex(in->v.c[j], in->v.c[j+1]);
	break;
      case SILC_DOUBLE_COMPLEX:
	j = j * 2;
	modified = silc_object_new_double_complex(in->v.z[j], in->v.z[j+1]);
	break;
      default:
	should_not_reach_here;
      }
    }
    free(v0);
    free(v1);
  } else if (length0 == 1) {
    v0 = clist_expand(subscript[0]);
    v1 = clist_expand(subscript[1]);
    modified = silc_object_new_row_vector(object->precision, length1);
    switch (object->precision) {
    case SILC_INT:
      _crs_to_row_vector_1(i, int32_t); break;
    case SILC_LONG:
      _crs_to_row_vector_1(l, int64_t); break;
    case SILC_FLOAT:
      _crs_to_row_vector_1(s, float); break;
    case SILC_DOUBLE:
      _crs_to_row_vector_1(d, double); break;
    case SILC_COMPLEX:
      _crs_to_row_vector_2(c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _crs_to_row_vector_2(z, double); break;
    default:
      should_not_reach_here;
    }
    free(v0);
    free(v1);
  } else if (length1 == 1) {
    v0 = clist_expand(subscript[0]);
    v1 = clist_expand(subscript[1]);
    modified = silc_object_new_column_vector(object->precision, length0);
    switch (object->precision) {
    case SILC_INT:
      _crs_to_column_vector_1(i, int32_t); break;
    case SILC_LONG:
      _crs_to_column_vector_1(l, int64_t); break;
    case SILC_FLOAT:
      _crs_to_column_vector_1(s, float); break;
    case SILC_DOUBLE:
      _crs_to_column_vector_1(d, double); break;
    case SILC_COMPLEX:
      _crs_to_column_vector_2(c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _crs_to_column_vector_2(z, double); break;
    default:
      should_not_reach_here;
    }
    free(v0);
    free(v1);
  } else if (clist_range(subscript[0], &s1, &e1) &&
	     clist_range(subscript[1], &s2, &e2)) {
    modified = silc_object_new_matrix(object->precision, object->format,
				      length0, length1); /* m, n */
    new_in = silc_crs_allocate(object->precision,
			       length0, in->nnz); /* m, nnz */
    modified->d.matrix.in = (void *)new_in;
    switch (object->precision) {
    case SILC_INT:
      _simple_subscript_crs_1(i, int32_t); break;
    case SILC_LONG:
      _simple_subscript_crs_1(l, int64_t); break;
    case SILC_FLOAT:
      _simple_subscript_crs_1(s, float); break;
    case SILC_DOUBLE:
      _simple_subscript_crs_1(d, double); break;
    case SILC_COMPLEX:
      _simple_subscript_crs_2(c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _simple_subscript_crs_2(z, double); break;
    default:
      should_not_reach_here;
    }
  } else {
    v0 = clist_expand(subscript[0]);
    v1 = clist_expand(subscript[1]);
    modified = silc_object_new_matrix(object->precision, object->format,
				      length0, length1); /* m, n */
    new_in = silc_crs_allocate(object->precision,
			       length0, in->nnz); /* m, nnz */
    modified->d.matrix.in = (void *)new_in;
    switch (object->precision) {
    case SILC_INT:
      _complex_subscript_crs_1(i, int32_t); break;
    case SILC_LONG:
      _complex_subscript_crs_1(l, int64_t); break;
    case SILC_FLOAT:
      _complex_subscript_crs_1(s, float); break;
    case SILC_DOUBLE:
      _complex_subscript_crs_1(d, double); break;
    case SILC_COMPLEX:
      _complex_subscript_crs_2(c, float); break;
    case SILC_DOUBLE_COMPLEX:
      _complex_subscript_crs_2(z, double); break;
    default:
      should_not_reach_here;
    }
    free(v0);
    free(v1);
  }
  return modified;
}

silc_format_t silc_crs_proc_table = {
  proc_new,
  proc_recv,
  proc_send,
  proc_free,
  proc_dup,
  proc_pconv,
  proc_to_dense,
  NULL, /* proc_to_crs */
  NULL, /* proc_assign */
  proc_subscript,
  proc_pprint
};

void init_sparse_crs_op(); /* sparse_crs_op.c */

void
init_sparse_crs()
{
  int format_id;

  format_id = silc_register_format(SILC_FORMAT_CRS, &silc_crs_proc_table);

  init_sparse_crs_op(MODULE_NAME);
}
