/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sparse_crs_sort.c,v 1.3 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "object.h"
#include "sparse_crs.h"
#include "wtime.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

/* helper functions for sorting non-zero elements in a row */

static int
partition(silc_object_t *A, int i, int j, int x)
{
  silc_crs_t *in;
  int L, R, tmp;

  in = A->d.matrix.in;
  L = i;
  R = j;
  while (L <= R) {
    while (L <= j && in->index[L] < x)
      L++;
    while (R >= i && in->index[R] >= x)
      R--;
    if (L > R)
      break;
    switch (A->precision) {
    case SILC_INT:
      {
	int32_t v = in->v.i[L];
	in->v.i[L] = in->v.i[R];
	in->v.i[R] = v;
      }
      break;
    case SILC_LONG:
      {
	int64_t v = in->v.l[L];
	in->v.l[L] = in->v.l[R];
	in->v.l[R] = v;
      }
      break;
    case SILC_FLOAT:
      {
	float v = in->v.s[L];
	in->v.s[L] = in->v.s[R];
	in->v.s[R] = v;
      }
      break;
    case SILC_DOUBLE:
      {
	double v = in->v.d[L];
	in->v.d[L] = in->v.d[R];
	in->v.d[R] = v;
      }
      break;
    case SILC_COMPLEX:
      {
	float v[2];
	v[0] = in->v.c[L*2];
      	v[1] = in->v.c[L*2+1];
	in->v.c[L*2]   = in->v.c[R*2];
	in->v.c[L*2+1] = in->v.c[R*2+1];
	in->v.c[R*2]   = v[0];
	in->v.c[R*2+1] = v[1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double v[2];
	v[0] = in->v.z[L*2];
      	v[1] = in->v.z[L*2+1];
	in->v.z[L*2]   = in->v.z[R*2];
	in->v.z[L*2+1] = in->v.z[R*2+1];
	in->v.z[R*2]   = v[0];
	in->v.z[R*2+1] = v[1];
      }
      break;
    default:
      should_not_reach_here;
    }
    tmp = in->index[L];
    in->index[L] = in->index[R];
    in->index[R] = tmp;
    L++;
    R--;
  }
  return L;
}

static void
sort_row(silc_object_t *A, int i, int j)
{
  int p, k, *index;

  if (i == j)
    return;
  index = SILC_CRS(A)->index;
  p = i + 1;
  while (p <= j && index[i] == index[p])
    p++;
  if (p > j)
    return;
  if (index[i] > index[p])
    p = i;
  k = partition(A, i, j, index[p]);
  sort_row(A, i, k-1);
  sort_row(A, k, j);
}

int
silc_crs_sort(silc_object_t *object)
{
  silc_crs_t *in;
  ssize_t i;
#ifdef DEBUG
  double t, elapsed;

  t = silc_wtime();
#endif
  in = SILC_CRS(object);
#pragma omp parallel for
  for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
    if (in->row[i] < in->row[i+1])
      sort_row(object, in->row[i], in->row[i+1]-1);
#ifdef DEBUG
  elapsed = silc_wtime() - t;
  printf("silc_crs_sort: %e sec.\n", elapsed);
#endif
  return 0;
}
