/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sparse_jds_format.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "object.h"
#include "sparse_jds.h"
#include "module.h"
#include "net.h"

#include "dense.h" /* for proc_to_dense() */

#define MODULE_NAME "sparse_jds"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

extern int origin; /* XXX */

silc_jds_t *
silc_jds_allocate(int precision, size_t m, size_t nnz, size_t maxnzr)
{
  silc_jds_t *in;
  int num_threads;

  in = (silc_jds_t *)malloc(sizeof(silc_jds_t));
  in->nnz = nnz;
  in->maxnzr = maxnzr;
  if (nnz == 0)
    nnz = 1;  /* avoid doing malloc(0) */
  switch (precision) {
  case SILC_INT:
    in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz); break;
  case SILC_LONG:
    in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz); break;
  case SILC_FLOAT:
    in->v.s = (float *)malloc(sizeof(float) * nnz); break;
  case SILC_DOUBLE:
    in->v.d = (double *)malloc(sizeof(double) * nnz); break;
  case SILC_COMPLEX:
    in->v.c = (float *)malloc(sizeof(float) * 2 * nnz); break;
  case SILC_DOUBLE_COMPLEX:
    in->v.z = (double *)malloc(sizeof(double) * 2 * nnz); break;
  default:
    should_not_reach_here;
  }
#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  in->jd_ptr = (int *)malloc(sizeof(int) * (maxnzr + 1) * num_threads);
  in->index = (int *)malloc(sizeof(int) * nnz);
  in->perm = (int *)malloc(sizeof(int) * m);
  in->row = (int *)malloc(sizeof(int) * (num_threads + 1));
  return in;
}

#ifdef ENABLE_NUMA
static void
init(silc_jds_t *in, int precision, size_t m, size_t n)
{
#pragma omp parallel
  {
    int i, nnz;

    i = (in->maxnzr + 1) * omp_get_thread_num();
    nnz = in->jd_ptr[i + in->maxnzr] - in->jd_ptr[i];
    switch (precision) {
    case SILC_INT:
      memset(in->v.i + in->jd_ptr[i], 0, sizeof(int32_t) * nnz);
      break;
    case SILC_LONG:
      memset(in->v.l + in->jd_ptr[i], 0, sizeof(int64_t) * nnz);
      break;
    case SILC_FLOAT:
      memset(in->v.s + in->jd_ptr[i], 0, sizeof(float) * nnz);
      break;
    case SILC_DOUBLE:
      memset(in->v.d + in->jd_ptr[i], 0, sizeof(double) * nnz);
      break;
    case SILC_COMPLEX:
      memset(in->v.c + in->jd_ptr[i] * 2, 0, sizeof(float) * 2 * nnz);
      break;
    case SILC_DOUBLE_COMPLEX:
      memset(in->v.z + in->jd_ptr[i] * 2, 0, sizeof(double) * 2 * nnz);
      break;
    default:
      should_not_reach_here;
    }
  }
}
#endif /* ENABLE_NUMA */

/* proc_new: create a zero matrix */
static int
proc_new(silc_object_t *object)
{
  silc_jds_t *in;
  int i, num_threads;

  in = silc_jds_allocate(object->precision, object->d.matrix.m, 0, 0);
  if (!in)
    return -1;
#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  for (i = 0; i <= object->d.matrix.m; i++)
    in->perm[i] = i;
  for (i = 0; i <= num_threads; i++)
    in->jd_ptr[i] = 0;
  object->d.matrix.in = (void *)in;
  return 0;
}

static int
proc_recv(int sock, silc_object_t *object)
{
  return 0;
}

static int
proc_send(int sock, silc_object_t *object)
{
  return 0;
}

static int
proc_free(silc_object_t *object)
{
  silc_jds_t *in;

  in = SILC_JDS(object);
  if (!in)
    return -1;
  free(in->v.i);
  free(in->jd_ptr);
  free(in->index);
  free(in->perm);
  free(in);
  return 0;
}

static int
proc_dup(silc_object_t *object, silc_object_t *duplicated, int copy)
{
  silc_jds_t *in, *new_in;
  int num_threads;

  in = SILC_JDS(object);
  new_in = silc_jds_allocate(object->precision,
			     object->d.matrix.m,
			     (!copy) ? 0 : in->nnz,
			     (!copy) ? 0 : in->maxnzr);
  if (!new_in)
    return -1;
  if (copy) {
#ifdef _OPENMP
    num_threads = omp_get_max_threads();
#else
    num_threads = 1;
#endif
    memcpy(new_in->perm, in->perm, sizeof(int) * object->d.matrix.m);
    memcpy(new_in->row, in->row, sizeof(int) * (num_threads + 1));
    memcpy(new_in->jd_ptr, in->jd_ptr,
	   sizeof(int) * (in->maxnzr + 1) * num_threads);
#pragma omp parallel
    {
      int p, n1, n2;
#ifdef _OPENMP
      p = (in->maxnzr + 1) * omp_get_thread_num();
#else
      p = 0;
#endif
      n1 = in->jd_ptr[p];
      n2 = in->jd_ptr[p + in->maxnzr] - n1;
      memcpy(new_in->index + n1, in->index + n1, sizeof(int) * n2);
      switch (object->precision) {
      case SILC_INT:
	memcpy(new_in->v.i + n1, in->v.i + n1, sizeof(int32_t) * n2);
	break;
      case SILC_LONG:
	memcpy(new_in->v.l + n1, in->v.l + n1, sizeof(int64_t) * n2);
	break;
      case SILC_FLOAT:
	memcpy(new_in->v.s + n1, in->v.s + n1, sizeof(float) * n2);
	break;
      case SILC_DOUBLE:
	memcpy(new_in->v.d + n1, in->v.d + n1, sizeof(double) * n2);
	break;
      case SILC_COMPLEX:
	n1 *= 2;
	memcpy(new_in->v.s + n1, in->v.s + n1, sizeof(float) * 2 * n2);
	break;
      case SILC_DOUBLE_COMPLEX:
	n1 *= 2;
	memcpy(new_in->v.d + n1, in->v.d + n1, sizeof(double) * 2 * n2);
	break;
      default:
	should_not_reach_here;
      }
    }
  }
  duplicated->d.matrix.in = (void *)new_in;
  return 0;
}

static int
proc_pconv(silc_object_t *object, int new_precision)
{
  return 0;
}

static int
proc_to_dense(silc_object_t *object, silc_object_t *new)
{
  return 0;
}

static int
proc_pprint(FILE *file, silc_object_t *object, int level)
{
  silc_jds_t *in;
  int i, j, k, l, n, p, num_threads;

#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  in = SILC_JDS(object);
  for (i = 0; i < in->maxnzr; i++) {
    for (j = 0; j < num_threads; j++) {
      n = in->row[j];
      p = (in->maxnzr + 1) * j + i;
      for (k = in->jd_ptr[p]; k < in->jd_ptr[p+1]; k++) {
	for (l = 0; l < level; l++)
	  fprintf(file, "  ");
	switch (object->precision) {
	case SILC_INT:
	  fprintf(file, "v.i[%d,%d] = %d\n",
		  in->perm[n], in->index[k], in->v.i[k]);
	  break;
	case SILC_LONG:
	  fprintf(file, "v.l[%d,%d] = %lld\n",
		  in->perm[n], in->index[k], in->v.l[k]);
	  break;
	case SILC_FLOAT:
	  fprintf(file, "v.s[%d,%d] = %e\n",
		  in->perm[n], in->index[k], in->v.s[k]);
	  break;
	case SILC_DOUBLE:
	  fprintf(file, "v.d[%d,%d] = %e\n",
		  in->perm[n], in->index[k], in->v.d[k]);
	  break;
	case SILC_COMPLEX:
	  fprintf(file, "v.c[%d,%d] = %e%+ei\n",
		  in->perm[n], in->index[k], in->v.c[k*2], in->v.c[k*2+1]);
	  break;
	case SILC_DOUBLE_COMPLEX:
	  fprintf(file, "v.z[%d,%d] = %e%+ei\n",
		  in->perm[n], in->index[k], in->v.z[k*2], in->v.z[k*2+1]);
	  break;
	default:
	  should_not_reach_here;
	}
	n++;
      }
    }
  }
  return 0;
}

static silc_object_t *
proc_subscript(silc_object_t *object, clist_t *subscript[])
{
  return NULL;
}

silc_format_t silc_jds_proc_table = {
  proc_new,
  NULL, /* proc_recv */
  NULL, /* proc_send */
  proc_free,
  proc_dup,
  NULL, /* proc_pconv */
  NULL, /* proc_to_dense */
  NULL, /* proc_to_crs */
  NULL, /* proc_assign */
  NULL, /* proc_subscript */
  proc_pprint,
};

void init_sparse_jds_op(); /* sparse_jds_op.c */

void
init_sparse_jds()
{
  int format_id;

  format_id = silc_register_format(SILC_FORMAT_JDS, &silc_jds_proc_table);

  init_sparse_jds_op(MODULE_NAME);
}
