/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sparse_jds_op.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "object.h"
#include "sparse_crs.h"
#include "sparse_jds.h"
#include "module.h"
#include "wtime.h"

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

/* out: 2 */
int
silc_jds_mul_matrix_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *z;
  silc_modifier_t *x_mods, *y_mods;
  silc_jds_t *in;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_jds_mul_matrix_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_JDS(x);
  z = silc_object_new_column_vector(y->precision, x->d.matrix.m);

#pragma omp parallel
  {
    int d1, d2, i, j, jj, k1, k2, p, row;

#ifdef _OPENMP
    p = omp_get_thread_num();
#else
    p = 0;
#endif
    d1 = (in->maxnzr + 1) * p;
    d2 = d1 + in->maxnzr;
    switch (x->precision) {
    case SILC_INT:
      {
	int32_t *v;
	v = (int32_t *)malloc(sizeof(int32_t) * z->d.vector.length);
#pragma omp for
	for (i = 0; i < (ssize_t)z->d.vector.length; i++)
	  v[i] = 0.0;
	for (i = d1; i < d2; i++) {
	  row = in->row[p];
#pragma vdir nodep
	  for (j = in->jd_ptr[i]; j < in->jd_ptr[i+1]; j++)
	    v[row++] += in->v.i[j] * y->d.vector.v.i[in->index[j]];
	}
	for (row = in->row[p]; row < in->row[p+1]; row++)
	  z->d.vector.v.i[in->perm[row]] = v[row];
	free(v);
      }
      break;
    case SILC_LONG:
      {
	int64_t *v;
	v = (int64_t *)malloc(sizeof(int64_t) * z->d.vector.length);
#pragma omp for
	for (i = 0; i < (ssize_t)z->d.vector.length; i++)
	  v[i] = 0.0;
	for (i = d1; i < d2; i++) {
	  row = in->row[p];
#pragma vdir nodep
	  for (j = in->jd_ptr[i]; j < in->jd_ptr[i+1]; j++)
	    v[row++] += in->v.l[j] * y->d.vector.v.l[in->index[j]];
	}
	for (row = in->row[p]; row < in->row[p+1]; row++)
	  z->d.vector.v.l[in->perm[row]] = v[row];
	free(v);
      }
      break;
    case SILC_FLOAT:
      {
	float *v;
	v = (float *)malloc(sizeof(float) * z->d.vector.length);
#pragma omp for
	for (i = 0; i < (ssize_t)z->d.vector.length; i++)
	  v[i] = 0.0;
	for (i = d1; i < d2; i++) {
	  row = in->row[p];
#pragma vdir nodep
	  for (j = in->jd_ptr[i]; j < in->jd_ptr[i+1]; j++)
	    v[row++] += in->v.s[j] * y->d.vector.v.s[in->index[j]];
	}
	for (row = in->row[p]; row < in->row[p+1]; row++)
	  z->d.vector.v.s[in->perm[row]] = v[row];
	free(v);
      }
      break;
    case SILC_DOUBLE:
      {
	double *v;
	v = (double *)malloc(sizeof(double) * z->d.vector.length);
#pragma omp for
	for (i = 0; i < (ssize_t)z->d.vector.length; i++)
	  v[i] = 0.0;
	for (i = d1; i < d2; i++) {
	  row = in->row[p];
#pragma vdir nodep
	  for (j = in->jd_ptr[i]; j < in->jd_ptr[i+1]; j++)
	    v[row++] += in->v.d[j] * y->d.vector.v.d[in->index[j]];
	}
	for (row = in->row[p]; row < in->row[p+1]; row++)
	  z->d.vector.v.d[in->perm[row]] = v[row];
	free(v);
      }
      break;
    case SILC_COMPLEX:
      {
	float *v;
	v = (float *)malloc(sizeof(float) * 2 * z->d.vector.length);
#pragma omp for
	for (i = 0; i < (ssize_t)z->d.vector.length * 2; i++)
	  v[i] = 0.0;
	for (i = d1; i < d2; i++) {
	  row = in->row[p];
#pragma vdir nodep
	  for (j = in->jd_ptr[i]; j < in->jd_ptr[i+1]; j++) {
	    jj = j * 2;
	    k1 = row * 2;
	    k2 = in->index[j] * 2;
	    v[k1]   += (in->v.c[jj]   * y->d.vector.v.c[k2]   -
			in->v.c[jj+1] * y->d.vector.v.c[k2+1]);
	    v[k1+1] += (in->v.c[jj]   * y->d.vector.v.c[k2+1] +
			in->v.c[jj+1] * y->d.vector.v.c[k2]);
	  }
	}
	for (row = in->row[p]; row < in->row[p+1]; row++) {
	  k1 = in->perm[row] * 2;
	  k2 = row * 2;
	  z->d.vector.v.c[k1]   = v[k2];
	  z->d.vector.v.c[k1+1] = v[k2+1];
	}
	free(v);
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double *v;
	v = (double *)malloc(sizeof(double) * 2 * z->d.vector.length);
#pragma omp for
	for (i = 0; i < (ssize_t)z->d.vector.length * 2; i++)
	  v[i] = 0.0;
	for (i = d1; i < d2; i++) {
	  row = in->row[p];
#pragma vdir nodep
	  for (j = in->jd_ptr[i]; j < in->jd_ptr[i+1]; j++) {
	    jj = j * 2;
	    k1 = row * 2;
	    k2 = in->index[j] * 2;
	    v[k1]   += (in->v.z[jj]   * y->d.vector.v.z[k2]   -
			in->v.z[jj+1] * y->d.vector.v.z[k2+1]);
	    v[k1+1] += (in->v.z[jj]   * y->d.vector.v.z[k2+1] +
			in->v.z[jj+1] * y->d.vector.v.z[k2]);
	  }
	}
	for (row = in->row[p]; row < in->row[p+1]; row++) {
	  k1 = in->perm[row] * 2;
	  k2 = row * 2;
	  z->d.vector.v.z[k1]   = v[k2];
	  z->d.vector.v.z[k1+1] = v[k2+1];
	}
	free(v);
      }
      break;
    default:
      should_not_reach_here;
    }
  }
  param[2].o = z;
  return 0;
}

typedef struct {
  int row;
  int nzr;
} perm_t;

static int
perm_compare(const void *p1, const void *p2)
{
  int i = ((perm_t *)p1)->nzr;
  int j = ((perm_t *)p2)->nzr;

  return (i > j) ? -1 : (i < j) ? 1 : 0;
}

int
silc_jds_func_crs_to_jds(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_crs_t *in;
  silc_jds_t *new_in;
  perm_t *perm;
  int i, m, n, maxnzr, num_threads;
  double t, elapsed;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_jds_func_crs_to_jds(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  if (object->format != silc_find_format_id(SILC_FORMAT_CRS)) {
    fprintf(stderr, "sparse_jds: not in %s format\n", SILC_FORMAT_CRS);
    return -1;
  }
#ifdef TIME_CRS_TO_JDS
  t = silc_wtime();
#endif
  in = SILC_CRS(object);
  perm = (perm_t *)malloc(sizeof(perm_t) * object->d.matrix.m);
  maxnzr = 0;
  for (i = 0; i < object->d.matrix.m; i++) {
    perm[i].row = i;
    perm[i].nzr = in->row[i+1] - in->row[i];
    if (maxnzr < perm[i].nzr)
      maxnzr = perm[i].nzr;
  }
  new_in = (silc_jds_t *)malloc(sizeof(silc_jds_t));
  new_in->nnz = in->nnz;
  new_in->maxnzr = maxnzr;
  switch (object->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * in->nnz);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * in->nnz);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)malloc(sizeof(float) * in->nnz);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)malloc(sizeof(double) * in->nnz);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)malloc(sizeof(float) * 2 * in->nnz);
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)malloc(sizeof(double) * 2 * in->nnz);
    break;
  default:
    should_not_reach_here;
  }
#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  new_in->jd_ptr = (int *)malloc(sizeof(int) * (maxnzr + 1) * num_threads);
  new_in->index = (int *)malloc(sizeof(int) * in->nnz);
  new_in->perm = (int *)malloc(sizeof(int) * object->d.matrix.m);
  new_in->row = (int *)malloc(sizeof(int) * (num_threads + 1));
  new_in->row[0] = 0;
  m = object->d.matrix.m % num_threads;
  n = object->d.matrix.m / num_threads;
  if (m == 0) {
    for (i = 1; i < num_threads; i++)
      new_in->row[i] = n * i;
  } else {
    for (i = 1; i < num_threads; i++) {
      if (i < m)
	new_in->row[i] = (n + 1) * i;
      else
	new_in->row[i] = (n + 1) * m + n * (i - m);
    }
  }
  new_in->row[num_threads] = object->d.matrix.m;
#pragma omp parallel
  {
    int j, k, p, q, pos;

#ifdef _OPENMP
    p = omp_get_thread_num();
#else
    p = 0;
#endif
    qsort((void *)(perm + new_in->row[p]), new_in->row[p+1] - new_in->row[p],
	  sizeof(perm_t), perm_compare);
    for (j = new_in->row[p]; j < new_in->row[p+1]; j++)
      new_in->perm[j] = perm[j].row;
    pos = 0;
    for (j = 0; j < new_in->row[p]; j++)
      pos += perm[j].nzr;

#define _crs_to_jds_1(pre) { \
      for (j = 0; j < maxnzr; j++) {				\
	new_in->jd_ptr[(maxnzr + 1) * p + j] = pos;		\
	for (k = new_in->row[p]; k < new_in->row[p+1]; k++) {	\
	  if (j < perm[k].nzr) {				\
	    q = in->row[perm[k].row] + j;			\
	    new_in->v.pre[pos] = in->v.pre[q];			\
	    new_in->index[pos] = in->index[q];			\
	    pos ++;						\
	  }							\
	}							\
      }								\
    }
#define _crs_to_jds_2(pre) { \
      for (j = 0; j < maxnzr; j++) {				\
	new_in->jd_ptr[(maxnzr + 1) * p + j] = pos;		\
	for (k = new_in->row[p]; k < new_in->row[p+1]; k++) {	\
	  if (j < perm[k].nzr) {				\
	    q = in->row[perm[k].row] + j;			\
	    new_in->v.pre[pos*2]   = in->v.pre[q*2];		\
	    new_in->v.pre[pos*2+1] = in->v.pre[q*2+1];		\
	    new_in->index[pos]     = in->index[q];		\
	    pos ++;						\
	  }							\
	}							\
      }								\
    }
    switch (object->precision) {
    case SILC_INT:
      _crs_to_jds_1(i); break;
    case SILC_LONG:
      _crs_to_jds_1(l); break;
    case SILC_FLOAT:
      _crs_to_jds_1(s); break;
    case SILC_DOUBLE:
      _crs_to_jds_1(d); break;
    case SILC_COMPLEX:
      _crs_to_jds_2(c); break;
    case SILC_DOUBLE_COMPLEX:
      _crs_to_jds_2(z); break;
    default:
      should_not_reach_here;
    }
    new_in->jd_ptr[(maxnzr + 1) * p + maxnzr] = pos;
  }
  new_object = silc_object_new_matrix(object->precision,
				      silc_find_format_id(SILC_FORMAT_JDS),
				      object->d.matrix.m,
				      object->d.matrix.n);
  if (!new_object)
    return -1;
  new_object->d.matrix.in = new_in;
#ifdef TIME_CRS_TO_JDS
  elapsed = silc_wtime() - t;
  printf("sparse_jds: func_crs_to_jds: %8lf sec.\n", elapsed);
#endif
  param[1].o = new_object; /* return value */
  return 0;
}

#define SILC_ANY_INTEGER   (SILC_INT|SILC_LONG)
#define SILC_ANY_REAL      (SILC_FLOAT|SILC_DOUBLE)
#define SILC_ANY_COMPLEX   (SILC_COMPLEX|SILC_DOUBLE_COMPLEX)
#define SILC_ANY_PRECISION (SILC_ANY_INTEGER|SILC_ANY_REAL|SILC_ANY_COMPLEX)

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {silc_jds_mul_matrix_vector, "silc_jds_mul_matrix_vector",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_JDS, SILC_COLUMN_VECTOR_TYPE, 0,
   2, 0},
  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {silc_jds_func_crs_to_jds, "silc_jds_func_crs_to_jds",
   "matrix JDS(matrix)", 0},
  {NULL} /* sentinel */
};

void
init_sparse_jds_op(char *module_name)
{
  silc_register_module(module_name,
		       NULL, /* _unary_op_table */
		       _binary_op_table,
		       _function_table,
		       NULL); /* _procedure_table */
}
