/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: variable.c,v 1.4 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#ifdef _WIN32
#define _EXPORTING 1
#endif
#include "variable.h"
#include "wtime.h"

#ifdef TIME_VAR_LOOKUP
extern double _var_lookup_time; /* server.c */
#endif

typedef struct _varlist {
  struct _varlist *prev, *next;
  char name[SILC_MAXNAMELEN+1];
  silc_object_t *object;
} varlist_t;

static varlist_t varlist_top;

#ifdef SINGLE_THREAD
#define init_mutex_varlist()
#define lock_mutex_varlist()
#define unlock_mutex_varlist()
#define destroy_mutex_varlist()
#else
#ifdef _WIN32
#include <windows.h>
static HANDLE mutex_varlist;
#define init_mutex_varlist() { mutex_varlist = CreateMutex(NULL, FALSE, NULL); }
#define lock_mutex_varlist() WaitForSingleObject(mutex_varlist, INFINITE)
#define unlock_mutex_varlist() ReleaseMutex(mutex_varlist)
#define destroy_mutex_varlist() CloseHandle(mutex_varlist)
#else
#include <pthread.h>
static pthread_mutex_t mutex_varlist;
#define init_mutex_varlist() pthread_mutex_init(&mutex_varlist, NULL)
#define lock_mutex_varlist() pthread_mutex_lock(&mutex_varlist)
#define unlock_mutex_varlist() pthread_mutex_unlock(&mutex_varlist)
#define destroy_mutex_varlist() pthread_mutex_destroy(&mutex_varlist)
#endif
#endif

void
var_init(void)
{
  varlist_top.next = varlist_top.prev = &varlist_top;
  init_mutex_varlist();
}

void
var_finalize(void)
{
  destroy_mutex_varlist();
#ifdef DEBUG
  printf("varlist finalized\n");
#endif
}

static void
add(const char *name, silc_object_t *object)
{
  varlist_t *p;

  p = (varlist_t *)malloc(sizeof(varlist_t));
  strcpy(p->name, name);
  p->object = object;
  p->next = &varlist_top;
  p->prev = varlist_top.prev;
  varlist_top.prev->next = p;
  varlist_top.prev = p;
}

DECLSPEC void
var_add(const char *name, silc_object_t *object)
{
  lock_mutex_varlist();
  add(name, object);
  unlock_mutex_varlist();
}

static varlist_t *
lookup(const char *name)
{
  varlist_t *p;
#ifdef TIME_VAR_LOOKUP
  double t, elapsed;

  t = silc_wtime();
#endif
  p = varlist_top.next;
  while (p != &varlist_top) {
    if (strcmp(p->name, name) == 0) {
#ifdef TIME_VAR_LOOKUP
      elapsed = silc_wtime() - t;
      _var_lookup_time += elapsed;
#endif
      return p;
    }
    p = p->next;
  }
  return NULL;
}

DECLSPEC silc_object_t *
var_get(const char *name)
{
  varlist_t *p;

  lock_mutex_varlist();
  p = lookup(name);
  unlock_mutex_varlist();
  if (!p)
    return NULL;
  return p->object;
}

DECLSPEC silc_object_t *
var_get_free(const char *name)
{
  varlist_t *p;
  silc_object_t *object;

  lock_mutex_varlist();
  p = lookup(name);
  if (p) {
    object = p->object;
    p->next->prev = p->prev;
    p->prev->next = p->next;
    free(p);
#ifdef DEBUG
    printf("unnamed 0x%lx\n", object);
#endif
  }
  unlock_mutex_varlist();
  return object;
}

DECLSPEC void
var_free(const char *name)
{
  varlist_t *p;

  lock_mutex_varlist();
  p = lookup(name);
  if (p) {
    silc_object_free(p->object);
    p->next->prev = p->prev;
    p->prev->next = p->next;
    free(p);
#ifdef DEBUG
    printf("released %s\n", name);
#endif
  }
  unlock_mutex_varlist();
}

void
var_free_all(void)
{
  varlist_t *p = varlist_top.next;

  while (p != &varlist_top) {
    varlist_t *next = p->next;
    silc_object_free(p->object);
    free(p);
    p = next;
  }
  var_init();
}

static int count = 0;

char *
var_add_temp(silc_object_t *object, char *name)
{
  lock_mutex_varlist();
  while (1) {
    sprintf(name, "_TMP%06d", count++);
    if (lookup(name) == NULL)
      break;
    if (count > 10000) {
      fprintf(stderr, "max temp variables.\n");
      exit(0);
    }
  }
  add(name, object);
  unlock_mutex_varlist();
  return name;
}

#ifdef DEBUG
int
var_count_temp()
{
  int n;
  varlist_t *p;

  n = 0;
  p = varlist_top.next;
  while (p != &varlist_top) {
    if (strncmp(p->name, "_TMP", 4) == 0)
      n++;
    p = p->next;
  }
  return n;
}
#endif
