/*
 * Copyright (C) 2002-2007 The SSI Project. All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the project nor the names of its contributors 
 *       may be used to endorse or promote products derived from this software 
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE SSI PROJECT ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE SSI PROJECT BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: sparse_crs_op.c,v 1.6 2007/10/31 05:50:18 kajiyama Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#include "object.h"
#include "sparse_crs.h"
#include "module.h"
#include "wtime.h"

#include "dense.h" /* for addition of dense and sparse matrices */

#define should_not_reach_here  fprintf(stderr, \
  "*** should not reach here (%s, line %d)\n", __FILE__, __LINE__)

static int
_complex_conjugate(silc_object_t *object)
{
  silc_crs_t *in;
  int i;

#define _complex_conjugate_one(v) ((v) = -(v))

#define _complex_conjugate_all(pre) \
	for (i = 0; i < in->nnz; i++) \
	  _complex_conjugate_one(in->v.pre[i*2+1])

  in = SILC_CRS(object);
  switch (object->precision) {
  case SILC_INT:
  case SILC_LONG:
  case SILC_FLOAT:
  case SILC_DOUBLE:
    break;
  case SILC_COMPLEX:
    _complex_conjugate_all(c); break;
  case SILC_DOUBLE_COMPLEX:
    _complex_conjugate_all(z); break;
  default:
    should_not_reach_here;
  }
  return 0;
}

static silc_object_t *
_transpose(silc_object_t *object)
{
  silc_object_t *modified;
  silc_crs_t *in, *new_in;
  int num_threads, *nnz_column, i;
  double t, elapsed;

#ifdef TIME_CRS_TRANSPOSE
  t = silc_wtime();
#endif
  in = SILC_CRS(object);
  modified = silc_object_new_matrix(object->precision,
				    object->format,
				    object->d.matrix.n,  /* m */
				    object->d.matrix.m); /* n */
  modified->d.matrix.in = (void *)silc_crs_allocate(object->precision,
						    object->d.matrix.n,
						    in->nnz);
  new_in = SILC_CRS(modified);
#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  nnz_column = (int *)calloc(object->d.matrix.n * num_threads, sizeof(int));
#pragma omp parallel
  {
    int offset;

#ifdef _OPENMP
    offset = object->d.matrix.n * omp_get_thread_num();
#else
    offset = 0;
#endif
#pragma omp for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      int j;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	++nnz_column[offset + in->index[j]];
      }
    }
#pragma omp barrier
#pragma omp single
    {
      int c, j, k, n;
      c = new_in->row[0] = 0;
      for (i = 0; i < object->d.matrix.n; i++) {
	if (i == 0) 
	  new_in->row[i+1] = 0;
	else
	  new_in->row[i+1] = new_in->row[i];
	for (j = 0; j < num_threads; j++) {
	  k = i + object->d.matrix.n * j;
	  n = nnz_column[k];
	  if (n > 0) {
	    new_in->row[i+1] += n;
	    nnz_column[k] = c;
	    c += n;
	  }
	}
      }
      new_in->row[modified->d.matrix.m] = in->nnz;
    }
#ifdef ENABLE_NUMA
#define _transpose_crs_init_1(pre) {				\
        int j;							\
        for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
          new_in->v.pre[j] = 0;					\
          new_in->index[j] = 0;					\
        }							\
      }
#define _transpose_crs_init_2(pre) {				\
        int j;							\
        for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
          new_in->v.pre[j*2]   = 0;				\
          new_in->v.pre[j*2+1] = 0;				\
          new_in->index[j] = 0;					\
        }							\
      }
    switch (modified->precision) {
    case SILC_INT:
#pragma omp for
      for (i = 0; i < (ssize_t)modified->d.matrix.m; i++)
	_transpose_crs_init_1(i);
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)modified->d.matrix.m; i++)
	_transpose_crs_init_1(l);
      break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)modified->d.matrix.m; i++)
	_transpose_crs_init_1(s);
      break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)modified->d.matrix.m; i++)
	_transpose_crs_init_1(d);
      break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)modified->d.matrix.m; i++)
	_transpose_crs_init_2(c);
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)modified->d.matrix.m; i++)
	_transpose_crs_init_2(z);
      break;
    default:
      should_not_reach_here;
    }
#endif /* ENABLE_NUMA */
#define _transpose_crs_copy_1(pre) {				\
        int j, k;						\
        for (j = in->row[i]; j < in->row[i+1]; j++) {		\
          k = nnz_column[offset + in->index[j]]++;		\
          new_in->v.pre[k] = in->v.pre[j];			\
          new_in->index[k] = i;					\
        }							\
      }
#define _transpose_crs_copy_2(pre) {				\
        int j, k;						\
        for (j = in->row[i]; j < in->row[i+1]; j++) {		\
          k = nnz_column[offset + in->index[j]]++;		\
          new_in->v.pre[k*2]   = in->v.pre[j*2];		\
          new_in->v.pre[k*2+1] = in->v.pre[j*2+1];		\
          new_in->index[k] = i;					\
        }							\
      }
    switch (modified->precision) {
    case SILC_INT:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_transpose_crs_copy_1(i);
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++) 
	_transpose_crs_copy_1(l);
      break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++) 
	_transpose_crs_copy_1(s);
      break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++) 
	_transpose_crs_copy_1(d);
      break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++) 
	_transpose_crs_copy_2(c);
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++) 
	_transpose_crs_copy_2(z);
      break;
    default:
      should_not_reach_here;
    }
  }
  free(nnz_column);
#ifdef TIME_CRS_TRANSPOSE
  elapsed = silc_wtime() - t;
  printf("sparse_crs_op: _transpose: %lf sec.\n", elapsed);
#endif
  return modified;
}

/* out = 0 */
int
silc_crs_complex_conjugate(int n_params, silc_mfunc_param_t *param)
{
#ifdef DEBUG
  printf("silc_crs_complex_conjugate(0x%lx)\n", param[0].o);
#endif
  return _complex_conjugate(param[0].o);
}

/* out = 1 */
int
silc_crs_transpose(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *modified;

#ifdef DEBUG
  printf("silc_crs_transpose(0x%lx)\n", param[0].o);
#endif
  modified = _transpose(param[0].o);
  if (!modified)
    return -1;
  param[1].o = modified;
  return 0;
}

/* out = 1 */
int
silc_crs_conjugate_transpose(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *modified;

#ifdef DEBUG
  printf("silc_crs_conjugate_transpose(0x%lx)\n", param[0].o);
#endif
  modified = _transpose(param[0].o);
  if (!modified)
    return -1;
  if (_complex_conjugate(modified) < 0)
    return -1;
  param[1].o = modified;
  return 0;
}

/* out = 0 */
int
silc_crs_negate(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  silc_crs_t *in;
  ssize_t i;

#define _negate(v) ((v) = -(v))

#define _negate_all_1(pre) { \
    size_t j;					\
    for (j = in->row[i]; j < in->row[i+1]; j++)	\
      _negate(in->v.pre[j]);			\
  }
#define _negate_all_2(pre) { \
    size_t j;					\
    for (j = in->row[i]; j < in->row[i+1]; j++)	\
      _negate(in->v.pre[j*2]);			\
  }

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_negate(0x%lx)\n", object);
#endif
  in = SILC_CRS(object);
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _negate_all_1(i);
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _negate_all_1(l);
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _negate_all_1(s);
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _negate_all_1(d);
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _negate_all_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
      _negate_all_2(z);
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 2 */
int
silc_crs_mul_matrix_vector(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *p;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *in;
  ssize_t i;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_matrix_vector(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_CRS(x);
  p = silc_object_new_column_vector(y->precision, x->d.matrix.m);
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      p->d.vector.v.i[i] = 0;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	p->d.vector.v.i[i] += in->v.i[j] * y->d.vector.v.i[in->index[j]];
    }
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      p->d.vector.v.l[i] = 0;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	p->d.vector.v.l[i] += in->v.l[j] * y->d.vector.v.l[in->index[j]];
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      p->d.vector.v.s[i] = 0.0;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	p->d.vector.v.s[i] += in->v.s[j] * y->d.vector.v.s[in->index[j]];
    }
    break;
  case SILC_DOUBLE:
#if 1
    {
      int *index = in->index;
      double *dx = in->v.d;
      double *dy = y->d.vector.v.d;
      double *dz = p->d.vector.v.d;
#pragma omp parallel for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
	int j;
	int start = in->row[i];
	int end = in->row[i+1];
	double t = 0.0;
	for (j = start; j < end; j++)
	  t += dx[j] * dy[index[j]];
	dz[i] = t;
      }
    }
#else
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      p->d.vector.v.d[i] = 0.0;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	p->d.vector.v.d[i] += in->v.d[j] * y->d.vector.v.d[in->index[j]];
    }
#endif
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j, i2, j2, k2;
      float real, imag;
      real = imag = 0.0;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	j2 = j * 2;
	k2 = in->index[j] * 2;
	real += (in->v.c[j2]   * y->d.vector.v.c[k2]   -
		 in->v.c[j2+1] * y->d.vector.v.c[k2+1]);
	imag += (in->v.c[j2]   * y->d.vector.v.c[k2+1] +
		 in->v.c[j2+1] * y->d.vector.v.c[k2]);
      }
      i2 = i * 2;
      p->d.vector.v.c[i2]   = real;
      p->d.vector.v.c[i2+1] = imag;
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j, i2, j2, k2;
      double real, imag;
      real = imag = 0.0;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	j2 = j * 2;
	k2 = in->index[j] * 2;
	real += (in->v.z[j2]   * y->d.vector.v.z[k2]   -
		 in->v.z[j2+1] * y->d.vector.v.z[k2+1]);
	imag += (in->v.z[j2]   * y->d.vector.v.z[k2+1] +
		 in->v.z[j2+1] * y->d.vector.v.z[k2]);
      }
      i2 = i * 2;
      p->d.vector.v.z[i2]   = real;
      p->d.vector.v.z[i2+1] = imag;
    }
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = p;
  return 0;
}

/* out: 2 */
int
silc_crs_mul_vector_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *p;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *in;
  ssize_t i, n;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_vector_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_CRS(y);
  p = silc_object_new_row_vector(y->precision, y->d.matrix.n);
#ifdef _OPENMP
  n = omp_get_max_threads();
#endif
  switch (x->precision) {
  case SILC_INT:
#ifdef _OPENMP
    /* parallel version */
    {
      int32_t *buf;
      buf = (int32_t *)malloc(sizeof(int32_t) * p->d.vector.length * n);
#pragma omp parallel for
      for (i = 0; i < n; i++) {
	size_t j;
	int32_t *v;
	v = buf + p->d.vector.length * i;
	for (j = 0; j < p->d.vector.length; j++)
	  v[j] = 0;
      }
#pragma omp parallel
      {
	int32_t *v;
	v = buf + p->d.vector.length * omp_get_thread_num();
#pragma omp for
	for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
	  size_t j;
	  for (j = in->row[i]; j < in->row[i+1]; j++)
	    v[in->index[j]] += x->d.vector.v.i[i] * in->v.i[j];
	}
      }
      for (i = 0; i < p->d.vector.length; i++)
	p->d.vector.v.i[i] = buf[i];
      for (i = 1; i < n; i++) {
	size_t j;
	int32_t *v;
	v = buf + p->d.vector.length * i;
	for (j = 0; j < p->d.vector.length; j++)
	  p->d.vector.v.i[j] += v[j];
      }
      free(buf);
    }
#else      
    /* sequential version */
    for (i = 0; i < (ssize_t)p->d.vector.length; i++)
      p->d.vector.v.i[i] = 0;
    for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
      size_t j, k;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	k = in->index[j];
	p->d.vector.v.i[k] += x->d.vector.v.i[k] * in->v.i[j];
      }
    }
#endif
    break;
  case SILC_LONG:
    for (i = 0; i < (ssize_t)p->d.vector.length; i++)
      p->d.vector.v.l[i] = 0;
    for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
      size_t j, k;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	k = in->index[j];
	p->d.vector.v.l[k] += x->d.vector.v.l[k] * in->v.l[j];
      }
    }
    break;
  case SILC_FLOAT:
    for (i = 0; i < (ssize_t)p->d.vector.length; i++)
      p->d.vector.v.s[i] = 0.0;
    for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
      size_t j, k;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	k = in->index[j];
	p->d.vector.v.s[k] += x->d.vector.v.s[k] * in->v.s[j];
      }
    }
    break;
  case SILC_DOUBLE:
    for (i = 0; i < (ssize_t)p->d.vector.length; i++)
      p->d.vector.v.d[i] = 0.0;
    for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
      size_t j, k;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	k = in->index[j];
	p->d.vector.v.d[k] += x->d.vector.v.d[k] * in->v.d[j];
      }
    }
    break;
  case SILC_COMPLEX:
    for (i = 0; i < (ssize_t)p->d.vector.length * 2; i++)
      p->d.vector.v.c[i] = 0.0;
    for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
      size_t j, j2, k2;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	j2 = j * 2;
	k2 = in->index[j] * 2;
	p->d.vector.v.c[k2]   += (x->d.vector.v.c[k2]   * in->v.c[j2]   -
				  x->d.vector.v.c[k2+1] * in->v.c[j2+1]);
	p->d.vector.v.c[k2+1] += (x->d.vector.v.c[k2]   * in->v.c[j2+1] +
				  x->d.vector.v.c[k2+1] * in->v.c[j2]);
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
    for (i = 0; i < (ssize_t)p->d.vector.length * 2; i++)
      p->d.vector.v.z[i] = 0.0;
    for (i = 0; i < (ssize_t)y->d.matrix.m; i++) {
      size_t j, j2, k2;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	j2 = j * 2;
	k2 = in->index[j] * 2;
	p->d.vector.v.z[k2]   += (x->d.vector.v.z[k2]   * in->v.z[j2]   -
				  x->d.vector.v.z[k2+1] * in->v.z[j2+1]);
	p->d.vector.v.z[k2+1] += (x->d.vector.v.z[k2]   * in->v.z[j2+1] +
				  x->d.vector.v.z[k2+1] * in->v.z[j2]);
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = p;
  return 0;
}

static int
matrix_scalar_product(silc_mfunc_param_t *matrix,
		      silc_mfunc_param_t *scalar)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *in;
  ssize_t i;

  x = matrix->o; x_mods = matrix->m;
  y = scalar->o; y_mods = scalar->m;
#ifdef DEBUG
  printf("sparse_crs_op: matrix_scalar_product(0x%lx, 0x%lx)\n", x, y);
#endif
  in = SILC_CRS(x);
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	in->v.i[j] *= y->d.scalar.i;
    }
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	in->v.l[j] *= y->d.scalar.l;
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	in->v.s[j] *= y->d.scalar.s;
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	in->v.d[j] *= y->d.scalar.d;
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j, k;
      float real, imag;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	k = j * 2;
	real = (in->v.c[k]   * y->d.scalar.c[0] -
		in->v.c[k+1] * y->d.scalar.c[1]);
	imag = (in->v.c[k]   * y->d.scalar.c[1] +
		in->v.c[k+1] * y->d.scalar.c[0]);
	in->v.c[k]   = real;
	in->v.c[k+1] = imag;
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j, k;
      double real, imag;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	k = j * 2;
	real = (in->v.z[k]   * y->d.scalar.z[0] -
		in->v.z[k+1] * y->d.scalar.z[1]);
	imag = (in->v.z[k]   * y->d.scalar.z[1] +
		in->v.z[k+1] * y->d.scalar.z[0]);
	in->v.z[k]   = real;
	in->v.z[k+1] = imag;
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  return 0;
}

/* out: 0 */
int
silc_crs_mul_matrix_scalar(int n_params, silc_mfunc_param_t *param)
{
  return matrix_scalar_product(&param[0], &param[1]);
}

/* out: 1 */
int
silc_crs_mul_scalar_matrix(int n_params, silc_mfunc_param_t *param)
{
  return matrix_scalar_product(&param[1], &param[0]);
}

#if 1 /* silc_crs_{add,sub}_matrix_matrix (v2) */

/* out: 2 */
int
silc_crs_add_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  ssize_t i;
  size_t nnz;
#ifdef DEBUG
#ifdef _WIN32
  unsigned long long flop;
#else
  size_t flop;
#endif
#endif

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_add_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
#ifdef DEBUG
  for (i = 0; i < x->d.matrix.m; i++) {
    int jx, jjx, jy, jjy;
    jx = x_in->row[i]; jjx = x_in->row[i+1];
    jy = y_in->row[i]; jjy = y_in->row[i+1];
    while (jx < jjx && jy < jjy) {
      if (x_in->index[jx] == y_in->index[jy]) {
	flop++; jx++; jy++; continue;
      }
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {
	jx++;
      }
      while (jy < jjy && x_in->index[jx] > y_in->index[jy]) {
	jy++;
      }
    }
  }
  printf("silc_crs_add_matrix_matrix: flop = %ld\n", flop);
#endif
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
  new_in->row[0] = nnz = 0;
  for (i = 0; i < x->d.matrix.m; i++) {
    int jx, jjx, jy, jjy;
    jx = x_in->row[i]; jjx = x_in->row[i+1];
    jy = y_in->row[i]; jjy = y_in->row[i+1];
    while (jx < jjx && jy < jjy) {
      if (x_in->index[jx] == y_in->index[jy]) {
	nnz++; jx++; jy++; continue;
      }
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {
	nnz++; jx++;
      }
      while (jy < jjy && x_in->index[jx] > y_in->index[jy]) {
	nnz++; jy++;
      }
    }
    if (jx < jjx) nnz += jjx - jx;
    if (jy < jjy) nnz += jjy - jy;
    new_in->row[i+1] = nnz;
  }
  new_in->nnz = nnz;
  new_in->index = (int *)malloc(sizeof(int) * nnz);

#define _add_1(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j] = x_in->v.pre[jx] + y_in->v.pre[jy];	\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
	new_in->v.pre[j] = x_in->v.pre[jx];			\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++;						\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	new_in->v.pre[j] = y_in->v.pre[jy];			\
	new_in->index[j] = y_in->index[jy];			\
	j++; jy++;						\
      }								\
    }								\
    while (jx < jjx) {						\
      new_in->v.pre[j] = x_in->v.pre[jx];			\
      new_in->index[j] = x_in->index[jx];			\
      j++; jx++;						\
    }								\
    while (jy < jjy) {						\
      new_in->v.pre[j] = y_in->v.pre[jy];			\
      new_in->index[j] = y_in->index[jy];			\
      j++; jy++;						\
    }								\
  }
#define _add_2(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j*2]   = x_in->v.pre[jx*2]   + y_in->v.pre[jy*2];	  \
	new_in->v.pre[j*2+1] = x_in->v.pre[jx*2+1] + y_in->v.pre[jy*2+1]; \
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
	new_in->v.pre[j*2]   = x_in->v.pre[jx*2];		\
	new_in->v.pre[j*2+1] = x_in->v.pre[jx*2+1];		\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++;						\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	new_in->v.pre[j*2]   = y_in->v.pre[jy*2];		\
	new_in->v.pre[j*2+1] = y_in->v.pre[jy*2+1];		\
	new_in->index[j] = y_in->index[jy];			\
	j++; jy++;						\
      }								\
    }								\
    while (jx < jjx) {						\
      new_in->v.pre[j*2]   = x_in->v.pre[jx*2];			\
      new_in->v.pre[j*2+1] = x_in->v.pre[jx*2+1];		\
      new_in->index[j] = x_in->index[jx];			\
      j++; jx++;						\
    }								\
    while (jy < jjy) {						\
      new_in->v.pre[j*2]   = y_in->v.pre[jy*2];			\
      new_in->v.pre[j*2+1] = y_in->v.pre[jy*2+1];		\
      new_in->index[j] = y_in->index[jy];			\
      j++; jy++;						\
    }								\
  }

  switch (x->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_1(i);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_1(l);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)malloc(sizeof(float) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_1(s);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)malloc(sizeof(double) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_1(d);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_2(z);
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      x->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  silc_crs_sort(param[2].o);
  return 0;
}

/* out: 2 */
int
silc_crs_sub_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  ssize_t i;
  size_t nnz;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_sub_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
#ifdef DEBUG
  printf("silc_crs_sub_matrix_matrix: flop = %d\n", y_in->nnz);
#endif
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
  new_in->row[0] = nnz = 0;
  for (i = 0; i < x->d.matrix.m; i++) {
    int jx, jjx, jy, jjy;
    jx = x_in->row[i]; jjx = x_in->row[i+1];
    jy = y_in->row[i]; jjy = y_in->row[i+1];
    while (jx < jjx && jy < jjy) {
      if (x_in->index[jx] == y_in->index[jy]) {
	nnz++; jx++; jy++; continue;
      }
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {
	nnz++; jx++;
      }
      while (jy < jjy && x_in->index[jx] > y_in->index[jy]) {
	nnz++; jy++;
      }
    }
    if (jx < jjx) nnz += jjx - jx;
    if (jy < jjy) nnz += jjy - jy;
    new_in->row[i+1] = nnz;
  }
  new_in->nnz = nnz;
  new_in->index = (int *)malloc(sizeof(int) * nnz);
  
#define _sub_1(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j] = x_in->v.pre[jx] - y_in->v.pre[jy];	\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
	new_in->v.pre[j] = x_in->v.pre[jx];			\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++;						\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	new_in->v.pre[j] = - y_in->v.pre[jy];			\
	new_in->index[j] =   y_in->index[jy];			\
	j++; jy++;						\
      }								\
    }								\
    while (jx < jjx) {						\
      new_in->v.pre[j] = x_in->v.pre[jx];			\
      new_in->index[j] = x_in->index[jx];			\
      j++; jx++;						\
    }								\
    while (jy < jjy) {						\
      new_in->v.pre[j] = - y_in->v.pre[jy];			\
      new_in->index[j] =   y_in->index[jy];			\
      j++; jy++;						\
    }								\
  }
#define _sub_2(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j*2]   = x_in->v.pre[jx*2]   - y_in->v.pre[jy*2];	  \
	new_in->v.pre[j*2+1] = x_in->v.pre[jx*2+1] - y_in->v.pre[jy*2+1]; \
	new_in->index[j]     = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
	new_in->v.pre[j*2]   = x_in->v.pre[jx*2];		\
	new_in->v.pre[j*2+1] = x_in->v.pre[jx*2+1];		\
	new_in->index[j]     = x_in->index[jx];			\
	j++; jx++;						\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	new_in->v.pre[j*2]   = - y_in->v.pre[jy*2];		\
	new_in->v.pre[j*2+1] = - y_in->v.pre[jy*2+1];		\
	new_in->index[j]     =   y_in->index[jy];		\
	j++; jy++;						\
      }								\
    }								\
    while (jx < jjx) {						\
      new_in->v.pre[j*2]   = x_in->v.pre[jx*2];			\
      new_in->v.pre[j*2+1] = x_in->v.pre[jx*2+1];		\
      new_in->index[j]     = x_in->index[jx];			\
      j++; jx++;						\
    }								\
    while (jy < jjy) {						\
      new_in->v.pre[j*2]   = - y_in->v.pre[jy*2];		\
      new_in->v.pre[j*2+1] = - y_in->v.pre[jy*2+1];		\
      new_in->index[j]     =   y_in->index[jy];			\
      j++; jy++;						\
    }								\
  }

  switch (x->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _sub_1(i);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _sub_1(l);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)malloc(sizeof(float) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _sub_1(s);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)malloc(sizeof(double) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _sub_1(d);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _sub_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _sub_2(z);
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      x->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  silc_crs_sort(param[2].o);
  return 0;
}

#else /* silc_crs_{add,sub}_matrix_matrix (v1) */

/* out: 2 */
int
silc_crs_add_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  size_t nnz, nnz_row;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_add_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
#pragma omp parallel
  {
    int *w;
    ssize_t i;
    size_t j, k, pos;

    w = (int *)calloc(x->d.matrix.n, sizeof(int));
#pragma omp for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {
	w[x_in->index[j]] = 1;
      }
      for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {
	w[y_in->index[j]] = 1;
      }
      new_in->row[i] = 0;
      for (j = 0; j < x->d.matrix.n; j++) {
	if (w[j]) {
	  new_in->row[i]++;
	  w[j] = 0;
	}
      }
    }
#pragma omp barrier
#pragma omp single
    {
      nnz = 0;
      for (i = 0; i < x->d.matrix.m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[x->d.matrix.m] = nnz;
      switch (x->precision) {
      case SILC_INT:
	new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
	break;
      case SILC_LONG:
	new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
	break;
      case SILC_FLOAT:
	new_in->v.s = (float *)malloc(sizeof(float) * nnz);
	break;
      case SILC_DOUBLE:
	new_in->v.d = (double *)malloc(sizeof(double) * nnz);
	break;
      case SILC_COMPLEX:
	new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
	break;
      default:
	should_not_reach_here;
      }
      new_in->index = (int *)malloc(sizeof(int) * nnz);
      new_in->nnz = nnz;
    }

#define _mmadd_1(pre) { \
      k = new_in->row[i];					\
      for (j = 0; j < x->d.matrix.n; j++)			\
        w[j] = nnz;						\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {		\
	new_in->v.pre[k] = x_in->v.pre[j];			\
	new_in->index[k] = x_in->index[j];			\
	w[x_in->index[j]] = k++;				\
      }								\
      for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {		\
        pos = w[y_in->index[j]];				\
	if (pos == nnz) {					\
	  new_in->v.pre[k] = y_in->v.pre[j];			\
	  new_in->index[k] = y_in->index[j];			\
	  k++;							\
	} else {						\
	  new_in->v.pre[pos] += y_in->v.pre[j];			\
	}							\
      }								\
    }
#define _mmadd_2(pre) { \
      k = new_in->row[i];					\
      for (j = 0; j < x->d.matrix.n; j++)			\
        w[j] = nnz;						\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {		\
	new_in->v.pre[k*2]   = x_in->v.pre[j*2];		\
	new_in->v.pre[k*2+1] = x_in->v.pre[j*2+1];		\
	new_in->index[k]     = x_in->index[j];			\
	w[x_in->index[j]] = k++;				\
      }								\
      for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {		\
        pos = w[y_in->index[j]];				\
	if (pos == nnz) {					\
	  new_in->v.pre[k*2]   = y_in->v.pre[j*2];		\
	  new_in->v.pre[k*2+1] = y_in->v.pre[j*2+1];		\
	  new_in->index[k]     = y_in->index[j];		\
	  k++;							\
	} else {						\
	  new_in->v.pre[pos*2]   += y_in->v.pre[j*2];		\
	  new_in->v.pre[pos*2+1] += y_in->v.pre[j*2+1];		\
	}							\
      }								\
    }

    switch (x->precision) {
    case SILC_INT: 
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmadd_1(i);
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmadd_1(l);
      break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmadd_1(s);
      break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmadd_1(d);
      break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmadd_2(c);
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmadd_2(z);
      break;
    default:
      should_not_reach_here;
    }
    free(w);
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      x->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  silc_crs_sort(param[2].o);
  return 0;
}

/* out: 2 */
int
silc_crs_sub_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  size_t nnz, nnz_row;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_sub_matrix_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
#pragma omp parallel
  {
    int *w;
    ssize_t i;
    size_t j, k, pos;

    w = (int *)calloc(x->d.matrix.n, sizeof(int));
#pragma omp for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {
	w[x_in->index[j]] = 1;
      }
      for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {
	w[y_in->index[j]] = 1;
      }
      new_in->row[i] = 0;
      for (j = 0; j < x->d.matrix.n; j++) {
	if (w[j]) {
	  new_in->row[i]++;
	  w[j] = 0;
	}
      }
    }
#pragma omp barrier
#pragma omp single
    {
      nnz = 0;
      for (i = 0; i < x->d.matrix.m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[x->d.matrix.m] = nnz;
      switch (x->precision) {
      case SILC_INT:
	new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
	break;
      case SILC_LONG:
	new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
	break;
      case SILC_FLOAT:
	new_in->v.s = (float *)malloc(sizeof(float) * nnz);
	break;
      case SILC_DOUBLE:
	new_in->v.d = (double *)malloc(sizeof(double) * nnz);
	break;
      case SILC_COMPLEX:
	new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
	break;
      default:
	should_not_reach_here;
      }
      new_in->index = (int *)malloc(sizeof(int) * nnz);
      new_in->nnz = nnz;
    }

#define _mmsub_1(pre) { \
      k = new_in->row[i];					\
      for (j = 0; j < x->d.matrix.n; j++)			\
        w[j] = nnz;						\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {		\
	new_in->v.pre[k] = x_in->v.pre[j];			\
	new_in->index[k] = x_in->index[j];			\
	w[x_in->index[j]] = k++;				\
      }								\
      for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {		\
        pos = w[y_in->index[j]];				\
	if (pos == nnz) {					\
	  new_in->v.pre[k] = -y_in->v.pre[j];			\
	  new_in->index[k] =  y_in->index[j];			\
	  k++;							\
	} else {						\
	  new_in->v.pre[pos] -= y_in->v.pre[j];			\
	}							\
      }								\
    }
#define _mmsub_2(pre) { \
      k = new_in->row[i];					\
      for (j = 0; j < x->d.matrix.n; j++)			\
        w[j] = nnz;						\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {		\
	new_in->v.pre[k*2]   = x_in->v.pre[j*2];		\
	new_in->v.pre[k*2+1] = x_in->v.pre[j*2+1];		\
	new_in->index[k]     = x_in->index[j];			\
	w[x_in->index[j]] = k++;				\
      }								\
      for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {		\
        pos = w[y_in->index[j]];				\
	if (pos == nnz) {					\
	  new_in->v.pre[k*2]   = -y_in->v.pre[j*2];		\
	  new_in->v.pre[k*2+1] = -y_in->v.pre[j*2+1];		\
	  new_in->index[k]     =  y_in->index[j];		\
	  k++;							\
	} else {						\
	  new_in->v.pre[pos*2]   -= y_in->v.pre[j*2];		\
	  new_in->v.pre[pos*2+1] -= y_in->v.pre[j*2+1];		\
	}							\
      }								\
    }

    switch (x->precision) {
    case SILC_INT: 
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmsub_1(i);
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmsub_1(l);
      break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmsub_1(s);
      break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmsub_1(d);
      break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmsub_2(c);
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	_mmsub_2(z);
      break;
    default:
      should_not_reach_here;
    }
    free(w);
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      x->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  silc_crs_sort(param[2].o);
  return 0;
}

#endif /* silc_crs_{add,sub}_matrix_matrix */

#if 0 /* silc_crs_mul_matrix_matrix (v4) */

/* out: 2 */
int
silc_crs_mul_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_matrix_matrix(0x%lx, 0x%lx) [v4]\n", x, y);
#endif
  y = _transpose(y); /* XXX */
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
#pragma omp parallel
  {
    union {
      int32_t *i;
      int64_t *l;
      float *s, *c;
      double *d, *z;
    } v;
    int *index;
    size_t max_nnz, nnz_row, k;
    ssize_t i, row_start;

    max_nnz = x_in->nnz;
    index = (int *)malloc(sizeof(int) * max_nnz);

    k = 0; row_start = -1;

#define _mmprod_1(pre,type) { \
      size_t j, n, ii, iix, jj, jjx;					\
									\
      if (row_start < 0)						\
	row_start = i;							\
      new_in->row[i] = 0;						\
      if (x_in->row[i] == x_in->row[i+1])				\
	continue;							\
      nnz_row = 0;							\
      for (j = 0; j < y->d.matrix.m; j++) {				\
	value = 0;							\
	n = 0;								\
	ii = x_in->row[i]; iix = x_in->row[i+1];			\
	jj = y_in->row[j]; jjx = y_in->row[j+1];			\
	while (ii < iix && jj < jjx) {					\
	  while (ii < iix && x_in->index[ii] < y_in->index[jj])		\
	    ii++;							\
	  if (ii == iix)						\
	    break;							\
	  while (jj < jjx && x_in->index[ii] > y_in->index[jj])		\
	    jj++;							\
	  if (jj == jjx)						\
	    break;							\
	  if (x_in->index[ii] == y_in->index[jj]) {			\
	    if (k == max_nnz) {						\
	      max_nnz += x_in->nnz;					\
	      v.pre = (type *)realloc(v.pre, sizeof(type) * max_nnz);	\
	      index = (int *)realloc(index, sizeof(int) * max_nnz);	\
	    }								\
	    value += x_in->v.pre[ii++] * y_in->v.pre[jj++];		\
            n++;							\
	  }								\
	}								\
	if (n > 0) {							\
	  v.pre[k] = value;						\
	  index[k] = j;							\
	  nnz_row++;							\
	  k++;								\
	}								\
      }									\
      new_in->row[i] = nnz_row;						\
    }
#define _mmprod_2(pre,type) { \
      size_t j, n, ii, iix, jj, jjx, ii2, jj2;				\
									\
      if (row_start < 0)						\
	row_start = i;							\
      new_in->row[i] = 0;						\
      if (x_in->row[i] == x_in->row[i+1])				\
	continue;							\
      nnz_row = 0;							\
      for (j = 0; j < y->d.matrix.m; j++) {				\
	real = imag = 0;						\
	n = 0;								\
	ii = x_in->row[i]; iix = x_in->row[i+1];			\
	jj = y_in->row[j]; jjx = y_in->row[j+1];			\
	while (ii < iix && jj < jjx) {					\
	  while (ii < iix && x_in->index[ii] < y_in->index[jj])		\
	    ii++;							\
	  if (ii == iix)						\
	    break;							\
	  while (jj < jjx && x_in->index[ii] > y_in->index[jj])		\
	    jj++;							\
	  if (jj == jjx)						\
	    break;							\
	  if (x_in->index[ii] == y_in->index[jj]) {			\
	    if (k == max_nnz) {						\
	      max_nnz += x_in->nnz;					\
	      v.pre = (type *)realloc(v.pre, sizeof(type) * 2 * max_nnz); \
	      index = (int *)realloc(index, sizeof(int) * max_nnz);	\
	    }								\
	    ii2 = ii * 2; ii++;						\
	    jj2 = jj * 2; jj++;						\
	    real += (x_in->v.pre[ii2]   * y_in->v.pre[jj2]   -		\
		     x_in->v.pre[ii2+1] * y_in->v.pre[jj2+1]);		\
	    imag += (x_in->v.pre[ii2]   * y_in->v.pre[jj2+1] +		\
		     x_in->v.pre[ii2+1] * y_in->v.pre[jj2]);		\
	    n++;							\
	  }								\
	}								\
	if (n > 0) {							\
	  v.pre[k*2]   = real;						\
	  v.pre[k*2+1] = imag;						\
	  index[k] = j;							\
	  nnz_row++;							\
	  k++;								\
	}								\
      }									\
      new_in->row[i] = nnz_row;						\
    }
    switch (x->precision) {
    case SILC_INT:
      {
	int32_t value;
	v.i = (int32_t *)malloc(sizeof(int32_t) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(i, int32_t);
      }
      break;
    case SILC_LONG:
      {
	int64_t value;
	v.l = (int64_t *)malloc(sizeof(int64_t) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(l, int64_t);
      }
      break;
    case SILC_FLOAT:
      {
	float value;
	v.s = (float *)malloc(sizeof(float) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(s, float);
      }
      break;
    case SILC_DOUBLE:
      {
	double value;
	v.d = (double *)malloc(sizeof(double) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(d, double);
      }
      break;
    case SILC_COMPLEX:
      {
	float real, imag;
	v.c = (float *)malloc(sizeof(float) * 2 * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(c, float);
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double real, imag;
	v.z = (double *)malloc(sizeof(double) * 2 * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(z, double);
      }
      break;
    default:
      should_not_reach_here;
    }
#pragma omp barrier
#pragma omp single
    {
      size_t nnz;

      nnz = 0;
      for (i = 0; i < x->d.matrix.m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[x->d.matrix.m] = nnz;
      switch (x->precision) {
      case SILC_INT:
	new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
	break;
      case SILC_LONG:
	new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
	break;
      case SILC_FLOAT:
	new_in->v.s = (float *)malloc(sizeof(float) * nnz);
	break;
      case SILC_DOUBLE:
	new_in->v.d = (double *)malloc(sizeof(double) * nnz);
	break;
      case SILC_COMPLEX:
	new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
	break;
      default:
	should_not_reach_here;
      }
      new_in->index = (int *)malloc(sizeof(int) * nnz);
      new_in->nnz = nnz;
    }
    switch (x->precision) {
    case SILC_INT:
      memcpy(new_in->v.i + new_in->row[row_start], v.i,
	     sizeof(int32_t) * k);
      free(v.i);
      break;
    case SILC_LONG:
      memcpy(new_in->v.l + new_in->row[row_start], v.l,
	     sizeof(int64_t) * k);
      free(v.l);
      break;
    case SILC_FLOAT:
      memcpy(new_in->v.s + new_in->row[row_start], v.s,
	     sizeof(float) * k);
      free(v.s);
      break;
    case SILC_DOUBLE:
      memcpy(new_in->v.d + new_in->row[row_start], v.d,
	     sizeof(double) * k);
      free(v.d);
      break;
    case SILC_COMPLEX:
      memcpy(new_in->v.c + new_in->row[row_start] * 2, v.c,
	     sizeof(float) * 2 * k);
      free(v.c);
      break;
    case SILC_DOUBLE_COMPLEX:
      memcpy(new_in->v.z + new_in->row[row_start] * 2, v.z,
	     sizeof(double) * 2 * k);
      free(v.z);
      break;
    default:
      should_not_reach_here;
    }
    memcpy(new_in->index + new_in->row[row_start], index, sizeof(int) * k);
    free(index);
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      y->d.matrix.m); /* n */
  param[2].o->d.matrix.in = (void *)new_in;
  silc_object_free(y);
  return 0;
}

#else

#if 1 /* silc_crs_mul_matrix_matrix (v3) */

/* out: 2 */
int
silc_crs_mul_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  double t, elapsed;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_matrix_matrix(0x%lx, 0x%lx) [v3]\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
#ifdef TIME_CRS_MMUL
  t = silc_wtime();
#endif

  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
#pragma omp parallel
  {
    union {
      int32_t *i;
      int64_t *l;
      float *s, *c;
      double *d, *z;
    } v;
    int *index;
    size_t max_nnz, nnz_row, k;
    ssize_t *w, pos, i, row_start;
#ifdef _WIN32
    unsigned long long flop;
#else
    size_t flop;
#endif

    max_nnz = x_in->nnz;
    index = (int *)malloc(sizeof(int) * max_nnz);
    w = (ssize_t *)malloc(sizeof(ssize_t) * y->d.matrix.n);

    k = 0; row_start = -1; flop = 0;

#define _mmprod_1(pre, type) { \
      size_t j, ii, jj, column;						\
									\
      if (row_start < 0)						\
	row_start = i;							\
      for (j = 0; j < y->d.matrix.n; j++)				\
        w[j] = -1;							\
      nnz_row = 0;							\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {			\
	ii = x_in->index[j];						\
	for (jj = y_in->row[ii]; jj < y_in->row[ii+1]; jj++) {		\
	  if (k == max_nnz) {						\
	    max_nnz += x_in->nnz;					\
	    v.pre = (type *)realloc(v.pre, sizeof(type) * max_nnz);	\
	    index = (int *)realloc(index, sizeof(int) * max_nnz);	\
	  }								\
	  value = x_in->v.pre[j] * y_in->v.pre[jj];			\
          column = y_in->index[jj];					\
          pos = w[column];						\
	  if (pos < 0) {						\
	    v.pre[k] = value;						\
	    index[k] = column; w[column] = k++; nnz_row++;		\
	  } else {							\
	    v.pre[pos] += value;					\
	  }								\
	  flop += 2;							\
	}								\
      }									\
      new_in->row[i] = nnz_row;						\
    }
#define _mmprod_2(pre, type) { \
      size_t j, ii, jj, column, j2, jj2;				\
									\
      if (row_start < 0)						\
	row_start = i;							\
      for (j = 0; j < y->d.matrix.n; j++)				\
        w[j] = -1;							\
      nnz_row = 0;							\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {			\
	j2 = j * 2;							\
	ii = x_in->index[j];						\
	for (jj = y_in->row[ii]; jj < y_in->row[ii+1]; jj++) {		\
	  jj2 = jj * 2;							\
	  if (k == max_nnz) {						\
	    max_nnz += x_in->nnz;					\
	    v.pre = (type *)realloc(v.pre, sizeof(type) * 2 * max_nnz);	\
	    index = (int *)realloc(index, sizeof(int) * max_nnz);	\
	  }								\
	  real = (x_in->v.pre[j2]   * y_in->v.pre[jj2]   -		\
		  x_in->v.pre[j2+1] * y_in->v.pre[jj2+1]);		\
	  imag = (x_in->v.pre[j2]   * y_in->v.pre[jj2+1] +		\
		  x_in->v.pre[j2+1] * y_in->v.pre[jj2]);		\
          column = y_in->index[jj];					\
          pos = w[column];						\
	  if (pos < 0) {						\
	    v.pre[k*2]   = real;					\
	    v.pre[k*2+1] = imag;					\
	    index[k] = column; w[column] = k++; nnz_row++;		\
	  } else {							\
	    v.pre[pos*2]   += real;					\
	    v.pre[pos*2+1] += imag;					\
	  }								\
	  flop += 8;							\
	}								\
      }									\
      new_in->row[i] = nnz_row;						\
    }

    switch (x->precision) {
    case SILC_INT:
      {
	int32_t value;
	v.i = (int32_t *)malloc(sizeof(int32_t) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(i, int32_t);
      }
      break;
    case SILC_LONG:
      {
	int64_t value;
	v.l = (int64_t *)malloc(sizeof(int64_t) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(l, int64_t);
      }
      break;
    case SILC_FLOAT:
      {
	float value;
	v.s = (float *)malloc(sizeof(float) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(s, float);
      }
      break;
    case SILC_DOUBLE:
      {
	double value;
	v.d = (double *)malloc(sizeof(double) * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(d, double);
      }
      break;
    case SILC_COMPLEX:
      {
	float real, imag;
	v.c = (float *)malloc(sizeof(float) * 2 * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(c, float);
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double real, imag;
	v.z = (double *)malloc(sizeof(double) * 2 * max_nnz);
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(z, double);
      }
      break;
    default:
      should_not_reach_here;
    }
#ifdef DEBUG
    printf("silc_crs_mul_matrix_matrix: flop = %ld\n", flop);
#endif
    free(w);
#pragma omp barrier
#pragma omp single
    {
      size_t nnz;

      nnz = 0;
      for (i = 0; i < x->d.matrix.m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[x->d.matrix.m] = nnz;
      switch (x->precision) {
      case SILC_INT:
	new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
	break;
      case SILC_LONG:
	new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
	break;
      case SILC_FLOAT:
	new_in->v.s = (float *)malloc(sizeof(float) * nnz);
	break;
      case SILC_DOUBLE:
	new_in->v.d = (double *)malloc(sizeof(double) * nnz);
	break;
      case SILC_COMPLEX:
	new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
	break;
      default:
	should_not_reach_here;
      }
      new_in->index = (int *)malloc(sizeof(int) * nnz);
      new_in->nnz = nnz;
    }
    switch (x->precision) {
    case SILC_INT:
      memcpy(new_in->v.i + new_in->row[row_start], v.i,
	     sizeof(int32_t) * k);
      free(v.i);
      break;
    case SILC_LONG:
      memcpy(new_in->v.l + new_in->row[row_start], v.l,
	     sizeof(int64_t) * k);
      free(v.l);
      break;
    case SILC_FLOAT:
      memcpy(new_in->v.s + new_in->row[row_start], v.s,
	     sizeof(float) * k);
      free(v.s);
      break;
    case SILC_DOUBLE:
      memcpy(new_in->v.d + new_in->row[row_start], v.d,
	     sizeof(double) * k);
      free(v.d);
      break;
    case SILC_COMPLEX:
      memcpy(new_in->v.c + new_in->row[row_start] * 2, v.c,
	     sizeof(float) * 2 * k);
      free(v.c);
      break;
    case SILC_DOUBLE_COMPLEX:
      memcpy(new_in->v.z + new_in->row[row_start] * 2, v.z,
	     sizeof(double) * 2 * k);
      free(v.z);
      break;
    default:
      should_not_reach_here;
    }
    memcpy(new_in->index + new_in->row[row_start], index, sizeof(int) * k);
    free(index);
  }
#ifdef TIME_CRS_MMUL
  elapsed = silc_wtime() - t;
  printf("sparse_crs_op: silc_crs_mul_matrix_matrix: %lf sec.\n", elapsed);
  printf("sparse_crs_op: nnz of x = %ld, of y = %ld, of xy = %ld\n",
	 x_in->nnz, y_in->nnz, new_in->nnz);
#endif
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      y->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  silc_crs_sort(param[2].o);
  return 0;
}

#else

#if 1 /* silc_crs_mul_matrix_matrix (v2) */

/* out: 2 */
int
silc_crs_mul_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  size_t nnz, nnz_row;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_matrix_matrix(0x%lx, 0x%lx) [v2]\n", x, y);
#endif
  y = _transpose(y); /* XXX */
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
#pragma omp parallel
  {
    ssize_t i;

#pragma omp for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j, ii, iix, jj, jjx;

      new_in->row[i] = 0;
      if (x_in->row[i] == x_in->row[i+1])
	continue;
      for (j = 0; j < y->d.matrix.m; j++) {
	ii = x_in->row[i]; iix = x_in->row[i+1];
	jj = y_in->row[j]; jjx = y_in->row[j+1];
	while (ii < iix && jj < jjx) {
	  while (ii < iix && x_in->index[ii] < y_in->index[jj])
	    ii++;
	  if (ii == iix)
	    break;
	  while (jj < jjx && x_in->index[ii] > y_in->index[jj])
	    jj++;
	  if (jj == jjx)
	    break;
	  if (x_in->index[ii] == y_in->index[jj]) {
	    new_in->row[i]++; ii++; jj++;
	    break;
	  }
	}
      }
    }
#pragma omp barrier
#pragma omp single
    {
      nnz = 0;
      for (i = 0; i < x->d.matrix.m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[x->d.matrix.m] = nnz;
      switch (x->precision) {
      case SILC_INT:
	new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
	break;
      case SILC_LONG:
	new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
	break;
      case SILC_FLOAT:
	new_in->v.s = (float *)malloc(sizeof(float) * nnz);
	break;
      case SILC_DOUBLE:
	new_in->v.d = (double *)malloc(sizeof(double) * nnz);
	break;
      case SILC_COMPLEX:
	new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
	break;
      default:
	should_not_reach_here;
      }
      new_in->index = (int *)malloc(sizeof(int) * nnz);
      new_in->nnz = nnz;
    }

#define _mmprod_1(pre) { \
      size_t j, k, n, ii, iix, jj, jjx;					\
      if (x_in->row[i] == x_in->row[i+1])				\
	continue;							\
      k = new_in->row[i];						\
      if (k == new_in->row[i+1])					\
	continue;							\
      for (j = 0; j < y->d.matrix.m; j++) {				\
	value = 0;							\
	n = 0;								\
	ii = x_in->row[i]; iix = x_in->row[i+1];			\
	jj = y_in->row[j]; jjx = y_in->row[j+1];			\
	while (ii < iix && jj < jjx) {					\
	  while (ii < iix && x_in->index[ii] < y_in->index[jj])		\
	    ii++;							\
	  if (ii == iix)						\
	    break;							\
	  while (jj < jjx && x_in->index[ii] > y_in->index[jj])		\
	    jj++;							\
	  if (jj == jjx)						\
	    break;							\
	  if (x_in->index[ii] == y_in->index[jj]) {			\
	    value += x_in->v.pre[ii++] * y_in->v.pre[jj++];		\
	    n++;							\
	  }								\
	}								\
	if (n > 0) {							\
	  new_in->v.pre[k] = value;					\
	  new_in->index[k] = j;						\
	  k++;								\
	}								\
      }									\
    }
#define _mmprod_2(pre) { \
      size_t j, k, n, ii, jj, iix, jjx, ii2, jj2;			\
      if (x_in->row[i] == x_in->row[i+1])				\
	continue;							\
      k = new_in->row[i];						\
      if (k == new_in->row[i+1])					\
	continue;							\
      for (j = 0; j < y->d.matrix.m; j++) {				\
	real = imag = 0;						\
	n = 0;								\
	ii = x_in->row[i]; iix = x_in->row[i+1];			\
	jj = y_in->row[j]; jjx = y_in->row[j+1];			\
	while (ii < iix && jj < jjx) {					\
	  while (ii < iix && x_in->index[ii] < y_in->index[jj])		\
	    ii++;							\
	  if (ii == iix)						\
	    break;							\
	  while (jj < jjx && x_in->index[ii] > y_in->index[jj])		\
	    jj++;							\
	  if (jj == jjx)						\
	    break;							\
	  if (x_in->index[ii] == y_in->index[jj]) {			\
	    ii2 = ii * 2; ii++;						\
	    jj2 = jj * 2; jj++;						\
	    real += (x_in->v.pre[ii2]	* y_in->v.pre[jj2]   -		\
		     x_in->v.pre[ii2+1] * y_in->v.pre[jj2+1]);		\
	    imag += (x_in->v.pre[ii2]	* y_in->v.pre[jj2+1] +		\
		     x_in->v.pre[ii2+1] * y_in->v.pre[jj2]);		\
	    n++;							\
	  }								\
	}								\
	if (n > 0) {							\
	  new_in->v.pre[k*2]   = real;					\
	  new_in->v.pre[k*2+1] = imag;					\
	  new_in->index[k] = j;						\
	  k++;								\
	}								\
      }									\
    }
    switch (x->precision) {
    case SILC_INT: 
      {
	int32_t value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(i);
      }
      break;
    case SILC_LONG:
      {
	int64_t value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(l);
      }
      break;
    case SILC_FLOAT:
      {
	float value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(s);
      }
      break;
    case SILC_DOUBLE:
      {
	double value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(d);
      }
      break;
    case SILC_COMPLEX:
      {
	float real, imag;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(c);
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double real, imag;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(z);
      }
      break;
    default:
      should_not_reach_here;
    }
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      y->d.matrix.m); /* n */
  param[2].o->d.matrix.in = (void *)new_in;
  silc_object_free(y);
  return 0;
}

#else /* silc_crs_mul_matrix_matrix (v1) */

/* out: 2 */
int
silc_crs_mul_matrix_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  size_t nnz, nnz_row;
  double t, elapsed;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_matrix_matrix(0x%lx, 0x%lx) [v1]\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
#ifdef TIME_CRS_MMUL
  t = silc_wtime();
#endif
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
#pragma omp parallel
  {
    size_t *w;
    ssize_t i;

    w = (size_t *)calloc(y->d.matrix.n, sizeof(size_t));
#pragma omp for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++) {
      size_t j, ii, jj;
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {
	ii = x_in->index[j];
	for (jj = y_in->row[ii]; jj < y_in->row[ii+1]; jj++)
	  w[y_in->index[jj]] = 1;
      }
      new_in->row[i] = 0;
      for (j = 0; j < y->d.matrix.n; j++) {
	if (w[j]) {
	  new_in->row[i]++;
	  w[j] = 0;
	}
      }
    }
#pragma omp barrier
#pragma omp single
    {
      nnz = 0;
      for (i = 0; i < x->d.matrix.m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[x->d.matrix.m] = nnz;
      switch (x->precision) {
      case SILC_INT:
	new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
	break;
      case SILC_LONG:
	new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
	break;
      case SILC_FLOAT:
	new_in->v.s = (float *)malloc(sizeof(float) * nnz);
	break;
      case SILC_DOUBLE:
	new_in->v.d = (double *)malloc(sizeof(double) * nnz);
	break;
      case SILC_COMPLEX:
	new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
	break;
      default:
	should_not_reach_here;
      }
      new_in->index = (int *)malloc(sizeof(int) * nnz);
      new_in->nnz = nnz;
    }

#define _mmprod_1(pre) { \
      size_t j, k, ii, jj, column, pos;				\
								\
      for (j = 0; j < y->d.matrix.n; j++)			\
        w[j] = nnz;						\
      k = new_in->row[i];					\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {		\
	ii = x_in->index[j];					\
	for (jj = y_in->row[ii]; jj < y_in->row[ii+1]; jj++) {	\
	  value = x_in->v.pre[j] * y_in->v.pre[jj];		\
          column = y_in->index[jj];				\
          pos = w[column];					\
	  if (pos == nnz) {					\
	    new_in->v.pre[k] = value;				\
	    new_in->index[k] = column;				\
	    w[column] = k++;					\
	  } else {						\
	    new_in->v.pre[pos] += value;			\
	  }							\
	}							\
      }								\
    }
#define _mmprod_2(pre) { \
      size_t j, k, ii, jj, column, pos, j2, jj2;		\
								\
      for (j = 0; j < y->d.matrix.n; j++)			\
        w[j] = nnz;						\
      k = new_in->row[i];					\
      for (j = x_in->row[i]; j < x_in->row[i+1]; j++) {		\
	j2 = j * 2;						\
	ii = x_in->index[j];					\
	for (jj = y_in->row[ii]; jj < y_in->row[ii+1]; jj++) {	\
	  jj2 = jj * 2;						\
	  real = (x_in->v.pre[j2]   * y_in->v.pre[jj2]   -	\
		  x_in->v.pre[j2+1] * y_in->v.pre[jj2+1]);	\
	  imag = (x_in->v.pre[j2]   * y_in->v.pre[jj2+1] +	\
		  x_in->v.pre[j2+1] * y_in->v.pre[jj2]);	\
          column = y_in->index[jj];				\
          pos = w[column];					\
	  if (pos == nnz) {					\
	    new_in->v.pre[k*2]   = real;			\
	    new_in->v.pre[k*2+1] = imag;			\
	    new_in->index[k] = column;				\
	    w[column] = k++;					\
	  } else {						\
	    new_in->v.pre[pos*2]   += real;			\
	    new_in->v.pre[pos*2+1] += imag;			\
	  }							\
	}							\
      }								\
    }

    switch (x->precision) {
    case SILC_INT: 
      {
	int32_t value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(i);
      }
      break;
    case SILC_LONG:
      {
	int64_t value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(l);
      }
      break;
    case SILC_FLOAT:
      {
	float value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(s);
      }
      break;
    case SILC_DOUBLE:
      {
	double value;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_1(d);
      }
      break;
    case SILC_COMPLEX:
      {
	float real, imag;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(c);
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      {
	double real, imag;
#pragma omp for
	for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
	  _mmprod_2(z);
      }
      break;
    default:
      should_not_reach_here;
    }
    free(w);
  }
#ifdef TIME_CRS_MMUL
  elapsed = silc_wtime() - t;
  printf("sparse_crs_op: silc_crs_mul_matrix_matrix: %lf sec.\n", elapsed);
  printf("sparse_crs_op: nnz of x = %ld, of y = %ld, of xy = %ld\n",
	 x_in->nnz, y_in->nnz, new_in->nnz);
#endif
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      y->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  silc_crs_sort(param[2].o);
  return 0;
}

#endif /* silc_crs_mul_matrix_matrix */

#endif /* silc_crs_mul_matrix_matrix */

#endif /* silc_crs_mul_matrix_matrix */

/* out: 2 */
int
silc_crs_mul_matrix_elementwise(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  ssize_t i;
  size_t nnz;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_mul_matrix_elementwise(0x%lx, 0x%lx)\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
  new_in->row[0] = nnz = 0;
  for (i = 0; i < x->d.matrix.m; i++) {
    int jx, jjx, jy, jjy;
    jx = x_in->row[i]; jjx = x_in->row[i+1];
    jy = y_in->row[i]; jjy = y_in->row[i+1];
    while (jx < jjx && jy < jjy) {
      if (x_in->index[jx] == y_in->index[jy]) {
	nnz++; jx++; jy++; continue;
      }
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {
	jx++;
      }
      while (jy < jjy && x_in->index[jx] > y_in->index[jy]) {
	jy++;
      }
    }
    new_in->row[i+1] = nnz;
  }
  new_in->nnz = nnz;
  new_in->index = (int *)malloc(sizeof(int) * nnz);
  
#define _mul_elementwise_1(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j] = x_in->v.pre[jx] * y_in->v.pre[jy];	\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
	jx++;							\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	jy++;							\
      }								\
    }								\
  }
#define _mul_elementwise_2(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j*2]   = (x_in->v.pre[jx*2]   * y_in->v.pre[jy*2]   - \
				x_in->v.pre[jx*2+1] * y_in->v.pre[jy*2+1]); \
	new_in->v.pre[j*2+1] = (x_in->v.pre[jx*2]   * y_in->v.pre[jy*2+1] + \
				x_in->v.pre[jx*2+1] * y_in->v.pre[jy*2]);   \
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
	jx++;							\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	jy++;							\
      }								\
    }								\
  }

  switch (x->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _mul_elementwise_1(i);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _mul_elementwise_1(l);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)malloc(sizeof(float) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _mul_elementwise_1(s);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)malloc(sizeof(double) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _mul_elementwise_1(d);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _mul_elementwise_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _mul_elementwise_2(z);
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      x->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  return 0;
}

/* out: 2 */
int
silc_crs_div_matrix_elementwise(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;
  silc_crs_t *x_in, *y_in, *new_in;
  ssize_t i;
  size_t nnz;
  int error;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_div_matrix_elementwise(0x%lx, 0x%lx)\n", x, y);
#endif
  x_in = SILC_CRS(x);
  y_in = SILC_CRS(y);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->row = (int *)malloc(sizeof(int) * (x->d.matrix.m + 1));
  new_in->row[0] = nnz = 0;
  error = 0;
  for (i = 0; i < x->d.matrix.m; i++) {
    int jx, jjx, jy, jjy;
    jx = x_in->row[i]; jjx = x_in->row[i+1];
    jy = y_in->row[i]; jjy = y_in->row[i+1];
    while (jx < jjx && jy < jjy) {
      if (x_in->index[jx] == y_in->index[jy]) {
	switch (x->precision) {
	case SILC_INT:
	  if (y_in->v.i[jy] == 0)
	    error = 1;
	  break;
	case SILC_LONG:
	  if (y_in->v.l[jy] == 0)
	    error = 1;
	  break;
	case SILC_FLOAT:
	  if (y_in->v.s[jy] == 0.0)
	    error = 1;
	  break;
	case SILC_DOUBLE:
	  if (y_in->v.d[jy] == 0.0)
	    error = 1;
	  break;
	case SILC_COMPLEX:
	  if (y_in->v.c[jy*2] == 0.0 && y_in->v.c[jy*2+1] == 0.0)
	    error = 1;
	  break;
	case SILC_DOUBLE_COMPLEX:
	  if (y_in->v.z[jy*2] == 0.0 && y_in->v.z[jy*2+1] == 0.0)
	    error = 1;
	  break;
	default:
	  should_not_reach_here;
	}
	nnz++; jx++; jy++; continue;
      }
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {
	error = 1;
	jx++;
      }
      while (jy < jjy && x_in->index[jx] > y_in->index[jy]) {
	jy++;
      }
    }
    new_in->row[i+1] = nnz;
  }
  if (error) {
    fprintf(stderr, "silc_crs_div_matrix_elementwise: division by zero\n");
    free(new_in->row);
    free(new_in);
    return -1;
  }
  new_in->nnz = nnz;
  new_in->index = (int *)malloc(sizeof(int) * nnz);

#define _div_elementwise_1(pre) { \
    int j, jx, jjx, jy, jjy;					\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
	new_in->v.pre[j] = x_in->v.pre[jx] / y_in->v.pre[jy];	\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
        should_not_reach_here;					\
	jx++;							\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	jy++;							\
      }								\
    }								\
  }
#define _div_elementwise_2(pre) { \
    int j, jx, jjx, jy, jjy;					\
    double div;							\
    j = new_in->row[i];						\
    jx = x_in->row[i]; jjx = x_in->row[i+1];			\
    jy = y_in->row[i]; jjy = y_in->row[i+1];			\
    while (jx < jjx && jy < jjy) {				\
      if (x_in->index[jx] == y_in->index[jy]) {			\
        div = (y_in->v.pre[jy*2]   * y_in->v.pre[jy*2] +	\
	       y_in->v.pre[jy*2+1] * y_in->v.pre[jy*2+1]);	\
	new_in->v.pre[j*2]					\
	  = (x_in->v.pre[jx*2]   * y_in->v.pre[jy*2] +		\
	     x_in->v.pre[jx*2+1] * y_in->v.pre[jy*2+1]) / div;	\
	new_in->v.pre[j*2+1]					\
	  = (x_in->v.pre[jx*2+1] * y_in->v.pre[jy*2] -		\
	     x_in->v.pre[jx*2]   * y_in->v.pre[jy*2+1]) / div;	\
	new_in->index[j] = x_in->index[jx];			\
	j++; jx++; jy++; continue;				\
      }								\
      while (jx < jjx && x_in->index[jx] < y_in->index[jy]) {	\
        should_not_reach_here;					\
	jx++;							\
      }								\
      while (jy < jjy && y_in->index[jy] < x_in->index[jx]) {	\
	jy++;							\
      }								\
    }								\
  }

  switch (x->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _div_elementwise_1(i);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _div_elementwise_1(l);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)malloc(sizeof(float) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _div_elementwise_1(s);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)malloc(sizeof(double) * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _div_elementwise_1(d);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _div_elementwise_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _div_elementwise_2(z);
    break;
  default:
    should_not_reach_here;
  }
  param[2].o = silc_object_new_matrix(x->precision,
				      x->format,
				      x->d.matrix.m,
				      x->d.matrix.n);
  param[2].o->d.matrix.in = (void *)new_in;
  return 0;
}

static void
_add_crs_to_dense(silc_object_t *x, silc_object_t *y)
{
  silc_dense_t *x_in;
  silc_crs_t *y_in;
  ssize_t i;

  x_in = SILC_DENSE(x);
  y_in = SILC_CRS(y);

#define _add_crs_to_dense_1(pre) {					  \
    size_t j;								  \
    for (j = y_in->row[i]; j < y_in->row[i+1]; j++)			  \
      x_in->v.pre[i + y_in->index[j] * x->d.matrix.m] += y_in->v.pre[j];  \
  }
#define _add_crs_to_dense_2(pre) {					  \
    size_t j, jj, kk;							  \
    for (j = y_in->row[i]; j < y_in->row[i+1]; j++) {			  \
      kk = (i + y_in->index[j] * x->d.matrix.m) * 2;			  \
      jj = j * 2;							  \
      x_in->v.pre[kk]   += y_in->v.pre[jj];				  \
      x_in->v.pre[kk+1] += y_in->v.pre[jj+1];				  \
    }									  \
  }
  switch (x->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_crs_to_dense_1(i);
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_crs_to_dense_1(l);
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_crs_to_dense_1(s);
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_crs_to_dense_1(d);
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_crs_to_dense_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)x->d.matrix.m; i++)
      _add_crs_to_dense_2(z);
    break;
  default:
    should_not_reach_here;
  }
}

int
silc_crs_add_dense_matrix(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y;
  silc_modifier_t *x_mods, *y_mods;

  x = param[0].o; x_mods = param[0].m;
  y = param[1].o; y_mods = param[1].m;
#ifdef DEBUG
  printf("silc_crs_add_dense_matrix(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->format == silc_find_format_id(SILC_FORMAT_DENSE))
    _add_crs_to_dense(x, y); /* out: 0 */
  else
    _add_crs_to_dense(y, x); /* out: 1 */
  return 0;
}

/* functions */

static silc_object_t *
_zeros(double v, size_t m, size_t n)
{
  silc_object_t *object;
  silc_crs_t *in;
  int format;
  size_t i, j;

  if (m == 1 && n == 1) {
    object = silc_object_new_double(v);
  } else if (m > 1 && n == 1) {
    object = silc_object_new_column_vector(SILC_DOUBLE, m);
    for (i = 0; i < object->d.vector.length; i++)
      object->d.vector.v.d[i] = v;
  } else if (m == 1 && n > 1) {
    object = silc_object_new_row_vector(SILC_DOUBLE, n);
    for (i = 0; i < object->d.vector.length; i++)
      object->d.vector.v.d[i] = v;
  } else if (m > 1 && n > 1) {
    format = silc_find_format_id(SILC_FORMAT_CRS);
    object = silc_object_new_matrix(SILC_DOUBLE, format, m, n);
    object->d.matrix.in = (void *)silc_crs_allocate(object->precision,
						    object->d.matrix.m,
						    0);
    in = SILC_CRS(object);
    for (i = 0; i <= object->d.matrix.m; i++)
      in->row[i] = 0;
  } else {
    return NULL;  
  }
  return object;
}

int
silc_crs_func_zeros_1(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  size_t n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_zeros_1(0x%lx)\n", object);
#endif
  if (object->type != SILC_SCALAR_TYPE)
    return -1;
  switch (object->precision) {
  case SILC_INT:
    n = object->d.scalar.i; break;
  case SILC_LONG:
    n = object->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "zeros: argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "zeros: invalid dimension\n");
    return -1;
  }
  param[1].o = _zeros(0, n ,n);
  return (param[1].o) ? 0 : -1;
}

int
silc_crs_func_zeros_2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *size1, *size2;
  size_t m, n;

  size1 = param[0].o;
  size2 = param[1].o;
#ifdef DEBUG
  printf("silc_crs_func_zeros_2(0x%lx, 0x%lx)\n", size1, size2);
#endif
  if (size1->type != SILC_SCALAR_TYPE ||
      size2->type != SILC_SCALAR_TYPE)
    return -1;
  switch (size1->precision) {
  case SILC_INT:
    m = size1->d.scalar.i; break;
  case SILC_LONG:
    m = size1->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "zeros: 1st argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (m < 1) {
    fprintf(stderr, "zeros: 1st dimension is invalid\n");
    return -1;
  }
  switch (size2->precision) {
  case SILC_INT:
    n = size2->d.scalar.i; break;
  case SILC_LONG:
    n = size2->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "zeros: 2nd argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "zeros: 2nd dimension is invalid\n");
    return -1;
  }
  param[2].o = _zeros(0, m ,n);
  return (param[2].o) ? 0 : -1;
}

int
silc_crs_func_eye(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  silc_crs_t *in;
  ssize_t i, n;
  int format;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_eye(0x%lx)\n", object);
#endif
  if (object->type != SILC_SCALAR_TYPE)
    return -1;
  switch (object->precision) {
  case SILC_INT:
    n = object->d.scalar.i; break;
  case SILC_LONG:
    n = object->d.scalar.l; break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "sparse_crs_op: eye: operand must be integer\n");
    return -1;
  default:
    should_not_reach_here;
  }
  in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  in->nnz = n;
  in->v.d = (double *)malloc(sizeof(double) * n);
  in->row = (int *)malloc(sizeof(int) * (n + 1));
  in->index = (int *)malloc(sizeof(int) * n);
  for (i = 0; i <= n; i++)
    in->row[i] = i;
#pragma omp parallel for
  for (i = 0; i < n; i++) {
    in->v.d[i] = 1.0;
    in->index[i] = i;
  }
  format = silc_find_format_id(SILC_FORMAT_CRS);
  param[1].o = silc_object_new_matrix(SILC_DOUBLE, format, n, n);
  param[1].o->d.matrix.in = (void *)in;
  return 0;
}

int
silc_crs_func_fliplr(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_crs_t *in, *new_in;
  const char *format_name;
  ssize_t i, m, n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_fliplr(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  format_name = silc_find_format_name(object->format);
  if (strcmp(format_name, SILC_FORMAT_CRS)) {
    /* XXX need a workaround */
    fprintf(stderr, "sparse_crs_op: fliplr: not in the CRS format\n");
    return -1;
  }
  in = SILC_CRS(object);
  m = object->d.matrix.m;
  n = object->d.matrix.n;
  new_in = silc_crs_allocate(object->precision, m, in->nnz);
  for (i = 0; i <= m; i++)
    new_in->row[i] = in->row[i];
#pragma omp parallel for
  for (i = 0; i < (ssize_t)m; i++) {
    int j, k;
    k = in->row[i+1] + in->row[i] - 1;
    for (j = in->row[i]; j < in->row[i+1]; j++)
      new_in->index[k-j] = n-1-in->index[j];
  }
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++) {
      int j, k;
      k = in->row[i+1] + in->row[i] - 1;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.i[k-j] = in->v.i[j];
    }
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++) {
      int j, k;
      k = in->row[i+1] + in->row[i] - 1;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.l[k-j] = in->v.l[j];
    }
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++) {
      int j, k;
      k = in->row[i+1] + in->row[i] - 1;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.s[k-j] = in->v.s[j];
    }
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++) {
      int j, k;
      k = in->row[i+1] + in->row[i] - 1;
      for (j = in->row[i]; j < in->row[i+1]; j++)
	new_in->v.d[k-j] = in->v.d[j];
    }
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++) {
      int j, k;
      k = in->row[i+1] + in->row[i] - 1;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	new_in->v.c[(k-j)*2]   = in->v.c[j*2];
	new_in->v.c[(k-j)*2+1] = in->v.c[j*2+1];
      }
    }
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++) {
      int j, k;
      k = in->row[i+1] + in->row[i] - 1;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	new_in->v.z[(k-j)*2]   = in->v.z[j*2];
	new_in->v.z[(k-j)*2+1] = in->v.z[j*2+1];
      }
    }
    break;
  default:
    should_not_reach_here;
  }
  param[1].o = silc_object_new_matrix(object->precision,
				      object->format,
				      object->d.matrix.m,
				      object->d.matrix.n);
  param[1].o->d.matrix.in = (void *)new_in;
  return 0;
}

int
silc_crs_func_flipud(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_crs_t *in, *new_in;
  const char *format_name;
  ssize_t i, m;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_flipud(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  format_name = silc_find_format_name(object->format);
  if (strcmp(format_name, SILC_FORMAT_CRS)) {
    /* XXX need a workaround */
    fprintf(stderr, "sparse_crs_op: flipud: not in the CRS format\n");
    return -1;
  }
  in = SILC_CRS(object);
  m = object->d.matrix.m;
  new_in = silc_crs_allocate(object->precision, m, in->nnz);
  new_in->row[0] = 0;
  for (i = 1; i <= m; i++)
    new_in->row[i] = new_in->row[i-1] + in->row[m-i+1] - in->row[m-i];
#pragma omp parallel for
  for (i = 0; i < (ssize_t)m; i++) {
    memcpy(new_in->index + new_in->row[i], in->index + in->row[m-i-1],
	   sizeof(int) * (in->row[m-i] - in->row[m-i-1]));
  }
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memcpy(new_in->v.i + new_in->row[i], in->v.i + in->row[m-i-1],
	     sizeof(int32_t) * (in->row[m-i] - in->row[m-i-1]));
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memcpy(new_in->v.l + new_in->row[i], in->v.l + in->row[m-i-1],
	     sizeof(int64_t) * (in->row[m-i] - in->row[m-i-1]));
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memcpy(new_in->v.s + new_in->row[i], in->v.s + in->row[m-i-1],
	     sizeof(float) * (in->row[m-i] - in->row[m-i-1]));
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memcpy(new_in->v.d + new_in->row[i], in->v.d + in->row[m-i-1],
	     sizeof(double) * (in->row[m-i] - in->row[m-i-1]));
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memcpy(new_in->v.c + 2 * new_in->row[i], in->v.c + 2 * in->row[m-i-1],
	     sizeof(float) * 2 * (in->row[m-i] - in->row[m-i-1]));
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < (ssize_t)m; i++)
      memcpy(new_in->v.z + 2 * new_in->row[i], in->v.z + 2 * in->row[m-i-1],
	     sizeof(double) * 2 * (in->row[m-i] - in->row[m-i-1]));
    break;
  default:
    should_not_reach_here;
  }
  param[1].o = silc_object_new_matrix(object->precision,
				      object->format,
				      object->d.matrix.m,
				      object->d.matrix.n);
  param[1].o->d.matrix.in = (void *)new_in;
  return 0;
}

static silc_object_t *
_diag(silc_object_t *x, int k)
{
  silc_object_t *new_object;
  silc_crs_t *new_in;
  int i, n;
  size_t nnz;

  n = x->d.vector.length + abs(k);
  new_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  new_in->index = (int *)malloc(sizeof(int) * x->d.vector.length);
  new_in->row = (int *)malloc(sizeof(int) * (n + 1));
  new_in->row[0] = nnz = 0;
  for (i = 0; i < n; i++) {
    if (0 <= i + k && i + k < n)
      new_in->index[nnz++] = i + k;
    new_in->row[i+1] = nnz;
  }
  new_in->nnz = nnz;
#define _diag_1(pre) { \
    size_t j;							\
    for (j = new_in->row[i]; j < new_in->row[i+1]; j++)		\
      new_in->v.pre[j] = x->d.vector.v.pre[j];			\
  }
#define _diag_2(pre) { \
    size_t j;							\
    for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
      new_in->v.pre[j*2]   = x->d.vector.v.pre[j*2];		\
      new_in->v.pre[j*2+1] = x->d.vector.v.pre[j*2+1];		\
    }								\
  }
  switch (x->precision) {
  case SILC_INT:
    new_in->v.i = (int32_t *)malloc(sizeof(int32_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diag_1(i);
    break;
  case SILC_LONG:
    new_in->v.l = (int64_t *)malloc(sizeof(int64_t) * nnz);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diag_1(l);
    break;
  case SILC_FLOAT:
    new_in->v.s = (float *)malloc(sizeof(float) * nnz);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diag_1(s);
    break;
  case SILC_DOUBLE:
    new_in->v.d = (double *)malloc(sizeof(double) * nnz);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diag_1(d);
    break;
  case SILC_COMPLEX:
    new_in->v.c = (float *)malloc(sizeof(float) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diag_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
    new_in->v.z = (double *)malloc(sizeof(double) * 2 * nnz);
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diag_2(z);
    break;
  default:
    should_not_reach_here;
  }
  new_object = silc_object_new_matrix(x->precision,
				      silc_find_format_id(SILC_FORMAT_CRS),
				      n, n);
  new_object->d.matrix.in = (void *)new_in;
  return new_object;
}

int
silc_crs_func_diag_1(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x;

  x = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_diag_1(0x%lx)\n", x);
#endif
  if (x->type != SILC_COLUMN_VECTOR_TYPE && x->type != SILC_ROW_VECTOR_TYPE)
    return -1;
  param[1].o = _diag(x, 0);
  return 0;
}

int
silc_crs_func_diag_2(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *x, *y, *tmp = NULL;
  int k;

  x = param[0].o;
  y = param[1].o;
#ifdef DEBUG
  printf("silc_crs_func_diag_2(0x%lx, 0x%lx)\n", x, y);
#endif
  if (x->type != SILC_COLUMN_VECTOR_TYPE &&
      x->type != SILC_ROW_VECTOR_TYPE &&
      x->type != SILC_SCALAR_TYPE)
    return -1;
  if (y->type != SILC_SCALAR_TYPE)
    return -1;
  switch (y->precision) {
  case SILC_INT:
    k = y->d.scalar.i;
    break;
  case SILC_LONG:
    k = y->d.scalar.l;
    break;
  case SILC_FLOAT:
  case SILC_DOUBLE:
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "diag: 2nd argument must be integer\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (x->type == SILC_SCALAR_TYPE) {
    tmp = silc_object_new_column_vector(x->precision, 1);
    memcpy(tmp->d.vector.v.i, &x->d.scalar.i, SILC_SCALAR_SIZE(x));
    switch (x->precision) {
    case SILC_INT:
      tmp->d.vector.v.i[0] = x->d.scalar.i; break;
    case SILC_LONG:
      tmp->d.vector.v.l[0] = x->d.scalar.l; break;
    case SILC_FLOAT:
      tmp->d.vector.v.s[0] = x->d.scalar.s; break;
    case SILC_DOUBLE:
      tmp->d.vector.v.d[0] = x->d.scalar.d; break;
    case SILC_COMPLEX:
      tmp->d.vector.v.c[0] = x->d.scalar.c[0];
      tmp->d.vector.v.c[1] = x->d.scalar.c[1];
      break;
    case SILC_DOUBLE_COMPLEX:
      tmp->d.vector.v.z[0] = x->d.scalar.z[0];
      tmp->d.vector.v.z[1] = x->d.scalar.z[1];
      break;
    default:
      should_not_reach_here;
      return -1;
    }
    x = tmp;
  }
  param[2].o = _diag(x, k);
  if (tmp)
    silc_object_free(tmp);
  return 0;
}

#if 0 /* v1, v2 */

#if 0 /* v1 */
static silc_object_t *
_sparse(silc_object_t *v1, silc_object_t *v2, silc_object_t *v3,
	size_t m, size_t n)
{
  silc_object_t *new_object;
  silc_crs_t *new_in;
  int *work;
  ssize_t i;

  work = (int *)malloc(m * sizeof(int));
  if (!work) {
    fprintf(stderr, "sparse_crs_op: _sparse: out of memory\n");
    return NULL;
  }
  new_in = silc_crs_allocate(v3->precision, m, v3->d.vector.length);
  if (!new_in) {
    fprintf(stderr, "sparse_crs_op: _sparse: out of memory\n");
    free(work);
    return NULL;
  }
#pragma omp parallel
  {
    /* count the number of non-zero elements per row */
#define _count_nnz_per_row(pre) { \
      size_t j;					\
      work[i] = 0;				\
      for (j = 0; j < new_in->nnz; j++) {	\
	if (v1->d.vector.v.pre[j] - 1 == i)	\
	  work[i] ++;				\
      }						\
    }
    switch (v1->precision) {
    case SILC_INT:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) 
	_count_nnz_per_row(i);
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) 
	_count_nnz_per_row(l);
      break;
    default:
      should_not_reach_here;
      break;
    }
    /* initialize the `row' array with appropriate values */
#ifdef ENABLE_NUMA
    for (i = 0; i < (ssize_t)m; i++)
      new_in->row[i] = 0;
#endif
#pragma omp barrier
#pragma omp single
    {
      new_in->row[0] = 0;
      for (i = 0; i < (ssize_t)m; i++)
	new_in->row[i+1] = new_in->row[i] + work[i];
    }
    
    /* initialize the `index' and `v.pre' arrays */
#define _copy_1(pr1, pr2, pr3) { \
      size_t j, k;						\
      work[i] = new_in->row[i];					\
      for (j = 0; j < (ssize_t)new_in->nnz; j++) {		\
	if (v1->d.vector.v.pr1[j] - 1 == i) {			\
	  k = work[i];						\
	  new_in->index[k] = v2->d.vector.v.pr2[j] - 1;		\
	  new_in->v.pr3[k] = v3->d.vector.v.pr3[j];		\
	  work[i] = k + 1;					\
	}							\
      }								\
    }
#define _copy_2(pr1, pr2, pr3) { \
      size_t j, k;						\
      work[i] = new_in->row[i];					\
      for (j = 0; j < (ssize_t)new_in->nnz; j++) {		\
	if (v1->d.vector.v.pr1[j] - 1 == i) {			\
	  k = work[i];						\
	  new_in->index[k]     = v2->d.vector.v.pr2[j] - 1;	\
	  new_in->v.pr3[k*2]   = v3->d.vector.v.pr3[j*2];	\
	  new_in->v.pr3[k*2+1] = v3->d.vector.v.pr3[j*2+1];	\
	  work[i] = k + 1;					\
	}							\
      }								\
    }
    switch (v1->precision) {
    case SILC_INT:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, i, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, i, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, i, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, i, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(i, i, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(i, i, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, l, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, l, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, l, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(i, l, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(i, l, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(i, l, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    case SILC_LONG:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, i, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, i, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, i, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, i, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(l, i, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(l, i, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, l, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, l, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, l, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_1(l, l, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(l, l, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)m; i++) _copy_2(l, l, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    default:
      should_not_reach_here;
      break;
    }
  }
  free(work);
  new_object = silc_object_new_matrix(v3->precision,
				      silc_find_format_id(SILC_FORMAT_CRS),
				      m, n);
  new_object->d.matrix.in = (void *)new_in;
  silc_crs_sort(new_object);
  return new_object;
}
#else /* v2 */
static silc_object_t *
_sparse(silc_object_t *v1, silc_object_t *v2, silc_object_t *v3,
	size_t m, size_t n)
{
  silc_object_t *new_object;
  silc_crs_t *new_in;
  int *work, num_threads;
  ssize_t i;

#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  work = (int *)malloc(sizeof(int) * m * num_threads);
  if (!work) {
    fprintf(stderr, "sparse_crs_op: _sparse: out of memory\n");
    return NULL;
  }
  new_in = silc_crs_allocate(v3->precision, m, v3->d.vector.length);
  if (!new_in) {
    fprintf(stderr, "sparse_crs_op: _sparse: out of memory\n");
    free(work);
    return NULL;
  }
#pragma omp parallel
  {
    size_t j, k, offset, nnz, nnz_row;

    /* count the number of non-zero elements per row */
#ifdef _OPENMP
    offset = m * omp_get_thread_num();
#else
    offset = 0;
#endif
    for (i = 0; i < m; i++)
      work[offset + i] = 0;
    switch (v1->precision) {
    case SILC_INT:
#pragma omp for
      for (i = 0; i < (ssize_t)new_in->nnz; i++)
	work[offset + v1->d.vector.v.i[i] - 1] ++;
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)new_in->nnz; i++)
	work[offset + v1->d.vector.v.l[i] - 1] ++;
      break;
    default:
      should_not_reach_here;
      break;
    }

    /* initialize the `row' array with appropriate values */
#pragma omp barrier
#pragma omp for
    for (i = 0; i < (ssize_t)m; i++) {
      nnz_row = 0;
      for (j = k = 0; j < num_threads; j++, k+=m) {
	nnz = work[k + i];
	work[k + i] = nnz_row;
	nnz_row += nnz;
      }
      new_in->row[i] = nnz_row;
    }
#pragma omp barrier
#pragma omp single
    {
      nnz = 0;
      for (i = 0; i < (ssize_t)m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[m] = nnz;
    }

    /* initialize the `index' and `v.pre' arrays */
#ifdef ENABLE_NUMA
#define _sparse_init_1(pre) { \
      for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
	new_in->v.pre[j] = 0;					\
	new_in->index[j] = 0;					\
      }								\
    }
#define _sparse_init_2(pre) { \
      for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
	new_in->v.pre[j*2] = new_in->v.pre[j*2+1] = 0;		\
	new_in->index[j] = 0;					\
      }								\
    }
    switch (v3->precision) {
    case SILC_INT:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(i); break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(l); break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(s); break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(d); break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_2(c); break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_2(z); break;
    default:
      should_not_reach_here; break;
    }
#pragma omp barrier
#endif
#define _copy_1(pr1, pr2, pr3) { \
      j = v1->d.vector.v.pr1[i] - 1;				\
      k = new_in->row[j] + work[offset + j];			\
      new_in->index[k] = v2->d.vector.v.pr2[i] - 1;		\
      new_in->v.pr3[k] = v3->d.vector.v.pr3[i];			\
      work[offset + j] ++;					\
    }
#define _copy_2(pr1, pr2, pr3) { \
      j = v1->d.vector.v.pr1[i] - 1;				\
      k = new_in->row[j] + work[offset + j];			\
      new_in->index[k]     = v2->d.vector.v.pr2[i] - 1;		\
      new_in->v.pr3[k*2]   = v3->d.vector.v.pr3[i*2];		\
      new_in->v.pr3[k*2+1] = v3->d.vector.v.pr3[i*2+1];		\
      work[offset + j] ++;					\
    }
    switch (v1->precision) {
    case SILC_INT:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, i, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, i, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, i, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, i, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(i, i, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(i, i, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, l, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, l, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, l, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(i, l, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(i, l, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(i, l, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    case SILC_LONG:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, i, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, i, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, i, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, i, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(l, i, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(l, i, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, l, i); break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, l, l); break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, l, s); break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_1(l, l, d); break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(l, l, c); break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)new_in->nnz; i++) _copy_2(l, l, z); break;
	default:
	  should_not_reach_here; break;
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    default:
      should_not_reach_here;
      break;
    }
  }
  free(work);
  new_object = silc_object_new_matrix(v3->precision,
				      silc_find_format_id(SILC_FORMAT_CRS),
				      m, n);
  new_object->d.matrix.in = (void *)new_in;
  silc_crs_sort(new_object);
  return new_object;
}
#endif

int
_filter_zeros(silc_object_t *v1, silc_object_t *v2, silc_object_t *v3,
	      silc_object_t *t1, silc_object_t *t2, silc_object_t *t3)
{
  silc_object_t *u1, *u2, *u3;
  int num_threads, *nnz;
  ssize_t i;
  size_t length, ii;

  if (v3->type == SILC_SCALAR_TYPE) {
    switch (v3->precision) {
    case SILC_INT:
      if (v3->d.scalar.i == 0) {
	length = 0;
	goto end;
      }
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t3->d.vector.v.i[i] = v3->d.scalar.i;
      break;
    case SILC_LONG:
      if (v3->d.scalar.l == 0) {
	length = 0;
	goto end;
      }
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t3->d.vector.v.l[i] = v3->d.scalar.l;
      break;
    case SILC_FLOAT:
      if (v3->d.scalar.s == 0) {
	length = 0;
	goto end;
      }
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t3->d.vector.v.s[i] = v3->d.scalar.s;
      break;
    case SILC_DOUBLE:
      if (v3->d.scalar.d == 0) {
	length = 0;
	goto end;
      }
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t3->d.vector.v.d[i] = v3->d.scalar.d;
      break;
    case SILC_COMPLEX:
      if (v3->d.scalar.c[0] == 0 && v3->d.scalar.c[1] == 0) {
	length = 0;
	goto end;
      }
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++) {
	ii = i * 2;
	t3->d.vector.v.c[ii]   = v3->d.scalar.c[0];
	t3->d.vector.v.c[ii+1] = v3->d.scalar.c[1];
      }
      break;
    case SILC_DOUBLE_COMPLEX:
      if (v3->d.scalar.z[0] == 0 && v3->d.scalar.z[1] == 0) {
	length = 0;
	goto end;
      }
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++) {
	ii = i * 2;
	t3->d.vector.v.z[ii]   = v3->d.scalar.z[0];
	t3->d.vector.v.z[ii+1] = v3->d.scalar.z[1];
      }
      break;
    default:
      should_not_reach_here;
    }
    switch (v1->precision) {
    case SILC_INT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t1->d.vector.v.i[i] = v1->d.vector.v.i[i];
      break;
    case SILC_LONG:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t1->d.vector.v.l[i] = v1->d.vector.v.l[i];
      break;
    case SILC_FLOAT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t1->d.vector.v.i[i] = v1->d.vector.v.s[i];
      break;
    case SILC_DOUBLE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	t1->d.vector.v.l[i] = v1->d.vector.v.d[i];
      break;
    default:
      should_not_reach_here;
    }
    switch (v2->precision) {
    case SILC_INT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v2->d.vector.length; i++)
	t2->d.vector.v.i[i] = v2->d.vector.v.i[i];
      break;
    case SILC_LONG:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v2->d.vector.length; i++)
	t2->d.vector.v.l[i] = v2->d.vector.v.l[i];
      break;
    case SILC_FLOAT:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v2->d.vector.length; i++)
	t2->d.vector.v.i[i] = v2->d.vector.v.s[i];
      break;
    case SILC_DOUBLE:
#pragma omp parallel for
      for (i = 0; i < (ssize_t)v2->d.vector.length; i++)
	t2->d.vector.v.l[i] = v2->d.vector.v.d[i];
      break;
    default:
      should_not_reach_here;
    }
    return 0;
  }
  u1 = silc_object_duplicate(t1, 0);
  if (!u1) {
    fprintf(stderr, "_filter_zeros: out of memory (u1)\n");
    return -1;
  }
  u2 = silc_object_duplicate(t2, 0);
  if (!u2) {
    fprintf(stderr, "_filter_zeros: out of memory (u2)\n");
    silc_object_free(u1);
    return -1;
  }
  u3 = silc_object_duplicate(t3, 0);
  if (!u3) {
    fprintf(stderr, "_filter_zeros: out of memory (u3)\n");
    silc_object_free(u1);
    silc_object_free(u2);
    return -1;
  }
#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  nnz = (int *)malloc(sizeof(int) * num_threads);
  if (!nnz) {
    fprintf(stderr, "_filter_zeros: out of memory (nnz)\n");
    silc_object_free(u1);
    silc_object_free(u2);
    silc_object_free(u3);
    return -1;
  }
#pragma omp parallel
  {
    int start, k, n, offset;

    start = -1;
#define _filter_1(pre1s, pre1d, type1, pre2s, pre2d, type2, pre3) {	\
      if (start < 0)							\
        start = k = i;							\
      if (v3->d.vector.v.pre3[i] != 0) {				\
	u1->d.vector.v.pre1d[k] = (type1)v1->d.vector.v.pre1s[i];	\
	u2->d.vector.v.pre2d[k] = (type2)v2->d.vector.v.pre2s[i];	\
	u3->d.vector.v.pre3[k] = v3->d.vector.v.pre3[i];		\
	k++;								\
      }									\
    }
#define _filter_2(pre1s, pre1d, type1, pre2s, pre2d, type2, pre3) {	\
      int ii, kk;							\
      if (start < 0)							\
        start = k = i;							\
      ii = i * 2;							\
      if (v3->d.vector.v.pre3[ii] != 0 && v3->d.vector.v.pre3[ii+1] != 0) { \
	kk = k * 2;							\
	u1->d.vector.v.pre1d[k] = (type1)v1->d.vector.v.pre1s[i];	\
	u2->d.vector.v.pre2d[k] = (type2)v2->d.vector.v.pre2s[i];	\
	u3->d.vector.v.pre3[kk]   = v3->d.vector.v.pre3[ii];		\
	u3->d.vector.v.pre3[kk+1] = v3->d.vector.v.pre3[ii+1];		\
	k++;								\
      }									\
    }
    switch (v1->precision) { /* nested switches with 4 x 4 x 6 cases */
    case SILC_INT:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, i, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, i, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, i, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, i, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, i, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, i, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, l, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, l, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, l, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, l, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, l, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, l, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_FLOAT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, s, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, s, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, s, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, s, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, s, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, s, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_DOUBLE:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, d, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, d, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, d, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(i, i, int32_t, d, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, d, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(i, i, int32_t, d, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_LONG:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, i, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, i, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, i, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, i, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, i, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, i, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, l, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, l, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, l, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, l, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, l, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, l, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_FLOAT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, s, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, s, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, s, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, s, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, s, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, s, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_DOUBLE:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, d, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, d, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, d, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(l, l, int64_t, d, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, d, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(l, l, int64_t, d, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_FLOAT:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, i, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, i, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, i, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, i, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, i, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, i, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, l, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, l, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, l, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, l, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, l, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, l, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_FLOAT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, s, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, s, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, s, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, s, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, s, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, s, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_DOUBLE:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, d, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, d, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, d, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(s, i, int32_t, d, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, d, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(s, i, int32_t, d, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      default:
	should_not_reach_here;
      }
      break;
    case SILC_DOUBLE:
      switch (v2->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, i, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, i, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, i, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, i, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, i, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, i, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, l, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, l, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, l, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, l, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, l, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, l, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_FLOAT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, s, i, int32_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, s, i, int32_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, s, i, int32_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, s, i, int32_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, s, i, int32_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, s, i, int32_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      case SILC_DOUBLE:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, d, l, int64_t, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, d, l, int64_t, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, d, l, int64_t, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_1(d, l, int64_t, d, l, int64_t, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, d, l, int64_t, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v3->d.vector.length; i++)
	    _filter_2(d, l, int64_t, d, l, int64_t, z);
	  break;
	default:
	  should_not_reach_here;
	}
	break;
      default:
	should_not_reach_here;
      }
      break;
    default:
      should_not_reach_here;
    }
#ifdef _OPENMP
    n = omp_get_thread_num();
#else
    n = 0;
#endif
    if (start < 0)
      nnz[n] = 0;
    else
      nnz[n] = k - start;
#pragma omp barrier
    offset = 0;
    for (i = 0; i < n; i++)
      offset += nnz[i];
    switch (t1->precision) {
    case SILC_INT:
      memcpy(t1->d.vector.v.i + offset, u1->d.vector.v.i + start,
	     sizeof(int32_t) * nnz[n]);
      break;
    case SILC_LONG:
      memcpy(t1->d.vector.v.l + offset, u1->d.vector.v.l + start,
	     sizeof(int64_t) * nnz[n]);
      break;
    default:
      should_not_reach_here;
    }
    switch (t2->precision) {
    case SILC_INT:
      memcpy(t2->d.vector.v.i + offset, u2->d.vector.v.i + start,
	     sizeof(int32_t) * nnz[n]);
      break;
    case SILC_LONG:
      memcpy(t2->d.vector.v.l + offset, u2->d.vector.v.l + start,
	     sizeof(int64_t) * nnz[n]);
      break;
    default:
      should_not_reach_here;
    }
    switch (t3->precision) {
    case SILC_INT:
      memcpy(t3->d.vector.v.i + offset, u3->d.vector.v.i + start,
	     sizeof(int32_t) * nnz[n]);
      break;
    case SILC_LONG:
      memcpy(t3->d.vector.v.l + offset, u3->d.vector.v.l + start,
	     sizeof(int64_t) * nnz[n]);
      break;
    case SILC_FLOAT:
      memcpy(t3->d.vector.v.s + offset, u3->d.vector.v.s + start,
	     sizeof(float) * nnz[n]);
      break;
    case SILC_DOUBLE:
      memcpy(t3->d.vector.v.d + offset, u3->d.vector.v.d + start,
	     sizeof(double) * nnz[n]);
      break;
    case SILC_COMPLEX:
      memcpy(t3->d.vector.v.c + offset * 2, u3->d.vector.v.c + start * 2,
	     sizeof(float) * 2 * nnz[n]);
      break;
    case SILC_DOUBLE_COMPLEX:
      memcpy(t3->d.vector.v.z + offset * 2, u3->d.vector.v.z + start * 2,
	     sizeof(double) * 2 * nnz[n]);
      break;
    default:
      should_not_reach_here;
    }
  }
  silc_object_free(u1);
  silc_object_free(u2);
  silc_object_free(u3);
  length = 0;
  for (i = 0; i < num_threads; i++)
    length += nnz[i];
  free(nnz);
 end:
  t1->d.vector.length = length;
  t2->d.vector.length = length;
  t3->d.vector.length = length;
  return 0;
}

int
silc_crs_func_sparse(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *v1, *v2, *v3, *s1, *s2, *t1, *t2, *t3;
  ssize_t m, n;

  v1 = param[0].o;
  v2 = param[1].o;
  v3 = param[2].o;
  s1 = param[3].o;
  s2 = param[4].o;
#ifdef DEBUG
  printf("silc_crs_func_sparse(0x%lx, 0x%lx, 0x%lx, 0x%lx, 0x%lx)\n",
	 v1, v2, v3, s1, s2);
#endif
  if (v1->type != SILC_COLUMN_VECTOR_TYPE &&
      v1->type != SILC_ROW_VECTOR_TYPE)
    return -1;
  if (v2->type != SILC_COLUMN_VECTOR_TYPE &&
      v2->type != SILC_ROW_VECTOR_TYPE)
    return -1;
  if (v3->type != SILC_COLUMN_VECTOR_TYPE &&
      v3->type != SILC_ROW_VECTOR_TYPE &&
      v3->type != SILC_SCALAR_TYPE)
    return -1;
  if (v1->precision == SILC_COMPLEX || v1->precision == SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, "sparse: 1st argument cannot be complex vector\n");
    return -1;
  }
  if (v2->precision == SILC_COMPLEX || v2->precision == SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, "sparse: 2nd argument cannot be complex vector\n");
    return -1;
  }
  if (v1->d.vector.length != v2->d.vector.length) {
    fprintf(stderr, "sparse: vector size mismatch (1st and 2nd arguments)\n");
    return -1;
  }
  if (v3->type != SILC_SCALAR_TYPE &&
      v3->d.vector.length != v1->d.vector.length) {
    fprintf(stderr, "sparse: vector size mismatch (3rd argument)\n");
    return -1;
  }
  if (s1->type != SILC_SCALAR_TYPE)
    return -1;
  if (s2->type != SILC_SCALAR_TYPE)
    return -1;
  switch (s1->precision) {
  case SILC_INT:
    m = s1->d.scalar.i;
    break;
  case SILC_LONG:
    m = s1->d.scalar.l;
    break;
  case SILC_FLOAT:
    m = (ssize_t)s1->d.scalar.s;
    break;
  case SILC_DOUBLE:
    m = (ssize_t)s1->d.scalar.d;
    break;
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "sparse: 4th argument must be integer or real\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (m < 1) {
    fprintf(stderr, "sparse: 4th argument must be positive\n");
    return -1;
  }
  switch (s2->precision) {
  case SILC_INT:
    n = s2->d.scalar.i;
    break;
  case SILC_LONG:
    n = s2->d.scalar.l;
    break;
  case SILC_FLOAT:
    n = (ssize_t)s2->d.scalar.s;
    break;
  case SILC_DOUBLE:
    n = (ssize_t)s2->d.scalar.d;
    break;
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "sparse: 5th argument must be integer or real\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "sparse: 5th argument must be positive\n");
    return -1;
  }
  switch (v1->precision) {
  case SILC_INT:
  case SILC_LONG:
    t1 = silc_object_duplicate(v1, 0);
    break;
  case SILC_FLOAT:
    t1 = silc_object_new_column_vector(SILC_INT, v1->d.vector.length);
    break;
  case SILC_DOUBLE:
    t1 = silc_object_new_column_vector(SILC_LONG, v1->d.vector.length);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  if (!t1) {
    fprintf(stderr, "sparse: out of memory (t1)\n");
    return -1;
  }
  switch (v2->precision) {
  case SILC_INT:
  case SILC_LONG:
    t2 = silc_object_duplicate(v2, 0);
    break;
  case SILC_FLOAT:
    t2 = silc_object_new_column_vector(SILC_INT, v2->d.vector.length);
    break;
  case SILC_DOUBLE:
    t2 = silc_object_new_column_vector(SILC_LONG, v2->d.vector.length);
    break;
  default:
    should_not_reach_here;
    return -1;
  }
  if (!t2) {
    fprintf(stderr, "sparse: out of memory (t2)\n");
    silc_object_free(t1);
    return -1;
  }
  if (v3->type == SILC_SCALAR_TYPE)
    t3 = silc_object_new_column_vector(v3->precision, v1->d.vector.length);
  else
    t3 = silc_object_duplicate(v3, 0);
  if (!t3) {
    fprintf(stderr, "sparse: out of memory (t3)\n");
    silc_object_free(t1);
    silc_object_free(t2);
    return -1;
  }
  if (_filter_zeros(v1, v2, v3, t1, t2, t3) < 0) {
    silc_object_free(t1);
    silc_object_free(t2);
    silc_object_free(t3);
    return -1;
  }
  param[5].o = _sparse(t1, t2, t3, m, n);
  silc_object_free(t1);
  silc_object_free(t2);
  silc_object_free(t3);
  return (param[5].o) ? 0 : -1;
}

#else /* v3 */

static silc_object_t *
_sparse(silc_object_t *v1, silc_object_t *v2, silc_object_t *v3,
	size_t m, size_t n)
{
  silc_object_t *new_object;
  silc_crs_t *new_in;
  int *work, num_threads;
  ssize_t i;

#ifdef _OPENMP
  num_threads = omp_get_max_threads();
#else
  num_threads = 1;
#endif
  work = (int *)malloc(sizeof(int) * m * num_threads);
  if (!work) {
    fprintf(stderr, "sparse_crs_op: _sparse: out of memory\n");
    return NULL;
  }
  new_in = silc_crs_allocate(v3->precision, m, v1->d.vector.length);
  if (!new_in) {
    fprintf(stderr, "sparse_crs_op: _sparse: out of memory\n");
    free(work);
    return NULL;
  }
#pragma omp parallel
  {
    size_t j, k, ii, kk, offset, nnz, nnz_row;

    /* count the number of non-zero elements per row */
#ifdef _OPENMP
    offset = m * omp_get_thread_num();
#else
    offset = 0;
#endif
    for (i = 0; i < m; i++)
      work[offset + i] = 0;
#define count_nnz_row_s(pre) {						\
      work[offset + (int)v1->d.vector.v.pre[i] - 1] ++;			\
    }
#define count_nnz_row_v1(pr1, pr3) {					\
      if (v3->d.vector.v.pr3[i] != 0)					\
	work[offset + (int)v1->d.vector.v.pr1[i] - 1] ++;		\
    }
#define count_nnz_row_v2(pr1, pr3) {					\
      size_t ii = i * 2;						\
      if (v3->d.vector.v.pr3[ii] != 0 && v3->d.vector.v.pr3[ii+1] != 0) \
	work[offset + (int)v1->d.vector.v.pr1[i] - 1] ++;		\
    }
    if (v3->type == SILC_SCALAR_TYPE) {
      switch (v3->precision) {
      case SILC_INT:
	if (v3->d.scalar.i == 0) goto skip_count;
	break;
      case SILC_LONG:
	if (v3->d.scalar.l == 0) goto skip_count;
	break;
      case SILC_FLOAT:
	if (v3->d.scalar.s == 0) goto skip_count;
	break;
      case SILC_DOUBLE:
	if (v3->d.scalar.d == 0) goto skip_count;
	break;
      case SILC_COMPLEX:
	if (v3->d.scalar.c[0] == 0 && v3->d.scalar.c[1] == 0) goto skip_count;
	break;
      case SILC_DOUBLE_COMPLEX:
	if (v3->d.scalar.z[0] == 0 && v3->d.scalar.z[1] == 0) goto skip_count;
	break;
      default:
	should_not_reach_here;
      }
      switch (v1->precision) {
      case SILC_INT:
#pragma omp for
	for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	  count_nnz_row_s(i);
	break;
      case SILC_LONG:
#pragma omp for
	for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	  count_nnz_row_s(l);
	break;
      case SILC_FLOAT:
#pragma omp for
	for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	  count_nnz_row_s(s);
	break;
      case SILC_DOUBLE:
#pragma omp for
	for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	  count_nnz_row_s(d);
	break;
      default:
	should_not_reach_here;
      }
    skip_count:
      ;
    } else { /* SILC_COLUMN_VECTOR_TYPE, SILC_ROW_VECTOR_TYPE */
      switch (v1->precision) {
      case SILC_INT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(i, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(i, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(i, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(i, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(i, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(i, z);
	  break;
	default:
	  should_not_reach_here;
	  break;
	}
	break;
      case SILC_LONG:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(l, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(l, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(l, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(l, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(l, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(l, z);
	  break;
	default:
	  should_not_reach_here;
	  break;
	}
	break;
      case SILC_FLOAT:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(s, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(s, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(s, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(s, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(s, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(s, z);
	  break;
	default:
	  should_not_reach_here;
	  break;
	}
	break;
      case SILC_DOUBLE:
	switch (v3->precision) {
	case SILC_INT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(d, i);
	  break;
	case SILC_LONG:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(d, l);
	  break;
	case SILC_FLOAT:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(d, s);
	  break;
	case SILC_DOUBLE:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v1(d, d);
	  break;
	case SILC_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(d, c);
	  break;
	case SILC_DOUBLE_COMPLEX:
#pragma omp for
	  for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	    count_nnz_row_v2(d, z);
	  break;
	default:
	  should_not_reach_here;
	  break;
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
    }

    /* initialize the `row' array with appropriate values */
#pragma omp barrier
#pragma omp for
    for (i = 0; i < (ssize_t)m; i++) {
      nnz_row = 0;
      for (j = k = 0; j < num_threads; j++, k+=m) {
	nnz = work[k + i];
	work[k + i] = nnz_row;
	nnz_row += nnz;
      }
      new_in->row[i] = nnz_row;
    }
#pragma omp barrier
#pragma omp single
    {
      nnz = 0;
      for (i = 0; i < (ssize_t)m; i++) {
	nnz_row = new_in->row[i];
	new_in->row[i] = nnz;
	nnz += nnz_row;
      }
      new_in->row[m] = new_in->nnz = nnz;
    }
    if (new_in->nnz == 0)
      goto skip_copy;

    /* initialize the `index' and `v.pre' arrays */
#ifdef ENABLE_NUMA
#define _sparse_init_1(pre) { \
      for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
	new_in->v.pre[j] = 0;					\
	new_in->index[j] = 0;					\
      }								\
    }
#define _sparse_init_2(pre) { \
      for (j = new_in->row[i]; j < new_in->row[i+1]; j++) {	\
	new_in->v.pre[j*2] = new_in->v.pre[j*2+1] = 0;		\
	new_in->index[j] = 0;					\
      }								\
    }
    switch (v3->precision) {
    case SILC_INT:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(i); break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(l); break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(s); break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_1(d); break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_2(c); break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)m; i++) _sparse_init_2(z); break;
    default:
      should_not_reach_here; break;
    }
#pragma omp barrier
#endif
#define _copy_s1(pr1, pr2, pr3) {				\
      j = (int)v1->d.vector.v.pr1[i] - 1;			\
      k = new_in->row[j] + work[offset + j];			\
      new_in->index[k] = (int)v2->d.vector.v.pr2[i] - 1;	\
      new_in->v.pr3[k] = v3->d.scalar.pr3;			\
      work[offset + j] ++;					\
    }
#define _copy_s2(pr1, pr2, pr3) {				\
      j = (int)v1->d.vector.v.pr1[i] - 1;			\
      k = new_in->row[j] + work[offset + j]; kk = k * 2;	\
      new_in->index[k]    = (int)v2->d.vector.v.pr2[i] - 1;	\
      new_in->v.pr3[kk]   = v3->d.scalar.pr3[0];		\
      new_in->v.pr3[kk+1] = v3->d.scalar.pr3[1];		\
      work[offset + j] ++;					\
    }
#define _copy_v1(pr1, pr2, pr3) {				\
      if (v3->d.vector.v.pr3[i] != 0) {				\
	j = (int)v1->d.vector.v.pr1[i] - 1;			\
	k = new_in->row[j] + work[offset + j];			\
	new_in->index[k] = (int)v2->d.vector.v.pr2[i] - 1;	\
	new_in->v.pr3[k] = v3->d.vector.v.pr3[i];		\
	work[offset + j] ++;					\
      }								\
    }
#define _copy_v2(pr1, pr2, pr3) {				\
      ii = i * 2;						\
      if (v3->d.vector.v.pr3[ii]   != 0 &&			\
	  v3->d.vector.v.pr3[ii+1] != 0) {			\
	j = (int)v1->d.vector.v.pr1[i] - 1;			\
	k = new_in->row[j] + work[offset + j]; kk = k * 2;	\
	new_in->index[k]    = (int)v2->d.vector.v.pr2[i] - 1;	\
	new_in->v.pr3[kk]   = v3->d.vector.v.pr3[ii];		\
	new_in->v.pr3[kk+1] = v3->d.vector.v.pr3[ii+1];		\
	work[offset + j] ++;					\
      }								\
    }

    switch (v1->precision) {
    case SILC_INT:
      switch (v2->precision) {
      case SILC_INT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_LONG:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_FLOAT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_DOUBLE:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(i, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(i, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(i, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(i, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    case SILC_LONG:
      switch (v2->precision) {
      case SILC_INT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_LONG:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_FLOAT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_DOUBLE:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(l, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(l, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(l, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(l, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    case SILC_FLOAT:
      switch (v2->precision) {
      case SILC_INT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_LONG:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_FLOAT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_DOUBLE:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(s, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(s, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(s, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(s, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    case SILC_DOUBLE:
      switch (v2->precision) {
      case SILC_INT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, i, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, i, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, i, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, i, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, i, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, i, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_LONG:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, l, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, l, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, l, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, l, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, l, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, l, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_FLOAT:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, s, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, s, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, s, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, s, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, s, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, s, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      case SILC_DOUBLE:
	if (v3->type == SILC_SCALAR_TYPE) {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s1(d, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_s2(d, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	} else {
	  switch (v3->precision) {
	  case SILC_INT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, d, i);
	    break;
	  case SILC_LONG:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, d, l);
	    break;
	  case SILC_FLOAT:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, d, s);
	    break;
	  case SILC_DOUBLE:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v1(d, d, d);
	    break;
	  case SILC_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, d, c);
	    break;
	  case SILC_DOUBLE_COMPLEX:
#pragma omp for
	    for (i = 0; i < (ssize_t)v1->d.vector.length; i++)
	      _copy_v2(d, d, z);
	    break;
	  default:
	    should_not_reach_here; break;
	  }
	}
	break;
      default:
	should_not_reach_here;
	break;
      }
      break;
    default:
      should_not_reach_here;
      break;
    }
  skip_copy:
    ;
  }
  free(work);
  if (new_in->nnz == 0)
    new_in->nnz = 1;  /* avoid doing realloc(*, 0) */
  new_in->v.i = (int32_t *)
    realloc(new_in->v.i, SILC_SCALAR_SIZE(v3) * new_in->nnz);
  if (!new_in->v.i) {
    free(new_in->index);
    free(new_in->row);
    free(new_in);
    return NULL;
  }
  new_in->index = (int *)realloc(new_in->index, sizeof(int) * new_in->nnz);
  if (!new_in->index) {
    free(new_in->v.i);
    free(new_in->row);
    free(new_in);
    return NULL;
  }
  new_object = silc_object_new_matrix(v3->precision,
				      silc_find_format_id(SILC_FORMAT_CRS),
				      m, n);
  new_object->d.matrix.in = (void *)new_in;
  silc_crs_sort(new_object);
  return new_object;
}

int
silc_crs_func_sparse(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *v1, *v2, *v3, *s1, *s2, *t1, *t2, *t3;
  ssize_t m, n;

  v1 = param[0].o;
  v2 = param[1].o;
  v3 = param[2].o;
  s1 = param[3].o;
  s2 = param[4].o;
#ifdef DEBUG
  printf("silc_crs_func_sparse(0x%lx, 0x%lx, 0x%lx, 0x%lx, 0x%lx)\n",
	 v1, v2, v3, s1, s2);
#endif
  if (v1->type != SILC_COLUMN_VECTOR_TYPE &&
      v1->type != SILC_ROW_VECTOR_TYPE)
    return -1;
  if (v2->type != SILC_COLUMN_VECTOR_TYPE &&
      v2->type != SILC_ROW_VECTOR_TYPE)
    return -1;
  if (v3->type != SILC_COLUMN_VECTOR_TYPE &&
      v3->type != SILC_ROW_VECTOR_TYPE &&
      v3->type != SILC_SCALAR_TYPE)
    return -1;
  if (v1->precision == SILC_COMPLEX || v1->precision == SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, "sparse: 1st argument cannot be complex vector\n");
    return -1;
  }
  if (v2->precision == SILC_COMPLEX || v2->precision == SILC_DOUBLE_COMPLEX) {
    fprintf(stderr, "sparse: 2nd argument cannot be complex vector\n");
    return -1;
  }
  if (v1->d.vector.length != v2->d.vector.length) {
    fprintf(stderr, "sparse: vector size mismatch (1st and 2nd arguments)\n");
    return -1;
  }
  if (v3->type != SILC_SCALAR_TYPE &&
      v3->d.vector.length != v1->d.vector.length) {
    fprintf(stderr, "sparse: vector size mismatch (3rd argument)\n");
    return -1;
  }
  if (s1->type != SILC_SCALAR_TYPE)
    return -1;
  if (s2->type != SILC_SCALAR_TYPE)
    return -1;
  switch (s1->precision) {
  case SILC_INT:
    m = s1->d.scalar.i;
    break;
  case SILC_LONG:
    m = s1->d.scalar.l;
    break;
  case SILC_FLOAT:
    m = (ssize_t)s1->d.scalar.s;
    break;
  case SILC_DOUBLE:
    m = (ssize_t)s1->d.scalar.d;
    break;
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "sparse: 4th argument must be integer or real\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (m < 1) {
    fprintf(stderr, "sparse: 4th argument must be positive\n");
    return -1;
  }
  switch (s2->precision) {
  case SILC_INT:
    n = s2->d.scalar.i;
    break;
  case SILC_LONG:
    n = s2->d.scalar.l;
    break;
  case SILC_FLOAT:
    n = (ssize_t)s2->d.scalar.s;
    break;
  case SILC_DOUBLE:
    n = (ssize_t)s2->d.scalar.d;
    break;
  case SILC_COMPLEX:
  case SILC_DOUBLE_COMPLEX:
    fprintf(stderr, "sparse: 5th argument must be integer or real\n");
    return -1;
  default:
    should_not_reach_here;
    return -1;
  }
  if (n < 1) {
    fprintf(stderr, "sparse: 5th argument must be positive\n");
    return -1;
  }
  param[5].o = _sparse(v1, v2, v3, m, n);
  return (param[5].o) ? 0 : -1;
}

#endif /* v3 */

int
silc_crs_func_diagvec(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_crs_t *in;
  ssize_t i, n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_diagvec(0x%lx)\n", object);
#endif
  in = SILC_CRS(object);
  if (object->d.matrix.m < object->d.matrix.n)
    n = object->d.matrix.m;
  else
    n = object->d.matrix.n;
  new_object = silc_object_new_column_vector(object->precision, n);

#define _diagvec_1(pre) {					\
    int j;							\
    new_object->d.vector.v.pre[i] = 0;				\
    for (j = in->row[i]; j < in->row[i+1]; j++) {		\
      if (in->index[j] == i) {					\
        new_object->d.vector.v.pre[i] = in->v.pre[j];		\
        break;							\
      }								\
    }								\
  }
#define _diagvec_2(pre) {					\
    int j;							\
    new_object->d.vector.v.pre[i*2]   = 0;			\
    new_object->d.vector.v.pre[i*2+1] = 0;			\
    for (j = in->row[i]; j < in->row[i+1]; j++) {		\
      if (in->index[j] == i) {					\
        new_object->d.vector.v.pre[i*2]   = in->v.pre[j*2];	\
        new_object->d.vector.v.pre[i*2+1] = in->v.pre[j*2+1];	\
        break;							\
      }								\
    }								\
  }
  switch (object->precision) {
  case SILC_INT:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diagvec_1(i);
    break;
  case SILC_LONG:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diagvec_1(l);
    break;
  case SILC_FLOAT:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diagvec_1(s);
    break;
  case SILC_DOUBLE:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diagvec_1(d);
    break;
  case SILC_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diagvec_2(c);
    break;
  case SILC_DOUBLE_COMPLEX:
#pragma omp parallel for
    for (i = 0; i < n; i++)
      _diagvec_2(z);
    break;
  default:
    should_not_reach_here;
  }
  param[1].o = new_object;
  return 0;
}

int
silc_crs_func_nnz(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object, *new_object;
  silc_crs_t *in, *new_in;
  const char *format_name;
  ssize_t i, m, n;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_func_nnz(0x%lx)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  format_name = silc_find_format_name(object->format);
  if (strcmp(format_name, SILC_FORMAT_CRS)) {
    /* XXX need a workaround */
    fprintf(stderr, "sparse_crs_op: nnz: not in the CRS format\n");
    return -1;
  }
  in = SILC_CRS(object);
  param[1].o = silc_object_new_int(in->nnz);
  return (param[1].o) ? 0 : -1;
}

/* procedures */

int
silc_crs_proc_split(int n_params, silc_mfunc_param_t *param)
{
  silc_object_t *object;
  silc_crs_t *in, *L_in, *D_in, *U_in;
  size_t L_nnz, D_nnz, U_nnz;
  const char *format_name;
  ssize_t i;

  object = param[0].o;
#ifdef DEBUG
  printf("silc_crs_proc_split(0x%lx, L, D, U)\n", object);
#endif
  if (object->type != SILC_MATRIX_TYPE)
    return -1;
  format_name = silc_find_format_name(object->format);
  if (strcmp(format_name, SILC_FORMAT_CRS)) {
    /* XXX need a workaround */
    fprintf(stderr, "sparse_crs_op: split: not in the CRS format\n");
    return -1;
  }
  if (object->d.matrix.m != object->d.matrix.n) {
    fprintf(stderr, "sparse_crs_op: split: not a square matrix\n");
    return -1;
  }

  in = SILC_CRS(object);

  L_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  L_in->row = (int *)malloc(sizeof(int) * (object->d.matrix.m + 1));

  D_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  D_in->row = (int *)malloc(sizeof(int) * (object->d.matrix.m + 1));

  U_in = (silc_crs_t *)malloc(sizeof(silc_crs_t));
  U_in->row = (int *)malloc(sizeof(int) * (object->d.matrix.m + 1));

#pragma omp parallel
  {
    /* count non-zero elements of each row */
#pragma omp for
    for (i = 0; i < (ssize_t)object->d.matrix.m; i++) {
      size_t j;

      L_in->row[i] = 0;
      D_in->row[i] = 0;
      U_in->row[i] = 0;
      for (j = in->row[i]; j < in->row[i+1]; j++) {
	if (i > in->index[j])
	  L_in->row[i]++;
	else if (i == in->index[j])
	  D_in->row[i]++;
	else
	  U_in->row[i]++;
      }
    }
#pragma omp barrier
#pragma omp single
    {
      size_t j;

      /* sum up the numbers of non-zero elements per row */
      L_in->nnz = 0;
      D_in->nnz = 0;
      U_in->nnz = 0;
      for (i = 0; i < object->d.matrix.m; i++) {
	j = L_in->row[i]; L_in->row[i] = L_in->nnz; L_in->nnz += j;
	j = D_in->row[i]; D_in->row[i] = D_in->nnz; D_in->nnz += j;
	j = U_in->row[i]; U_in->row[i] = U_in->nnz; U_in->nnz += j;
      }
      L_in->row[object->d.matrix.m] = L_in->nnz;
      D_in->row[object->d.matrix.m] = D_in->nnz;
      U_in->row[object->d.matrix.m] = U_in->nnz;
      switch (object->precision) {
      case SILC_INT:
	L_in->v.i = (int32_t *)malloc(sizeof(int32_t) * L_in->nnz);
	D_in->v.i = (int32_t *)malloc(sizeof(int32_t) * D_in->nnz);
	U_in->v.i = (int32_t *)malloc(sizeof(int32_t) * U_in->nnz);
	break;
      case SILC_LONG:
	L_in->v.l = (int64_t *)malloc(sizeof(int64_t) * L_in->nnz);
	D_in->v.l = (int64_t *)malloc(sizeof(int64_t) * D_in->nnz);
	U_in->v.l = (int64_t *)malloc(sizeof(int64_t) * U_in->nnz);
	break;
      case SILC_FLOAT:
	L_in->v.s = (float *)malloc(sizeof(float) * L_in->nnz);
	D_in->v.s = (float *)malloc(sizeof(float) * D_in->nnz);
	U_in->v.s = (float *)malloc(sizeof(float) * U_in->nnz);
	break;
      case SILC_DOUBLE:
	L_in->v.d = (double *)malloc(sizeof(double) * L_in->nnz);
	D_in->v.d = (double *)malloc(sizeof(double) * D_in->nnz);
	U_in->v.d = (double *)malloc(sizeof(double) * U_in->nnz);
	break;
      case SILC_COMPLEX:
	L_in->v.c = (float *)malloc(sizeof(float) * 2 * L_in->nnz);
	D_in->v.c = (float *)malloc(sizeof(float) * 2 * D_in->nnz);
	U_in->v.c = (float *)malloc(sizeof(float) * 2 * U_in->nnz);
	break;
      case SILC_DOUBLE_COMPLEX:
	L_in->v.z = (double *)malloc(sizeof(double) * 2 * L_in->nnz);
	D_in->v.z = (double *)malloc(sizeof(double) * 2 * D_in->nnz);
	U_in->v.z = (double *)malloc(sizeof(double) * 2 * U_in->nnz);
	break;
      default:
	should_not_reach_here;
      }
      L_in->index = (int *)malloc(sizeof(int) * L_in->nnz);
      D_in->index = (int *)malloc(sizeof(int) * D_in->nnz);
      U_in->index = (int *)malloc(sizeof(int) * U_in->nnz);
    }

#define _split_1(pre) { \
      size_t L_k, D_k, U_k, j;					\
								\
      L_k = L_in->row[i];					\
      D_k = D_in->row[i];					\
      U_k = U_in->row[i];					\
      for (j = in->row[i]; j < in->row[i+1]; j++) {		\
	if (i > in->index[j]) {					\
	  L_in->v.pre[L_k] = in->v.pre[j];			\
	  L_in->index[L_k] = in->index[j];			\
	  L_k++;						\
	} else if (i == in->index[j]) {				\
	  D_in->v.pre[D_k] = in->v.pre[j];			\
	  D_in->index[D_k] = in->index[j];			\
	  D_k++;						\
	} else {						\
	  U_in->v.pre[U_k] = in->v.pre[j];			\
	  U_in->index[U_k] = in->index[j];			\
	  U_k++;						\
	}							\
      }								\
    }

#define _split_2(pre) { \
      size_t L_k, D_k, U_k, j;					\
								\
      L_k = L_in->row[i];					\
      D_k = D_in->row[i];					\
      U_k = U_in->row[i];					\
      for (j = in->row[i]; j < in->row[i+1]; j++) {		\
	if (i > in->index[j]) {					\
	  L_in->v.pre[L_k*2]   = in->v.pre[j*2];		\
	  L_in->v.pre[L_k*2+1] = in->v.pre[j*2+1];		\
	  L_in->index[L_k]     = in->index[j];			\
	  L_k++;						\
	} else if (i == in->index[j]) {				\
	  D_in->v.pre[D_k*2]   = in->v.pre[j*2];		\
	  D_in->v.pre[D_k*2+1] = in->v.pre[j*2+1];		\
	  D_in->index[D_k]     = in->index[j];			\
	  D_k++;						\
	} else {						\
	  U_in->v.pre[U_k*2]   = in->v.pre[j*2];		\
	  U_in->v.pre[U_k*2+1] = in->v.pre[j*2+1];		\
	  U_in->index[U_k]     = in->index[j];			\
	  U_k++;						\
	}							\
      }								\
    }

    switch (object->precision) {
    case SILC_INT: 
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_split_1(i);
      break;
    case SILC_LONG:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_split_1(l);
      break;
    case SILC_FLOAT:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_split_1(s);
      break;
    case SILC_DOUBLE:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_split_1(d);
      break;
    case SILC_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_split_2(c);
      break;
    case SILC_DOUBLE_COMPLEX:
#pragma omp for
      for (i = 0; i < (ssize_t)object->d.matrix.m; i++)
	_split_2(z);
      break;
    default:
      should_not_reach_here;
    }
  }

  param[1].o = silc_object_new_matrix(object->precision,
				      object->format,
				      object->d.matrix.m,
				      object->d.matrix.n);
  param[1].o->d.matrix.in = (void *)L_in;
  param[2].o = silc_object_new_matrix(object->precision,
				      object->format,
				      object->d.matrix.m,
				      object->d.matrix.n);
  param[2].o->d.matrix.in = (void *)D_in;
  param[3].o = silc_object_new_matrix(object->precision,
				      object->format,
				      object->d.matrix.m,
				      object->d.matrix.n);
  param[3].o->d.matrix.in = (void *)U_in;
  return 0;
}

#define SILC_ANY_INTEGER   (SILC_INT|SILC_LONG)
#define SILC_ANY_REAL      (SILC_FLOAT|SILC_DOUBLE)
#define SILC_ANY_COMPLEX   (SILC_COMPLEX|SILC_DOUBLE_COMPLEX)
#define SILC_ANY_PRECISION (SILC_ANY_INTEGER|SILC_ANY_REAL|SILC_ANY_COMPLEX)

static unary_op_table_t _unary_op_table[] = {
  /* name, op, precision, type, format, out, flags */
  {silc_crs_negate, "silc_crs_negate",
   SILC_NEGATE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   0, 0},
  {silc_crs_complex_conjugate, "silc_crs_complex_conjugate",
   SILC_COMPLEX_CONJUGATE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   0, 0},
  {silc_crs_transpose, "silc_crs_transpose",
   SILC_TRANSPOSE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   1, 0},
  {silc_crs_conjugate_transpose, "silc_crs_conjugate_transpose",
   SILC_CONJUGATE_TRANSPOSE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   1, 0},
  {NULL} /* sentinel */
};

static binary_op_table_t _binary_op_table[] = {
  /* name, op, precision, x_type, x_format, y_type, y_format, out, flags */
  {silc_crs_mul_matrix_vector, "silc_crs_mul_matrix_vector",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_COLUMN_VECTOR_TYPE, 0,
   2, 0},
  {silc_crs_mul_vector_matrix, "silc_crs_mul_vector_matrix",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_ROW_VECTOR_TYPE, 0, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   2, 0},
  {silc_crs_mul_matrix_scalar, "silc_crs_mul_matrix_scalar",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_SCALAR_TYPE, 0,
   0, 0},
  {silc_crs_mul_scalar_matrix, "silc_crs_mul_scalar_matrix",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_SCALAR_TYPE, 0, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   1, 0},
  {silc_crs_add_matrix_matrix, "silc_crs_add_matrix_matrix",
   SILC_ADD_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   2, 0},
  {silc_crs_sub_matrix_matrix, "silc_crs_sub_matrix_matrix",
   SILC_SUBTRACT_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   2, 0},
  {silc_crs_mul_matrix_matrix, "silc_crs_mul_matrix_matrix",
   SILC_MULTIPLY_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   2, 0},
  {silc_crs_mul_matrix_elementwise, "silc_crs_mul_matrix_elementwise",
   SILC_MULTIPLY_ELEMENTWISE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   2, 0},
  {silc_crs_div_matrix_elementwise, "silc_crs_div_matrix_elementwise",
   SILC_DIVIDE_ELEMENTWISE_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   2, 0},

  /* addition of dense and sparse matrices */
  {silc_crs_add_dense_matrix, "silc_crs_add_dense_matrix",
   SILC_ADD_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_DENSE, SILC_MATRIX_TYPE, SILC_FORMAT_CRS,
   0, 0},
  {silc_crs_add_dense_matrix, "silc_crs_add_dense_matrix",
   SILC_ADD_OP, SILC_ANY_PRECISION,
   SILC_MATRIX_TYPE, SILC_FORMAT_CRS, SILC_MATRIX_TYPE, SILC_FORMAT_DENSE,
   1, 0},

  {NULL} /* sentinel */
};

static function_table_t _function_table[] = {
  /* name, sig_text, flags */
  {silc_crs_func_zeros_1, "silc_crs_func_zeros_1",
   "matrix zeros(scalar)", 0},
  {silc_crs_func_zeros_2, "silc_crs_func_zeros_2",
   "matrix zeros(scalar, scalar)", 0},
  {silc_crs_func_eye, "silc_crs_func_eye",
   "matrix eye(scalar)", 0},
  {silc_crs_func_fliplr, "silc_crs_func_fliplr",
   "matrix fliplr(matrix)", 0},
  {silc_crs_func_flipud, "silc_crs_func_flipud",
   "matrix flipud(matrix)", 0},
  {silc_crs_func_diag_1, "silc_crs_func_diag_1",
   "matrix diag(vector)", 0},
  {silc_crs_func_diag_2, "silc_crs_func_diag_2",
   "matrix diag(vector, scalar)", 0},
  {silc_crs_func_diag_2, "silc_crs_func_diag_2",
   "matrix diag(scalar, scalar)", 0},
  {silc_crs_func_diagvec, "silc_crs_func_diagvec",
   "vector diagvec(matrix)", 0},
  {silc_crs_func_sparse, "silc_crs_func_sparse",
   "matrix sparse(vector, vector, vector, scalar, scalar)", 0},
  {silc_crs_func_sparse, "silc_crs_func_sparse",
   "matrix sparse(vector, vector, scalar, scalar, scalar)", 0},
  {silc_crs_func_nnz, "silc_crs_func_nnz",
   "scalar nnz(matrix)", 0},
  {NULL} /* sentinel */
};

static procedure_table_t _procedure_table[] = {
  /* name, sig_text, flags */
  {silc_crs_proc_split, "silc_crs_proc_split",
   "split(matrix in, matrix out, matrix out, matrix out)", 0},
  {NULL} /* sentinel */
};

void
init_sparse_crs_op(char *module_name)
{
  silc_register_module(module_name,
		       _unary_op_table,
		       _binary_op_table,
		       _function_table,
		       _procedure_table);
}
